/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, romanick@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: cccp_deliver.c,v 1.9 2001/10/09 20:22:43 idr Exp $"

/*
 *   COMPONENT_NAME: LIB
 *
 *   FUNCTIONS:
 */

/* PROJECT:	Cluster Lock Manager communications facility*/
/**/
/* MODULE:	protocol-specific grunge*/

#if defined( __KERNEL__ )
# ifndef MODULE
#  define MODULE
# endif

# include <linux/kernel.h>
# define __NO_VERSION__
# include <linux/module.h>
# ifdef USING_FULL_VERSIONING
#  if CONFIG_MODVERSIONS==1
#   define MODVERSIONS
#   include <linux/modversions.h>
#  endif
# endif /* USING_FULL_VERSIONING */

#else

# include <stdio.h>
# include <errno.h>
# include <string.h>
# include <unistd.h>
# include <sys/param.h>
# include <sys/poll.h>
# include <sys/uio.h>

#endif /* __KERNEL__ */

#include <cltrc.h>
#include <dlm_cluster.h>
#include <cccp.h>

#include "cccp_pkt.h"
#include "cccp_queue.h"
#include "cccp_proto.h"
#include "cccp_private.h"

/* Semaphore used to let the delivery thread know that there maybe another
 * message available that needs to be processed.
 */

cccp_sem_t   cccp_delivery_sem;




/****************************************************************************/
/* localForward - Pass a message to that callback function for a port
 *
 * RETURNS
 * 
 * If message delivered successfully TRUE is returned.  Otherwise, FALSE is
 * returned.
 */

static bool_t
local_forward(
	      /* Message to be delivered.
	       */

	      clm_msg_t * msg,


	      /* Node ID of the sender of the message.
	       */

	      unsigned sender 
	    )
{
    int		(*recvcb)(clm_msg_t *);
    uint        self;
    bool_t   msg_done;


    self = thisNode();

    /* Loop as long as the local node is the destination node of the message.
     */

    msg_done = FALSE;
    while ( !msg_done && (msg->cm_node == self) ) {
#ifdef MASSIVE_DEBUG
	bsdlog( LOG_DEBUG, "[%s:%s] in msg delivery loop\n",
	        PROCESS_NAME, __FUNCTION__ );
#endif

	/* cm_node is over-loaded.  In some places it is the receiving node,
	 * in others it is the sending node.  Here, we switch.  On input it
	 * is the receiving node (which must be the local node ID), but now
	 * we want to set it to be the sending node ID (so that the client
	 * can determine who sent the message).
	 */

	msg->cm_node = sender;


	recvcb = getCB(msg->cm_port);
	if ( recvcb != NULL )
	{
	    TRCHKL3T(CPTRC_RCVCB_INVOKE,msg,msg->cm_port,recvcb);

	    if ( (*recvcb)( msg ) == 0 )
	    {
		TRCHKL1T(CPTRC_RCVCB_RETURN,msg);
		msg_done = TRUE;
	    }
	    else
	    {
		msg->cm_hops++;
		TRCHKL2T(CPTRC_FORWARDED,msg->cm_node,msg->cm_port);
	    }
	}
	else
	{
	    msg_done = TRUE;
	    bsdlog( LOG_WARNING, "[%s:%s] NULL receive callback for port %u",
		    PROCESS_NAME, __FUNCTION__, msg->cm_port );
	}
    }

    return( msg_done );
}




/****************************************************************************/

bool_t
cccp_deliver_all_from_port( unsigned port_index )
{
    clm_msg_t   * msg_to_deliver;
    clm_pkt_t   * pkt;
    unsigned      sender;
    cccp_forward_callback_t * action[2];
    bool_t        got_a_message;


    got_a_message = FALSE;

    while ( (msg_to_deliver = cccp_port_dequeue_message( port_index )) 
	    != NULL )
    {
	got_a_message = TRUE;

	sender = msg_to_deliver->cm_sender;
	pkt = msg_to_deliver->cm_packet;

	action[0] = msg_to_deliver->cm_action[0];
	action[1] = msg_to_deliver->cm_action[1];

	bsdlog( LOG_DEBUG, "[%s:%s] preparing to process message @ "
	        "0x%08lx [0x%08lx, 0x%08lx]\n",
	        PROCESS_NAME, __FUNCTION__,
	        (long) msg_to_deliver, (long) action[0], (long) action[1] );

	if ( local_forward( msg_to_deliver, sender ) )
	{
	    bsdlog( LOG_DEBUG, "[%s:%s] message being retired\n",
		    PROCESS_NAME, __FUNCTION__ );

	    finishMessage( msg_to_deliver, CPERR_OK );
	    msg_to_deliver = NULL;

	    if ( action[0] != NULL )
	    {
		(*action[0])( NULL, sender, pkt );
	    }

	    if ( sender != thisNode() )
	    {
		ASSERT( pkt != NULL, "NULL packet pointer" );
		CLM_FREE( pkt );
	    }
	}
	else
	{
	    bsdlog( LOG_DEBUG, "[%s:%s] redirect/remote send requested\n",
		    PROCESS_NAME, __FUNCTION__ );

	    if ( action[1] != NULL )
	    {
		(*action[1])( msg_to_deliver, sender, pkt );
	    }
	}
    }


    return( got_a_message );
}


/****************************************************************************/

static unsigned cccp_next_port_to_deliver = 0;

void *
cccp_msg_delivery_loop( void * arg )
{
    unsigned   port;
    unsigned   count;
    bool_t     got_a_message;


#ifdef __KERNEL__
    dlm_set_thread_name( current, "CCCP delivery" );
    MOD_INC_USE_COUNT;
#endif

    while ( ! cccp_time_to_die )
    {
	CCCP_SEM_DOWN( & cccp_delivery_sem );

	if ( ! cccp_time_to_die )
	{
	    /* Deliver the messages */

	    for ( count = 0 ; count <= CCCP_MAX_PORT ; count++ )
	    {
		port = (cccp_next_port_to_deliver + count) & CCCP_MAX_PORT;
		if ( cccp_port_lock_delivery( port ) )
		{
		    got_a_message = cccp_deliver_all_from_port( port );
		    cccp_port_unlock_delivery( port );

		    if ( got_a_message )
		    {
			cccp_next_port_to_deliver = port;
			break;
		    }
		}
	    }
	}
    }


#ifdef __KERNEL__
    MOD_DEC_USE_COUNT;
#endif

    bsdlog( LOG_INFO, __FUNCTION__ " termination, CCCP shutdown." );
    return( NULL );
}




/****************************************************************************/
/* cccp_msg_deliver_local - Locally deliver a message
 *
 * The message will be queued for delivery to the callback function set for
 * the port specified in the message (i.e., msg->cm_port).  Upon delivery,
 * the callback can return either zero or non-zero.  If zero is returned,
 * then the message data will be sent to the 'not_forwarded' function.  If
 * non-zero is returned, the message data will be passed to the 'forwarded'
 * function.
 * 
 * If the message is NOT forwarded and the sender of the message is NOT the
 * local node, then the clm_pkt_t associated with the message
 * (i.e., msg->cm_packet) will automatically be freed after calling the
 * 'not_forwarded' function.
 */

void
cccp_msg_deliver_local(
		        /* Message to be delivered.
			 */

		        clm_msg_t * msg,


		        /* Function to execute if the client does not request
			 * a re-direct of the message.  This pointer may be
			 * NULL.
			 */

		        cccp_forward_callback_t * not_forwarded,


		        /* Function to execute if the client request a
			 * re-direct of the message.  This pointer may be
			 * NULL.
			 */

		        cccp_forward_callback_t * forwarded
		      )
{
    msg->cm_action[0] = not_forwarded;
    msg->cm_action[1] = forwarded;


    /* Queue the message and signal the delivery thread.
     */

    cccp_port_queue_message( msg );
    (void) CCCP_SEM_UP( & cccp_delivery_sem );
}




/****************************************************************************/
/* cccp_msg_queue_ack_and_free_pkt - Queues an ack for the message, frees pkt
 *
 * This function is used when a redirect is NOT requested.
 */

void
cccp_msg_queue_ack_and_free_pkt( clm_msg_t * msg,
				 unsigned sender,
				 clm_pkt_t * pkt 
			       )
{
    cccp_node_t * node;


    LOG_PACKET( pkt );

    CCCP_MUTEX_LOCK( & cccp_global_lock );
    node = getNodeInfo( sender );
    queueAck( node, pkt->cp_channel );
    CCCP_MUTEX_UNLOCK( & cccp_global_lock );
}




/****************************************************************************/
/* cccp_msg_just_queue - Passes the msg/pkt to queuePacket
 *
 */

void
cccp_msg_just_queue(
		     clm_msg_t * msg,
		     unsigned sender,
		     clm_pkt_t * pkt 
		   )
{
    CCCP_MUTEX_LOCK( & cccp_global_lock );
    pkt->cp_port = msg->cm_port;
    pkt->cp_hops = msg->cm_hops;
    (void) queuePacket( msg, pkt, msg->cm_node );
    CCCP_MUTEX_UNLOCK( & cccp_global_lock );
}




/****************************************************************************/
/* cccp_msg_queue_redirect - Queues the message as a re-direct
 *
 */

void
cccp_msg_queue_redirect(
			 clm_msg_t * msg,
			 unsigned sender,
			 clm_pkt_t * pkt 
		       )
{
    cccp_node_t * node;
    
    
    CCCP_MUTEX_LOCK( & cccp_global_lock );
    node = getNodeInfo( sender );
    queueRedirect( msg, pkt, node, sender );
    CCCP_MUTEX_UNLOCK( & cccp_global_lock );
}




/****************************************************************************/
/* cccp_msg_do_remote_send - Sends the message to a remote node
 * 
 */

void
cccp_msg_do_remote_send(
			 clm_msg_t * msg,
			 unsigned sender,
			 clm_pkt_t * pkt 
		       )
{
    int   err;


    /* allocate a packet to go with the message*/
    pkt = cccp_msg_marshal( msg );
    msg->cm_packet = pkt;
    msg->cm_sender = thisNode();

    
    err = CPERR_NOMEM;
    if ( pkt != NULL )
    {
	bool_t   packet_sent_successfully;


	/* Try to send the packet.  If the packet can't be queued, destroy
	 * message and notify the message's originator via the message's
	 * callback function.
	 */

	CCCP_MUTEX_LOCK( & cccp_global_lock );
	packet_sent_successfully = queuePacket( msg, pkt, msg->cm_node );
	CCCP_MUTEX_UNLOCK( & cccp_global_lock );

	err = ( packet_sent_successfully ) ? 0 : CPERR_QPKT;
    }
    

    if ( err != 0 )
    {
	finishMessage( msg, err );
	if ( pkt != NULL )
	{
	    CLM_FREE(pkt);
	}
    }
}
