/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, romanick@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: cccp_init.c,v 1.16 2001/11/07 23:56:18 idr Exp $"

/*
 *   COMPONENT_NAME: LIB
 *
 *   FUNCTIONS: cccp_dumpinfo
 *		cccp_init
 *		cccp_stop
 */

/* PROJECT:	Cluster Lock Manager communications facility*/
/**/
/* MODULE:	initialization functions*/

#if defined( __KERNEL__ )
# ifndef MODULE
#  define MODULE
# endif
# include <linux/kernel.h>
# include <linux/module.h>
# ifdef USING_FULL_VERSIONING
#  if CONFIG_MODVERSIONS==1
#   define MODVERSIONS
#   include <linux/modversions.h>
#  endif
# endif /* USING_FULL_VERSIONING */

# include <linux/init.h>
# include <linux/errno.h>
# include <linux/socket.h>
# include <linux/proc_fs.h> 
#else
# include <errno.h>
# include <sys/socket.h>
# include <signal.h>
# include <unistd.h>
#endif /* __KERNEL__ */

#include <cltrc.h>
#include <dlm_cluster.h>
#include <cccp.h>

#include "cccp_pkt.h"
#include "cccp_proto.h"
#include "cccp_private.h"

#include <iterator.h>

/* our node ID*/
int cccp_our_id = (-1);


/* socket for accepting connections (TCP) or reading packets (UDP)*/

cccp_sock_t cccp_our_sock	= CCCP_SOCK_INVALID;

DECLARE_MUTEX( cccp_global_lock );

cccp_pid_t     cccp_poll_thread;
cccp_pid_t     cccp_retransmit_thread;
cccp_pid_t     cccp_msg_delivery_thread;

volatile bool_t cccp_time_to_die = FALSE;

void cccp_minimal_stop( void );

#ifdef __KERNEL__

static const char * cccp_proc_name = "cccp";

# if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)

static int cccp_dumpinfo22( char * buf, char ** start, off_t offset,
    int len, int unused );

static struct proc_dir_entry cccp_proc_entry = {
    0,                                    /* low inode, dynamic */
    0,                                    /* length of name */
    NULL,                                 /* name string */
    S_IFREG | S_IRUGO,                    /* mode */
    1, 0, 0,                              /* nlinks, owner, group */
    0,                                    /* size, unused */
    NULL,                                 /* default operations */
    & cccp_dumpinfo22                     /* function to read data */
};

# else

static struct proc_dir_entry *cccp_proc_entry = NULL;
static int cccp_read_proc( char * page, char **start, off_t off, int count,
    int *eof, void *data );

# endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) */

extern kmem_cache_t * cccp_timer_cache;
extern kmem_cache_t * cccp_msg_cache;

#endif /* __KERNEL__ */


/****************************************************************************/
/* cccp_minimal_init - Perform basic initialization of CCCP
 * 
 */

int
cccp_minimal_init( void )
{
    bsdlog( LOG_INFO, __FUNCTION__ " started" );
    cccp_our_id = -1 ;
    initPorts() ;
    initProto() ;

#ifdef __KERNEL__

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)

    cccp_proc_entry = create_proc_read_entry( cccp_proc_name, 0, NULL,
					      cccp_read_proc, NULL );

#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) */

    cccp_proc_entry.name = cccp_proc_name;
    cccp_proc_entry.namelen = strlen( cccp_proc_name );
    proc_register( & proc_root, & cccp_proc_entry );

#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) */

    cccp_timer_cache = kmem_cache_create( "CCCP timer", 
					  sizeof( struct timer_list ),
					  0, 0, NULL, NULL );
    cccp_msg_cache = kmem_cache_create( "CCCP message",
				        sizeof( clm_msg_t ), 
				        0, 0, NULL, NULL );

    if ( (cccp_timer_cache == NULL) || (cccp_timer_cache == NULL) )
    {
	bsdlog( LOG_CRIT, "[%s:%s] Could not create caches!\n",
	        PROCESS_NAME, __FUNCTION__ );

	cccp_minimal_stop();
	return( -1 );
    }
#endif /* __KERNEL__ */


    return( 0 );
}




/****************************************************************************/
/* cccp_init - Initialize the CCCP subsystem
 *
 * Initializes the communications module.  In the kernel version, it's
 * responsible for starting the kproc.
 *
 * RETURNS
 *
 * If initialization successful CPERR_OK is returned, error code otherwise.
 */

int
cccp_init( void )
{
    iterator_t * iter;
    int	  status;
    int   i;


    TRCHKL0T(CPTRC_UINIT_ENTRY);
    CCCP_MUTEX_LOCK( & cccp_global_lock );

    bsdlog( LOG_INFO, "[%s:%s]\n",
           PROCESS_NAME, __FUNCTION__);

#if ! defined( __KERNEL__ )
    (void) cccp_minimal_init();
    read_node_cfg_file();


    /* This is only here to work around the strange way that the heartbeat
     * interface library works.
     */
    
    (void) clm_node_reachable( 1 );
#endif

    status = clm_getlocalid( & cccp_our_id );

    bsdlog( LOG_DEBUG, "cccp_our_id = %d, status = %d\n", cccp_our_id,
	    status);

    ASSERT((status == DLM_NORMAL),"doTopology failed in init");


    status = cccp_ip_init( cccp_our_id );

    bsdlog( LOG_DEBUG, "cccp_ip_init status = %d\n", status );

    ASSERT( status == CPERR_OK, "failed to get socket for IP between "
	    "cllockds" );

    bsdlog(LOG_DEBUG, "[%s:%s] cccp_our_id [%d] \n",
           PROCESS_NAME, __FUNCTION__,
           cccp_our_id);

    iter = dlm_nodeid_iterator_create();
    while ( ! iterator_next( iter, & i ) )
    {
	if ( i != cccp_our_id )
	{
	    (void) cccpConn( i );
	}
    }

    iterator_destroy( iter );
    

    /* Make any necessary new connections.
     */

    openConnections();

    CCCP_SEM_INIT( & cccp_poll_sem );
    CCCP_SEM_INIT( & cccp_delivery_sem );

    CCCP_NEW_THREAD( cccp_poll_thread, NULL, cccp_poll_loop, NULL );
    CCCP_NEW_THREAD( cccp_retransmit_thread, NULL, cccp_retransmit_loop,
		     NULL );
    CCCP_NEW_THREAD( cccp_msg_delivery_thread, NULL, cccp_msg_delivery_loop,
		     NULL );

    CCCP_MUTEX_UNLOCK( & cccp_global_lock );

    TRCHKL2T(CPTRC_UINIT_EXIT,status,cccp_our_id);
    return( status );
}




/****************************************************************************/
/* cccp_minimal_stop - Perform basic, final shutdown of CCCP
 * 
 */

void
cccp_minimal_stop( void )
{
#ifdef __KERNEL__
#ifdef CONFIG_PROC_FS
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)

    if ( cccp_proc_entry != NULL )
    {
	remove_proc_entry( cccp_proc_entry->name, cccp_proc_entry->parent );
	cccp_proc_entry = NULL;
    }

#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) */

    proc_unregister( & proc_root, cccp_proc_entry.low_ino );

#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) */
#endif /* CONFIG_PROC_FS */

    if ( cccp_timer_cache != NULL )
    {
	kmem_cache_destroy( cccp_timer_cache );
	cccp_timer_cache = NULL;
    }

    if ( cccp_msg_cache != NULL )
    {
	kmem_cache_destroy( cccp_msg_cache );
	cccp_msg_cache = NULL;
    }
#endif /* __KERNEL__ */
}




/* FUNCTION:	cccp_stop*/
/**/
/* PURPOSE:	Shuts down the communications module.*/
/**/
/* INPUT:	none*/
/**/
/* OUTPUT:	none*/

void
cccp_stop ( void )
{
    int   err;
#if ! defined( __KERNEL__ )
    int   i;
#endif


    TRCHKL0T(CPTRC_USTOP_ENTRY);

    if ( checkInit() )
    {
	CCCP_MUTEX_LOCK( & cccp_global_lock );

	cccp_time_to_die = TRUE;
	cccp_our_id = (-1);

	/* This lets the re-transmit loop thread and the delivery loop thread
	 * know that they need to stop running.
	 */

	CCCP_SEM_UP( & cccp_poll_sem );
	CCCP_SEM_UP( & cccp_delivery_sem );

	err = kill_proc( cccp_poll_thread, SIGINT, 1 );
	if ( err < 0 )
	{
	    bsdlog( LOG_CRIT, "[%s:%s] Unable to kill CCCP poll thread "
		    "(pid=%u), error = %u\n", PROCESS_NAME, __FUNCTION__,
		    cccp_poll_thread, -err );
	}


#if ! defined( __KERNEL__ )
	/* Shut down the protocol subsystem, pretending that all nodes went
	 * down at once.
	 */

	{
	    iterator_t * iter;


	    iter = dlm_nodeid_iterator_create();
	    while ( ! iterator_next( iter, & i ) )
	    {
		cccp_node_down( i, FALSE );
	    }

	    iterator_destroy( iter );
	}

#endif /* __KERNEL__ */

	CCCP_MUTEX_UNLOCK( & cccp_global_lock );
    }


    TRCHKL0T(CPTRC_USTOP_EXIT);
}



/****************************************************************************/
/* cccp_dumpinfo - Dumps comm layer info to buffer for /proc filesystem
 *
 */

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)) || !defined( __KERNEL__ )

int 
cccp_dumpinfo23( char * buf, unsigned len )

#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) */

static int
cccp_dumpinfo22(
		 char * buf,
		 char ** start,
		 off_t offset,
		 int len,
		 int unused
	       )

#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) */
{
    static const char  header[] =
	"node id: %d\n"
	"{session:retries:queue size:acks rcvd}\n";

    char *p = buf;
    unsigned   bytes;


    bytes = snprintf( p, len, header, cccp_our_id );
    bytes += cccp_node_stats( p + bytes, len - bytes );
    bytes += cccp_node_colors( p + bytes, len - bytes );
    bytes += cccp_timeout_status( p + bytes, len - bytes );
    return( bytes );
}



#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)

static int 
cccp_read_proc( char * page, char **start, off_t off,
	        int count, int *eof, void *data )
{
    int len;

    len = cccp_dumpinfo23( page + off, count - off );
    if ( len <= (off + count) )
	*eof = 1;

    *start = page + off;
    len -= off;

    if ( len > count )
	len = count;

    if ( len < 0 )
	len = 0;

    return len;
}

#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) */

/****************************************************************************/

#ifndef module_init
# define module_init( func ) int init_module( void ) { return( func () ); }
# define module_exit( func ) void cleanup_module( void ) { func (); }
#endif

module_init( cccp_minimal_init )
module_exit( cccp_minimal_stop )
