/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: cccp_input.c,v 1.9 2001/11/07 23:56:18 idr Exp $"

/* PROJECT:	Cluster Lock Manager communications facility*/
/**/
/* MODULE:	incoming packet handler*/

#if defined( __KERNEL__ )
# ifndef MODULE
#  define MODULE
# endif

# include <linux/kernel.h>
# define __NO_VERSION__
# include <linux/module.h>
# ifdef USING_FULL_VERSIONING
#  if CONFIG_MODVERSIONS==1
#   define MODVERSIONS
#   include <linux/modversions.h>
#  endif
# endif /* USING_FULL_VERSIONING */
#endif /* __KERNEL__ */

#include <cltrc.h>
#include <dlm_cluster.h>
#include <cccp.h>

#include "cccp_pkt.h"
#include "cccp_proto.h"
#include "cccp_private.h"



	    
/****************************************************************************/
/* handlePacket - Read a packet from a link and run it through the protocol
 *
 * This routine is responsible for freeing the msg and pkt parameters when
 * TRUE is returned (caller is responsible when FALSE is returned).  This is
 * done so that the packet can be recycled in case of a redirect.  The
 * message body is also freed if it exists.
 * 
 * RETURNS
 *
 * If packet read and processed TRUE is returned.  Otherwise, FALSE is
 * returned (error).
 */

bool_t
handlePacket (
	       /* Sender's HACMP node ID.
		*/

	       uint node,


	       /* Pointer to an uninitialized message.
		*/

	       clm_msg_t *msg,
	      

	       /* Pointer to a clm_pkt_t.
		*/

	       clm_pkt_t *pkt
	     )
{
    cccp_node_t * ni;
    const char  * text_msg;
    char          msg_buf[32];
    bool_t        status;


    LOG_PACKET( pkt );

    ASSERT_GLOBAL_LOCKED();

    if ( pkt->cp_version != CCCP_VERSION )
    {
	bsdlog( LOG_WARNING, "[%s:%s] packet with invalid version (%d) "
	        "received from node %u\n",
	        PROCESS_NAME, __FUNCTION__,
	        pkt->cp_version, pkt->cp_srcnode );

	CLM_FREE( pkt );
	cccp_msg_destroy( msg );
	
	return( FALSE );
    }


    status = TRUE;

    /* get the node_info_t for the sender*/
    ni = cccpConn( node );

    if ( ni == NULL ) 
    {
	text_msg = "couldn't get node info";
	status = FALSE;
    }
    else if ( checkSession( ni, pkt ) )
    {
	/* The packet is from a dropped session. 
	 */

	text_msg = "msg from old session";
    }
    else if ( (pkt->cp_flags & CPF_NACK_VALID) != 0 )
    {
	/* handle NACKs first */

	nackRetrans(ni, pkt->cp_channel, pkt->cp_hops, pkt->cp_acks);
	text_msg = "NACK processed";
    }
    else if ( checkDups( ni, pkt->cp_flags, pkt->cp_channel ) ) 
    {
	/* The message is a duplicate.  Just drop it.
	 */

	TRCHKL1T(CPTRC_DUPLICATE,pkt);

	snprintf( msg_buf, sizeof( msg_buf ), "dup dropped, slot = %u", 
		  pkt->cp_channel );
	text_msg = msg_buf;
    }
    else
    {
	/* handle acks*/
	handleAcks(ni,pkt->cp_acks);


	text_msg = "no msg/redir";
	if ( (pkt->cp_flags & (CPF_REDIR_VALID | CPF_MSG_PRESENT)) != 0 )
	{
	    /* make sure we don't have a redirect queued for this already*/
	    if ( checkRedirs( ni, pkt->cp_channel ) ) 
	    {
#if defined( SHOW_ORDERING )
		bsdlog( LOG_DEBUG, "dropping packet frm %d - already "
		        "redirected slot=%d\n",
		       pkt->cp_srcnode, pkt->cp_channel);
#endif

		text_msg = "already redirected";
	    }
	    else if ( checkOrder( ni, pkt, msg ) )
	    {
#if defined( SHOW_ORDERING )
		bsdlog( LOG_DEBUG, "msg out of order from %u on channel %u\n",
		        pkt->cp_srcnode, pkt->cp_channel );
#endif

		/* checkOrder returns TRUE if the message is being processed
		 * in the wrong order.  In this case it will also put the
		 * message in the proper queue so that it can be delivered
		 * later.
		 */

		/* This path MUST NOT free the msg/pkt.  checkOrder will
		 * either queue the msg/pkt (in inmsgs/inpkts) or free them.
		 */

		msg = NULL;
		pkt = NULL;

		text_msg = "msg out of order";
	    }
	    else if ( (pkt->cp_flags & CPF_REDIR_VALID) != 0 ) 
	    {
		/* Handle the redirect.
		 */

		queueAck( ni, pkt->cp_channel );
		handleRedirect( ni, & pkt->cp_redirect );

		text_msg = "redirect processed";
	    }
	    else
	    {
		/* Pass the message payload off to the appropriate user
		 * call-back.
		 */

		msg->cm_packet = pkt;
		msg->cm_sender = node;

		bsdlog( LOG_DEBUG, "[%s:%s] delivering message\n",
		        PROCESS_NAME, __FUNCTION__ );

		CCCP_MUTEX_UNLOCK( & cccp_global_lock );
		cccp_msg_deliver_local( msg, 
		  (cccp_forward_callback_t *) cccp_msg_queue_ack_and_free_pkt,
	          (cccp_forward_callback_t *) cccp_msg_queue_redirect );
		CCCP_MUTEX_LOCK( & cccp_global_lock );

		msg = NULL;
		pkt = NULL;

		text_msg = "msg delivered";
	    }
	}

	reOrder( ni, node );
    }

	
    if ( pkt != NULL )
    {
	cccp_msg_destroy(msg);
	CLM_FREE(pkt);
    }


    bsdlog( LOG_DEBUG, "[%s:%s] leaving (%s,%s)\n",
	    PROCESS_NAME, __FUNCTION__,
	    status ? "true" : "false", text_msg );

    return( status );
}
