/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, romanick@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: cccp_kernel.c,v 1.9 2001/11/07 23:56:18 idr Exp $"

/*
 *   COMPONENT_NAME: LIB
 *
 *   FUNCTIONS:
 *		cccp_time
 *		random
 *		cccp_sock_write
 */

/* PROJECT:	Cluster Lock Manager communications facility*/
/**/
/* MODULE:	Kernel equivalents for various user-level routines*/

#if defined( __KERNEL__ )
# ifndef MODULE
#  define MODULE
# endif

# include <linux/kernel.h>
# define __NO_VERSION__
# include <linux/module.h>
# ifdef USING_FULL_VERSIONING
#  if CONFIG_MODVERSIONS==1
#   define MODVERSIONS
#   include <linux/modversions.h>
#  endif
# endif /* USING_FULL_VERSIONING */

# include <linux/stddef.h>
# include <linux/timer.h>
#else

# include <errno.h>
# include <signal.h>
# include <stdarg.h>

# include <sys/types.h>
# include <sys/param.h>
# include <sys/time.h>
# include <sys/uio.h>

#endif /* __KERNEL__ */



#include <cltrc.h>
#include <dlm_cluster.h>
#include <cccp.h>

#include "cccp_private.h"

#if defined( __KERNEL__ ) && defined( STRESS )

long
random ( void )
{
    return( (long) cccp_time() );
}

#endif /* __KERNEL__ */

int
cccp_sock_write( cccp_sock_t sock, void * data, size_t bytes )
{
    struct iovec	iov;
    struct msghdr       msg_header;

    
    iov.iov_base = data;
    iov.iov_len  = bytes;

    msg_header.msg_name = NULL;
    msg_header.msg_namelen = 0;
    msg_header.msg_iov = & iov;
    msg_header.msg_iovlen = 1;
    msg_header.msg_control = NULL;
    msg_header.msg_controllen = 0;
    msg_header.msg_flags = 0;

    return( CCCP_SOCK_SENDMSG( sock, & msg_header, bytes ) );
}




int
cccp_time ( void )
{
    struct timeval   tv;
    int              now;


#if defined( __KERNEL__ )
    do_gettimeofday( & tv );
#else
    gettimeofday( & tv, NULL );
#endif /* __KERNEL__ */


    now = (tv.tv_sec * CCCP_HZ) + (tv.tv_usec / (1000000 / CCCP_HZ));
    return( now );
}




#if defined( __KERNEL__ )
static long cccp_timer_expires = 0L;
kmem_cache_t * cccp_timer_cache = NULL;
static atomic_t timer_counter;
#endif

void
cccp_set_timer( unsigned ticks )
{
#if defined( __KERNEL__ )
    long   expires;
    long   now;
    struct timer_list * timer;
    
    
    now = jiffies;
    expires = now + ticks;

    timer = kmem_cache_alloc( cccp_timer_cache, 0 );
    init_timer( timer );

    timer->function = (void *) cccp_poll_timer_handler;
    timer->data = (ulong) timer;
    timer->expires = expires;

    add_timer( timer );
    atomic_inc( & timer_counter );

    if ( (cccp_timer_expires < now) || (cccp_timer_expires > expires) )
    {
	cccp_timer_expires = expires;
    }

#else
    struct sigaction   action;
    struct itimerval   tick_tock;
    unsigned           curr;


    if ( ticks <= 10 )
    {
	cccp_poll_timer_handler( SIGALRM );
    }
    else
    {
	getitimer( ITIMER_REAL, & tick_tock );

	curr = (tick_tock.it_value.tv_sec * 1000 * 1000) 
	    + tick_tock.it_value.tv_usec;

	if ( (curr == 0) || (curr > (ticks * 1000)) )
	{
	    sigemptyset( & action.sa_mask );
	    sigaddset( & action.sa_mask, SIGALRM );

	    action.sa_handler = cccp_poll_timer_handler;
	    action.sa_flags = 0;

	    sigaction( SIGALRM, & action, NULL );

	    tick_tock.it_interval.tv_sec = 1000;
	    tick_tock.it_interval.tv_usec = 0;
	    tick_tock.it_value.tv_sec = 0;
	    tick_tock.it_value.tv_usec = ticks * 1000;

	    setitimer( ITIMER_REAL, & tick_tock, NULL );
	}
    }
#endif /* __KERNEL__ */
}




void
cccp_timer_release( ulong timer )
{
#if defined( __KERNEL__ )
    kmem_cache_free( cccp_timer_cache, (struct timer_list *) timer );
    atomic_dec( & timer_counter );
#endif
}




size_t
cccp_timeout_status( char * buf, unsigned limit )
{
    size_t   offset;
    long     val;

    val = cccp_timer_expires - jiffies;
    
    if ( val > 0 )
    {
	offset = snprintf( buf, limit, "Next timeout in %ld jiffies.\n", 
			   val );
    }
    else
    {
	offset = snprintf( buf, limit, "Timeout not set (%ld jiffies ago).\n",
			   -val );
    }

    offset += snprintf( & buf[ offset ], limit - offset,
		        "Timer queue has %u entries.\n\n",
		        atomic_read( & timer_counter ) );
    return( offset );
}
