/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, romanick@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: cccp_msg.c,v 1.7 2001/06/13 22:38:15 idr Exp $"

/*
 *   COMPONENT_NAME: LIB
 *
 *   FUNCTIONS: cccp_msg_free
 *              cccp_init_pkt
 *              cccp_msg_unmarshal
 *              cccp_msg_marshal
 */

#if defined( __KERNEL__ )
# ifndef MODULE
#  define MODULE
# endif

# include <linux/kernel.h>
# define __NO_VERSION__
# include <linux/module.h>
# ifdef USING_FULL_VERSIONING
#  if CONFIG_MODVERSIONS==1
#   define MODVERSIONS
#   include <linux/modversions.h>
#  endif
# endif /* USING_FULL_VERSIONING */

#else

# include <errno.h>
# include <unistd.h>

#endif

#include <cltrc.h>
#include <dlm_cluster.h>
#include <cccp.h>

#include "cccp_private.h"


kmem_cache_t * cccp_msg_cache = NULL;

/****************************************************************************/
/* cccp_msg_create - Allocate and initialize a clm_msg_t structure
 *
 * Allocates and initializes a clm_msg_t structure.  The supplied values are
 * filled into the fields of the clm_msg_t, and the "internally used" fields
 * are zeroed out.
 * 
 * RETURNS
 * 
 * A pointer to a new clm_msg_t object on success, or NULL on failure.
 * 
 * NOTES
 * 
 * In the future this function will be augmented witha cccp_msg_destroy
 * function that frees a message allocated by cccp_msg_create.  After that
 * is done, this function will be modified to use the Linux slab allocator.
 */

clm_msg_t *
cccp_msg_create( unsigned port, unsigned prio, unsigned node, unsigned type,
		 unsigned length, clm_msg_callback_t * cb, void * data,
		 void * private )
{
    clm_msg_t * msg;
    

    msg = kmem_cache_alloc( cccp_msg_cache, 0 );
    if ( msg != NULL )
    {
	(void) memset( msg, 0, sizeof( *msg ) );
	msg->cm_port = port;
	msg->cm_prio = prio;
	msg->cm_node = node;
	msg->cm_type = type;
	msg->cm_length = length;
	msg->cm_callback = cb;
	msg->cm_data = data;
	msg->cm_private = private;
    }

    return( msg );
}




/****************************************************************************/

void
cccp_msg_destroy( clm_msg_t * msg )
{
    kmem_cache_free( cccp_msg_cache, msg );
}




/****************************************************************************/
/* cccp_msg_duplicate - Make a copy of a message
 * 
 * Makes a copy of the given message so that the caller can modify the copy
 * to make a reply.  The linkage information, the callback pointer, and the
 * private data pointer will always be set to NULL in the new message.
 *
 * RETURNS
 *
 * If the copy is successful, a pointer to the new message is returned.
 * Otherwise, NULL is returned.
 */


clm_msg_t *
cccp_msg_duplicate(
		    /* Pointer to the original message.
		     */

		    const clm_msg_t * src 
		  )
{
    clm_msg_t * tmp;


    tmp = cccp_msg_create( src->cm_port, src->cm_prio, src->cm_node,
			   src->cm_type, src->cm_length, NULL,
			   NULL, NULL );
    if ( tmp != NULL ) 
    {
	if ( src->cm_length != 0 ) 
	{
	    tmp->cm_data = CLM_MALLOC(src->cm_length);
	    if ( tmp->cm_data == NULL ) 
	    {
		cccp_msg_destroy( tmp );
		tmp = NULL;
	    }
	    else
	    {
		(void) memcpy( tmp->cm_data, src->cm_data, src->cm_length );
		tmp->cm_sender = src->cm_sender;
	    }
	}
	else 
	{
	    SUGGEST((src->cm_data != NULL),"cm_data set with cm_length==0");
	}
    }

    return( tmp );
}




/****************************************************************************/

void
cccp_msg_free( clm_msg_t * msg, int status )
{
    ASSERT((status == CPERR_OK),"bad status in freeit");
    cccp_msg_destroy( msg );
}




/****************************************************************************/

void
cccp_init_pkt(
	       clm_pkt_t * pkt
	     )
{
    (void) memset( pkt, 0, sizeof( *pkt ) );

    pkt->cp_version = CCCP_VERSION;
    pkt->cp_header_length = sizeof( clm_pkt_t );
    pkt->cp_srcnode = thisNode();
}




/****************************************************************************/

clm_msg_t *
cccp_msg_unmarshal( 
		    clm_pkt_t * pkt
		  )
{
    clm_msg_t * msg;


    msg = cccp_msg_create( pkt->cp_port, pkt->cp_prio, thisNode(),
			   pkt->cp_type, pkt->cp_length, NULL,
			   NULL, NULL );
    if ( (msg != NULL)
	 && ( (pkt->cp_flags & CPF_MSG_PRESENT) && pkt->cp_length ) )
    {
	msg->cm_callback = cccp_msg_free;
	msg->cm_data = ((__u8 *) pkt) + pkt->cp_header_length;
    }

    return( msg );
}




/****************************************************************************/

clm_pkt_t *
cccp_msg_marshal(
		  const clm_msg_t * msg
		)
{
    clm_pkt_t * pkt;


    pkt = CLM_MALLOC( sizeof( clm_pkt_t ) );
    if ( pkt != NULL )
    {
	pkt->cp_flags   = CPF_MSG_PRESENT;
	pkt->cp_version = CCCP_VERSION;
	pkt->cp_port    = msg->cm_port;
	pkt->cp_prio    = msg->cm_prio;
	pkt->cp_type    = msg->cm_type;
	pkt->cp_length  = msg->cm_length;
	pkt->cp_hops    = msg->cm_hops;
	pkt->cp_node    = msg->cm_node;
	pkt->cp_srcnode = thisNode();
    }


    return( pkt );
}
