/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, romanick@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: cccp_ports.c,v 1.6 2001/03/27 19:44:25 idr Exp $"

/*
 *   COMPONENT_NAME: LIB
 *
 *   FUNCTIONS: cccp_listen
 *		getCB
 *		initPorts
 */

/*
 * PROJECT:	Cluster Lock Manager communications facility
 *
 * MODULE:	port information
 */

#if defined( __KERNEL__ )
# ifndef MODULE
#  define MODULE
# endif

# include <linux/kernel.h>
# define __NO_VERSION__
# include <linux/module.h>
# ifdef USING_FULL_VERSIONING
#  if CONFIG_MODVERSIONS==1
#   define MODVERSIONS
#   include <linux/modversions.h>
#  endif
# endif /* USING_FULL_VERSIONING */

#else

# include <stddef.h>

#endif

#include <cltrc.h>
#include <dlm_cluster.h>
#include <cccp.h>

#include "cccp_private.h"

struct cccp_port
{
    cccp_callback_t * callback;
    
    spinlock_t        queue_mutex;
    dlm_list_t        queue;
    
    cccp_mutex_t      delivery_mutex;
};

typedef struct cccp_port cccp_port_t;

static cccp_port_t port_table[ (CCCP_MAX_PORT + 1) ];
static int ports_inited = 0 ;

/*
 * FUNCTION:	initPorts
 *
 * PURPOSE:	Initialize the module-private data structures.
 *
 * INPUT:	none
 *
 * OUTPUT:	none
 */

void
initPorts ( void )
{
    unsigned   i;
    
    
    for ( i = 0 ; i < (CCCP_MAX_PORT + 1) ; i++ )
    {
	port_table[ i ].callback = NULL;
	port_table[ i ].queue_mutex = SPIN_LOCK_UNLOCKED;
	dlm_list_init( & port_table[ i ].queue );
	CCCP_MUTEX_INIT( & port_table[ i ].delivery_mutex );
    }

    ports_inited = 1 ;
}

/*
 * FUNCTION:	getCB
 *
 * PURPOSE:	Translate a port number to a callback function address.
 *
 * INPUT:	port number (1-255)
 *
 * OUTPUT:	receive callback function pointer
 */

cccp_callback_t * getCB( uint port )
{
    ASSERT( (port != 0) && (port <= CCCP_MAX_PORT), 
	    "getCB called for port zero" );
    return( port_table[port].callback );
}

/*
 * FUNCTION:	cccp_listen
 *
 * PURPOSE:	Register a receive callback function for a particular port
 *		(deregister by specifying NULL for function pointer).
 *
 * INPUT:	port		port number (1-255)
 *		cb		pointer to callback function
 *
 * OUTPUT:	CPERR_OK if no errors, CCCP error code otherwise
 */

int
cccp_listen( uint port, cccp_callback_t * cb )
{
    int    status;


    TRCHKL2T(CPTRC_LISTEN_ENTRY, port, cb) ;

    ASSERT((ports_inited), "cccp_listen called before ports init") ;
    ASSERT( !cb || (port_table[port].callback == NULL),
	    "duplicate receive callbacks" );

    status = CPERR_ILLEGAL;
    if (port <= CCCP_MAX_PORT )
    {
	if ( port != 0 )
	{
	    port_table[port].callback = cb ;
	    status = CPERR_OK;
	}
	NO_ELSE("cccp_listen called for port zero") ;
    }


    TRCHKL1T(CPTRC_LISTEN_EXIT, status) ;
    return( status );
}




/****************************************************************************/

void
cccp_port_queue_message(
			 /* Message that is to be queued.
			  */

			 clm_msg_t * msg
		       )
{
    cccp_port_t * port;
    
    port = & port_table[ msg->cm_port ];
    spin_lock( & port->queue_mutex );

    dlm_list_add_tail( & port->queue, (dlm_listnode_t *) msg );

    spin_unlock( & port->queue_mutex );
}




/****************************************************************************/

clm_msg_t *
cccp_port_dequeue_message(
			   /* Port from which a message is to be dequeued.
			    */

			   unsigned   port_index
			 )
{
    cccp_port_t * port;
    clm_msg_t   * msg;


    port = & port_table[ port_index ];
    spin_lock( & port->queue_mutex );

    msg = (clm_msg_t *) dlm_list_rem_head( & port->queue );

    spin_unlock( & port->queue_mutex );

    return( msg );
}




int
cccp_port_lock_delivery( unsigned port )
{
    return( CCCP_MUTEX_TRYLOCK( & port_table[ port ].delivery_mutex ) 
	     == CCCP_MUTEX_GOT_LOCK );
}

void
cccp_port_unlock_delivery( unsigned port )
{
    CCCP_MUTEX_UNLOCK( & port_table[ port ].delivery_mutex );
}
