/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */

#if !defined( CCCP_PRIVATE_H )
#define CCCP_PRIVATE_H 1

#ident "$Id: cccp_private.h,v 1.16 2001/11/07 23:56:19 idr Exp $"

#ifndef BOOL_T_DEFINED
#define BOOL_T_DEFINED
typedef int bool_t;
#endif /* BOOL_T_DEFINED */

#if defined( __KERNEL__ )
# include <linux/in.h>
# include <linux/inet.h>
# include <linux/net.h>
# include <linux/socket.h>
# include <asm/semaphore.h>
# include <asm/processor.h>
# include <asm/uaccess.h>
#else
# include <netinet/in.h>
# include <arpa/inet.h>
# include <sys/socket.h>
# include <pthread.h>
# include <errno.h>
#endif

#include <sys/syslog.h>

#include <cccp.h>
#include <dlm_cluster.h>
#include <dlm_kernel.h>
#include "cccp_pkt.h"

/****************************************************************************/

#define CCCP_PROTO	PF_INET, SOCK_DGRAM, IPPROTO_UDP
#define SUBCHANNELS	32


#if defined(DEBUG)

# if defined( __KERNEL__ )
#  define ABORT(m)    do { int * xx = 0; *xx = 0xdeadbeefUL; } while( 0 )
# else
#  define ABORT(m) \
        do { extern void stack_dump( void ); \
            stack_dump(); \
            exit( 1 ); \
        } while( 0 )
# endif

# ifdef ASSERT
#  undef ASSERT
# endif

# define as_str(a) as_str_b(a)
# define as_str_b(a) #a
# define ASSERT(cond,msg)					\
    if (cond) {							\
	;							\
    } else {							\
	bsdlog( LOG_CRIT, "ASSERT FAILED: %s, file %s, line %d\n", \
	   (msg), __FILE__, __LINE__ );				\
	ABORT("ASSERT: " __FILE__ ", line " as_str(__LINE__));	\
    }

# define NO_ELSE(msg)						\
    else {							\
	bsdlog(LOG_ERR,"WARNING: %s, file %s, line %d\n",	\
	   (msg), __FILE__, __LINE__);				\
    }

# define SUGGEST(cond,msg)					\
    if (cond) {							\
	;							\
    } NO_ELSE(msg);

#else /* DEBUG */

# ifdef ASSERT
#  undef ASSERT
# endif
# define ASSERT(cond,msg)
# define SUGGEST(cond,msg)
# define NO_ELSE(msg)

#endif /* DEBUG */


/* macros for manipulating bits*/
#define GETBIT(x,i)	(((x) >> (i)) & 1UL)
#define _MASK(i)	(1UL << (i))
#define _SETBIT(x,i)	((x) |=  _MASK(i))
#define _CLRBIT(x,i)	((x) &= ~_MASK(i))
#if defined(DEBUG)
#define ERR_SETBIT	"bit set twice"
#define ERR_CLRBIT	"bit cleared twice"
#define SETBIT(x,i)	{ ASSERT((GETBIT(x,i)==0),ERR_SETBIT); _SETBIT(x,i); }
#define CLRBIT(x,i)	{ ASSERT((GETBIT(x,i)!=0),ERR_CLRBIT); _CLRBIT(x,i); }
#else
#define SETBIT(x,i)	_SETBIT(x,i)
#define CLRBIT(x,i)	_CLRBIT(x,i)
#endif
#define FLIPBIT(x,i)	((x) ^= _MASK(i))



/****************************************************************************/
/* This is to make the various calls to inet_ntoa work correctly.
 */

#if ! defined( __KERNEL__ )
static inline const char * in_ntoa( __u32 addr )
{
    struct in_addr   in;

    in.s_addr = addr;
    return( inet_ntoa( in ) );
}

#endif

#if defined( __KERNEL__ )

typedef struct socket * cccp_sock_t;
#define CCCP_SOCK_INVALID   NULL
#define CCCP_SOCK_IS_VALID(s)  ((s) != NULL)

#define CCCP_SOCK_CREATE( sock, err ) \
	err = sock_create( CCCP_PROTO, sock )

#define CCCP_SOCK_BIND( sock, sin, size ) \
	(sock)->ops->bind( sock, sin, size )

#define CCCP_SOCK_CLOSE( sock ) \
	sock_release( sock )

#define CCCP_SOCK_CONNECT( sock, sin, size ) \
	(sock)->ops->connect( sock, sin, size, 0 )

static __inline__ int CCCP_SOCK_RECVMSG( cccp_sock_t sock, 
					 struct msghdr * sock_msg,
					 size_t buf_size,
					 int flags )
{
    mm_segment_t   oldfs;
    int            status;

    oldfs = get_fs(); set_fs( KERNEL_DS );
    status = sock_recvmsg( sock, sock_msg, buf_size, flags );
    set_fs( oldfs );

    return( status );
}

static __inline__ int CCCP_SOCK_SENDMSG( cccp_sock_t sock, 
					 struct msghdr * sock_msg,
					 size_t buf_size )
{
    mm_segment_t   oldfs;
    int            status;
    
    oldfs = get_fs(); set_fs( KERNEL_DS );
    status = sock_sendmsg( sock, sock_msg, buf_size );
    set_fs( oldfs );

    return( status );
}

#else

typedef int cccp_sock_t;
#define CCCP_SOCK_INVALID   (-1)
#define CCCP_SOCK_IS_VALID(s)  ((s) >= 0)

#define CCCP_SOCK_CREATE( sock, err ) \
	do { \
	    *(sock) = socket( CCCP_PROTO ); \
	    err = (*(sock) < 0) ? -errno : 0; \
	} while( 0 )

#define CCCP_SOCK_BIND( sock, sin, size ) \
	({int tmp_status = bind( sock, sin, size ); \
          (tmp_status < 0) ? -errno : 0; })

#define CCCP_SOCK_CLOSE( sock ) \
	(void) close( sock )

#define CCCP_SOCK_CONNECT( sock, sin, size ) \
	connect( sock, sin, size )

#define CCCP_SOCK_RECVMSG( sock, sock_msg, bug_size, flags ) \
	({int tmp_bytes = recvmsg( sock, sock_msg, flags ); \
          (tmp_bytes < 0) ? -errno : tmp_bytes; })

#define CCCP_SOCK_SENDMSG( sock, sock_msg, buf_size ) \
	({int tmp_bytes = sendmsg( sock, sock_msg, 0 ); \
          (tmp_bytes < 0) ? -errno : tmp_bytes; })

#endif


/****************************************************************************/

#if defined( __KERNEL__ )

/* Somewhere between 2.2.14 and 2.2.18 DECLARE_MUTEX was added to
 * asm/semaphore.h.  We need to be able to compile on both sets of kernel
 * header files.
 */

# ifndef DECLARE_MUTEX
#  define DECLARE_MUTEX(name)    cccp_mutex_t (name) = MUTEX
# endif

# define CCCP_SEM_INIT( s )     sema_init( s, 0 )
# define CCCP_SEM_DOWN( s )     down( s )
# define CCCP_SEM_UP( s )       up( s )

# define CCCP_MUTEX_INIT( m )   sema_init( m, 1 )
# define CCCP_MUTEX_LOCK( m )   down( m )

# define CCCP_MUTEX_GOT_LOCK     0
# define CCCP_MUTEX_NO_LOCK      1
# define CCCP_MUTEX_TRYLOCK( m ) down_trylock( m )
# define CCCP_MUTEX_UNLOCK( m ) up( m )
typedef struct semaphore cccp_mutex_t;
typedef struct semaphore cccp_sem_t;
typedef pid_t cccp_pid_t;

# define CCCP_NEW_THREAD( pid, flags, func, arg ) \
        do { pid = kernel_thread( (int (*)(void *)) func, \
				  (void *) arg, \
				  (unsigned long) flags ); } while( 0 )

# if defined( DEBUG )
#  define ASSERT_GLOBAL_LOCKED() \
	ASSERT( down_trylock( & cccp_global_lock ) != 0, \
                "global lock not held" )
# else
#  define ASSERT_GLOBAL_LOCKED()
# endif

# define PROCESS_NAME current->comm

#else

#define __SMP__
#include <asm/atomic.h>
#undef __SMP__


# undef DECLARE_MUTEX
# define DECLARE_MUTEX(name)    cccp_mutex_t name = PTHREAD_MUTEX_INITIALIZER

# define CCCP_SEM_INIT( s )     semaphore_init( s )
# define CCCP_SEM_DOWN( s )     semaphore_down( s )
# define CCCP_SEM_UP( s )       semaphore_up( s )
# define CCCP_MUTEX_LOCK( m )   (void) pthread_mutex_lock( m )
# define CCCP_MUTEX_UNLOCK( m ) (void) pthread_mutex_unlock( m )
# define CCCP_MUTEX_GOT_LOCK     1
# define CCCP_MUTEX_NO_LOCK      0

typedef pthread_mutex_t cccp_mutex_t;
typedef pthread_t cccp_pid_t;

struct cccp_sem
{
    pthread_mutex_t   lock;
    pthread_cond_t    cond;
    atomic_t          count;
};

typedef struct cccp_sem cccp_sem_t;

# define CCCP_NEW_THREAD( pid, flags, func, arg ) \
        do { (void) pthread_create( & (pid), flags, func, arg ); } while( 0 )

# if defined( DEBUG )
#  define ASSERT_GLOBAL_LOCKED() \
	ASSERT( pthread_mutex_trylock( & cccp_global_lock ) == EBUSY, \
                "global lock not held" )
# else
#  define ASSERT_GLOBAL_LOCKED()
# endif

# define PROCESS_NAME "CCCP"

#endif




/****************************************************************************/

#if defined( __KERNEL__ )
# define CCCP_HZ HZ
#else
# define CCCP_HZ 1000
#endif /* defined( __KERNEL__ ) */




/****************************************************************************/

#ifdef DEBUG

# define LOG_PACKET(P) \
    bsdlog( LOG_DEBUG, "[%s:%s] %u->%u, port/slot/flag = %u/%u/%02x, " \
	    "ACKs = 0x%08lx\n", \
	    PROCESS_NAME, __FUNCTION__, \
	    (unsigned) (P)->cp_srcnode, (unsigned) (P)->cp_node, \
	    (unsigned) (P)->cp_port, (unsigned) (P)->cp_channel, \
	    (unsigned) (P)->cp_flags, (P)->cp_acks )

#else

# define LOG_PACKET(P)

#endif /* DEBUG */

/* from cccp_init.c*/
extern int cccp_our_id;
extern volatile bool_t cccp_time_to_die;

static __inline__ unsigned thisNode( void ) { return( cccp_our_id ); }
static __inline__ int checkInit ( void ) { return (cccp_our_id != (-1)); }


/* from cccp_kernel.c */
extern int cccp_time ( void );
extern int cccp_sock_write( cccp_sock_t sock, void * data, size_t bytes );
extern void cccp_set_timer( unsigned msec );
extern size_t cccp_timeout_status( char * buf, unsigned limit );
extern void cccp_timer_release( ulong timer );

/* from cccp_init.c */
extern int cccp_init( void );
extern void cccp_stop( void );
extern cccp_sock_t cccp_our_sock;
extern cccp_mutex_t cccp_global_lock;

/* from cccp_udp.c */
extern int cccp_ip_init( uint id );
extern void * cccp_poll_loop( void * arg );
extern void * cccp_retransmit_loop( void * arg );
extern void cccp_poll_timer_handler( unsigned unused );
extern cccp_sem_t cccp_poll_sem;

/* from cccp_proto.c */
extern size_t cccp_node_stats( char * buf, unsigned limit );
extern size_t cccp_node_colors( char * buf, unsigned limit );
extern void finishMessage ( clm_msg_t * msg, int acktype );

/* from cccp_input.c */
extern int handlePacket( uint, clm_msg_t *, clm_pkt_t * );

/* from cccp_msg.c */
extern clm_msg_t * cccp_msg_unmarshal( clm_pkt_t * pkt );
extern clm_pkt_t * cccp_msg_marshal( const clm_msg_t * msg );
extern void cccp_msg_free( clm_msg_t * msg, int status );
extern void cccp_init_pkt( clm_pkt_t * pkt );

/* from cccp_ports.h */
extern void initPorts( void );
extern cccp_callback_t * getCB( uint port );
extern void cccp_port_queue_message( clm_msg_t * msg );
extern clm_msg_t * cccp_port_dequeue_message( unsigned port_index );
extern int cccp_port_lock_delivery( unsigned port );
extern void cccp_port_unlock_delivery( unsigned port );


/* from cccp_deliver.c */
extern void cccp_msg_deliver_local( clm_msg_t * msg,
    cccp_forward_callback_t * not_forwarded,
    cccp_forward_callback_t * forwarded );
void cccp_msg_just_free_pkt( clm_msg_t * m, unsigned sndr, clm_pkt_t * p );
void cccp_msg_just_queue( clm_msg_t * m, unsigned sndr, clm_pkt_t * p );
void cccp_msg_queue_ack_and_free_pkt( clm_msg_t * m, unsigned sndr, clm_pkt_t * p );
void cccp_msg_queue_redirect( clm_msg_t * m, unsigned sndr, clm_pkt_t * p );
void cccp_msg_do_remote_send( clm_msg_t * m, unsigned sndr, clm_pkt_t * p );
extern void * cccp_msg_delivery_loop( void * arg );
extern cccp_sem_t cccp_delivery_sem;


/* from glue.c */
#if ! defined( __KERNEL__ )
extern void no_trace( int hook );
extern void stack_dump( void );
extern void read_node_cfg_file( void );
extern void semaphore_up( cccp_sem_t * sem );
extern void semaphore_down( cccp_sem_t * sem );
extern void semaphore_init( cccp_sem_t * sem );
#endif

/****************************************************************************/

#define offset_of( structure, field ) \
    (unsigned) (& ((( structure * ) 0)-> field))

/****************************************************************************/
/* id2client - Convert node ID to a client TCP/UDP port
 *
 * Convert a node id to a port number for connecting (TCP) or sending (UDP)
 * to that node.
 *
 * RETURNS
 *
 * Port number.
 */

static __inline__ uint
id2client (
	    /* Client ID.
	     */
	    uint id 
	  )
{
    return( 7000 + NODEID_TO_INDEX( id ) );
}




/****************************************************************************/
/* id2server - Convert node ID to a server TCP/UDP port
 *
 * Convert a node ID to a port number for accepting connections (TCP) or
 * receiving packets (UDP).
 *
 * RETURNS
 *
 * Port number.
 *
 * NOTES
 *
 * The port number returned by this function is a constant not derived from
 * the node id.  I don't really know if there'll ever be a case where we want
 * to return a node-specific value, but this function is included anyway for
 * symmetry and possible future debugging.
 */

static __inline__ uint
id2server ( uint id )
{
    return( 8000 );
}

#endif /* CCCP_PRIVATE_H */
