/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, romanick@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: cccp_proto.c,v 1.20 2001/11/07 23:56:18 idr Exp $"

/*
 *   COMPONENT_NAME: LIB
 *
 *   FUNCTIONS: ackTimeout
 *		cccpConn
 *		checkDups
 *		checkRedirs
 *		doAckTimeouts
 *		dumpTable
 *		finishMessage
 *		getNodeInfo
 *		getPollList
 *		handleAcks
 *		handleRedirect
 *		ifsDown
 *		initProto
 *		ipConn
 *		nakMessage
 *		nodeDown
 *		openConnections
 *		queueAck
 *		queuePacket
 *		queueRedirect
 *		sendPacket
 *		transmitPacket
 *		updateNodeInfo
 */

/* PROJECT:	Cluster Lock Manager communications facility*/
/**/
/* MODULE:	protocol-specific grunge*/

#if defined( __KERNEL__ )
# ifndef MODULE
#  define MODULE
# endif

# include <linux/kernel.h>
# define __NO_VERSION__
# include <linux/module.h>
# ifdef USING_FULL_VERSIONING
#  if CONFIG_MODVERSIONS==1
#   define MODVERSIONS
#   include <linux/modversions.h>
#  endif
# endif /* USING_FULL_VERSIONING */

#else

# include <stdio.h>
# include <errno.h>
# include <string.h>
# include <unistd.h>
# include <sys/param.h>
# include <sys/poll.h>
# include <sys/uio.h>

#endif /* __KERNEL__ */

#include <cltrc.h>
#include <dlm_cluster.h>
#include <cccp.h>

#include "cccp_pkt.h"
#include "cccp_queue.h"
#include "cccp_proto.h"
#include "cccp_private.h"

/* seconds (actually, one-second boundaries) to wait before retransmitting*/
#define RETRY_SECONDS (2 * CCCP_HZ)	/* only really used in UDP version*/

struct cccp_node {
    int		state;			/* flags defined below*/
    cccp_sock_t fd;			/* fd for socket to this node*/
    cccp_sock_t old_fd;			/* fd that needs to be closed */
    ulong	node_addr;		/* addr of remote node */
    int		nextmsgslot;		/* next slot to send msg on*/
    int		nextredirslot;		/* next slot to send redir on*/
    int		readmsgslot;		/* next slot to read msg on*/
    int		readredirslot;		/* next slot to read redir on*/
    u_int32_t	incolor;		/* mask of expected inbound colors*/
    u_int32_t	outcolor;		/* mask of outbound colors*/
    u_int32_t	redirs;			/* mask of redirected inbound channels*/
    unsigned	nodeid;			/* node id destination node */
    long	session;		/* session id of latest connection */
    long	in_session;		/* most recent session id read */
    long	ack_time;		/* used to delay empty acks to node */
    clm_msg_t	*msgs[SUBCHANNELS];	/* messages for each outbound channel*/
    clm_pkt_t	*pkts[SUBCHANNELS];	/* packets for each outbound channel*/
    clm_msg_t	*inmsgs[SUBCHANNELS];	/* messages for each outbound channel*/
    clm_pkt_t	*inpkts[SUBCHANNELS];	/* packets for each outbound channel*/
    dlm_list_t	queue;			/* deferred requests*/
    ulong	deadline[SUBCHANNELS];	/* retransmission times*/
    long	ask[SUBCHANNELS];	/* retransmission times*/
    ulong	retries;
    unsigned    acks_received;
};


/* node state values*/
#define NS_NEED2SENDACK	0x01	/* new or idle channel*/
#define NS_NEED2RECVACK	0x02	/* new channel, cannot send until ack received*/
#define NS_NEEDCLOSE	0x04	/* socket hosed, need new one*/
#define NS_WARM		0x08	/* at least one successful send over socket */
#define NS_NEED2RESET	0x10	/* reset this conn if no ack in 4 secs */

#define ORIG_STATE	NS_NEED2SENDACK
#define RESTART_STATE	NS_NEED2SENDACK
#define MARK_SOCK_CLOSED(n) \
	do { \
	    ASSERT( (n)->old_fd == CCCP_SOCK_INVALID, \
		    __FUNCTION__ ": old_fd set twice" ); \
	    (n)->old_fd = node->fd; \
	    (n)->fd = CCCP_SOCK_INVALID; \
	} while( 0 )

#if defined( SHOW_RETRANS )
# define LOG_RETRANS(s, i, n ) \
    bsdlog( LOG_DEBUG, "%s %d to node %d next = %d", (s), \
            (i), (n)->nodeid, (i < 16) ? (n)->nextmsgslot : \
	    (n)->nextredirslot )
#else
# define LOG_RETRANS( s, i, n )
#endif

#define OFFSET_MSG    0
#define OFFSET_REDIR  16
#define NEXT_CHANNEL( s, o ) ((((s) - (o) + 1) % (SUBCHANNELS / 2)) + (o))


static cccp_node_t * nodeList[MAXNODES];
static int liveNodes;


/* FUNCTION:	initProto*/
/**/
/* PURPOSE:	Initialize the module-private data structures.*/
/**/
/* INPUT:	none*/
/**/
/* OUTPUT:	none*/

void
initProto ( void )
{
    (void) memset( nodeList, 0, sizeof( nodeList ) );
    liveNodes = 0;
}




/****************************************************************************/
/* cccp_change_topology - Change node topology
 * 
 * Changes node topology.  Called from clm_change_conf() after new
 * nodeid<->index established
 */

void
cccp_change_topology( void )
{
    struct cccp_node *oldl[MAXNODES] ;
    uint n ;
    int i ;


    CCCP_MUTEX_LOCK( & cccp_global_lock );
    (void) memcpy( oldl, nodeList, sizeof( oldl ) );
    (void) memset( nodeList, 0, sizeof( nodeList ) );

    for (i = 0 ; i < MAXNODES ; i++)
    {
	if ( oldl[ i ] != NULL )
	{
	    n = NODEID_TO_INDEX( oldl[ i ]->nodeid );
	    if ( (n < 0) || (n >= MAXNODES) )
	    {
		bsdlog( LOG_ERR, "[%s:%s] Node %u missing!\n",
		        PROCESS_NAME, __FUNCTION__,
		        oldl[ i ]->nodeid );
		CLM_FREE( oldl[ i ] );
	    }
	    else
	    {
		nodeList[n] = oldl[i];
	    }
	}
    }
    CCCP_MUTEX_UNLOCK( & cccp_global_lock );
}




/* FUNCTION:	getNodeInfo*/
/**/
/* PURPOSE:	Retrieve a pointer to the cccp_node_t for the given node ID.*/
/**/
/* INPUT:	node ID*/
/**/
/* OUTPUT:	pointer to cccp_node_t, NULL if not found*/

static __inline__ unsigned node_idx( unsigned node_id )
{
    unsigned   idx;


    ASSERT( node_id != 0, "zero node ID" );
    SUGGEST( node_id != thisNode(), "processing local node ID" );

    idx = NODEID_TO_INDEX( node_id );

#ifdef MASSIVE_DEBUG
    bsdlog( LOG_DEBUG, "[%s:%s] node_id[%d] idx [%d] this node[%d] \n",
            PROCESS_NAME, __FUNCTION__,
            node_id, idx, thisNode() );
#endif

    ASSERT( idx < MAXNODES, "invalid node ID" );
    
    return( idx );
}

__inline__ cccp_node_t *
getNodeInfo ( uint node_id )
{
    return nodeList[ node_idx( node_id ) ];
}




static __inline__ void
setNodeInfo ( uint node_id, cccp_node_t * node )
{
    nodeList[ node_idx( node_id ) ] = node;
}




size_t
cccp_node_stats( char * buf, unsigned limit )
{
    cccp_node_t	* node;
    unsigned      offset;
    unsigned      i;
    const char  * format;


    offset = 0;
    for ( i = 0; (i < MAXNODES) && (offset < limit) ; i++ )
    {
	node = nodeList[i];
	format = ( node != NULL ) ? "{%ld:%ld:%d:%u}" : "{-:-:-:-}";

	offset += snprintf( & buf[ offset ], limit - offset, format,
			    node->session, node->retries,
			    countQueue( & node->queue ),
			    node->acks_received );
    }

    offset += snprintf( & buf[ offset ], limit - offset, "\n\n" );
    return( offset );
}




size_t
cccp_node_colors( char * buf, unsigned limit )
{
    cccp_node_t	* node;
    unsigned      offset;
    unsigned      i;
    u_int32_t     color;


    offset = snprintf( & buf[ 0 ], limit, "in:  " );
    for (i = 0; (i < MAXNODES) && (offset < limit) ; i++) 
    {
	node = nodeList[i];
	color = (node == NULL) ? 0UL : node->incolor;
	offset += snprintf( & buf[ offset ], limit - offset, "%08lx ",
			    (unsigned long) color );
    }

    offset += snprintf( & buf[ offset ], limit - offset, "\nout: " );
    for (i = 0; (i < MAXNODES) && (offset < limit) ; i++) 
    {
	node = nodeList[i];
	color = (node == NULL) ? 0UL : node->outcolor;
	offset += snprintf( & buf[ offset ], limit - offset, "%08lx ",
			    (unsigned long) color );
    }

    offset += snprintf( & buf[ offset ], limit - offset, "\n\n" );

    return( offset );
}




/****************************************************************************/
/* updateNodeInfo - Update socket associated with a node
 *
 * Update the file descriptor associated with a node, creating the
 * node_info_t if necessary.
 * 
 * RETURNS
 *
 * A pointer to cccp_node_t for node is returned.
 */

cccp_node_t *
updateNodeInfo( 
	        /* ID of the node to update.
		 */

	        uint nodeID,


	        /* New socket for the node.  This may be CCCP_SOCK_INVALID.
		 */

	        cccp_sock_t sock
	      )
{
    cccp_node_t *ni ;
    ulong node_addr ;


    /* handle the easy case first*/
    node_addr = clm_id_to_addr( nodeID );
    ni = getNodeInfo( nodeID );
    if ( ni != NULL ) 
    {
	/* make sure the fd is up to date*/
	if ( CCCP_SOCK_IS_VALID( ni->fd ) ) 
	{
	    bsdlog( LOG_DEBUG, __FUNCTION__ ": close 0x%08lx",
		    (unsigned long) ni->fd );
	    CCCP_SOCK_CLOSE( ni->fd );
	}

	/* process that resets connection may not have been able to close fd */
	if ( CCCP_SOCK_IS_VALID( ni->old_fd ) ) 
	{
	    bsdlog( LOG_DEBUG, __FUNCTION__ ": close 0x%08lx",
		    (unsigned long) ni->old_fd );
	    CCCP_SOCK_CLOSE( ni->old_fd );

	    ni->old_fd = CCCP_SOCK_INVALID;
	}

	if (ni->state & NS_WARM)
		ni->session++;

	ni->state = RESTART_STATE;
	ni->fd = sock;
	ni->node_addr = node_addr;

#ifdef DEBUG
	bsdlog( LOG_DEBUG, "[%s:%s] Updating read session for %u to "
	        "%lu fd=0x%08lx\n",
	        PROCESS_NAME, __FUNCTION__,
	        nodeID, ni->session, (unsigned long) ni->fd );
#endif
    }
    else
    {
	/* allocate a new cccp_node_t*/
	ni = (cccp_node_t *) CLM_MALLOC(sizeof(cccp_node_t));
	if ( ni != NULL ) 
	{
	    unsigned   i;

#ifdef DEBUG
	    bsdlog(LOG_DEBUG, "Init new node struct for %d\n", nodeID);
#endif

	    /* initialize it*/

	    (void) memset( ni, 0, sizeof( *ni ) );
	    dlm_list_init( & ni->queue );

	    ni->state = ORIG_STATE;
	    ni->fd = sock;
	    ni->old_fd = CCCP_SOCK_INVALID;
	    ni->node_addr = node_addr;
	    ni->nextredirslot = OFFSET_REDIR;
	    ni->readredirslot = OFFSET_REDIR;
	    ni->nodeid = nodeID;

	    for ( i = 0 ; i < SUBCHANNELS ; i++ )
	    {
		ni->deadline[i] = ULONG_MAX;
	    }

	    setNodeInfo( nodeID, ni );
	    ++liveNodes;
	}
    }


    return( ni );
}

/****************************************************************************/
/* finishMessage -
 *
 * Handle the send callback for a message, or free stuff ourselves.
 */

void
finishMessage (
	        /* Message to ack.
		 */

	        clm_msg_t * msg,

	        /* Value for callback function.
		 */

	        int acktype 
	      )
{
    /* we could be called for a redirect w/ no message*/
    if ( msg != NULL )
    {
	if ( msg->cm_callback != NULL ) {
	    TRCHKL2T(CPTRC_SNDCB_INVOKE,msg,acktype);

	    bsdlog( LOG_DEBUG, "[%s:%s] calling message callback "
		    "@ 0x%08lx( 0x%08lx, %d )\n",
		    PROCESS_NAME, __FUNCTION__,
		    (long) msg->cm_callback, (long) msg, acktype );

	    (*msg->cm_callback)( msg, acktype );
	    TRCHKL1T(CPTRC_SNDCB_RETURN,msg);
	} else {
	    /* free everything ourselves*/
	    SUGGEST( msg->cm_private == NULL,
		     "cm_private with no send callback" );

	    if ( msg->cm_data != NULL ) {
		CLM_FREE(msg->cm_data);
	    }
	    cccp_msg_destroy(msg);
	}
    }
}

/* FUNCTION:	ifsDown*/
/**/
/* PURPOSE:	Close socket and reset fd if the ifs we were using went down*/
/**/
/* INPUT:	node ID*/
/*		IFSnode pointer*/
/**/
/* OUTPUT:	none*/

void
ifsDown ( uint id, ulong ifs_addr)
{
    cccp_node_t	*node;
    int		i;


    bsdlog(LOG_DEBUG, "ifsDown id %d  addr %s\n", id, in_ntoa(ifs_addr)) ;
    
    /* If the ID is for the local node, shutdown ALL of the connections.
     * The zero case should never happen, but if it does, try to deal with
     * it in a sensible way.
     */

    if ( (id == thisNode()) || (id == 0) )
    {
	/* check any nodes we may have connected to */
	for (i = 0; i < MAXNODES ; i++)
	{
	    /* We do not need to explicitly check that i is not
	     * thisNode() because the nodeList entry for thisNode() will
	     * ALWAYS be NULL.
	     * 
	     * Check for same net on connected node.
	     */

	    node = nodeList[i];
	    if ( (node != NULL) && CCCP_SOCK_IS_VALID( node->fd ) )
	    {
#if defined(DEBUG)
		bsdlog( LOG_DEBUG, "ifsDown: addr=0x%lx node=%u (0x%lx) "
		        "w=%lu r=%lu\n",
		        ifs_addr, node->nodeid, node->node_addr,
		        node->session, node->in_session );
#endif
		ASSERT( node->old_fd == CCCP_SOCK_INVALID,
		        "ifsDown: old_fd set twice" );

		node->old_fd = node->fd;
		node->fd = CCCP_SOCK_INVALID;
	    }
	}
    }
    else
    {
	/* here if the interface that went down was on a remote node */

	node = getNodeInfo( id );
	if ( (node != NULL) && (node->node_addr == ifs_addr) 
	     && (CCCP_SOCK_IS_VALID( node->fd )) )
	{
#if defined(DEBUG)
	    bsdlog( LOG_DEBUG, "ifsDown: addr=0x%lx node=%u (0x%lx) w=%lu "
		    "r=%lu\n", ifs_addr, node->nodeid, node->node_addr,
		    node->session, node->in_session );
#endif

	    ASSERT( node->old_fd == CCCP_SOCK_INVALID,
		    "ifsDown: old_fd set twice" );

	    node->old_fd = node->fd;
	    node->fd = CCCP_SOCK_INVALID;
	}
    }
}




static int
nakMessage ( clm_msg_t *msg, clm_pkt_t *pkt, void * body )
{
    CLM_FREE(pkt);
    finishMessage(msg,CPERR_NODEDOWN);

    return TRUE;
}

/****************************************************************************/
/* cccp_node_down - Clean up a cccp_node_t after a node goes down
 *
 * Cleans up a cccp_node_t after a node goes down.
 */

void
cccp_node_down ( 
		 /* Node ID of the node that went down.
		  */

		 uint id, 
	 
		 /* TRUE if the node failed, FALSE if "normal" shutdown.
		  */

		 bool_t node_failed
	       )
{
    cccp_node_t	*node;
    int		i;

    bsdlog( LOG_INFO, "[%s:%s] id [%d] failed [%d]\n",
            PROCESS_NAME, __FUNCTION__,
            id, node_failed );

    node = getNodeInfo( id );
    if ( node != NULL ) 
    {
	for (i = 0; i < SUBCHANNELS; ++i) {
	    if (node->inpkts[i])
		CLM_FREE(node->inpkts[i]);
	    if (node->inmsgs[i])
		cccp_msg_destroy(node->inmsgs[i]);
	}


	/* If nodeDown() is due to a node failure rather than a normal
	 * shutdown, fire off the send callbacks in the order they were
	 * made 
	 */

	if ( node_failed )
	{
	    i = node->nextmsgslot;
	    do {
		if (node->pkts[i]) {
		    CLM_FREE(node->pkts[i]);
		    finishMessage(node->msgs[i],CPERR_NODEDOWN);
		}
		i = (i + 1) % SUBCHANNELS;
	    } while (i != node->nextmsgslot);
	}


	/* If we have a connection, close it.
	 */

	if ( CCCP_SOCK_IS_VALID( node->fd ) )
	{
	    CCCP_SOCK_CLOSE( node->fd );
	    node->fd = CCCP_SOCK_INVALID;
	}

	if ( CCCP_SOCK_IS_VALID( node->old_fd ) )
	{
	    CCCP_SOCK_CLOSE( node->old_fd );
	    node->old_fd = CCCP_SOCK_INVALID;
	}

	flushQueue( & node->queue, nakMessage, 0 );

	ASSERT( dlm_list_is_empty( & node->queue ),
	        "out-bound message queue not empty!" );

	CLM_FREE(node);

	setNodeInfo( id, NULL );
    }
}




/****************************************************************************/
/* cccp_node_up - Create a connection to a new node
 *
 * Creates a connection to a node that has just joined the cluster.
 */

void
cccp_node_up(
	      /* Node ID of the node that went down.
	       */

	      uint id
	    )
{
    CCCP_MUTEX_LOCK( & cccp_global_lock );
    cccpConn( id );
    openConnections();
    CCCP_MUTEX_UNLOCK( & cccp_global_lock );
}




/****************************************************************************/
/* ipConn - Make a socke connection to a node
 *
 * Make a TCP connection to a particular node.
 *
 * RETURNS
 *
 * Zero on success, or a negative-errno value on failure.
 */

static int
ipConn(
        /* Node ID to whom we will open a connection.
	 */

        uint nodeID,


        /* Buffer to hold the socket for the connection.
	 */

        cccp_sock_t * sock 
      )
{
    int                  status;
    struct sockaddr_in	 addr;
    struct sockaddr_in	 peer;
    cccp_node_t        * node;


    ASSERT_GLOBAL_LOCKED();
    ASSERT( nodeID != thisNode(), "connecting to ourselves" );
    ASSERT( sock != NULL, "NULL socket pointer" );


    /* If the stale (probably dead) socket is still open, it must be closed
     * before the new one can be opened.
     */

    node = nodeList[ NODEID_TO_INDEX( nodeID ) ];
    ASSERT( node != NULL, "NULL node pointer!" );
    if ( CCCP_SOCK_IS_VALID( node->old_fd ) )
    {
	CCCP_SOCK_CLOSE( node->old_fd );
	node->old_fd = CCCP_SOCK_INVALID;
    }


    CCCP_SOCK_CREATE( sock, status );
    if ( status < 0 ) {
	bsdlog(LOG_CRIT, "Unable to obtain new socket for node %d\n", nodeID);
	return( status );
    }


    /* bind to a well-known address so the other side will recognize us*/
    addr.sin_family		= AF_INET;
    addr.sin_addr.s_addr	= INADDR_ANY;
    addr.sin_port		= id2client(nodeID);

    status = CCCP_SOCK_BIND( *sock, (struct sockaddr *) &addr,sizeof( addr ) );
    if ( status < 0 )
    {
	bsdlog( LOG_CRIT, __FUNCTION__ " failed sock 0x%08lx bind for node "
	        "%u, errno %u", (unsigned long) *sock, nodeID, -status );

	CCCP_SOCK_CLOSE( *sock );
	return(status);
    }

    /* connect to the remote node*/
    peer.sin_family		= AF_INET;
    peer.sin_port		= id2server(nodeID);
    peer.sin_addr.s_addr	= clm_id_to_addr( nodeID );
    if ( peer.sin_addr.s_addr == 0 ) {
	bsdlog(LOG_DEBUG, "unable to get address for node %d", nodeID);

	CCCP_SOCK_CLOSE( *sock );
	return -1;
    }

    status = CCCP_SOCK_CONNECT( *sock, (struct sockaddr *) & peer,
			        sizeof( peer ) );
    if (status < 0) {
	bsdlog( LOG_DEBUG, "[%s:%s] failed to connect to node %u at "
	        "%s:%d on 0x%08lx, errno %u",
	        PROCESS_NAME, __FUNCTION__,
	        nodeID,
	        in_ntoa(peer.sin_addr.s_addr), ntohs(peer.sin_port),
	        (unsigned long) *sock, -status );

	CCCP_SOCK_CLOSE( *sock );
	return( status );
    }

#if defined(DEBUG)
    bsdlog(LOG_DEBUG, "connected to node %d at %s:%d on 0x%08lx\n", nodeID,
	   in_ntoa(peer.sin_addr.s_addr), ntohs(peer.sin_port), 
	   (unsigned long) *sock );
#endif

    /* everything's OK*/
    return( 0 );
}




/****************************************************************************/
/* FUNCTION:	cccpConn*/
/**/
/* PURPOSE:	Ensure that a CCCP connection exists to the specified node.*/
/**/
/* INPUT:	node ID*/
/**/
/* OUTPUT:	pointer to cccp_node_t if connected, NULL otherwise*/

cccp_node_t *
cccpConn ( uint nodeID )
{
    cccp_node_t	* node;


    ASSERT_GLOBAL_LOCKED();

    node = getNodeInfo( nodeID );

    bsdlog( LOG_DEBUG, "[%s:%s] node %d 0x%08lx update fd (0x%08lx) to "
	    "0x%08lx\n",
	    PROCESS_NAME, __FUNCTION__,
	    nodeID, (unsigned long) node,
	    (unsigned long) ((node != NULL) ? node->fd : CCCP_SOCK_INVALID),
	    (unsigned long) CCCP_SOCK_INVALID );

    if ( (node == NULL) || !(CCCP_SOCK_IS_VALID( node->fd )) )
    {
	/* This creates the cccp_node_t so openConnections will see it.
	 */

	node = updateNodeInfo( nodeID, CCCP_SOCK_INVALID );
    }

    return( node );
}




/* FUNCTION:	transmitPacket*/
/**/
/* PURPOSE:	Handle the actual transmission of a packet after a channel*/
/*		has been allocated for it.*/
/**/
/* INPUT:	fp		file pointer for write()*/
/*		pkt		pointer to clm_pkt_t*/
/*		udata		user portion of packet (may be NULL)*/
/**/
/* OUTPUT:	TRUE if whole packet sent OK, FALSE otherwise*/

static bool_t
transmitPacket( cccp_node_t * node, clm_pkt_t *pkt, void *udata )
{
    int             status;
    struct iovec    iov[2];
    size_t          total_len;
    struct msghdr   msg_header;
#if defined( STRESS )
    clm_pkt_t       corrupt_pkt;
#endif

    TRCHKL5T( CPTRC_TRANSMIT, pkt, pkt->cp_node, pkt->cp_port,
	      pkt->cp_channel, pkt->cp_flags );

    pkt->cp_session = ++(node->session);
    pkt->cp_acks = node->incolor;
    pkt->cp_header_length = sizeof( clm_pkt_t );

    LOG_PACKET( pkt );

    if ( ! CCCP_SOCK_IS_VALID( node->fd ) )
    {
	bsdlog( LOG_WARNING, "[%s:%s] Invalid socket handle, node = %u\n",
	        PROCESS_NAME, __FUNCTION__,
	        pkt->cp_node );

	if ( ipConn( pkt->cp_node, & node->fd ) < 0 )
	{
	    bsdlog( LOG_WARNING, "[%s:%s] got error from ipConn\n",
		    PROCESS_NAME, __FUNCTION__ );
	    return( FALSE );
	}
    }


    msg_header.msg_name = NULL;
    msg_header.msg_namelen = 0;
    msg_header.msg_iov = iov;
    msg_header.msg_control = NULL;
    msg_header.msg_controllen = 0;
    msg_header.msg_flags = MSG_NOSIGNAL;

    iov[0].iov_base = pkt;
    iov[0].iov_len  = pkt->cp_header_length;

    total_len = 0;
    if ( ((pkt->cp_flags & CPF_MSG_PRESENT) != 0) 
	 && (pkt->cp_length != 0) )
    {
	ASSERT( udata != NULL, "missing message body in transmitPacket" );

	iov[1].iov_base	= udata;
	iov[1].iov_len	= pkt->cp_length;
	msg_header.msg_iovlen = 2;
    }
    else
    {
	msg_header.msg_iovlen = 1;

	pkt->cp_length = 0;
	if ( (pkt->cp_flags & CPF_REDIR_VALID) == 0)
	{
	    /* If there is no message data, and there is no redirect, then we
	     * don't need to send that part of the header.
	     */

	    pkt->cp_header_length = offset_of( clm_pkt_t, cp_redirect );
	}
    }

    total_len = pkt->cp_length + pkt->cp_header_length;

#if defined( STRESS )
    if ( (msg_header.msg_iovlen == 2) && ((random() % 1000) < 7) )
    {
	corrupt_pkt = *pkt;
	corrupt_pkt.cp_port = 0;
	iov[0].iov_base = & corrupt_pkt;

	bsdlog( LOG_DEBUG, "corrupting port in packet" );
    }
#endif /* STRESS */


    status = CCCP_SOCK_SENDMSG( node->fd, & msg_header, total_len );
    if (status < 0)
    {
	bsdlog( LOG_ERR, "[%s:%s] failed sock 0x%08lx (old = 0x%08lx), "
	        "errno %d",
	        PROCESS_NAME, __FUNCTION__,
	        (unsigned long) node->fd, (unsigned long) node->old_fd,
	        -status );

	MARK_SOCK_CLOSED( node );

	status = 0;
    }


    TRCHKL1T(CPTRC_WRITE,pkt);

    bsdlog( LOG_DEBUG, "[%s:%s] returning %s\n",
	    PROCESS_NAME, __FUNCTION__,
	    (status >= 0) ? "success" : "failure" );
    return( status >= 0 );
}

/* sendPacket - Allocate a subchannel and send a packet
 *
 * Allocates a subchannel for the message to be send and sends it to the
 * destination node.  This function can fail if the required subchannel is
 * not available (i.e., there is already a message in that subchannel waiting
 * for an ACK) or if the message cannot be sent to the destination node.
 *
 * If the message is sent, it will be placed in the pkts and/or msgs buffer
 * for the destination node.  It will remain there until an ACK is received.
 *
 * RETURNS
 *
 * If the packet is successfully sent, TRUE is returned.  Otherwise, FALSE
 * is returned.
 * 
 * NOTES
 * 
 * All messages to be sent by this function MUST either be redirect messages
 * or messages with user payload.  Standalone ACKs are NOT sent using this
 * function!
 */

static bool_t
sendPacket(
	    /* Pointer to the message (if any) associated with the packet
	     * that is to be sent.  Will be NULL if-and-only-if the packet
	     * is a redirect.
	     */

	    clm_msg_t * msg,


	    /* Packet that is to be sent.
	     */

	    clm_pkt_t * pkt,


	    /* Pointer to the node info stucture for the destination node.
	     */

	    cccp_node_t * node 
	  )
{
    unsigned   slot;
    unsigned * slot_ptr;
    unsigned   next_slot;
    bool_t     status;
    void     * user_data;


    LOG_PACKET( pkt );

    if ( ! CCCP_SOCK_IS_VALID( node->fd ) ) {
	return(FALSE);
    }

    /* can't send anything until we get the first ack*/
    if (node->state & NS_NEED2RECVACK) {
	return(FALSE);
    }

#if defined(EXTRA_DEBUG)
    /* integrity check - should _never_ be called twice for the same pkt*/

    for ( slot = 0 ; slot < SUBCHANNELS ; slot++ )
    {
	ASSERT( pkt != node->pkts[ slot ], "packet already pending" );
    }
#endif


    ASSERT( ((~pkt->cp_flags & CPF_REDIR_VALID) ||
		(~pkt->cp_flags & CPF_MSG_PRESENT)),
		"Packet contains redirect AND message");

    /* big slot, or little slot */
    if ( (pkt->cp_flags & CPF_REDIR_VALID) != 0 )
    {
	slot_ptr = & node->nextredirslot;

	slot = *slot_ptr;
	ASSERT( slot >= OFFSET_REDIR, __FUNCTION__ ": Invalid slot for "
	        "redirect!" );
	next_slot = NEXT_CHANNEL( slot, OFFSET_REDIR );
    }
    else
    {
	slot_ptr = & node->nextmsgslot;

	slot = *slot_ptr;
	ASSERT( slot < OFFSET_REDIR, __FUNCTION__ ": Invalid slot for "
	        "message!" );
	next_slot = NEXT_CHANNEL( slot, OFFSET_MSG );

	ASSERT( msg != NULL, "No message for non-redirect!" );
    }


    /* Slot values are either on the range [0,15] or [16,31].  Make sure
     * that bit 4 is either 0 (i.e., range [0,15]) or 1 (i.e., range [16,31])
     * for both slot and next_slot.
     */

    if ( ((slot ^ next_slot) & 0x10) != 0 )
    {
	bsdlog( LOG_CRIT, "[%s:%s] next_slot (%u) not valid for slot (%u)!\n",
	        PROCESS_NAME, __FUNCTION__,
	        next_slot, slot );
	ASSERT( ((slot ^ next_slot) & 0x10) == 0, 
	        "Invalid next slot value!" );
    }

    ASSERT( ((pkt->cp_flags & CPF_MSG_PRESENT) == 0) || (slot <  16),
	    "Invalid message slot" );
    ASSERT( ((pkt->cp_flags & CPF_REDIR_VALID) == 0) || (slot >= 16),
	    "Invalid redirect slot" );


    /* If the packet pointer for this slot is non-NULL, then there is already
     * a message waiting for an ACK in that slot.  In this case we cannot
     * send out the new message.  We have to wait until the slot becomes
     * available.
     */

    status = FALSE;
    if ( node->pkts[slot] == NULL )
    {
	/* fill in the remaining packet fields*/
	pkt->cp_channel = slot;

	
	/* The "blue" flag needs to be set to the same value as the color
	 * for the current slot.
	 */

	pkt->cp_flags &= ~CPF_BLUE;
	pkt->cp_flags |= GETBIT( node->outcolor, slot ) << CPB_BLUE;


	/* Send the data over the wire.  The worst that can happen is that
	 * it's not acked and we retry later.
	 */

	user_data = (msg == NULL) ? NULL : msg->cm_data;
	status = transmitPacket( node, pkt, user_data );
	if ( status ) 
	{
	    node->state |= NS_WARM;

	    /* mark the slot as allocated*/
	    node->deadline[slot] = cccp_time() + RETRY_SECONDS;
	    cccp_set_timer( RETRY_SECONDS );

	    *slot_ptr = next_slot;

	    /* store the message/packet info in the slot*/

	    ASSERT( node->msgs[ slot ] == NULL, "Message already in slot!" );
	    ASSERT( node->pkts[ slot ] == NULL, "Packet already in slot!" );

	    node->msgs[ slot ] = msg;
	    node->pkts[ slot ] = pkt;

	    node->state &= ~NS_NEED2SENDACK;
	}
    }


    return( status );
}

/* FUNCTION:	queuePacket*/
/**/
/* PURPOSE:	Allocate a packet and send it using sendPacket (above).*/
/*		This routine is basically a wrapper to handle queuing when*/
/*		the link is full.*/
/**/
/* INPUT:*/
/*		msg	pointer to clm_msg_t stored in node_info_t*/
/*		pkt	pointer to clm_pkt_t, stored and used to send*/
/*		udata	pointer to user data for packet*/
/*		id	node ID of destination*/
/**/
/* OUTPUT:	TRUE if packet queued/sent, FALSE otherwise*/

int
queuePacket ( clm_msg_t *msg, clm_pkt_t *pkt, uint dest_node )
{
    cccp_node_t	*node;
    int          status;

    ASSERT( pkt != NULL, "NULL packet" );
    ASSERT( dest_node != thisNode(), "bad node in queuePacket" );

    /* try to deliver using sendPacket*/
    status = FALSE;
    node = cccpConn( dest_node );
    if ( node != NULL ) {
#if defined( OLD_MARSHAL )
	pkt->cp_node    = dest_node;
	pkt->cp_srcnode = thisNode();
#endif

	status = sendPacket( msg, pkt, node );
	if ( ! status && (msg != NULL) )
	{
	    /* could not send - queue instead*/
	    ASSERT( (pkt->cp_flags & CPF_REDIR_VALID) == 0,
		    "Attempting to queue redirect" );
	    
	    msg->cm_packet = pkt;
	    msg->cm_node_ptr = node;
	    dlm_list_add_tail( & node->queue, (dlm_listnode_t *) msg );
	    status = TRUE;
	}
    }

    return( status );
}

/* FUNCTION:	handleAcksGuts */
/**/
/* PURPOSE:	This function was pulled out of handleAcks to loop through*/
/*		the subchannels for the node 16 at a time.  It is called */
/*		once for the message half and once for the redirect half. */
/**/
/* INPUT:	node		acknowledging node*/
/*		start		subchannel to start on */
/*		offset		either 0 or 16, depeing on which half */
/*		acks		acknowledgement mask to check vs. outcolor*/
/**/
/* OUTPUT:	none*/
void
handleAcksGuts(cccp_node_t *node, int start, int offset, int acks)
{
    int		i = start;
    clm_pkt_t	*pp;
    clm_msg_t	*mp;
    unsigned    ack_mask;


    /* Each bit in acks that does not match the bit in outcolor represents
     * a new ack that must be processed.
     */

    ack_mask = acks ^ node->outcolor;


    do {
	if ( GETBIT( ack_mask, i ) != 0 ) {
	    /* take care of the clm_pkt_t
	     */
	    pp = node->pkts[i];
	    mp = node->msgs[i];

	    ASSERT( pp != NULL, "ack received on idle subchannel" );

	    /* update the redirect mask
	     */
	    if (pp->cp_flags & CPF_REDIR_VALID) {
		CLRBIT(node->redirs,pp->cp_redirect.cr_oldchan);
	    }

	    /* Take care of the callback and clm_msg_t.  The message pointer
	     * could be NULL if this was a redirect.
	     */
	    if (mp) {
		finishMessage( mp, CPERR_OK );
		node->msgs[i] = NULL;
	    }

	    CLM_FREE( node->pkts[i] );
	    node->pkts[i] = NULL;


	    /* Change state for this subchannel.  At this point, it should
	     * match the value (for the ith bit) of acks.
	     */

	    FLIPBIT(node->outcolor,i);


	    node->deadline[i] = ULONG_MAX;
	    node->acks_received++;


	    /* Call flushQueue here in case finishMessage on future ack
	     * tries to send.  It could find a free slot from previous ack and
	     * send packet out of order when packet on queue should go first.
	     */
	    if (mp) {
		flushQueue( & node->queue,
			    (queue_flush_func_t *) sendPacket, 0);
	    }
	}
	ASSERT( i >= offset, __FUNCTION__ ": Invalid slot!" );
	i = NEXT_CHANNEL( i, offset );
    } while (i != start);
}


/* FUNCTION:	handleAcks*/
/**/
/* PURPOSE:	Process acknowledgements contained in a packet by changing*/
/*		the color for that subchannel, invoking the send callback,*/
/*		and so forth as described in the spec.*/
/**/
/* INPUT:	node		acknowledging node*/
/*		acks		acknowledgement mask to check vs. outcolor*/
/**/
/* OUTPUT:	none*/

int
handleAcks ( cccp_node_t *node, u_int32_t acks )
{
#if defined( SHOW_ACKS )
    static u_int32_t oldacks = 0;
#endif

    if (node->outcolor ^ acks) {
	TRCHKL1T(CPTRC_GOTACKS,node->outcolor^acks);
    }

#if defined(SHOW_ACKS)
    if (acks != oldacks) {
	oldacks = acks;
	bsdlog( LOG_DEBUG, "[%s:%s] acks 0x%08lx ^ 0x%08lx = 0x%08lx\n",
	        PROCESS_NAME, __FUNCTION__,
	        node->outcolor, acks, node->outcolor ^ acks );
    }
#endif

    /* if we received an ACK, clear bits to indicate connection is alive */

    node->state &= ~(NS_NEED2RECVACK|NS_NEED2RESET);

    /* handle all acks first*/
    handleAcksGuts(node, node->nextmsgslot,   OFFSET_MSG,   acks);
    handleAcksGuts(node, node->nextredirslot, OFFSET_REDIR, acks);

    ASSERT( node->outcolor == acks, "ack botch" );

    flushQueue( & node->queue, (queue_flush_func_t *) sendPacket, 0 );

    return 0;
}

/* FUNCTION:	queueAck*/
/**/
/* PURPOSE:	Change the color on an inbound subchannel and mark the node*/
/*		as needing an ack.  If NS_NEED2SENDACK is still set when*/
/*		ackTimeout (below) is called, a separate ack packet will be*/
/*		created; more commonly, the ack will be piggybacked.*/
/**/
/* INPUT:	node		node to receive ack*/
/*		chan		subchannel to acknowledge*/
/**/
/* OUTPUT:	none*/

void
queueAck ( cccp_node_t *node, uint chan )
{
    FLIPBIT(node->incolor,chan);
    node->ask[chan] = 0;
    node->state |= NS_NEED2SENDACK;

    /* Set a timer so that we will send a raw ACK if there is no other
     * messaging activity.
     */

    cccp_set_timer( 4 * CCCP_HZ );

    bsdlog( LOG_DEBUG, "[%s:%s] new ACK mask for node %u: %08lx\n",
	    PROCESS_NAME, __FUNCTION__,
	    node->nodeid, (unsigned long) node->incolor );
}




/****************************************************************************/
/* checkSession - Check session ID of incoming packet
 * 
 * Checks session id of incoming packet.  Updates node if new.
 * 
 * RETURNS
 *
 * If the session has been dropped, TRUE will be returned.  If the session
 * is good and the node->in_session has been updated, FALSE will be
 * returned.
 */

bool_t
checkSession(
	      /* Pointer to cccp_node_t for this node.
	       */

	      cccp_node_t * node, 

	      /* Packet to check.
	       */

	      clm_pkt_t *pkt
	    )
{
    /* Only process packets that are destined for this node.
     */

    if ( pkt->cp_node == thisNode() )
    {
	/* check for new session id */
	if ( pkt->cp_session > node->in_session ) 
	{
#if defined(SHOW_SESSIONS)
	    bsdlog( LOG_DEBUG, "Read session from %d to %d for node %d\n",
		    node->in_session, pkt->cp_session, node->nodeid );
#endif
	    node->in_session = pkt->cp_session;
	    return FALSE;
	}
	else
	{
	    /* this packet is from a dropped connection.  See ya */
	    bsdlog( LOG_DEBUG, "drop - got %lu expect %lu from %d "
		    "acks=0x%08lx f=%02x c=%d\n", 
		    pkt->cp_session, (node->in_session+1),
		    node->nodeid, pkt->cp_acks, pkt->cp_flags, 
		    pkt->cp_channel );
	}
    }

    return TRUE;
}




/* FUNCTION:	nackRetrans */
/**/
/* PURPOSE:	Retransmit all packets specified in the NACK mask */
/**/
/* INPUT:	node		pointer to cccp_node_t for this node*/
/*		nacks		nack mask */
/**/
/* OUTPUT:	none*/

void
nackRetrans ( cccp_node_t *node, ulong start, ulong got, ulong nacks )
{
    unsigned    index;
    unsigned    slot;
    ulong	now = cccp_time();
    ulong       next_timeout;
    bool_t      message_logged;

    unsigned    spread;
    unsigned    offset;

#if defined(SHOW_RETRANS)
    bsdlog( LOG_DEBUG, "nackRetrans to %d nacks= 0x%08lx\n", node->nodeid,
	    nacks );
#endif


    /* If start falls outside SUBCHANNELS, the NACK is bad, reject it.
     */

    if ( start >= SUBCHANNELS )
    {
        return;
    }

    offset = (start >= OFFSET_REDIR) ? OFFSET_REDIR : OFFSET_MSG;

    spread = (start >= got) ? (got + (SUBCHANNELS / 2)) : got;
    spread -= start;


    message_logged = FALSE;
    next_timeout = RETRY_SECONDS * 2;
    for ( index = 0 ; index < spread ; index++ )
    {
	slot = NEXT_CHANNEL( (start + index), offset );

	if ( GETBIT( nacks, slot ) != 0 ) 
	{
	    clm_msg_t * msg;
	    clm_pkt_t * pkt;
	    void      * udata;


	    msg = node->msgs[ slot ];
	    pkt = node->pkts[ slot ];
	    udata = (msg == NULL) ? NULL : msg->cm_data;


	    /* This used to be an assertion, but that opens the door for
	     * a DoS attack.  There is no harm in processing false NACKs,
	     * so if there is an invalid bit, ignore just that bit.
	     */

	    if ( pkt == NULL )
	    {
		bsdlog( LOG_ERR, "[%s:%s] NACK mask (0x%08lx) has NACK for "
		        "slot %u, but there is no packet there!\n",
		        PROCESS_NAME, __FUNCTION__,
		        nacks, slot );

		continue;
	    }


	    if (pkt->cp_flags & CPF_MSG_PRESENT) 
	    {
		ASSERT( msg != NULL,
		        "missing message header detected in nackRetrans" );

		/* If the length is non-zero (which the length may be 0),
		 * then the pointer to the message data must not be NULL.
		 */

		ASSERT( (pkt->cp_length == 0) || (udata != NULL),
		        "missing message body detected in nackRetrans" );

		node->retries++;
	    }

	    LOG_RETRANS( "retransmitting", slot, node );

	    /* Retransmit the packet.  If this fails, it just means that an
	     * ACK won't be received.  Since this is basically the current
	     * state, this isn't a problem.  However, it is VERY important
	     * to reset the ACK timeout for this message!
	     */

	    if ( ! message_logged && ! transmitPacket( node, pkt, udata ) )
	    {
		bsdlog( LOG_ERR, __FUNCTION__ ": transmit to %d failed",
		        msg->cm_node );
		message_logged = TRUE;
	    }
	    else
	    {
		node->state |= NS_WARM;
	    }

	    node->deadline[ slot ] = now + RETRY_SECONDS;

	} /* end if nack bit was set */


	if ( (node->deadline[ slot ] - now) < next_timeout )
	{
	    next_timeout = (node->deadline[ slot ] - now);
	}
    }


    cccp_set_timer( next_timeout );
}




/****************************************************************************/
/* ackTimeoutGuts - Retransmit any messages that have timed-out
 * 
 * This function was pulled out of handleAcks to loop through the subchannels
 * for the node 16 at a time.  It is called once for the message half and
 * once for the redirect half.
 *
 * RETURNS
 * 
 * The amount of time (in CCCP_HZ units) remaining until the next timeout.
 */

ulong
ackTimeoutGuts(
	        /* Destination node that is to be processed.
		 */

	        cccp_node_t * node,


	        /* Index of the subchannel to start on.
		 */

	        int start
	      )
{
    unsigned    index;
    unsigned    slot;
    ulong       now = cccp_time();
    ulong       next_timeout;
    bool_t      message_logged;
    unsigned    offset;


    offset = (start >= OFFSET_REDIR) ? OFFSET_REDIR : OFFSET_MSG;

    message_logged = FALSE;
    next_timeout = RETRY_SECONDS * 2;
    for ( index = 0 ; index < (SUBCHANNELS / 2) ; index++ )
    {
	slot = NEXT_CHANNEL( (start + index), offset );


	/* Has the retransmit timeout for this message slot expired?  The
	 * comparison is "fuzzy" so that messages that are about to timeout
	 * will be retransmitted.  The thinking is that if most of the
	 * timeout period has expired, the odds are pretty good that all of
	 * it will.
	 * 
	 * NOTICE that the value of 'now' is NOT modified!  This could be done
	 * outside the loop to speed things up a small bit.  HOWEVER, the
	 * value of 'now' is used below, and it must be the actual value.
	 * 
	 * Testing the 'deadline' before testing the 'pkt' is safe because
	 * when a message is retired the deadline MUST BE SET to ULONG_MAX.
	 * In this case, this comparison will ALWAYS fail, hence the assertion
	 * that 'pkt' must be NULL.
	 */

	if ( (now + (RETRY_SECONDS / 8)) >= node->deadline[ slot ] ) 
	{
	    clm_msg_t * msg;
	    clm_pkt_t * pkt;
	    void * udata;


	    msg = node->msgs[ slot ];
	    pkt = node->pkts[ slot ];
	    udata = (msg == NULL) ? NULL : msg->cm_data;

	    ASSERT( pkt != NULL, "NULL packet for timed-out slot!" );

	    if ( (pkt->cp_flags & CPF_MSG_PRESENT) != 0 )
	    {
		ASSERT( msg != NULL,
		        "missing message header detected in ackTimeout" );

		ASSERT( (pkt->cp_length == 0) || (udata != NULL),
		        "missing message body detected in ackTimeout" );
	    }


	    LOG_RETRANS( "timeout retran", slot, node );


	    /* Retransmit the packet.  If this fails, it just means that an
	     * ACK won't be received.  Since this is basically the current
	     * state, this isn't a problem.  However, it is VERY important
	     * to reset the ACK timeout for this message!
	     */

	    if ( ! message_logged && ! transmitPacket( node, pkt, udata ) )
	    {
		bsdlog( LOG_ERR, __FUNCTION__ ": transmit to %d failed",
		        pkt->cp_node );
		message_logged = TRUE;
	    }

	    node->deadline[ slot ] = now + RETRY_SECONDS;


	    /* Turn on bit in state to indicate possible failed
	     * connection.
	     */

	    node->state |= NS_NEED2RECVACK;
	}


	if ( (node->deadline[ slot ] - now) < next_timeout )
	{
	    next_timeout = (node->deadline[ slot ] - now);
	}
    }


    return( next_timeout );
}


/* FUNCTION:	ackTimeout*/
/**/
/* PURPOSE:	Check whether we've exceeded our time limit for piggybacking*/
/*		an ack and, if so, generate a separate ack packet.*/
/**/
/* INPUT:	node		pointer to cccp_node_t for this node*/
/**/
/* OUTPUT:	none*/

unsigned long
ackTimeout ( cccp_node_t *node )
{
    clm_pkt_t   mypkt;
    ulong       now;
    int         status;
    ulong       next_timeout[2];


    now = cccp_time();

    ASSERT( node != NULL, "null node in ackTimeout" );
    ASSERT( CCCP_SOCK_IS_VALID( node->fd ), "bad fd in ackTimeout" );

    /* ensure that we send periodic acks even with no other activity*/
    if ( ((now - node->ack_time) < 4) 
	 && ((node->state & NS_NEED2SENDACK) == 0) )
    {
	return( 5 * CCCP_HZ );
    }

    /* if reset bit still set, then close and restart connection */
    if (node->state & NS_NEED2RESET)
    {
	bsdlog( LOG_DEBUG, "[%s:%s] Reset %d - stale",
	        PROCESS_NAME, __FUNCTION__,
	        node->nodeid );
	MARK_SOCK_CLOSED( node );
	return( 5 * CCCP_HZ );
    }

    /* no ack recvd since retransmit, mark conn for possible reset */
    if (node->state & NS_NEED2RECVACK)
    {
	bsdlog( LOG_DEBUG, "[%s:%s] Need reset?  %d - stale",
	        PROCESS_NAME, __FUNCTION__,
	        node->nodeid );
	node->state |= NS_NEED2RESET;
    }

    node->ack_time = now;
    node->state &= ~NS_NEED2SENDACK;

    cccp_init_pkt( & mypkt );

    mypkt.cp_session	= ++node->session;
    mypkt.cp_node	= node->nodeid;
    mypkt.cp_acks	= node->incolor;
    mypkt.cp_header_length = offset_of( clm_pkt_t, cp_redirect );

    status = cccp_sock_write( node->fd, & mypkt, mypkt.cp_header_length );
    if ( status < 0 )
    {
	bsdlog( LOG_ERR, __FUNCTION__ ": write to %d failed, errno %d",
	        node->nodeid, -status );
	MARK_SOCK_CLOSED( node );
	return( 5 * CCCP_HZ );
    }
    node->state |= NS_WARM;
    TRCHKL1T(CPTRC_WRITE,&mypkt);


    next_timeout[0] = ackTimeoutGuts( node, node->nextmsgslot );
    next_timeout[1] = ackTimeoutGuts( node, node->nextredirslot );
    
    return( (next_timeout[0] < next_timeout[1]) 
	   ? next_timeout[0] 
	   : next_timeout[1] );
}


/* FUNCTION:	doAckTimeouts*/
/**/
/* PURPOSE:	Do ack timeouts on all active links*/
/**/
/* INPUT:	none*/
/**/
/* OUTPUT:	none*/

unsigned long
doAckTimeouts ( void )
{
    int		  i;
    cccp_node_t	* node;
    ulong         next_timeout;
    ulong         temp;


    next_timeout = RETRY_SECONDS * 2;
    for (i = 0; i < MAXNODES; ++i) 
    {
	node = nodeList[i];
	if ( (node != NULL) && CCCP_SOCK_IS_VALID( node->fd ) ) 
	{
	    temp = ackTimeout( node );
	    if ( temp < next_timeout )
	    {
		next_timeout = temp;
	    }
	}
    }

    return( next_timeout );
}

/* FUNCTION:	queueRedirect*/
/**/
/* PURPOSE:	Generate a redirect, allocate a subchannel for it, and send*/
/*		it to the original packet sender.*/
/**/
/* INPUT:	msg		clm_msg_t being redirected*/
/*		pkt		clm_pkt_t being redirected*/
/*		node		pointer to the cccp_node_t*/
/*		id		node ID for sender/destination*/
/**/
/* OUTPUT:	none*/
/**/
/* NOTES:	We send immediately even though the protocol allows us to*/
/*		piggyback, on the theory that they've already waited one*/
/*		transit time and we shouldn't make them wait any more than*/
/*		we have to.*/
/**/
/*		This routine frees the message passed to it, including the*/
/*		cm_data pointer.*/

void
queueRedirect ( clm_msg_t *msg, clm_pkt_t *pkt, cccp_node_t *node, uint id )
{
    uint	chan	= pkt->cp_channel;


    ASSERT( NODEID_TO_INDEX( msg->cm_node ) >= 0, 
	    "invalid node ID in getNodeInfo" );


    /* modify the clm_pkt_t*/
    pkt->cp_flags		= CPF_REDIR_VALID;
    pkt->cp_redirect.cr_hops	= msg->cm_hops;
    pkt->cp_redirect.cr_node	= msg->cm_node;
    pkt->cp_redirect.cr_port	= msg->cm_port;
    pkt->cp_redirect.cr_oldchan	= chan;

    /* free the clm_msg_t*/
    cccp_msg_free( msg, CPERR_OK );


    if ( queuePacket( NULL, pkt, id ) ) 
    {
	SETBIT(node->redirs,chan);
    }
    else
    {
	bsdlog( LOG_WARNING, "WARNING: unable to queue redirect packet "
	        "(in %s)\n", __FUNCTION__ );
    }
}

/* FUNCTION:	handleRedirect*/
/**/
/* PURPOSE:	Handle a redirect by retransmitting to the new address.*/
/**/
/* INPUT:	node		the node from which we got the redirect*/
/*		redir		new destination node/port, plus subchannel*/
/**/
/* OUTPUT:	none*/

void
handleRedirect ( cccp_node_t *node, const clm_redir_t * redir )
{
    clm_msg_t * msg;
    clm_pkt_t * pkt;
    uint        chan;



    chan = redir->cr_oldchan;

    /* Make sure the subchannel # is reasonable.
     */

    if ( chan >= SUBCHANNELS )
    {
	bsdlog( LOG_ERR, "Invalid subchannel (%u) for redirect", chan );
	return;
    }


    /* Make sure we actually had a user message going out on that channel.
     */

    pkt = node->pkts[chan];
    msg = node->msgs[chan];

    if ( pkt == NULL )
    {
	bsdlog( LOG_ERR, "Redirect received for idle subchannel (%u)", chan );
	ASSERT( msg == NULL, "non-NULL msg pointer" );
	return;
    }
    
    if ( (pkt->cp_flags & CPF_MSG_PRESENT) == 0 )
    {
	bsdlog( LOG_ERR, "Redirect received on subchannel (%u) for internal "
	        "packet", chan );
	return;
    }


    if (pkt->cp_flags & CPF_REDIR_VALID) {
	CLRBIT(node->redirs, pkt->cp_redirect.cr_oldchan);
        pkt->cp_flags &= ~CPF_REDIR_VALID;
    }

    /* remove the message/packet from this node/subchannel*/

    ASSERT( msg != NULL, "NULL msg pointer" );

    node->msgs[chan] = NULL;
    node->pkts[chan] = NULL;
    node->deadline[chan] = ULONG_MAX;

    TRCHKL3T(CPTRC_GOTREDIR,msg,redir->cr_node,redir->cr_port);

    /* send it to the new destination*/
    msg->cm_node = redir->cr_node;
    msg->cm_port = redir->cr_port;
    msg->cm_hops = redir->cr_hops;

    msg->cm_packet = pkt;

    CCCP_MUTEX_UNLOCK( & cccp_global_lock );
    cccp_msg_deliver_local( msg, NULL, 
			    (cccp_forward_callback_t *) cccp_msg_just_queue );
    CCCP_MUTEX_LOCK( & cccp_global_lock );

    /* if redir freed slot, give the queue 1st crack at it */
    flushQueue( & node->queue, (queue_flush_func_t *) sendPacket, 0 );
}

/* openConnections - Look for node w/a cccp_node_t but no open socket
 * 
 * Look for any nodes for which we have a cccp_node_t but no valid file
 * descriptor.  Call makeConn for any such nodes.
 */

void
openConnections ( void )
{
    int           i;
    int           id;
    cccp_node_t * node;
    cccp_node_t * ni ;
    cccp_sock_t   sock;
    int           status;


    ASSERT_GLOBAL_LOCKED();
    for (i = 0; i < MAXNODES; ++i)
    {
	id = INDEX_TO_NODEID(i);
	node = nodeList[i];
	if ( (node != NULL) && (clm_node_reachable( id ))
	     && !(CCCP_SOCK_IS_VALID( node->fd )) )
	{
#ifdef SHOW_CONNECTIONS
	    bsdlog(LOG_DEBUG, "opening connection to index %d\n",i);
#endif


	    /* no way this should fail*/
	    status = ipConn( id, & sock );

	    /* connect can fail because of Linux SNMP protocols;
	     * TBD: there may be a way to turn this off
	     */
	    if ( status == 0 )
	    {
		ni = updateNodeInfo( id, sock );
		if ( ni != NULL )
		{
		    ASSERT( ni == node, "New and old node objects do not "
			    "match" );
		    flushQueue( & node->queue,
			        (queue_flush_func_t *) sendPacket, 0 );
		}
		else
		{
		    bsdlog( LOG_WARNING, "[%s:%s] unable to update node info "
			    "for %u, closing stale socket\n",
			    PROCESS_NAME, __FUNCTION__, id );
		    CCCP_SOCK_CLOSE( sock );
		}
	    }
	    else
	    {
		bsdlog(LOG_DEBUG, "openConnections: ipConn failed") ;
	    }
	}
    }
}




/****************************************************************************/
/* checkRedirs - Determine if a redirect is already queues for a channel
 *
 * See whether a redirect is already queued for this channel, which would
 * mean we've already seen it and should not process it at all this time
 * around.
 *
 * RETURNS
 * 
 * If a redirect is queue on the channel, TRUE is returned.  FALSE is
 * returned otherwise.
 */

bool_t
checkRedirs( 
	     /* Pointer to the node to be tested.
	      */

	     cccp_node_t * node, 

	     /* Channel to check.
	      */

	     uint chan
	   )
{
    return( GETBIT( node->redirs, chan ) != 0 );
}

/* FUNCTION:	sendNack */
/**/
/* PURPOSE:	Transmit a NACK mask to the specified node to generate */
/*		retransmissions of the packets indicated. */
/**/
/* INPUT:	node		pointer to the cccp_node_t */
/*		first		first missing channel */
/*		got		channel that we received on */
/**/
/* OUTPUT:	Nothing */
void
sendNack ( cccp_node_t *node, uint first, uint got )
{
    clm_pkt_t   mypkt;
    unsigned    offset;
    ulong     * ackptr;
    ulong       now = cccp_time();
    int         spread = 0;
    int         status;
#if defined(SHOW_NACKS)
    int         ofirst = first, ogot = got;
#endif

    /*
     * it is possible to generate a NACK before connection is open.  If file
     * descriptor is invalid, we can't send a NACK
     */
    if ( ! CCCP_SOCK_IS_VALID( node->fd ) )
	return;
    
    cccp_init_pkt( & mypkt );

    mypkt.cp_session = ++(node->session);
    mypkt.cp_node = node->nodeid;

    mypkt.cp_flags = CPF_NACK_VALID;
    mypkt.cp_channel = first;
    mypkt.cp_hops = got;

    ackptr = &mypkt.cp_acks;

    /* If first is bigger than got, then the mask is wrapped, but on a
     * 16-bit boundary. offset is either 16 or 32, depending on which half
     * of the mask we're dealing with.
     */

    if (first > got) {
        offset = got / (SUBCHANNELS/2);
        offset = offset * (SUBCHANNELS/2) + (SUBCHANNELS/2);
        for ( /* empty */ ; first < offset ; first++ ) {
	    /*
	     * send NACKS for pkts we don't already have & haven't asked for in
	     * the last second - don't send NACKs for redir'd msgs
	     */
	    if (node->ask[first] != now && node->inpkts[first] == NULL &&
					(GETBIT(node->redirs,first) == 0)) {
		    *ackptr |= (1 << first);
	    }
	    spread++;
	}
	first = offset - (SUBCHANNELS/2);
    }


    for ( /* empty */ ; first < got ; first++ ) {
	    /* Send NACKS for pkts we don't already have & haven't asked for
	     * in the last second - don't send NACKs for redir'd msgs
	     */

	    if (node->ask[first] != now && node->inpkts[first] == NULL &&
					(GETBIT(node->redirs,first) == 0)) {
		    *ackptr |= (1 << first);
	    }
	    spread++;
    }

    /* why bother ? */
    if (spread < 8 || *ackptr == 0) {
	return;
    }

#if defined(SHOW_NACKS)
    bsdlog( LOG_DEBUG, "[%s:%s] Send NACK to node %d for %d to %d -> "
	    "0x%08lx\n", PROCESS_NAME, __FUNCTION__,
	    node->nodeid, ofirst, ogot, *ackptr );
#endif

    /* timestamp the NACKS */
    for (first = 0; first < SUBCHANNELS; first++) {
	if (*ackptr & (1 << first)) {
		node->ask[first] = now;
	}
    }

    status = cccp_sock_write(node->fd, &mypkt, sizeof mypkt);
    if ( status < 0 )
    {
	bsdlog( LOG_ERR, "[%s:%s] could not send NACKs to node %u, "
	        "errno %d\n",
	        PROCESS_NAME, __FUNCTION__,
	        node->nodeid, -status );
	MARK_SOCK_CLOSED( node );
	return;
    }
    node->state |= NS_WARM;
}

/* FUNCTION:	reOrder*/
/**/
/* PURPOSE:	Check the saved packets for this node and handle any */
/*		that were put on hold because of ordering. */
/**/
/* INPUT:	node		pointer to the cccp_node_t*/
/**/
/* OUTPUT:	Nothing */
void
reOrder(cccp_node_t *node, uint sender)
{
    clm_pkt_t	*pkt;
    clm_msg_t	*msg;
    int		slot;


    for ( slot = node->readredirslot 
	  ; node->inpkts[ slot ] != NULL
	  ; slot = node->readredirslot )
    {
	pkt = node->inpkts[slot];
	msg = node->inmsgs[slot];

	ASSERT( msg != NULL, "NULL msg pointer" );

        node->inmsgs[slot] = NULL;
        node->inpkts[slot] = NULL;

	ASSERT( slot >= OFFSET_REDIR, __FUNCTION__ ": Invalid slot for "
	        "redirect!" );
        node->readredirslot = NEXT_CHANNEL( slot, OFFSET_REDIR );

        /* handle redirect*/
        if ( (pkt->cp_flags & CPF_REDIR_VALID) != 0 ) {
	    queueAck(node,pkt->cp_channel);
	    handleRedirect(node,&pkt->cp_redirect);
        }

        cccp_msg_destroy(msg);
        CLM_FREE(pkt);
    }


    for ( slot = node->readmsgslot
	  ; node->inpkts[ slot ] != NULL
	  ; slot = node->readmsgslot )
    {
	pkt = node->inpkts[slot];
	msg = node->inmsgs[slot];

	ASSERT( msg != NULL, "NULL msg pointer" );

        node->inmsgs[slot] = NULL;
        node->inpkts[slot] = NULL;

	ASSERT( slot <  OFFSET_REDIR, __FUNCTION__ ": Invalid slot for "
	        "message!" );
        node->readmsgslot = NEXT_CHANNEL( slot, OFFSET_MSG );

        /* handle user message*/
        if (pkt->cp_flags & CPF_MSG_PRESENT) {
	    ASSERT( pkt->cp_channel == slot, "Packet in wrong slot!" );

	    msg->cm_sender = sender;
	    msg->cm_packet = pkt;

	    CCCP_MUTEX_UNLOCK( & cccp_global_lock );
	    cccp_msg_deliver_local( msg,
		  (cccp_forward_callback_t *) cccp_msg_queue_ack_and_free_pkt,
		  (cccp_forward_callback_t *) cccp_msg_queue_redirect );
	    CCCP_MUTEX_LOCK( & cccp_global_lock );
        }
    }
}

/****************************************************************************/
/* checkOrder - Check whether a packet is the next one expected
 *
 * Check whether a packet is the next one that we expect.  If not, save it
 * and return TRUE, otherwise do nothing and return FALSE
 *
 * RETURNS
 *
 * If packet is out of order TRUE is returned.  Otherwise, FALSE is returned.
 */

#if defined( SHOW_ORDERING )
#define ORDERING_LOG(s)	bsdlog( LOG_DEBUG, s " - read %d frm %d exp %d\n", \
                                chan, node->nodeid, node->readmsgslot );
#else
#define ORDERING_LOG(s)
#endif /* SHOW_ORDERING */

int
checkOrder ( cccp_node_t * node, clm_pkt_t * pkt, clm_msg_t * msg )
{
    int	  chan = pkt->cp_channel;
    int   next_chan;
    int * slot;

    ASSERT( (~pkt->cp_flags & CPF_REDIR_VALID) || 
	    (~pkt->cp_flags & CPF_MSG_PRESENT),
	    "Packet contains redirect AND message" );

    /* Depending on whether there is a message pay-load or a redirect in the
     * packet, select either the read-message slot or the read-redirect slot.
     */

    if ( (pkt->cp_flags & CPF_MSG_PRESENT) != 0 )
    {
	slot = & node->readmsgslot;
	ASSERT( chan <  OFFSET_REDIR, __FUNCTION__ ": Invalid slot for "
	        "message!" );
	next_chan = NEXT_CHANNEL( chan, OFFSET_MSG );
    }
    else
    {
	slot = & node->readredirslot;
	ASSERT( *slot >= OFFSET_REDIR, __FUNCTION__ ": Invalid slot for "
	        "redirect!" );
	next_chan = NEXT_CHANNEL( chan, OFFSET_REDIR );
    }


    if ( chan != *slot ) {
	if ( node->inpkts[chan] != NULL ) {
	    CLM_FREE( pkt );
	    cccp_msg_destroy( msg );

	    ORDERING_LOG( "dup in checkOrder" );
	} else {
	    node->inmsgs[chan] = msg;
	    node->inpkts[chan] = pkt;

	    ORDERING_LOG( "packet out of order" );
	}
	sendNack( node, *slot, chan );
	return TRUE;
    }

    *slot = next_chan;
    return FALSE;
}

/* FUNCTION:	checkDups*/
/**/
/* PURPOSE:	Check whether a packet is a duplicate of one already*/
/*		received, using the CPF_BLUE flag.*/
/**/
/* INPUT:	node		pointer to the cccp_node_t*/
/*		flags		flags for the packet*/
/*		chan		channel packet was received on*/
/**/
/* OUTPUT:	TRUE if packet is a duplicate, FALSE otherwise*/

bool_t
checkDups ( cccp_node_t *node, ulong flags, uint chan )
{
    bool_t     is_dup;
    unsigned   blue;
    unsigned   color;


    is_dup = FALSE;
#ifdef DEBUG
    blue = 0;
    color = 0;
#endif

    if ( (flags & (CPF_REDIR_VALID|CPF_MSG_PRESENT)) != 0 )
    {
	blue  = GETBIT( flags, CPB_BLUE );
	color = GETBIT( node->incolor, chan );

	is_dup = (color != blue);
    }

#ifdef DEBUG
    if ( is_dup )
    {
	bsdlog( LOG_DEBUG, "[%s:%s] msg is dup: %u != %u on slot %u, "
	        "incolor = 0x%08lx\n",
	        PROCESS_NAME, __FUNCTION__,
	        color, blue, chan, (unsigned long) node->incolor );
    }
#endif /* DEBUG */

    return is_dup;
}
