/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, romanick@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: cccp_queue.c,v 1.8 2001/07/31 23:48:24 idr Exp $"

/*
 *   COMPONENT_NAME: LIB
 *
 *   FUNCTIONS: dfunc
 *		dumpQueue
 *		flushQueue
 *		traverseQueue
 */

/* PROJECT:	Cluster Lock Manager communications facility*/
/**/
/* MODULE:	queue handling*/
/**/
/* NOTES:	Queues are maintained as a singly-linked ring, with the head*/
/*		being an instance of the same structure _but not itself part*/
/*		of the ring_ with it's "next" pointer pointing to the _tail_*/
/*		of the queue.  Thus an empty queue consists of only the head*/
/*		pointing to itself.  A single-element queue consists of the*/
/*		head pointing to the only element, which points to itself.*/
/*		A multiple-element queue consists of the head, pointing to*/
/*		the last element, pointing to the first element, all the way*/
/*		back to the last element.*/
/**/
/*		The declarations in the header file use a lot of "void *" to*/
/*		maintain opacity; nobody outside of this module should ever*/
/*		know what a clm_q_t looks like.*/

#if defined( __KERNEL__ )
# ifndef MODULE
#  define MODULE
# endif

# include <linux/kernel.h>
# define __NO_VERSION__
# include <linux/module.h>
# ifdef USING_FULL_VERSIONING
#  if CONFIG_MODVERSIONS==1
#   define MODVERSIONS
#   include <linux/modversions.h>
#  endif
# endif /* USING_FULL_VERSIONING */

#else

# include <stdio.h>

#endif

#include <cltrc.h>
#include <dlm_cluster.h>
#include <cccp.h>

#include "cccp_pkt.h"
#include "cccp_queue.h"
#include "cccp_private.h"

/* FUNCTION:	traverseQueue*/
/**/
/* PURPOSE:	Go through all the elements in a queue, applying a given*/
/*		function to each.  This function sort of double duty, as*/
/*		it can be either destructive or non-destructive depending*/
/*		on how the "freeit" argument is set.*/
/**/
/* INPUT:	head		pointer to queue head*/
/*		fp		pointer to function to apply to queue members*/
/*		freeit		TRUE for destructive operation*/
/**/
/* OUTPUT:	none*/
/**/
/* NOTE:	Elements are only removed if the "fp" function returns TRUE*/
/*		(i.e. non-zero) even if destructive behaviour is specified.*/

static void
traverseQueue( dlm_list_t *head, queue_flush_func_t * fp, int freeit, 
	       uint prio )
{
    clm_msg_t      * msg;
    clm_msg_t      * next;


    msg = (clm_msg_t *) dlm_list_get_head( head );
    while( msg != NULL )
    {
	next = (clm_msg_t *) dlm_list_succ( (dlm_listnode_t *) msg );

	if (freeit) {
	    TRCHKL1T(CPTRC_DEQUEUE,msg);
	}

	/* if the priority of the current element is high enough */
	if (msg->cm_prio >= prio) {
	    /* free the one we just processed, if told to*/
	    if (freeit) {
		dlm_list_remove( (dlm_listnode_t *) msg );
	    }

	    /* run the function on this element*/
	    if ( !(*fp)( msg, msg->cm_packet, msg->cm_node_ptr ) ) {
		return;
	    }
	}

	msg = next;
    }
}

/* FUNCTION:	countQueue*/
/**/
/* PURPOSE:	Return a count of the number of items on the queue for */
/*		debugging purposes */
/**/
/* INPUT:	qp		pointer to queue head*/
/**/
/* OUTPUT:	none*/

unsigned
countQueue ( dlm_list_t *head )
{
    dlm_listnode_t *qp;	/* points to current element */
    unsigned	count = 0;

    qp = (dlm_listnode_t *) head;
    while ( (qp = dlm_list_succ( qp )) != NULL )
    {
	count++;
    }

    return count;
}

/* FUNCTION:	flushQueue*/
/**/
/* PURPOSE:	Retry all the messages/packets on the queue, until immediate*/
/*		delivery fails or the queue is exhausted.*/
/**/
/* INPUT:	qp		pointer to queue head*/
/*		fp		function to use for delivery*/
/**/
/* OUTPUT:	none*/

void
flushQueue ( dlm_list_t *qp, queue_flush_func_t * fp, uint prio )
{
    traverseQueue(qp,fp,TRUE, prio);
}

/* FUNCTION:	dfunc*/
/**/
/* PURPOSE:	Utility function passed to traverseQueue() by dumpQueue() to*/
/*		print out queue elements.*/
/**/
/* INPUT:	all the stuff that _would_ be passed to the delivery function*/
/*		if we were in flushQueue() rather than dumpQueue()*/
/**/
/* OUTPUT:	always TRUE so traverseQueue() doesn't terminate*/

static int
dfunc( clm_msg_t * msg, clm_pkt_t * pkt, void * node )
{
#if defined( __KERNEL__ )
    printk( "  <0x%lx>\n", (long) msg );
#else
    printf( "  <0x%lx>\n", (long) msg );
#endif

    return(TRUE);
}

/* FUNCTION:	dumpQueue*/
/**/
/* PURPOSE:	Dump the elements of the specified queue for debugging.*/
/**/
/* INPUT:	pointer to queue head*/
/**/
/* OUTPUT:	none*/

void
dumpQueue ( dlm_list_t *qp )
{
    traverseQueue(qp,dfunc,TRUE,0);
}
