/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, romanick@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: cccp_udp.c,v 1.14 2001/10/09 20:23:31 idr Exp $"

/*
 *   COMPONENT_NAME: LIB
 *
 *   FUNCTIONS: cccp_poll_loop
 *		cccp_ip_init
 *		readPacket
 */

#if defined( __KERNEL__ )
# ifndef MODULE
#  define MODULE
# endif

# include <linux/kernel.h>
# define __NO_VERSION__
# include <linux/module.h>
# ifdef USING_FULL_VERSIONING
#  if CONFIG_MODVERSIONS==1
#   define MODVERSIONS
#   include <linux/modversions.h>
#  endif
# endif /* USING_FULL_VERSIONING */

# include <linux/sched.h>
#else

# include <errno.h>
# include <signal.h>
# include <unistd.h>
# include <string.h>
# include <sys/poll.h>
# include <sys/uio.h>

# include <sys/time.h>

#endif

#include <cltrc.h>
#include <dlm_cluster.h>
#include <cccp.h>

#include "cccp_pkt.h"
#include "cccp_proto.h"
#include "cccp_private.h"

static const int bufferSize = 4096;
cccp_sem_t cccp_poll_sem;




/****************************************************************************/
/* cccp_ip_init - Initialize IP subsystem
 *
 * Initialize the IP subsystem by opening our socket etc.
 *
 * RETURNS
 *
 * If initialization is successful, CPERR_OK will be returned.  Otherwise a
 * CCCP error code will be returned.
 */

int
cccp_ip_init(
	      /* The local node ID.
	       */

	      uint id 
	    )
{
    cccp_sock_t         sock;
    struct sockaddr_in	addr;
    int                 err;

    addr.sin_family      = AF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;
    addr.sin_port        = id2server( id );


    /* First, create the socket.
     */

    CCCP_SOCK_CREATE( & sock, err );
    if ( err < 0 )
    {
	bsdlog( LOG_ERR, __FUNCTION__ ": unable to create server socket" );
	return(CPERR_NOSOCKET);
    }


    /* Now bind the socket.
     */

    err = CCCP_SOCK_BIND( sock, (struct sockaddr *) & addr, sizeof( addr ) );
    if (  err < 0 )
    {
	bsdlog( LOG_ERR, __FUNCTION__ ": unable to bind to server socket, "
	        "err = %u", -err );
	return(CPERR_CNBIND);
    }

    cccp_our_sock = sock;
    return(CPERR_OK);
}




/****************************************************************************/
/* readPacket - Read in a packet and run it through the protocol
 *
 * Reads in a packet and runs it through the protocol.  Most of the real work
 * is done in handlePacket, with this function just handling allocation etc.
 */

static clm_msg_t *
readPacket( __u8 * buf, int * err )
{
    struct sockaddr_in   addr;
    int                  bytes;
    clm_pkt_t          * pkt;
    uint                 id;
    struct iovec         io_vector;
    struct msghdr        msg_header;
    clm_msg_t          * msg;


    io_vector.iov_base = buf;
    io_vector.iov_len  = bufferSize;

    msg_header.msg_name = & addr;
    msg_header.msg_namelen = sizeof( struct sockaddr_in );
    msg_header.msg_iov = & io_vector;
    msg_header.msg_iovlen = 1;
    msg_header.msg_control = NULL;
    msg_header.msg_controllen = 0;
    msg_header.msg_flags = 0;

    bytes = CCCP_SOCK_RECVMSG( cccp_our_sock, & msg_header, bufferSize, 0 );
    if ( bytes < 0 )
    {
	/* The poll task gets hit with a signal and cccp_time_to_die gets
	 * set when it's time for CCCP to shutdown.  In this case, we should
	 * NOT log message that the recvmsg call failed.  Otherwise, log a
	 * message.
	 */

	if ( ((-bytes != EINTR) && (-bytes != ERESTARTSYS))
	     || !cccp_time_to_die )
	{
	    bsdlog( LOG_ERR, "Error %d on socket read\n", -bytes );
	}


	*err = -bytes;
	return NULL;
    }

    TRCHKL1T(CPTRC_READ,buf);

    ASSERT( msg_header.msg_namelen == sizeof( struct sockaddr_in ),
	    "address size mismatch" );


    *err = 0;
    pkt = (clm_pkt_t *)buf;
    id = pkt->cp_srcnode;


    /* Packets that have either a message payload or a redirect *MUST* have
     * a valid port number.  Other messages (i.e., standalone ACKs) will
     * have a port of 0.
     */

    if ( (pkt->cp_port == 0) 
	 && (pkt->cp_flags & (CPF_MSG_PRESENT | CPF_REDIR_VALID)) )
    {
	bsdlog( LOG_WARNING, "Message received for port 0 (from node %u).\n",
	        id );
	return( NULL );
    }


    SUGGEST((id != thisNode()),"received packet from ourself");
#if defined(SHOW_CONNECTIONS)
    bsdlog( LOG_DEBUG, "[%s:%s] %s:%d,%s reachable, size = %u/%lu, "
	    "type = %08lx\n",
	    PROCESS_NAME, __FUNCTION__,
	    in_ntoa( addr.sin_addr.s_addr ), addr.sin_port,
	    clm_node_reachable( id ) ? "" : " not",
	    (unsigned) pkt->cp_header_length, pkt->cp_length, pkt->cp_type );
    LOG_PACKET( pkt );
#endif


    /* don't allow reads from inactive nodes */
    msg = NULL;
    if ( clm_node_reachable( id ) && (id != thisNode()) ) 
    {
#if defined(STRESS)
	/* If stress-testing is compiled in, drop about 0.7% of the packets
	 * to simulate network errors.
	 */

	if ((random() % 1000) < 7) {
	    bsdlog( LOG_DEBUG,"DROPPING PACKET, flags = 0x%08x\n",
		   pkt->cp_flags );
	    return( NULL );
	}
#endif

	if ( bytes == (pkt->cp_length + pkt->cp_header_length) ) 
	{
	    msg = cccp_msg_unmarshal( pkt );
	    if ( msg == NULL )
	    {
		*err = ENOMEM;
	    }
	}
	else
	{
	    bsdlog( LOG_WARNING, "Message body size mis-match.  Got "
		    "%u, but expected %u (from node %u).",
		    bytes, 
		    (unsigned) (pkt->cp_length + pkt->cp_header_length), id );
	}
    }
#ifdef DEBUG
    else
    {
	bsdlog( LOG_DEBUG, "[%s:%s] dropping msg from %u (not connected)\n",
	        PROCESS_NAME, __FUNCTION__, id );
    }

#endif


    bsdlog( LOG_DEBUG, "[%s:%s] returning (clm_msg_t *) 0x%08lx\n",
	    PROCESS_NAME, __FUNCTION__, (unsigned long) msg );
    return( msg );
}




/****************************************************************************/
/* cccp_poll_loop -
 * 
 * Checks all file descriptors for new messages, assembles any needed
 * clm_msg_t structures, and invokes receive callbacks.  Currently called
 * explicitly, but will be called from the kproc in the kernel version.
 */

void *
cccp_poll_loop( void * arg )
{
    void       * buf;
    clm_msg_t  * msg;
    int          err;
#if !defined( __KERNEL__ )
    sigset_t        sigs;


    sigemptyset( & sigs );
    sigaddset( & sigs, SIGALRM );
    sigprocmask( SIG_BLOCK, & sigs, NULL );
#else
    MOD_INC_USE_COUNT;
    dlm_set_thread_name( current, "CCCP poll" );
#endif /* __KERNEL__ */


    TRCHKL0T(CPTRC_POLL_ENTRY);

    buf = NULL;
    msg = NULL;
    while( ! cccp_time_to_die )
    {
	CCCP_MUTEX_LOCK( & cccp_global_lock );

	/* make any necessary new connections*/
	openConnections();

	CCCP_MUTEX_UNLOCK( & cccp_global_lock );


	if ( buf == NULL )
	{
	    buf = CLM_MALLOC( bufferSize );
	}

	if ( buf != NULL )
	{
	    (void) memset( buf, 0, bufferSize );
	    msg = readPacket( buf, & err );
	    if ( msg != NULL )
	    {
		clm_pkt_t  * pkt;
		bool_t       status;


		pkt = (clm_pkt_t *) buf;

		CCCP_MUTEX_LOCK( & cccp_global_lock );
		status = handlePacket( pkt->cp_srcnode, msg, pkt );
		CCCP_MUTEX_UNLOCK( & cccp_global_lock );

		if ( ! status )
		{
		    cccp_msg_destroy( msg );
		    CLM_FREE( buf );
		}

		/* Either we just freed both of these buffers or handlePacket
		 * still owns them.  In either case, buf CANNOT be re-used.
		 */

		msg = NULL;
		buf = NULL;
	    }
	    else if ( err != 0 )
	    {
		cccp_time_to_die = TRUE;
	    }
	}
    }


    if ( buf != NULL )
    {
	CLM_FREE( buf );
	buf = NULL;
    }


    /* Close our socket so that we won't receive any more incoming
     * messages.  This *MUST* be done here instead of in cccp_stop because
     * when cccp_stop is called, this thread may be blocked reading data from
     * the socket.  If the socket is closed at that time, bad things (i.e., an
     * oops on Linux) will happen.
     */

    CCCP_SOCK_CLOSE( cccp_our_sock );
    cccp_our_sock = CCCP_SOCK_INVALID;


#ifdef __KERNEL__
    MOD_DEC_USE_COUNT;
#endif
    bsdlog( LOG_INFO, "%s termination, CCCP shutdown\n", __FUNCTION__  );
    return( NULL );
}




/****************************************************************************/
/* cccp_poll_timer_handler - Handle timer interrupts
 *
 * Each time the timer interrupt fires, this routine will execute.  The
 * timer is set to trigger message retransmitts (after the timeout period).
 * The sole purpose of this function is to "wake-up" the retransmit loop
 * (see cccp_retransmit_loop).
 */

void
cccp_poll_timer_handler( unsigned timer )
{
    (void) CCCP_SEM_UP( & cccp_poll_sem );
    cccp_timer_release( timer );
}




/****************************************************************************/
/* cccp_retransmit_loop - Retransmit messages that have timed-out
 * 
 * Waits on the cccp_poll_sem semaphore and retransmits any out-bound
 * messages that have timed-out.  That is, if an ACK has not been received
 * for a message within the timeout period, it will be retransmitted.
 */

void *
cccp_retransmit_loop( void * arg )
{
    unsigned long   next_timeout;

#ifdef __KERNEL__
    dlm_set_thread_name( current, "CCCP retrans" );
    MOD_INC_USE_COUNT;
#endif

    next_timeout = 0;
    while( ! cccp_time_to_die )
    {
	CCCP_SEM_DOWN( & cccp_poll_sem );

	CCCP_MUTEX_LOCK( & cccp_global_lock );
	if ( ! cccp_time_to_die )
	{
	    /* make any necessary new connections*/
	    openConnections();

	    /* handle ack timeouts*/
	    next_timeout = doAckTimeouts();

	    if ( next_timeout != 0 )
	    {
		cccp_set_timer( next_timeout );
	    }
	}

	CCCP_MUTEX_UNLOCK( & cccp_global_lock );

#ifdef DEBUG
	if ( next_timeout < 4 )
	{
	    bsdlog( LOG_DEBUG, "next_timeout = %lu", next_timeout );
	}
#endif
    }


#ifdef __KERNEL__
    MOD_DEC_USE_COUNT;
#endif
    bsdlog( LOG_INFO, __FUNCTION__ " termination, CCCP shutdown." );
    return( NULL );
}
