/*
 * Cluster interface modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_heartbeat.c,v 1.6 2001/11/29 21:46:27 idr Exp $"

/* ============================================================ */
/*
 * Distributed Lock Manager for Linux -
 *  Heartbeat cluster manager interface
 *
 * This section is relatively generic, in that all modules
 * the DLM will work with will essentially work this way.
 *
 * Provides setup and subscription services for dealing with
 * the hearbeat clustering support.  Uses the Heartbeat API
 * to:
 * - intialize a connection (ll_cluster_new() etc.)
 * - subscribe for node and comm events (set_nstatus_callback(), etc.)
 * - handle asynchronous notification (subscribe callback)
 * - inform the mainline DLM user-space code events have
 *  arrived.
 *
 * This module is dynamically loaded by the mainline code, so
 * needs to provide a "handle" that is used by the mainline
 * code to tell us to get going.  We do that through the
 * cluster_interface_handling() function.
 *
 * If the mainline code decides we need to shut down, it
 * should call the cluster_interface_shutdown() function.
 *
 * The execution model is as follows:
 * - mainline code dynamically loads this 'library'.
 * - it creates a thread and that thread calls into
 *  cluster_interface_handling().
 * - it sets up the group services connection.
 * - it then basically waits for notifications.
 * - when events arrive, they are packaged using
 *  dlm_node_t blocks, and passed into the mainline code
 *  using:
 *  - dlm_notify_begin(void);
 *  - dlm_notify_deliver(dlm_node_t *info);
 *  - dlm_notify_end(void);
 *
 * Any errors that cause us to lose the connection to
 * the clustering code (group services in this case) causes
 * the thread to return to the mainline code, which signals
 * an error.  Normally it will deal with it by shutting
 * down the DLM on this node.
 *
 * NOTE: for now, just do direct linkage, although the
 * routines and their operations will be the same.
 */
/* ============================================================ */

#include <pthread.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <sys/utsname.h>
#include <sys/syslog.h>

#include <heartbeat.h>                  /* Heartbeat api info */
#include <hb_api_core.h>
#include <hb_api.h>

#include "dlm.h"                        
#include "dlm_devmsg.h"                 /* Messages to dlm */

/* ============================================================ */

static ll_cluster_t * clust = NULL;
static void node_status(const char * node_name,
                        const char * status, 
                        void * priv );
static int find_local( void );
static void * monitor_cluster(void);
static int init_clust(void );

static dlm_node_array_t *configured_nodes = NULL;
static pthread_t listening_thread;
static struct utsname uts_info;
static int need_to_quit = 0;

/* ============================================================ */
/*
 * The public routine.  Thread comes in, does not leave unless
 * some fatal error occurs.
 *
 * Input: pointer to configured nodes block.
 * Output: nothing, return itself is bad enough news.
 *
 * We use the configured nodes block to determine that the
 * nodes that the DLM cares about show up in the notificaitons,
 * as it is possible that the DLM is running on only a subset of
 * cluster nodes, likewise, the DLM may be configured for more
 * nodes than are clustered.  A key is that if we never get an
 * event showing a node coming up, the DLM will never try to
 * allow that node into the locking setup.
 *
 * Note that if the admin misconfigured the node numbers or
 * names that Heartbeat uses and that the DLM uses, bad things
 * will happen, but there is nothing we can do about that.
 */
void *cluster_interface_handling(void *arg)
{
  int rc;
  configured_nodes = (dlm_node_array_t *)arg;

  (void) _heartbeat_h_Id;
  (void) _ha_msg_h_Id;

  /*
   * Initialize with Heartbeat.  It will loop forever, or,
   * until the main routine gets bored...  The main routine
   * will not allow the DLM function to open up until we 
   * start sending notifications, which require us to get 
   * past the initialization.
   */

  listening_thread = pthread_self();

  printf("2nd thread, pid/tid [%d/%ld]\n",
         getpid(),
         listening_thread);

  rc = init_clust();

  if (HA_OK != rc) {
    fprintf(stderr,
            "Error on Heartbeat init [%d]\n",
            rc);
    return(NULL);
  }

  /*
   * Get the identity of the local node and tell the mainline.
   */

  rc = find_local();

  if (-1 == rc) {
    fprintf(stderr,
            "Cannot find local node info!\n");
    return(NULL);
  } else {
    fprintf(stderr,
            "Local node is [%s]\n",
            uts_info.nodename);
  }

  /*
   * Collect initial node/interface status info and subscribe
   * for updates.  Control will stay in that routine until we
   * are told to quit.
   */

  monitor_cluster();

  return(NULL);
}


/* ------------------------------------------------------------ */
/*
 * Shutting down, cleanly quiesce the cluster interface.
 * Quiesce the listening thread, and it will shutdown the
 * Heartbeat interface.
 *
 * This function is called from the mainline code when it
 * wants to shut down cleanly.
 */
void *cluster_interface_shutdown(void)
{
  if (0 != listening_thread) {
    need_to_quit = 1;
    pthread_kill(listening_thread, SIGKILL);
  }
  listening_thread = 0;
  return(NULL);
}

/* ============================================================ */
/*
 * Callback for each change in node status.  Report to the
 * mainline.
 */
void
node_status( const char * node_name,
             const char * status,
             void * priv )
{
  haDLM_state_t state;
  dlm_node_t *node;

  if (0 == strcmp( status, ACTIVESTATUS )) {
    state = haDLM_state_up;
  } else if (0 == strcmp( status, DEADSTATUS )) {
    state = haDLM_state_down;
  } else {
    return;                             /* Nothing to do. */
  }

  node = (dlm_node_t *)malloc(sizeof(dlm_node_t));
  if (NULL == node) {
    assert(!"Cannot allocate local node block!");
  }

  /*
   * If the cluster code does NOT report node number, then
   * this field must be set to -1!  If it does not report
   * node name, then that field must be set to null in the
   * first character.  If it does not report a network address,
   * then set dlm_comm_state to haDLM_state_null and the
   * dlm_node_address to zero.
   */
  node->dlm_node_state = state;
  node->dlm_node_number = -1;

  strncpy(&node->dlm_node_name[0], node_name, MAX_NODENAME_LEN-1);
  node->dlm_node_name[MAX_NODENAME_LEN-1] = '\0';

  node->dlm_comm_state = haDLM_state_null;
  node->dlm_node_address.s_addr = 0;

  dlm_notify_begin();
  dlm_notify_deliver(node);
  dlm_notify_end();

  syslog( LOG_INFO, "node %s is %s", node_name, status );
}

/* ============================================================ */
/*
 * First, run the list of nodes and deliver the current status
 * to the mainline.  Then sit on the Heartbeat message queue
 * and wait for status updates to arrive.  As they come in,
 * process them and see if they're for "our" nodes/interfaces.
 */
void *
monitor_cluster(void)
{
  struct ha_msg * reply;
  int   err;
  const char * node_name;
  const char * status;
  const char * type;
  const char * orig;
  haDLM_state_t state;
  dlm_node_t *node;

  /*
   * Set callback to deal with future node events.
   */
  err = clust->llc_ops->set_nstatus_callback( clust, node_status, NULL );
  if (HA_OK != err) {
    assert(!"could not set node-status callback" );
  }

  err = clust->llc_ops->init_nodewalk( clust );
  if (HA_OK != err) {
    assert(!"could not init nodewalk" );
  }

  /*
   * Collect initial (from our POV) status of all nodes.
   */
  dlm_notify_begin();
  while( (node_name = clust->llc_ops->nextnode( clust )) != NULL ) {
    status = clust->llc_ops->node_status( clust, node_name );

    if ( 0 == strcmp( status, ACTIVESTATUS ) ) {
      state = haDLM_state_up;
    } else if (0 == strcmp( status, DEADSTATUS )) {
      state = haDLM_state_down;
    } else {
      continue;                         /* Nothing to do. */
    }

    node = (dlm_node_t *)malloc(sizeof(dlm_node_t));
    if (NULL == node) {
      assert(!"Cannot allocate local node block!");
    }
    node->dlm_node_state = state;
    node->dlm_node_number = -1;

    strncpy(&node->dlm_node_name[0], node_name, MAX_NODENAME_LEN-1);
    node->dlm_node_name[MAX_NODENAME_LEN-1] = '\0';

    node->dlm_comm_state = haDLM_state_null;
    node->dlm_node_address.s_addr = 0;

    dlm_notify_deliver(node);
  }
  dlm_notify_end();
  clust->llc_ops->end_nodewalk( clust );

  /*
   * Simply wait for things to happen.  The callback will be
   * called when events arrive.
   */
  pthread_testcancel();
  while (!need_to_quit &&
         (NULL != (reply = clust->llc_ops->readmsg( clust, 1 )))) {
    pthread_testcancel();

    if ((type = ha_msg_value(reply, F_TYPE)) == NULL) {
      type = "?";
    }
    if ((orig = ha_msg_value(reply, F_ORIG)) == NULL) {
      orig = "?";
    }
	
    syslog( LOG_DEBUG, "Got a message of type [%s] from [%s]\n",
            type, orig );

    ha_log_message(reply);
    ha_msg_del(reply); 

    reply = NULL;
  }

  syslog( LOG_INFO, __FUNCTION__ " terminating" );
  return( NULL );
}

/* ============================================================ */
/*
 * Establish a persistent connection to Heartbeat.  Not much
 * more to do once that is done.
 */
int
init_clust( void )
{
  int   err;
  int   init_tries = 0;

  for(err = HA_FAIL;
      ((HA_FAIL == err) && (10 > init_tries));
      init_tries++) {

    if ( clust == NULL ) {
      clust = ll_cluster_new( "heartbeat" );
      if (NULL == clust) {
        fprintf(stderr,
                "Init try %d, cannot allocate Heartbeat cluster handle\n",
                init_tries);
      } else {
        err = clust->llc_ops->signon( clust, NULL );
        if (HA_OK != err) {
          fprintf(stderr,
                  "Init try %d, cannot signon to Heartbeat cluster handle\n",
                  init_tries);
        }
      }
    }
    if (HA_OK == err) break;
    sleep(10);
  }
  if (HA_OK != err) {
    assert(!"could not connect to Heartbeat cluster manager" );
  }

  return(err);
}

/* ------------------------------------------------------------ */
/*
 * Determine the local node.  Get the name from uname, then do
 * a request to ensure that Heartbeat recognizes the name.
 * At this time, we don't care about the other nodes in the
 * config.
 *
 * If not found, we are in bad shape, assert.
 */
int
find_local( void )
{
  int   err;
  dlm_node_t *node;
  const char * state;

  err = uname(&uts_info);
  if (0 != err) {
    fprintf(stderr,
            "Error on uname, %d [%s]\n",
            errno,
            strerror(errno));
    assert(!"cannot get node info from uname");
  }

  state = clust->llc_ops->node_status( clust, uts_info.nodename );
  if (NULL == state) {
    fprintf(stderr,
            "Cannot get status for node %s from Heartbeat\n",
            uts_info.nodename);
    assert(!"Cannot get node status");
  }

  node = (dlm_node_t *)malloc(sizeof(dlm_node_t));
  if (NULL == node) {
    fprintf(stderr,
            "Cannot allocate local node block!\n");
    assert(!"Cannot allocate local node block!");
  }
  node->dlm_node_state = haDLM_state_local;
  node->dlm_node_number = -1;
  strcpy(&node->dlm_node_name[0], uts_info.nodename);
  node->dlm_comm_state = haDLM_state_null;
  node->dlm_node_address.s_addr = 0;

  dlm_notify_begin();
  dlm_notify_deliver(node);
  dlm_notify_end();

  return(HA_OK);
}

