/*
 * Cluster interface modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_rsct.c,v 1.2 2001/02/20 02:39:36 wombat Exp $"

/* ============================================================ */
/*
 * Distributed Lock Manager for Linux -
 *  RSCT cluster manager interface
 *
 * This section is relatively generic, in that all modules
 * the DLM will work with will essentially work this way.
 *
 * Provides setup and subscription services for dealing with
 * the RSCT clustering support.  Uses the Group Services API
 * to:
 * - intialize a connection (ha_gs_init())
 * - subscribe for node and comm events (ha_gs_subscribe())
 * - handle asynchronous notification (subscribe callback)
 * - inform the mainline DLM user-space code events have
 *  arrived.
 *
 * This module is dynamically loaded by the mainline code, so
 * needs to provide a "handle" that is used by the mainline
 * code to tell us to get going.  We do that through the
 * cluster_interface_handling() function.
 *
 * If the mainline code decides we need to shut down, it
 * should call the cluster_interface_shutdown() function.
 *
 * The execution model is as follows:
 * - mainline code dynamically loads this 'library'.
 * - it creates a thread and that thread calls into
 *  cluster_interface_handling().
 * - it sets up the group services connection.
 * - it then basically waits for notifications.
 * - when events arrive, they are packaged using
 *  dlm_node_t blocks, and passed into the mainline code
 *  using:
 *  - dlm_notify_begin(void);
 *  - dlm_notify_deliver(dlm_node_t *info);
 *  - dlm_notify_end(void);
 *
 * Any errors that cause us to lose the connection to
 * the clustering code (group services in this case) causes
 * the thread to return to the mainline code, which signals
 * an error.  Normally it will deal with it by shutting
 * down the DLM on this node.
 *
 * NOTE: for now, just do direct linkage, although the
 * routines and their operations will be the same.
 */

/* ------------------------------------------------------------ */

/*
 * Using RSCT:
 *
 * For RSCT interactions, the above mentioned routines are
 * the ones we use, to repeat:
 * - intialize a connection (ha_gs_init())
 * - subscribe for node and comm events (ha_gs_subscribe())
 * - handle asynchronous notification (subscribe callback)
 * - inform the mainline DLM user-space code events have
 *  arrived.
 *
 * In this discussion if we use RSCT generically, we are
 * strictly speaking using the only RSCT API we use directly
 * is the Group Services API (GSAPI).
 *
 * RSCT works on an asynchronous notification model, we, the
 * client, connect to it, register one or more subscriptions,
 * and then sit back and wait to be notified of iteresting
 * things happening.
 *
 * We use only a small, very, very small, portion of the GSAPI,
 * since we do not care about forming a peer group.  We use
 * the subscription feature - ha_gs_subscribe() - and register
 * for the following information:
 * - node membership: i.e., node up and down events for all
 *  nodes in the cluster.
 * - adapter membership: i.e., communications adapter up and
 *  down events for all adapters in the cluster.
 *
 * The former is easy.  The latter has some interesting
 * questions to it:
 * - RSCT only reports for the comm adapters which it has been
 *  told to monitor.
 *  - the DLM cares about the IP address configured to it for
 *      each node.
 *  - this IP address may not have been given to RSCT, so may
 *      never show up.
 *  - RSCT can give us all IP addresses that are aliased to a
 *      comm adapter, so we may see the IP address that way.
 * - Other clustering packages may not even give us this level
 *  of information.
 * Bottom line for immediate use, we will not subscribe for
 * adapter information until later.
 *
 * An additional question here, RSCT can monitor more than just
 * IP (i.e, network) links, it can also monitor serial links,
 * target-mode comm links over disk adapters, in fact, almost
 * literally tin-cans and string.  But, the subscription model
 * that RSCT (GSAPI) presents for this data is rather poor (and,
 * I must admit it, I designed it...):
 * - the subscription is by network "type", i.e., subscribe "for
 *  all ethernet adapters" or "for all serial links".
 * - the client (us) must then parse through ALL events for ALL
 *  adapters for the ones we care about.
 *
 * On the DLM side, right now it basically has an IP address,
 * and doesn't much care what it is on.  Also, the cccp library
 * can only use IP networks, so doesn't much care about the
 * serial or tin-can links.  But, if it gives that IP address
 * to this module, the only way to possibly find the address
 * is to subscribe for all possible IP network type, and dig
 * through (and it may still not show up.)
 *
 * Bottom-bottom line.  Given this - that the info exists but
 * we don't quite know how to deal with it - and questions
 * about other clustering packages, we will deal ONLY with
 * NODE failure events for now.  This means that the DLM will
 * still recover from node failures, and will be able to
 * reconfig when a node boots, but it will not be able to
 * deal with an IP adapter failure, even if the node has a
 * redundant adapter.  Note though that the code will remain
 * in the DLM to do this, we'll get to it later.
 */
/* ============================================================ */

#include <pthread.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <ha_gs.h>                      /* GSAPI info */
#include "dlm_devmsg.h"

/* ============================================================ */
/*
 * Set up parameters for initialization.
 */
static  ha_gs_responsiveness_t responsiveness = {HA_GS_PING_RESPONSIVENESS,
                                                 600,
                                                 10,
                                                 (char *)0,
                                                 0};

/*
 * Callbacks for various notifications.
 */
static void query_cb(const ha_gs_query_notification_t *note);
static void delayed_error_cb(const ha_gs_delayed_error_notification_t *note);
static ha_gs_callback_rc_t responsive_cb(const ha_gs_responsiveness_notification_t *note);
static void subscribe_cb_HostMbr(const ha_gs_subscription_notification_t *note);
#ifdef DOING_ENET_MEMBERSHIP
static void subscribe_cb_EnetMbr(const ha_gs_subscription_notification_t *note);
#endif /* DOING_ENET_MEMBERSHIP */

/*
 * Forward declarations of working functions.
 */
static int init_gs(void);
static int find_local(int *nodenum);
static int subscribe(char *group,
                     void (callback)(const ha_gs_subscription_notification_t *note),
                     ha_gs_token_t *token);

/*
 * Keep track of RSCT connection.
 */
static int gs_filedes = -1;
static ha_gs_token_t host_subscription = -1;
#ifdef DOING_ENET_MEMBERSHIP
static ha_gs_token_t enet_subscription = -1;
#endif /* DOING_ENET_MEMBERSHIP */

static dlm_node_array_t *configured_nodes = NULL;
static int our_node = -1;

/* ============================================================ */
/*
 * The public routine.  Thread comes in, does not leave unless
 * some fatal error occurs.
 *
 * Input: pointer to configured nodes block.
 * Output: nothing, return itself is bad enough news.
 *
 * We use the configured nodes block to determine that the
 * nodes that the DLM cares about show up in the notificaitons,
 * as it is possible that the DLM is running on only a subset of
 * cluster nodes, likewise, the DLM may be configured for more
 * nodes than are clustered.  A key is that if we never get an
 * event showing a node coming up, the DLM will never try to
 * allow that node into the locking setup.
 *
 * Note that if the admin misconfigured the node numbers that
 * RSCT uses and that the DLM uses, bad things will happen,
 * but there is nothing we can do about that.
 */
void *cluster_interface_handling(void *arg)
{
  ha_gs_rc_t rc;
  configured_nodes = (dlm_node_array_t *)arg;

  /*
   * Initialize with RSCT.  It will loop forever, or, until
   * the main routine gets bored...  The main routine will
   * not allow the DLM function to open up until we start
   * sending notifications, which require us to get past
   * the initialization.
   */

  rc = init_gs();

  if (HA_GS_OK != rc) {
    fprintf(stderr,
            "Error on RSCT init [%d]\n",
            rc);
    return(NULL);
  }

  /*
   * Get the identity of the local node and tell the mainline.
   */

  rc = find_local(&our_node);

  if (-1 == rc) {
    fprintf(stderr,
            "Cannot find local node info!\n");
    return(NULL);
  } else {
    fprintf(stderr,
            "Local node is [%d]\n",
            our_node);
  }

  /*
   * Subscribe for host membership.
   */

  rc = subscribe(HA_GS_HOST_MEMBERSHIP_GROUP,
                 subscribe_cb_HostMbr,
                 &host_subscription);

  if (HA_GS_OK != rc) {
    fprintf(stderr,
            "Error on Host subscription [%d]\n",
            rc);
    return(NULL);
  }

#ifdef DOING_ENET_MEMBERSHIP
  /*
   * Subscribe for host membership.
   */

  rc = subscribe(HA_GS_ENET_MEMBERSHIP_GROUP,
                 subscribe_cb_EnetMbr,
                 &enet_subscription);

  if (HA_GS_OK != rc) {
    fprintf(stderr,
            "Error on Enet subscription [%d]\n",
            rc);
    return(NULL);
  }
#endif /* DOING_ENET_MEMBERSHIP */

  /*
   * Simply dedicate the thread to ha_gs_dispatch().
   * It will only return on error or shutdown.
   */

  rc = ha_gs_dispatch(HA_GS_BLOCKING);

  if (HA_GS_OK != rc) {
    fprintf(stderr,
            "Error in dispatch [%d]\n",
            rc);
    return(NULL);
  }

  return(NULL);
}


/* ------------------------------------------------------------ */
/*
 * Shutting down, cleanly quiesce the cluster interface.
 * For RSCT, this means ha_gs_quit() and not much more.
 * This will cause the thread to return from dispatch,
 * but without signalling an error.
 *
 * This function is called from the mainline code when it
 * wants to shut down cleanly.
 */
void *cluster_interface_shutdown(void)
{
  ha_gs_rc_t rc;

  rc = ha_gs_quit();

  fprintf(stderr,
          "Issued RSCT quit, return code [%d]\n",
          rc);

  return(NULL);
}

/* ============================================================ */

int init_gs(void)
{
  ha_gs_rc_t rc;
  int init_tries;

  /*
   * Attempt to initialize with Group Services.
   */
  rc = HA_GS_NOT_OK;
  init_tries = 0;

  while(rc != HA_GS_OK) {

    rc = ha_gs_init(&gs_filedes,
                    HA_GS_SOCKET_NO_SIGNAL,
                    &responsiveness,
                    NULL,
                    responsive_cb,
                    delayed_error_cb,
                    query_cb);

    if (HA_GS_OK == rc) {
      fprintf(stderr,
              "Connected to group services, socket fd [%d]\n",
              gs_filedes);
      break;
    } else if (HA_GS_EXISTS == rc) {
      fprintf(stderr,
              "You have already initialized!  Why do it again?");
      rc = HA_GS_OK;
      break;
    } else if (HA_GS_CONNECT_FAILED == rc) {
      /*
       * Perhaps GS has not yet started, loop and try again.
       */
      init_tries++;
      sleep(10);
    } else {
      fprintf(stderr,
              "Fatal error trying to init group services [%d]\n",
              rc);
      break;
    }
  }

  return(rc);
}

/* ------------------------------------------------------------ */
/*
 * Determine the local node number, and inform the mainline
 * code.
 */
int find_local(int *nodenum)
{
  ha_gs_rc_t rc;
  dlm_node_t *node;

  rc = ha_gs_get_node_number(nodenum);

  if (HA_GS_OK != rc) {
    fprintf(stderr,
            "Cannot get local node number, error [%d]\n",
            rc);
    *nodenum = -1;
    return(-1);
  }

  node = (dlm_node_t *)malloc(sizeof(dlm_node_t));
  if (NULL == node) {
    fprintf(stderr,
            "Cannot allocate local node block!\n");
    return(-1);
  }
  node->dlm_node_state = haDLM_state_local;
  node->dlm_node_number = *nodenum;
  node->dlm_node_name[0] = '\0';
  node->dlm_comm_state = haDLM_state_null;
  node->dlm_node_address.s_addr = 0;

  dlm_notify_begin();
  dlm_notify_deliver(node);
  dlm_notify_end();

  return(HA_GS_OK);
}

/* ------------------------------------------------------------ */
/*
 * Subscribe to the given group name, with the specified callback.
 * If all goes well, put the returned token into the given slot.
 *
 * Note that success here means ONLY that GS has accepted the
 * request, we have to get a callback to determine complete
 * success.
 */
int subscribe(char *group,
              void (callback)(const ha_gs_subscription_notification_t *note),
              ha_gs_token_t *token)
{
  ha_gs_rc_t rc;
  ha_gs_proposal_info_t proposal;

  proposal.gs_subscribe_request.gs_subscription_control =
    HA_GS_SUBSCRIBE_ALL_MEMBERSHIP;
  proposal.gs_subscribe_request.gs_subscription_group = group;
  proposal.gs_subscribe_request.gs_subscription_callback = callback;

  rc = ha_gs_subscribe(token,
                       &proposal);

  fprintf(stderr,
          "Subscribe to group [%s] rc [%d] token [%d]\n",
          proposal.gs_subscribe_request.gs_subscription_group,
          rc,
          *token);

  return(rc);
}

/* ============================================================ */
/*
 * A delayed error means that our subscription request has been
 * denied.  Rare, but could happen in various odd situations
 * where either:
 * - cluster not yet stable, or
 * - all adapters of the specified type are down, or
 * - the cluster has NO adapters of the specified type
 */
void delayed_error_cb(const ha_gs_delayed_error_notification_t *note)
{
  ha_gs_rc_t rc;
  ha_gs_token_t note_token;

  note_token = note->gs_request_token;

  if (host_subscription == note_token) {
    /*
     * Cluster not stable, need to reissue subscribe request.
     * But, best to do after a few seconds...
     */
    sleep(5);
    rc = subscribe(HA_GS_HOST_MEMBERSHIP_GROUP,
                   subscribe_cb_HostMbr,
                   &host_subscription);

    if (HA_GS_OK != rc) {
      fprintf(stderr,
              "Error on Host subscription [%d]\n",
              rc);
    }
    return;
  }

#ifdef DOING_ENET_MEMBERSHIP
  if (enet_subscription == note_token) {
    /*
     * Cluster not stable, need to reissue subscribe request.
     * But, best to do after a few seconds...
     */
    sleep(5);
    rc = subscribe(HA_GS_ENET_MEMBERSHIP_GROUP,
                   subscribe_cb_EnetMbr,
                   &enet_subscription);

    if (HA_GS_OK != rc) {
      fprintf(stderr,
              "Error on enet subscription [%d]\n",
              rc);
    }
    return;
  }
#endif /* DOING_ENET_MEMBERSHIP */

  return;
}

/* ------------------------------------------------------------ */
/*
 * Nodes have come up, or gone down.  GS will never combine
 * both up and down in one notification.  Go through the data
 * and package it up for calling the mainline code with the
 * list of affected nodes.
 */
void subscribe_cb_HostMbr(const ha_gs_subscription_notification_t *note)
{
  ha_gs_membership_t *membership;
  ha_gs_provider_t   *provider;
  dlm_node_t *node;
  int count, i;

  if ((HA_GS_SUBSCRIPTION_DELTA_JOIN & note->gs_subscription_type) ||
      (HA_GS_SUBSCRIPTION_DELTA_LEAVE & note->gs_subscription_type)) {
    membership = note->gs_changing_membership;
    count = membership->gs_count;

    if (0 == count) {
      fprintf(stderr,
              "HostMbr callback, but no nodes included?\n");
      return;
    }

    /*
     * Yes, each node is represented by a provider block...
     */
    dlm_notify_begin();
    provider = membership->gs_providers;
    
    for (i = 0; i < membership->gs_count; i++) {
      node = (dlm_node_t *)malloc(sizeof(dlm_node_t));
      if (NULL == node) {
        fprintf(stderr,
                "Cannot allocate local node block!\n");
        exit(65);
      }
      if (HA_GS_SUBSCRIPTION_DELTA_JOIN & note->gs_subscription_type) {
        node->dlm_node_state = haDLM_state_up;
      } else {
        node->dlm_node_state = haDLM_state_down;
      }
      /*
       * If the cluster code does NOT report node number, then
       * this field must be set to -1!  If it does not report
       * node name, then that field must be set to null in the
       * first character.
       */
      node->dlm_node_number = provider->gs_node_number;
      node->dlm_node_name[0] = '\0';
      node->dlm_comm_state = haDLM_state_null;
      node->dlm_node_address.s_addr = 0;

      dlm_notify_deliver(node);

      provider++;
    }
    dlm_notify_end();
  }

  return;
}

/* ------------------------------------------------------------ */
/*
 * Will look similar to above, only will deal with adapter
 * portion.
 */
#ifdef DOING_ENET_MEMBERSHIP
void subscribe_cb_EnetMbr(const ha_gs_subscription_notification_t *note)
{
  return;
}
#endif /* DOING_ENET_MEMBERSHIP */

/* ------------------------------------------------------------ */
/*
 * Called by GS on whatever time basis we specified, just to
 * be sure we are still working ok.  Just say so.
 */
ha_gs_callback_rc_t responsive_cb(const ha_gs_responsiveness_notification_t *note)
{
  fprintf(stderr,
          "Responsiveness check, we are ok\n");

  return(HA_GS_CALLBACK_OK);
}

/* ------------------------------------------------------------ */
/*
 * Why I defined this in the GSAPI, and why it is still there,
 * someone will have to tell me...
 */
void query_cb(const ha_gs_query_notification_t *note)
{
  return;
}
  
