/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: ast_kern.c,v 1.10 2001/11/28 14:12:26 bkon Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: ASTallocate
 *		ASTcall
 *		ASTcount
 *		ASTfree
 *		ASTget
 *		ASTpurge
 *		ASTregister
 *		k_ASTget
 *		kernapi_k_ASTget
 *		valid_handle
 */
/*
 *
 * A client registers a callback via clmlock(), at which time a handle
 * is set.  Information is stored in the handle until such a time as
 * the actual AST is registered.  The ast_handle token can be passed to an
 * arbitrary process. 
 * When that process wishes to cause the callback to occur in the client it 
 * invokes ASTcall. ASTcall registers the AST on behalf of the process, and
 * makes a notation in a shared table indicating 
 * that an AST is pending. The client must invoke ASTpoll to execute any 
 * pending callbacks.
 *
 * All pending ASTs are on the pending list for a given client
 * All allocated ASTs that have not fired are not on any list
 *
 * It is expected that only one AST will be delivered to a given slot
 * When an AST is to be executed, it is first removed from the pending list.
 * This allows another process to modify the pending list while an AST is
 * in progress.
 * 
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/stddef.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <asm/uaccess.h>

#include "dlm_kernel.h"
#include "dlm.h"
#include "clmstructs.h"
#include "ast_stub.h"
#include "cccp.h"

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */

static kmem_cache_t * ast_cache = NULL;

/* ============================================================ */

bool_t valid_handle( const struct ast_handle * hp );
static int ASTget (struct poll_info *user_pi, int clpid);

static int ASTregister(struct ast_handle *handlep);

__inline__ void ASTfree( void * p )
{
    kmem_cache_free( ast_cache, (p) );
}

#define ASTallocate() kmem_cache_alloc( ast_cache, 0 )

/* ------------------------------------------------------------ */
/*
 * Name
 *	ast_init
 * Description
 * 	Initialize AST variables
 * Parameters
 * 	None
 * Global Data Affected
 *	astmem
 *	ast_segcount
 * Returns
 *   	Nothing
 */
void
ast_init(void)
{
    ast_cache = kmem_cache_create( "DLM AST",
				   sizeof( struct astinfo ),
				   0, 0, NULL, NULL );
}


void
ast_cleanup( void )
{
    if ( ast_cache != NULL )
    {
	kmem_cache_destroy( ast_cache );
	ast_cache = NULL;
    }
}

/*
 * Name
 *	ASTregister
 * Description
 *	Allocate an AST slot for a callback. Create a client record is
 *	necessary. Initialize the allocated slot with the supplied
 *	information. The AST is not on any queue when this function returns
 * Parameters
 *	handlep:location through which ast_handle should be returned
 *
 * returns:
 *	0 for success
 *	negative for error
 */
static int
ASTregister(struct ast_handle *handlep)
{
    struct astinfo * a;
    int              status;


    status = -1;
    if (handlep != NULL) 
    {
	/* Grab an AST from the pool and initialize it.
	 */

	a = ASTallocate();
	if ( a != NULL )
	{
	    /* Set data in AST record.
	     */

	    a->ast_funcp = handlep->asth_funcp;
	    a->ast_arg = handlep->asth_arg;
	    a->ast_dest = handlep->asth_dest;

	    handlep->asth_aptr = a;

	    status = 0;
	}
    }

    return( status );
}




/****************************************************************************/
/* ASTcall_unlck - Queue and unlock AST
 *
 * Mark an unlock AST as pending by adding to the pending queue for the
 * specified client.  The application will invoke the function itself via
 * ASTpoll().  Update client information.
 *
 * RETURNS
 * 
 * On success, zero is returned.  If no astinfo structure could be allocated,
 * then ASTE_NOASTS is returned.
 * 
 * NOTES
 * 
 * This function assumes that the client structure is properly locked.
 */
int
ASTcall_unlck(
	       /* Pointer to the client to receive the AST.
		*/

	       clm_client_t *cl,


	       /* Status value to be passed to the AST.
		*/

	       dlm_stats_t status,


	       /* AST function pointer.
		*/

	       dlm_astfunc_t * ast_funcp,


	       /* Context argument pointer.
		*/

               void *arg,


	       /* Extra pointer value.
		*/

	       void *extrap
	     )
{
    struct astinfo *astp;

    /* grab an AST from the pool */
    astp = ASTallocate();
    if ( astp == NULL) 
    {
	return( ASTE_NOASTS );
    }

    /* set data in AST record */
    astp->ast_funcp = ast_funcp;
    astp->ast_arg = arg;
    astp->ast_dest = (struct lockstatus *)extrap;
    astp->ast_lksb.status = status;
    astp->ast_lksb.lockid = 0;	/* zero lockid indicates unlock AST */

    /* queue the AST on the client's pending queue */
    dlm_list_add_tail( & (cl->cli_ast_pnd), (dlm_listnode_t *) astp );

    return 0;
}




/****************************************************************************/
/* ASTcall - Mark an AST as pending and queue in a client
 *
 * Mark an AST as pending by adding to the pending queue for the appropriate
 * client.  If the flags in the AST record indicate a persistent AST, allocate
 * a new ast slot to handle this call.  The application will invoke the
 * function itself via ASTpoll().  Update client information.
 *
 * RETURNS
 * 
 * On success, zero is returned.  If the AST handle is invalid, ASTE_BADHANDLE
 * will be returned.  If an astinfo block cannot be allocated, ASTE_NOASTS
 * will be returned.
 */

int
ASTcall(
	 /* Pointer to the AST handle.
	  */

	 struct ast_handle * hp,

	
	 /* Lock status block to return.  May be NULL.
	  */

	 const struct lockstatus * lksbp,


	 /* Optional mode argument to pass to callback function.  Only used
	  * for BASTs.
	  */

	 int mode
       )
{
    struct astinfo * astp;
    clm_client_t   * clientp;


    if ( !valid_handle( hp ) || (hp->asth_client == NULL) )
    {
	return ASTE_BADHANDLE;
    }

    if ( ASTregister( hp ) < 0 )
    {
	return ASTE_NOASTS;
    }

    astp = hp->asth_aptr;
    clientp = hp->asth_client;

    CLM_ASSERT( astp != NULL, "ASTregister succeeded, but asth_aptr NULL!" );


    /* copy the lock status block to the location specified by the client */
    if ( lksbp != NULL )
    {
	(void) memcpy( & astp->ast_lksb, lksbp, sizeof( *lksbp ) );
    }
    else 
    {
	/* lockid of 0 means unlock AST.  Set to -1.  */
	astp->ast_lksb.lockid = -1;
    }

    /* pass back blocking mode (only used for blocking AST) */
    astp->ast_mode = mode;

    /* queue the AST on the client's pending queue */
    dlm_list_add_tail( & clientp->cli_ast_pnd, (dlm_listnode_t *) astp );

    return 0;
}

/*
 * Name
 *	ASTpurge
 * Description
 *	Purge all ASTs associated with the given
 *	owner
 * Arguments
 *	pid - process ID
 *	tid - thread ID
 * Returns
 *	The number of ASTs purged
 *	-1 if ASTs not initialized
 * Global Data Affected
 *	astmem->ast_table
 *	astmem->ast_clients
 *	astmem->ast_free
 */
int
ASTpurge(int pid, int tid)
{
    clm_client_t *client;
    dlm_listnode_t * astp;
    int npurged = 0;
    extern int OurSiteID;


    /* 
     * pick up client record 
     * Free any pending ASTs
     * Free any in-use ASTs
     */

    /* BUGBUG If findclientbypid locks the client structure for us, then we
     * BUGBUG need to unlock it after the loop.
     */

    if (NULL != (client = findclientbypid(OurSiteID, pid))) 
    {
	while ( (astp = dlm_list_rem_head( & client->cli_ast_pnd )) != NULL )
	{
	    ASTfree( astp );
	    npurged++;
	} 
    }

    return npurged;
}

/*
 * Name
 *	valid_handle
 * Description
 *	Validate an AST handle
 * Argument
 *	hp - pointer to handle
 * Returns
 *	1 if all data in handle is within expected ranges
 *	0 otherwise
 * Global Data Affected
 *	NONE
 */
bool_t
valid_handle( const struct ast_handle * hp )
{
    return ( (hp != NULL) && (hp->asth_pid > 0) && (hp->asth_funcp != NULL) );
}

/*
 * Function:
 *	k_ASTget
 *
 * Description:
 *	Wrapper for ASTget, which does all the real work.
 *
 * Arguments:
 *	Same as for ASTget.
 *
 * Return value:
 *	Same as for ASTget.
 *
 * Side effects:
 *	NONE
 */

int
k_ASTget (struct poll_info *user_pi, int clpid)
{
    extern int clm_active;
    int status;


    acquire_lock(LK_ENTRY, clpid);
    status = 0;

    /* if no lock manager, do nothing! */
    if ( clm_active )
    {
	att_seg() ;
	status = ASTget(user_pi, clpid);
	det_seg() ;
    }

    release_lock(LK_ENTRY, clpid);

    return(status);
}

/*
 * Function:
 *	kernapi_k_ASTget
 *
 * Description:
 *	Wrapper for k_ASTget, in order to meet rc and argument requirements
 *      of function passed to keventd thread, and to handle status from 
 *      k_ASTget.  Used only for kernel API clients.
 *
 * Arguments:
 *	none.
 *
 * Return value:
 *	NONE 
 *
 * Side effects:
 *	NONE
 */

void
kernapi_k_ASTget ()
{
        int status;

        status = k_ASTget(NULL, DLM_KERNEL_PID);
        if (0 > status) {
                bsdlog( LOG_WARNING, 
                       "[KAPI:%s] Error [%d] while handling AST for pid %d \n",
                        __FUNCTION__, status, DLM_KERNEL_PID);

        }

}

/*
 * Function:
 *	ASTget
 *
 * Description:
 *	Retrieve information about the next AST to fire for this client.
 *
 * Arguments:
 *	struct poll_info *user_pi;	-- location for AST info
 *  clpid -- pid of the client making request
 *
 * Return value:
 *	>0 if AST fetched
 *	 0 if no AST fetched
 *	<0 if error
 *
 * Side effects:
 *	NONE
 */

static int
ASTget (struct poll_info *user_pi, int clpid)
{
    clm_client_t		*client;	/* index to client record */
    struct astinfo *	astp;		/* ptr to first pending AST */
    struct poll_info 	pi;
    extern int		OurSiteID;

    /* pick up client record */

    /* BUGBUG If findclientbypid locks the client structure for us, then we
     * BUGBUG need to unlock before returning.
     */

    client = findclientbypid(OurSiteID, clpid);
    if ( NULL == client )
    {
        return(-1);
    }

    do
    {
         astp = (struct astinfo *) dlm_list_rem_head( & client->cli_ast_pnd );
  	 if (astp == NULL) 
	 {
	     return(0);
	 }


	/* BUGBUG This is the original commentary for this function.  However,
	 * BUGBUG I do not understand why this could even possibly be correct.
	 * BUGBUG It seems to me that this is a memory leak.
	 */

	/* This should never happen, but ignoring it is safest.  -jjd- */
	if (!astp->ast_funcp) 
	{
	    DBPRINTF((DBG_CTI, "Null AST func (0x%x)\n",astp)) ;
	    /* ignoring it may be safe, but freeing the pointer is not. -bg- */
	    /* ASTfree(astp); */
	}
    } while ( astp->ast_funcp == NULL );

    if (DLM_KERNEL_PID == clpid) {

        if (astp->ast_lksb.lockid == 0) {
            /* AST handling for unlocks */
            (*astp->ast_funcp)(astp->ast_arg, astp->ast_lksb.status, astp->ast_dest);
        } else {
	    /* AST handling for locks */
    	    if (astp->ast_dest)
	    {
                (void)memcpy(astp->ast_dest, &astp->ast_lksb,
                     (sizeof(struct lockstatus)-sizeof(unsigned int)));
            }
                (*astp->ast_funcp)(astp->ast_arg, astp->ast_mode, NULL);
            } 

    } else {
            if (astp->ast_lksb.lockid == 0) 
            {
                 /* AST handling for unlocks */
                pi.pi_type = AST_UNLOCK;
	        pi.pi_extrap = astp->ast_dest;
	        pi.pi_status = astp->ast_lksb.status;
            }
            else
            {
	        /* AST handling for locks */
	        pi.pi_type = AST_LOCK;

	        /* copy out the lock status structure, but stop short of the
	         * timeout value so we don't overwrite it. 
	         */

    	        if (astp->ast_dest)
	        {
	             copy_to_user(astp->ast_dest, &astp->ast_lksb, 
		            (sizeof(struct lockstatus)-sizeof(unsigned int)));
	        }

	        pi.pi_mode = astp->ast_mode;
            }

        pi.pi_args = astp->ast_arg;
        pi.pi_func = astp->ast_funcp;
        copy_to_user(user_pi, &pi, sizeof(struct poll_info));
    }

    ASTfree(astp);
    return(1);
}
