/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_alloc.c,v 1.8 2001/10/16 18:55:50 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		find_reclock
 *		free_le
 *		get_le
 *		growlocktab
 */

#ifndef lint
#endif
/*
 * This file implements an allocation pool for reclock structures.
 * 
 * This is now implemented using a Linux kmem_cache_t and a hash
 * table.  When a reclock is requested, if the ID is specified, then
 * that particular reclock will be returned.
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/signal.h>
#include <linux/types.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "clm_data.h"			/* external data definitions */
#include "dlm_kernel.h"
#include "cltrc.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */

struct rl_hash
{
    dlm_listnode_t   link;
    struct reclock * lock;
};

kmem_cache_t * dlm_reclock_cache = NULL;
kmem_cache_t * dlm_rlhash_cache = NULL;

/* BUGBUG Spin lock needed.  hash_insert and hash_find need an explicitly
 * BUGBUG stated lock policy and it must be followed by all callers.
 */
static dlm_list_t  rlh_table[ 16 ];

/* used to generate "gen" number for unique lock ids */

static __inline__ unsigned next_generation( void )
{
    static atomic_t   generation = ATOMIC_INIT( 0 );
    
    /* BUGBUG This isn't correct.  It would be possible for two processors
     * BUGBUG to get the same value for generation.
     */

    atomic_inc( & generation );
    return( atomic_read( & generation ) );
}


static __inline__ unsigned make_lock_id( unsigned site_id, unsigned gen )
{
    /* Give 13-bits for the site ID and use the remaining 19 bits for the
     * "generation" number.
     */
    
    return( (site_id & 0x00001fff) | ((gen & 0x0007ffff) << 13) );
}


#define HASH_CALC(id) ( ((id) >> 13) & 0x0f )

static void hash_insert( struct rl_hash * node )
{
    dlm_list_t * bucket;


    bucket = & rlh_table[ HASH_CALC( node->lock->lck.lockid ) ];
    dlm_list_add_head( bucket, (dlm_listnode_t *) node );
}


static struct rl_hash * hash_find( unsigned lock_id )
{
    dlm_list_t * bucket;
    struct rl_hash * curr;


    bucket = & rlh_table[ HASH_CALC( lock_id ) ];
    for ( EACH_DLM_LISTNODE( bucket, curr ) )
    {
	if ( curr->lock->lck.lockid == lock_id )
	{
	    return( curr );
	}
    }

    return( NULL );
}


/*
 * Name
 *	clm_alloc_init
 * Description
 * 	Init all static structures in clm_alloc
 * Arguments
 *	None
 * Returns
 *	Nothing
 * Global Data Affected
 *	freelocks
 */

void
clm_alloc_init(void)
{
    unsigned   i;
    int        err;
    static const char * const cache_names[] =
    {
	"DLM reclock", "DLM rl hash"
    };


    /* Initialize the hash buckets first so that dlm_destroy_all_le can be
     * called safely.
     */

    for ( i = 0 ; i < 16 ; i++ )
    {
	dlm_list_init( & rlh_table[ i ] );
    }

    clm_reclocks_used = 0 ;	/* reclock structs currently in use */
    clm_reclocks_denied = 0 ;	/* # lock denied - no more space */
    clm_reclocks_max = 0 ;	/* reclock structs allocd from lock_heap */


    err = 0;
    dlm_reclock_cache = kmem_cache_create( cache_names[ err ],
					   sizeof( struct reclock ),
					   0, 0, NULL, NULL );
    if ( dlm_reclock_cache != NULL )
    {
	err++;
	dlm_rlhash_cache = kmem_cache_create( cache_names[ err ],
					      sizeof( struct rl_hash ),
					      0, 0, NULL, NULL );
	if ( dlm_rlhash_cache != NULL )
	{
	    err = -1;
	}
    }


    if ( err != -1 )
    {
	dlm_destroy_all_le();
	bsdlog( LOG_CRIT, "[%s:%s] Unable to allocate %s cache\n",
	        current->comm, __FUNCTION__, cache_names[ err ] );
	BUG();
    }
}


void
dlm_destroy_all_le( void )
{
    dlm_list_t     * bucket;
    struct rl_hash * node;
    unsigned         i;


    for ( i = 0 ; i < 16 ; i++ )
    {
	bucket = & rlh_table[ i ];
	while ( (node = (struct rl_hash *) dlm_list_get_head( bucket )) 
	       != NULL )
	{
	    free_le( node->lock->lck.type, node->lock );
	}
    }

    if ( dlm_rlhash_cache != NULL )
    {
	kmem_cache_destroy( dlm_rlhash_cache );
	dlm_rlhash_cache = NULL;
    }

    if ( dlm_reclock_cache != NULL )
    {
	kmem_cache_destroy( dlm_reclock_cache );
	dlm_reclock_cache = NULL;
    }
}




/*
 * Name
 *	free_le
 * Description
 * 	Put a reclock back on the freelist for the indicated lock family
 * 	If we are the primary and the reclock was for a remote lock, 
 * 	Free the lockid hash entry
 * Arguments
 *	locktype - lock family
 *	a - the reclock to be freed
 * Returns
 *	Nothing
 * Global Data Affected
 *	freelocks
 */
void
free_le(dlm_restype_t locktype, struct reclock *a)
{
    unsigned int   id = a->lck.lockid;
    struct rl_hash   * node;


    cltrc(L_USR, CLLOCKD_HKID, T_ENTR_FUNC, T_CLLOCKD_FREE_LE, 0x00, 0x00, 0x00, 0x00);

    /*
     * Unix locks never call addmap so they never should call freemap
     */
    if (locktype != DLM_RES_UNIX)
    {
	if (a->lck.remlockid != 0 && a->lck.site != OurSiteID)
	    freemap(a->lck.remlockid, a->lck.lockid, a->lck.site) ;
    }

    /* free up any cancels that may be pending for this lock */
    trash_cancels(a);

    /*  Remove it from the client list. */
    delete_client_reclock (a);

    /*  If it is on the timeout queue, remove it.  */
    delete_timeout_reclock (a);

    node = hash_find( id );
    dlm_list_remove( (dlm_listnode_t *) node );

    kmem_cache_free( dlm_rlhash_cache, node );
    kmem_cache_free( dlm_reclock_cache, a );
    clm_reclocks_used--;  /* global count */

    cltrc(L_USR, CLLOCKD_HKID, T_EXIT_FUNC, T_CLLOCKD_FREE_LE, 0x00, 0x00, 0x00, 0x00);
}


			
/*
 * Name
 * 	get_le
 * Description
 * 	Return a reclock from private pool
 * 	all available reclocks are linked into free lists by lock family
 * 	Reclocks are allocated for each family from separate pools
 * Arguments
 *	locktype - The lock family for which a lock is desired
 * Returns
 * 	A pointer to a reclock. The only field considered valid is the lockid
 * 	field. 
 * 	No errors are reported back because they are all fatal
 * Global Data Affected
 * 	freelocks
 */	
struct reclock *
get_le(dlm_restype_t locktype)
{
    struct reclock * a;
    struct rl_hash  * rlh;


    cltrc(L_USR, CLLOCKD_HKID, T_ENTR_FUNC, T_CLLOCKD_GET_LE, 0x00, 0x00, 0x00, 0x00);

    CLM_ASSERT( dlm_reclock_cache != NULL, "reclock cache NULL!" );
    CLM_ASSERT( dlm_rlhash_cache != NULL, "rlhash cache NULL!" );

    rlh = NULL;
    a = kmem_cache_alloc( dlm_reclock_cache, 0 );
    if ( a != NULL )
    {
	(void) memset( a, 0, sizeof( struct reclock ) );

	a->lck.lockid = make_lock_id( OurSiteID, next_generation() );
	a->lck.type = locktype;
	a->lck.state = 0;
	clm_reclocks_used++ ;    /* global count */

	rlh = kmem_cache_alloc( dlm_rlhash_cache, 0 );
	if ( rlh != NULL )
	{
	    rlh->lock = a;
	    hash_insert( rlh );
	}
	else
	{
	    kmem_cache_free( dlm_reclock_cache, a );
	    a = NULL;
	}
    }

    if ( a == NULL )
    {
	bsdlog( LOG_CRIT, "[%s:%s] No more locks", current->comm,
	        __FUNCTION__ );
	clm_reclocks_denied++;   /* global count */
	
	CLM_ASSERT( rlh == NULL, "rl_hash not NULL, but reclock is?" );
    }
#ifdef DEBUG
    else
    {
	CLM_ASSERT( rlh != NULL, "rl_hash is NULL, but reclock is not?" );
    }
#endif
    cltrc(L_USR, CLLOCKD_HKID, T_EXIT_FUNC, T_CLLOCKD_GET_LE, a, 0x00, 0x00, 0x00);
    return a;
}


/*
 * Name
 *	find_reclock
 * Description
 * 	given a lockid return a pointer to the reclock
 * Arguments
 * 	locktype - lock family
 * 	lockid - the lockid associatyed with the reclock to be fetched
 * Returns
 *	a pointer to the desired reclock 
 *	or NULL  if lockid specifies out of range index
 * Global Data Affected
 *	NONE
 */
struct reclock *
find_reclock(dlm_restype_t locktype, uint lockid)
{
    struct rl_hash  * node;
    struct reclock * rp;


    cltrc(L_USR, CLLOCKD_HKID, T_ENTR_FUNC, T_FIND_RECLOCK,
	  0x00, 0x00, 0x00, 0x00);
	
    if (locktype < 0 || locktype > DLM_RES_NTYPES) 
    {
	bsdlog(LOG_ERR, "find_reclock: locktype %d out of range\n",
	       locktype);
	return NULL;
    }

    node = hash_find( lockid );
    if ( (node != NULL) && (node->lock != NULL) )
    {
	rp = node->lock;
	rp->lck.type = locktype;
    }
    else
    {
	rp = NULL;
	DBPRINTF((DBG_LOCK, "find_reclock: got wrong reclock"));
    }

    cltrc(L_USR, CLLOCKD_HKID, T_EXIT_FUNC, T_FIND_RECLOCK, rp,
	  0x00, 0x00, 0x00);

    return rp;
}
