/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_batch.c,v 1.11 2001/10/16 18:56:23 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: BATCH_MUTEX
 *		MSGSTR
 *		clm_alloc_trans
 *		clm_free_trans
 *		clm_queue_reqmsg
 *		clm_queue_respmsg
 *		clm_queue_response
 *		clm_response_complete
 *		clm_writev
 *		copy_trans
 *		match_request
 *		pti_flush_resend
 *		pti_flush_responses
 *		resend_inflight
 *		unresend
 */

#ifndef lint
#endif

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/types.h>

#include "dlm.h"
#include "dlm_kernel.h"
#include "dlm_version.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "cltrc.h"
#include "clm_data.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/*
 * This file implements an allocation and batching scheme for 
 * requests (transactions) from clients.
 * Transactions are allocated from the "available" queue
 * and put on the "sendable" queue when they are ready for
 * transmission to the primary. On being sent, they are moved 
 * to the "inflight" queue. When a response is received or a 
 * buffer is no longer needed, it  is returned to the "available" queue.
 * When a buffer is in use on the local machine, it does not appear
 * on any of the three queues.
 * 
 * A transaction is marked as sendable by a call to clm_push_trans(). 
 * Transmission  occurs when one of two conditions are true:
 * 1. We have reached our configured batching size
 * 2. A timer has expired. Flush out any pending writes
 *
 * Buffer entries are marked ready for re-use by calling clm_free_trans(). 
 * 
 */
#define SENDBUFSIZE             (DLM_BATCHSIZE*9)
#define ALLOC_LIMIT             (DLM_BATCHSIZE*3)

#ifdef DEBUG_ALL_PROTO /* hide from prototype file but not debugger */
#define STATIC
#else
#define STATIC static
#endif

/* file globals start here */

STATIC atomic_t t_allocated = ATOMIC_INIT( 0 );


/* Notice: If MEMORY_DEBUG is enabled, then linux/mm/slab.c *MUST* be built
 * with DEBUG enabled.  Otherwise, kmem_cache_create will oops.
 */

#ifdef MEMORY_DEBUG
# define TRANS_CACHE_FLAGS  (SLAB_RED_ZONE | SLAB_POISON)
#else
# define TRANS_CACHE_FLAGS  (0)
#endif

static kmem_cache_t * clm_trans_cache = NULL;


/* New entries are put on the end of these two queues to preserve temporal
 * order. Transactions are process in a single threaded fashion on the
 * primary, so we expect the responses to come back in the same order
 */

/* BUGBUG Both of these queues should be protected by a sleep lock.  Both
 * BUGBUG could probably be protected by the same mutex, but I don't see
 * BUGBUG any requirement either way.
 */

STATIC DECLARE_DLM_LIST( sendable_queue );
STATIC DECLARE_DLM_LIST( resend_queue );


/*
 * These two used elsewhere in module...
 */

/* BUGBUG The 'inflight' queue should also be protected by a sleep lock.
 */

clm_msg_t *inflight = NULL;
clm_msg_t *inflight_last = NULL;

/* end of file globals */


/*
 * Name
 *	clm_batch_init
 * Description
 * 	Initialize static in clm_batch.c
 * Arguments
 *	NONE
 * Returns
 *	Nothing
 */
void
clm_batch_init(void)
{
    inflight = NULL;
    inflight_last = NULL;

    clm_trans_cache = kmem_cache_create( "DLM transaction",
					 sizeof( clm_trans_t ),
					 TRANS_CACHE_FLAGS, 0,
					 NULL, NULL );
    CLM_ASSERT( clm_trans_cache != NULL, "Unable to create transaction "
	        "cache" );
}




/****************************************************************************/
/* clm_batch_cleanup - Clean up all resources used by transaction batches
 *
 * Cleans up all of the resources used by transaction batches.  Currently,
 * this primarilly amounts to destroying the clm_trans_cache.  The should
 * only be called during DLM shutdown.
 */

void
clm_batch_cleanup( void )
{
    if ( clm_trans_cache != NULL )
    {
	kmem_cache_destroy( clm_trans_cache );
	clm_trans_cache = NULL;
    }
}




/*
 * Name
 *	clm_buff_check
 * Description
 * 	Return TRUE if there are more than TR_BUFF_LIMIT transaction
 *	buffers in use.  This is used to hold off clients from queuing large
 *	numbers of requests faster than a remote resource master can
 *	return responses.
 * Arguments
 *	NONE
 * Returns
 *	TRUE if limit is reached
 * Global Data Affected
 *	None
 */
int
clm_buff_check(void)
{
    return ( atomic_read( & t_allocated ) > TR_BUFF_LIMIT);
}




/****************************************************************************/
/* clm_alloc_trans - Allocate a transaction buffer
 *
 * Allocates, but does not initialize, a new transaction buffer.  The memory
 * returned by this function will contain random values and must NOT be used
 * by the caller.  In DEBUG builds, the memory is intentionally set to values
 * that are likely to cause a system crash if used (as a pointer).
 *
 * RETURNS
 *
 * A pointer to an uninitialized transaction on success or NULL on failure.
 */

clm_trans_t *
clm_alloc_trans( void )
{
    clm_trans_t * t;


    t = kmem_cache_alloc( clm_trans_cache, 0 );
    if ( t != NULL )
    {
	atomic_inc( & t_allocated );
    }

    return( t );
}

/*
 * Name
 *	copy_trans
 * Description
 * 	return a copy of a tansaction buffer
 * Arguments
 *	transaction to copy
 * Returns
 *	a transaction buffer
 *	NULL if no buffers are available
 */
clm_trans_t *
copy_trans( const clm_trans_t * trans )
{
    clm_trans_t * tbp;


    tbp = clm_alloc_trans();
    if ( tbp != NULL )
    {
	(void) memcpy( tbp, trans, sizeof( clm_trans_t ) );
    }

    return( tbp );
}




/****************************************************************************/
/* clm_free_trans - Free a transaction buffer
 *
 * Releases a transaction buffer, previously allocated by clm_alloc_trans,
 * back to the available pool.  This function assumes that the buffer is not
 * linked into any list.
 */

void
clm_free_trans( clm_trans_t * transp )
{
    atomic_dec( & t_allocated );
    kmem_cache_free( clm_trans_cache, transp );
}


/* clm_writev - Send all messages in a message queue
 *
 * Call cccp_send for each message on the provided queue.  This function
 * has been changed so that before calling cccp_send on a queue entry,
 * it removes the entry from the queue in case that send somehow
 * wraps around to a different function that leads us back here.  This
 * prevents the comm layer from putting us in an infinite loop.
 */

void
clm_writev(
	    /* Pointer to the queue of messages to be sent.
	     */

	    dlm_list_t * msg_queue 
	  )
{
    clm_msg_t	*msg;
    int		ret;


    while ( ! dlm_list_is_empty( msg_queue ) ) 
    {
	msg = (clm_msg_t *) dlm_list_rem_head( msg_queue );

	STAMP((clm_trans_t *)(msg->cm_data));

	/* adjust the output message to destination version */
	clm_vers_snd(msg, clm_version(msg->cm_node));

	/*
	 * If the destination value is set to 0, then the message
	 * was reset because of a node failure.  Use downRetry to
	 * reasign the destination, otherwise, just call cccp_send()
	 */
	if (msg->cm_node == 0) {
	    ret = downRetry(msg);
	} else {
	    ret = cccp_send(msg);
	}

	/* if the send indicates a node node, use downRetry to resend */
	if (CPERR_NODEDOWN == ret) {
	    ret = downRetry(msg);
	}
	DBPRINTF((DBG_COMM, "writev: cccp_send returned %d\n", ret));

	
	/* If the message really could not be sent, put it back at the head
	 * of the queue.  In doing this, the message will be the first one
	 * sent (as it should be) on the next try.  This will maintain the
	 * proper ordering of the messages.
	 */

	if (CPERR_OK != ret) {
	    CLM_ASSERT((msg->cm_private == NULL),
		       "writev: cm_private not NULL");

	    dlm_list_add_head( msg_queue, (dlm_listnode_t *) msg );

	    DBPRINTF((DBG_COMM, "dest=%d port=%d\n", msg->cm_node,
		      msg->cm_port));
	    break;
	}
    }


    return;
}

/*
 * Name
 *	pti_flush_resend
 * Description
 *	This function walks the queue of all the requests that failed when
 *	their destination node bought the farm.  Hopefully, node rebuild has
 *	completed successfully, so resend the requests.
 * Arguments
 *	NONE
 * Returns
 *	NOTHING
 * Global Data Affected
 *	resend
 *	resend_last
 */
void
pti_flush_resend(void)
{
    DBPRINTF((DBG_FUNC|DBG_RECOV, "enter %s\n", __FUNCTION__));

    clm_writev( & resend_queue );

    DBPRINTF((DBG_FUNC|DBG_RECOV, "leave %s\n", __FUNCTION__));
    return;
}


/*
 * Name
 *	pti_flush_responses
 * Description
 *	Check all of the secondary response queues and send any replies
 *	back to the appropriate node.
 * Arguments
 *	NONE
 * Returns
 *	NOTHING
 * Global Data Affected
 *	sendable
 *	sendable_last
 */

void
pti_flush_responses(void)
{
    clm_writev( & sendable_queue );
    clmm_trigger();
    return;
}


/*
 * Name
 *	reset_inflight
 * Description
 *	This function is called at the beginning of reconfiguration to check
 *	for any messages on the inflight queue with a destination of the
 *	specified node id.  These messages should have their destinations
 *	reset to zero so they will be resent when reconfig is done.
 * Arguments
 *	node id of the destination to reset
 * Returns
 *	NOTHING
 * Global Data Affected
 *	None
 */
void
reset_inflight(int node)
{
	clm_msg_t	*msg = inflight, *next;

	/* don't dink messages queued to active node */
	if (clm_node_reachable(node)) {
		return;
	}

	DBPRINTF((DBG_FUNC|DBG_RECOV, "enter %s\n", __FUNCTION__));
	while (msg) {
		next = ((clm_trans_t *)(msg->cm_data))->clm_next;

		if (msg->cm_node == node) {
			msg->cm_node = 0;

			/* must be 0 on resend, or we try to free msg twice */
			msg->cm_private = NULL;
		}

		msg = next;
	}

	DBPRINTF((DBG_FUNC|DBG_RECOV, "leave %s\n", __FUNCTION__));

	return;
}



/*
 * Name
 *	resend_inflight
 * Description
 *	This function is called after reconfiguration has been completed for
 *	a node failure.  It walks the inflight and redirects or resends any
 *	requests that were sent to the failed node and are awaiting a
 *	response that isn't likely to come.
 * Arguments
 *	NONE
 * Returns
 *	NOTHING
 * Global Data Affected
 *	None
 */
void
resend_inflight(void)
{
	clm_msg_t	*msg = inflight, *next;

	DBPRINTF((DBG_FUNC|DBG_RECOV, "enter %s\n", __FUNCTION__));
	while (msg) {
		next = ((clm_trans_t *)(msg->cm_data))->clm_next;

		/*
		 * The following section of code should be split out into
		 * a separate function that takes a message structure pointer
		 * and resets the destination node to the appropriate
		 * directory node.	XXX
		 */
		if (msg->cm_node <= 0 || !clm_node_reachable(msg->cm_node)) {
			clm_trans_t	*req = msg->cm_data;
			int			site = 0, ret;

			CLM_ASSERT(((int)msg->cm_private < 16),
				"Msg queued somewhere else");
			DBPRINTF((DBG_CONFIG|DBG_RECOV,
					"%s: %d message lost to failed node",
					__FUNCTION__, msg->cm_type));

			switch (msg->cm_type) {
			case DLM_LOCK:
			case DLM_UNLOCK:
				site = clm_direct(
					req->clm_lockreq.nameinfo.name,
					req->clm_lockreq.nameinfo.namelen);
				break;
			case DLM_REGISTER:
				site = clm_direct(
					&req->clm_register.unix_name[0],
					req->clm_register.namelen);
				break;
			case DLM_SCN_OP:
				site = clm_direct(
					&req->clm_scninfo.nameinfo.name[0],
					req->clm_scninfo.nameinfo.namelen);
				break;
			default:
				/* best we can do is punt */
				bsdlog(LOG_CRIT, "%s: Node failed.  Can't "
				       "redirect type %ld",
					__FUNCTION__, msg->cm_type);

				/* pull myself off the inflight queue */
				req = match_request(req);

				clm_free_trans(req);
				continue;
			}

			if (0 == site) {
				msg = next;
				bsdlog(LOG_CRIT,
					"%s: Node failed.  Can't get site", __FUNCTION__);

				/* pull myself off the inflight queue */
				req = match_request(req);

				clm_free_trans(req);
				continue;
			}

			msg->cm_node = site;
			msg->cm_port = PORT_DIRECT;
			msg->cm_private = NULL;

			DBPRINTF((DBG_RECOV, "%s: resend to %d\n", __FUNCTION__, site));

			ret = cccp_send(msg);
			if (ret != CPERR_OK) {
				/* XXX re-interpret and redirect */
				bsdlog(LOG_CRIT,
					"%s: Node failed.  Send failed with %d",
					__FUNCTION__, ret);

				/* pull myself off the inflight queue */
				req = match_request(req);

				clm_free_trans(req);
			} else {
				DBPRINTF((DBG_CONFIG,
					"%s: resent to %d pid %d",
					__FUNCTION__, site, req->clm_pid));
			}

		}

		msg = next;
	}

	DBPRINTF((DBG_FUNC|DBG_RECOV, "leave %s\n", __FUNCTION__));

	return;
}


/*
 * Name
 *	unresend
 * Description
 *	Remove requests from the resend queue that belong to dead clients.
 * Arguments
 *	pid - pid of the deceased client
 * Returns
 *	Nothing
 * Global Data Affected
 *	resend
 *	resend_last
 */
void
unresend(int	dead_pid)
{
    clm_msg_t	   * msg;
    dlm_listnode_t * next;
    clm_trans_t	   * req;

    DBPRINTF((DBG_PURGE, "Enter unresend for %d", dead_pid ));

    /* Walk through the resend list.  If a transaction is for the client
     * that died, remove it and destroy it.
     */

    for ( msg = (clm_msg_t *) dlm_list_get_head( & resend_queue )
	  ; msg != NULL
	  ; msg = (clm_msg_t *) next )
    {
	next = dlm_list_succ( (dlm_listnode_t *) msg );
	req = (clm_trans_t *)msg->cm_data;

	if (req->clm_pid == dead_pid) 
	{
	    DBPRINTF((DBG_PURGE, "Un-resend req from %d\n", dead_pid));

	    /* unlink the message from the list */
	    dlm_list_remove( (dlm_listnode_t *) msg );

	    /* prevent match_request from doing something stupid */
	    msg->cm_private = NULL;

	    /* going to free it, so take it off the inflight queue */
	    req = match_request(req);

	    /* free the memory */
	    clm_free_trans(req);
	    cccp_msg_destroy( msg );
	}
    }

    return;
}


/*
 * Name
 *	clm_queue_reqmsg
 * Description
 *	I discovered early on that if you put requests that need to be
 *	redirected on the same queue as the master responses so they'll get
 *	flushed after reconfig, you get into some very weird infinite loops.
 *	For that reason, requests whose callback runs where the status
 *	indicates node failure are put onto a "bounced request" queue to be
 *	flushed out when reconfiguration is completed.  This is the function
 *	used to put those requests on the queue.
 *	NOTE:  The requests that are being bounced are never removed from the
 *	inflight queue.  That is why we use the cm_private field to string
 *	these into a separate queue.
 * Arguments
 *	msg - message pointer to be queued
 * Returns
 *	Nothing
 * Global Data Affected
 *	resend
 *	resend_last
 */
void
clm_queue_reqmsg(clm_msg_t *msg)
{
    clm_trans_t *transp = (clm_trans_t *)msg->cm_data;

    if (NULL == transp) {
	bsdlog(LOG_CRIT, "NULL data ptr in clm_queue_reqmsg");
	abort("clm_queue_reqmsg");
    }

    dlm_list_add_tail( & resend_queue, (dlm_listnode_t *) msg );

    DBPRINTF((DBG_RECOV|DBG_BATCH, "Queued request 0x%x to site %d\n",
	      transp, msg->cm_node));

    return;
}


/*
 * Name
 *	clm_queue_respmsg
 * Description
 *	Mark the current message as a RESPONSE (PORT_RESPONSE)
 *	Put the provided message pointer on the queue of responses
 *	to be sent.
 * Arguments
 *	msg - message pointer to be queued
 * Returns
 *	Nothing
 * Global Data Affected
 *	sendable
 *	sendable_last
 */
int
clm_queue_respmsg(clm_msg_t *msg)
{
    msg->cm_port = PORT_RESPONSE ;
    return clm_queue_msg(msg) ;
}


/*
 * Name
 *	clm_queue_msg
 * Description
 *	Put the provided message pointer on the queue of messages
 *	to be sent.
 *	This is done to prevent recursive calls through the 'interesting'
 *	guts of cccp.
 * Arguments
 *	msg - message pointer to be queued
 * Returns
 *	Nothing
 * Global Data Affected
 *	sendable
 *	sendable_last
 */
int
clm_queue_msg(struct clm_msg *msg)
{
#ifdef DBPRINT
    clm_trans_t *transp = (clm_trans_t *)msg->cm_data;
#endif /* DBPRINT */

    dlm_list_add_tail( & sendable_queue, (dlm_listnode_t *) msg );

    DBPRINTF((DBG_BATCH,"Queued message 0x%x to site %d\n",transp, 
	      msg->cm_node));

    return 0 ;
}


/*
 * Name
 * 	clm_response_complete
 * Description
 *	This function is call by the communications layer when it has
 *  	received an acknowledgement for the outbound response.  A response
 *	is the transaction structure that is returned from the master to the
 *	secondary, so if we are in this function, it is because we are
 *	receiving confirmation that a response that we sent was processed
 *	by the destination node.  We can now free that transaction structure.
 * Arguments
 *	msg - pointer to the message structure that was acknowledged
 *	stat - status returned
 * Returns
 *	Nothing
 * Global Data Affected
 *	NONE
 */
void
clm_response_complete(clm_msg_t	*msg, int	stat)
{
	clm_trans_t *req = (clm_trans_t *)msg->cm_data;

	if (stat) {
		bsdlog(LOG_ERR, MSGSTR(ERROR_DETECTED,
				"Error detected in clm_response_complete\n"));
	}

	clm_free_trans(req);
	cccp_msg_destroy( msg );

	return;

}

/*
 * Name
 *	clm_queue_response
 * Description
 *	Put the provided transaction pointer on the queue of responses
 *	to be sent back to the secondary specified by the site argument.
 *	This function should generally be used to send notifies and other
 *	asynchronous messages to another node.
 *
 *	NOTE:   WARNING!  This function assumes that the transaction
 *	structure pointer that has been passed to it has been allocated
 *	from clm_alloc_trans().  It sets up a callback function for the
 *	send and call clm_free_trans to free the transaction.  This
 *	queue function should *NOT* be used if the transaction structure
 *	has not been allocated from clm_alloc_trans()!!!!!!!!!!
 * Arguments
 *	site - nodeid of the secondary to queue the response for
 *	transp - a transaction prevously allocated by clm_alloc_trans()
 * Returns
 *	Nothing
 * Global Data Affected
 *	None
 */
void
clm_queue_response(int site, clm_trans_t *transp)
{
    clm_msg_t * msg;


    msg = cccp_msg_create( 0, 1, site, transp->clm_type, sizeof( *transp ),
			   clm_response_complete, transp, NULL );
    if (NULL == msg)
    {
	bsdlog(LOG_CRIT, MSGSTR(MALLOC_ERR,
				"malloc error in clm_queue_response\n"));
	abort("malloc in clm_queue_response");
    }

    transp->clm_next = NULL;
    transp->clm_vers = DLM_VERSION;

    DBPRINTF((DBG_BATCH,"Enter clm_queue_response: dir=%d\n",
	      transp->clm_direction));

    clm_queue_respmsg(msg);
    return;
}


/*
 * Name
 *	match_request
 * Description
 * 	Match an incoming response with the associated request. We
 * 	will eventually deal with "stale" requests here (requests that never
 * 	got received a reply). Since execution is single threaded on the 
 * 	primary, we expect to get responses back in the same order that the
 * 	requests went out. If the response does not correspond to the first
 * 	unresolved request, something is amiss. We don't handle this gracefully
 * 	yet, because it is not clear what to do. 
 * Arguments
 *	resp - an incoming transaction to be paired with its request
 * Returns
 *	the associated request
 *	abort() if not found
 * Global Data Affected
 *	inflight
 * 	inflight_last
 */
clm_trans_t *
match_request(clm_trans_t *resp)
{
	clm_msg_t		*tr_prev = NULL;
	clm_msg_t		*tr_curr = inflight;
#ifdef DBPRINT
	int	match_count = 0;
#endif

	DBPRINTF((DBG_FUNC, "[%s:%s] %s\n", current->comm, __FUNCTION__,
		  "enter"));
	while (tr_curr && tr_curr->cm_data != resp->clm_sender) {
#ifdef DBPRINT
		match_count++;
#endif
		tr_prev = tr_curr;
		tr_curr = ((clm_trans_t *)(tr_curr->cm_data))->clm_next;
	}

#ifdef DBPRINT
        DBPRINTF((DBG_FORCE, "[%s:%s] match_count = %d",
		  current->comm, __FUNCTION__, match_count));
#endif

#if 0
	if (500 < match_count)
		bsdlog(LOG_ERR, MSGSTR(QUE_OVERFLOW,
				"inflight queue exceeds 500 entries"));
#endif
	if (resp->clm_sender != tr_curr->cm_data) {
		if (findclientbypid(OurSiteID, resp->clm_authpid) == NULL) {
			DBPRINTF((DBG_BASIC,
			  "found a leftover transaction from a dead client"));
			return (NULL);
		}
		/*
		 * found a request that never got a reply
		 */
		DBPRINTF((DBG_FORCE,"Mismatch trans seq = %x inf = 0x%x"
				" sender = 0x%x",
				resp->clm_sequence, inflight,
				resp->clm_sender));
		bsdlog(LOG_CRIT, MSGSTR(LOST_REPLY,
			"ERROR: lost a reply from the primary. Aborting"));
		TRCHKGT(CLLOCKD_HKID, T_MISMATCH_RESP, resp->clm_sender,
				resp->clm_sequence, inflight, 0x00);
		abort("match_request");
	}
	else {
#ifdef DBPRINT
		TRCHKGT(CLLOCKD_HKID, T_MATCH_RESP, resp->clm_sender,
				match_count, 0x00, 0x00);
#endif
		if (NULL != tr_prev) {
			((clm_trans_t *)(tr_prev->cm_data))->clm_next =
					((clm_trans_t *)
					(tr_curr->cm_data))->clm_next;
			if (inflight_last == tr_curr)
				inflight_last = tr_prev;
		} else {
			inflight = ((clm_trans_t *)
					(tr_curr->cm_data))->clm_next;
			if (inflight == NULL)
				inflight_last = NULL;
		}
		((clm_trans_t *)(tr_curr->cm_data))->clm_next = NULL;
	}

	CLM_ASSERT(((ulong)tr_curr->cm_private < 16),
			"match_request: msg on queue");
	/*
	 * free the message structure - we check cm_private because here we
	 * use it like a flag.  The message structure either has to be freed
	 * by the send callback or by this function, but we never know which
	 * function is going to be called first.  Whichever one *is* called
	 * first, sets the cm_private field of the message to be non-zero.
	 * So, if cm_private is not zero, then the other function has finished
	 * with the message and it can be safely freed.  Got it?
	 */
	if ((ulong)tr_curr->cm_private & MSG_CONFIRMED) {
		cccp_msg_destroy( tr_curr );
	} else {
		tr_curr->cm_private = (void *)((ulong)tr_curr->cm_private |
					MSG_MATCHED);
	}

	DBPRINTF((DBG_FUNC, "[%s:%s] %s\n", current->comm, __FUNCTION__,
		  "exit"));

	return ((clm_trans_t *)resp->clm_sender);
}
