/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_client.c,v 1.17 2001/10/19 22:39:51 idr Exp $"

/*
 *   COMPONENT_NAME: CLLOCKD
 *
 *   FUNCTIONS: CMP_VALUE
 *		MSGSTR
 *		add_client_reclock
 *		add_timeout_reclock
 *		addclient
 *		checkclient
 *		delete_client_reclock
 *		delete_timeout_reclock
 *		dumpf_timeout_queue
 *		findclient
 *		icmp
 *		removeclient
 *		trash_replies
 */

#ifndef lint
#endif

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/stddef.h>
#include <linux/signal.h>
#include <linux/types.h>
#include <linux/param.h>
#include <linux/errno.h>

#include "dlm.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "cltrc.h"
#include "clm_data.h"		/* external data definitions */
#include "dlm_kernel.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/*
 * A database of lock manager clients is maintained on both the
 * primary and secondary. Each node tracks the clients locally and
 * maintains a record of the sequence number contained in the most
 * recent request received. In this way the lock manager can detect
 * when a UDP packet has been dropped. On the primary, all clients
 * are recorded in this database. Again, only local clients are checked
 * for dropped packets. However, every client in the database on the 
 * primary has associated with its entry a list of all reclock structures
 * held by that client. This structure is traversed when performing global 
 * deadlock detection.
 */
/*
 * How much do we bump the client array each time we have to increase its
 * size?  
 */
#define CLIENT_INCR	100

/* Array of client reference pointers that is malloc'ed in addclient.  */

static DECLARE_DLM_SPINLOCK( cli_refs_lock );
static clm_client_t ** cli_refs = NULL;

/* Size of above array in terms of number of client structures.  */
static unsigned    maxclients = 0;

static void clm_remove_client_by_array_pointer( clm_client_t ** p );

#define IS_ON_QUEUE(n) ((n)->succ != NULL)



/*
 * Name
 * 	clm_client_init 
 * Desription
 *	Initialize static in clm_client.c
 * Arguments
 *	None
 * Returns
 *	Nothing
 * Global Data Affected
 *	NONE
 */
void
clm_client_init(void)
{
    DLM_SPINLOCK_INIT( & cli_refs_lock );
    cli_refs = NULL;
    maxclients = 0;
}



/*
 * Name
 * 	check_client
 * Desription
 *	Useful client-debug function.  This function examines all of the
 *	locks queued on the client structure for each registered client and
 *	confirms that each lock that is not INFLIGHT_WAITING is queued
 *	somewhere on the resource that it references.
 *	This function aborts if it finds something bad.
 *	I removed all calls to it, but they can be added back for debugging.
 * Arguments
 *	cp - pointer to client structure that cannot be checked.  Skip it
 *		This call be and usually is NULL.
 * Returns
 *	Nothing
 * Global Data Affected
 *	NONE
 */
void
check_client(clm_client_t *cp)
{
    dlm_listnode_t  * curr;
    dlm_listnode_t  * next;
    struct reclock  * a;
    struct resource * rp;
    clm_client_t    * ptr;
    unsigned          i;
    unsigned          k;


    DLM_SPINLOCK_LOCK( & cli_refs_lock );

    for (k = 0; k < nclients; k++) 
    {
	ptr = cli_refs[k];

	if (ptr != cp)
	{
	    for ( curr = dlm_list_get_head( & ptr->cli_queue )
		  ; curr != NULL
		  ; curr = next)
	    {
		next = dlm_list_succ( curr );
		a = CLIENT_LINK_TO_RECLOCK( curr );

		if ((a->lck.site == OurSiteID) &&
		    (~a->lck.state & INFLIGHT_WAITING)) 
		{
		    /* here if inflight is clear and lock is local */

		    rp = resptr(a->lck.rh);
		    CLM_ASSERT((rp), "No resource in check_client");

		    for (i = 0; i < RLQ_MAX; i++) 
		    {
			if ( check_queue( &rp->lockqs[i], a, RL_RQ_OFS ) )
			{
			    DLM_SPINLOCK_UNLOCK( & cli_refs_lock );
			    return;
			}
		    }

		    /* lock not inflight and not found.  Bad, bad, bad */
		    bsdlog(LOG_CRIT, "Lock 0x%x state 0x%x not on queue",
			   a->lck.lockid, a->lck.state);
		    abort("Bad lock state in check_client");
		}
	    }
	}
    }

    DLM_SPINLOCK_UNLOCK( & cli_refs_lock );
}


/*
 * Name
 * 	icmp 
 * Desription
 *	Comparison routine for qsort and bsearch.  Comparison based on the 
 *      site and client id in the client structure;  
 * Arguments
 *	c1,c2 - the two clients to compare
 * Returns
 *
 * 	0 if clients are identical
 *      > 0 if c1 if longer or alphabetically > than c2
 *      < 0 if c1 if short or alphabetically < than c2
 * Global Data Affected
 *	NONE
 */
int
icmp(clm_client_t **c1, clm_client_t **c2)
{
    int rc ;

    if ((rc = (*c1)->cli_site - (*c2)->cli_site))
	return rc ;		/* sites are different */

    return memcmp(&(*c1)->cli_id, &(*c2)->cli_id, 
		  sizeof(clm_client_id_t));
}

/*
 * Function:
 *	kill_clients
 *
 * Description:
 *	Send a SIGKILL to any clients that we know about.  It isn't much of a
 * problem while we're dead, but if the lock manager is restarted and
 * they're still around, the system call that they're stuck in may try to
 * reference memory that we free when we exit.
 *
 * Arguments:
 *	NONE
 *
 * Return value:
 *	NONE
 *
 * Side effects:
 *	Lots of little, dead, client processes.
 */
void
kill_clients (void)
{
    int i, pid, corpse = FALSE;

    DBPRINTF((DBG_FUNC|DBG_CONFIG, "Entering kill_clients\n"));

    DLM_SPINLOCK_LOCK( & cli_refs_lock );

    for (i = 0; i < nclients; i++) 
    {
	/* if client structure refs a process, KILL it */
	if ( (pid = CLID2PID(cli_refs[i]->cli_id)) 
	     && (cli_refs[i]->cli_type == CLM_CLI_PROCESS)
	     && (cli_refs[i]->cli_site == OurSiteID) )
	{
	    DBPRINTF((DBG_CONFIG, "Send SIGKILL to %d\n", pid));

	    /* BUGBUG Not 100% sure about this.  Can dlm_kill_user_proc
	     * BUGBUG sleep?  If so, then that messes up the mutex.
	     */

	    (void) dlm_kill_user_proc( pid, SIGKILL );
	    corpse = TRUE;
	}
	
	/* Once the client has been killed, free the client structure.
	 */

	clm_remove_client_by_array_pointer( & cli_refs[ i ] );
    }

    DLM_SPINLOCK_UNLOCK( & cli_refs_lock );


    /* if we killed anyone, we observe a moment of silence */
    if (corpse) 
    {
	clm_sleep(1);
    }

    return;
}

/*
 * addclientbypid
 *
 * add client by site and pid rather than client_id_t.
 * see addclient for what it does
 */
bool_t
addclientbypid(
	        /* Node ID on which the client is running.
		 */

	        int site,


	        /* Process or group ID.
		 */

	        int pid,


	        /* Client type (i.e., process, group, or transaction).
		 */

	        clm_client_type_t type, 


	        /* (optional) Buffer to hold a pointer to the new client.
		 */

	        clm_client_t ** cli_ptr
	       )
{
    clm_client_id_t   clid;


    set_client_id( clid, pid );
    return( addclient( site, &clid, type, 0, cli_ptr ) );
}




/****************************************************************************/
/* addclient - Add a client to the list of valid clients
 *
 * Add another client to the list of valid clients.  Possibly increase the
 * size of the clients array by CLIENT_INCR.
 * 
 * If cli_ptr is non-NULL, the buffer it points to will be filled with a
 * pointer to the client structure if the call is successful (i.e., TRUE is
 * returned).  Otherwise, (i.e., if FALSE is returned) the buffer will be
 * filled with NULL.
 * 
 * RETURNS
 * 
 * If the new client is successfully added, TRUE is returned.  Otherwise,
 * FALSE is returned.
 *
 * NOTES
 *
 * The design of cli_ptr may seem seem strange, but there is a method to the
 * madness.  In the future it is expected that each client will have an
 * associated mutex.  When the client pointer is returned, the mutex will
 * be held.  By using this strange pointer-to-pointer scheme, callers that
 * don't need the client pointer will not have to bother with unlocking the
 * mutex.
 *
 * GLOBALS
 *
 * nclients, clients
 */

bool_t
addclient(
	   /* Node ID on which the client is running.
	    */

	   int site,


	   /* Pointer to the client ID of the client.
	    */

           const clm_client_id_t * pclid,


	   /* Client type (i.e., process, group, or transaction).
	    */

	   clm_client_type_t type, 


	   /* Initial sequence number for the client.
	    */

	   unsigned int sequence,


	   /* (optional) Buffer to hold a pointer to the new client.
	    */

	   clm_client_t ** cli_ptr
	 )
{
    clm_client_t  * new_cli;
    clm_client_t ** addp;
    unsigned        i;


    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_ADDCLIENT,
	    site, pclid->clid_i[1], 0x00, 0x00);

    DBPRINTF((DBG_CLIENT, "addclient client site = %d id: %x %x (%d)\n", site,
	      db_clidp(pclid), pclid->clid_i[1]));

    if ( cli_ptr != NULL )
    {
	*cli_ptr = NULL;
    }

    DLM_SPINLOCK_LOCK( & cli_refs_lock );

    if (nclients >= maxclients) 
    {
	clm_client_t * clients;
	int	old_size;
	int	new_size;

	old_size = maxclients * sizeof(clm_client_t *);
	maxclients += CLIENT_INCR;
	new_size = maxclients * sizeof(clm_client_t *);


	/* BUGBUG This is bad.  If the REALLOC call fails, the DLM is up
	 * BUGBUG the creek without a paddle.
	 */
	/* BUGBUG Can malloc/realloc sleep?  If so, that will break the
	 * BUGBUG spinlock_t.
	 */

	clients = (clm_client_t *) MALLOC(CLIENT_INCR * sizeof(clm_client_t)); 
	if (cli_refs)
	    cli_refs = (clm_client_t **)REALLOC(cli_refs, old_size,new_size);
	else
	    cli_refs = (clm_client_t **)MALLOC(new_size);

	if (!clients || !cli_refs)
	{
	    DLM_SPINLOCK_UNLOCK( & cli_refs_lock );
	    return FALSE;
	}


	for (i=nclients; i<maxclients; i++)
	    cli_refs[i] = &clients[i-nclients];
    }

    new_cli = cli_refs[ nclients ];

    (void) memset( new_cli, 0, sizeof( clm_client_t ) );
    (void) memcpy( & new_cli->cli_id, pclid, sizeof( clm_client_id_t ) );
    new_cli->cli_site = site;
    new_cli->cli_type = type;
    new_cli->cli_sequence = sequence;
    dlm_list_init( & new_cli->sync_seq );
    dlm_list_init( & new_cli->cli_queue );
    dlm_list_init( & new_cli->cli_ast_pnd );
    init_waitqueue_head( & new_cli->cli_reply_event );


    /* get pointer to item *after which* this should be added */

    addp = (clm_client_t **) bsearchp( & new_cli, cli_refs,
				       nclients, sizeof cli_refs[0],
				       (int(*)(const void*,const void*)) icmp);
    if ( addp == NULL )
    {
	/* Here if addclient called for pid/site pair that is already in
	 * the list.  This can happen if a pid is recycled by AIX before
	 * purge of previous incarnation is complete.  Notify caller
	 * of failure and let them take appropriate action.
	 */

	DLM_SPINLOCK_UNLOCK( & cli_refs_lock );
	return FALSE;
    }

    addp++ ;	/* point at place addition should be done */

#ifdef CONFIG_PROC_FS
    new_cli->proc_entry = dlm_proc_create_client_entry( site, new_cli );
#endif

    /* if not last, move ptr table down one */
    if ( addp != & cli_refs[ nclients ] )
    {
	(void) memmove( & addp[1], & addp[0],
		        (size_t) ( ((char *) & cli_refs[ nclients ]) 
				   - ((char *) addp) ) );

	/* copy first free ptr into addition spot */
	*addp = new_cli;
    }

    if ( cli_ptr != NULL )
    {
	*cli_ptr = new_cli;
    }

    nclients++;
    DLM_SPINLOCK_UNLOCK( & cli_refs_lock );

    return TRUE;
}


/*
 * Name
 *	findclient
 * Description
 *	Given a site and client id, find the entry in the corresponding client 
 *	in the client list.  
 * Arguments
 *	site - site id of the desired client
 *	clid - client id of the desired client
 * Returns
 *	a pointer to the desired client struct
 *	NULL if not found
 * Global Data Affected
 *	NONE
 */
clm_client_t  *
findclient(int site, const clm_client_id_t *pclid)
{
    clm_client_t **cp = NULL;
    clm_client_t   c;
    clm_client_t   *pc = &c;


    c.cli_site = site;
    (void) memcpy( & pc->cli_id, pclid, sizeof( clm_client_id_t ) );

    DBPRINTF((DBG_CLIENT,
	      "[%s:%s] site %d clid %08lx%08lx nclients %d max %d\n",
	      current->comm, __FUNCTION__,
              site, db_clidp(pclid), nclients, maxclients));

    DLM_SPINLOCK_LOCK( & cli_refs_lock );
    if (nclients > 0) 
    {
	cp = (clm_client_t **) bsearch((void *) &pc,
                                       (void *) cli_refs,
                                       nclients,
                                       sizeof(clm_client_t *), 
                                       (int(*)(const void*, const void*))icmp);
    }
    DLM_SPINLOCK_UNLOCK( & cli_refs_lock );

    return( (NULL == cp) ? NULL : *cp );
}

/*
 * Name
 *	findclientbypid
 * Description
 *	Given a site and pid, find the entry in the corresponding client 
 *	in the client list.  
 * Arguments
 *	site - site id of the desired client
 *	pid - process id of the desired client
 * Returns
 *	a pointer to the desired client struct
 *	NULL if not found
 * Global Data Affected
 *	NONE
 */
clm_client_t  *
findclientbypid(int site, int pid)
{
	clm_client_id_t   clid;

	set_client_id(clid, pid);
	return (findclient(site, &clid));
}

/*
 * Name
 * 	checkclient 
 * Description
 * 	Validate a local client and check for packet lossage.  This client is 
 *	local so we use our site number with pid to find the correct client 
 *	entry.  With the current client entry, we verify that SEQUENCE is 
 *	valid.  If it is valid, we adjust the sequence number in the client 
 *	entry. 
 * Arguments
 *	pid - process id of client to be validated
 *	sequence - the sequence number to be validate
 * Returns
 *	TRUE if client is valid and sequence is valid
 *	FALSE otherwise
 *
 */
int
checkclient(int pid, unsigned int sequence)
{
    clm_client_t * cp = findclientbypid(OurSiteID, pid);
    int   id;

    DBPRINTF((DBG_CTI, "checkclient:pid %d, seq %d\n", pid, sequence));

    id = 0;
    if ( cp != NULL )
    {
	cp->cli_sequence = sequence;
	id = (cp->cli_groupid != 0) ? cp->cli_groupid : pid;
    }

    return( id );
}

/*
 * Name
 *	trash_replies
 * Description
 *	Remove all queued replies from a client record
 * Arguments
 *	cp - pointer to client record
 * Returns
 *	void
 * Global Data Affected
 *	NONE
 */

static void
trash_replies (clm_client_t *cp)
{
    clm_trans_t *tp;
#if defined(ANNOYING_DEBUG_MESSAGES)
    int n = 0;
#endif

    while (cp->cli_reply) {
	tp = cp->cli_reply->clm_next;
	clm_free_trans(cp->cli_reply);
	cp->cli_reply = tp;
#if defined(ANNOYING_DEBUG_MESSAGES)
	++n;
#endif
    }

#if defined(ANNOYING_DEBUG_MESSAGES)
    bsdlog(LOG_DEBUG, MSGSTR(NUM_TRSH_REPLY, "trashed %d replies"),n);
#endif
}

/*
 * Name
 *	removeclientbypid
 * Description
 *	Given a site and pid, remove the entry for the corresponding client 
 *	in the client list.  
 * Arguments
 *	site - site id of the desired client
 *	pid - process id of the desired client
 * Returns
 *	nonzero if not found
 *	0 for success
 * Global Data Affected
 *	see removeclient
 */
int
removeclientbypid(int site, int pid)
{
    clm_client_id_t clid ;

    set_client_id(clid, pid) ;
    return removeclient(site, &clid) ;
}




/****************************************************************************/
/* clm_remove_client_by_array_pointer - Private routine to remove a client
 * 
 * Given a pointer to an entry in the cli_refs array, the client specified at
 * that location will be removed.
 * 
 * NOTES
 * 
 * This function assumes that the global client table lock will already be
 * held.
 */

static void
clm_remove_client_by_array_pointer( clm_client_t ** p )
{
    clm_client_t   * psearch;
    dlm_listnode_t * ast;


    CLM_ASSERT( nclients > 0, "Can't remove!  No known clients!" );
    CLM_ASSERT( p >= & cli_refs[ 0 ], "Invalid client array pointer!" );
    CLM_ASSERT( p < & cli_refs[ nclients ], "Invalid client array pointer!" );

    if ( ! dlm_list_is_empty( & (*p)->cli_queue ) )
	abort("deleting client with locks on queue") ;

    DBPRINTF((DBG_CLIENT, "[%s:%s] deleting 0x%08lx (%08lx%08lx)\n",
	      current->comm, "removeclient",
	      *p, (*p)->cli_id.clid_i[0], (*p)->cli_id.clid_i[1] ));
    trash_replies(*p);


    /* Destroy any pending ASTs for this client.  Typically, there will only
     * be ASTs queued at this point if the DLM is being shutdown.
     */

    while ( (ast = dlm_list_rem_head( & (*p)->cli_ast_pnd )) != NULL )
    {
	ASTfree( ast );
    }


#ifdef CONFIG_PROC_FS
    if ( (*p)->proc_entry != NULL )
    {
	dlm_proc_destroy_client_entry( (*p)->proc_entry );
	(*p)->proc_entry = NULL;
    }
#endif

    /* Remove the client for the global client table.
     */

    --nclients;
    if (nclients > 0) 
    {
	psearch = *p;
	(void) memmove( p, p + 1, 
		        ((void *) & cli_refs[ nclients ]) - ((void *) p) );
	cli_refs[ nclients ] = psearch;
    }
}




/****************************************************************************/
/*
 * Name
 *	removeclient
 * Description
 *	Remove a client from the client (site/client id) list. Usually 
 *      called after lock purge.
 * Arguments
 *	site - site id of the client to remove
 *	pclid - pointer to client id of the client to remove
 * Returns
 *	< 0 if specified client doesn't exist
 *	0 for sucess
 * Global Data Affected
 *	nclients
 *	cli_refs
 */
int
removeclient(int site, clm_client_id_t *pclid)
{
    clm_client_t  ** p;
    clm_client_t     search;
    clm_client_t   * psearch = &search;
    int              status;

    memcpy(&(psearch->cli_id), pclid, sizeof(clm_client_id_t));
    search.cli_site = site ;

    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_REMOVECLIENT,
	    site, pclid->clid_i[1], 0x00, 0x00);

    status = -1;
    DLM_SPINLOCK_LOCK( & cli_refs_lock );

    /* It is possible to receive purge requests for unregistered clients 
     */

    if (nclients >= 1)
    {
	DBPRINTF((DBG_CLIENT, "removeclient: site %d clid %x %x\n", site, 
		  db_clidp(pclid)));

	/* Specific client.  */
	p = (clm_client_t **) bsearch((void *) &psearch, (void *) cli_refs,
				      nclients, sizeof(clm_client_t *), 
				      (int(*)(const void*, const void*))icmp);
	if (p == NULL) 
	{
	    bsdlog(LOG_ERR, MSGSTR(NO_CLIENT, 
				   "removeclient: no such client: %x %x"), 
		   db_clidp(pclid));
	}
	else
	{
	    clm_remove_client_by_array_pointer( p );
	    status = 0;
	}
    }


    DLM_SPINLOCK_UNLOCK( & cli_refs_lock );
    return( status );
}


/*
 * Name
 *	add_client_reclock 
 * Description
 * 	Each client has an entry in the client list.  This entry contains 
 *	a list of all locks owned by the said client, whether they are on 
 *	the grant, convert or wait queues.  This routine finds the client 
 *	entry for a given (site, clid) pair and adds the reclock passed onto 
 *	this lock list.  
 * Arguments
 *	site - the site id on which the client is running
 *	pclid - pointer to client id of the client
 *      type - client type: process, group or transaction
 *	reclock - the relock to add to the lock list
 * Returns
 *	0
 * Global Data Affected
 *	cli_refs
 */
int
add_client_reclock(int site,
                   clm_client_id_t *pclid, 
                   clm_client_type_t type,
                   struct reclock *reclock)
{
    clm_client_t  *cl = findclient(site, pclid) ;
	    
    DBPRINTF((DBG_DEADLOCK, "add_client_reclock: site %d clid %x %x\n", site, 
	      db_clidp(pclid)));

    /* If new client, add it */
    if ( cl == NULL )
    {
	if( ! addclient( site, pclid, type, 0, & cl ) )
	{
	    abort("addclient failed");
	}

	CLM_ASSERT( cl != NULL, "client added, but pointer is NULL!" );
    }


    /* If we are reusing a client record created by a purge, promote it to a
     * type which may hold a lock on its queue.
     */

    if (cl->cli_type == CLM_CLI_PURGE)
	cl->cli_type = type ;

    add_queue(&cl->cli_queue, reclock, RL_CL_OFS) ;
    return 0 ;
}


/*
 * Name
 *	delete_client_reclock 
 * Description
 * 	Called when a lock is being discarded.  This finds the client 
 *	associated with this lock and deletes the lock from the clients lock 
 *	list.  This doesn't affect the reclock except for the cl_prev and 
 *	cl_next values. 
 * Arguments
 *	reclock - the lock to remove from the lock list
 * Returns
 *	NOTHING
 * Global Data Affected
 * 	cli_refs
 */
void
delete_client_reclock(struct reclock *reclock)
{
    clm_client_t * cl;


    /* If transaction id defined */
    cl = (reclock->lck.lk_flags & LKM_XID)
	? findclient( reclock->lck.site, 
		      (clm_client_id_t *) & reclock->lck.xid )
	: findclientbypid( reclock->lck.site, reclock->lck.pid );


    /* this may be an orphan lock with no client - if so, do nothing */
    if ( (cl != NULL) && (CLM_CLI_PURGE != cl->cli_type) )
    {
	DBPRINTF((DBG_DEADLOCK, 
		  "delete_client_reclock: site %d clid %x %x reclock 0x%x\n",
		  reclock->lck.site, db_clid(cl->cli_id), 
		  reclock->lck.lockid));


	/* Unix locks don't get put on client queues OR */
	/* Make sure we haven't already removed it.  */
	if ( IS_ON_QUEUE( & reclock->dl.cl ) )
	{
	    remove_queue(&cl->cli_queue, reclock, RL_CL_OFS) ;
	    
	    /* If client is a transaction and the list is empty, delete
	     * the client. 
	     */

	    if ( dlm_list_is_empty( & cl->cli_queue ) 
		 && (CLM_CLI_TRANSACTION == cl->cli_type) )
	    {
		removeclient( reclock->lck.site,
			      (clm_client_id_t *) & reclock->lck.xid );
	    }
	}
    }
}


/*
 * Name
 *	add_timeout_reclock 
 * Description
 *	Add the given reclock to the deadlock timeout queue. 
 * Arguments
 *	rl - the reclock to add to the deadlock timeout queue
 * Returns
 *	NOTHING
 * Global Data Affected
 *	cli_refs
 */
void
add_timeout_reclock(struct reclock *rl)
{
    DBPRINTF((DBG_DEADLOCK, 
	      "add_timeout_reclock: site %d pid %d reclock 0x%x\n",
	      rl->lck.site, rl->lck.pid, rl->lck.lockid));

    /* We don't care about deadlock if the caller specified the no
     * deadlock flag.
     */
    if ( (rl->lck.lk_flags & LKM_NODLCKWT) != 0 )
	return;

    /* make sure we're not already on the deadlock timeout queue */
    if ( IS_ON_QUEUE( & rl->dl.tq ) )
	return;

    /* Stamp it with the current time.  */
    (void) gettimeofday(&(rl->dl.timestamp), NULL);
    rl->dl.checkstamp = rl->dl.timestamp;

    add_queue(&dl_timeout, rl, RL_TQ_OFS) ;
}

/*
 * Name
 *	delete_timeout_reclock
 * Description
 * 	If the given reclock is on the deadlock timeout queue, delete it 
 *	from the queue. 
 * Arguments
 *	rl - the reclock to delete
 * Returns
 *	FALSE if reclock not on timeout queue
 *	TRUE otherwise
 * Global Data Affected
 *	cli_refs
 */
int
delete_timeout_reclock(struct reclock *rl)
{
    DBPRINTF((DBG_DEADLOCK, "delete_timeout_reclock: 0x%x\n",
	      rl->lck.lockid));

    if ( ! IS_ON_QUEUE( & rl->dl.tq ) )
    {
	DBPRINTF((DBG_DEADLOCK,
		  "delete_timeout_reclock: not on list.\n"));
	return FALSE;
    }

    /* The reclock is on the list, delete it.  
     */

    remove_queue(&dl_timeout, rl, RL_TQ_OFS) ;
    return TRUE;
}


/*
 * Name
 *	dumpf_timeout_queue
 * Description
 * 	Prints the contents of the deadlock timeout queue to debug output
 * Arguments
 *	Dump file pointer
 * Returns
 *	NOTHING of consequence
 * Global Data Affected
 *	NONE
 */

unsigned
dlm_print_timeout_queue( char * buf, unsigned limit )
{
    dlm_listnode_t * curr;
    struct reclock * a;
    unsigned         bytes;


    bytes = 0;
    for ( EACH_DLM_LISTNODE( & dl_timeout, curr ) )
    {
	a = DEADLOCK_LINK_TO_RECLOCK( curr );
	bytes += dlm_lock_print( & buf[ bytes ], limit - bytes, a );
    }

    return( bytes );
}




/****************************************************************************/
/* client_apply - Apply a function to each known client
 *
 * Calls the specified function for each of the known clients.  The iteration
 * terminates either when all of the clients have been processed or when the
 * supplied function returns non-zero.
 * 
 * RETURNS
 *
 * If all clients are processed without 'func' returning non-zero, then zero
 * will be returned.  Otherwise, the index+1 of the offending client will be
 * returned.
 * 
 * NOTES
 * 
 * Makes use of the global client table lock.
 * 
 * The function passed to in (i.e., func) must NOT call any other function
 * that uses the global client table lock.  If it does, then the system will
 * likely deadlock.
 */

int
client_apply(
	      /* Function to be applied to each of the clients.
	       */

	      int (*func)(clm_client_t *, void *),


	      /* Extra parameter to be passed to 'func'.
	       */

	      void * arg,


	      /* If non-NULL, then the failure code of 'func' will be
	       * stored here.
	       */

	      int * retp
	    )
{
    clm_client_t * cp;
    unsigned       i;
    int            result;

    
    DLM_SPINLOCK_LOCK( & cli_refs_lock );
    for (i = 0 ; i < nclients ; i++)
    {
	cp = cli_refs[i] ;
	result = (*func)(cp, arg);
	if ( result != 0 )
	{
	    if (retp)
		*retp = result ;

	    DLM_SPINLOCK_UNLOCK( & cli_refs_lock );
	    return i + 1 ;
	}
    }

    DLM_SPINLOCK_UNLOCK( & cli_refs_lock );
    return 0 ;
}


/*
 * reconfig_del_dead_clients
 *
 * scan the client pointer array
 * if any clients are sited at a dead node, delete the client
 * depends on knowing the structure of cli_refs and how removeclient works
 *
 * returns 0
 */

void
reconfig_del_dead_clients( void )
{
    unsigned   i;


    DLM_SPINLOCK_LOCK( & cli_refs_lock );

    for ( i = 0 ; i < nclients ; /* empty */ )
    {
	/* If the node is active, then we can't deem its clients dead.  Skip
	 * the client.
	 */

	if ( clm_node_active( cli_refs[ i ]->cli_site ) )
	{
	    i++;
	}
	else
	{
	    /* depends on removeclient preserving ordering */

	    clm_remove_client_by_array_pointer( & cli_refs[ i ] );
	}
    }

    DLM_SPINLOCK_UNLOCK( & cli_refs_lock );
}
