/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_clmlock.c,v 1.16 2001/11/09 20:04:02 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		begin_lock_clm
 *		blocked_convert
 *		blocked_lock
 *		clear_value_block
 *		clm_convert
 *		clm_count_resources
 *		clm_countres
 *		clm_dumpres
 *		clm_lock
 *		clm_printlock
 *		clm_printlocks
 *		clm_printmode
 *		clm_reset_valueblocks
 *		clmcompatible
 *		cont_lock_clm
 *		convert_lock_clm
 *		convert_locks
 *		dump_value_block
 *		find_incompatible
 *		grant_convert
 *		grant_lock
 *		makename
 *		master_lock_clm
 *		purge_clm
 *		send_bast
 *		send_basts
 *		send_basts_again
 *		valueblock
 *		wake_all_locks
 *		wakeup_clm
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/stddef.h>
#include <linux/signal.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/smp.h>
#include <linux/sysrq.h>
#include <linux/interrupt.h>
#include <asm/semaphore.h>
#include <asm/softirq.h>

#include "dlm.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "clm_data.h"			/* external data definitions */
#include "dlm_kernel.h"
#include "clm_locksw.h"
#include "dlm_version.h"
#include "cltrc.h"
#include <iterator.h>

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */

static int begin_clm_lock_convert(struct transaction *req, struct reclock **ap, int *master) ;
static int begin_clm_lock_new(struct transaction *req, struct reclock **ap) ;
static int cont_lock_clm_convert(struct transaction *req,
                                 struct transaction *resp,
                                 struct reclock *a) ;
static int cont_lock_clm_new(struct transaction *resp, struct reclock *a) ;
static struct resource *
	makename(int site, int pid, clm_trans_t *req, int create_flag) ;

/*
 * Lock level array used to determine down-conversions.
 */
static const int lk_level[] = { 0,	/* LKM_NLMODE */
				1,	/* LKM_CRMODE */
				2,	/* LKM_CWMODE */
				2,	/* LKM_PRMODE */
				3,	/* LKM_PWMODE */
				4 };	/* LKM_EXMODE */

/*
 * these routines all utilize the clm_lockinfo to pass data in and out
 * when returning status to the application
 * It is left to the API to construct
 * a lock status block to be returned to the user
 */

/*
 * Name
 *	master_lock_clm
 * Description
 *	this function is called on a CLM lock request on the master node
 *	for the resource in question
 *
 *	this routine determines whether
 *	  the client record exists,
 *	  any conflict with purges exist
 *	  and if the request is for a new lock or to convert an existing lock,
 *	then calls the appropriate routine.
 * Arguments
 *	req - the lock/convert request
 * Returns
 *	return status from clm_convert or clm_lock
 * 	any data to be returned is put into the incoming request 
 *	structure
 * Global Data Affected
 *	NONE
 */
int
master_lock_clm(clm_msg_t *msg, int flags)
{
	clm_trans_t *req = msg->cm_data;
	struct lockinfo *li = &req->clm_lockreq.lockinfo ;
	clm_client_t *cp ;
	int site ;

	site = msg->cm_node ;
	/*
	 * be sure client & client record exist
	 * create client record if remote request & not convert
	 * if purge of client is in progress, bounce request
	 */
	if ( ! (cp = findclientbypid(li->site, req->clm_authpid)))
	{
	    /*
	     * If this request is from a local client and the client no longer
	     * exists, then it was probably being processed by a remote
	     * directory at the time the client was purged.  Drop it.
	     */
	    if (li->site == OurSiteID)
	    {
		DBPRINTF((DBG_CLMLOCK,
			"clm_master_lock: Drop req from old client 0x%x %s",
			req, printbin(&req->clm_lockreq.nameinfo.name[0],
			req->clm_lockreq.nameinfo.namelen)));
		clm_lock_drop(req, flags) ;
		return 0 ;
	    }
	    if ( ! addclientbypid( li->site, req->clm_authpid,
				   CLM_CLI_PROCESS, NULL ) )
	    {
		DBPRINTF((DBG_LOCK, "can't add client %d/%d\n",
				li->site, req->clm_pid)) ;

		/* if local lock no directory to delete */
		if (flags & DLM_RES_CREATE &&
			! (req->clm_lockinfo.lk_flags & LKM_LOCAL))
		{
		    struct resource res;

		    res.name = &req->clm_lockreq.nameinfo.name[0];
		    res.namelen = req->clm_lockreq.nameinfo.namelen;
		    res.type = DLM_RES_VMS;
		    res.rsrc_flags = 0;

		    /* couldn't create, tell directory */
		    update_directory(&res, DIR_DELETE, RESRPRI, NULL );
		}
		
		/* here if we couldn't create client record */
		msg = cccp_msg_duplicate( msg );
		req = msg->cm_data ;
		msg->cm_node = site ;
		req->clm_direction = PTI_RESPONSE ;
		req->clm_status = DLM_SYSERR ;

		clm_queue_respmsg(msg) ;
		return 0 ;
	    }
	}
	else
	if (cp->cli_state & CLI_CODE)	/* any purge in progress ? */
	{
	    if (flags & DLM_RES_CREATE) {
		struct resource res;

		res.name = &req->clm_lockreq.nameinfo.name[0];
		res.namelen = req->clm_lockreq.nameinfo.namelen;
		res.type = DLM_RES_VMS;
		res.rsrc_flags = 0;

		/* couldn't create, tell directory */
		update_directory(&res, DIR_DELETE, RESRPRI, NULL );
	    }

	    DBPRINTF((DBG_LOCK|DBG_PURGE,
				"can't add lock for %d/%d - purging\n",
				li->site, req->clm_pid)) ;
	    msg->cm_port = PORT_BOUNCE ;	/* try again later - CRUDE */
	    return TRUE ;
	}

	if (req->clm_lockinfo.lk_flags & LKM_CONVERT)
		return clm_convert(msg);
	else
		return clm_lock(msg, flags);
}
 
/*
 * Name
 *	clm_lock
 * Description
 *	This is the routine that will be called on the master node for the
 *	resource that the lock request is being made against regardless of
 *	whether the master is local or remote to the requesting client.
 *	
 *	Check to see if the resource in question is supposed to be mastered
 *	here.  If it isn't, then calculate the directory node id and return
 *	that.
 *	
 *	Create or update the master copy of the resource being referenced.
 *
 *	If the request was local, then a lock record was allocated by the
 *	begin_lock routine, otherwise allocate a local copy of the reclock
 *	and map it to the slave lockid.
 *
 * Arguments
 *	req - the lock request
		Input structure: clm_lockinfo
 * 		Output structure: clm_lockinfo
 * Returns in req->clm_status: 
 *	DLM_DENIED_NOLOCKS	resource failure
 *	DLM_IVLOCKID		invalid lockid
 *	DLM_NORMAL		lock granted
 * 	DLM_NOTQUEUED		LKM_NOQUEUE specified and request would block
 * Returns: 
 *	0 if successful
 *	Site id of directory node if resource mastered elsewhere.
 * Global Data Affected
 *	NONE
 */
int
clm_lock(clm_msg_t *msg, int flags)
{
	clm_trans_t		*req = msg->cm_data;
	struct reclock		*reclock;
	struct lockinfo		*lip = &req->clm_lockreq.lockinfo;
	register int		pid = lip->pid;
	register int		site = lip->site;
	struct resource		*rp;
	uint			lockid;
	rl_info_t		rl_entry;

	STAMP(req);
	DBPRINTF((DBG_CLMLOCK, "clm_lock: reqmode=0x%x\n", lip->reqmode));
	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_CLM_LOCK,
		req->clm_pid, req->clm_sequence, req->clm_lockinfo.reqmode,
		req->clm_lockinfo.lk_flags);

	/*
	 * Call makename to update the local LRDB.  Create the resource if
	 * it doesn't exist.  Update the refcount if it does.
	 */
	STAMP(req);
	rp = makename(site, pid, req, flags);
	if (rp == NULL) {
		struct resource res;

		/*
		 * Local Lock: Reached node with LKM_FINDLOCAL set. 
		 *             If (resource not found && !DLM_RES_CREATE)
		 *                send request node a NACK
		 */
		if ((req->clm_lockinfo.lk_flags & LKM_FINDLOCAL) &&
		    !(flags & DLM_RES_CREATE)) {
		        if (0 != clmll_send_nack_response (msg)) {
		                /* NACK requests failed */
			        DBPRINTF((DBG_LOCAL,
					  "clm_lock: Nack request failed\n"));
			}
			goto done;
		}
		
		res.name = &req->clm_lockreq.nameinfo.name[0];
		res.namelen = req->clm_lockreq.nameinfo.namelen;
		res.type = DLM_RES_VMS;
		res.rsrc_flags = 0;

		/* here if resource table full */
		msg = cccp_msg_duplicate( msg );
		req = msg->cm_data;
		msg->cm_node = site;
		req->clm_direction = PTI_RESPONSE;
		req->clm_status = DLM_DENIED_NOLOCKS;

		/* if local lock no directory to delete */
		if (!(req->clm_lockinfo.lk_flags & LKM_LOCAL)) {
		        /* couldn't create, tell directory */
			update_directory(&res, DIR_DELETE, RESRPRI, NULL );
		}
		

		clm_queue_respmsg(msg);
		goto done;
	} else if ((struct resource *)(-1) == rp) {

		/*
		 * Local Lock: Found 2ndary copy of resource with
		 *             LKM_FINDLOCAL set, send NACK to requesting node
		 */
		if (req->clm_lockinfo.lk_flags & LKM_FINDLOCAL) {
		        if (0 != clmll_send_nack_response (msg)) {
		                /* NACK requests failed */
			        DBPRINTF((DBG_LOCAL,
					  "clm_lock: Nack request failed\n"));
			}
			goto done;
		}
		
		/*
		 * here if we are not master of the resource, so determine
		 * the site id of the directory node so we can send a redirect
		 */
		site = clm_direct(&req->clm_lockreq.nameinfo.name[0],
				req->clm_lockreq.nameinfo.namelen);

		DBPRINTF((DBG_DIRECT,
			"clm_lock: Not the master of %s.  Redirect to %d\n",
			printbin(&req->clm_lockreq.nameinfo.name[0],
			req->clm_lockreq.nameinfo.namelen), site));

		if (!site) {
			bsdlog(LOG_CRIT, MSGSTR(NULL_RET,
					"clm_lock:clm_direct() returned NULL"));
			return -1;
		}
		return site;
	}

	/*
	 * If the resource is migrating, bounce message to apply back
	 * pressure.  Sender will retry.
	 */
	if (rp->rsrc_flags & (RSRC_FRZEN | RSRC_STUCK | RSRC_UPDATING))
	{
	    decr_delete_resource(rp, 0) ;
	    msg->cm_port = PORT_BOUNCE ;
	    return TRUE ;
	}
	STAMP(req);

	/* update RLDB if not doing LOCAL_LOCK */
	if (!(req->clm_lockinfo.lk_flags & (LKM_LOCAL | LKM_FINDLOCAL))) {
		rl_entry.rl_name = rp->name;
		rl_entry.rl_namelen = rp->namelen;
		rl_entry.rl_type = rp->type;
		rl_entry.rl_node = OurSiteID;
		rl_update(&rl_entry);
	}

	STAMP(req);
	/*
	 * request is not going to be re-directed.  call cccp_msg_duplicate() to insure
	 * that we have a writeable copy of the message, reset the req pointer
	 * and enqueue the response back to the secondary.
	 */

	msg = cccp_msg_duplicate( msg );
	req = msg->cm_data;
	msg->cm_node = site;
	req->clm_direction = PTI_RESPONSE;
	clm_queue_respmsg(msg);

	STAMP(req);

	clmm_cache(rp, msg->cm_node);	/* saves resource & node for migr */

	STAMP(req);
	/*
	 * At this point, we know that we are the resource master that
	 * is going to handle this request.  Set the master field in the
	 * route information of the request so when the secondary lock
	 * manager reads the response, it'll know where it was mastered.
	 */
	req->clm_rt_master = OurSiteID;

	/*
	 * Resource Statistics:
	 *        cs_requests - lock request was made
	 *        cs_same     - from same node as last request
	 *        lastnode    - update lastnode to be current node
	 */
	rp->stats.cs_requests++;
	if (rp->lastnode == req->clm_lockreq.lockinfo.site) {
	        rp->stats.cs_same++;
	}
	rp->lastnode = req->clm_lockreq.lockinfo.site;


#ifdef DBPRINT
	/* verify that a BAST was registered */
	if (valid_handle(&req->clm_route.basth)) {
		DBPRINTF((DBG_AST, "Registering BAST 0x%x\n",
			req->clm_route.basth));
	}
#endif

	/*
	 * if the lock request originated on a different node, then the master
	 * node has to create a local copy of the lock and map the remote
	 * lockid to the local one.
	 */
	if (req->clm_lockreq.lockinfo.site != OurSiteID)
	{
		reclock = get_le(DLM_RES_VMS);
		if (reclock == NULL) {
			req->clm_status = DLM_DENIED_NOLOCKS;
			decr_delete_resource(rp, 0) ;
			goto done;
		}
		if (addmap(req->clm_lockreq.lockinfo.lockid, 	
			reclock->lck.lockid,
			req->clm_lockreq.lockinfo.site) < 0)
		{
			req->clm_status = DLM_DENIED_NOLOCKS;
			decr_delete_resource(rp, 0) ;
			free_le(DLM_RES_VMS, reclock);
			goto done;
		}
		/*
		 * Resource Statistics:
		 *        cs_remote - lock request from remote node
		 */
		rp->stats.cs_remote++;

	} else {
		/*
		 * Resource is being mastered on requesting node.
		 * We already allocated a reclock structure before calling 
		 * the master layer so fetch it.
		 */
		reclock = find_reclock(DLM_RES_VMS, lip->lockid);
		
		/*
		 * Resource Statistics:
		 *        cs_local - lock request from local node
		 */
		rp->stats.cs_local++;
		
	}

	if (reclock == NULL) {
		req->clm_status = DLM_IVLOCKID;
		decr_delete_resource(rp, 0) ;
		goto done;
	}

	/*
	 * At this point, we need to copy all of the information from the
	 * lockinfo structure in the request into the lockinfo structure in
	 * the newly acquired reclock, but we don't want to carry over
	 * certain flags and we want to keep the local lockid.
	 */
	lockid = reclock->lck.lockid;
	req->clm_lockreq.lockinfo.state &= ~INFLIGHT_WAITING;
        req->clm_lockinfo.request = reclock->lck.request;
	
	reclock->lck = req->clm_lockreq.lockinfo;

	/* restore lockid */
	reclock->lck.lockid = lockid;

	/* copy some more needed information into the reclock structure */
	reclock->lck.rh = rp->rh;
	reclock->route = req->clm_route;

	/* pass resource handle back to secondary */
	req->clm_lockreq.lockinfo.rh = rp->rh;

	/*
	 * The master node needs to know what the corresponding lockid is
	 * on the secondary node.
	 */
	if (req->clm_lockreq.lockinfo.site != OurSiteID)
	{
		reclock->lck.remlockid = req->clm_lockreq.lockinfo.lockid;
	}
	else
		reclock->lck.remlockid = 0;

	STAMP(req);
	/*
	 * if the request is for a NLMODE lock OR the wait and convert queues 
	 * are empty AND the request is compatible, grant the request 
	 */
	if ((reclock->lck.reqmode == LKM_NLMODE) ||
	    (( dlm_list_is_empty( & rp->lockqs[RLQ_CONV] )
	      && dlm_list_is_empty( & rp->lockqs[RLQ_WAIT] )) &&
	     find_incompatible(rp, reclock->lck.reqmode, NULL) == NULL)) {
		/*
		 * here if compatible, move lock to grant queue, update the
		 * valueblock to be returned
		 */
		req->clm_status = grant_lock(rp, reclock, 
				req->clm_lockreq.lockinfo.lkvalue);
		reclock->lck.lk_flags &= ~LKM_BLOCK;

		/* copy the lock sequence number to be returned to secondary */
		req->clm_lockinfo.seqnum = reclock->lck.seqnum;

		/*
		 * Resource Statistics:
		 *        cs_compat - lock request not placed on wait queue
		 */
		rp->stats.cs_compat++;

	} else {

	        /*
		 * Resource Statistics:
		 *        cs_incompat - lock request placed on wait queue
		 */
		rp->stats.cs_incompat++;

		/* here if request was not grantable */
		if (reclock->lck.lk_flags & LKM_NOQUEUE) {
			/*
			 * here if request was specified with NOQUEUE.  Free
			 * reclock, resource reference and return status.
			 */
			DBPRINTF((DBG_CLMLOCK,
				"clm_lock: rejecting lock because NOQUEUE"));

			decr_delete_resource(rp, 0) ;	/* may delete */
			free_le(DLM_RES_VMS, reclock);
			req->clm_status = DLM_NOTQUEUED;	
			goto done;
		}

		/* Add lock to proper queues and set blocked states */
		blocked_lock(rp, reclock);

		/* copy the lock sequence number to be returned to secondary */
		req->clm_lockinfo.seqnum = reclock->lck.seqnum;

		req->clm_status = DLM_WORKING;
	}

	check_handles(rp, 0) ;

#ifdef DBPRINT
	clm_dumpres(rp);
#endif

	DBPRINTF((DBG_CLMLOCK, "clm_lock: status %d lockid 0x%x\n", 
		req->clm_status, req->clm_lockreq.lockinfo.lockid));
/*
 * MUST NOT REFERENCE rp here
 */
done:
	STAMP(req);
	return 0;
}

/*
 * Name
 *	begin_lock_clm
 * Description
 * 	this routine handles the initial processing of a lock request before
 *	it is forwarded to the distributed layer.
 * 		o if the request is a conversion, try to reject it locally
 *   		  else, send to the master layer
 * 		o if request is a lock create, allocate local lockid 
 *		  before sending to the master layer
 * Arguments
 *	req - the lock request
 * 		Input Structure: clm_lockinfo
 * 		Output Structure: clm_lockinfo
 * Returns
 *	 0 for success
 * 	-1 for failure
 * Global Data Affected
 *	NONE
 */
int
begin_lock_clm(clm_trans_t *req)
{
	struct lockinfo *lip = &req->clm_lockinfo;
	struct reclock *a ;
	dlm_stats_t stat;
	int master = 0;
	clm_client_t *cp ;
	int rc ;

	STAMP(req);
	DBPRINTF((DBG_FUNC, "enter begin_lock_clm\n"));
	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_BEGIN_LOCK_CLM,
		req->clm_pid, req->clm_lockinfo.lockid,
		req->clm_lockinfo.reqmode, req->clm_lockinfo.lk_flags);

	/* remove need for clinfo in API - client doesn't know nodeid */
	req->clm_lockinfo.site = OurSiteID;

	/*
	 * if creating a lock and PROC_OWNED is set, then authpid should
	 * be the client's actual process id
	 */
	if ((req->clm_lockinfo.lk_flags & LKM_PROC_OWNED) &&
			(~req->clm_lockinfo.lk_flags & LKM_CONVERT)) {
		req->clm_authpid = req->clm_pid;
	}

	/* set ownership value for lock request */
	req->clm_lockinfo.pid = req->clm_authpid;

	/* but someone has to get the AST */
	req->clm_lockinfo.realpid = req->clm_pid;

	/*
	 * be sure a client purge is not in progress here
	 * if so apply back pressure (crude version used here,
	 * should queue request on client
	 */
	if ( ! (cp = findclientbypid(OurSiteID, req->clm_pid)))
	{
	    DBPRINTF((DBG_LOCK, "PID %d has no client record\n", req->clm_pid)) ;
	    req->clm_status = DLM_BADARGS ;
	    goto badreq ;
	}
	if (cp->cli_state & CLI_CODE)		/* purge in progress? */
	{
	    DBPRINTF((DBG_LOCK|DBG_PURGE, "PID %d purge in progress\n",
							req->clm_pid)) ;
	    req->clm_status = DLM_DENIED_GRACE_PERIOD ;
	    goto badreq ;
	}
#ifdef DBPRINT
	/* verify that a BAST was registered */
	if ((lip->lk_flags & LKM_CONVERT) &&
					valid_handle(&req->clm_route.cbasth)){
		DBPRINTF((DBG_AST, "Registering BAST 0x%x for convert\n",
			req->clm_route.cbasth));
	} else if (valid_handle(&req->clm_route.basth)) {
		DBPRINTF((DBG_AST, "Registering BAST 0x%x for lock request\n",
			req->clm_route.basth));
	}
#endif

	if (lip->lk_flags & LKM_CONVERT)
	    rc = begin_clm_lock_convert(req, &a, &master) ;
	else 
	    rc = begin_clm_lock_new(req, &a) ;
	switch (rc)
	{
	case 0:	break ;
	case 1: goto badreq ;
	case -1: return -1 ;
	default:
	    abort("bad return from begin") ;
	    break ;
	}

	req->clm_status = DLM_NORMAL;	/* is this needed? */

	/*
	 * send reply back to client  - don't free the request because it is
	 * enqueued on the inflight queue to be matched when the response comes
	 * back
	 */
	sendudp(req);

	/*
	 * If LKM_FINDLOCAL set then do a clmll_broadcast, if LKM_LOCAL set
	 * create lock locally, else call pti_call
	 * Call clmll_broadcast () which allocates LLB structure and request
	 * a lock on resource from every node in the cluster.
	 */
	if (lip->lk_flags & LKM_FINDLOCAL) {
                /* Request lock on resource on every node w/ LKM_FINDLOCAL */
                DBPRINTF((DBG_LOCAL,
			  "begin_lock_clm: get %d lock on resource (%s) "
			  "created Locally\n",
			  req->clm_lockinfo.reqmode,
			  printbin(req->clm_lockreq.nameinfo.name,
				   req->clm_lockreq.nameinfo.namelen)));
		
		if ((stat = clmll_broadcast (req)) != 0) {
		        /* broadcasting lock requests failed */
		        DBPRINTF((DBG_LOCAL, 
				  "begin_lock_clm: clmll_broadcast on "
				  "resource (%s) failed\n",
				  printbin(req->clm_lockreq.nameinfo.name,
					   req->clm_lockreq.nameinfo.namelen)
				  ));

			/* clear UDP flag and send AST */
			req->clm_respflags &= ~RESP_UDP;
			req->clm_respflags |= RESP_AST;	
			clm_reply(req, 0);
		}
	} else if (lip->lk_flags & LKM_LOCAL) {
                /* goes straight to master lock funcion */
                DBPRINTF((DBG_LOCAL,
			  "begin_lock_clm: create resource (%s) locally on "
			  "node %d\n",
			  printbin(req->clm_lockreq.nameinfo.name,
				   req->clm_lockreq.nameinfo.namelen),
			  OurSiteID));
		if ((stat = clmll_create (req)) != 0) {
                        /* local lock request failed */
		        DBPRINTF((DBG_LOCAL, 
				  "begin_lock_clm: local lock request on "
				  "resource (%s) failed\n",
				  printbin(req->clm_lockreq.nameinfo.name,
					   req->clm_lockreq.nameinfo.namelen)));

			/* clear UDP flag and send AST */
			req->clm_respflags &= ~RESP_UDP;
			req->clm_respflags |= RESP_AST;	
			clm_reply(req, 0);

			/* free the reclock */
			free_le(DLM_RES_VMS, a);
		}
	} else {
                stat = pti_call(req, master);
	} 
    
	DBPRINTF((DBG_FUNC, "leave begin_lock_clm\n"));
	return stat;

badreq:
	req->clm_respflags &= ~RESP_AST;
	clm_reply(req, 0);
	return -1;
}

/*
 * begin_clm_lock_convert
 *
 * processing for begin_clm_lock() specific to conversion requests
 *
 * checks for existence of reclock, migration in progress, and
 *   (some) valid state of reclock
 *
 * returns
 *	-> reclock via ap
 *	nodeid of master via pmaster
 *
 *	0 if OK to continue processing
 *	1 if error but reply not yet sent to client
 *	-1 if error and reply sent to client
 */
static int
begin_clm_lock_convert(
    struct transaction *req,	/* client transaction request */
    struct reclock **ap,	/* place to store pointer to reclock found */
    int *pmaster)		/* place to store nodeid of resource master */
{
    struct lockinfo *lip = &req->clm_lockinfo;
    struct reclock *a;
    struct resource *rp;

    DBPRINTF((DBG_CLMLOCK, "Convert 0x%x to %d\n",
			 req->clm_lockinfo.lockid, req->clm_lockinfo.reqmode));
		
    /* clear the LKM_LOCAL & LKM_FINDLOCAL flags before convert */
    lip->lk_flags = ((lip->lk_flags & ~LKM_LOCAL) & ~LKM_FINDLOCAL);

    /*
     * Conversion requests can only be made on an existing,
     * valid lock record.  See if it we can find it.
     */
    a = find_reclock(DLM_RES_VMS, lip->lockid) ;
    *ap = a ;
    if (! a || !mylock(a, req) || a->lck.state & LOCK_GONE)
    {
	req->clm_status = DLM_IVLOCKID;	
	req->clm_lockinfo.lk_flags &= ~LKM_BLOCK;
	return 1 ;
    }

    /*
     * If INFLIGHT is still set from last request, then client is being
     * awfully pushy.  So push back.
     */
    if (a->lck.state & (INFLIGHT_CONVERTING | INFLIGHT_WAITING))
    {
	req->clm_status = DLM_DENIED_GRACE_PERIOD ;
	return 1 ;
    }

    rp = resptr(a->lck.rh);
    if (!rp)
    {
	bsdlog(LOG_ERR, MSGSTR(CANT_CONVERT,
			       "begin_clm_lock_convert: can't get resource "
			       "for converting lock rh=0x%08lx\n"),
				a->lck.rh.rh );
	abort("begin_clm_lock_convert");
    }
    /*
     * Apply back pressure if lock migrating
     */
    if (rp->rsrc_flags & (RSRC_FRZEN | RSRC_STUCK | RSRC_UPDATING))
    {
	req->clm_status = DLM_DENIED_GRACE_PERIOD ;
	return 1 ;
    }

    if (a->lck.lk_flags & LKM_XID)
    {
	/* Preserve XID info for convert */
	req->clm_lockinfo.lk_flags |= LKM_XID;
	memcpy(&lip->xid, &a->lck.xid, sizeof(clm_client_id_t));
    }

    /*
     * Conversion requests can not be made on a lock that has
     * not been granted yet.
     */
    if (a->lck.state & LOCK_WAITING)
    {
	ast_handle_t asth_save;
	/*
	 * In this situation, client expects both UDP and
	 * AST response without going through cont_remote...
	 * so we have to send UDP response from here
	 */
	req->clm_status = DLM_CVTUNGRANT;	
	/*
	 * sendast now looks up the handle from the reclock, so
	 * we have to fake it out or it'll call the wrong AST.
	 * 4/17/94: changed calling order so that clm_reply
	 * is called before sendast so that if the LKM_SYNCSTS
	 * flag is used, the call to sendsync will work
	 * correctly
	 */
	clm_reply(0, req);
	asth_save = a->route.asth;
	a->route.asth = req->clm_asth;
	sendast(req);
	a->route.asth = asth_save;
	clm_free_trans(req);
	return -1;
    }

    /*
     * Conversion requests can not be made on a lock that is
     * already converting.
     */
    if (a->lck.state & LOCK_CONVERTING)
    {
	ast_handle_t asth_save;
	/*
	 * In this situation, client expects both UDP and
	 * AST response without going through cont_remote...
	 * so we have to send UDP response from here
	 */
	req->clm_status = DLM_DENIED;	
	/*
	 * sendast now looks up the handle from the reclock, so
	 * we have to fake it out or it'll call the wrong AST.
	 * 4/17/94: changed calling order so that clm_reply
	 * is called before sendast so that if the LKM_SYNCSTS
	 * flag is used, the call to sendsync will work
	 * correctly
	 */
	clm_reply(0, req);
	asth_save = a->route.asth;
	a->route.asth = req->clm_asth;
	sendast(req);
	a->route.asth = asth_save;
	clm_free_trans(req);
	return -1;
    }

    /*
     * set lock state to LOCK_CONVERTING in case someone tries
     * to unlock it before the master's response is received
     */
    a->lck.state |= INFLIGHT_CONVERTING;
    /*
     * Local copy of resource probably knows where the master is
     */
    *pmaster = rp->master ? rp->master : OurSiteID;

    /*
     * Copy resource nameinfo into request for lookup in case
     * master has moved.  We could optimize this by not doing
     * an unnecessary copy for a locally mastered resource.
     */
    req->clm_lockreq.nameinfo.namelen = rp->namelen;
    resname_bcopy(rp->name, &req->clm_lockreq.nameinfo.name[0],
					rp->namelen);
    DBPRINTF((DBG_CLMLOCK,"convert: mst = %d\n", *pmaster));
    return 0 ;
}

/*
 * begin_clm_lock_new
 *
 * processing for begin_clm_lock() specific to new lock requests
 *
 * allocates reclock, marks state, and links lock to client struct
 *
 * returns
 *	-> reclock via ap
 *
 *	0 if OK to continue processing
 *	1 if error
 */

static int
begin_clm_lock_new(
    struct transaction *req,	/* client transaction request */
    struct reclock **ap)	/* place to store pointer to reclock found */
{
    struct lockinfo *lip = &req->clm_lockinfo;
    struct reclock *a;

    /* 
     * allocate reclock so we can return lockid to the client
     * before forwarding request to the master layer
     */
    a = get_le(DLM_RES_VMS);
    *ap = a ;
    if ( ! a)
    {
	req->clm_status = DLM_DENIED_NOLOCKS;
	req->clm_lockinfo.lk_flags &= ~LKM_BLOCK;
	return 1 ;
    }

    /*
     * set lock state to LOCK_WAITING in case someone tries
     * to unlock it before the master's response is received
     * also set the pid in lockinfo in case client tries
     * to convert lock before primary's response
     */
    a->lck.state |= INFLIGHT_WAITING;
		
    /* set ownership value for lock */
    a->lck.pid = req->clm_authpid;

    a->lck.site = OurSiteID ;

    /* if XID is defined */
    if (lip->lk_flags & LKM_XID)
    {
	a->lck.lk_flags |= LKM_XID;

	/* Store XID in lockinfo */
	memcpy(&a->lck.xid, &lip->xid, sizeof(lip->xid));
    }
    xid_add_client_reclock(a) ;

    req->clm_lockinfo.lockid = a->lck.lockid;
    return 0 ;
}

/*
 * Name
 *	cont_lock_clm
 * Description
 * 	Continuation routine for lock request returned from master
 * 	At this point we have received status back from the master.
 * 	Set up the lock locally if this node was not the master.
 *	Send any notification to the client.
 * Arguments
 *      req - the request that was sent to the primary
 * 		Input Structure: clm_lockinfo
 *      resp - the response from the primary
 * 		Output Structure: clm_lockinfo
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
cont_lock_clm(clm_trans_t *req, clm_trans_t *resp)
{
	struct resource	*rp = NULL;
	struct reclock	*a = NULL;

	DBPRINTF((DBG_FUNC, "enter cont_lock_clm: rh=0x%08lx id=0x%x stat=%d\n", 
		 resp->clm_lockreq.lockinfo.rh.rh, 
		 resp->clm_lockreq.lockinfo.lockid,
		 resp->clm_status));
	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_CONT_LOCK_CLM,
		resp->clm_pid, resp->clm_sequence, resp->clm_lockinfo.lockid,
		resp->clm_status);

	/*
	 * If this request is from a local client and the client no longer
	 * exists, then it was probably being processed by a remote directory
	 * at the time the client was purged.  Drop it.
	 */
	if (NULL == findclientbypid(OurSiteID, resp->clm_authpid)) {

		/* purge will not clean this lock if it's not on a resource */
		if ((a = find_reclock(DLM_RES_VMS,
					resp->clm_lockreq.lockinfo.lockid)) &&
					mylock(a, resp)) {
			DBPRINTF((DBG_LOCK|DBG_PURGE,
				"Freeing remote inflighted reclock 0x%x r=%s",
				resp->clm_lockreq.lockinfo.lockid,
				printbin(&resp->clm_lockreq.nameinfo.name[0],
				resp->clm_lockreq.nameinfo.namelen)));
			free_le(DLM_RES_VMS, a);
		}

		DBPRINTF((DBG_CLMLOCK,
			"cont_lock_clm: Drop resp from leftover client 0x%x %s",
			resp, printbin(&resp->clm_lockreq.nameinfo.name[0],
			resp->clm_lockreq.nameinfo.namelen)));
		if (req)
			clm_free_trans(req);
		return;
	}

	/* some response type can skip to the end and send answers to client */
	if (resp->clm_status != DLM_NORMAL && resp->clm_status != DLM_WORKING
			&& resp->clm_status != DLM_VALNOTVALID
			&& resp->clm_status != DLM_DENIED_NOLOCKS
			&& resp->clm_status != DLM_NOTQUEUED
			&& resp->clm_status != DLM_DEADLOCK) {
		goto done;
	}
	/* only handle NOTQUEUED locally if mastered elsewhere */
	if (resp->clm_status == DLM_NOTQUEUED &&
				resp->clm_rt_master == OurSiteID)
		goto done;

	/*
	 * first pick up reclock and resource
	 */
	a = find_reclock(DLM_RES_VMS, 
			 resp->clm_lockreq.lockinfo.lockid);
	if (a == NULL) {
		bsdlog(LOG_ERR, MSGSTR(CANT_RELOCK, 
			"cont_lock_clm: ERROR: can't find reclock "
			"for blocked lock id=0x%x"),
			resp->clm_lockreq.lockinfo.lockid);
		resp->clm_status = DLM_IVLOCKID;
		goto done;
	}

	/* if converting, save old BAST, new one is in cbasth */
	if (resp->clm_lockinfo.lk_flags & LKM_CONVERT)
		resp->clm_route.basth = a->route.basth;
	a->route = resp->clm_route; /* copy route info, ASTs, etc */

	/* if our response from the master got NOLOCKS, then just free it */
	if (resp->clm_status == DLM_DENIED_NOLOCKS)
	{
		free_le(DLM_RES_VMS, a);
		goto done;
	}

	/*
	 * If the current request was mastered on the local node, then any
	 * state information in the local LRDB should already be up to date.
	 * Skip to the end...
	 */
	if (resp->clm_rt_master == OurSiteID) {
		a->lck.state &= (~INFLIGHT_CONVERTING & ~INFLIGHT_WAITING);
		goto done;
	}

	/*
	 * PTI_RESPONSES should only be read if INFLIGHT flags are set
	 * We need to handle the situation where a master copy of a resource
	 * can migrate to a different node, grant a request on this node, and
	 * send us the ASYNC grant message before we get the PTI_RESPONSE
	 * message from the old master node.
	 */
	if (a->lck.state & (INFLIGHT_CONVERTING | INFLIGHT_WAITING)) {
		/* here if we get an async grant before a PTI_RESPONSE */
		if (resp->clm_direction == PTI_ASYNC) {
			bsdlog(LOG_ERR,
				"ASYNC for 0x%x recv'd before response",
				a->lck.lockid);
		}
	} else if (resp->clm_direction == PTI_RESPONSE) {
		/* if we get a response and inflight is clear, toss it */
		bsdlog(LOG_ERR, "Inflight clear when  0x%x response recv'd",
				a->lck.lockid);
		resp->clm_status = DLM_VOID;
		goto done;
	}

	/*
	 * We've gotten our answer from the master, so the request in no
	 * longer inflight -> clear INFLIGHT_ bits from state
	 */
	a->lck.state &= (~INFLIGHT_CONVERTING & ~INFLIGHT_WAITING);

	/* 
	 * we're now ready to handle the transaction locally.
	 * we'll pick up the state from the master and put
	 * the lock into the same state locally
	 */
	if (resp->clm_lockinfo.lk_flags & LKM_CONVERT)
	    cont_lock_clm_convert(req, resp, a) ;
	else
	    cont_lock_clm_new(resp, a) ;
	check_handles(rp, 0) ;

done:
        if (a && a->lck.request)
            submit_cancels(a);
#ifdef DBPRINT
        if (rp)
	    clm_dumpres(rp);
#endif
	resp->clm_respflags = RESP_AST;

	clm_reply(req,resp);
	DBPRINTF((DBG_FUNC, "leave cont_lock_clm\n"));
}

/*
 * cont_lock_clm_convert
 *
 * processing specific for conversion requests for cont_lock_clm
 * (client processing after receipt of response from master)
 *
 * can abort() if truly invalid state
 * sets master nodeid in reclock
 * handles conversion granted or blocked state from master
 *
 * returns 1 if error, 0 if OK
 */
static int
cont_lock_clm_convert(
	struct transaction *req,	/* request from user client */
	struct transaction *resp,	/* response to client being filled in */
	struct reclock *a)		/* lock being converted */
{
    int state;
    int mode;
    int orphan;
    struct resource *rp ;
    rl_info_t rl_entry ;

    /*
     * Here if completing a conversion request.  Lock record
     * should have a valid resource handle.
     */
    DBPRINTF((DBG_CLMLOCK, "cont_lock_clm: CONVERT\n"));
    if (a->lck.rh.rh == 0)
    {
	bsdlog(LOG_CRIT, MSGSTR(NULL_HANDLE,
			"cont_lock_clm: NULL resource handle for 0x%x"),
			resp->clm_lockreq.lockinfo.lockid);
	req->clm_status = DLM_IVRESHANDLE;	
	abort("cont_lock_clm");
	return 1 ;
    }
    if (!(rp = resptr(a->lck.rh)))
    {
	resp->clm_status = DLM_IVRESHANDLE;	
	bsdlog(LOG_CRIT, MSGSTR(CANT_GET_RES, 
			"cont_lock_clm: can't get resource "
			"for converting lock rh=0x%08lx\n"),
			a->lck.rh.rh );
	abort("cont_lock_clm");
	return 1 ;
    }

    /*
     * update resource's notion of who is master - if master is
     * local, then this response came from an old master copy
     * that have moved here since it was sent.
     */
    if (rp->master)
    {
	CLM_ASSERT((resp->clm_rt_master &&
				resp->clm_rt_master != OurSiteID),
				"cont_lck: Set master local");
	rp->master = resp->clm_rt_master;
    }

    /* update RLDB */
    rl_entry.rl_name = rp->name;
    rl_entry.rl_namelen = rp->namelen;
    rl_entry.rl_node = rp->master;
    rl_entry.rl_type = rp->type;
    rl_update(&rl_entry);

    /*
     * blocked convert - a lock's mode, state, and
     * orphan bit have to be preserved
     */
    mode = a->lck.mode;
    state = a->lck.state;
    orphan = a->lck.lk_flags & LKM_ORPHAN;
    resp->clm_lockinfo.rh.rh = a->lck.rh.rh;
    resp->clm_lockinfo.request = a->lck.request;
    a->lck = resp->clm_lockinfo;
    a->lck.mode = mode;	/* restore lock state */
    a->lck.lk_flags |= orphan;
    a->lck.state = state;

    if (resp->clm_status == DLM_WORKING)
    {
	blocked_convert(rp, a);
    }
    else
    if (resp->clm_status == DLM_DEADLOCK)
    {
	/*
	 * here if master sent us DEADLOCK response.
	 * update local resource copy
	 */
	check_handles(rp, 0) ;
	/*
	 * Migration may have syncd us up with the master
	 * copy before the deadlock response got here.  Check
	 * lock state to ensure it is on convert queue before
	 * trying to move it.
	 */
	if (a->lck.state & LOCK_CONVERTING) {
	    remove_queue(&rp->lockqs[RLQ_CONV], a, RL_RQ_OFS);
	    a->lck.state &= ~LOCK_CONVERTING;
	    a->lck.state |= LOCK_GRANTED;
	    add_queue(&rp->lockqs[RLQ_GRANT], a, RL_RQ_OFS);
	    delete_timeout_reclock(a);
	    unset_lock_timer(a);
	    a->lck.reqmode = a->lck.mode;
	    check_handles(rp, 0) ;
	}
    }
    else
    {
	/*
	 * we can get here for two reasons,
	 * either a convert request succeeded,
	 * or a blocked convert was granted 
	 */
	a->lck.reqmode = resp->clm_lockinfo.reqmode;
	/*
	 * assume that we're getting the correct
	 * status back from the primary and pass that
	 * back to the client rather than status from
	 * grant_lock
	 */
	grant_convert(rp, a, resp->clm_lockinfo.lkvalue);
    }
    return 0 ;
}

/*
 * cont_lock_clm_new
 *
 * processing specific for new lock requests for cont_lock_clm
 * (client processing after receipt of response from master)
 *
 * if no resource structure yet, create the resource
 * sets master nodeid in reclock
 * handles lock granted or blocked state from master
 *
 * returns 1 if error, 0 if OK
 */

static int
cont_lock_clm_new(
    struct transaction *resp,	/* response from master/sent to user */
    struct reclock *a)		/* lock being added */
{
    uint lockid;
    struct resource *rp ;
    rl_info_t rl_entry ;
    int decrflag = 0 ;	/* true if need to --refcount */

    DBPRINTF((DBG_CLMLOCK, "cont_lock_clm: LOCK\n"));
    if (a->lck.state & (LOCK_WAITING | LOCK_GRANTED))
    {
	rp = resptr(a->lck.rh);
    }
    else
    {
	/*
	 * register the name - secondary copy is always
	 * allowed to create the resource
	 */
	rp = makename((int)resp->clm_lockreq.lockinfo.site, 
			resp->clm_lockreq.lockinfo.pid, 
			resp, DLM_RES_CREATE);
    } /* end if */
    if (rp == NULL || rp == (struct resource *)(-1))
    {
	bsdlog(LOG_ERR, MSGSTR(CANT_GET_RES_LOCK, 
			"cont_lock_clm: can't get resource for new "
			"lock rh=0x%08lx"),
			resp->clm_lockreq.lockinfo.rh.rh );
	resp->clm_status = DLM_DENIED_NOLOCKS;
	free_le(DLM_RES_VMS, a);
	return 1 ;
    }

    /*
     * local lock ugliness - if we get an ASYNC response
     * for a local lock resource and it's not from the
     * master that we expect, then toss it 'cause it can
     * really screw up local state.
     */
    if ((rp->rsrc_flags & RSRC_LOCAL) &&
			resp->clm_rt_master != rp->master)
    {
	bsdlog(LOG_ERR, "%s: rejecting local lock response from "
	       "wrong master copy", __FUNCTION__);
	resp->clm_status = DLM_VOID;
	return 1 ;
    }

    /* update RLDB */
    rl_entry.rl_name = rp->name;
    rl_entry.rl_namelen = rp->namelen;
    rl_entry.rl_node = rp->master;
    rl_entry.rl_type = rp->type;
    rl_update(&rl_entry);

    /* save lockid */
    lockid = a->lck.lockid;

    /*
     * if a blocked lock is completing we don't want to
     * overwrite our local copy of lockinfo. We need to
     * reference the state below On the other hand, if this is
     * an initial lock request coming back from the primary, we
     * don't yet have any local info
     */
    if (!(a->lck.state & LOCK_WAITING))
    {
	/* copy entire lockinfo struct */
        resp->clm_lockinfo.request = a->lck.request;
	a->lck = resp->clm_lockreq.lockinfo;
	a->lck.state &= ~LOCK_GRANTED;	/* don't miscopy grant state */
    }
    else
    {
	/* just copy sequence number updated by master */
	a->lck.seqnum = resp->clm_lockreq.lockinfo.seqnum;
    }

    /* restore lockid */
    a->lck.lockid = lockid;
    a->lck.rh = rp->rh;

    if (resp->clm_status == DLM_WORKING)
	blocked_lock(rp, a);
    else if (resp->clm_status == DLM_NOTQUEUED ||
			resp->clm_status == DLM_TIMEOUT) {
	decrflag = 1 ;		/* we're removing a reference */
	free_le(DLM_RES_VMS, a);
    } else if (resp->clm_status == DLM_DEADLOCK) {
	/*
	 * here if master sent us DEADLOCK response.
	 * update local resource copy
	 */
	check_handles(rp, 0) ;
	if (a->lck.state & LOCK_WAITING)
	{
	    /* dont' remove from wait queue if not on it */
	    remove_queue(&rp->lockqs[RLQ_WAIT], a, RL_RQ_OFS);
	    unset_lock_timer(a);
	    decrflag = 1 ;		/* we're removing a reference */
	    free_le(DLM_RES_VMS, a);
	}
    }
    else
    {
	/*
	 * we can get here for two reasons, either a
	 * lock request succeeded, or a blocked lock was
	 * granted
	 */
	if (a->lck.state & LOCK_WAITING)
	{
	    check_handles(rp, 0) ;
	    remove_queue(&rp->lockqs[RLQ_WAIT], a, RL_RQ_OFS);
	}
	/*
	 * assume that we're getting the correct status
	 * back from the primary and pass that back to the
	 * client rather than status from grant_lock
	 */
	grant_lock(rp, a, resp->clm_lockreq.lockinfo.lkvalue);
    }
    /*
     * must be last code referencing resource!!
     */
	if (decrflag)
	    decr_delete_resource(rp, 0) ;

    return(0); /* If here, all succeeded. */
}


/*
 * Name
 *	clm_convert
 * Description
 *	This is the routine that will be called on the master node for the
 *	resource that the convert request is being made against regardless of
 *	whether the master is local or remote to the requesting client.
 *	
 *	Collect data structure associated with lockid
 * 	then call convert_lock_clm to attempt conversion 
 *	Update the master copy of the resource being referenced.
 *
 * Arguments
 *	req - the conversion request
 * 		Input Structure: clm_lockinfo
 * Returns in req->clm_status:
 *	DLM_IVLOCKID	bad lockid
 *	DLM_IVRESHANDLE	bad resptr
 * 		Output Structure: clm_lockinfo
 * Returns:
 *	0 if successful
 *	Site id of directory node if resource mastered elsewhere.
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
clm_convert(clm_msg_t *msg)
{
	clm_trans_t *req = msg->cm_data;
	register struct resource *rp;
	register struct reclock *a;
	struct ast_handle pasth;
	struct ast_handle pcbasth;
	register uint id;

	DBPRINTF((DBG_FUNC, "enter clm_convert:"));
	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_CLM_CONVERT,
		req->clm_pid, req->clm_sequence, req->clm_lockinfo.lockid,
		req->clm_lockinfo.lk_flags);

#ifdef DBPRINT
	/* verify that a BAST was registered */
	if (valid_handle(&req->clm_route.cbasth)) {
		DBPRINTF((DBG_AST, "Registering BAST 0x%x\n",
			req->clm_route.cbasth));
	}
#endif

	/*
	 * if we are processing a request from another node, map lockid 
	 */
	if (req->clm_lockinfo.site != OurSiteID)
		id = mapid(req->clm_lockinfo.lockid, req->clm_lockinfo.site);
	else
		id = req->clm_lockinfo.lockid;

	DBPRINTF((DBG_CLMLOCK, "0x%x convert to %d\n", id,
			req->clm_lockinfo.reqmode));

	/*
	 * Confirm lockid, find reclock, get resource handle, pointer
	 */
	if (id == 0 || ((a = find_reclock(DLM_RES_VMS, id)) == NULL) ||
				!mylock(a, req)) {
		int	site;
		/*
		 * Here if we are acting as a remote master but the master
		 * has moved.  Determine the directory node for the resource
		 * and return that to pti_proc_request() so this request will
		 * be forwarded.
		 */
		site = clm_direct(&req->clm_lockreq.nameinfo.name[0],
				req->clm_lockreq.nameinfo.namelen);
		if (!site) {
			bsdlog(LOG_CRIT, MSGSTR(NULL_RET_CONVERT, 
				"clm_convert:clm_direct() returned NULL"));
			return -1;
		}
		DBPRINTF((DBG_CLMLOCK|DBG_CONFIG,
			"clm_convert: bad id 0x%x -> not master of %s -> %d", 
			req->clm_lockinfo.lockid,
			printbin(&req->clm_lockreq.nameinfo.name[0],
			req->clm_lockreq.nameinfo.namelen), site));

		return site;
	} else if (a->lck.rh.rh == 0 || !(rp = resptr(a->lck.rh))) {
		bsdlog(LOG_CRIT, MSGSTR(NULL_HANDLE_CONVERT, 
			"clm_convert: NULL resource handle for 0x%x"), id);
		req->clm_status = DLM_IVRESHANDLE;	
		abort("clm_convert");
	} else if (rp->master != 0) {
		int	site;
		/*
		 * Here if we are acting as a remote master but the master
		 * has moved.  Determine the directory node for the resource
		 * and return that to pti_proc_request() so this request will
		 * be forwarded.
		 */
		site = clm_direct(&req->clm_lockreq.nameinfo.name[0],
				req->clm_lockreq.nameinfo.namelen);
		if (!site) {
			bsdlog(LOG_CRIT, MSGSTR(NULL_RET_MASTER, 
				"master_unlock_clm:clm_direct() returned NULL"));
			return -1;
		}
		DBPRINTF((DBG_CLMLOCK|DBG_CONFIG,
			"clm_convert: bad id 0x%x -> not master of %s -> %d", 
			req->clm_lockinfo.lockid, printbin(rp->name,
			rp->namelen), site));

		return site;
	} else {
	/*
	 * If the resource is migrating, bounce message to apply back
	 * pressure.  Caller will retry.
	 */
		if (rp->rsrc_flags & (RSRC_FRZEN | RSRC_STUCK | RSRC_UPDATING))
		{
		    msg->cm_port = PORT_BOUNCE ;
		    return TRUE ;
		}

		/*
		 * request is not going to be re-directed.  call cccp_msg_duplicate()
		 * to insure that we have a writeable copy of the message,
		 * reset the req pointer and enqueue the response back to the
		 * secondary.
		 */

		msg = cccp_msg_duplicate( msg );
		req = msg->cm_data;
		req->clm_direction = PTI_RESPONSE;
		msg->cm_node = req->clm_lockinfo.site;
		clm_queue_respmsg(msg);
		clmm_cache(rp, msg->cm_node);

		/*
		 * Resource Statistics:
		 *        cs_requests - lock request was made
		 *        cs_same     - from same node as last request
		 *        lastnode    - update lastnode to be current node
		 *        cs_local    - local lock request
		 *        cs_remote   - remote lock request
		 */
		rp->stats.cs_requests++;
		if (rp->lastnode == req->clm_lockreq.lockinfo.site) {
		        rp->stats.cs_same++;
		}
		rp->lastnode = req->clm_lockreq.lockinfo.site;
		if (req->clm_lockreq.lockinfo.site == OurSiteID) 
		{		    
		        rp->stats.cs_local++;
		} else 
		{		    
		        rp->stats.cs_remote++;
		}
		
		/*
		 * Store master id in routeinfo for cache update.
		 */
		req->clm_rt_master = OurSiteID;

		req->clm_lockinfo.rh = a->lck.rh;

		/* update value block in reclock since it might be
		 * a while before the resources gets updated (if at all)
		 * also copy the timeout value in case it's used
		 */
		(void) memcpy( a->lck.lkvalue, req->clm_lockinfo.lkvalue,
			sizeof(a->lck.lkvalue));
		a->lck.li_to = req->clm_lockreq.lockinfo.li_to;

		/* preserve BAST handle in lock record, new BAST in cbasth */
		req->clm_route.basth = a->route.basth;

		/* clear secondary's inflight flag (if any) */
		req->clm_lockreq.lockinfo.state &= ~INFLIGHT_CONVERTING;

		/* save AST handle, if request returns CVTUNGRANT, restore it */
		pasth = a->route.asth;
		pcbasth = a->route.cbasth;
		a->route = req->clm_route;

		req->clm_status = convert_lock_clm(rp, a, 
			req->clm_lockinfo.reqmode, req->clm_lockinfo.lk_flags,
			req->clm_lockinfo.lkvalue);
		if (req->clm_status == DLM_CVTUNGRANT ||
				req->clm_status == DLM_DENIED) {
			a->route.asth = pasth;
			a->route.cbasth = pcbasth;
		} else if (req->clm_status != DLM_NOTQUEUED) {

			/* try to convert some more blocked requests */
			convert_locks(rp, a);
		}

#ifdef DBPRINT
		if (a->lck.state & LOCK_CONVERTING) {
			DBPRINTF((DBG_CLMLOCK, "TRANSACTION %d BLOCKING\n",
				req->clm_sequence));
		}
#endif

		/* copy the lock sequence number to be returned to secondary */
		req->clm_lockinfo.seqnum = a->lck.seqnum;
	}
		
#ifdef DBPRINT
	clm_dumpres(rp);
#endif
	return 0;
}
	
/*
 * Name
 *	convert_lock_clm
 * Description
 *	Process a lock conversion request
 * Arguments
 *	rp - the associated resource
 *	a - the reclock to be converted
 *	mode - the target mode
 *	flags - flags from the convert request
 *	valblk - value block 
 * Returns
 *	DLM_CVTUNGRANT - lock is on wait queue
 *	DLM_DENIED - lock is already converting
 *	DLM_NOTQUEUED - would block and LKM_NOQUEUE was specified
 *	DLM_WORKING - convert blocked
 *	DLM_NORMAL - granted
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
convert_lock_clm(struct resource *rp, struct reclock *a, int mode, int flags,
		char *valblk)
{
	register dlm_stats_t status;

	if (a->lck.state & LOCK_WAITING) {
		status =  DLM_CVTUNGRANT;	
		goto done;
	}

	/* is this correct? */
	if (a->lck.state & LOCK_CONVERTING) {
		DBPRINTF((DBG_CLMLOCK, "Can't convert converting lock"));
		status = DLM_DENIED;
		goto done;
	}

	a->lck.reqmode = mode;
	a->lck.state &= ~INFLIGHT_CONVERTING;
	
	/*
	 * If the level is lower, then it's a downgrade -> grant convert
	 * or if request mode == grant mode -> grant convert
	 * or if the convert queue is NULL
	 *		and there are no incompatible grants -> grant convert
	 */
	if (lk_level[mode] < lk_level[a->lck.mode] || mode == a->lck.mode ||
			(dlm_list_is_empty( & rp->lockqs[RLQ_CONV] ) &&
			NULL == find_incompatible(rp, mode, a))) {
		/*
		 * Here if the convert request is to be granted.
		 */
	        /* 
		 * Resource Statistics:
		 *        cs_compat    - compatible up convert request
		 *        cs_downgrade - downconvert
		 */
	        if ((lk_level[mode] >= lk_level[a->lck.mode]) &&
						    (mode != a->lck.mode)) {
			rp->stats.cs_compat++; /* up convert */
		} else {
			rp->stats.cs_downgrade++;
		}
	  
		a->lck.lk_flags = flags | (a->lck.lk_flags & LKM_ORPHAN);
		a->lck.lk_flags &= ~LKM_BLOCK;
		status = grant_convert(rp, a, valblk);
	} else {
		/*
		 * Here if the convert request is to be blocked.
		 */
	        /*
		 * Resource Statistics:
		 *        cs_incompat - incompatible convert request
		 */
	         rp->stats.cs_incompat++;

		/* 
		 * if lock can't be successfully converted, return 
		 * error if NOQUEUE is specified 
		 */
		if (flags & LKM_NOQUEUE) {
			DBPRINTF((DBG_CLMLOCK,
			   "convert_lock_clm: rejecting because LKM_NOQUEUE"));
			status = DLM_NOTQUEUED;
			goto done;
		} else {
			status = DLM_WORKING;	
			a->lck.lk_flags = flags | (a->lck.lk_flags &
								LKM_ORPHAN);
			blocked_convert(rp, a);
		}
	}
done:
	return status;
}

int
clm_lock_drop(struct transaction *req, unsigned int flags)
{
    struct reclock	*a;

    DBPRINTF((DBG_CLMLOCK,
	    "clm_lock: Drop req from leftover client 0x%x %s",
	    req, printbin(&req->clm_lockreq.nameinfo.name[0],
	    req->clm_lockreq.nameinfo.namelen)));
    /*
     * If the directory authorized us to create the master copy,
     * then it will think that we own it unless we tell it
     * otherwise.
     */
    if (flags & DLM_RES_CREATE)
    {
	struct resource res;

	res.name = &req->clm_lockreq.nameinfo.name[0];
	res.namelen = req->clm_lockreq.nameinfo.namelen;
	res.type = DLM_RES_VMS;
	res.rsrc_flags = 0;

	update_directory(&res, DIR_DELETE, RESRPRI, NULL );
    }

    /* purge will not clean this lock if it's not on a resource */
    if ((a = find_reclock(DLM_RES_VMS, req->clm_lockreq.lockinfo.lockid))
	    && mylock(a, req))
    {
	DBPRINTF((DBG_CLMLOCK,
		"Freeing local inflighted reclock 0x%x r=%s"
		" f=0x%x",
		req->clm_lockreq.lockinfo.lockid,
		printbin(&req->clm_lockreq.nameinfo.name[0],
		req->clm_lockreq.nameinfo.namelen), flags));
	free_le(DLM_RES_VMS, a);
    }

    /*
     * But WAIT!  There's more!  This request is *probably* on
     * our inflight queue.
     */
    req = match_request(req);

    if (req)
    {
	clm_free_trans(req);
    }
    return 0 ;
}


/*
 * Name
 *	convert_locks
 * Description
 *	For each entry in convert queue, try to grant conversion 
 * 	then try to grant waiting locks
 * Arguments
 *	rp - resource against which conversions should be tested
 *	a1 - reclock that triggered conversions, don't send notifications
 *	  for this lock, let the higher level handle that
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
convert_locks(struct resource *rp, struct reclock *a1)
{
	struct reclock *a;
	struct reclock * next;
	struct reclock *incompat;
	dlm_stats_t status;

	if (!rp || !rp->refcount)
		return;

	/*
	 * The secondary only shadows state returned from the master.
	 * We will therefore not be granting any conversion requests,
	 * we only have state for local locks.  This function is called
	 * from master_unlock_clm which is called on the master and 2ndary
	 */
	if (rp->master != 0)
		return;

	DBPRINTF((DBG_FUNC, "entering convert_locks\n"));
	
    for ( a = (struct reclock *) dlm_list_get_head( & rp->lockqs[RLQ_CONV] )
	  ; a != NULL
	  ; a = next )
    {
	/* save a->next here because grant_convert blows it away */

	next = (struct reclock *) dlm_list_succ( (dlm_listnode_t *) a );

	incompat = find_incompatible(rp, a->lck.reqmode, a);
	if (incompat == NULL) 
	{
	    /* move to grant queue */
	    DBPRINTF((DBG_CLMLOCK,
		      "convert_locks: waking sleeping converter\n"));

	    status = grant_convert(rp, a, a->lck.lkvalue);
	    if (a1 != a)  
	    {
		(void) memcpy( a->lck.lkvalue, rp->value, MAXLOCKVAL );
		notify(a, RESP_AST, status);
	    }

	    /* BASTs are now sent whenever the queues change */
	    send_basts_again(rp, a);
	}
	else
	{
	    /* this queue is FIFO */
	    break;
	}
    }

    if ( ! dlm_list_is_empty( & rp->lockqs[RLQ_CONV] ) 
	 && (a1 != NULL) )
    {
	send_basts_again(rp, a1);
    }

    /* 
     * don't grant a waiting lock while any are waiting to convert
     */
    if ( ! dlm_list_is_empty( & rp->lockqs[RLQ_CONV] ) )
	return;

    for ( a = (struct reclock *) dlm_list_get_head( & rp->lockqs[RLQ_WAIT] )
	  ; a != NULL 
	  ; a = next )
    {
	/* save a->next here because grant_convert blows it away */
	next = (struct reclock *) dlm_list_succ( (dlm_listnode_t *) a );

	incompat = find_incompatible(rp, a->lck.reqmode, a);
	if (incompat == NULL) 
	{
	    /* move to grant queue */	
	    DBPRINTF((DBG_CLMLOCK,
		      "convert_locks: granting sleeping locker\n"));
	    remove_queue(&rp->lockqs[RLQ_WAIT], a, RL_RQ_OFS) ;
	    status = grant_lock(rp, a, a->lck.lkvalue);
	    if (a1 != a)  {
		(void) memcpy( a->lck.lkvalue, rp->value, MAXLOCKVAL );
		notify(a, RESP_AST, status);
	    }

	    /* BASTs are now sent whenever the queues change */
	    send_basts_again(rp, a);
	}
	else
	{
	    /* this queue is FIFO */
	    break;
	}
    }

    /* send reminder BASTs if we are still blocking someone */
    if (a1)
	send_basts_again(rp, a1);

    DBPRINTF((DBG_FUNC, "leaving convert_locks\n"));
}
	
/*
 * Name
 *	blocked_convert
 * Description
 *	Move the named lock (struct reclock) from the GRANT queue to the
 *	CONVERT queue.  This includes changing the state and flags of the
 *	lock, adding it to the deadlock timeout queue. Send blocking ASTs
 *	to anyone who cares.
 * Arguments
 *	rp - the associated resource
 *	a - the blocked lock
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
blocked_convert(struct resource *rp, struct reclock *a)
{
	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_BLOCKED_CONVERT,
		a->lck.pid, a->lck.lockid, a->lck.remlockid, a->lck.reqmode);

	DBPRINTF((DBG_CLMLOCK,
		"blocked_convert: move from GRANT to CONVERT queue\n"));

	check_handles(rp, 0) ;

	/*
	 * Make sure that this lock is not already on the convert queue.
	 * Radical migrations can cause resource state to be updated before
	 * response transactions from old requests have to chance to get
	 * back here.
	 */
	if (~a->lck.state & LOCK_CONVERTING) {

		/* move from lock to convert queue */
		remove_queue(&rp->lockqs[RLQ_GRANT], a, RL_RQ_OFS) ;
		a->lck.state &= ~LOCK_GRANTED;
		a->lck.state |= LOCK_CONVERTING;
		/*
		 * we need to set this so that the transaction
		 * will be recognized as a conversion
		 * in cont_remote_lock_clm on the secondary
		 */
		a->lck.lk_flags |= LKM_CONVERT;
		lck_set_seq(rp, a);
		add_queue(&rp->lockqs[RLQ_CONV], a, RL_RQ_OFS) ;
	}

	check_handles(rp, 0) ;
	add_timeout_reclock(a);
	/* check if we should set a timer */
	if ((a->lck.lk_flags & LKM_TIMEOUT) && (a->lck.site == OurSiteID) &&
			(a->lck.li_to > 0)) {
		set_lock_timer(a);
	}
	if (rp->master == 0) {
		send_basts(rp, a);
	}
}

/*
 * Name
 *	blocked_lock
 * Description
 *	Add the specified reclock to the wait queue and set the LOCK_WAITING 
 *	attribute. Assumes that the reclock is not currently on any queue. 
 *	Send blocking ASTs to anyone who cares.
 * Arguments
 *	rp -the associated resource
 *	a - the blocked lock
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void 
blocked_lock(struct resource *rp, struct reclock *a)
{
	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_BLOCKED_LOCK,
		a->lck.pid, a->lck.lockid, a->lck.remlockid, a->lck.reqmode);

	DBPRINTF((DBG_CLMLOCK, "blocked_lock: adding lock to WAIT queue\n"));

	/* don't add to wait queue if we're already on it */
	if (~a->lck.state & LOCK_WAITING) {
		a->lck.state |= LOCK_WAITING;
		lck_set_seq(rp, a);
		add_queue(&rp->lockqs[RLQ_WAIT], a, RL_RQ_OFS) ;
	}

	check_handles(rp, 0) ;
	add_timeout_reclock(a);
	/* check if we should set a timer */
	if ((a->lck.lk_flags & LKM_TIMEOUT) && (a->lck.site == OurSiteID) &&
			(a->lck.li_to > 0)) {
		set_lock_timer(a);
	}
	if (rp->master == 0) {
		send_basts(rp, a);
	}
}

/*
 * Name
 *	grant_lock
 * Description
 *	Grant the specified lock 
 * 	Add reclock to grant queue, update grantmode of reclock.
 * 	Process lock value block if appropriate. Assumes that the reclock is
 *	 not currently on any queue.
 * Arguments
 *	rp - the associated resource
 *	a - the reclock to be converted
 *	valblk - value block 
 * Returns
 *	DLM_VALNOTVALID if valueblock() returns such
 *	DLM_NORMAL otherwise
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
grant_lock(struct resource *rp, struct reclock *a, char *valblk)
{
	dlm_stats_t status = DLM_NORMAL;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_GRANT_LOCK,
		a->lck.pid, a->lck.lockid, a->lck.remlockid, a->lck.reqmode);

	DBPRINTF((DBG_CLMLOCK, "grant_lock: adding lock to GRANT queue\n"));

	a->lck.state &= ~LOCK_WAITING;
	a->lck.mode = a->lck.reqmode;
	lck_set_seq(rp, a);
	if (~a->lck.state & LOCK_GRANTED) {
		add_queue(&rp->lockqs[RLQ_GRANT], a, RL_RQ_OFS) ;
		/* copy resource handle to reclock structure */
		a->lck.rh = rp->rh;
		a->lck.state &= ~LOCK_WAITING;
		a->lck.state |= LOCK_GRANTED;
	}
	status = valueblock(rp, LKM_NLMODE, LKM_NLMODE, valblk,
				a->lck.lk_flags);

	/* stop timer if there is one */
	if ((a->lck.lk_flags & LKM_TIMEOUT) && (a->lck.site == OurSiteID) &&
				(a->lck.li_to > 0)) {
		unset_lock_timer(a);
		a->lck.lk_flags &= ~LKM_TIMEOUT;
		a->lck.li_to = 0;
	}

	/*  If it is on the deadlock timeout queue, take it off.  */
	delete_timeout_reclock(a);

	if (a->lck.lk_flags & LKM_VALBLK) {
		return status;
	} else {
		return DLM_NORMAL;
	}
}

/*
 * Name
 *	grant_convert
 * Description
 *	grant the specified convert
 * 	update grant mode
 * 	process value block as appropriate
 *	If reclock was previosly blocked, remove from convert queue
 *	and add to grant queue
 *	
 * Arguments
 *	rp - the associated resource
 *	a - the reclock to be converted
 *	valblk - value block 
 * Returns
 *	DLM_VALNOTVALID if valueblock() returns such
 *	DLM_NORMAL otherwise
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
grant_convert(struct resource *rp, struct reclock *a, char *valblk)
{
	dlm_stats_t status = DLM_NORMAL;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_GRANT_CONVERT,
		a->lck.pid, a->lck.lockid, a->lck.remlockid, a->lck.reqmode);

	DBPRINTF((DBG_CLMLOCK,
		 "grant_convert: converting lockid 0x%x from %d to %d\n", 
		 a->lck.lockid, a->lck.mode, a->lck.reqmode));


	status = valueblock(rp, a->lck.mode, a->lck.reqmode, valblk,
			a->lck.lk_flags);

	/* stop timer if there is one */
	if ((a->lck.lk_flags & LKM_TIMEOUT) && (a->lck.site == OurSiteID) &&
				(a->lck.li_to > 0)) {
		unset_lock_timer(a);
		a->lck.lk_flags &= ~LKM_TIMEOUT;
		a->lck.li_to = 0;
	}

	a->lck.mode = a->lck.reqmode;
	a->route.basth = a->route.cbasth;

	/*
	 * don't add to grant queue if it is already on that queue
	 * this is true for an in-place conversion
	 */
	if (a->lck.state & LOCK_CONVERTING) {
		remove_queue(&rp->lockqs[RLQ_CONV], a, RL_RQ_OFS) ;
		lck_set_seq(rp, a);
		a->lck.state &= ~LOCK_CONVERTING;
		add_queue(&rp->lockqs[RLQ_GRANT], a, RL_RQ_OFS) ;
		a->lck.state |= LOCK_GRANTED;
	}

	/* clear BAST bit to allow */
	a->lck.state &= ~LOCK_BASTED;

	/*  If it's on the deadlock timeout queue, take it off.  */
	delete_timeout_reclock(a);

	if (a->lck.lk_flags & (LKM_VALBLK | LKM_INVVALBLK)) {
		return status;
	} else {
		return DLM_NORMAL;
	}
}

/*
 * Name
 *	find_incompatible
 * Description
 * 	Given a resource and a requested mode, walk the GRANT and CONVERT
 *	queues and return the first incompatible lock found. If recp is 
 * 	non-NULL ignore this lock if encountered in a queue. 
 * Arguments
 *	rp - the resource to check
 *	req_mode - the requested mode against which to test compatibility
 *	recp - the reclock of the requester
 *	
 * Returns
 *	pointer to first incompatible lock found
 *	NULL if none found, *always returns NULL if req_mdoe == NLMODE*
 * Global Data Affected
 *	NONE
 */
struct reclock *
find_incompatible(struct resource *rp, int req_mode, struct reclock *recp)
{
    struct reclock * a;
    struct reclock * incompatible = NULL;
    unsigned         i;


    DBPRINTF((DBG_FUNC, "enter find_incompatible\n"));

    /* check grant queue */
    for (i = RLQ_GRANT ; i < RLQ_WAIT ; i++)
    {
	for ( EACH_DLM_LISTNODE( & rp->lockqs[i], a ) )
	{
	    if ( (a != recp) && ! clmcompatible( req_mode, a->lck.mode ) )
	    {
		incompatible = a;
		goto out ;
	    }
	}
    }

out:

    DBPRINTF((DBG_FUNC, "leave find_incompatible(0x%x)\n", incompatible));
    return incompatible;
}


/*
 * Name
 *	send_basts
 * Description
 * 	For the given resource, send a blocking AST to each lock holder
 *	who has requested a blocking AST and is incompatible with the specified
 *	mode.
 *	- Send a BAST to any lock holders that are blocking me -
 * Arguments
 *	rp - the resource
 *	ba - the lock that is causing the request to be sent
 * Returns
 *	zero (was The number of ASTs posted)
 * Global Data Affected
 *	NONE
 */
int
send_basts(struct resource *rp, struct reclock *ba)
{
    struct reclock * a;
    int              i;

    DBPRINTF((DBG_FUNC, "enter send_basts\n"));

	/* check convert THEN grant queue ???? */
    for (i = RLQ_CONV ; i >= RLQ_GRANT ; i--)
    {
	for ( EACH_DLM_LISTNODE( & rp->lockqs[i], a ) )
	{
	    if ( (ba != a) && ! clmcompatible(ba->lck.reqmode, a->lck.mode) )
	    {
		send_bast(a, ba->lck.reqmode);
	    }
	}
    }
    return 0 ;
}


/*
 * Name
 *	send_basts_again
 * Description
 *	This function is the same a the one above, except that it is called
 *	with the reclock of the lock that was just granted.
 * 	For the given resource, send a blocking AST to myself for each lock
 *	request that I'm blocking.
 *	- Send myself a BAST for any locks that I'm blocking -
 * Arguments
 *	rp - the resource
 *	ba - The lock that might be blocking someone else.
 * Returns
 *	zero (comment was The number of ASTs posted)
 * Global Data Affected
 *	NONE
 */
int
send_basts_again(struct resource *rp, struct reclock *ba)
{
    struct reclock * a;
    unsigned         i;


    DBPRINTF((DBG_FUNC, "enter send_basts_again\n"));

    /* check convert then wait queue */
    for (i = RLQ_CONV ; i < RLQ_MAX ; i++)
    {
	for ( EACH_DLM_LISTNODE( & rp->lockqs[i], a ) )
	{
	    if ( (ba != a) && ! clmcompatible(a->lck.reqmode, ba->lck.mode) )
	    {
		send_bast(ba, a->lck.reqmode);
	    }
	}
    }
    return 0 ;
}

/*
 * Name
 *	send_bast
 * Description
 * 	send a blocking AST to the owner of the specified reclock
 * Arguments
 *	a - the reclock to receive the BAST
 *	mode - the mode requested
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
send_bast(struct reclock *a, int mode)
{
	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_SEND_BAST,
		a->lck.pid, a->lck.lockid, a->lck.remlockid, mode);

	/*
	 * we now only send one BAST to a lock, so check to see if one has
	 * already been sent.  We also can't send one if there is no BAST
	 * function
	 */
	if ((a->lck.state & LOCK_BASTED) || a->route.basth.asth_funcp == NULL)
		return;

	DBPRINTF((DBG_AST, 
	  "Sending blocking AST to %d mode %s blocked request %s\n",
	  a->lck.pid, clm_printmode(a->lck.mode), clm_printmode(mode)));
	a->lck.bastmode = mode;

	/* set BAST bit to prevent another BAST */
	a->lck.state |= LOCK_BASTED;

	notify(a, RESP_BAST, DLM_NORMAL);
}

 

#define NLMASK		(1<<LKM_NLMODE)
#define CRMASK		(1<<LKM_CRMODE)
#define CWMASK		(1<<LKM_CWMODE)
#define PRMASK		(1<<LKM_PRMODE)
#define PWMASK		(1<<LKM_PWMODE)
#define EXMASK		(1<<LKM_EXMODE)

/*
 * The lock compatibility table
 * To use it, index in by the current mode
 * And check if the bit position set by the desired mode
 * is set, if so, the locks are compatible
 */
static const int clmlockcompat[] = {
	NLMASK|CRMASK|CWMASK|PRMASK|PWMASK|EXMASK,
	NLMASK|CRMASK|CWMASK|PRMASK|PWMASK,
	NLMASK|CRMASK|CWMASK,
	NLMASK|CRMASK|PRMASK,
	NLMASK|CRMASK,
	NLMASK
};


/*
 * Name
 *	clmcompatible
 * Description
 *	determine if two lock modes are compatible
 * Arguments
 *	m1, m2 - the two modes to check for compatibility
 * Returns
 *	1 if compatible
 * 	0 if not compatible
 * Global Data Affected
 *	NONE
 */
int
clmcompatible(int m1, int m2)
{
	return ((clmlockcompat[m1] & (1<<m2)) != 0);
}


/*
 * Name
 *	wake_all_locks
 * Description
 *	This routine is run on each node after a reconfiguration.
 * 	It walks through the whole resource database and attempts to 
 *	complete any stalled transactions for resources that are
 *	mastered locally
 * Arguments
 *	NONE
 * Returns
 *	NOTHING
 */
void
wake_all_locks(void)
{
    iterator_t * iter;
    struct resource * res;


    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_WAKE_ALL_LOCKS,
	    0x00, 0x00, 0x00, 0x00);

    DBPRINTF((DBG_CLMLOCK, "wake_all_locks: attempting to wake sleepers\n"));

    iter = resource_table_iterator_create();
    CLM_ASSERT( iter != NULL, "Unable to create resource iterator" );

    while ( ! iterator_next( iter, & res ) )
    {
	if ( (res->master == 0) && (res->refcount > 0) )
	{
	    convert_locks( res, NULL );
	}
    }

    iterator_destroy( iter );
}
	

/*
 * Name
 *	wakeup_clm
 * Description
 *	A lock has been purged, wake any sleepers.
 * Arguments
 *	rp - the associated resource
 *	a - the lock that was purged
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void 
wakeup_clm(struct resource *rp, struct reclock *a)
{
	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_WAKEUP_CLM,
		0x00, 0x00, 0x00, 0x00);

	convert_locks(rp, a);
}

/*
 * Name
 *	clm_printmode
 * Description
 *	Return a character string corresponding to the specified
 *	lock mode, returns "JUNK" if mode out of range
 * Arguments
 *	m - the mode
 * Returns
 *	a character string
 * Global Data Affected
 *	NONE
 */
char *
clm_printmode(int m)
{
	static char const *modes[] = {
	"NL", "CR", "CW", "PR", "PW", "EX", "JUNK" };
	
	if (m < 0 || m > 5)
		m = 6;
	return (char *) modes[m];
}
/*
 * Name
 * 	clm_printlock
 * Description
 *	Display lockid, pid, grant mode, request mode, flags, and state
 *	on the debugging stream for the specified lock
 * Arguments
 *	a - the lock to print
 * Returns
 *	NOTHING
 */
void
clm_printlock(struct reclock *a)
{
	DBPRINTF((DBG_CLMLOCK, 
		" id=0x%x rmid=0x%x pid=%d gt=%s rq=%s f=%d st=%d"
		" nd=%d xid=0x%x%08x\n",
		a->lck.lockid, a->lck.remlockid, a->lck.pid,
		clm_printmode(a->lck.mode), clm_printmode(a->lck.reqmode),
		a->lck.lk_flags, a->lck.state, a->lck.site,
		(uint)a->lck.xid.xid_i[0], (uint)a->lck.xid.xid_i[1]));
}

/*
 * Name
 *	clm_printlocks
 * Description
 *	Print all locks in a lock queue on the debugging stream 
 *	after first printing the supplied string
 * Arguments
 *	s - the string
 *	a - the first lock in the queue
 * Returns
 *	the number of locks printed
 * Global Data Affected
 *	NONE
 */
static int
clm_printlocks( const char *s, dlm_list_t * list )
{
    struct reclock * a;
    int n = 0;
    DBPRINTF((DBG_CLMLOCK, "%s\n", s));

    for ( EACH_DLM_LISTNODE( list, a ) )
    {
	clm_printlock(a);
	n++;
    }
    return n;
}

/*
 * Name
 *	clm_dumpres
 * Description
 *	Print information about a resource on the debugging stream
 *	and then dump all lock queues associated with the resource
 * Arguments
 *	rp - the resource
 * Returns
 *	the number of locks printed
 * Global Data Affected
 *	NONE
 */
int
clm_dumpres(struct resource *rp)
{
	int n = 0;
	if (rp && rp->rsrc_flags & RSRC_ACTIVE) {
		DBPRINTF((DBG_CLMLOCK, "\n*******\nRESOURCE (%s) (0x%x) \n", 
			printbin(rp->name, rp->namelen), rp->rh.rh));
		DBPRINTF((DBG_CLMLOCK, "VALUE BLOCK %s\n", 
			printbin(rp->value, 
			sizeof(rp->value))));
		DBPRINTF((DBG_CLMLOCK, "refcount=%d flags=0x%x master=%d\n",
			rp->refcount, rp->rsrc_flags, rp->master));
		n += clm_printlocks("WAIT", & rp->lockqs[RLQ_WAIT] );
		n += clm_printlocks("CONV", & rp->lockqs[RLQ_CONV] );
		n += clm_printlocks("GRANT", & rp->lockqs[RLQ_GRANT] );
		DBPRINTF((DBG_CLMLOCK, "\n"));
	}
	return n;
}

/*
 * Name
 *	makename
 * Description
 *	Register a resource name, or establish another reference to
 *	an existing resource and return a pointer to the resource
 * Arguments
 *	site - the site of the client
 *	pid - the process id of the client
 *	req - the lock request being processed
 * Returns
 *	the resource or NULL if there's an error
 * Side Effect
 *	(sometimes?) sets master of resource to originator of req
 *	Increments reference count on resource
 * Global Data Affected
 *	NONE
 */
static struct resource *
makename(int site, int pid, clm_trans_t *req, int create_flag)
{
	struct resource *rp;
	struct clm_regreq regreq;

	regreq.name = &req->clm_lockreq.nameinfo.name[0];
	regreq.namelen = req->clm_lockreq.nameinfo.namelen;

	/*
	 * This master value will only have a non-zero value when it is called
	 * from the continue routine on a secondary node.  Therefore, the
	 * master field in a resource structure will have a value of zero
	 * on the master node, and the master's site id on secondary nodes.
	 */
	CLM_ASSERT((req->clm_rt_master || req->clm_rt_master != OurSiteID),
			"makename: Set master local");
	regreq.master = req->clm_rt_master;
	
	regreq.type = DLM_RES_VMS;
	regreq.pid = pid;
	regreq.site = site;
	regreq.count = 1;
	
	/*
	 * create resource if necessary
	 */
	if (NULL != (rp = lookup_resource(&regreq))) {
		check_handles(rp, 0) ;
		DBPRINTF((DBG_CLMLOCK, "makename: found rh 0x%x refcount %d\n",
			rp->rh.rh, rp->refcount));

		/*
		 * if site != OurSiteID then this is a remote request
		 * that thinks it's on the master.  If the lookup returns a
		 * resource pointer that is mastered elsewhere, then this
		 * request has been sent to the wrong master.  Return error.
		 * NOTE: There is a *remote* possibility that there is a
		 * leftover secondary copy of a resource that is about to be
		 * purged when we get a new create request from a directory
		 * server.  If the create bit is set, then do not bounce the
		 * request.
		 */
		if (regreq.master == 0 && (~create_flag & DLM_RES_CREATE) &&
						(rp->master != 0 ||
						rp->rsrc_flags & RSRC_FRZEN))
			return (struct resource *)(-1);

		/*
		 * Here if lookup was successful.  Update references.
		 */
		if (rp->master) 
		{
		    if (regreq.master == 0) 
		    {
			struct reclock	*a;


			/* Here if we're creating a master copy of a resource
			 * on top of an old secondary copy.  If there are
			 * leftover locks, then they shouldn't migrate.  Set
			 * the flag to prevent this from happening.
			 */

			for ( EACH_DLM_LISTNODE( & rp->lockqs[ RLQ_GRANT ], a ) )
			{
			    if (a->lck.state & LOCK_GONE) 
			    {
				a->lck.state |= LOCK_LOCAL_ONLY;
			    }
			}

			/* prevent master from having FRZEN bit set */
			rp->rsrc_flags &= ~RSRC_FRZEN;
		    }

		    CLM_ASSERT((regreq.master != OurSiteID),
			       "makename2: Set master local");
		    rp->master = regreq.master;
		}
		rp->refcount++;
	} else if (req->clm_lockinfo.lk_flags & LKM_FINDLOCAL &&
				!(create_flag & DLM_RES_CREATE)) {
	
	        return (NULL);
       	
	} else if (site != OurSiteID || !(create_flag & DLM_RES_CREATE)) {
		/*
		 * Here if we couldn't find the resource and the request is
		 * from a different node.  A master copy of a resource should
		 * not be *created* on a node other than the one that made the
		 * request to cause the creation.  So if we're here, then
		 * either the mastery of the resource moved since the
		 * directory told us where it was, or the master copy of the
		 * resource went away since we got our directory information.
		 * So we need to check with the directory again.  Also,
		 * check the creation bit in the flags.  If we are not
		 * authorized to create the master copy of the resource, then
		 * the directory information that we got was out of date and
		 * we need to check with the directory again.  Return error.
		 */
		return (struct resource *)(-1);

	} else if (NULL != (rp = allocate_resource(&regreq))) {
		DBPRINTF((DBG_CLMLOCK, "makename: allocated rh 0x%x\n", rp->rh));
		if (req->clm_lockinfo.lk_flags & (LKM_LOCAL | LKM_FINDLOCAL)) {
			rp->rsrc_flags |= RSRC_LOCAL;
		}
	}
	
	return rp;
}


/*
 * Name
 *	valueblock
 * Description
 *	Handle lock value block setting or retrieving based on modes.
 * 	If moving from >= LKM_PWMODE to lower, update resource value block
 * 	else update valblk from resource.
 * 	If this node is a secondary, update resource value with value 
 *	coming back from primary
 * Arguments
 *	rp -the associated resource
 *	oldmode - the currently held mode (NLMODE if new lock)
 *	newmode - the desired mode
 *	valblk - the valueblock from the client
 *	flags - option flags, currently only LKM_INVVALBLK
 * Returns
 *	DLM_VALNOTVALID if returning invalid value block
 *	DLM_NORMAL otherwise
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
valueblock(struct resource *rp, int oldmode, int newmode, char *valblk,
		int flags)
{
    dlm_stats_t status = DLM_NORMAL;

    if (valblk == NULL)
	return DLM_VALNOTVALID;

    /* setting lock value block */	

    /*
     * NOTE: LVB set behavior is affected by LKM_OLD_LVB flag added for
     * CR960021.  If flag is set, setting LVB requires downconvert.
     * If flag is clear, a conversion to the same mode can set LVB.
     *
     * The user side will now never send in LKM_OLD_LVB, so remove
     * the check for it, this should give us the normal behavior.
     */

    if (oldmode >= LKM_PWMODE 
	 && ((oldmode > newmode) 
	     || (oldmode == newmode /* && (~flags & LKM_OLD_LVB) */))) 
    {
	if (flags & LKM_INVVALBLK) 
	{
	    DBPRINTF((DBG_CLMLOCK, "Invalidating resource value block\n"));
	    clear_value_block(rp);	
	}
	else if (flags & LKM_VALBLK) 
	{
	    DBPRINTF((DBG_CLMLOCK, "Updating resource value block\n"));
#ifdef DBPRINT
	    dump_value_block (valblk, rp);
#endif
	    (void) memcpy( rp->value, valblk, sizeof(rp->value) );
	    rp->rsrc_flags |= RSRC_VALID;
	}
    }
    else 
    {
	/* if we are on a secondary, assume that value 
	 * block has already been passed back from resource 
	 * on the master, so update our local copy and
	 * leave value in valblk intact
	 * -- this leftover code is questionable -bg- XXX
	 */
	if (0 != rp->master) 
	{
	    (void) memcpy( rp->value, valblk, sizeof(rp->value) );
	    DBPRINTF((DBG_CLMLOCK, 
		      "Updating resource value block from primary\n"));
#ifdef DBPRINT
	    dump_value_block (rp->value, rp);
#endif
	    rp->rsrc_flags |= RSRC_VALID;
	}
	else
	{
	    /* 
	     * we are the primary, so hand back 
	     * current resource value block 
	     */
	    if (rp->rsrc_flags & RSRC_VALID) {
		(void) memcpy( valblk, rp->value, sizeof(rp->value) );
		DBPRINTF((DBG_CLMLOCK, "Returning value block to user\n"));
#ifdef DBPRINT
		dump_value_block (rp->value, rp);
#endif
	    }
	    else 
	    {
		DBPRINTF((DBG_CLMLOCK, 
			  "Returning DLM_VALNOTVALID to user\n"));
		status = DLM_VALNOTVALID;
	    }
	}
    }

    return status;
}

/*
 * Name
 *	dump_value_block
 * Description
 *	Display lock value block for debugging purposes
 * Argument
 *	value - the value block
 *	rp - associated resource
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
dump_value_block( const char * value, const struct resource * rp )
{
#define DVB_CHARS    16U
#define DVB_BUF_SIZE ((3 * DVB_CHARS) + 1)
    unsigned   i;
    unsigned   j;
    unsigned   bytes;
    char       buf[ DVB_BUF_SIZE ];

	
    DBPRINTF((DBG_CLMLOCK, "Value block for resource handle 0x%x: ",
	      rp->rh.rh));

    if (dlm_debug & DBG_CLMLOCK)
    {
	for ( i = sizeof( rp->value ) ; i != 0 ; /* empty */ )
	{
	    bytes = 0;


	    /* On each line, format at most DVB_CHARS.  The value of 'i'
	     * must be updated here because the loop destroys the value
	     * calculated for j.
	     */

	    j = min( i, DVB_CHARS );
	    i -= j;

	    for ( /* empty */ ; j != 0 ; j-- )
	    {
		bytes += snprintf( buf + bytes, DVB_BUF_SIZE - bytes, " %02X",
				   (unsigned char) *value );
		value++;
	    }

	    DBPRINTF((DBG_CLMLOCK, "%s", buf));
	}
    }

#undef DVB_BUF_SIZE
#undef DVB_CHARS
}

/*
 * Name
 *	clm_reset_valueblocks
 * Description
 *	For each resource in the resource table clear its lock value block
 * 	If no local clients hold the resource at EX or PW
 * Arguments
 *	NONE
 * Returns
 *	NOTHING
 * Global Data Affected
 *	restab
 */
void
clm_reset_valueblocks(void)
{
    iterator_t * iter;
    struct resource * res;
    unsigned   k;
    struct reclock * t;


    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_CLM_RESET_VALUEBLOCKS,
	    0x00, 0x00, 0x00, 0x00);

    DBPRINTF((DBG_CLMLOCK, 
	    "clm_reset_valueblocks: resetting lock value blocks\n"));

    iter = resource_table_iterator_create();
    while ( ! iterator_next( iter, & res ) )
    {
	if ( res->refcount > 0 )
	{
	    /* scan grant & convert queues */
	    for (k = RLQ_GRANT ; k < RLQ_WAIT ; k++)
	    {
		for ( EACH_DLM_LISTNODE( & res->lockqs[k], t ) )
		{
		    if (t->lck.mode >= LKM_CWMODE) 
		    {
			goto out1 ;
		    }
		}
	    }

	    clear_value_block( res );
	    
out1: ;	/* found a mode >= CW mode */

	}
    }

    iterator_destroy( iter );
}


/*
 * Name
 *	clear_value_block
 * Description
 *	Reset the lock value information associated with the specified
 *	resource
 * Arguments
 *	rp - the resource whose lock value information should be reset
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
clear_value_block(struct resource *rp)
{
    (void) memset( rp->value, 0, sizeof( rp->value ) );
    rp->rsrc_flags &= ~(RSRC_VALID);
}

/*
 * Name
 *	purge_clm
 * Description
 *	Purge the specified lock. Invalidate the lock value block
 *	if appropriate.
 * Arguments
 *	rp - resource on which lock resides
 * 	rlqp - pointer to current queue for lock
 * 	a - the lock itself
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
purge_clm(struct resource *rp, dlm_list_t * queue, struct reclock *a)
{
	check_handles(rp, CH_NOREF) ;
	remove_queue( queue, a, RL_RQ_OFS );
	if ((a->lck.lk_flags & LKM_TIMEOUT) && (a->lck.site == OurSiteID) &&
				(a->lck.li_to > 0)) {
		unset_lock_timer(a);
		a->lck.lk_flags &= ~LKM_TIMEOUT;
		a->lck.li_to = 0;
	}
	if (a->lck.mode >= LKM_PWMODE) {
		clear_value_block(rp);
	}
	(lockswitch[rp->type].wakeup)(rp, a);
	free_le(rp->type, a);
}

/*
 * xid_add_client_reclock
 *
 * wrapper around add_client_reclock
 * adds a reclock onto a possibly new client structure
 * sets id, site & type consistent with the reclock as used in
 * clm_clmlock()
 *
 * used to maintain state info for purge & migration
 *
 * arg: struct reclock *
 * ret: return from add_reclock
 */
int
xid_add_client_reclock(struct reclock *a)
{
    struct clm_client_id clid ;
    int rc ;

    /*
     * record client info on master for purge & migration
     */
    if (a->lck.lk_flags & LKM_XID)
    {
	/* Use XID for client id */
	rc = add_client_reclock(a->lck.site,
		    ((clm_client_id_t *)&(a->lck.xid)), 
		    CLM_CLI_TRANSACTION, a) ;
    }
    else
    {
	/* Set up client id for process */
	set_client_id(clid, a->lck.pid) ;
	rc = add_client_reclock(a->lck.site, &clid, CLM_CLI_PROCESS, a) ;
    }
    return rc ;
}
