/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_clmunlock.c,v 1.7 2001/10/04 20:29:55 idr Exp $"

/*
 *   COMPONENT_NAME: CLLOCKD
 *
 *   FUNCTIONS: MSGSTR
 *		begin_unlock_clm
 *		cont_unlock_clm
 *		local_unlock_clm
 *		master_unlock_clm
 *		pre_proc_unlock_clm
 *
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/stddef.h>
#include <linux/signal.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/smp.h>
#include <linux/sysrq.h>
#include <linux/interrupt.h>
#include <asm/semaphore.h>
#include <asm/softirq.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "clm_data.h"			/* external data definitions */
#include "dlm_kernel.h"
#include "clm_locksw.h"
#include "dlm_version.h"
#include "cltrc.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/*
 * Name
 *      hold_cancel
 * Description
 *      The function is called when a cancel request is received for a lock
 *      that does yet have a local resource because the master response
 *      for the initial lock request hasn't been received yet.  This
 *      function queues the cancel request on the reclock structure to
 *      be sent when there is a local copy of the resource.
 * Arguments
 *      a - the reclock that is being canceled
 *      req - the cancel request
 * Returns
 *      nothing
 * Global Data Affected
 *      NONE
 */
static void
hold_cancel(struct reclock *a, struct transaction *req)
{
    struct transaction  **preq = &a->lck.request;

    /* this req will be on the tail */
    req->clm_next = NULL;

    CLM_ASSERT((req->clm_type == DLM_UNLOCK), "Bad type in hold_cancel");

    /* in the unlikely event that there's a list of cancels, walk to the end */
    while (*preq)
        preq = (struct transaction **)&((*preq)->clm_next);

    /* point the tail pointer to this cancel request */
    *preq = req;
}

/*
 * Name
 *      trash_cancels
 * Description
 *      This routine is called when the master response message has been
 *      handled for a new lock.  It checks the cancel list on the
 *      lock for held cancel requests and resends them.
 * Arguments
 *      a - the reclock to check for cancels
 * Returns
 *      nothing
 * Global Data Affected
 *      NONE
 */
void
trash_cancels(struct reclock *a)
{
    struct transaction  *next, *req;

    /* do we need to be here? */
    if (!a || NULL == a->lck.request)
        return;

    /* stash the whole queue, ensure cancel list is NULL */
    req = a->lck.request;
    a->lck.request = NULL;

    /* walk list and send each cancel */
    while (req)
    {
        next = req->clm_next;
        req->clm_next = NULL;

        req->clm_status = DLM_IVLOCKID;
        clm_reply(req, NULL);

        req = next;
    }
}

/*
 * Name
 *      submit_cancels
 * Description
 *      This routine is called when the master response message has been
 *      handled for a new lock.  It checks the cancel list on the
 *      lock for held cancel requests and resends them.
 * Arguments
 *      a - the reclock to check for cancels
 * Returns
 *      nothing
 * Global Data Affected
 *      NONE
 */
void
submit_cancels(struct reclock *a)
{
    struct transaction  *next, *req;
    struct resource     *rp;

    /* do we need to be here? */
    if (!a || NULL == a->lck.request)
        return;

    /* stash the whole queue, ensure cancel list is NULL */
    req = a->lck.request;
    a->lck.request = NULL;

    /* walk list and send each cancel */
    while (req)
    {
        next = req->clm_next;
        req->clm_next = NULL;

        if (a->lck.rh.rh && (rp = resptr(a->lck.rh)))
        {
            /*
             * local resource exists, use its info, but copy the
             * name and length into the nameinfo of the request
             * in case the resource has moved from the master.
             */
	    (void) memcpy( &req->clm_lockreq.nameinfo.name[0],
			   rp->name, rp->namelen);
            req->clm_lockreq.nameinfo.namelen = rp->namelen;

            CLM_ASSERT((req->clm_type == DLM_UNLOCK), "Bad type in submit");

            pti_call(req, rp->master ? rp->master : OurSiteID);
        }
        else
            bsdlog(LOG_DEBUG, "Dropping cancel of 0x%x", a->lck.lockid);

        req = next;
    }
}

/*
 * Name
 *	begin_unlock_clm
 * Description
 * 	This routine handles the initial processing of an unlock request before
 *	it is forwarded to the distributed layer.  Well, at least that's
 *	what you'd expect it do.  What it REALLY does is this:
 *
 *	In the interest of optimization, we call the pre-process unlock function
 *	first with the intent of sending a successful status back to the
 *	client before we forward the request to the master.  This call can
 *	have four different outcomes:
 *
 *	1.  Error detected locally.  Unlock aborted.  Notify client.
 *	    No interaction with dist layer needed.
 *	2.  The resource is mastered locally, processing is complete.  Send
 *	    client notification.  No interaction with dist layer needed.
 *	3.  The resource is mastered remotely.  Local unlock complete.
 *	    Forward request to distributed layer.  Notify client.
 *	4.  The resource is mastered remotely.  Local unlock not possible
 *	    because of local lock state.  Forward request.  Notify client
 *	    when response comes back.
 *
 *	For unlocks on a remotely mastered resource that are successful, this
 *	optimization yields a significant performance improvement.  Without
 *	it, a request has to make a network roundtrip to the master node
 *	before the clmunlock function can return.
 *
 * Arguments
 *	req - the unlock request
 * Returns
 *	return status from pti_call 
 * Global Data Affected
 * 	NONE
 */
int
begin_unlock_clm(clm_trans_t *req)
{
	int	stat = 0;
	int	site = 0;

	DBPRINTF((DBG_FUNC, "enter begin_unlock_clm\n"));
	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_BEGIN_UNLOCK_CLM,
		req->clm_pid, req->clm_sequence, req->clm_lockinfo.lockid,
		req->clm_lockinfo.lk_flags);

	/* Add this to remove need for clinfo in API */
	req->clm_lockinfo.site = OurSiteID;

	/*
	 * Call the pre_proc_unlock_clm function to see if the unlock can
	 * be processed completed on this node or at least before it gets
	 * passed to the master.
	 */
	site = pre_proc_unlock_clm(req);

	if (req->clm_status == DLM_FORWARD) {
		struct reclock	*a;
		struct resource	*rp;

		/*
		 * Here if the request has to be sent to the master before
		 * it can be processed locally.  Local resource and master
		 * information may or may not exist.  If local resource
		 * exists, send to master.  If it doesn't exist, then the
		 * initial lock request that we are trying to cancel hasn't
		 * come back from the master yet, so the name information in
		 * the request structure used to create the lock is still on
		 * the inflight queue and the reclock has a pointer to it.
		 */
		a = find_reclock(DLM_RES_VMS, req->clm_lockinfo.lockid);
		/* no need to check result - pre_unlock_clm() succeeded */
	
		/*
		 * If this is an async unlock call, then send the UDP response
		 * now to cause the function call to complete
		 */
		if (req->clm_respflags & RESP_AST) {
			req->clm_status = DLM_NORMAL;
			sendudp(req);
			req->clm_status = DLM_FORWARD;
		}

		/* get the resource pointer for the lock being referenced */
		if (a->lck.rh.rh && (rp = resptr(a->lck.rh))) {
			/*
			 * local resource exists, use its info, but copy the
			 * name and length into the nameinfo of the request
			 * in case the resource has moved from the master.
			 */
			resname_bcopy(rp->name, &req->clm_lockreq.nameinfo.name[0],
					rp->namelen);
			req->clm_lockreq.nameinfo.namelen = rp->namelen;
			stat = pti_call(req,
				rp->master ? rp->master : OurSiteID);
		} else {
                        /* save cancel til we have local resource */
                        hold_cancel(a, req);
                }
	} else if (req->clm_status == DLM_NORMAL) {
		/*
		 * Unlock successful.  If AST flag is set, then send the
		 * unlock AST.
		 */
		if (req->clm_respflags & RESP_AST) {
			send_unlck_ast(req);
		}
		/*
		 * Here if pre_proc_unlock_clm completed processing locally.
		 * If site is zero, local node was master and we're done,
		 * otherwise, forward request to remote master
		 */
		if (0 != site) {
			stat = pti_call(req, site);
			clm_reply(0,req);	/* don't free request */
		} else
			clm_reply(req,0);	/* free request struct */
	} else {
		/*
		 * Here if pre_proc_unlock_clm returned some kind of error.
		 * Send reply to client and free the request.
		 */
		clm_reply(req,0);
	}

	DBPRINTF((DBG_FUNC, "leave begin_unlock_clm\n"));
	return stat;
}

/*
 * Name
 *      cont_unlock_clm
 * Description
 *      Continue an unlock request coming back from the primary
 *      Detailed description contained in code.
 * Arguments
 *      req - the request that was sent to the master
 *      resp - the response from the master
 * Returns
 *      NOTHING
 * Global Data Affected
 *      NONE
 */
void
cont_unlock_clm(clm_trans_t *req, clm_trans_t *resp)
{

	DBPRINTF((DBG_FUNC, "enter cont_unlock_clm\n"));
	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_CONT_UNLOCK_CLM,
		req->clm_pid, req->clm_sequence, req->clm_lockinfo.lockid,
		resp->clm_status);

	/*
	 * if the request was forwarded, cancelled, or aborted, then
	 * processing still has to happen on the local node,
	 * otherwise, just free the original request transaction
	 */
        if (resp->clm_status == DLM_TIMEOUT) {
		/*
		 * Here if processing a response to a timeout request.  Call
		 * the master unlock to send the timeout AST, but no client
		 * response for the cancel is required because a client didn't
		 * send it - the timeout subsystem did.  Free the request.
		 */
		local_unlock_clm(resp);
		clm_free_trans(req);
        } else if (req->clm_respflags & (RESP_UDP | RESP_AST)) {
		/*
		 * Here to handle an unlock request that couldn't be handled
		 * locally before master processing.  Call local unlock to
		 * handle the response then send the reply to the client.
		 * We want to pass the response into local unlock because it
		 * contains the correct status that needs to be returned to
		 * the client.
		 */
		local_unlock_clm(resp);
		/*
		 * If the response has both the FORCE and CANCEL flags set,
		 * then we need to resend this request as a normal unlock
		 * without sending a reply to the client.
		 */
		if (resp->clm_lockinfo.lk_flags & LKM_FORCE) {
			req->clm_lockinfo.lk_flags &= ~LKM_CANCEL;
			begin_unlock_clm(req);
		} else {
			/* here for a normal CANCEL response */
			clm_reply(req, resp);
		}
	} else {
		/*
		 * Here after reading a master's response to an unlock
		 * request that has already been handled locally.  Toss it.
		 */
		local_unlock_clm(resp);
		clm_free_trans(req);
	}
	DBPRINTF((DBG_FUNC, "leave cont_unlock_clm\n"));
}

/*
 * Name
 *	pre_proc_unlock_clm
 * Description
 *	See if the unlock can be processed completely on this node or at
 *	least before it gets passed to the master.  This routine is always
 *	called locally to the unlock request and always before control
 *	has been passed to the master for the resource.
 * Arguments
 *	req - the unlock request
 * 		input structure: lockinfo
 * Returns in req->clm_status:
 *	DLM_IVLOCKID - bad lock id specified
 *	DLM_IVRESHANDLE - bad resource handle in lock
 *	DLM_DENIED - tried to cancel ungranted lock
 *	DLM_DENIED - tried to cancel converting lock without LKM_CANCEL set
 *	DLM_FORWARD - local processing not possible, forward request
 *	DLM_NORMAL - okay
 * 		Output Structure: lockinfo
 * Returns:
 *	site id of master if called on secondary & local processing complete
 *	site id of master if called on secondary requests needs forwarding
 *	otherwise, 0 - status should be obtained from req->clm_status
 * Global Data Affected
 *	NONE
 */
int
pre_proc_unlock_clm(clm_trans_t *req)
{
    struct reclock *a;
    struct resource *rp;
    int site;
    int master = 0;
    const char * msg_str;


    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_PRE_UNLOCK_CLM,
	    req->clm_pid, req->clm_sequence, req->clm_lockinfo.lockid,
	    req->clm_lockinfo.lk_flags);

    DBPRINTF((DBG_FUNC, "enter pre_proc_unlock_clm\n"));

    /* check for an invalid lock id */
    a = NULL;
    msg_str = "got bad id";

    if ( req->clm_lockinfo.lockid != 0 )
    {
	a = find_reclock( DLM_RES_VMS, req->clm_lockinfo.lockid );
	msg_str = "can't find reclock for";
    }

 
    if ( (a == NULL) || !mylock(a, req) || (a->lck.state & LOCK_GONE) )
    {
	DBPRINTF((DBG_CLMLOCK, "[%s:%s] %s 0x%x", current->comm, __FUNCTION__,
		  msg_str, req->clm_lockinfo.lockid));
	req->clm_status = DLM_IVLOCKID;	
	return( 0 );
    }


    /* Cancel requests have to be passed to the master before they can
     * be handled locally
     */

    if (req->clm_lockinfo.lk_flags & LKM_CANCEL) 
    {
	DBPRINTF((DBG_CLMLOCK, "pre_proc_unlock_clm: cancelling lock 0x%x\n", 
		  req->clm_lockinfo.lockid));

	req->clm_status = DLM_FORWARD;	

	/* clear FORCE flag if present */
	req->clm_lockinfo.lk_flags &= ~LKM_FORCE;

	/* forwarding is a special case recognized by the calling
	 * routine.  We do not want to jump to the 'done' label.
	 * Instead, we will return from here.
	 */

	return 0;
    }


    /* If inflight flags are set then a request is still inflight to the
     * master for this resource.  If converting or waiting flags are set, then
     * request is blocked.  Unlock at this point is invalid.
     */

    if (a->lck.state & (INFLIGHT_WAITING | INFLIGHT_CONVERTING |
			LOCK_CONVERTING |  LOCK_WAITING)) 
    {
	DBPRINTF((DBG_CLMLOCK,
		  "pre_proc_unlock_clm: lock 0x%x not granted yet\n",
		  req->clm_lockinfo.lockid));
	if (~req->clm_lockinfo.lk_flags & LKM_FORCE) 
	{
	    req->clm_status = DLM_DENIED;	
	}
	else
	{
	    /* Here if the FORCE flag is going to cause a cancel followed by
	     * an unlock.  Change the request to a CANCEL that has the FORCE
	     * flag also set and pass it on to the master.  When the response
	     * comes back, send it *again* without the CANCEL flag to unlock
	     * the lock.
	     */

	    req->clm_lockinfo.lk_flags |= LKM_CANCEL;
	    req->clm_status = DLM_FORWARD;
	}

	return 0;
    }


    /* If we get to this point, we are either on the master node for the
     * resource, or we are on the secondary and in a position where we can
     * complete the unlock before forwarding the request to the master node.
     */

    /* get the resource pointer for the lock being referenced */
    if (a->lck.rh.rh == 0) 
    {
	bsdlog(LOG_CRIT, MSGSTR(NULL_LOCK_HANDLE, 
				"pre_proc_unlock_clm: NULL resource handle for 0x%x"),
	       req->clm_lockinfo.lockid);
	abort("pre_proc_unlock_clm");
	return( 0 );
    }

    if (!(rp = resptr(a->lck.rh))) 
    {
	bsdlog(LOG_ERR, MSGSTR(NO_HANDLE, 
			       "pre_proc_unlock_clm: can't find handle for 0x%x"),
	       req->clm_lockinfo.lockid);
        bsdlog(LOG_ERR, __FUNCTION__" handle[%lx/%u/%u/%u/%u]\n",
               a->lck.rh.rh,
               a->lck.rh.handle.site,
               a->lck.rh.handle.type,
               a->lck.rh.handle.seg,
               a->lck.rh.handle.cookie);
	req->clm_status = DLM_IVRESHANDLE;	
	return( 0 );
    }

    check_handles(rp, 0) ;

    /*
     * if not the master, go with plan B 
     */

    if ( rp->master == 0 )
    {
	/*
	 * Resource Statistics:
	 *        cs_requests  - unlock request made
	 *        cs_downgrade - unlock resource
	 *        cs_local     - unlock from local node
	 *        cs_same      - unlock from same node as last request
	 *        lastnode     - update lastnode to be current node
	 */
	DBPRINTF((DBG_STATS, 
		  "pre_proc_unlock_clm: Update Resource Statistics\n"));
	rp->stats.cs_requests++;
	rp->stats.cs_local++;
	rp->stats.cs_downgrade++;
	if (rp->lastnode == req->clm_lockreq.lockinfo.site) {	    
	        rp->stats.cs_same++;
	}
	rp->lastnode = req->clm_lockreq.lockinfo.site;


	/* kill granted lock */
	check_handles(rp, 0) ;
	remove_queue(&rp->lockqs[RLQ_GRANT], a, RL_RQ_OFS) ;

	/* 
 	 * if releasing PW or EX lock, update resource lock value
	 * block if requested
	 */ 
	if ((req->clm_lockinfo.lk_flags & (LKM_VALBLK | LKM_INVVALBLK)) &&
		(req->clm_lockinfo.lk_flags & LKM_CANCEL) == 0) {
		valueblock(rp, a->lck.mode, LKM_NLMODE, 
			req->clm_lockinfo.lkvalue, req->clm_lockinfo.lk_flags);
	}
	/*
	 * Get some information from the resource pointer before calling
	 * decr_delete_resource because it could potentially wipe the pointer.
	 * We need to copy the master information and the resource name
	 * information that will/could be needed to find the master.
	 */
	master = rp->master;
	site = a->lck.site;
	req->clm_lockreq.nameinfo.namelen = rp->namelen;
	resname_bcopy(rp->name, &req->clm_lockreq.nameinfo.name[0], rp->namelen);

	free_le(DLM_RES_VMS, a);
	convert_locks(rp, NULL);
#ifdef DBPRINT
	clm_dumpres(rp);
#endif
	DBPRINTF((DBG_FUNC, "leave pre_proc_unlock_clm\n"));
	req->clm_status = DLM_NORMAL;	
/*
 * must be last code referencing resource!!
 */
	decr_delete_resource(rp, 0) ;

	return 0;
    }
    else
    {
	/*
	 * plan B:  Change the grant mode of the local copy of the lock to
	 * 0 so it will not conflict with any future requests and notify the
	 * client that its unlock was successful.  Set the LOCK_GONE bit in
	 * the lock's state field.  Subsequent lookups on that lockid will
	 * check for the bit and return DLM_IVLOCKID if it is set.  Then send
	 * the unlock request to the master copy of the resource.  When the
	 * master's response comes back, it will lookup the local copy of the
	 * lock and blow it away.
	 */

	/* change grant mode to NULL to prevent conflicts */
	a->lck.mode = a->lck.reqmode = LKM_NLMODE;

	/* set LOCK_GONE flag to prevent lookup before removal */
	a->lck.state |= LOCK_GONE;

	/* 
 	 * if releasing PW or EX lock, update resource lock value
	 * block if requested
	 */ 
	if ((req->clm_lockinfo.lk_flags & (LKM_VALBLK | LKM_INVVALBLK)) &&
		(req->clm_lockinfo.lk_flags & LKM_CANCEL) == 0) {
		valueblock(rp, a->lck.mode, LKM_NLMODE, 
			req->clm_lockinfo.lkvalue, req->clm_lockinfo.lk_flags);
	}
	/*
	 * Get some information from the resource pointer before calling
	 * decr_delete_resource because it could potentially wipe the pointer.
	 * We need to copy the master information and the resource name
	 * information that will/could be needed to find the master.
	 */
	master = rp->master;
	site = a->lck.site;
	req->clm_lockreq.nameinfo.namelen = rp->namelen;
	resname_bcopy(rp->name, &req->clm_lockreq.nameinfo.name[0], rp->namelen);

#ifdef DBPRINT
	clm_dumpres(rp);
#endif
	DBPRINTF((DBG_FUNC, "leave pre_proc_unlock_clm\n"));
	req->clm_status = DLM_NORMAL;	


	/* We are not the master for the resource involved, so the request has
	 * to be fowarded to the master.  To get that to happen, we have to
	 * return the site id of the master node for this resource.
	 */

	return( master );
    }
}


/*
 * Name
 *	master_unlock_clm
 * Description
 * 	remove a lock or cancel a request
 *	This routine is run on the node that is supposed to be the master
 *	for the resource referenced by the lock that is being operated on.
 *	If a request is made to operate on a resource that is not mastered
 *	on this node, then the directory node id for the resource is
 *	calculated and the id is returned to the calling function which will
 *	forward the request.
 * Arguments
 *	msg - the message containing the unlock request
 * 		input structure: lockinfo
 * Returns in req->clm_status:
 *	DLM_IVLOCKID - bad lock id specified
 *	DLM_IVRESHANDLE - bad resource handle in lock
 *	DLM_CANCELGRANT - tried to cancel granted lock
 *	DLM_DENIED - tried to cancel ungranted lock
 *	DLM_DENIED - tried to cancel converting lock without LKM_CANCEL set
 *	DLM_FORWARD - local processing not possible, forward request
 *	DLM_NORMAL - okay
 * 		Output Structure: lockinfo
 * Returns:
 *	site id of directory if called as master on wrong node
 *	otherwise, 0 - status should be obtained from req->clm_status
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
master_unlock_clm(clm_msg_t *msg)
{
	clm_trans_t *req = msg->cm_data;
	struct reclock *a;
	struct resource *rp;
	register uint id;
	int	nodeid;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_MASTER_UNLOCK_CLM,
		req->clm_pid, req->clm_sequence, req->clm_lockinfo.lockid,
		req->clm_lockinfo.lk_flags);

	DBPRINTF((DBG_FUNC, "enter master_unlock_clm\n"));
	/*
	 * if we are processing a request from another node, map lockid 
	 */
	if (req->clm_lockinfo.site != OurSiteID)
		id = mapid(req->clm_lockinfo.lockid, req->clm_lockinfo.site) ;
	else
		id = req->clm_lockinfo.lockid;

	if (id == 0 || !(a = find_reclock(DLM_RES_VMS, id)) ||
				!mylock(a, req)) {
		int	site;
		/*
		 * check first for the psycho unlock request that was sent to
		 * a dead master and if so, boot it.
		 */
		if (msg->cm_hops > 80) {
			bsdlog(LOG_ERR, 
				"Master unlock request for %d:%s exceeds 80 "
				"hops.  Dropping request.",
				req->clm_lockreq.nameinfo.namelen,
				printbin(&req->clm_lockreq.nameinfo.name[0],
				req->clm_lockreq.nameinfo.namelen));
			return 0;
		}

		/*
		 * Here if we are acting as a remote master but the master
		 * has moved.  Determine the directory node for the resource
		 * and return that to pti_proc_request() so this request will
		 * be forwarded.
		 */
		site = clm_direct(req->clm_lockreq.nameinfo.name,
				req->clm_lockreq.nameinfo.namelen);
		if (!site) {
			bsdlog(LOG_CRIT, MSGSTR(RET_NULL_DIRECT,
					"master_unlock_clm:clm_direct() "
					"returned NULL"));
			return -1;
		}
		DBPRINTF((DBG_CLMLOCK|DBG_CONFIG,
			"master_unlock_clm: got bad id 0x%x -> not master", 
			req->clm_lockinfo.lockid));

		return site;
	}

	/* get the resource pointer for the lock being referenced */
	if (a->lck.rh.rh == 0) {
		bsdlog(LOG_CRIT, MSGSTR(NO_RES_HANDL, 
		    "master_unlock_clm: NULL resource handle for 0x%x"),
			 id);
		if (req->clm_lockinfo.site == OurSiteID) {
			int	site;
			/*
			 * We were misled to believe that this request was
			 * mastered locally.  Redirect the request to the
			 * directory.
			 */
			site = clm_direct(req->clm_lockreq.nameinfo.name,
					req->clm_lockreq.nameinfo.namelen);
			if (!site) {
				bsdlog(LOG_CRIT, MSGSTR(NULL_RETURN, 
						"master_unlock_clm:"
						"clm_direct() returned NULL"));
				return -1;
			}
			return site;
		}
		/*
		 * Here if we found a copy of our lock on a remote node that
		 * didn't have a valid resource handle.  Bad bad bad.
		 */
		abort("master_unlock_clm");
	}
	if (!(rp = resptr(a->lck.rh))) {
		bsdlog(LOG_ERR, MSGSTR(UNLOCK_NO_HANDLE, 
			       "master_unlock_clm: can't find handle for 0x%x"),
				req->clm_lockinfo.lockid);
		abort("master_unlock_clm");
	}

	/* here if we think we are on the master and not local */
	if (rp->master != 0) {
		int	site;
		/*
		 * Here if we are acting as a remote master but the
		 * master has moved.  Determine the directory node for
		 * the resource and return that to pti_proc_request()
		 * so this request will be forwarded.  This should be
		 * caught when looking up lockid and this check is
		 * unneeded.
		 */
		site = clm_direct(rp->name, rp->namelen);
		if (!site) {
			bsdlog(LOG_CRIT, MSGSTR(NULL_RETURN, 
					"master_unlock_clm:clm_direct() "
					"returned NULL"));
			return -1;
		}
		return site;
	} else {
		/*
		 * If the resource is migrating, bounce message to apply back
		 * pressure.  Caller will retry.
		 */
		if (rp->rsrc_flags & (RSRC_FRZEN | RSRC_STUCK | RSRC_UPDATING))
		{
		    msg->cm_port = PORT_BOUNCE;
		    return msg->cm_node;
		}
		if (a->lck.state & (INFLIGHT_WAITING | INFLIGHT_CONVERTING))
		{
		    /*
		     * here if master is 0 and inflight bit is set.  That
		     * means that we are trying to handle a local unlock
		     * request on a resource that has recently migrated to
		     * this node for a lock that has an outstanding request.
		     * We don't want to deal with this unlock request until
		     * the inflight request is taken care of.
		     */
		    msg->cm_port = PORT_BOUNCE;
		    return msg->cm_node;
		}
		/*
		 * request is not going to be re-directed.  call
		 * cccp_msg_duplicate() to insure that we have a writeable copy
		 * of the message, reset the req pointer and enqueue
		 * the response back to the secondary.
		 */

		msg = cccp_msg_duplicate( msg );
		req = msg->cm_data;
		req->clm_direction = PTI_RESPONSE;
		req->clm_rt_master = OurSiteID;

		/*
		 * Resource Statistics:
		 *        cs_requests  - unlock request made
		 *        cs_downgrade - unlock resource
		 *        cs_remote    - unlock from remote node
		 *        cs_same      - unlock from same node as last request
		 *        lastnode     - update lastnode to be current node
		 */
		DBPRINTF((DBG_STATS, 
			  "master_unlock_clm: Update Resource Statistics\n"));
		rp->stats.cs_requests++;
		if (req->clm_lockreq.lockinfo.site == OurSiteID) {
		        rp->stats.cs_local++;
		} else {
		        rp->stats.cs_remote++;
		}
		rp->stats.cs_downgrade++;
		if (rp->lastnode == req->clm_lockreq.lockinfo.site) {
		        rp->stats.cs_same++;
		}
		rp->lastnode = req->clm_lockreq.lockinfo.site;

		clm_queue_respmsg(msg);
	}

	nodeid = a->lck.site;

	if (req->clm_lockinfo.lk_flags & LKM_CANCEL) {
		/* 
		 * LKM_CANCEL specifies that a lock conversion
		 * should be cancelled and its old mode restored
		 */
		if ((a->lck.state & LOCK_CONVERTING) != 0) {
			/* here to cancel conversion request */

			/*
			 * if this is a local cancel, then we want to let
			 * the local routine handle it so it can send the UDP
			 * reply that we are not in a position to send.
			 */
			if (a->lck.site == OurSiteID)
				goto done;

			/* put converter back on grant q at old mode */
			check_handles(rp, 0) ;
			a->lck.state &= ~LOCK_CONVERTING;
			remove_queue(&rp->lockqs[RLQ_CONV], a, RL_RQ_OFS) ;
			/* replace at old mode */
			lck_set_seq(rp, a);
			add_queue(&rp->lockqs[RLQ_GRANT], a, RL_RQ_OFS) ;
			a->lck.state |= LOCK_GRANTED;
			check_handles(rp, 0) ;
			delete_timeout_reclock(a);
			if (a->lck.site == OurSiteID) {
			    (void) memset( & a->route.cbasth, 0,
					   sizeof( a->route.cbasth ) );
			    unset_lock_timer(a);
			}

			/* only send notification to local clients
			 * cont_unlock will eventually lead back 
			 * to this routine on the secondary
			 */
			if (a->lck.site == OurSiteID) {
			    if (req->clm_lockinfo.lk_flags & LKM_TIMEOUT)
			        notify(a, RESP_AST, DLM_TIMEOUT);
			    else 
			        notify(a, RESP_AST, DLM_CANCEL);
			}

			req->clm_status =
				(req->clm_lockinfo.lk_flags & LKM_TIMEOUT) ?
				DLM_TIMEOUT : DLM_NORMAL;	

			/*
			 * blocked converts can block other requests.  Call
			 * convert_locks() to check for any grantable converts
			 */
			convert_locks(rp, NULL);

			goto done;

		} else if (a->lck.state & LOCK_WAITING) {
			/*
			 * if this is a local cancel, then we want to let
			 * the local routine handle it so it can send the UDP
			 * reply that we are not in a position to send.
			 */
			if (a->lck.site == OurSiteID)
				goto done;

			/* kill a blocked lock request */
			check_handles(rp, 0) ;
			remove_queue(&rp->lockqs[RLQ_WAIT], a, RL_RQ_OFS) ;

			/* only send notification to local clients
			 * cont_unlock will eventually lead back 
			 * to this routine on the secondary
			 */
			if (a->lck.site == OurSiteID) {
			    if (req->clm_lockinfo.lk_flags & LKM_TIMEOUT)
			        notify(a, RESP_AST, DLM_TIMEOUT);
			    else 
			        notify(a, RESP_AST, DLM_ABORT);
			}
			/*
			 * clear the LKM_FORCE flag in the response to
			 * prevent the attempt to unlock the lock
			 */
			req->clm_lockinfo.lk_flags &= ~LKM_FORCE;

			/*
			 * This block does not jump to the 'done' label as
			 * the other blocks do because the reclock that it
			 * cancelled has to be cleaned up and freed.
			 */

		} else {
			/* request already granted */
			DBPRINTF((DBG_CLMLOCK,
				"master_unlock_clm: CANCELGRANT for 0x%x\n", 
				req->clm_lockinfo.lockid));
			req->clm_status = DLM_CANCELGRANT;

			/* if this was a timeout CANCEL, drop it on
			 * the floor and kick it under the rug */
			if (req->clm_lockinfo.lk_flags & LKM_TIMEOUT) {
				DBPRINTF((DBG_TIMEOUT,
					"Timed-out request already granted"));
				req->clm_respflags = 0;
			}

			goto done;
		}

	/* can't cancel ungranted or converting lock w/o LKM_CANCEL */
	} else if ((a->lck.state & LOCK_CONVERTING) ||
			(a->lck.state & LOCK_WAITING)) {
		req->clm_status = DLM_DENIED;	
		goto done;
	} else {
		/* kill granted lock */
		check_handles(rp, 0) ;
		remove_queue(&rp->lockqs[RLQ_GRANT], a, RL_RQ_OFS) ;

		/*
		 * set the GONE flag if master is local so local_unlock_clm
		 * won't try to unlock after us.  This is necessary if the
		 * master copy migrates back to the local node after the
		 * unlock is processed.
		 */
		if (nodeid == OurSiteID)
			req->clm_lockinfo.state |= LOCK_GONE;

	}
	/* 
 	 * if releasing PW or EX lock, update resource lock value
	 * block if requested
	 */ 
	if ((req->clm_lockinfo.lk_flags & (LKM_VALBLK | LKM_INVVALBLK)) &&
		(req->clm_lockinfo.lk_flags & LKM_CANCEL) == 0) {
		valueblock(rp, a->lck.mode, LKM_NLMODE, 
			req->clm_lockinfo.lkvalue, req->clm_lockinfo.lk_flags);
	}
	free_le(DLM_RES_VMS, a);
	convert_locks(rp, NULL);
#ifdef DBPRINT
	clm_dumpres(rp);
#endif
	DBPRINTF((DBG_FUNC, "leave master_unlock_clm\n"));
	req->clm_status = (req->clm_lockinfo.lk_flags & LKM_TIMEOUT) ?
			DLM_TIMEOUT : DLM_NORMAL;	
    /*
     * this code MUST BE LAST reference to resource
     */
	decr_delete_resource(rp, 0) ;
done:

	return 0;
}


/*
 * Name
 *	local_unlock_clm
 * Description
 * 	remove a lock or cancel a request
 *	This unlock handler wil run as a continue function for a request
 *	that has been handled by the master and needs to be continued on the
 *	local machine.
 * Arguments
 *	msg - the message containing the unlock request
 * 		input structure: lockinfo
 * Returns in req->clm_status:
 *	DLM_IVLOCKID - bad lock id specified
 *	DLM_IVRESHANDLE - bad resource handle in lock
 *	DLM_CANCELGRANT - tried to cancel granted lock
 *	DLM_DENIED - tried to cancel ungranted lock
 *	DLM_DENIED - tried to cancel converting lock without LKM_CANCEL set
 *	DLM_FORWARD - local processing not possible, forward request
 *	DLM_NORMAL - okay
 * 		Output Structure: lockinfo
 * Returns:
 *	site id of directory if called as master on wrong node
 *	otherwise, 0 - status should be obtained from req->clm_status
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
local_unlock_clm(clm_trans_t *req)
{
	struct reclock *a;
	struct resource *rp;
	register uint id;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_LOCAL_UNLOCK_CLM,
		req->clm_pid, req->clm_sequence, req->clm_lockinfo.lockid,
		req->clm_lockinfo.lk_flags);

	DBPRINTF((DBG_FUNC, "enter local_unlock_clm\n"));

	/* if LOCK_GONE in response, master is local, unlock already done */
	if (req->clm_lockinfo.state & LOCK_GONE)
		goto done;

	id = req->clm_lockinfo.lockid;

	if (id == 0 || !(a = find_reclock(DLM_RES_VMS, id)) ||
				!mylock(a, req)) {
		goto done;
	}

	/* get the resource pointer for the lock being referenced */
	if (a->lck.rh.rh == 0) {
		bsdlog(LOG_CRIT, MSGSTR(LOCAL_NULL_RES, 
		    "local_unlock_clm: NULL resource handle for 0x%x"),
			 id);
		abort("local_unlock_clm");
		goto done;
	}
	if (!(rp = resptr(a->lck.rh))) {
		bsdlog(LOG_ERR, MSGSTR(LOCAL_UNLK_ERR,
			"local_unlock_clm: can't find handle for 0x%x"),
			req->clm_lockinfo.lockid);
		abort("local_unlock_clm");
		goto done;
	}

	/*
	 * Update master pointer on local copy of the resource.  This should
	 * be done on all continue functions that have received direct
	 * responses from the node with the master copy of a resource so that
	 * this information is kept up to date.  The migration algorithms rely
	 * on this field to be accurate on nodes that a resource is likely to
	 * migrate to because of frequent accesses
	 */
	if (rp->master && req->clm_rt_master != OurSiteID) {
	        /* For timeout, rp could be request, so check TIMEOUT flag */
	        CLM_ASSERT(((req->clm_rt_master &&
			     req->clm_rt_master != OurSiteID) ||
			    (req->clm_lockinfo.lk_flags & 
			     (LKM_CANCEL|LKM_TIMEOUT))),
			   "Set master local");
		rp->master = req->clm_rt_master;
	}

	if (req->clm_lockinfo.lk_flags & LKM_CANCEL) {
		/* 
		 * LKM_CANCEL specifies that a lock conversion
		 * should be cancelled and its old mode restored
		 */
		if ((a->lck.state & LOCK_CONVERTING) != 0) {
			/* here to cancel conversion request */
			/* put converter back on grant q at old mode */
			check_handles(rp, 0) ;
			a->lck.state &= ~LOCK_CONVERTING;
			remove_queue(&rp->lockqs[RLQ_CONV], a, RL_RQ_OFS) ;
			/* reset sequence number if this is master copy */
			if (0 == rp->master)
				lck_set_seq(rp, a);
			/* replace at old mode */
			add_queue(&rp->lockqs[RLQ_GRANT], a, RL_RQ_OFS) ;
			a->lck.state |= LOCK_GRANTED;
			check_handles(rp, 0) ;
			delete_timeout_reclock(a);
			if (a->lck.site == OurSiteID) {
			    unset_lock_timer(a);
			    (void) memset( & a->route.cbasth, 0,
					   sizeof( a->route.cbasth ) );
			}

			/* only send notification to local clients
			 * cont_unlock will eventually lead back 
			 * to this routine on the secondary
			 */
			if (a->lck.site == OurSiteID) {
			    if (req->clm_lockinfo.lk_flags & LKM_TIMEOUT)
			        notify(a, RESP_AST, DLM_TIMEOUT);
			    else 
			        notify(a, RESP_AST, DLM_CANCEL);
			}

			req->clm_status =
				(req->clm_lockinfo.lk_flags & LKM_TIMEOUT) ?
				DLM_TIMEOUT : DLM_NORMAL;	

			/*
			 * blocked converts can block other requests.  Call
			 * convert_locks() to check for any grantable converts
			 */
			convert_locks(rp, NULL);

			goto done;

		} else if (a->lck.state & LOCK_WAITING) {
			/* kill a blocked lock request */
			check_handles(rp, 0) ;
			remove_queue(&rp->lockqs[RLQ_WAIT], a, RL_RQ_OFS) ;

			/* only send notification to local clients
			 * cont_unlock will eventually lead back 
			 * to this routine on the secondary
			 */
			if (a->lck.site == OurSiteID) {
			    unset_lock_timer(a);
			    if (req->clm_lockinfo.lk_flags & LKM_TIMEOUT)
			        notify(a, RESP_AST, DLM_TIMEOUT);
			    else 
			        notify(a, RESP_AST, DLM_ABORT);
			}
			/*
			 * clear the LKM_FORCE flag in the response to
			 * prevent the attempt to unlock the lock
			 */
			req->clm_lockinfo.lk_flags &= ~LKM_FORCE;

			/*
			 * This block does not jump to the 'done' label as
			 * the other blocks do because the reclock that it
			 * cancelled has to be cleaned up and freed.
			 */

		} else {
			/* request already granted */
			DBPRINTF((DBG_CLMLOCK,
				"local_unlock_clm: CANCELGRANT for 0x%x\n", 
				req->clm_lockinfo.lockid));
			req->clm_status = DLM_CANCELGRANT;

			/* if this was a timeout CANCEL, drop it on
			 * the floor and kick it under the rug */
			if (req->clm_lockinfo.lk_flags & LKM_TIMEOUT) {
				DBPRINTF((DBG_TIMEOUT,
					"Timed-out request already granted"));
				req->clm_respflags = 0;
			}

			goto done;
		}

	/* can't cancel ungranted or converting lock w/o LKM_CANCEL */
	} else if ((a->lck.state & LOCK_CONVERTING) ||
			(a->lck.state & LOCK_WAITING)) {
		req->clm_status = DLM_DENIED;	
		goto done;
	} else {
		/* kill granted lock */
		check_handles(rp, 0) ;
		remove_queue(&rp->lockqs[RLQ_GRANT], a, RL_RQ_OFS) ;
	}
	/* 
 	 * if releasing PW or EX lock, update resource lock value
	 * block if requested
	 */ 
	if ((req->clm_lockinfo.lk_flags & (LKM_VALBLK | LKM_INVVALBLK)) &&
		(req->clm_lockinfo.lk_flags & LKM_CANCEL) == 0) {
		valueblock(rp, a->lck.mode, LKM_NLMODE, 
			req->clm_lockinfo.lkvalue, req->clm_lockinfo.lk_flags);
	}
	free_le(DLM_RES_VMS, a);
	convert_locks(rp, NULL);
#ifdef DBPRINT
	clm_dumpres(rp);
#endif
	DBPRINTF((DBG_FUNC, "leave local_unlock_clm\n"));
	req->clm_status = (req->clm_lockinfo.lk_flags & LKM_TIMEOUT) ?
			DLM_TIMEOUT : DLM_NORMAL;	
/*
 * must be last code referencing resource!!
 */
	decr_delete_resource(rp, 0) ;

done:
	return 0;
}
