/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_compat.c,v 1.2 2001/03/12 19:59:58 wombat Exp $"

/*
 *   COMPONENT_NAME: CLLOCKD
 *
 *   FUNCTIONS: 
 *		clm_vers_adj
 */

#ifdef __linux__
/*
 * As you may guess, these currently have no function.  In fact, as
 * you can see, they were essentially gutted before we even started
 * the port to Linux.  I assume these were left in here earlier to
 * avoid having to change the code in other areas that called these.
 * We, for now, follow this same approach.
 */
#endif /* __linux__ */

#ifndef lint
#endif

#include <linux/types.h>

#include "dlm.h"
#include "clmstructs.h"
#include "cccp.h"

/* ============================================================ */

/*
 * This file contains the translation routines for each global struct that
 * has different versions.  These old versions of the structs are
 * kept as separate structs called: 
 *	{structname}{old version}
 * For example, the 3.1 version of lockinfo is named lockinfo31.
 *
 * The routines called
 *      from_{structname}{old version} 
 * convert a old version of a structure to its current format.
 *
 * The routines called:
 *      to_{structname}{old version}
 * convert the current structure to its old format.
 *
 * Note that clm_vers_snd() and clm_vers_rcv() do the conversions for
 * the struct transaction.
 * The conversion routines also exist in clm_migrate.c and
 * clm_reconfig.c where they handle structs only defined in those modules.
 */


/*
 * Name
 *	clm_vers_snd
 * Description
 *	Examine the outgoing message and make any adjustments to compensate
 *	for version skew.  The version to adjust the message to is specified
 *	by the calling function because of situations where a message may get
 *	redirected to a node other than the one in the cm_node field.  In
 *	this case, messages have to be adjusted to the oldest active version.
 * Arguments
 *	message pointer
 *	destination version
 * Returns 
 *	DLM_NORMAL		Compatibility changes successful.
 *	DLM_VERS_CONFLICT	Message cannot be sent to destination
 * Global Data Affected
 *	None
 */
dlm_stats_t
clm_vers_snd(clm_msg_t	*msg, int destversion)
{
#ifndef __linux__
	clm_trans_t	*req;			/* request pointer */
	int		ourversion;
	int		type;
#endif /* !__linux__ */
	CLM_ASSERT((msg && msg->cm_data),
               "No message present in clm_vers_snd");

	/* no version skew, nothing to do */
	return DLM_NORMAL;
}


/*
 * Name
 *	clm_vers_rcv
 * Description
 *	Examine the incoming message and make any adjustments to compensate
 *	for version skew.  This is accomplished by taking the original
 *	request pointer and moving it to the cm_private field and then
 *	allocating a current version of the transaction structure and copying
 *	the appropriate data into the new structure.  This translation must
 *	be undone before the receive callback relinquishes the message pointer
 *	to the comms layer because the comms layer will free whatever
 *	cm_data is pointing to.
 * Arguments
 *	message pointer
 * Returns 
 *	DLM_NORMAL		Compatibility changes successful.
 *	DLM_VERS_CONFLICT	Message cannot be sent to destination
 * Global Data Affected
 *	None
 */
dlm_stats_t
clm_vers_rcv(clm_msg_t	*msg)
{
#ifndef __linux__
	clm_trans_t	*req = (clm_trans_t *)msg->cm_data;/* request pointer */
	static const char fn[] = "clm_vers_rcv";
#endif /* !__linux__ */
	CLM_ASSERT((msg && msg->cm_data), "No message present in clm_vers_rcv");

	return DLM_NORMAL;
}


/*
 * Name
 *	clm_vers_reset
 * Description
 *	Examine the provided message and if it was translated, reset the
 *	pointers to what they were.
 * Arguments
 *	message pointer
 * Returns 
 *	DLM_NORMAL		Compatibility changes successful.
 * Global Data Affected
 *	None
 */
dlm_stats_t
clm_vers_reset(clm_msg_t	*msg)
{
#ifndef __linux__
	clm_trans_t	*req;
#endif /* !__linux__ */
	CLM_ASSERT((msg && msg->cm_data),
			"No message present in clm_vers_reset");
#ifndef __linux__
	req = (clm_trans_t *)msg->cm_data;
#endif /* !__linux__ */
	/* if copy was made, original will be oldtrans ptr */

	return DLM_NORMAL;
}
