/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_cti.c,v 1.20 2001/11/06 13:32:46 bkon Exp $"

/*
 *   COMPONENT_NAME: CLLOCKD
 *
 *   FUNCTIONS: MSGSTR
 *		clm_process
 *		clm_recv
 *		clm_reply
 *		clm_request
 *		clm_send
 *		grab_resp
 *		k_clm_recv
 *		k_clm_send
 *		k_dump_resources
 *		k_local_primary
 *		k_res_count
 *		k_set_debug
 *		notify
 *              pop_seq
 *              push_seq
 *		sendast
 *		sendsync
 *		sendudp
 */

/* ============================================================ */
/*
 * NOTE: the system call "catch" routines in here need to be
 * properly tied into the write routine in dlm_kerndd and then
 * they should mostly "just work" (famous last words...)
 *
 * Oh yeah, these are the routines named "k_...".
 */
/* ============================================================ */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/signal.h>
#include <linux/tqueue.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/param.h>	/* for MAXHOSTNAMELEN */
#include <linux/sched.h>
#include <asm/uaccess.h>

#include "dlm.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "dlm_kernel.h"
#include "pti.h"
#include "cltrc.h"
#include "clm_data.h"
#include "clm_locksw.h"
#include "dlm_version.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */
struct tq_struct dlm_keventd_task;

/* ============================================================ */

#ifdef REQ_STATS
#define INC_COUNT_TIMESTAMP( word, counter ) \
	do { \
	    struct timestruc_t foo; \
	    unsigned now = gettimeofday(&foo, NULL); \
	    static unsigned then = 0; \
\
	    (counter)++; \
	    if ((now - then) > 5) { \
		then = now; \
		bsdlog( LOG_DEBUG, "%s count = %d\n", (word), (counter) ); \
	    } \
	} while ( 0 )
#endif /* REQ_STATS */

/*
 *  stack of sync sequence numbers, used for clmlock_sync call
 */
typedef struct sync_seq 
{
    dlm_listnode_t  link;
    unsigned int seq;       /* seq number */
    int tid;                /* thread id */
} seq_t;


/* ============================================================ */

static int clm_send(clm_trans_t *usrreq, int clpid);
static int clm_recv(clm_trans_t *usrresp, unsigned int seq, int clpid);

/* ------------------------------------------------------------ */
/*
 * Name: push_seq
 *
 * Description:
 *	Pushes the sequence number given and the client thread id 
 *      onto the client's synchronous sequence number stack.  This sync seq 
 *      number should not be confused with seq number.  The sync seq
 *      number is used when the clmlock_sync call is made.  The thread id 
 *      stored in order to prevent the clmlock_sync call reply from being 
 *      sent to the wrong thread in the process.  For non-thread-safe clients,
 *      the tid should be 0.
 *
 * Parameters:
 *	cp              Ptr to the client
 *	seq             Seq number to push on the stack
 *      tid             Thread id to push on the stack.
 *
 * Return value:
 *	void
 *
 * Side effects:
 *	
 */
static void
push_seq(clm_client_t *cp, int seq, int tid)
{
    /* MALLOC more space on stack for sequence # */
    seq_t *tmp = (seq_t *)MALLOC(sizeof(seq_t));

    DBPRINTF((DBG_FUNC, "enter push_seq\n"));

    if (!tmp) {
	abort("push_seq");
    }
	
    DBPRINTF((DBG_SYNC, "push_seq: push %d onto sequence stack\n", seq));  

    /* Update stack */
    tmp->seq = seq;
    tmp->tid = tid;
    dlm_list_add_head( & cp->sync_seq, (dlm_listnode_t *) tmp );

    DBPRINTF((DBG_FUNC, "exit push_seq\n"));
    return;
}

/*
 * Name: pop_seq
 *
 * Description:
 *	Search for last sequence number with a matching tid in the client's
 *      synchronous sequence number stack.  Pop this sequence number 
 *      off the stack.  This sync seq number  should not be confused with 
 *      seq number.  The sync seq number is used when the clmlock_sync call 
 *      is made.  Specifying a specific thread id ensures that the correct 
 *      sequence number is popped off for the client.
 *
 * Parameters:
 *	cp              Ptr to the client
 *      tid             Thread id to search for.
 *
 * Return value:
 *	int             The sync sequence number popped off the stack
 *
 * Side effects:
 *	
 */
static unsigned int
pop_seq( clm_client_t * cp, int tid )
{
    seq_t    * curr;
    unsigned   seq;

    DBPRINTF((DBG_FUNC, "enter pop_seq\n"));
	
    for ( EACH_DLM_LISTNODE( & cp->sync_seq, curr ) )
    {
	/* Check for matching tids */
	if ( curr->tid == tid )
	{
	    dlm_list_remove( (dlm_listnode_t *) curr );

	    seq = curr->seq;
	    FREE( curr );
		
	    DBPRINTF((DBG_SYNC, "pop_seq: pop %d off sequence stack\n", seq));
	    DBPRINTF((DBG_FUNC, "exit pop_seq\n"));
	    return (seq);
	}
    }

    return (0);
}


/*
 * Name: k_clm_send
 *
 * Description:
 *	A wrapper for clm_send, to check structure validity and guarantee
 * correct locking.
 *
 * Parameters:
 *	usrreq	Address of transaction structure in user space
 *	len		Size of transaction structure
 *  clpid   Pid of client making request
 *
 * Return value:
 *	CLM error code
 *
 * Side effects:
 *	Takes/releases global entry-point lock
 */

int
k_clm_send(clm_trans_t *usrreq, int len, int clpid)
{
    int status;

    att_seg();

    bsdlog(LOG_INFO,
           "k_clm_send, pointer/len/sizeof/pid/kproc/ID[%p/%d/%d/%d/%d/%d]\n",
           usrreq, len, sizeof(*usrreq), clpid, kproc_pid, OurSiteID);

    status = -1;
    if (len == sizeof(*usrreq))
    {
again:
	acquire_lock(LK_ENTRY, clpid);
	if (kproc_pid) 
	{
	    if (-1 == OurSiteID) 
	    {
		release_lock(LK_ENTRY, clpid);
		bsdlog(LOG_DEBUG, "OurSiteID is -1, sleeping");
		clm_sleep(1);
		goto again;
	    }
	    status = clm_send(usrreq, clpid);
	} 
	else 
	{
	    status = -2;
	}

	pti_flush_responses();
	release_lock(LK_ENTRY, clpid);
    }

    det_seg();
    return(status);
}

/*
 * Name: clm_send
 *
 * Description:
 *	"Send" transaction structure from user to kernel space for processing.
 * This is equivalent to the UDP receive in the user-level lock manager.
 *
 * Parameters:
 *	usrreq		Address of transaction structure in user space
 *
 * Return value:
 *	CLM error code
 *
 * Side effects:
 *	Locks taken/released, etc.
 */

static int
clm_send(clm_trans_t *usrreq, int clpid)
{
	clm_trans_t *req;

	/* Make sure we're not getting too far ahead of ourselves.  -jjd- */

	DBPRINTF((DBG_CTI, "[%s:%s] %s (%d)\n", current->comm,
		  __FUNCTION__, "enter", clpid));

	req = clm_alloc_trans();
	if (req == NULL) {
	    bsdlog(LOG_ERR, MSGSTR(BUFFER_ERR,
			"clm_send: can't get receive buffer"));
	    return(-3);
	}

	DBPRINTF((DBG_CTI,"Copyin from 0x%x to 0x%x\n",usrreq,req));
        if (DLM_KERNEL_PID == clpid) {
            (void) memcpy(req, usrreq, sizeof(clm_trans_t));
        } else {
          if (copy_from_user(req, usrreq, sizeof(clm_trans_t))) {
              bsdlog(LOG_ERR, MSGSTR(COPY_FAILED, "Copyin failed"));
              clm_free_trans(req);
              return(-4);
          }
       }



#if defined(TIME_STATS)
	req->stats.slot = 0;
	STAMP(req);
#endif

	clm_process(req);

	DBPRINTF((DBG_CTI, "[%s:%s] %s (%d)\n", current->comm,
		  __FUNCTION__, "exit", clpid));
	return (0);
}

/*
 * Name: k_clm_recv
 *
 * Description:
 *	A wrapper for clm_recv, to check structure validity and guarantee
 * correct locking.
 *
 * Parameters:
 *	usrreq	Address of transaction structure in user space
 *	len		Size of transaction structure
 *  seq     Sequence number of client's request
 *  clpid   Pid of client making request
 *
 * Return value:
 *	CLM error code
 *
 * Side effects:
 *	Takes/releases global entry-point lock
 */

int
k_clm_recv(clm_trans_t *usrresp, int len, unsigned int seq, int clpid)
{
    int status;


    att_seg();

    bsdlog(LOG_INFO,
           "k_clm_recv, pointer/len/sizeof/seq/pid[%p/%d/%d/%d/%d]\n",
           usrresp, len, sizeof(*usrresp), seq, clpid);

    status = -1;
    if ( len == sizeof(*usrresp) )
    {
	acquire_lock(LK_ENTRY, clpid);
	status = (kproc_pid != 0) ? clm_recv( usrresp, seq, clpid ) : -2;
	release_lock(LK_ENTRY, clpid);
    }

    det_seg();
    return(status);
}

/*
 * Name: clm_recv
 *
 * Description:
 *	"Receive" transaction structure in user from kernel space after
 * processing.  This is equivalent to the UDP send in the user-level lock
 * manager.
 *
 * Parameters:
 *	usrreq		Address of transaction structure in user space
 *
 * Return value:
 *	CLM error code
 *
 * Side effects:
 *	None
 */

static clm_trans_t *
grab_resp(clm_client_t *cp, unsigned int seq)
{
    clm_trans_t *p;
    clm_trans_t **q = &cp->cli_reply;

    DBPRINTF((DBG_CTI,"[%s:%s] enter( seq=%d )\n",
	      current->comm, __FUNCTION__, seq));

    p = NULL;
    while (NULL != *q)
    {
	bsdlog( LOG_INFO, "[%s:%s] *q = 0x%08lx (%d ?= %d)\n",
	        current->comm, __FUNCTION__,
	        (unsigned long) *q, (*q)->clm_sequence, seq );

	if ( (*q)->clm_sequence == seq) 
	{
	    p = *q;
	    *q = p->clm_next;
	    break;
	}

	q = (clm_trans_t **) & (*q)->clm_next;
    }

    DBPRINTF((DBG_CTI,"[%s:%s] exit trans=%p\n",
	      current->comm, __FUNCTION__, p));
    return( p );
}

static int
clm_recv(clm_trans_t *usrresp, unsigned int seq, int clpid)
{
    int status;
    clm_client_t *cp;
    clm_trans_t *tmp;

    DBPRINTF((DBG_CTI,"enter %s pid=%d seq=%d\n",
              __FUNCTION__, clpid, seq));

    cp = findclientbypid(OurSiteID,clpid);
    if (!cp) {
	bsdlog(LOG_ERR, MSGSTR(CLIENT_NOT_FOUND,
			       __FUNCTION__ "could not find client %d"),clpid);
	return(-3);
    }

    for (;;) {
	tmp = grab_resp(cp,seq);
	if (tmp)
	    break;
	status = wait_for_it(LK_ENTRY,cp, clpid);
	if (status)
	    return status ;
    }

    DBPRINTF((DBG_CTI, "[%s:%s] Copyout from 0x%x to 0x%x\n",
                current->comm, __FUNCTION__, tmp, usrresp));

    if (DLM_KERNEL_PID == clpid) {
        (void) memcpy(usrresp, tmp, sizeof(clm_trans_t));
    } else {
      if (copy_to_user(usrresp, tmp, sizeof(clm_trans_t))) {
            bsdlog(LOG_ERR,MSGSTR(COPY_OUT_FAILED,
                   __FUNCTION__ "Copyout failed"));
            return(-4);
      }
    }


    clm_free_trans(tmp);

    DBPRINTF((DBG_CTI,"leave clm_recv (%d)\n",clpid));
    return(0);
}


/*
 * This routine is the common code, called regardless of whether we came in
 * through clm_getreq() or clm_send().  -jjd-
 */

/*
 * Name
 *	clm_process
 * Description 
 *	Do some initial request validation before handing the request
 * 	off to clm_request()
 * Arguments
 *	req - the request
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
clm_process(clm_trans_t *req)
{
    DBPRINTF((DBG_CTI, "[%s:%s] BEGIN TRANSACTION %s\n",
	      current->comm, __FUNCTION__,
	      transaction_type_string( req->clm_type )));
    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_CLM_PROCESS,
	    req->clm_pid, req->clm_sequence, 0x00, 0x00);

    if (!(req->clm_authpid = checkclient(req->clm_pid, req->clm_sequence))
	&& req->clm_type != DLM_APPREG) 
    {
	bsdlog(LOG_ERR, MSGSTR(ILLEGAL_CLIENT,
			       "clm_process: Unrecognized client %d"), 
	       req->clm_pid);
	req->clm_status = DLM_REJECTED;
	clm_reply(req, NULL);	

	/*
	 * we can't have gotten here if the lock manager isn't running, but
	 * if clm_active isn't set yet, then we are not yet handling client
	 * requests.  clm_buff_check() returns TRUE is we've already queued
	 * lot's of requests.  grace_period is true while the lock manager
	 * is in reconfig state.  Return grace status to make client sleep
	 * and retry.
	 */
    }
    else if ((!clm_active || clm_buff_check() || (grace_period))
	     && req->clm_type != DLM_APPREG)
    {
	bsdlog(LOG_INFO,
	       "[%s:%s] clm_process reject active/grace/buff[%d/%d/%d/] \n",
	       current->comm, __FUNCTION__,
	       clm_active, clm_buff_check(), grace_period);

	req->clm_status = DLM_DENIED_GRACE_PERIOD;
	clm_reply(req, NULL);
    }
    else 
    {
	clm_request(req);
    }

}
		
/*
 * Name
 *	clm_request
 * Description 
 *	Process a request packet and invoke the appropriate
 * 	local and remote functions
 *	Send response back to client if the service routine
 *	was local (didn't go across the net to the primary).
 * Arguments
 *	req - the request
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
clm_request(clm_trans_t *req)
{
    int (*callfunc)(clm_trans_t *);
    int type;
    int site;
    clm_client_id_t clid;


    DBPRINTF((DBG_FUNC|DBG_CTI,"[%s:%s] %s type = %s/%d", current->comm,
	      __FUNCTION__, "enter",
	      transaction_type_string( req->clm_type ), req->clm_type ));

    callfunc = NULL;
    type = req->clm_locktype;
    switch ( req->clm_type )
    {
	case DLM_LOCK:
	{
	    if (type > DLM_RES_NTYPES) 
	    {
		req->clm_status = DLM_BADARGS;
	    }
	    else
	    {
		callfunc = lockswitch[type].begin_lock;
		req->clm_callback = lockswitch[type].cont_lock;
		req->clm_lockinfo.site = OurSiteID;
		req->clm_lockinfo.state = 0;
	    }

	    break;
	}

	case DLM_UNLOCK:
	{
	    if (type > DLM_RES_NTYPES) 
	    {
		req->clm_status = DLM_BADARGS;
	    }
	    else
	    {
		callfunc = lockswitch[type].begin_unlock;
		req->clm_callback = lockswitch[type].cont_unlock;
		site = req->clm_lockinfo.site;
		if ( type == DLM_RES_UNIX )
		{
		    req->clm_lockinfo.site = OurSiteID;
		    req->clm_lockinfo.state = 0;
		}
	    }

	    break;
	}

	case DLM_REGISTER:
	{
	    callfunc = begin_register;	
	    req->clm_callback = cont_register;
	    break;
	}

	case DLM_APPREG:
	{
	    /* Use pid for client id */
	    set_client_id(clid, req->clm_pid);
			
	    if ( ! addclient(OurSiteID, &clid, CLM_CLI_PROCESS,
			     req->clm_sequence, NULL ) )
	    {
		/*
		 * Can't add client.  Condition is likely
		 * due to purge in progress.  Wait and try
		 * again after sleep.
		 */
		req->clm_status = DLM_DENIED_GRACE_PERIOD;
	    }
	    else	
		req->clm_status = DLM_NORMAL;

	    break;
	}

	case DLM_PURGELOCKS:
	    DBPRINTF((DBG_CTI,"Purge requested for %d\n",
		      req->clm_purgelocks.purgepid));
	    callfunc = begin_purge;
	    req->clm_callback = NULL;
	    break;

	case DLM_SCN_OP:
	    callfunc = begin_scn_op;
	    req->clm_callback = cont_scn_op;
	    break;

	case DLM_RMIGR_OP:
	    callfunc = begin_tune_op;
	    req->clm_callback = cont_tune_op;
	    break;

	case DLM_GMIGR_OP:
	    callfunc = begin_glob_tune_op;
	    break;

	case DLM_STATS:
	    callfunc = begin_stats_op;
	    req->clm_callback = cont_stats_op;
	    break;

	default:
	{
	    bsdlog(LOG_CRIT, MSGSTR(BAD_REQUEST,
				    "clm_request: bad request type %d"),
		   req->clm_type);
	    req->clm_status = DLM_BADARGS;
	    abort("clm_request");
	    break;
	}
    }


    if ( callfunc != NULL )
    {
	/*
	 * call the begin function for the request.  the individual
	 * functions are responsible for responses to the client for the
	 * purpose handling the different needs that depend on the request type
	 * and validity
	 */

	(*callfunc)(req);
    }
    else
    {
	/*
	 * here after processing a request that needs a reply sent or
	 * if we received an error processing a request
	 */
	clm_reply(req, NULL);
    }

    DBPRINTF((DBG_FUNC|DBG_CTI,"[%s:%s] %s", current->comm, __FUNCTION__,
	      "exit" ));
}

/****************************************************************************/

void
dlm_reply_guts( clm_trans_t * p )
{
    if ( p->clm_respflags & RESP_UDP) 
    {
	DBPRINTF((DBG_CTI, "[%s:%s] RESP_UDP\n", current->comm,
		  __FUNCTION__));
	sendudp( p );
    }
    else if (p->clm_respflags & (RESP_AST | RESP_BAST)) 
    {
	DBPRINTF((DBG_CTI, "[%s:%s] RESP_AST\n", current->comm,
		  __FUNCTION__));

	if (p->clm_type == DLM_LOCK)
	{
	    sendast( p );
	}
	else 
	{
	    send_unlck_ast( p );
	}
    }
    else 
    {
	DBPRINTF((DBG_CTI, "[%s:%s] no response path specified\n",
		  current->comm, __FUNCTION__));
    }
}




/*
 * Name
 *	clm_reply
 * Description
 * 	Send a reply to a client. if the response pointer is NULL, 
 * 	assume this was a local transaction or that an error occurred 
 * 	and  that the request packet status indicates this fact. 
 * 	Free the request packet if provided
 * Arguments
 *	request - the request transaction
 *	response  - the response transaction from the primary
 * Returns
 *	NOTHING
 * Global Data Affected 
 *	NONE
 */
void
clm_reply(clm_trans_t *request, clm_trans_t *response)
{
	clm_trans_t *p;

	DBPRINTF((DBG_FUNC, "enter clm_reply\n"));
	if (response != NULL)
		p = response;
	else 
		p = request;

	/* sanity check */
	if (p == NULL) {
		bsdlog(LOG_ERR, MSGSTR(NULL_RESP,
				"clm_reply: response and request NULL"));
		return;
	}

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_CLM_REPLY,
                p->clm_pid, 0x00, 0x00, 0x00);

	DBPRINTF((DBG_CTI, "[%s:%s] END TRANSACTION %s  status = %d\n",
		  current->comm, __FUNCTION__,
		  transaction_type_string( p->clm_type ), p->clm_status));

	dlm_reply_guts( p );

	/* Do not free request if its LKM_FINDLOCAL flag is set */
	if (request && (request->clm_type != DLM_LOCK ||
			(~request->clm_lockinfo.lk_flags & LKM_FINDLOCAL)))
		clm_free_trans(request);

	DBPRINTF((DBG_FUNC, "leave clm_reply\n"));
}

/*
 * Name
 *	sendudp
 * Description
 *	send a UDP reply back to a client after adding some validation
 *	info to the transaction header
 * Arguments
 *	p - the transaction to send
 * 
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */

void
sendudp(clm_trans_t *oldp)
{
	clm_trans_t *p;
	clm_client_t *cp;
#if defined(REQ_STATS)
	static uint count[] = {0,0,0,0,0,0,0,0};
#endif

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_SENDUDP,
                oldp->clm_pid, 0x00, 0x00, 0x00);

	oldp->clm_respflags &= ~RESP_UDP;

	/* RESP_NULL is used by the purge code in the kernel.  -jjd- */
	if (oldp->clm_respflags & RESP_NULL) {
		DBPRINTF((DBG_CTI, "sendudp: Null response"));
		return;
	}

	cp = findclientbypid(OurSiteID,oldp->clm_pid);
	if (!cp) {
		bsdlog(LOG_ERR, MSGSTR(CLIENT_NOT_FOUND2,
				"sendudp: could not find client %d"),
			oldp->clm_pid);
		return;
	}

	/* For clmlock_sync: if blocking lock request then return, */
	/* sendast will unset LKM_BLOCK flag & run sendudp */
	if ((oldp->clm_type == DLM_LOCK) &&
			(oldp->clm_lockinfo.lk_flags & LKM_BLOCK)) {
	    /*
	     * if the lock manager is coming up, and we are in the grace period,
	     * the udp response must be sent back to the client.  Only return
	     * if the clmlock_sync request can be handled at this point (ie, we
	     * are not in the grace period).
	     */
	    if ( oldp->clm_status != DLM_DENIED_GRACE_PERIOD) {

	        /* Push seq number and thread id onto stack */
	        push_seq(cp, oldp->clm_sequence, 
			 oldp->clm_asth.asth_tid);
		return;
	    }
	}

    if ( oldp->clm_status != DLM_VOID )
    {
	/* If this UDP response is being sent from notify, then it won't have
	 * a sequence number, but the API expects one, so give it one.
	 */

	if ( oldp->clm_sequence == 0 ) {
	    oldp->clm_sequence = ++cp->cli_sequence;
	}


	/* We have to make a copy because the transaction gets munged a bit
	 * between here and where we pass it on to the primary.
	 */

	p = copy_trans( oldp );
	if ( p == NULL ) {
		bsdlog(LOG_ERR, MSGSTR(BUFFER_SHORTAGE,
				"sendudp: could not allocate buffer"));
		return;
	}

	DBPRINTF((DBG_CTI,"Response %d in buffer 0x%x\n", p->clm_sequence, p));
	DBPRINTF((DBG_FUNC, "enter sendudp\n"));


#if defined(TIME_STATS)
	STAMP(p);
	if ((p->clm_respflags & RESP_AST) == 0) {
	    PRINTTIME(p);
	}
#endif

#if defined(REQ_STATS)
	if (p->clm_status != DLM_NORMAL && p->clm_status != DLM_SYNC) {
	    DBPRINTF((DBG_UNIXLOCK, "Bad REQ type=%d id=0x%x stat=%d",
		      p->clm_type, p->clm_lockinfo.lockid,
		      p->clm_status));
	} else {
	    INC_COUNT_TIMESTAMP( "UDP", count[ p->clm_type ] );
	}
#endif

	/* Wake up the client who's waiting on this.  -jjd- */
	p->clm_next = cp->cli_reply;
	cp->cli_reply = p;

	wake_up_interruptible( & cp->cli_reply_event );
	
    }
    else 
    {
	bsdlog(LOG_ERR, MSGSTR(DLM_VOID_INSEND, "DLM_VOID in sendudp\n"));
    }

    DBPRINTF((DBG_FUNC, "leave sendudp\n"));
}

/*
 * Name
 *	notify
 * Description
 * 	Send an asynchronous notification to a client.
 *	Construct a transaction record from a reclock.
 *	If remote client, send message to secondary.
 * 	If local client, send message directly
 *	Used for completion ASTs, blocking ASTs, and generic blocked
 *	completion.
 *	Depends on send_async(), sendast(), and sendudp() to do real work
 * Arguments
 *	a - the transaction that has completed
 *	how - method of notification, if local
 *	status - status of transaction
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
notify( const struct reclock * a, int how, dlm_stats_t status )
{
	clm_trans_t trans;
	clm_trans_t *tptr;
	struct resource *rp;
	dlm_ops_t cmd = DLM_LOCK;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_NOTIFY,
                a->lck.pid, 0x00, 0x00, 0x00);

	DBPRINTF((DBG_FUNC, "enter notify how=%d status=%d\n", how, status));
	
	trans.clm_sequence = 0;
	trans.clm_type = cmd;
	trans.clm_vers = DLM_VERSION;
	trans.clm_route = a->route;
	/* these are slightly incorrect */
	trans.clm_direction = PTI_ASYNC;
	trans.clm_status = status;
	trans.clm_lockinfo = a->lck;
	/*
	 * This condition assumes that all notify call for the CLM lock
	 * model are AST responses and UNIX lock model calls are UDP.  The
	 * realpid field has to be set for the CLM lock model so that the
	 * version compatibility translator doesn't trash BAST notifies.
	 */
	if (a->lck.type == DLM_RES_UNIX) {
		trans.clm_pid = a->lck.pid;
	} else {
		trans.clm_pid = trans.clm_lockinfo.realpid =
				a->route.asth.asth_pid;
	}
	trans.clm_authpid = a->lck.pid;

	if (a->lck.site == OurSiteID) {
		trans.clm_lockinfo.lockid = a->lck.lockid;
	} else {
		/* replace lockid from originating host */
		trans.clm_lockinfo.lockid = a->lck.remlockid;
	}

	if (a->lck.rh.rh == 0) {
		bsdlog(LOG_ERR, MSGSTR(NOTIFY_NULL_HANDLE,
				"notify: NULL resource handle"));
		return;
	}

    if (!(rp = resptr(a->lck.rh)))
    {
	bsdlog(LOG_ERR, MSGSTR(BAD_RESPTR,
			       "notify: bad resptr, rh=0x%lx\n"),
	       a->lck.rh.rh );
	return;
    }

	resname_bcopy(rp->name, &(trans.clm_lockreq.nameinfo.name[0]), rp->namelen);
	trans.clm_lockreq.nameinfo.namelen = rp->namelen;

	/* continue routines depend on this being set correctly */
	if (rp->master == 0) {
		trans.clm_rt_master = OurSiteID;
	} else {
		trans.clm_rt_master = 0;
	}

    trans.clm_respflags = how;
    DBPRINTF((DBG_CTI, "NOTIFY trans %s\n",
	      transaction_type_string( trans.clm_type )));

    if (a->lck.site == OurSiteID && how != RESP_BAST) 
    {
	/* here if request was local */
	dlm_reply_guts( & trans );
    }
    else 
    {
	/* here if request was remote */

	/*
	 * blocking ASTs are recognized on the secondary by 
	 * their NULL callback
	 */
	if (how == RESP_BAST)
	{
	    trans.clm_callback = NULL;
	}

	DBPRINTF((DBG_CTI, "notify: sending to site %d\n", a->lck.site));
	tptr = copy_trans(&trans);
	clm_queue_response(a->lck.site, tptr);
    }

    DBPRINTF((DBG_FUNC, "leave notify\n"));
}

/****************************************************************************/

clm_client_t *
dlm_get_client_from_transaction(
				 clm_trans_t * p,
				 struct reclock * a,
				 struct routeinfo ** route
			       )
{
    int   pid;
    struct routeinfo * ri;


    route = (route == NULL) ? & ri : route;

    /* For some SYNC replies, there won't be a reclock, so we may have to get
     * the route information from the transaction structure instead.
     */

    if ( p->clm_pid != p->clm_lockinfo.pid )
    {
	/* Here if lock owned by a group.  Use clm_pid in trans struct because
	 * lck.pid will have the group id.
	 */

	pid = p->clm_pid;
	*route = &(p->clm_route);
    }
    else if ( a != NULL )
    {
	pid = a->lck.pid;
	*route = a->route.respflags ? &(a->route) : &(p->clm_route);
    }
    else 
    {
	pid = p->clm_lockinfo.pid;
	*route = &(p->clm_route);
    }


    return( findclientbypid( OurSiteID, pid ) );
}

    


/****************************************************************************/
/*
 * Name
 *	sendsync
 * Description
 *	Take the provided AST notification transaction structure, find the
 *	matching structure on the reply queue of the client, copyout the
 *	lockstatus block structure to the client's destination pointer, and
 *	change the return status to DLM_SYNC.
 * Arguments
 *	p - the transaction associated with the AST
 * Returns
 *	TRUE if DLM_SYNC return successful, otherwise FALSE
 * Global Data Affected
 *	NONE
 */
int
sendsync(clm_trans_t *p, struct reclock *a, struct lockstatus *lksb)
{
	clm_trans_t	*tptr;
	clm_client_t		*cl;
	struct routeinfo	*route;


    cl = dlm_get_client_from_transaction( p, a, & route );
	if (NULL == cl) {
	        /* client must have died, so just return */
		bsdlog(LOG_CRIT, "Can't find client rec for DLM_SYNC return");
		return TRUE;
	}

	/* walk the list, find matching trans ptr.  It should be on top */
	tptr = cl->cli_reply;
	while (tptr && tptr->clm_sequence != p->clm_sequence) {
		tptr = tptr->clm_next;
	}

	/* if we can't find the transaction, DLM_SYNC return not possible */
	if (NULL == tptr) {
		bsdlog(LOG_CRIT, "Can't find transaction for DLM_SYNC return");
		return FALSE;
	}

	tptr->clm_status = DLM_SYNC;
	if (copy_to_user(route->asth.asth_dest, lksb, sizeof(*lksb))) {
		bsdlog(LOG_CRIT, "Copyout failed for DLM_SYNC return");
		abort("sendsync");
	}

	return TRUE;
}

/*
 * Name
 *	send_unlck_ast
 * Description
 *	queue the unlock ast specified in the provided transaction
 * Arguments
 *	p - the transaction associated with the AST
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
send_unlck_ast(clm_trans_t *p)
{
  	clm_client_t *cl;       /* Used for clmlock_sync */

	/* find the client record for whoever issued this unlockd request */
	cl = findclientbypid(OurSiteID, p->clm_pid);
		  
	if (NULL == cl) {
		/* client must have died, so just return */
		bsdlog(LOG_CRIT, "Can't find client rec for unlock AST");
		return;
	}

	DBPRINTF((DBG_AST, "Sending unlock AST stat=%d id=0x%x\n",
		p->clm_status, p->clm_lockinfo.lockid));

	if (ASTcall_unlck(cl, p->clm_status, p->clm_asth.asth_funcp,
			p->clm_asth.asth_arg, p->clm_basth.asth_arg) != 0) {
		bsdlog(LOG_ERR, MSGSTR(AST_FAILED,
				"sendast: send failed AST"));
	} else {
                if (DLM_KERNEL_PID == p->clm_pid) {
                        schedule_task(&dlm_keventd_task);
                }else{
                      (void) dlm_kill_user_proc(p->clm_pid, p->clm_signo);
                }

		PRINTTIME(p);
	}

	/* clear AST flag from response bitmask */
	p->clm_respflags &= ~RESP_AST;
}

/*
 * Name
 *	sendast
 * Description
 *	queue the ast specified in the provided transaction
 * Arguments
 *	p - the transaction associated with the AST
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
sendast(clm_trans_t *p)
{
    struct lockstatus lksb;
    struct lockstatus * lock_status;
    struct reclock * a;
    ast_handle_t * ast_handle;
    clm_client_t * cl;       /* Used for clmlock_sync */
    int pid;
    int bast_mode;
#if defined(REQ_STATS)
    struct resource	*rp;
    static uint count = 0;
#endif


    ast_handle = NULL;


    /* Strictly speaking, this is unneeded.  However, without it GCC will
     * complain that these variables 'might be used uninitialized in this
     * function'.
     */

    lock_status = NULL;
    bast_mode = 0;
    pid = 0;


    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_SENDAST,
	    p->clm_pid, 0x00, 0x00, 0x00);

    /* don't send an AST for a WORKING or VOID status */
    if (p->clm_status == DLM_WORKING || p->clm_status == DLM_VOID) 
    {
	DBPRINTF((DBG_CTI,"sendast: clm_status = %d\n",p->clm_status));
	return;
    }

    (void) memset(&lksb, 0, sizeof(lksb));

#if defined(TIME_STATS)
    if ((p->clm_respflags & RESP_AST) && valid_handle(&p->clm_asth)) 
    {
	STAMP(p);
    }
#endif

    /* 
     * since a completion AST must always precede any Blocking ASTs,
     * check for normal AST first.  If clmlock_sync call is being
     * made (LKM_BLOCK flag set) then don't have to validate handle
     */

    a = find_reclock( DLM_RES_VMS, p->clm_lockinfo.lockid );

    if ((p->clm_respflags & RESP_AST) &&
	((valid_handle(&p->clm_asth)) ||
	 (p->clm_lockinfo.lk_flags & LKM_BLOCK)) ) 
    {
	lksb.status = p->clm_status;
	lksb.lockid = p->clm_lockinfo.lockid;


	/* We shouldn't send an AST if there's no local copy of the
	 * lock - unless it's a NOTQUEUED or DEADLOCK in which case
	 * we'll need to use the routeinfo from the transaction struct.
	 */

	if ( (a == NULL) || ! mylock( a, p ) )
	{
	    a = NULL;

	    if ( (p->clm_status != DLM_NOTQUEUED)
		 && (p->clm_status != DLM_DEADLOCK)
		 && (p->clm_status != DLM_DENIED_NOLOCKS) )
	    {
		DBPRINTF((DBG_AST, "sendast: can't find reclock 0x%x",
			  lksb.lockid));
		return;
	    }
	}


#if defined(REQ_STATS)
	if (lksb.status != DLM_NORMAL) 
	{
	    rp = (a != NULL) ? resptr(a->lck.rh) : NULL;
	    DBPRINTF((DBG_UNIXLOCK, "Bad AST stat=%d id=0x%x pid=%d (%s)",
		      lksb.status, lksb.lockid,
		      p->clm_pid,
		      (rp != NULL) ? printbin(rp->name, rp->namelen)
		      : "?" ));
	}
	else 
	{
	    INC_COUNT_TIMESTAMP( "AST", count );
	}
#endif

	(void) memcpy( &lksb.value[0], p->clm_lockinfo.lkvalue,
		      sizeof( lksb.value ) );


	/* For clmlock_sync: If blocking lock request clear bit,
	 * copyout route info, and call sendudp
	 */

	if (p->clm_lockinfo.lk_flags & LKM_BLOCK) 
	{
	    /* unset LKM_BLOCK */
	    p->clm_lockinfo.lk_flags &= ~LKM_BLOCK;


	    DBPRINTF((DBG_SYNC, "sendast: LKM_BLOCK was set 0x%x",
		      lksb.lockid));


	    /* We need to get a pointer to the client structure that
	     * initiated this SYNC request so we can get a sequence
	     * number from it.
	     */

	    cl = dlm_get_client_from_transaction( p, a, NULL );
	    if (NULL == cl) 
	    {
		/* client must have died, so just return */
		bsdlog( LOG_CRIT, "Can't find client rec for LKM_BLOCK" );
	    }
	    else
	    {
		/* CALL SENDUDP */
		p->clm_sequence = pop_seq(cl, p->clm_asth.asth_tid);
		sendudp(p);
	    }
	}
	else
	{
	    /* If getpid() returns the client's pid, then this AST is
	     * being sent from the context of the client's process which
	     * means that it was processed completely on the local machine
	     * - unless the direction is ASYNC meaning that a client did
	     * something to unblock one of its own locks.
	     */

	    if (p->clm_lockinfo.lk_flags & LKM_SYNCSTS &&
		p->clm_direction != PTI_ASYNC &&
		getpid() == p->clm_pid &&
		sendsync(p, a, &lksb)) 
	    {
		return;
	    }


	    /* Get the route and client information.  If there is no reclock,
	     * then use OurSiteID and the PID from the transaction.
	     */

	    if ( a == NULL )
	    {
		ast_handle = & p->clm_route.asth;
		ast_handle->asth_client = findclientbypid( OurSiteID,
							   p->clm_pid );
	    }
	    else
	    {
		ast_handle = & a->route.asth;
		ast_handle->asth_client = findclientbypid( a->lck.site,
							ast_handle->asth_pid );
	    }

	    lock_status = & lksb;
	    bast_mode = 0;
	    pid = p->clm_pid;
	}
    }
    else if ( (p->clm_respflags & RESP_BAST)
	     && valid_handle( & p->clm_basth ) ) 
    {
	/* check if local copy has been nuked - geeeerosssss */
	if( (a == NULL) || (a->lck.state & LOCK_GONE) )
	{
	    DBPRINTF((DBG_AST, "BAST wrong reclock 0x%x",
		      p->clm_lockinfo.lockid));
	}
	else
	{
	    ast_handle = & a->route.basth;
	    ast_handle->asth_client = findclientbypid(a->lck.site,
						      a->route.basth.asth_pid);

	    lock_status = NULL;
	    bast_mode = p->clm_lockinfo.bastmode;
	    pid = ast_handle->asth_pid;
	}
    }
    else if ( (p->clm_respflags & RESP_BAST) && (a == NULL) )
    {
	bsdlog(LOG_ERR, "sendast: Can't find lock 0x%x",
	       p->clm_lockinfo.lockid);

	/*
	 * we hope that this was a BAST for a lock that the
	 * secondary already unlocked and we ignore it - no error
	 */
    }
    else 
    {
	DBPRINTF((DBG_AST, "send_ast: no valid handle   B: p=%d t=%d",
		  p->clm_basth.asth_pid, p->clm_basth.asth_tid));

	DBPRINTF((DBG_AST, "C: p=%d t=%d:",
		  p->clm_asth.asth_pid, p->clm_asth.asth_tid));
    }


    if ( ast_handle != NULL )
    {
	CLM_ASSERT( ast_handle->asth_client != NULL, "No client in sendast" );
	if ( ASTcall( ast_handle, lock_status, bast_mode ) != 0 )
	{
	    bsdlog(LOG_ERR, MSGSTR(BAST_FAILED,
				   "sendast: send failed for BAST"));
	}
	else 
	{
            if (DLM_KERNEL_PID == p->clm_pid) {
                    schedule_task(&dlm_keventd_task);
            } else {
                    (void) dlm_kill_user_proc(a->route.basth.asth_pid, p->clm_signo);
            }

	}
	PRINTTIME(p);
    }
}

/*
 * Function:
 *	k_set_debug
 * Description:
 *	Sets the debug flags of the lock manager to the mask passed in
 *	by the API.
 * Arguments:
 *	debug_flags - debug mask built in API
 * Return value:
 *	0 - unused
 * Side effects:
 *	Will enable or disable debug output on the fly if the code was
 *	compiled with DBPRINT defined.
 */
int
k_set_debug(int debug_flags)
{

	att_seg();

	bsdlog(LOG_DEBUG,"Setting debug flags to %d\n", debug_flags);
	dlm_debug = debug_flags;

	det_seg();

	return 0;
}
