/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_deadlock.c,v 1.13 2001/10/29 21:50:39 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: DDX_CHECK
 *		MSGSTR
 *		TIMEDIFF
 *		cccpx_null
 *		cccpx_recv
 *		cccpx_send
 *		clmdd_analyze
 *		clmdd_check
 *		clmdd_decode
 *		clmdd_dostart
 *		clmdd_encode
 *		clmdd_encodelen
 *		clmdd_incoming
 *		clmdd_init
 *		clmdd_massage
 *		clmdd_primo
 *		clmdd_printstack1864
 *		clmdd_printtwfg
 *		clmdd_processddx1114
 *		clmdd_pushc
 *		clmdd_pushr
 *		clmdd_pushxr
 *		clmdd_startstart
 *		clmdd_suspect
 *		clmdd_task
 *		clmdd_validate
 *		clmdd_walk
 *		ddx_alloc
 *		ddx_expand
 *		ddx_free
 *		dnext
 *		free_iovec
 *		jabber
 *		jabber_recv
 *		clmdd_purgeclnt
 *		tpop2
 *		tpop4
 *		tpush2
 *		tpush4
 */
#ifndef lint
#endif

/*
 *	clm_deadlock.c
 *
 *	  This module implements the distributed deadlock detection
 *	  capability in the lock manager.  The design can be found in
 *	  <somewhere to be designated later.>
 *
 *	Contents
 *	  description
 *	  contents
 *	  to do
 *	  includes
 *	  deadlock stack data structures and random defines
 *	  message data structures and random defines
 *	  static data
 *	  function prototypes
 *	  module entry points
 *	  received message callback
 *	  cluster-wide deadlock cycle coordination
 *	  pushing resources and clients on the deadlock stack
 *	  local deadlock checking
 *	  deadlock stack analysis and pruning
 *	  deadlock stack allocation/grow/free
 *	  shipping and receiving TWFG fragments
 *	  curing deadlock
 *	  debugging utilities
 *	  cluster configuration utilities
 *	  communication module steroids
 *
 *	To do (this list here prior to Linux port...)
 *	  finish adding NLS messages - search for "XXX - NLS ms"
 *	  verify type == DLM_RES_VMS necessary?
 *	  scrub: optimize { A } -> A
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/stddef.h>
#include <linux/param.h>
#include <linux/time.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "clm_data.h"
#include "dlm_kernel.h"
#include "dlm_version.h"
#include <iterator.h>

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/* Gross module configuration. */
#undef	DISABLE_CLMDD		/* disable the module */
#undef	UNIT_TEST		/* some stuff to make unit testing easier */
#define	COMMUNICATION_STEROIDS	/* enable large message fragmentation */

#define LOCKIDTOSITE(a) ((a) >> 28)

#ifdef DEBUG_ALL_PROTO /* hide from prototype file but not debugger */
#define	STATIC
#else
#define STATIC static
#endif

/*
 *	Return code for malloc failure; this is the same as
 *	CPERR_NOMEM in cccp.h.
 */
#define	MEM_ERR		6		/* error return for malloc fail */

/*
 *	Deadlock detection stack size.  The size is relatively small
 *	and the increment relatively large because I suspect the only
 *	"real" time the stack will have to be grown is during a stress
 *	test.
 */
#define	DDX_SIZE	512		/* initial size of the stack */
#define	DDX_INCR	2048		/* amount to grow it each time */

/*
 *	The in-core wait-for graph.  A node in the graph is a
 *	discriminated union; permissible types are given in the enum
 *	ddx_flavor_t.
 *		INVALID		invalid entry, for bombproofing
 *		CLIENT		known client reference
 *		XCLIENT		unknown client reference
 *		RSRC		resource reference
 *		XRSRC		unknown resource reference
 *		BRANCH		beginning of a set
 *		RETURN		end of a set
 *		SCRUBBED	null entry, but not invalid
 */
typedef enum {
	INVALID = 0, CLIENT, XCLIENT, RSRC, XRSRC, BRANCH, RETURN, SCRUBBED
} ddx_flavor_t;

typedef struct ddx_s {
	ddx_flavor_t	ddx_type;
	union {
		struct {
			int		cli_site;
			clm_client_id_t cli_id;
			unsigned long	cli_rid;
			int		cli_marked;
		} un_cli;
		struct {
			int		rsc_namelen;
			char		*rsc_name;
			short		rsc_mode;
			int		rsc_lockid;
			int		rsc_locknode ;
		} un_rsc;
		struct {
			unsigned long	blk_lockid;
			int		blk_site;
			unsigned long	blk_age;
		} un_blk;
	} ddx_un;
} ddx_t;

#define	ddx_site	ddx_un.un_cli.cli_site
#define	ddx_clid	ddx_un.un_cli.cli_id
#define	ddx_rid		ddx_un.un_cli.cli_rid
#define	ddx_marked	ddx_un.un_cli.cli_marked
#define	ddx_namelen	ddx_un.un_rsc.rsc_namelen
#define	ddx_name	ddx_un.un_rsc.rsc_name
#define	ddx_mode	ddx_un.un_rsc.rsc_mode
#define	ddx_lockid	ddx_un.un_rsc.rsc_lockid
#define	ddx_locknode	ddx_un.un_rsc.rsc_locknode
#define	ddx_blkid	ddx_un.un_blk.blk_lockid
#define	ddx_blksite	ddx_un.un_blk.blk_site
#define	ddx_age		ddx_un.un_blk.blk_age

/* BUGBUG A spinlock is need to protect the DDX stack.
 */

ddx_t	*ddx_stack, *ddx_sp, *ddx_end;
int	ddx_count;
int	ddx_rnames_alloced = 0;

#define	DDX_CHECK()							\
  CLM_ASSERT(ddx_stack && ddx_end > ddx_stack && ddx_sp >= ddx_stack	\
  && ddx_sp < ddx_end, "DDX stack is having problems!" )

#define	dnext(label)	{					\
			DDX_CHECK();				\
			if (++ddx_sp == ddx_end)		\
				if (ddx_expand())		\
					goto label;		\
			}

#define	tpush2(cp, s)	\
	cp[0] = ((s) >> 8) & 0xff; cp[1] = (s) & 0xff

#define	tpush4(cp, s)	\
	cp[0] = ((s) >> 24) & 0xff; cp[1] = ((s) >> 16) & 0xff;	\
	cp[2] = ((s) >> 8) & 0xff; cp[3] = (s) & 0xff

typedef struct {
	char		*iov_base;
	unsigned long	iov_off;
	unsigned long	iov_len;
} iovec_t;

/*
 *	Messages
 */
typedef struct {
	unsigned long	sdd_cycle;	/* deadlock cycle # */
	int		sdd_instigator;	/* instigator of this cycle */
	int		sdd_node;	/* node of oldest waiting lock req */
	struct timeval	sdd_tv;		/* age of oldest waiting lock req */
} start_msg_t;
#define	NO_NODE		(-1)
#define	SDD_FORMAT	\
  "%s: STRT cycle=%ld instigator=%d node=%d sec=%ld usec=%ld\n"

/* The purge message */
typedef struct {
	int		prg_site;	/* client's site */
	clm_client_id_t	prg_clid;	/* client's id */
	int		prg_tid;	/* reserved--future thread id */
	unsigned long	prg_rid;	/* victim reclock id */
} purge_msg_t;
#define	PRG_FORMAT	"%s: PURG site=%d clid=0x%x %x tid=%d rid=0x%lx\n"

/* The victim message */
typedef struct {
	int		blw_site;	/* victim site */
	clm_client_id_t	blw_clid;	/* victim client id */
	unsigned long	blw_rid;	/* victim reclock id */
	int		blw_namelen;	/* length of resource name */
	int		blw_nameoff;	/* offset to beginning of name */
					/* resource name follows */
} blow_msg_t;
#define	BLW_FORMAT "%s: BLOW site=%d clid=0x%x %x rid=0x%lx len=%d name='%s'\n"

typedef	char twfg_pgm_t;		/* encoded TWFG program step */

/* Message types */
#define	MSG_STRT	0x53545254	/* request to start cycle */
#define	MSG_TWFG	0x54574647	/* TWFG fragment */
#define	MSG_PURG	0x50555247	/* client request to purge */
#define	MSG_BLOW	0x424c4f57	/* blow away request to master */
#define	MSG_SEGM	0x5345474d	/* large message segment */
#define	MSG_GORD	0x474f5244	/* big junk message, for testing */

/* TWFG program opcodes */
#define	TWFG_SITECLID	1		/* resolved (site, clid) */
#define	TWFG_XCLIENT	2		/* external client reference */
#define	TWFG_XRSRC	3		/* external resource reference */
#define	TWFG_BRANCH	4		/* beginning of a branch */
#define	TWFG_RETURN	5		/* end of fragment */

/* Data expected after each opcode */
#define	TWFG_SITECLID_LEN	16	/* site id, client id, reclock id */
#define	TWFG_SITEPID_LEN	12	/* site id, process id, reclock id */
#define	TWFG_XCLIENT_LEN	12	/* site id, client id */
#define	TWFG_XCLIENT31_LEN	8	/* site id, process id */
#define	TWFG_XRSRC_LEN		12	/* mode, lockid, namelen, locksite */
#define	TWFG_XRSRC41_LEN	8
	/* mode, lockid, namelen PLUS namelen bytes -- not NULL terminated */
#define	TWFG_BRANCH31_LEN	0	/* 3.1 version, nothing in branch */
#define	TWFG_BRANCH_LEN		12	/* lock id, site, age */
#define	TWFG_RETURN_LEN		0

#define TIMEDIFF(A, B)  (((A).tv_sec + ((A).tv_usec > 500000 ? 1 : 0)) \
  - ((B).tv_sec + ((B).tv_usec > 500000 ? 1 : 0)))

/*
 * Deadlock timeout check.  This is set to the number of seconds a lock
 * should stay on the WAIT or CONVERT queue before being checked for
 * deadlock.
 */
STATIC int deadlock_timeout = 30;

/*
 * Number of the cluster-wide deadlock cycle.  This gets incremented
 * on the lowest numbered node in the cluster at any given time.
 */
STATIC unsigned long deadlock_cycle = 0;

/*
 * Paranoia.
 */
STATIC int clmdd_initialized = 0;

int clmdd_init(void);
void clmdd_task(void);
int clmdd_incoming(clm_msg_t *);
STATIC ddx_t *shoot_me_now(ddx_t *ddxs, ddx_t *ddxe);
STATIC purge_msg_t *clmdd_prg_msg(clm_msg_t *);
STATIC blow_msg_t *clmdd_blow_msg(clm_msg_t *);
STATIC int clmdd_startstart(void);
STATIC int clmdd_dostart(start_msg_t *);
STATIC int clmdd_suspect(struct timeval *);
STATIC void clmdd_check(void);
STATIC int clmdd_pushr(struct reclock *);
STATIC int clmdd_pushxr(char *, int, int, int, int);
STATIC int clmdd_pushc(struct reclock *);
STATIC void clmdd_analyze(struct reclock *);
STATIC void clmdd_processddx(void);
STATIC ddx_t *clmdd_walk(ddx_t *, ddx_t *);
STATIC void clmdd_setid(clm_client_id_t *, struct reclock *);
STATIC int clmdd_massage(void);
STATIC int ddx_alloc(void);
STATIC int ddx_expand(void);
STATIC void ddx_free(void);
STATIC void clmdd_decode(iovec_t *, int, int);
STATIC twfg_pgm_t *clmdd_encode(unsigned long *, int);
STATIC unsigned long clmdd_encodelen(int);
STATIC void clmdd_purgeclnt(int, unsigned long);
STATIC int clmdd_purgersrc(int, clm_client_id_t *, unsigned long, int, char *);
#ifdef	DEBUG
STATIC int clmdd_validate(void);
STATIC void clmdd_printstack(void);
STATIC void clmdd_printtwfg(iovec_t *, int, int);
STATIC int on_convert_q(struct resource *, struct reclock *);
#endif /* DEBUG */

STATIC bool_t clmdd_primo(int);

#ifdef	COMMUNICATION_STEROIDS
STATIC void cccpx_null(clm_msg_t *, int);
STATIC int cccpx_send(clm_msg_t *);
STATIC void cccpx_recv(clm_msg_t *, void (*)(iovec_t *, int, int));
STATIC void free_iovec(iovec_t *, int);
#ifdef	UNIT_TEST
STATIC int jabber(int);
STATIC void jabber_recv(iovec_t *, int);
#endif /* UNIT_TEST */
#else
#define	cccpx_send	cccp_send
#endif /* COMMUNICATION_STEROIDS */

/* ============================================================ */
/*
 *	Externally visible functions.
 */

/*
 * Name
 *	clmdd_init
 * Description
 *	Initializes the deadlock detection module.  Assumes that the
 *	communications module has been initialized (or can at least
 *	accept a call to cccp_listen).
 * Arguments
 *	None.
 * Returns
 *	0	Success
 *	!0	Failure
 * Global Data Affected
 * 	clmdd_initialized	set to 1 if clmdd_init succeeded
 */
int
clmdd_init(void)
{
	int rv;

#ifndef	DISABLE_CLMDD
	DBPRINTF((DBG_DEADLOCK, "clmdd_init()\n"));
	rv = cccp_listen((unsigned char) PORT_DEADLOCK, clmdd_incoming);
	if (rv == CPERR_OK)
		clmdd_initialized = 1;
	return (rv);
#else
	DBPRINTF((DBG_DEADLOCK,
	  "Deadlock detection (intentionally) disabled\n"));
	clmdd_initialized = 0;
	return (CPERR_OK);
#endif
}

/*
 * Name
 *	clmdd_task
 * Description
 *	Called periodically from the lock manager's main loop.  On the
 *	first node in a cluster, kicks off a ddx cycle; other nodes
 *	are driven by receiving a STRT message.
 * Arguments
 *	None.
 * Returns
 *	None.
 * Global Data Affected
 * 	None.
 */
void
clmdd_task(void)
{
	static int then = 0;
	struct timeval tv;

	if (!clmdd_initialized)
		return;
	/*
	 *	If the lock manager is quiescent, we can be called
	 *	as often as once a second.  This makes sure we don't
	 *	start a cycle that often.
	 */
	(void) gettimeofday(&tv, NULL);
	if ((tv.tv_sec - then) < (deadlock_timeout / 2))
		return;
	then = tv.tv_sec;

	if (clmdd_primo(OurSiteID))
		(void) clmdd_startstart();
}

/*
 * Name
 *	clmdd_incoming
 * Description
 *	Callback function for processing incoming messages from other
 *	nodes' deadlock detection modules.
 * Arguments
 *	cm - the incoming message
 * Returns
 *	1 - please forward STRT or BLOW message to the selected node.
 *	0 - otherwise.
 * Global Data Affected
 * 	None.
 */
int
clmdd_incoming(clm_msg_t *cm)
{
	int node;
	start_msg_t *sdd;
	purge_msg_t *prg;
	blow_msg_t *blw;
	iovec_t iov;

	CLM_ASSERT( cm != NULL, "NULL message received" );
	if (!clmdd_initialized)
		return (0);
	switch (cm->cm_type) {
	case MSG_STRT:			/* request to start cycle */
		sdd = (start_msg_t *) cm->cm_data;
		CLM_ASSERT( cm->cm_length >= sizeof (start_msg_t),
			    "Bad start message size" );
		DBPRINTF((DBG_DEADLOCK, SDD_FORMAT,
		  "> clmdd_incoming", sdd->sdd_cycle, sdd->sdd_instigator,
		  sdd->sdd_node, sdd->sdd_tv.tv_sec, sdd->sdd_tv.tv_usec));
		if (sdd->sdd_instigator == OurSiteID) {
			if (sdd->sdd_node == NO_NODE)
				;	/* end of cycle */
			else if (sdd->sdd_node == OurSiteID)
				clmdd_check();
			else {
				cm->cm_node = sdd->sdd_node;
				return (1);
			}
		} else {
			if (sdd->sdd_node == OurSiteID)
				clmdd_check();
			else
				clmdd_dostart((start_msg_t *) cm->cm_data);
		}
		break;
	case MSG_TWFG:			/* TWFG fragment */
		DBPRINTF((DBG_DEADLOCK, "> clmdd_incoming: TWFG length = %d\n",
		  cm->cm_length));
		iov.iov_base = cm->cm_data;
		iov.iov_off = 0;
		iov.iov_len = cm->cm_length;
		clmdd_decode(&iov, 1, clm_version(cm->cm_node));
		break;
	case MSG_PURG:			/* victim to purge */
		prg = clmdd_prg_msg(cm); /* translate if necessary */

		DBPRINTF((DBG_DEADLOCK, PRG_FORMAT, "> clmdd_incoming",
		  prg->prg_site, db_clid(prg->prg_clid), prg->prg_tid, 
		  prg->prg_rid));

		clmdd_purgeclnt(prg->prg_site, prg->prg_rid);
		break;
	case MSG_BLOW:			/* master blow-away request */
		blw = clmdd_blow_msg(cm); /* translate if necessary */

		DBPRINTF((DBG_DEADLOCK, BLW_FORMAT, "> clmdd_incoming",
		  blw->blw_site, db_clid(blw->blw_clid), blw->blw_rid, 
		  blw->blw_namelen, (char *) cm->cm_data + blw->blw_nameoff));

		node = clmdd_purgersrc(blw->blw_site, &blw->blw_clid,
		  blw->blw_rid, blw->blw_namelen,
		  (char *) cm->cm_data + blw->blw_nameoff);

		/* check for redirect */
		if (node) {
			cm->cm_node = node;
			return (1);
		}
		break;
#ifdef	COMMUNICATION_STEROIDS
	case MSG_SEGM:
		cccpx_recv(cm, clmdd_decode);
		break;
#endif
	default:
		DBPRINTF((DBG_DEADLOCK, "> clmdd_incoming: '%.4s' msg?\n",
		  (char *) &cm->cm_type));
		break;
	}
	return (0);
}

/*
 *	Inter-node coordination.
 *
 *	This stuff is necessary because if you have a deadlock
 *	condition it is possible that it will block several clients'
 *	requests at about the same time.  If we kicked off a periodic
 *	ddx cycle on all nodes without some kind of coordination, this
 *	would start deadlock detection cycles on all of the affected
 *	nodes at roughly the same time.
 *
 *	This code is used to synchronize the beginning of a cycle.
 *	The cycle begins on the lowest numbered site, triggered by the
 *	periodic task clmdd_task.  The first node checks its deadlock
 *	detection queue for suspicious requests.  If it finds one it
 *	sets the start message timestamp to the age of the oldest
 *	request, otherwise it sets sdd->sdd_node to NO_NODE.  The
 *	start message is then sent to the next node in the cluster.
 *	The receiving node checks its queue, and if its oldest request
 *	is older than the one in the start message it sets the
 *	timestamp and node ID to its own values.  If a node receives a
 *	start message with its own node ID in the message, it begins a
 *	ddx cycle.
 *
 *	Note that we need to use the *age* of requests, rather than
 *	the timestamp, due to the implementation of gettimeofday() in
 *	kernel.c.
 */

/*
 * Name
 *	clmdd_startstart
 * Description
 *	Kicks off a clusterwide ddx cycle.  The caller determines
 *	whether or not it is appropriate to call this function.
 * Arguments
 *	None.
 * Returns
 *	0	Success
 *	!0	Failure
 * Global Data Affected
 * 	deadlock_cycle
 * Assumptions
 *	Assumes that cccp_send can be called to send a message
 *	to ourself; otherwise code will have to be added to
 *	handle a one-node cluster.
 *
 *	Assumes that the communications module will free both
 *	the message and the data if cccp_send fails.
 */
STATIC int
clmdd_startstart(void)
{
	clm_msg_t *cm;
	start_msg_t *sdd;
	struct timeval tv;
	int rv;

	if ((sdd = (start_msg_t *) CLM_MALLOC(sizeof (start_msg_t))) == NULL)
	{
		bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC,
					"Malloc error in %s()"), __FUNCTION__);
		abort("Malloc error in clmdd_startstart()");
	}
	sdd->sdd_cycle = ++deadlock_cycle;
	sdd->sdd_instigator = OurSiteID;
	if (clmdd_suspect(&tv)) {
		sdd->sdd_node = OurSiteID;
		sdd->sdd_tv.tv_sec = tv.tv_sec;
		sdd->sdd_tv.tv_usec = tv.tv_usec;
	} else {
		sdd->sdd_node = NO_NODE;
#ifdef	DEBUG
		sdd->sdd_tv.tv_sec = 12345;
		sdd->sdd_tv.tv_usec = 67890;
#endif
	}

    cm = cccp_msg_create( PORT_DEADLOCK, 0, dlm_nodeid_next( OurSiteID ),
			  MSG_STRT, sizeof( start_msg_t ), NULL, sdd, NULL );
    if ( cm == NULL )
    {
	bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC,
				"Malloc error in %s()"), __FUNCTION__);
	abort("Malloc error in clmdd_startstart()");
    }

    DBPRINTF((DBG_DEADLOCK, SDD_FORMAT,
	      "< clmdd_startstart", sdd->sdd_cycle, sdd->sdd_instigator,
	      sdd->sdd_node, sdd->sdd_tv.tv_sec, sdd->sdd_tv.tv_usec));
    if ((rv = cccp_send(cm)) != CPERR_OK) {
	DBPRINTF((DBG_DEADLOCK,
		  "clmdd_startstart: can't send STRT message\n"));
	return (rv);
    }

    return (0);
}

/*
 * Name
 *	clmdd_dostart
 * Description
 *	Process a start packet.
 * Arguments
 *	sdd - the received start packet.
 * Returns
 *	0	Success
 *	!0	Failure
 * Global Data Affected
 * 	deadlock_cycle
 * Assumptions
 *	Assumes that cccp_send can be called to send a message
 *	to ourself; otherwise code will have to be added to
 *	handle a one-node cluster.
 *
 *	Assumes that the communications module will free both
 *	the message and the data if cccp_send fails.
 */
STATIC int
clmdd_dostart(start_msg_t *sdd)
{
	struct timeval tv;
	clm_msg_t *cm;
	start_msg_t *nsdd;
	int rv;

	CLM_ASSERT( sdd != NULL, "NULL message received" );
	if ((nsdd = (start_msg_t *) CLM_MALLOC(sizeof (start_msg_t))) == NULL) {
		bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC1,
		  "Malloc error in clmdd_dostart()"));
		abort("Malloc error in clmdd_dostart()");
	}
	deadlock_cycle = nsdd->sdd_cycle = sdd->sdd_cycle;
	nsdd->sdd_instigator = sdd->sdd_instigator;
	if (clmdd_suspect(&tv) &&
	    (sdd->sdd_node == NO_NODE
	     || (tv.tv_sec > sdd->sdd_tv.tv_sec
		 || (tv.tv_sec == sdd->sdd_tv.tv_sec
		     && tv.tv_usec > sdd->sdd_tv.tv_usec)))) {
		/*
		 * 	We have a suspicious lock request that's
		 *	been waiting longer.
		 */
		DBPRINTF((DBG_DEADLOCK,
		  "clmdd_dostart: we have an older suspect\n"));
		nsdd->sdd_node = OurSiteID;
		nsdd->sdd_tv.tv_sec = tv.tv_sec;
		nsdd->sdd_tv.tv_usec = tv.tv_usec;
	} else {
		/*
		 *	Forward message.  Possible refinement: use
		 *	the communications module forwarding feature
		 *	to do this.
		 */
		DBPRINTF((DBG_DEADLOCK,
		  "clmdd_dostart: we don't have an older suspect\n"));
		nsdd->sdd_node = sdd->sdd_node;
		nsdd->sdd_tv.tv_sec = sdd->sdd_tv.tv_sec;
		nsdd->sdd_tv.tv_usec = sdd->sdd_tv.tv_usec;
	}

    cm = cccp_msg_create( PORT_DEADLOCK, 0, dlm_nodeid_next( OurSiteID ),
			  MSG_STRT, sizeof( start_msg_t ), NULL, nsdd, NULL );
    if ( cm == NULL )
    {
	bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC1,
				"Malloc error in clmdd_dostart()"));
	abort("malloc error in clmdd_dostart()");
    }

    DBPRINTF((DBG_DEADLOCK, SDD_FORMAT,
	      "< clmdd_dostart", nsdd->sdd_cycle, nsdd->sdd_instigator,
	      nsdd->sdd_node, nsdd->sdd_tv.tv_sec, nsdd->sdd_tv.tv_usec));
    if ((rv = cccp_send(cm)) != CPERR_OK) {
	DBPRINTF((DBG_DEADLOCK,
		  "clmdd_dostart: can't send STRT message\n"));
	return (rv);
    }

    return (0);
}

/* ============================================================ */
/*
 *	Version compatibility translators
 *
 * See earlier note on version information above where the
 * data structures were defined.  Keeping this code to
 * essentially remind us to worry about future migration
 * issues and to act as examples of how to deal with it.
 *
 * I leave it up to the reader to determine if these are
 * "good" examples!
 */

/*
 * Name
 *	clmdd_prg_msg
 * Description
 *	Take the message passed in and return a pointer to the 3.1.1
 *	version of the purge structure.  If the message came from a
 *	backwater node, then its pid needs to be changed to a client id.
 * Arguments
 *	msg - pointer to the message to translate
 * Returns
 *	pointer to the translated structure
 * Global Data Affected
 * 	None.
 */
STATIC purge_msg_t *
clmdd_prg_msg(clm_msg_t *msg)
{
    return (purge_msg_t *)msg->cm_data;
}

/*
 * Name
 *	clmdd_blow_msg
 * Description
 *	Take the message passed in and return a pointer to the 3.1.1
 *	version of the purge structure.  If the message came from a
 *	backwater node, then its pid needs to be changed to a client id.
 * Arguments
 *	msg - pointer to the message to translate
 * Returns
 *	pointer to the translated structure
 * Global Data Affected
 * 	None.
 */
STATIC blow_msg_t *
clmdd_blow_msg(clm_msg_t *msg)
{
    return (blow_msg_t *)msg->cm_data;
}

/* ============================================================ */

/*
 *	Local deadlock checking.
 */

/*
 * Name
 *	clmdd_suspect
 * Description
 *	Search the deadlock timeout queue for the oldest suspicious
 *	lock request.  ("Suspicious" == waiting for longer than
 *	deadlock_timeout seconds).  Fills in the tv argument with
 *	the age of the oldest request encountered.
 * Arguments
 *	tv - place to stick the age of the oldest request
 * Returns
 *	0	Nothing suspicious was found
 *	>0	Something suspicious was found
 * Global Data Affected
 * 	None.
 */
STATIC int
clmdd_suspect(struct timeval *tv)
{
    dlm_listnode_t * node;
    struct reclock * curr;
	struct timeval now, thresh;
	int rv = 0;

	CLM_ASSERT( tv != NULL, "NULL timeval" );
    
    node = dlm_list_get_head( & dl_timeout );
    if ( node == NULL )
	return (rv);


    /* Calculate the threshold.  Timestamps at or before the threshold time
     * are deemed suspicious.
     */

    (void) gettimeofday(&now, NULL);
    CLM_ASSERT( now.tv_sec >= deadlock_timeout, "Timeout not yet expired!" );
    thresh.tv_sec = now.tv_sec - deadlock_timeout;
    thresh.tv_usec = now.tv_usec;


    /* Find the first local suspicious request, if present.
     */

    while ( node != NULL )
    {
	curr = DEADLOCK_LINK_TO_RECLOCK( node );
	node = dlm_list_succ( node );

	if ((TIMEDIFF(thresh, curr->dl.checkstamp) >= 0 ) &&
	    (curr->lck.site == OurSiteID)) 
	{
	    rv++;
	    tv->tv_sec = curr->dl.checkstamp.tv_sec;
	    tv->tv_usec = curr->dl.checkstamp.tv_usec;
	    break;
	}
    }


    if (rv) 
    {
	/* Now find the oldest local request.
	 */

	while ( node != NULL )
	{
	    curr = DEADLOCK_LINK_TO_RECLOCK( node );

	    if ((curr->dl.checkstamp.tv_sec < tv->tv_sec) &&
		(curr->lck.site == OurSiteID)) 
	    {
		tv->tv_sec = curr->dl.checkstamp.tv_sec;
		tv->tv_usec = curr->dl.checkstamp.tv_usec;
	    }

	    node = dlm_list_succ( node );
	}

	/* Make tv the age of the request rather than its time.
	 */

	timeval_sub(&now, tv, tv);
    }


    CLM_ASSERT( node == NULL, "Timeout queue not totally processed" );
    CLM_ASSERT( rv >= 0, "Impossible value!" );

    return (rv);
}

/*
 * Name
 *	clmdd_check
 * Description
 *	Checks each candidate for deadlock detection on the local
 *	machine in turn.
 * Arguments
 *	None.
 * Returns
 *	Nothing.
 * Global Data Affected
 *
 */
STATIC void
clmdd_check(void)
{
    dlm_listnode_t * node;
    struct reclock * curr;
    struct timeval thresh;	/* threshold time */
    struct timeval now;	/* current time */


    /* Calculate the threshold.  Timestamps at or before the threshold time
     * are deemed suspicious.
     */

    (void) gettimeofday(&thresh, NULL);
    now = thresh;

    CLM_ASSERT( thresh.tv_sec >= deadlock_timeout, 
	        "Timeout not yet expired!" );

    thresh.tv_sec -= deadlock_timeout;

    for ( EACH_DLM_LISTNODE( & dl_timeout, node ) )
    {
	curr = DEADLOCK_LINK_TO_RECLOCK( node );

	/* run deadlock check on all local blocked reqs */
	if ((TIMEDIFF(thresh, curr->dl.checkstamp) >= 0) &&
	    (curr->lck.site == OurSiteID))
	{
	    clmdd_analyze(curr);
	    if ((curr->lck.state & LOCK_WAITING) ||
		(curr->lck.state & LOCK_CONVERTING)) 
	    {
		curr->dl.checkstamp = now;
	    }
	    else 
	    {
		delete_timeout_reclock(curr);
	    }
	}
    }

    CLM_ASSERT( node == NULL, "Timeout queue not totally processed" );
}

/*
 * Name
 *	clmdd_pushr
 * Description
 *	Pushes a resource corresponding to a lock request onto the
 *	deadlock detection stack.  A resource will eventually expand
 *	into a list of clients holding locks on that resource (i.e.,
 *	in the wait/convert queues) that are incompatible with the
 *	lock request passed in as an argument.  If the resource is
 *	mastered locally, the expansion is done here.  If the resource
 *	is remote, an XRSRC entry is pushed onto the deadlock detection
 *	stack.  In case of an error the deadlock stack is freed.
 * Arguments
 *	lock - the lock request
 * Returns
 *	-1	an error
 *	0	success, resource is mastered locally
 *	>0	success, resource is mastered on the returned node ID
 * Global Data Affected
 *	ddx_stack, ddx_sp
 */
STATIC int
clmdd_pushr(struct reclock *lock)
{
	struct timeval now;	/* current time */
	struct resource *rp;
	struct reclock *r;
	int rv;

	CLM_ASSERT( lock != NULL, "Invalid input lock" );
	DBPRINTF((DBG_DEADLOCK, "entering clmdd_pushr:"));
	clm_printlock(lock);

	/* If a local cycle exists, just return.  It will be processed in
	 * clmdd_processddx().
	 */
	if (clmdd_walk(ddx_stack, ddx_end) != (ddx_t *) 0)
		return (0);

	/* Backtrack to the resource we're waiting on.
	 */
	rp = (lock->lck.rh.rh) ? resptr(lock->lck.rh) : NULL;
	if (rp == NULL) {
		/* XXX - NLS me */
		bsdlog(LOG_ERR, "clmdd_pushr: can't get resource handle");
		goto blech;
	}

	rv = rp->master ? rp->master : OurSiteID;

	if (rp->master == 0) {
		/* Resource is mastered locally.  Push its holders onto the
		 * stack.
		 */
		DDX_CHECK();
		ddx_sp->ddx_type = BRANCH;
		/*
		 * BRANCH must contain lockid, site and age of blocked lock
		 */
		ddx_sp->ddx_blkid = lock->lck.site == OurSiteID ?
				lock->lck.lockid : lock->lck.remlockid;
		ddx_sp->ddx_blksite = lock->lck.site;
		(void) gettimeofday(&now, NULL);
		ddx_sp->ddx_age = TIMEDIFF(now, lock->dl.timestamp);
		dnext(blech);

	    for ( EACH_DLM_LISTNODE( & rp->lockqs[ RLQ_GRANT ], r ) )
	    {
		if (!(r->lck.lk_flags & LKM_NODLCKWT)
		    && !clmcompatible(lock->lck.reqmode, r->lck.mode)
		    && clmdd_pushc(r))
		    goto blech;
	    }

	    r = (struct reclock *) dlm_list_get_head( & rp->lockqs[RLQ_CONV] );
	    if (lock->lck.state & LOCK_CONVERTING) 
	    {
		/* Lock is converting -- push all of the requests ahead of us
		 * onto the stack.
		 */

		CLM_ASSERT( on_convert_q(rp, lock), 
			    "Lock marked converting, but not on convert "
			    "queue" );

		while (r && r != lock) 
		{
		    if (!(r->lck.lk_flags & LKM_NODLCKWT)
			&& clmdd_pushc(r))
			goto blech;

		    r = (struct reclock *) dlm_list_succ( (dlm_listnode_t *) 
							  r );
		}

		/* skip ourselves */
		r = (struct reclock *) dlm_list_succ( (dlm_listnode_t *) r );
	    }

	    while ( r != NULL )
	    {
		if (!(r->lck.lk_flags & LKM_NODLCKWT)
		    && !clmcompatible(lock->lck.reqmode, r->lck.mode)
		    && clmdd_pushc(r))
		    goto blech;

		r = (struct reclock *) dlm_list_succ( (dlm_listnode_t *) r );
	    }

		DDX_CHECK();
		/*
		 *	If there was no expansion since the last
		 *	BRANCH, pop it off.
		 */

		/* so ddx_sp[-1] is valid */
		CLM_ASSERT( ddx_sp > ddx_stack, "Deadlock stack insane!" );

		if ((ddx_sp - 1)->ddx_type == BRANCH) {
			ddx_sp--;
			ddx_sp->ddx_type = INVALID;
		} else {
			ddx_sp->ddx_type = RETURN;
			dnext(blech);
		}
	} else {
		/*
		 *	Resource is mastered elsewhere.  Push an
		 *	external resource reference onto the stack.
		 */
		DDX_CHECK();
		ddx_sp->ddx_namelen = rp->namelen;
		ddx_sp->ddx_name = CLM_MALLOC(rp->namelen + 1);
		if (!ddx_sp->ddx_name) {
			bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC2,
			  "Malloc error in clmdd_pushr()"));
			abort("Malloc error in clmdd_pushr()");
		}
		ddx_rnames_alloced++;
		resname_bcopy(rp->name, ddx_sp->ddx_name, rp->namelen);
		*(ddx_sp->ddx_name + ddx_sp->ddx_namelen) = '\0';
		ddx_sp->ddx_mode = lock->lck.reqmode;
		ddx_sp->ddx_lockid = lock->lck.lockid;
		ddx_sp->ddx_locknode = OurSiteID ;
		ddx_sp->ddx_type = XRSRC;
		dnext(blech);
	}
	return (rv == OurSiteID ? 0 : rv);
blech:
	ddx_free();
	return (-1);
}

/*
 * Name
 *	clmdd_pushxr
 * Description
 *	Push an external resource reference (i.e., one received in a
 *	partial TWFG message) onto the stack.  A resource reference
 *	will eventually be expanded into a list of clients as
 *	described above in clmdd_pushr().  If the resource is mastered
 *	locally expand it.  Otherwise push an XRSRC reference onto the
 *	stack.
 * Arguments
 *	name - the resource name
 *	namelen - length of name
 *	mode - mode of the incompatible request
 * Returns
 *	-1	an error
 *	0	success, resource is mastered locally
 *	>0	success, resource is mastered on the returned node ID
 * Global Data Affected
 *	ddx_stack, ddx_sp
 */
STATIC int
clmdd_pushxr(char *name, int namelen, int mode, int remlockid, int locknode)
{
	struct resource *rp;
	struct reclock *r, *a = NULL;
	uint id, node;
	struct timeval now;	/* current time */

	CLM_ASSERT( (name != NULL) && (namelen != 0), 
		    "Invalid resource name" );

	DBPRINTF((DBG_DEADLOCK, "entering clmdd_pushxr: name=%s mode=%s "
		  "remlockid=0x%x", printbin(name, namelen), 
		  clm_printmode(mode), remlockid));

	/* lookup resource */
	rp = find_resource( name, namelen, DLM_RES_VMS );

	/* if mastered here, expand branch */
	if (rp && rp->master == 0) {
		/*
		 *	Resource is mastered locally.  Push its
		 *	holders onto the stack.
		 */
		DDX_CHECK();
		ddx_sp->ddx_type = BRANCH;
		/*
		 * BRANCH must contain lockid, site and age of blocked lock
		 */
		ddx_sp->ddx_blkid = remlockid;
		if (remlockid && (id = mapid(remlockid, locknode)) && 
				(a = find_reclock(DLM_RES_VMS, id))) {
			(void) gettimeofday(&now, NULL);
			ddx_sp->ddx_age = TIMEDIFF(now, a->dl.timestamp);
			ddx_sp->ddx_blksite = a->lck.site;
		}
		dnext(blech);

	    for ( EACH_DLM_LISTNODE( & rp->lockqs[ RLQ_GRANT ], r ) )
	    {
		if (!(r->lck.lk_flags & LKM_NODLCKWT)
		    && !clmcompatible(mode, r->lck.mode)
		    && clmdd_pushc(r))
		    goto blech;
	    }

	    r = (struct reclock *) dlm_list_get_head( & rp->lockqs[RLQ_CONV] );
	    if (a && (a->lck.state & LOCK_CONVERTING)) 
	    {
		/* Lock is converting -- push all of the lock requests before
		 * us on the queue onto the stack.
		 */

		while (r && r->lck.remlockid != remlockid) 
		{
		    if (!(r->lck.lk_flags & LKM_NODLCKWT)
			&& clmdd_pushc(r))
			goto blech;

		    r = (struct reclock *) dlm_list_succ( (dlm_listnode_t *) 
							  r );
		}
		CLM_ASSERT( r != NULL, "Remote lock not queued!" );

		/* skip ourselves */
		r = (struct reclock *) dlm_list_succ( (dlm_listnode_t *) r );
	    }

	    while ( r != NULL )
	    {
		if (!(r->lck.lk_flags & LKM_NODLCKWT)
		    && !clmcompatible(mode, r->lck.mode)
		    && clmdd_pushc(r))
		{
		    goto blech;
		}

		r = (struct reclock *) dlm_list_succ( (dlm_listnode_t *) r );
	    }

		DDX_CHECK();
		/*
		 *	If there was no expansion since the last
		 *	BRANCH, pop it off.
		 */
		CLM_ASSERT( ddx_sp > ddx_stack, "Deadlock stack insane!" );
		if ((ddx_sp - 1)->ddx_type == BRANCH) {
			ddx_sp--;
			ddx_sp->ddx_type = INVALID;
		} else {
			ddx_sp->ddx_type = RETURN;
			dnext(blech);
		}
		return (0);
	} else {
		if (rp) {
			node = rp->master;
		} else {
			node = 0;
		}
		DDX_CHECK();
		ddx_sp->ddx_namelen = namelen;
		ddx_sp->ddx_name = CLM_MALLOC(namelen + 1);
		if (!ddx_sp->ddx_name) {
			/* XXX - NLS me */
			bsdlog(LOG_CRIT, "Malloc error in clmdd_pushxr()");
			abort("Malloc error in clmdd_pushxr()");
		}
		ddx_rnames_alloced++;
		resname_bcopy(name, ddx_sp->ddx_name, namelen);
		*(ddx_sp->ddx_name + ddx_sp->ddx_namelen) = '\0';
		ddx_sp->ddx_mode = mode;
		ddx_sp->ddx_lockid = remlockid;
		ddx_sp->ddx_type = XRSRC;
		dnext(blech);
		return (node);
	}
blech:
	ddx_free();
	return (-1);
}

/*
 * Name
 *	been_there
 * Description
 *	Returns non-zero if lockid is in a client entry before it on the
 *	stack.  This function has been added to replace the deadlock stamp
 *	field that used to get set on a lock when it was traversed during
 *	the build of the deadlock graph.  It caused problems when two
 *	different searches were started from different blocked requests in
 *	the same graph.
 * Arguments
 *	r - the client reclock
 * Returns
 *	0	lockid not found
 *	!0	lockid found
 * Global Data Affected
 *	None
 */
STATIC int
been_there(struct reclock *r)
{
	ddx_t *ddx;

	for (ddx = &ddx_stack[1]; ddx < ddx_end; ddx++) {
		if (ddx->ddx_type == CLIENT &&
				ddx->ddx_rid == r->lck.lockid)
			return TRUE;
	}

	return FALSE;
}

/*
 * Name
 *	clmdd_pushc
 * Description
 *	Pushes a client onto the deadlock detection stack.
 *	Eventually, a client expands into a list of all its blocked
 *	lock requests.  If the client is local, marks and expands it.
 *	In case of an error, frees the deadlock stack.
 * Arguments
 *	r - the client reclock
 * Returns
 *	0	success
 *	!0	failure
 * Global Data Affected
 *	ddx_stack, ddx_sp
 */
STATIC int
clmdd_pushc(struct reclock *r)
{
	clm_client_t *cl;
	clm_client_id_t clid;  /* local copy of client id to pass 
				  to findclient */
    dlm_listnode_t * node;

	CLM_ASSERT( r != NULL, "Invalid input lock" );
	DBPRINTF((DBG_DEADLOCK, "entering clmdd_pushc:"));
	clm_printlock(r);
	/*
	 * We used to check the deadlock stamp here.  If the stamp was
	 * already set to the current deadlock_cycle value, then we assumed
	 * that this branch had already been traversed, so we didn't search
	 * down it.  The problem was that a different tree fragment might
	 * have touched the lock and the branch has not yet been traversed
	 * in our tree.  So instead, we look backwards to see if the branch
	 * is already in our tree.  We may get some repeat branches this
	 * way, but the alternatives are either very messy or they caused
	 * deadlocks to be missed.
	 */
	if (been_there(r)) {
		DBPRINTF((DBG_DEADLOCK,
			"clmdd_pushc: Already seen 0x%x - skip branch",
			r->lck.lockid));
		return (0);
	}
	r->dl.deadlock_stamp = deadlock_cycle;

	/*
	 *	If a local cycle exists, just return.  It
	 *	will be processed in clmdd_processddx().
	 */
	if (clmdd_walk(ddx_stack, ddx_end) != (ddx_t *) 0)
		return (0);

	if (r->lck.site == OurSiteID) {
		/*
		 *	The client list is complete here.  Push all
		 *	its blocked lock requests onto the
		 *	stack.
		 */
		DDX_CHECK();
		ddx_sp->ddx_type = CLIENT;
		ddx_sp->ddx_site = r->lck.site;
		ddx_sp->ddx_rid = r->lck.lockid;
		/* keep copy of client id */
		clmdd_setid(&clid, r);
		memcpy(&(ddx_sp->ddx_clid), &clid, sizeof(clm_client_id_t));
		ddx_sp->ddx_marked = 1;
		dnext(blech);
		ddx_sp->ddx_type = BRANCH;
		/*
		 * This is a BRANCH from a holder.  Contains NO info.
		 */
		ddx_sp->ddx_blkid = 0;
		dnext(blech);

	    cl = findclient(r->lck.site, &clid);
	    if ( cl != NULL )
	    {
		struct reclock *rw;


		for ( EACH_DLM_LISTNODE( & cl->cli_queue, node ) )
		{
		    rw = CLIENT_LINK_TO_RECLOCK( node );

		    if (rw->lck.state & (LOCK_WAITING | LOCK_CONVERTING)
			&& !(rw->lck.lk_flags & LKM_NODLCKWT))
			if (clmdd_pushr(rw) < 0)
			    goto blech;
		}
	    }


		DDX_CHECK();
		/*
		 *	If there was no expansion since the last
		 *	BRANCH, pop it off.
		 */
		CLM_ASSERT( ddx_sp > ddx_stack, "Deadlock stack insane!" );
		if ((ddx_sp - 1)->ddx_type == BRANCH) {
			ddx_sp--;
			ddx_sp->ddx_type = INVALID;
		} else {
			ddx_sp->ddx_type = RETURN;
			dnext(blech);
		}
	} else {
		/*
		 *	The client is located elsewhere.
		 */
		DDX_CHECK();
		ddx_sp->ddx_type = XCLIENT;
		ddx_sp->ddx_site = r->lck.site;
		clmdd_setid(&ddx_sp->ddx_clid, r);
		ddx_sp->ddx_rid = r->lck.remlockid;
		ddx_sp->ddx_marked = 0;
		dnext(blech);
	}
	return (0);
blech:
	return (1);
}

/*
 * Name
 *	clmdd_analyze
 * Description
 *	Analyzes the dependencies for the given lock request.  If
 *	all information is available locally and deadlock exists,
 *	blow away the lock request.  If there are external references,
 *	call clmdd_encode to send off the TWFG to the next node.
 * Arguments
 *	lock - the lock request to check
 * Returns
 *	Nothing.
 * Global Data Affected
 * 	None.
 */
STATIC void
clmdd_analyze(struct reclock *lock)
{

	CLM_ASSERT( lock != NULL, "Invalid input lock" );

	if (ddx_alloc())
		return;
	DDX_CHECK();
	lock->dl.deadlock_stamp = 0;		/* force */
	ddx_sp->ddx_type = CLIENT;		/* C -> */
	ddx_sp->ddx_site = lock->lck.site;
	clmdd_setid(&ddx_sp->ddx_clid, lock);
	ddx_sp->ddx_rid =
	  lock->lck.site == OurSiteID ? lock->lck.lockid : lock->lck.remlockid;
	ddx_sp->ddx_marked = 1;
	dnext(blech);
	clmdd_pushr(lock);			/* C -> R */
#ifdef	DEBUG
	DBPRINTF((DBG_DEADLOCK, "clmdd_analyze: constructed stack:\n"));
	clmdd_printstack();
#endif
	clmdd_processddx();
blech:
	return;
}

/*
 * Name
 *	clmdd_processddx
 * Description
 *	Handles the in-core deadlock stack after it's been built
 *	by clmdd_decode() (network) or clmdd_analyze() (local).
 * Arguments
 *	None.
 * Returns
 *	Nothing.
 * Global Data Affected
 *	ddx_stack and friends
 */
STATIC void
clmdd_processddx(void)
{
	ddx_t *ddx;
	clm_msg_t *cm;
	twfg_pgm_t *twfg;
	purge_msg_t *prg;
	unsigned long twfglen;
	uint node, lockid;


	if (!ddx_stack)
		return;

	/*
	 *	Look for cycles.
	 */
	ddx = clmdd_walk(ddx_stack, ddx_end);
	if (ddx != (ddx_t *) 0) {
		/*
		 *	Try to pick the the most recently blocked request
		 */
		ddx = shoot_me_now(ddx_stack, ddx);

		CLM_ASSERT( ddx->ddx_type == BRANCH, 
			    "Deadlock graph corrupt" );

		/* older version will try to cancel on CLIENT node */
		if (ddx->ddx_type == CLIENT) {
			/*
			 * Here if s_m_n() returned ptr to CLIENT because
			 * stack contained entries built by 3.1 lock manager
			 */
			ddx = ddx_stack;
			node = ddx->ddx_site;
			lockid = ddx->ddx_rid;
		} else {
			node = ddx->ddx_blksite;
			lockid = ddx->ddx_blkid;
		}
		if (node == OurSiteID) {
			/*
			 *	Deny local request.
			 */
			clmdd_purgeclnt(node, lockid);
		} else {
			/*
			 *	Deny remote request.
			 */
			if ((prg = (purge_msg_t *)
			  CLM_MALLOC(sizeof (purge_msg_t))) == NULL) {
				bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC3,
				  "Malloc error in clmdd_processddx()"));
				abort("Malloc error in clmdd_processddx()");
			}

            prg->prg_site = node;
            prg->prg_tid = 0;
            prg->prg_rid = lockid;

		    cm = cccp_msg_create( PORT_DEADLOCK, 0, node, MSG_PURG,
					  sizeof( purge_msg_t ), NULL, prg,
					  NULL );
		    if ( cm == NULL )
		    {
			CLM_FREE( prg );
			bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC3,
						"Malloc error in clmdd_processddx()"));
			abort("Malloc error in clmdd_processddx()");
		    }

		    DBPRINTF((DBG_DEADLOCK, PRG_FORMAT,
			      "< clmdd_processddx", prg->prg_site, 
			      db_clid(prg->prg_clid), prg->prg_tid, prg->prg_rid));
		    if (cccp_send(cm) != CPERR_OK) {
			DBPRINTF((DBG_DEADLOCK,
				  "clmdd_processddx: can't send message\n"));
		    }
		}
		ddx_free();
		return;
	}

	/*
	 *	Prune the stack, send off a TWFG message
	 *	if any remains.
	 */
    node = clmdd_massage();
    if ( ddx_stack != NULL )
    {
	twfg = clmdd_encode(&twfglen, clm_version(node));
	if ( twfg != NULL )
	{
	    CLM_ASSERT( twfglen != 0, "Invalid message length" );

	    cm = cccp_msg_create( PORT_DEADLOCK, 0, node, MSG_TWFG, twfglen,
				  NULL, twfg, NULL );
	    if ( cm == NULL )
	    {
		bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC3,
					"Malloc error in clmdd_processddx()"));
		abort("Malloc error in clmdd_processddx()");
	    }

	    DBPRINTF((DBG_DEADLOCK, "< clmdd_processddx: TWFG length = %d\n",
		      cm->cm_length));
	    if (cccpx_send(cm) != CPERR_OK) {
		DBPRINTF((DBG_DEADLOCK,
			  "clmdd_processddx: can't send message\n"));
	    }
	}
	ddx_free();
    }
}

/*
 * Name
 *	shoot_me_now
 * Description
 *	Follows the chain back from ddxe to ddxs to find the most recently
 *	blocked lock request and returns a pointer to it.  In conditions
 *	where there is an older version of the lock manager present, we
 *	may not be able to find the newest request because the information
 *	will not be in the branches of graphs that are passed through the
 *	3.1 nodes.  In this case we return a pointer to ddxe which will
 *	duplicate the 3.1 behavior.
 * Arguments
 *	ddxs - Start of deadlock chain in stack
 *	ddxe - Point in stack where deadlock was found
 * Returns
 *	NULL	Error
 *	!NULL	pointer to position on stack with latest blocked request.
 * Global Data Affected
 * 	None.
 */
STATIC ddx_t *
shoot_me_now(ddx_t *ddxs, ddx_t *ddxe)
{
    ddx_t *ddx;		/* working pointer */
    ddx_t *baby = NULL;	/* points to shortest-lived request in cycle */
    unsigned long age = -1;	/* age of youngest request */
    int level = 0;		/* branch level */

    CLM_ASSERT( ddxe >= ddxs, "Deadlock location impossible" );
    if ( (ddxs == NULL) || (ddxe == ddxs) )
	return ( NULL );

    CLM_ASSERT( ddxs->ddx_type == CLIENT, "Deadlock graph corrupt" );

    /* Loop back through stack.  Skip over complete branches. Compare age in
     * incomplete blocking branches and maintain pointer to the one with the
     * smallest age value.
     */

    for (ddx = ddxe; ddx > ddxs; ddx--) 
    {
	if (ddx->ddx_type == BRANCH) 
	{
	    if (level) 
	    {
		level--;
	    }
	    else if (ddx->ddx_blkid && ddx->ddx_age < age) 
	    {
		age = ddx->ddx_age;
		baby = ddx;
	    }
	}
	else if (ddx->ddx_type == RETURN) 
	{
	    level++;
	}
    }

    /* Return the youngest branch, if found.  If no branches containing an
     * age value were found, return the top of stack.  This mimicks the
     * HACMP 3.1 behavior.
     */

    return( (baby != NULL) ? baby : ddxe );
}

/*
 * Name
 *	clmdd_walk
 * Description
 *	Walks the ddx stack looking for cycles.  Stops at an
 *	unbalanced RETURN or an INVALID entry.
 * Arguments
 *	ddxs - where to start the search.
 *	ddxe - where to end the search.
 * Returns
 *	NULL	no deadlock
 *	!NULL	deadlock at the returned stack entry
 * Global Data Affected
 * 	None.
 */
STATIC ddx_t *
clmdd_walk(ddx_t *ddxs, ddx_t *ddxe)
{
    ddx_t *ddx;

    CLM_ASSERT( ddxe >= ddxs, "Deadlock location impossible" );
    if ( (ddxs == NULL) || (ddxe == ddxs) )
	return ( NULL );

    CLM_ASSERT( ddxs->ddx_type == CLIENT, "Deadlock graph corrupt" );

    /* Do comparison on site and client id */
    for (ddx = ddxs + 1; ddx < ddxe; ddx++) 
    {
	if (ddx->ddx_type == CLIENT
	    && ddx->ddx_site == ddxs->ddx_site
	    && (0 == memcmp(&ddx->ddx_clid, &ddxs->ddx_clid, 
			    sizeof(clm_client_id_t))))
	{
	    return (ddx);
	}
    }

    return( NULL );
}

/*
 * Name
 *	clmdd_massage
 * Description
 *	Removes unnecessary data from the deadlock stack, prior
 *	to encoding & sending it off.  If the graph is fully
 *	resolved, frees the stack and returns 0; otherwise returns
 *	the node to which the TWFG should be sent.
 * Arguments
 *	None.
 * Returns
 *	0	all done
 *	>0	send to this node
 *	<0	external references, but I don't know where to send
 * Global Data Affected
 * 	ddx_stack and friends
 */
STATIC int
clmdd_massage(void)
{
	ddx_t *ddx, *scrubddx;
	int rv, scrublevel;
	struct resource *rp;


	if (ddx_stack == (ddx_t *) 0)
		return (0);

    rv = scrublevel = 0;
    scrubddx = (ddx_t *) 0;
    for (ddx = ddx_stack; ddx < ddx_end; ddx++) 
    {
	/* invariant: scrubddx is valid only when rv==0 */
	CLM_ASSERT( rv == 0 || scrubddx == NULL, "Deadlock stack insane!" );
	CLM_ASSERT( scrublevel >= 0, "Deadlock stack insane!" );

	switch (ddx->ddx_type) 
	{
	    case BRANCH:
	    {
		if (!rv) 
		{
		    if (!scrubddx)
			scrubddx = ddx;
		    scrublevel++;
		}
		break;
	    }

	    case RETURN:
	    {
		if (scrubddx && --scrublevel == 0) {
		    while (scrubddx <= ddx) {
			scrubddx->ddx_type = SCRUBBED;
			scrubddx++;
		    }
		    scrubddx = NULL;
		}
		break;
	    }

	    case XCLIENT:
	    {
		if (rv <= 0) {
		    rv = ddx->ddx_site;
		    CLM_ASSERT( rv >= 0, "Invalid client site" );
		    scrubddx = NULL;
		}
		break;
	    }

	    case XRSRC:
	    {
		if (rv <= 0) 
		{
		    rp = find_resource( ddx->ddx_name, ddx->ddx_namelen,
				        DLM_RES_VMS );
		    if (NULL != rp )
		    {
			rv = (rp->master != 0) ? rp->master : OurSiteID;
		    } 
		    else 
		    {
			bsdlog( LOG_ERR, "clmdd_massage: can't find"
			        "res %s", printbin(ddx->ddx_name,
						   ddx->ddx_namelen) );
			rv = -1;
		    }

		    scrubddx = NULL;
		}
		break;
	    }

	    case CLIENT:
	    case RSRC:
	    case SCRUBBED:
	    {
		break;
	    }

	    case INVALID:
	    {
		goto later;
	    }
	}
    }

later:
    if (rv == 0)
	ddx_free();

    return (rv);
}

STATIC int
ddx_alloc(void)
{
	ddx_t *ddx;

	CLM_ASSERT( ddx_stack == NULL, "Deadlock stack already exists" );
	DBPRINTF((DBG_DEADLOCK, "ddx_alloc: %d entries\n", DDX_SIZE));
	ddx_stack = (ddx_t *) CLM_MALLOC(DDX_SIZE * sizeof (ddx_t));
	if (!ddx_stack) {
		bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC4,
		  "Malloc error in ddx_alloc()"));
		abort("malloc error in ddx_alloc()");
	}
	ddx_count = DDX_SIZE;
	ddx_end = (ddx_stack + ddx_count);
	ddx_sp = ddx_stack;

	/*
	 *	Paranoia -- set all the entries to INVALID.
	 */
	ddx = ddx_sp;
	do {
		ddx->ddx_type = INVALID;
	} while (++ddx < ddx_end);
	DDX_CHECK();
	return (0);
}

STATIC int
ddx_expand(void)
{
    ddx_t *ddx;
    ddx_t *nddx;


    CLM_ASSERT( ddx_stack != NULL, "No deadlock stack" );
    DBPRINTF((DBG_DEADLOCK, "ddx_expand: %d -> %d entries\n", ddx_count,
	      ddx_count + DDX_INCR));

    /* The following is essentially
     *	    ddx_stack = realloc(ddx_stack, ddx_count + DDX_INCR);
     *	but I need to use ddx_free() in case of an error.
     */

    nddx = (ddx_t *) CLM_MALLOC(ddx_count + DDX_INCR * sizeof (ddx_t));
    if (!nddx) {
	bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC5,
				"Malloc error in ddx_expand()"));
	abort("malloc error in ddx_expand()");
    }

    (void) memcpy( nddx, ddx_stack, ddx_count * sizeof( ddx_t ) );
    CLM_FREE(ddx_stack);

    ddx_stack = nddx;
    ddx_sp = ddx_stack + ddx_count;

    /* Paranoia -- set all the newly allocated entries to INVALID.
     */

    ddx = ddx_sp;
    ddx_count += DDX_INCR;
    ddx_end = (ddx_stack + ddx_count);
    do {
	ddx->ddx_type = INVALID;
    } while (++ddx < ddx_end);

    DDX_CHECK();
    return (0);
}

STATIC void
ddx_free(void)
{
    ddx_t * ddx;


    if ( ddx_stack != NULL )
    {
	DBPRINTF((DBG_DEADLOCK, "ddx_free: stack + %d names\n",
		  ddx_rnames_alloced));
	for (ddx = ddx_stack; ddx < ddx_end; ddx++) 
	{
	    CLM_ASSERT((ddx->ddx_type >= INVALID &&
			ddx->ddx_type <= SCRUBBED), "Bad type in ddx_free");
	    if ( (ddx->ddx_type == XRSRC) && (ddx->ddx_name != NULL) )
	    {
		CLM_FREE(ddx->ddx_name);
		ddx->ddx_name = NULL;
		ddx_rnames_alloced--;
	    }
	}

	CLM_FREE(ddx_stack);
	ddx_stack = NULL;

	CLM_ASSERT( ddx_rnames_alloced == 0, "Resource names still allocated "
		    "for deadlock stack, but stack is empty" );
    }
}

/*
 * Name
 *	cpop
 * Description
 *	Returns a pointer to the specified number of bytes from the 
 *	specified memory.  The message pointer is advanced and reset to
 *	the next iov record if necessary.
 * Arguments
 *	msg - pointer to the message pointer
 *	length - number of bytes to pull from the message
 *	msglen - pointer to the counter of # bytes left in segment
 *	iov - pointer to the iov pointer
 *	count - pointer to the length of the array
 * Returns
 *	Pointer to either the memory or a copy of the memory because it
 *	crossed between segments
 * Global Data Affected
 * 	ddx_stack, ddx_sp.
 */
STATIC char *
cpop(char **msg, int length, uint *msglen, iovec_t **iov, int *count)
{
    static char	buf[128];
    char *ptr = &buf[0];

    CLM_ASSERT((length > 0 && length <= 128), "Invalid length in cpop");
    CLM_ASSERT((*msglen >= 0), "Negative msglen in cpop");

    /* if size fits in this segment, just return pointer to it */
    if (*msglen >= length) {
	ptr = *msg;
	*msg += length;
	*msglen -= length;
	return ptr;
    }

    /* otherwise, copy fragments into static local buffer */
    memcpy(ptr, *msg, *msglen);
    ptr = &buf[*msglen];
    length -= *msglen;

    /* there should be more fragments */
    if (*count <= 0) {
	bsdlog(LOG_CRIT, "Bad count in cpop\n");
	return NULL;
    }

    /* advance iov to next block - assume length crosses only one boundary */
    (*iov)++;

    /* last entry in iov array is extra and set to NULLs */
    CLM_ASSERT(((*iov)->iov_base), "NULL addr in cpop");

    /* set msg and length values for new segment & adjust remaining count */
    *msg = (*iov)->iov_base + (*iov)->iov_off;
    *msglen = (*iov)->iov_len;
    (*count)--;

    /* copy rest of length from new fragment */
    memcpy(ptr, *msg, length);

    *msg += length;
    *msglen -= length;

    return &buf[0];
}

/*
 * Name
 *	tpop
 * Description
 *	Returns the specified number of bytes from the specified memory
 *	as an unsigned int.  The message pointer is advanced and reset to
 *	the next iov record if necessary.
 * Arguments
 *	msg - pointer to the message pointer
 *	length - number of bytes to pull from the message
 *	msglen - pointer to the counter of # bytes left in segment
 *	iov - pointer to the iov pointer
 *	count - pointer to the length of the array
 * Returns
 *	unsigned int containing value from buffer
 * Global Data Affected
 * 	ddx_stack, ddx_sp.
 */
STATIC unsigned long
tpop(char **msg, int length, uint *msglen, iovec_t **iov, int *count)
{
    unsigned int	value = 0;

    CLM_ASSERT((length > 0 && length <= 4), "Invalid length in tpop");
    CLM_ASSERT((*msglen >= 0), "Negative msglen in tpop");

    /* loop until we have the # of bytes wanted */
    while (length > 0) {

	/* if nothing left in this seg, advance iov */
	if (0 == *msglen) {

	    /* # of segs left should be > 0 */
	    if (*count <= 0) {
		bsdlog(LOG_CRIT, "Bad count in tpop\n");
		return 0;
	    }

	    /* advance iov to next block */
	    (*iov)++;

	    /* last entry in iov array is extra and set to NULLs */
	    CLM_ASSERT(((*iov)->iov_base), "NULL addr in tpop");

	    /* set msg and length values for new seg & adjust remaining count */
	    *msg = (*iov)->iov_base + (*iov)->iov_off;
	    *msglen = (*iov)->iov_len;
	    (*count)--;
	}

	/* shift next byte onto return value & update position values */
	value = (value << 8) | (unsigned int)(**msg);
	length--;
	(*msglen)--;
	(*msg)++;
    }

    return value;
}

/* routines that use these use the same var names */
#define tpop4(cp)	tpop(&cp, 4, &msglen, &iov, &count)
#define tpop2(cp)	tpop(&cp, 2, &msglen, &iov, &count)

/*
 *	Shipping and receiving TWFG fragments.
 */

/*
 * Name
 *	clmdd_decode
 * Description
 *	Unpacks and expands the data in a 'TWFG' message.  I was
 *	originally going to do the expansion as a separate step,
 *	but the data structure manipulation would have been far
 *	too complicated.
 * Arguments
 *	iov - an array of (base, offset, length) triples containing
 *	  the TWFG
 *	count - length of the array
 * Returns
 *	None, but if ddx_stack is non-NULL when this routine
 *	exits, it is valid.
 * Global Data Affected
 * 	ddx_stack, ddx_sp.
 */
STATIC void
clmdd_decode(iovec_t *iov, int count, int version)
{
	clm_client_t *cl;
	int site;
	clm_client_id_t clid;
	int mode;
	uint namelen, lockid;
	twfg_pgm_t *msg;
	uint msglen;
	char *name;
	int locknode ;
    dlm_listnode_t * node;
    struct reclock * rec;

	CLM_ASSERT( ddx_stack == NULL, "Deadlock stack already exists" );
	ddx_alloc();
#ifdef	DEBUG
	clmdd_printtwfg(iov, count, version);
#endif
	while (count--)
	{
	    /* last entry in iov array is extra and set to NULLs */
	    CLM_ASSERT((iov->iov_base), "NULL addr in clmdd_decode");
	    msg = iov->iov_base + iov->iov_off;
	    msglen = iov->iov_len;
	    while (msglen--)
	    {
		switch (*msg++)
		{
		case TWFG_SITECLID:
			DDX_CHECK();
			ddx_sp->ddx_type = CLIENT;
			ddx_sp->ddx_site = tpop4(msg);

            ddx_sp->ddx_clid.clid_i[0] = tpop4(msg);
            ddx_sp->ddx_clid.clid_i[1] = tpop4(msg);

			ddx_sp->ddx_rid = tpop4(msg);
			dnext(blech);
			break;
		case TWFG_XCLIENT:
			site = tpop4(msg);
			if (site == OurSiteID) {
			    clid.clid_i[0] = tpop4(msg);
			    clid.clid_i[1] = tpop4(msg);

			    cl = findclient(site, &clid);
			    if ( (cl == NULL)
				|| dlm_list_is_empty( & cl->cli_queue) )
				break;

			    node = dlm_list_get_head( & cl->cli_queue );
			    rec = CLIENT_LINK_TO_RECLOCK( node );
			    if (clmdd_pushc( rec ))
				goto blech;
			} else {
			    DDX_CHECK();
			    ddx_sp->ddx_type = XCLIENT;
			    ddx_sp->ddx_site = site;

			    ddx_sp->ddx_clid.clid_i[0] = tpop4(msg);
			    ddx_sp->ddx_clid.clid_i[1] = tpop4(msg);

			    dnext(blech);
			}
			break;
		case TWFG_XRSRC:
			mode = tpop2(msg);
			lockid = tpop4(msg);
			namelen = tpop2(msg);

		    locknode = tpop4(msg) ; 

			name = cpop(&msg, namelen, &msglen, &iov,
				    &count);
			if (clmdd_pushxr(name, namelen, mode,
					 lockid, locknode) < 0)
                            goto blech;
			break;
		case TWFG_BRANCH:
			DDX_CHECK();
			ddx_sp->ddx_type = BRANCH;

            ddx_sp->ddx_blkid = 0;
            ddx_sp->ddx_blksite = 0;
            ddx_sp->ddx_age = 0;

			dnext(blech);
			break;
		case TWFG_RETURN:
			DDX_CHECK();
			ddx_sp->ddx_type = RETURN;
			dnext(blech);
			break;
		default:
			DBPRINTF((DBG_DEADLOCK,
			  "clmdd_decode: unknown opcode %d\n", *msg));
			goto blech;
		}
	    }
	    iov++;
	}
#ifdef	DEBUG
	DBPRINTF((DBG_DEADLOCK, "clmdd_decode: received stack:\n"));
	clmdd_printstack();
#endif
	clmdd_processddx();
	return;
blech:
	ddx_free();
	CLM_ASSERT( ddx_stack == NULL, "Deadlock stack destroyed, but still "
		    "there" );
}

/*
 * Name
 *	clmdd_encode
 * Description
 *	Converts the in-core deadlock stack to the serially encoded
 *	format used in a 'TWFG' message.  Allocates the necessary
 *	space.
 * Arguments
 *	msglen (result) - the length of the resulting message.
 *	If clmdd_encode returns successfully, *msglen is guaranteed
 *	to be > 0.
 * Returns
 *	The message data, or NULL in case of an error.
 * Global Data Affected
 * 	None.
 */
STATIC twfg_pgm_t*
clmdd_encode(unsigned long *msglen, int version)
{
	twfg_pgm_t *rv, *rvp;
	ddx_t *ddx;
	int count;

	CLM_ASSERT( msglen != NULL, "Invalid message length buffer" );
#ifdef	DEBUG
	if (clmdd_validate())
		return( NULL );
#endif

	*msglen = clmdd_encodelen(version);
	DBPRINTF((DBG_DEADLOCK, "clmdd_encode entered, msglen=%d\n",
	  *msglen));
	if (!*msglen)
		return( NULL );
	if ((rv = CLM_MALLOC(*msglen * sizeof (twfg_pgm_t)))
	  == NULL) {
		bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC6,
		  "Malloc error in clmdd_encode()"));
		abort("Malloc error in clmdd_encode()");
	}
	ddx = ddx_stack;
	count = ddx_count;
	rvp = rv;
	while (count--) {
		switch (ddx->ddx_type) {
		case INVALID:
			goto break2;
		case CLIENT:
			*rvp++ = TWFG_SITECLID;
			tpush4(rvp, ddx->ddx_site);
			rvp += 4;

            tpush4(rvp, ddx->ddx_clid.clid_i[0]);
            rvp += 4;
            tpush4(rvp, ddx->ddx_clid.clid_i[1]);
            rvp += 4;

			tpush4(rvp, ddx->ddx_rid);
			rvp += 4;
			break;
		case XCLIENT:
			*rvp++ = TWFG_XCLIENT;
			tpush4(rvp, ddx->ddx_site);
			rvp += 4;

            tpush4(rvp, ddx->ddx_clid.clid_i[0]);
            rvp += 4;
            tpush4(rvp, ddx->ddx_clid.clid_i[1]);
            rvp += 4;

			break;
		case XRSRC:
			*rvp++ = TWFG_XRSRC;
			tpush2(rvp, ddx->ddx_mode);
			rvp += 2;
			tpush4(rvp, ddx->ddx_lockid);
			rvp += 4;
			tpush2(rvp, ddx->ddx_namelen);
			rvp += 2;

          tpush4(rvp, ddx->ddx_locknode) ; rvp += 4 ;

			resname_bcopy(&ddx->ddx_name[0], rvp, ddx->ddx_namelen);
			rvp += ddx->ddx_namelen;
			break;
		case BRANCH:
			*rvp++ = TWFG_BRANCH;

            tpush4(rvp, ddx->ddx_blkid);
            rvp += 4;
            tpush4(rvp, ddx->ddx_blksite);
            rvp += 4;
            tpush4(rvp, ddx->ddx_age);
            rvp += 4;

			break;
		case RETURN:
			*rvp++ = TWFG_RETURN;
			break;
		case SCRUBBED:
        case RSRC:
			break;
		}
		ddx++;
	}
break2:
	/* spec */
	CLM_ASSERT( rv == NULL || *msglen > 0, "Message, but no name" );

	/* expected length */
	CLM_ASSERT( rvp == (rv + (*msglen)), "Message tail pointer invalid" );

#ifdef	DEBUG
	if (rv != NULL) {
		iovec_t iov;

		iov.iov_base = rv;
		iov.iov_off = 0;
		iov.iov_len = *msglen;
		clmdd_printtwfg(&iov, 1, version);
	}
#endif
	return (rv);
}


/*
 * Name
 *	clmdd_encodelen
 * Description
 *	Figures out how long a 'TWFG' message will be, based on the
 *	in-core deadlock stack.
 * Arguments
 *	None.
 * Returns
 *	The length.
 * Global Data Affected
 * 	None.
 */
STATIC unsigned long
clmdd_encodelen(int version)
{
	unsigned long rv = 0;
	ddx_t *ddx;
	int count;


	if (ddx_stack) {
		ddx = ddx_stack;
		count = ddx_count;
		while (count--) {
			switch (ddx->ddx_type) {

			case CLIENT:
				rv += (TWFG_SITECLID_LEN + 1);
				break;
			case XCLIENT:
				rv += (TWFG_XCLIENT_LEN + 1);
				break;
			case XRSRC:
				CLM_ASSERT( ddx->ddx_namelen > 0, "No name" );
				rv += 1 + ddx->ddx_namelen ;
				rv += TWFG_XRSRC_LEN ;
				break;
			case BRANCH:
				rv += (TWFG_BRANCH_LEN + 1);
				break;
			case RETURN:
				rv += (TWFG_RETURN_LEN + 1);
				break;
            case RSRC:
            case SCRUBBED:
              break;
			case INVALID:
				goto out;
			}
			ddx++;
		}
	}
out:
	return (rv);
}

/*
 * Name
 *	clmdd_purgeclnt
 * Description
 *	This routine is called on a client site as the first step in
 *	curing deadlock.  It searches the client list for the client
 *	specified by (site, clid) for the lock request identified by rid,
 *	then looks up the master for the resource.  If the master is
 *	the local site, it calls clmdd_purgersrc() directly.  Otherwise
 *	it conses up a BLOW message to send to the master of the
 *	resource.
 * Arguments
 *	site - the client site
 *	pclid - client id of the client
 *	rid - reclock id to purge
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
STATIC void
clmdd_purgeclnt(int site, unsigned long rid)
{
	clm_client_t *cl;
	struct reclock *waiter;
	struct reclock *a;
	struct resource *rp;
	blow_msg_t *blw;
	clm_msg_t *cm;
	int len;
	clm_client_id_t	clid;
	clm_client_id_t	*pclid = &clid;
    dlm_listnode_t * node;

	DBPRINTF((DBG_DEADLOCK,
			"clmdd_purgeclnt: site=%d clid=%x %x rid=0x%lx\n",
			site, db_clidp(pclid), rid));
	a = find_reclock(DLM_RES_VMS, rid);
	if (!a)
		return;

	if ( a->lck.lk_flags & LKM_XID ) {
		/* Use XID as deadlock id */
		memcpy(pclid, &a->lck.xid, sizeof(clm_client_id_t));
	} else {
		/* Use pid as deadlock id */
		set_client_idp(pclid, a->lck.pid);
	}

	cl = findclient(site, pclid);
	if (!cl) {
		/*
		 *	Not an error; the client exited before
		 *	deadlock was detected.
		 */
		DBPRINTF((DBG_DEADLOCK,
			  "clmdd_purgeclnt: no client (%d, %x %x)\n", 
			  site, db_clidp(pclid)));
		return;
	}

    waiter = NULL;
    for ( EACH_DLM_LISTNODE( & cl->cli_queue, node ) )
    {
	waiter = CLIENT_LINK_TO_RECLOCK( node );
	if (waiter->lck.lockid == rid)
	    break;
    }

    if ( waiter == NULL )
    {
	/* Not an error; the client gunned this lock before deadlock was
	 * detected.
	 */

	DBPRINTF((DBG_DEADLOCK,
		  "clmdd_purgeclnt: no reclock (%d, %x %x, 0x%lx)\n", 
		  site, db_clidp(pclid), rid));
	return;
    }

	/*
	 *	Look up the resource.
	 */
	rp = (waiter->lck.rh.rh) ? resptr(waiter->lck.rh) : NULL;
	if ( rp == NULL )
		bsdlog(LOG_ERR, MSGSTR(PURG_DEADLOCK,
				       "clmdd_purgeclnt: bad resptr, rh=0x%lx\n"),
		       waiter->lck.rh.rh );

	if (rp->master == 0) {
		/*
		 *	We're the master for the resource too.
		 */
		clmdd_purgersrc(OurSiteID, pclid, rid, rp->namelen, rp->name);
	} else {
		/*
		 *	Send a BLOW message to the master.
		 */

        len = sizeof (blow_msg_t) + rp->namelen + 1;

		if ((blw = (blow_msg_t *) CLM_MALLOC(len)) == NULL) {
			bsdlog(LOG_CRIT, "Malloc error in clmdd_purgeclnt()");
			abort("Malloc error in clmdd_purgeclnt()");
		}

	    /* remove possible alignment cruft */
	    (void) memset( blw, 0, len );

        blw->blw_site = site;
        memcpy(&blw->blw_clid, pclid, sizeof(clm_client_id_t));
        blw->blw_rid = rid;
        blw->blw_namelen = rp->namelen;
        blw->blw_nameoff = sizeof (blow_msg_t);
        resname_bcopy(rp->name, (char *) blw + blw->blw_nameoff,
                      rp->namelen);

	    cm = cccp_msg_create( PORT_DEADLOCK, 0, rp->master, MSG_BLOW, len,
				  NULL, blw, NULL );
	    if ( cm == NULL ) 
	    {
		bsdlog(LOG_CRIT, "Can't malloc message in clmdd_purgeclnt");
		abort("Malloc error in clmdd_purgeclnt");
	    }

#if defined(DBPRINT)
        DBPRINTF((DBG_DEADLOCK, BLW_FORMAT, "< clmdd_purgeclnt",
                  blw->blw_site, db_clid(blw->blw_clid), blw->blw_rid, 
                  blw->blw_namelen,
                  (char *)cm->cm_data + blw->blw_nameoff));
#endif
		if (cccp_send(cm) != CPERR_OK) {
			DBPRINTF((DBG_DEADLOCK,
			  "clmdd_purgeclnt: can't send BLOW message\n"));
		}
	}
}

/*
 * Name
 *	clmdd_purgersrc
 * Description
 *	This routine is called on the master site for a resource in
 *	order to cure deadlock.
 * Arguments
 *	site - the client site
 *	rid - reclock id to purge (lockid if site == OurSiteID,
 *	  remlockid otherwise)
 *	namelen - length of the resource name in bytes
 *	name - the resource name
 * Returns
 *	!0 - resource no longer mastered on this site but on return value
 *	0 - did the deed or encountered an error
 */
STATIC int
clmdd_purgersrc(int site, clm_client_id_t *pclid, unsigned long rid, int namelen, char *name)
{
	uint		lid;
	struct reclock	*waiter;
	struct resource	*rp;
    dlm_listnode_t * node;

    DBPRINTF((DBG_DEADLOCK, "clmdd_purgersrc: site %d  rid %x  name %s\n",
		site, rid, printbin(name, namelen))) ;
	/*
	 *	Now find the reclock.  What a pain.
	 */

    lid = ( site == OurSiteID ) ? rid : mapid( rid, site );

    waiter = NULL;
    for ( EACH_DLM_LISTNODE( & dl_timeout, node ) )
    {
	waiter = DEADLOCK_LINK_TO_RECLOCK( node );
	if (waiter->lck.lockid == lid)
	    break;
    }

    if ( waiter == NULL )
    {
	return (0);
    }

	/*
	 *	Look up the resource.
	 */
	rp = (waiter->lck.rh.rh) ? resptr(waiter->lck.rh) : NULL;
	if ( rp == NULL )
		bsdlog(LOG_ERR, "clmdd_purgersrc: bad resptr, rh=0x%lx\n",
		       waiter->lck.rh.rh);

	/* if res not mastered here ... */
	if (rp->master)
		return rp->master;

	/* Now we can just dink the lock.  Which list is it on: CONV or WAIT? */
	if (waiter->lck.state & LOCK_CONVERTING) {
		check_handles(rp, 0) ;
		remove_queue(&rp->lockqs[RLQ_CONV], waiter, RL_RQ_OFS) ;
		if (site == OurSiteID) {
		    unset_lock_timer(waiter);
		}
		lck_set_seq(rp, waiter);
		add_queue(&rp->lockqs[RLQ_GRANT], waiter, RL_RQ_OFS) ;
		waiter->lck.reqmode = waiter->lck.mode;
		waiter->lck.state &= ~LOCK_CONVERTING;
		waiter->lck.state |= LOCK_GRANTED;
		check_handles(rp, 0) ;

		/*
		 * notify must go after lck_set_seq to preserve seq order
		 * on secondary
		 */
		notify(waiter, RESP_AST, DLM_DEADLOCK);
		convert_locks(rp, waiter) ;	/* promote any other locks */
	}
	else
	{
		notify(waiter, RESP_AST, DLM_DEADLOCK);
		check_handles(rp, 0) ;
		remove_queue(&rp->lockqs[RLQ_WAIT], waiter, RL_RQ_OFS) ;
		if (site == OurSiteID) {
		    unset_lock_timer(waiter);
		}

		/*
		 * Delete it off the other various lists and free up
		 * the reclock.
		 */
		free_le(rp->type, waiter);
		convert_locks(rp, NULL) ;	/* promote any other locks */
		decr_delete_resource(rp, 0) ;	/* LAST REF to rp */
	}

	delete_timeout_reclock(waiter);

	return (0);
}

/*
 *	Utilities.
 */

#ifdef	DEBUG
/*
 * Name
 *	clmdd_validate
 * Description
 *	Makes sure the in-core stack is OK for sending.
 * Arguments
 *	None.
 * Returns
 *	0	in-core stack is OK
 *	1	something's wrong
 * Global Data Affected
 *	None.
 */
STATIC int
clmdd_validate(void)
{
	int level = 0;
	int atend = 0;
	ddx_t *ddx;
	int count;

	if (ddx_stack) {
		ddx = ddx_stack;
		count = ddx_count;
		while (count--) {
			switch (ddx->ddx_type) {
			case INVALID:
				atend = 1;
			case SCRUBBED:
			case CLIENT:
			case XCLIENT:
			case XRSRC:
				break;
			case BRANCH:
				level++;
				break;
			case RETURN:
				level--;
				break;
			case RSRC:
				DBPRINTF((DBG_DEADLOCK,
				  "clmdd_validate: RSRC in stack\n"));
#ifndef __linux__
				brkpoint();
#endif /* !__linux__ */
				return (1);
			default:
				DBPRINTF((DBG_DEADLOCK,
				  "clmdd_validate: type %d in stack\n",
				  ddx->ddx_type));

			}
			if (ddx->ddx_type != INVALID && atend) {
				DBPRINTF((DBG_DEADLOCK,
				  "clmdd_validate: cruft in stack\n"));
				return (1);
			}
			ddx++;
		}
	}
	if (level) {
		DBPRINTF((DBG_DEADLOCK,
		  "clmdd_validate: unbalanced branch/return\n"));
		return (1);
	}
	return (0);
}

/*
 * Name
 *	clmdd_printstack
 * Description
 *	Prints the contents of the ddx stack, from the stack
 *	pointer to the base, for debugging.
 * Arguments
 *	None.
 * Returns
 *	Nothing.
 * Global Data Affected
 * 	None.
 */
STATIC void
clmdd_printstack(void)
{
	ddx_t *ddx;
	int n;

	if (!ddx_stack) {
		DBPRINTF((DBG_DEADLOCK, "dumping ddx_stack, it's null\n"));
		return;
	}
	n = (int) (ddx_sp - ddx_stack);
	DBPRINTF((DBG_DEADLOCK,
	  "dumping ddx stack, sp @ %d, allocs: ddx %d rnames %d\n", n,
	  ddx_count, ddx_rnames_alloced));
	ddx = ddx_sp;
	while (ddx >= ddx_stack) {
		switch (ddx->ddx_type) {
		case INVALID:
			DBPRINTF((DBG_DEADLOCK, "%4d INVALID\n", n));
			break;
		case CLIENT:
			DBPRINTF((DBG_DEADLOCK,
			  "%4d CLIENT site=%d clid=%x %x rid=0x%lx marked=%d\n",
			  n, ddx->ddx_site, db_clid(ddx->ddx_clid), 
			  ddx->ddx_rid, ddx->ddx_marked));
			break;
		case XCLIENT:
			DBPRINTF((DBG_DEADLOCK,
			  "%4d XCLIENT site=%d clid=%x %x rid=0x%lx marked=%d\n",
			  n, ddx->ddx_site, db_clid(ddx->ddx_clid), 
			  ddx->ddx_rid, ddx->ddx_marked));
			break;
		case RSRC:
			DBPRINTF((DBG_DEADLOCK,
			  "%4d RSRC namelen=%d name=%s mode=%d\n", n,
			  ddx->ddx_namelen, ddx->ddx_name, ddx->ddx_mode));
			break;
		case XRSRC:
			DBPRINTF((DBG_DEADLOCK,
			  "%4d XRSRC namelen=%d name=%s mode=%d"
						" lockid=0x%x  node=%d\n",
			  n, ddx->ddx_namelen, ddx->ddx_name, ddx->ddx_mode,
			  ddx->ddx_lockid, ddx->ddx_locknode));
			break;
		case BRANCH:
			DBPRINTF((DBG_DEADLOCK,
			  "%4d BRANCH lockid=0x%x site=%d age=%d\n", n,
			  ddx->ddx_blkid, ddx->ddx_blksite, ddx->ddx_age));
			break;
		case RETURN:
			DBPRINTF((DBG_DEADLOCK, "%4d RETURN\n", n));
			break;
		case SCRUBBED:
			DBPRINTF((DBG_DEADLOCK, "%4d SCRUBBED\n", n));
			break;
		default:
			DBPRINTF((DBG_DEADLOCK, "%4d unknown type=%d\n", n,
			  (int) ddx->ddx_type));
			break;

		}
		ddx--;
		n--;
	}
}

/*
 * Name
 *	clmdd_printtwfg
 * Description
 *	Prints the data in a 'TWFG' message.
 * Arguments
 *	iov - an array of (base, offset, length) triples containing
 *	  the TWFG
 *	count - length of the array
 * Returns
 *	None.
 * Global Data Affected
 */
STATIC void
clmdd_printtwfg(iovec_t *iov, int count, int version)
{
    uint site, rid, lockid, age;
    clm_client_id_t clid;
    int mode, namelen;
    twfg_pgm_t *msg;
    uint msglen;
    char buf[128];
    char	*name;

    while (count--)
    {
	DBPRINTF((DBG_DEADLOCK, "TWFG iov off=%d len=%d\n",
	  iov->iov_off, iov->iov_len));
	/* last entry in iov array is extra and set to NULLs */
	CLM_ASSERT((iov->iov_base), "NULL addr in clmdd_printtwfg");
	msg = iov->iov_base + iov->iov_off;
	msglen = iov->iov_len;
	while (msglen--)
	{
	    switch (*msg++) {
	    case TWFG_SITECLID:
		    site = tpop4(msg);

            clid.clid_i[0] = tpop4(msg);
            clid.clid_i[1] = tpop4(msg);

		    rid = tpop4(msg);
		    DBPRINTF((DBG_DEADLOCK,
		      "SITECLID site=%d clid=%x %x rid=0x%lx\n",
		      site, db_clid(clid), rid));
		    break;
	    case TWFG_XCLIENT:
		    site = tpop4(msg);

            clid.clid_i[0] = tpop4(msg);
            clid.clid_i[1] = tpop4(msg);

		    DBPRINTF((DBG_DEADLOCK,
		      "XCLIENT site=%d clid=%x %x\n", site, 
			      db_clid(clid)));
		    break;
	    case TWFG_XRSRC:
		    mode = tpop2(msg);
		    lockid = tpop4(msg);
		    namelen = tpop2(msg);
		    CLM_ASSERT( namelen < 128, "Resource name too big" );

			site = tpop4(msg) ; 

		    name = cpop(&msg, namelen, &msglen, &iov,
				&count);
		    resname_bcopy(name, buf, namelen);

		    *(buf + namelen) = '\0';
		    DBPRINTF((DBG_DEADLOCK,
		      "XRSRC mode=%d lockid=0x%x site=%d name=%s namelen=%d\n",
		      mode, lockid, site, buf, namelen));
		    break;
	    case TWFG_BRANCH:

            lockid = tpop4(msg);
            site = tpop4(msg);
            age = tpop4(msg);

		    DBPRINTF((DBG_DEADLOCK,
		      "BRANCH lockid=0x%x site=%d age=%d\n",
		      lockid, site, age));
		    break;
	    case TWFG_RETURN:
		    DBPRINTF((DBG_DEADLOCK, "RETURN\n"));
		    break;
	    default:
		    DBPRINTF((DBG_DEADLOCK,
		      "UNKNOWN OPCODE: %d\n", *msg));
	    }
	}
	iov++;
    }
}
#endif /* DEBUG */

/*
 * Name
 *	clmdd_setid
 * Description
 *      If XID exists in reclock, sets deadlock id to XID
 *      else sets deadlock id using pid
 * Arguments
 *	pdid - pointer to deadlock id
 *      preclock - pointer to reclock
 * Returns
 *      Nothing
 * Global Data Affected
 *	None.
 */
STATIC void 
clmdd_setid(clm_client_id_t *pdid, struct reclock *preclock)
{
    if ( preclock->lck.lk_flags & LKM_XID )
    {
	/* Use XID as deadlock id */
	memcpy(pdid, &preclock->lck.xid, sizeof(clm_client_id_t));
    }
    else
    {
        /* Use pid as deadlock id */
	set_client_idp(pdid, preclock->lck.pid);
    }
}

#ifdef	DEBUG
/*
 *	Return true if the given reclock is on the convert queue
 *	for the given resource.
 */
STATIC int
on_convert_q(struct resource *rp, struct reclock *lck)
{
    struct reclock *r;

    if ( (lck->lck.state & LOCK_CONVERTING) )
    {
	for ( EACH_DLM_LISTNODE( & rp->lockqs[RLQ_CONV], r ) )
	{
	    if (r == lck)
		return (1);
	}
    }


    return (0);
}
#endif

/* clmdd_primo - Determine if a node is has the lowest node ID in the cluster
 *
 * Determines if the specified node ID is the ID of the lowest numbered (with
 * respect to node ID) active node in the cluster.
 * 
 * RETURNS
 * 
 * If the specified node ID is the lowest, TRUE is returned.  Otherwise, FALSE
 * is returned.
 */

STATIC bool_t
clmdd_primo( int me )
{
    iterator_t * iter;
    unsigned     id;


    iter = dlm_nodeid_iterator_create();
    while ( ! iterator_next( iter, & id )
	    && (id != me) && ! clm_node_active( id ) )
    {
	/* empty */
    }

    iterator_destroy( iter );


    /* Technically speaking, it should be impossible for the loop to
     * terminate due to running out of nodes to test.  The loop should either
     * hit the local node or some other active node.  If this does not happen,
     * then there is something wrong with the internal cluster state.
     */

    CLM_ASSERT( (id == me) || clm_node_active( id ),
	        "Our node not found in active set" );

    return( id == me );
}

/* ============================================================ */

#ifdef	COMMUNICATION_STEROIDS
/*
 *	Extension to the communications package to handle large
 *	messages.
 *
 *	Shoot!  We can't send 4096 bytes of data, the total
 *	packet can't be >4096 bytes.  Aim low for safety.
 */
#define	CCCP_BUFFERSIZE		3584	/* this doesn't belong here */

typedef struct {
	unsigned long	sh_type;	/* real type of the message */
	unsigned long	sh_serial;	/* serial number */
	short		sh_seg;		/* segment number within message */
	unsigned short	sh_totseg;	/* total number of segments */
} seg_hdr_t;
#define	SEG_FORMAT	"%s: SEGM type=0x%lx serial=%ld seg=%d totseg=%d\n"
#define	CCCPX_ABORT		(-1)
#define	CCCPX_CHUNKSIZE		(CCCP_BUFFERSIZE - sizeof (seg_hdr_t))

/*
 *	Stuff to guarantee we can get an abort message out,
 *	even if mallocs are failing.
 */
/*ARGSUSED*/
STATIC void
cccpx_null(clm_msg_t *cm, int v)
{
}
static seg_hdr_t ssh;
static clm_msg_t scm;

/*
 * Name
 *	cccpx_send
 * Description
 *	Works like cccp_send, except the message can be indefinitely
 *	long and callbacks are not supprted.  WARNING: this function
 *	is not implemented as if it were going to be frequently used!
 * Arguments
 *	ocm - the message
 * Returns
 *	One of the error codes defined in cccp.h
 */
STATIC int
cccpx_send(clm_msg_t *ocm)
{
	clm_msg_t *cm;
	char *msgbuf, *omsgp;
	static unsigned long serial = 0;
	unsigned long nbytes, toxmit;
	unsigned int nextseg = 1;
	unsigned int totseg;
	int rv;

	if (ocm->cm_length <= CCCP_BUFFERSIZE)
		return (cccp_send(ocm));
	serial++;
	rv = CPERR_OK;
	totseg = ((ocm->cm_length - 1) / CCCPX_CHUNKSIZE) + 1;
	for (nbytes = ocm->cm_length, omsgp = ocm->cm_data; nbytes > 0; ) {
		toxmit = nbytes > CCCPX_CHUNKSIZE ? CCCPX_CHUNKSIZE : nbytes;

		msgbuf = CLM_MALLOC(sizeof (seg_hdr_t) + toxmit);
		if (msgbuf == NULL) {
		    abort("Malloc error in cccpx_send");
		}
		((seg_hdr_t *) msgbuf)->sh_type = ocm->cm_type;
		((seg_hdr_t *) msgbuf)->sh_serial = serial;
		((seg_hdr_t *) msgbuf)->sh_seg = nextseg;
		((seg_hdr_t *) msgbuf)->sh_totseg = totseg;
		(void) memcpy( msgbuf + sizeof (seg_hdr_t), omsgp, toxmit );

	 
	    CLM_ASSERT( ocm->cm_callback == NULL, "Callback not NULL" );
	    cm = cccp_msg_create( ocm->cm_port, 0, ocm->cm_node, MSG_SEGM,
				  sizeof( seg_hdr_t ) + toxmit, NULL, msgbuf,
				  NULL );
	    if ( cm == NULL )
	    {
		CLM_FREE( msgbuf );
		abort( "Malloc error in cccpx_send" );
	    }

	    if ((rv = cccp_send(cm)) != CPERR_OK)
		break;
	    nbytes -= toxmit;
	    omsgp += toxmit;
	    nextseg++;
	}
	if (rv == CPERR_OK)
		return (CPERR_OK);

	if (nextseg != 1) {
		/*
		 *	We actually sent part of the message,
		 *	so abort it.  Use static data since
		 *	we may have gotten here from a malloc
		 *	failure.
		 */
		ssh.sh_type = ocm->cm_type;
		ssh.sh_serial = serial;
		ssh.sh_seg = CCCPX_ABORT;
		ssh.sh_totseg = 0;
		scm.cm_type = MSG_SEGM;
		scm.cm_length = sizeof (seg_hdr_t);
		scm.cm_node = ocm->cm_node;
		scm.cm_port = ocm->cm_port;
		scm.cm_callback = cccpx_null;
		scm.cm_data = &ssh;
		scm.cm_private = (void *) 0;
		(void) cccp_send(&scm);
	}

	/* Free old message */
	CLM_FREE(ocm->cm_data);
	cccp_msg_destroy( ocm );
	return (rv);
}

/*
 * Name
 *	cccpx_recv
 * Description
 *	Collects segments of an oversized message until all the
 *	segments are received or an ABORT arrives.  When the message
 *	is complete, calls the specified callback function with the
 *	message segments and lengths stored in an array of iovec_t
 *	structures.
 *
 *	Restriction: only one oversized message can be collected
 *	at a time.  If this were a general facility, we'd want to
 *	register completion callbacks on a per-message-type basis.
 *	But this is just a bandage to keep an extreme condition
 *	from not working...
 * Arguments
 *	cm - the message
 *	completion_cb - the completion callback
 * Returns
 *	Nothing.
 */
STATIC void
cccpx_recv(clm_msg_t *cm, void (*completion_cb)(iovec_t *, int, int))
{
	seg_hdr_t *sh;
	static iovec_t *iov;
	static unsigned long serial = 0;
	static int node = -1;
	static int count = 0;
	int i;

	sh = (seg_hdr_t *) cm->cm_data;
	DBPRINTF((DBG_DEADLOCK, SEG_FORMAT, "> cccpx_recv",
	  sh->sh_type, sh->sh_serial, sh->sh_seg, sh->sh_totseg));
	if (sh->sh_seg == CCCPX_ABORT)
		goto abort;
	if (sh->sh_serial != serial || cm->cm_node != node) {
		/*
		 *	This is a new big message, so
		 *	throw away anything we're in the
		 *	process of collecting.
		 */
		if (count)
			free_iovec(iov, count);
		serial = sh->sh_serial;
		node = cm->cm_node;
		count = sh->sh_totseg;
		if (!count)
			goto abort;
		if ((iov = CLM_MALLOC((count+1) * sizeof (iovec_t)))
		  == (iovec_t *) 0) {
			/* XXX - NLS me */
		    abort("Malloc error in cccpx_recv");
		}
		(void) memset( iov, 0, (count+1) * sizeof (iovec_t) );
	}
	if (sh->sh_seg <= count) {
		CLM_ASSERT( sh->sh_seg > 0, "Invalid segment number" );
		/*
		 * copy message 'cause it gets freed after comm callback
		 * returns
		 */
		if ((iov[sh->sh_seg - 1].iov_base =
		    (void *)CLM_MALLOC(cm->cm_length)) == (void *) 0) {
		    /* XXX - NLS me */
		    abort("Malloc error in cccpx_recv");
		}
		memcpy(iov[sh->sh_seg - 1].iov_base, cm->cm_data,
				cm->cm_length);
		iov[sh->sh_seg - 1].iov_off = sizeof (seg_hdr_t);
		iov[sh->sh_seg - 1].iov_len = cm->cm_length
		  - sizeof (seg_hdr_t);
	}
	/*
	 *	Check to see whether we're finished collecting
	 *	messages or not.
	 */
	for (i = 0; i < count; i++)
		if (!iov[i].iov_len)
			return;
	(*completion_cb)(iov, count, clm_version(cm->cm_node));
abort:
	if (count)
		free_iovec(iov, count);
	serial = 0;
	node = -1;
	count = 0;
}

STATIC void
free_iovec(iovec_t *iov, int count)
{
    iovec_t *oiov;

    CLM_ASSERT( iov != NULL, "Invalid iovec pointer" );
    oiov = iov;
    while (count--) {
	if (iov->iov_base)
	    CLM_FREE( iov->iov_base );
	iov++;
    }

    CLM_FREE( oiov );
}
#endif	/* COMMUNICATION_STEROIDS */

#if	defined(UNIT_TEST) && defined(COMMUNICATION_STEROIDS)
#define	EL_GORDO	49199

/*
 *	Tests cccpx_send by consing up a colossal message to
 *	send to the specified node.
 */
STATIC int
jabber(int node)
{
    clm_msg_t * cm;
    char      * dater;
    int         rv;

    if ((dater = (char *) CLM_MALLOC(EL_GORDO)) == NULL) {
	bsdlog(LOG_CRIT, "Malloc error in jabber()");
	abort("malloc error in jabber()");
    }
    
    (void) memset( dater, 0x2a, EL_GORDO );

    cm = cccp_msg_create( PORT_DEADLOCK, 0, node, MSG_GORD, EL_GORDO,
			  NULL, dater, NULL );
    if ( cm == NULL ) 
    {
	CLM_FREE( dater );
	bsdlog(LOG_CRIT, "Malloc error in jabber()");
	abort("Malloc error in jabber()");
    }

    if ((rv = cccpx_send(cm)) != CPERR_OK) {
		DBPRINTF((DBG_DEADLOCK,
		  "jabber: can't send colossal message\n"));
		return (rv);
    }

    DBPRINTF((DBG_DEADLOCK,
	      "jabber: just sent %d bytes of garbage to node %d\n",
	      EL_GORDO, node));
    return (0);
}

STATIC void
jabber_recv(iovec_t *iov, int count)
{
	int i, j, bungle;
	char *dp;

	i = 0;
	bungle = 0;
	while (count--) {
		dp = iov->iov_base + iov->iov_off;
		for (j = 0; j < iov->iov_len; j++)
			if (*dp++ != 0x2a)
				bungle++;
		i += iov->iov_len;
		iov++;
	}
	DBPRINTF((DBG_DEADLOCK,
	  "jabber_recv: received %d bytes of garbage, thanks a lot.\n", i));
	if (bungle)
		DBPRINTF((DBG_DEADLOCK,
		  "jabber_recv: %d bungles\n", bungle));
}

#endif	/* defined(UNIT_TEST) && defined(COMMUNICATION_STEROIDS) */
