/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_debug.c,v 1.7 2001/11/02 22:14:45 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		clm_write_core_log
 *		dbprintf
 *		dbg_test
 *		dumprintf
 *		transaction_type_string
 */

#ifndef lint
#endif

/* ============================================================ */
/*
 * Linux note:  This file operates in two modes:
 *
 * - bsdlog - which we have converted to use printk, i.e., go
 *  into the system message log.  This is keyed through the
 *  clm_dbprintf() function below.
 *
 * - file - where, as with the "dump" portions of clm_cti.c,
 *  expects to do a 'standard' file write to dump out data,
 *  based on a system call request from the user side.  This
 *  is keyed through the dumprintf() function below, which
 *  expects a file descriptor.
 *
 * The former method is handled by us having provided a bsdlog()
 * that uses the printk facilities.  The latter is not, YET,
 * handled, but, it needs to be converted to using the /proc
 * filesystem.  In the future then, we will replace this latter
 * method with facilities to allow you to read the debug data
 * through /proc.
 */
/* ============================================================ */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/signal.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <stdarg.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "clm_data.h"			/* external data definitions */
#include "dlm_kernel.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */

/*
 * this file contains debugging support for the lock manager
 * The global variable "debug" contains the bitmask of all
 * modes of debugging output that have been requested by the user. 
 * The debug routines below accept a bitmask of debugging modes
 * at which a certain message should be output. If any of the specified
 * modes are present in the global "debug" variable, the message will be 
 * output. If debug is enabled for any mode, DBG_FORCE messages will
 * be printed (see clm_main.c:main).
 */

/*
 * Name
 *	DBPRINTF
 * Description
 *	debugging printf
 * 	put message to bsdlog
 *	ignore the FILE argument
 * 
 * Arguments
 *	mask - desired debug level
 *	fmt - printf style format string
 *	a,b,c etc - printf style arguments
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */

void clm_dbprintf (int mask, const char *fmt, ...)
{
    va_list   args;


    if ( (dlm_debug & mask) != 0 )
    {
	va_start( args, fmt );
	vbsdlog( LOG_DEBUG, fmt, args );
	va_end( args );
    }
}

int
dbg_test(int flag)
{
    return (dlm_debug & flag) != 0 ;
}

/****************************************************************************/
/* For some reason, cproto gets very confused by the 'const' part of the
 * return value.  Sigh...
 */

#ifndef CPROTO
const char *
transaction_type_string( int type )
{
    extern const char * const dlm_opstab[];
    extern const int dlm_nops;
    const char * str = "INVALID";
    
    if ( type > 0 )
    {
	type--;
	if ( type < dlm_nops )
	{
	    str = dlm_opstab[ type ];
	}
    }

    return( str );
}

#endif /* CPROTO */

#ifdef DLM_TRANSACTION_STRINGIFY

/*
 * See comments in str_trans() regarding these being ifdefed
 * out.
 */

/*
 * Structure to string conversions for everything that is passed
 * between systems.  All return a pointer to a static buffer
 * containing the converted string.
 */

static char
*str_direction_tab[] =
{
    "DLM_REQUEST",
    "DLM_RESPONSE",
    "PTI_REQUEST",
    "PTI_RESPONSE",
    "PTI_ASYNC",
    "PTI_STAT",
    "??BAD??"
} ;

static char
	*str_regreq(struct clm_regreq *),
	*str_lockreq(struct lockreq *),
	*str_lockinfo(struct lockinfo *),
	*str_purgelocks(struct purgereq *),
	*str_scninfo(struct scninfo *),
	*str_rmpinfo(struct res_migr_params *),
	*str_gmpinfo(struct glob_migr_params *),
	*str_statsinfo(struct getstats *),
	*str_lkvalue(char *r),
	*str_resparam(clm_resparams_t *r),
	*str_globparam(struct clm_globparams *r),
	*str_resparam(clm_resparams_t *r),
	*str_stats(struct clm_statistics *r),
	*str_ast_handle(struct ast_handle *r),
	*str_rh(union clm_rh *),
	*str_reuse_li(union reuse_li *r, int type),
	*str_route(struct routeinfo *r) ;

#ifndef __linux__
#define BUFCHK if (strlen(buf) >= sizeof buf) brkpoint() ;
#else
/*
 * Is there a Linux equivalent to brkpoint()?
 */
#define BUFCHK 
#endif /* !__linux__ */

/*
 * str_trans
 *
 * Format a 'struct transaction' (clm_trans_t for the tasteless)
 * into a static buffer & return the buffer.
 *
 * Linux note: Ifdef the stuff out for now, worry about it
 * later.
 */

char *
str_trans(struct transaction *t)
{
    static char buf[2500] ;
    char *p;
    char *s ;
    int i ;
    int type;
    int site;


    sprintf(p, "trans %x  ", t) ;
    p += strlen(p) ;
    if (t->clm_prog != DLM_PROG)
    {
	sprintf(p, "clm_prog (%x) != %x\n", t->clm_prog, DLM_PROG) ;
	p += strlen(p) ;
    }
    if (t->clm_vers != DLM_VERS)
    {
	sprintf(p, "vers %x  ", t->clm_vers) ;
	p += strlen(p) ;
    }
    sprintf(p, "seq c/p %x/%x  ", t->clm_sequence, t->pti_sequence) ;
    p += strlen(p) ;
    if (t->clm_authpid && t->clm_authpid != t->clm_pid)
    {
	sprintf(p, "apid %d  ", t->clm_authpid) ;
	p += strlen(p) ;
    }
    if (t->clm_sender)
    {
	sprintf(p, "sender %x  ", t->clm_sender) ;
	p += strlen(p) ;
    }
    if (t->clm_next)
    {
	sprintf(p, "next %x\n", t->clm_next) ;
	p += strlen(p) ;
    }
    sprintf(p, "type %s\nroute: %s\n",
		dlm_get_lock_type_string(t->clm_locktype), str_route(&t->clm_route)) ;
    p += strlen(p) ;

    switch (t->clm_type)
    {
    case DLM_REGISTER:
	if (t->clm_direction == PTI_RESPONSE)
	    s = str_rh(&t->clm_handle) ;
	else
	    s = str_regreq(&t->clm_register) ;
	break ;

    case DLM_LOCK:
	s = str_lockreq(&t->clm_lockreq) ;
	break ;

    case DLM_UNLOCK:
    case DLM_CANCEL:
	s = str_lockinfo(&t->clm_lockinfo) ;
	break ;

    case DLM_APPREG:
	s = "" ;
	break ;

    case DLM_PURGELOCKS:
	s = str_purgelocks(&t->clm_purgelocks) ;
	break ;

    case DLM_SCN_OP:
	s = str_scninfo(&t->clm_scninfo) ;
	break ;

    case DLM_RMIGR_OP:
	s = str_rmpinfo(&t->clm_rmpinfo) ;
	break ;

    case DLM_GMIGR_OP:
	s = str_gmpinfo(&t->clm_gmpinfo) ;
	break ;

    case DLM_STATS:
	s = str_statsinfo(&t->clm_statsinfo) ;
	break ;

    default:
	sprintf(p, "clm_request: bad request type %d", t->clm_type) ;
	p += strlen(p) ;
	return buf ;
    }

    strcpy( p, transaction_type_string( t->clm_type ) );
    p += strlen(p) ;
    sprintf(p, "/dir %d (%s):\n", t->clm_direction,
	str_direction_tab[t->clm_direction > PTI_STAT || t->clm_direction < 0
		? PTI_STAT + 1 : t->clm_direction]) ;
    p += strlen(p) ;
    i = strlen(s) + 1 ;
    if (p + i >= buf + sizeof buf - 2)
	i = buf + sizeof buf - p - 3 ;
    (void) memcpy( p, s, i );
    if (p[i])
    {
	p[i] = '\n' ;
	p[i + 1] = '\0' ;
    }
    BUFCHK
    return buf ;
}

/*
 * str_regreq
 *
 * format a registration request (struct regreq) into a static buffer
 * return a pointer to the buffer
 */
static char *
str_regreq(struct clm_regreq *r)
{
    char *p ;
    static char buf[MAXRESOURCELEN + 100] ;

    p = buf ;
    sprintf(p,
            "pid %d  type %s  site %d  count %d  rh %s  name '%s'\n",
            r->pid,
            dlm_get_lock_type_string(r->type),
            r->site, r->count,
            str_rh(&r->rh),
            printbin(r->name, r->namelen)) ;
    
    BUFCHK
    return buf ;
}

/*
 * str_lockreq
 *
 * format a lock request (resource name) into a static buffer
 * return a pointer to the buffer
 */
static char *
str_lockreq(struct lockreq *r)
{
    static char buf[MAXRESOURCELEN + 340] ;

    sprintf(buf,
            "%sresname: %s",
            str_lockinfo(&r->lockinfo),
            printbin(r->nameinfo.name,
                     r->nameinfo.namelen)) ;
    BUFCHK
    return buf ;
}

/*
 * str_lockinfo
 *
 * format lock information (struct lockinfo) into a static buffer
 * return a pointer to the buffer
 */

static char *
str_lockinfo(struct lockinfo *r)
{
    static char buf[400], mode[40], rmode[40], bmode[40] ;

    strcpy(mode, clm_printmode(r->mode)) ;
    strcpy(rmode, clm_printmode(r->reqmode)) ;
    strcpy(bmode, clm_printmode(r->bastmode)) ;
    sprintf(buf,
            "gm %s  rm %s  bm %s  site %d\n"
            "lkid 0x%x  flags 0x%x  pid %d  rlkid %x  state %x  rh %s\n"
            "li_reuse %s  type %s  seq %x  xid 0x%x%08x\nlkval:  %s\n",
            mode,
            rmode,
            bmode,
            r->site,
            r->lockid,
            r->lk_flags,
            r->pid,
            r->remlockid,
            r->state,
            str_rh(&r->rh),
            str_reuse_li(&r->li_reuse,
                         r->type),
            dlm_get_lock_type_string(r->type),
            r->seqnum,
            r->xid.xid_u.xid_int[0],
            r->xid.xid_u.xid_int[1],
            str_lkvalue(r->lkvalue)) ;
    BUFCHK
    return buf ;
}

/*
 * str_purgelocks
 *
 * format a purge request (struct purgereq) into a static buffer
 * return a pointer to the buffer
 */

static char *
str_purgelocks(struct purgereq *r)
{
    static char buf[200] ;

    sprintf(buf,
            "prgsite %d  prgpid %d  reqsite %d  reqpid %d\n"	/*paste*/
            "restype %s  flags 0x%x",
            r->purgesite,
            r->purgepid,
            r->site,
            r->pid,
            dlm_get_lock_type_string(r->type),
            r->flags) ;
    BUFCHK
    return buf ;
}

/*
 * str_scninfo
 *
 * format a SCN request (struct scninfo) into a static buffer
 * return a pointer to the buffer
 */

static char *
str_scninfo(struct scninfo *r)
{
    static char buf[MAXDLMNAMELEN + 200] ;

    sprintf(buf,
            "site %d  pid %d  bit_len %d  scn_op %d  lockid 0x%x\n"
            "name %s\nlkval %s\n",
            r->site,
            r->pid,
            r->bit_len,
            r->scn_op,
            r->lockid,
            printbin(r->nameinfo.name,
                     r->nameinfo.namelen),
            str_lkvalue(r->lkvalue)) ;
    BUFCHK
    return buf ;
}

/*
 * str_rmpinfo
 *
 * format a resource migration request (struct res_migr_params)
 * into a static buffer
 * return a pointer to the buffer
 */

static char *
str_rmpinfo(struct res_migr_params *r)
{
    static char buf[MAXDLMNAMELEN + 200] ;

    sprintf(buf,
#ifdef __linux__
            "name '%s'  type %s  reqsite %d  flags %#xd resptr %p\n"
#else
            "name '%s'  type %s  reqsite %d  flags 0x%x resptr 0x%x\n"
#endif /* __linux__ */
            "params %s\n",
            printbin(r->rmp_resname, r->rmp_namelen),
            dlm_get_lock_type_string(r->rmp_type),
            r->rmp_site,
            r->rmp_flags,
            r->rmp_rp,
            str_resparam(&r->rmp_params)) ;
    
    BUFCHK
    return buf ;
}

/*
 * str_gmpinfo
 *
 * format a global migration request (struct glob_migr_params)
 * into a static buffer
 * return a pointer to the buffer
 */

static char *
str_gmpinfo(struct glob_migr_params *r)
{
    static char buf[240] ;
    
    sprintf(buf,
            "rqflags 0x%x  params: %s\n",
            r->gmp_flags,
			str_globparam(&r->gmp_params)) ;
    BUFCHK
    return buf ;
}

/*
 * str_statsinfo
 *
 * format a resource statistics request (struct getstats) into a static buffer
 * return a pointer to the buffer
 */

static char *
str_statsinfo(struct getstats *r)
{
    static char buf[MAXDLMNAMELEN + 200] ;

    sprintf(buf,
            "name '%s'  type %s  stats: %s\n",
            printbin(r->resname, r->namelen),
            dlm_get_lock_type_string(r->type),
            str_stats(&r->statistics)) ;
    
    BUFCHK
    return buf ;
}

/*
 * str_rh
 *
 * format a resource handle (union clm_rh) into a static buffer
 * return a pointer to the buffer
 */

static char *
str_rh(union clm_rh *r)
{
    static char buf[60] ;

    if ( ! r->rh)
	strcpy(buf, "0") ;
    else
	sprintf(buf,
            "rh %#xu %#xu %#xu %#xu/%#xlu",
            r->handle.site,
            r->handle.type,
            r->handle.seg,
            r->handle.cookie,
            r->rh) ;

    BUFCHK
    return buf ;
}

/*
 * str_route
 *
 * format a route (struct routeinfo) into a static buffer
 * return a pointer to the buffer
 */

static char *
str_route(struct routeinfo *r)
{
    static char buf[540], asts[3][140] ;
    int i ;

    strcpy(asts[0], str_ast_handle(&r->asth)) ;
    strcpy(asts[1], str_ast_handle(&r->basth)) ;
    strcpy(asts[2], str_ast_handle(&r->cbasth)) ;
    for (i = 0 ; i < 3 ; i++)
	if (strcmp(asts[i], "0"))
	    goto longp ;
    if ( ! r->respflags && ! r->rt_master && ! r->callback
		&& ! strcmp(str_rh(&r->lcl_rh), "0"))
	strcpy(buf, "0") ;
    else
longp:
    sprintf(buf,
            "flags %#xd  master %d  cb* %p, reuse %s\n"
            "asth: %s\nbasth: %s\ncbasth: %s\n",
            r->respflags,
            r->rt_master,
            r->callback,
            str_rh(&r->lcl_rh),
            asts[0],
            asts[1],
            asts[2]) ;
    
    BUFCHK
    return buf ;
}

/*
 * str_ast_handle
 *
 * format an ast handle (struct ast_handle) into a static buffer
 * return a pointer to the buffer
 */

static char *
str_ast_handle(struct ast_handle *r)
{
    static char buf[140] ;

    if ( ! r->asth_pid && !  r->asth_tid && !  r->asth_dest && !  r->asth_funcp
		&& ! r->asth_arg && ! r->asth_aptr && ! r->asth_client)
	strcpy(buf, "0") ;
    else
	sprintf(buf,
            " ast pid %d  tid %d  lksb* %p  funcp *%p\n"
            "arg* %p aptr* %p  client* %p",
            r->asth_pid,
            r->asth_tid,
            r->asth_dest,
            r->asth_funcp,
            r->asth_arg,
            r->asth_aptr,
            r->asth_client) ;
    
    BUFCHK
    return buf ;
}

/*
 * str_lkval
 *
 * format a lock value block (char[]) into a static buffer
 * return a pointer to the buffer
 */

static char *
str_lkvalue(char *r)
{
    static char buf[MAXLOCKVAL * 4 + 10] ;
    int i ;
    char *p ;

    p = buf ;
    for (i = 0 ; i < MAXLOCKVAL ; i++)
    {
	sprintf(p, "<%x>", r[i]) ;
	p += strlen(p) ;
    }
    BUFCHK
    return buf ;
}

/*
 * str_resparam
 *
 * format resource migration params (clm_resparams_t) into a static buffer
 * return a pointer to the buffer
 */

static char *
str_resparam(clm_resparams_t *r)
{
    static char buf[200] ;

    sprintf(buf,
            "valid %d  stickiness %d",
            r->cr_valid,
            r->cr_stickiness) ;
    BUFCHK
    return buf ;
}

/*
 * str_globparam
 *
 * format global migration params (struct globparam) into a static buffer
 * return a pointer to the buffer
 */

static char *
str_globparam(struct clm_globparams *r)
{
    static char buf[200] ;

    sprintf(buf,
            "valid %d  recalc_time %d  decay_rate %s",
            r->cg_valid,
            r->cg_recalc_time,
            kfl_ftos(r->cg_decay_rate)) ;
    BUFCHK
    return buf ;
}

/*
 * str_stats
 *
 * NOTCODED
 * format lock statistics (struct clm_statistics) into a static buffer
 * return a pointer to the buffer
 */

static char *
str_stats(struct clm_statistics *r)
{
    static char buf[400] ;

    strcpy(buf, "no stats") ;
    BUFCHK
    return buf ;
}

/*
 * str_reuse_li
 *
 * format timeout-pid/bounds union (union reuse_li) into a static buffer
 * choose format from lock type
 * return a pointer to the buffer
 */

#undef lb
#undef ub
#undef to
#undef realpid

static char *
str_reuse_li(union reuse_li *r, int type)
{
    static char buf[80] ;

    if (type == DLM_RES_UNIX)
      sprintf(buf,
              "lb %x  ub %x",
              r->bounds_unix.lb,
              r->bounds_unix.ub) ;
    else
      sprintf(buf,
              "tmo %d  realpid %d",
              r->mod_clm.to,
              r->mod_clm.realpid) ;
    
    BUFCHK
    return buf ;
}
#endif /* DLM_TRANSACTION_STRINGIFY */
