/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_direct.c,v 1.12 2001/11/28 00:30:49 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		bounce_msg
 *		dir_proc_request
 *		dir_proc_update
 *		dir_read_request
 *		dir_read_request_locked
 *		dir_read_update
 *		dir_read_update_locked
 *		unbounce
 *		update_directory
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/signal.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/timer.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "cltrc.h"
#include "clm_data.h"			/* external data definitions */
#include "dlm_kernel.h"
#include "dlm_version.h"
#include "dlm_recover.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */

typedef struct update_hold_q 
{
    dlm_listnode_t    link;
    rl_update_t     * up_date;
} up_q_t;

static DECLARE_DLM_LIST( hold_up );

/*
 * Name
 *	clm_direct_init
 * Description
 *	clm_direct static initialization function.  Initializes static vars.
 * Arguments
 *	None
 * Returns
 *	Nothing
 * Global Data Affected
 * 	RLDB
 */
void
clm_direct_init(void)
{
    dlm_list_init( & hold_up );
}

/*
 * Name
 *	unbounce
 * Description
 *	This function is called with the pid of a local client that is
 *	being purged.  If there are any requests on the bounce queue that
 *	match this pid, remove them.
 * Arguments
 *	dead_pid - PID of client being purged
 * Returns
 *	Nothing
 * Global Data Affected
 * 	bounce_list
 */
void
unbounce(int	dead_pid)
{
  clm_msg_t	*msg, *next, *prev = NULL;
  clm_trans_t	*req;
  struct reclock *a;

  DBPRINTF((DBG_PURGE, "Enter unbounce for %d list = 0x%x", dead_pid,
			bounce_list));
  /*
   * Walk through the bounce_list.  Save enough pointers so that we
   * can remove entries as we go.
   */
    for (msg = bounce_list; msg; msg = next) 
    {
	next = msg->cm_private;
	req = (clm_trans_t *)msg->cm_data;

	if (req->clm_pid != dead_pid)
	{
	    prev = msg;
	}
	else
	{
	    DBPRINTF((DBG_PURGE, "Unbounce req from %d\n", dead_pid));

	    /* unlink the message from the list */
	    if (prev) {
		prev->cm_private = msg->cm_private;
	    } else {
		bounce_list = msg->cm_private;
	    }

	    /* If this req has a lock on the inflight queue, it must be freed */
	    a = find_reclock(DLM_RES_VMS, req->clm_lockreq.lockinfo.lockid);
	    if (NULL != a) {
		if (INFLIGHT_WAITING & a->lck.state) {
		    DBPRINTF((DBG_LOCK|DBG_PURGE,
			      "Free remote inflight reclock 0x%x r=%s from unbounce",
			      req->clm_lockreq.lockinfo.lockid,
			      printbin(&req->clm_lockreq.nameinfo.name[0],
				       req->clm_lockreq.nameinfo.namelen)));
		    free_le(DLM_RES_VMS, a);
		}
	    }

	    /* free the memory */
	    clm_free_trans(req);
	    cccp_msg_destroy( msg );
	}
  }

  return;
}

/*
 * Name
 *	resend_held
 * Description
 *	This function is called with a pointer to an rl_entry that contains
 *	the name information of a resource whose location was just updated
 *	on the local directory.  Resend any msgs on the bounce queue with
 *	the same name.  If called with a NULL pointer, resend all entries.
 * Arguments
 *	rl_entry - rl_entry containing nameinfo
 * Returns
 *	Nothing
 * Global Data Affected
 * 	RLDB
 */
void
resend_held(rl_info_t	*rl_entry)
{
	clm_msg_t	*msg, *next, *prev = NULL;
	clm_trans_t	*req;
	struct nameinfo	*ni;
	clm_msg_t	*bounce_save = bounce_list;
	clm_msg_t	*save_inf, *save_inf_lst;
	int		ret;

    extern clm_msg_t *inflight;
    extern clm_msg_t *inflight_last;

	/*
	 * We dequeue the entire bounce_list here in case the cccp_send call
	 * that we make ends up re-queueing the request back onto the list.
	 * This hopefully avoids infinite loops and queue corruption.
	 */
	if (bounce_list) {
		DBPRINTF((DBG_CONFIG,
				"Enter resend_held list = 0x%x", bounce_list));
		bounce_list = NULL;
	}

	/*
	 * Walk through the bounce_list.  Save enough pointers so that we
	 * can remove entries as we go.
	 */
	for (msg = bounce_save; msg; prev = msg, msg = next) {
		next = msg->cm_private;
		req = (clm_trans_t *)msg->cm_data;

		/* if no name specified, resend all entries */
		if (NULL == rl_entry)
			goto send_each;

		/* switch on request type to determine name */
		switch (req->clm_type) {
		case DLM_LOCK:
		case DLM_UNLOCK:
			ni = &req->clm_lockreq.nameinfo;
			if (rl_entry->rl_namelen != ni->namelen ||
				rl_entry->rl_type != req->clm_locktype ||
					strncmp(&ni->name[0], rl_entry->rl_name,
					ni->namelen))
				continue;
			break;
		case DLM_REGISTER:
			if (rl_entry->rl_namelen != req->clm_register.namelen ||
				rl_entry->rl_type != DLM_RES_UNIX ||
				strncmp(&req->clm_register.unix_name[0],
					rl_entry->rl_name, rl_entry->rl_namelen))
				continue;
			break;
		case DLM_SCN_OP:
			ni = &req->clm_scninfo.nameinfo;
			if (rl_entry->rl_namelen != ni->namelen ||
				rl_entry->rl_type != DLM_RES_VMS ||
					strncmp(&ni->name[0], rl_entry->rl_name,
					ni->namelen))
				continue;
			break;
		case DLM_PURGELOCKS:
			bsdlog(LOG_CRIT,
				"Purge request type=%d in resend_held",
				req->clm_type);
			abort("resend_held");
			break;
		case DLM_RMIGR_OP:
			if (rl_entry->rl_namelen != req->clm_rmpinfo.rmp_namelen ||
				rl_entry->rl_type != req->clm_rmpinfo.rmp_type ||
				strncmp(&req->clm_rmpinfo.rmp_resname[0],
					rl_entry->rl_name, rl_entry->rl_namelen))
				continue;
			break;
		case DLM_STATS:
			if (rl_entry->rl_namelen != req->clm_statsinfo.namelen ||
				rl_entry->rl_type != req->clm_statsinfo.type ||
				strncmp(&req->clm_statsinfo.resname[0],
					rl_entry->rl_name, rl_entry->rl_namelen))
				continue;
			break;
		default:
			bsdlog(LOG_CRIT,
				"Unknown request type=%d in resend_held",
				req->clm_type);
			abort("resend_held");
			break;
		}
send_each:
		msg->cm_private = NULL;
		if (prev) {
			prev->cm_private = next;
		} else {
			bounce_save = next;
		}
		/*
		 * new purge states can allow requests of DEAD clients to be
		 * queued on the bounce queue.  Check for these and don't send
		 * them out.
		 */
		if (!findclientbypid(OurSiteID,req->clm_pid)) {
		    bsdlog(LOG_CRIT,"Drop req in resend\n");
		    clm_free_trans(req);
		    cccp_msg_destroy( msg );
		    msg = prev;
		    continue;
		}
		/*
		 * Messages that are queued on the bounce list have been
		 * removed from the inflight queue to prevent the
		 * resend_inflight() logic from resending a transaction that
		 * is going to be bounced anyways.  Before we re-send it here,
		 * we need to requeue it on the inflight list.
		 */
		save_inf = inflight;
		save_inf_lst = inflight_last;
		req->clm_next = NULL;
		if (inflight_last != NULL) {
			((clm_trans_t *)(inflight_last->cm_data))->clm_next =
					msg;
			inflight_last = msg;
		} else {
			inflight = msg;
			inflight_last = msg;
		}
		ninflight++;
		DBPRINTF((DBG_PTI,"Enqueued 0x%x on inflight queue", req));

		/*
		 * If the destination value is set to 0, then the message
		 * was reset because of a node failure.  Use downRetry to
		 * reasign the destination, otherwise, just call cccp_send()
		 */
		if (msg->cm_node && clm_node_active(msg->cm_node)) {
			ret = cccp_send(msg);
		} else {
			ret = downRetry(msg);
		}

		if ( ret != CPERR_OK ) {
			bsdlog(LOG_ERR, "cccp_send error in resend_held");
		} else {
			DBPRINTF((DBG_CONFIG,
					"Resend 0x%x for pid %d successfull\n",
					msg, req->clm_pid));
		}

		/* let try not to really screw up the list */
		msg = prev;

	}

	/* relink saved list with bounce_list if necessary */
	if (prev) {
		CLM_ASSERT((prev->cm_private == NULL),
				"resend_held: cm_private not NULL");
		prev->cm_private = bounce_list;
		bounce_list = bounce_save;
	}
}

/*
 * Name
 *	bounce_msg
 * Description
 *	This function is called with a pointer to a message that needs to
 *	be queued until the local directory is updated with the new location
 *	of the master copy of the resource referenced by this request.
 * Arguments
 *	msg - message to be queued
 * Returns
 *	Nothing
 * Global Data Affected
 * 	RLDB
 */
void
bounce_msg(clm_msg_t	*msg)
{
  clm_trans_t	*req = msg->cm_data;
  clm_client_t	*cp;
  struct reclock  *a;

  DBPRINTF((DBG_CONFIG, "Queue bounce message 0x%x", msg));
	
  CLM_ASSERT(((ulong)msg->cm_private & MSG_BOUNCE),
             "Bad cm_private value in bounce_msg");

  /*
   * Before we put this message on the bounce list, take it off the
   * inflight queue.  If the message is queued on both lists during a
   * node rebuild, then it can potentially be re-sent twice and that
   * would be a bad thing.
   */
  msg->cm_private = NULL;
  req = match_request(req);

  /*
   * If this request is for a client that is being purged, then
   * there's not much point in queuing the msg to be resent
   */
  cp = findclientbypid(OurSiteID, req->clm_pid);
  if (!cp || cp->cli_state & CLI_CODE) {
    /* if client is DOA, so is request */
    DBPRINTF((DBG_PURGE,
              "Intercepting bounce req of dead client pid %d",
              req->clm_pid));

    /* If this req has a lock on the inflight queue, it must be freed */
    a = find_reclock(DLM_RES_VMS, req->clm_lockreq.lockinfo.lockid);
    if (NULL != a)
      {
        if (INFLIGHT_WAITING & a->lck.state)
          {
            DBPRINTF((DBG_LOCK|DBG_PURGE,
                      "Free remote inflight reclock 0x%x r=%s "
                      "from bounce_msg",
                      req->clm_lockreq.lockinfo.lockid,
                      printbin(&req->clm_lockreq.nameinfo.name[0],
                               req->clm_lockreq.nameinfo.namelen)));
            free_le(DLM_RES_VMS, a);
          }
      }

    clm_free_trans(req);
    cccp_msg_destroy( msg );
    return;
  }

  /* queue the message on the bounce list */
  msg->cm_private = bounce_list;
  bounce_list = msg;
}

/*
 * Name
 *	dir_bounce_request
 * Description
 *	This callback is used to bounce requests that would otherwise be
 *	stuck in infinite loops caused by a migrating resource.  If the
 *	directory for the resource is local, then queue the request to be
 *	flushed when we get the update from the new master.  If the
 *	directory is remote, then just redirect the request there.  It should
 *	have new information now.
 *
 *	This callback is also used to throw away requests from clients that
 *	have exited while we were galavanting across the cluster looking
 *	for their resource master.
 * Arguments
 *	msg - message that was redirected from the dir port
 * Returns
 *	explained above?
 * Global Data Affected
 * 	RLDB
 */
int
dir_bounce_request(clm_msg_t	*msg)
{
	rl_info_t		rl_entry;	/* RLDB entry structure */
	clm_trans_t		*req;
	int			dir_node;

	DBPRINTF((DBG_DIRECT,
			"entering dir_bounce_request 0x%x hop=%d",
			msg, msg->cm_hops));

	/* Massage incoming message into readable form */
	clm_vers_rcv(msg);

    req = msg->cm_data;
    if ( ! dlm_stuff_rl_info( req, & rl_entry ) )
    {
	bsdlog( LOG_CRIT, "[%s:%s] Invalid directory request type=%d",
	        current->comm, __FUNCTION__, req->clm_type );
	BUG();
    }

    DBPRINTF((DBG_DIRECT, "[%s:%s] dir - %u: \"%s\", type = %d\n",
	      current->comm, __FUNCTION__, req->clm_type,
	      printbin( rl_entry.rl_name, rl_entry.rl_namelen ),
	      rl_entry.rl_type) );


    if ( (msg->cm_hops > 80) && (req->clm_type == DLM_UNLOCK) )
    {
	bsdlog(LOG_ERR,	"Master unlock request for %d:%s exceeds 80 "
	       "hops.  Dropping request.",
	       req->clm_lockreq.nameinfo.namelen,
	       printbin(&req->clm_lockreq.nameinfo.name[0],
			req->clm_lockreq.nameinfo.namelen));
	clm_vers_reset(msg);
	return 0;
    }

    /* If this message has been bouncing around a lot, then there is a
     * problem.  Messages that can safely be dropped should be recognized
     * before this point and dropped.
     */

    if (msg->cm_hops > 100) 
    {
	bsdlog(LOG_ERR, "dir_bounce_request: hop count exceeds 100");
	abort("dir_bounce_request");
    }


    acquire_lock( LK_ENTRY, kproc_pid );

    /* If a request comes back and the client's dead, drop it
     */

    if ( NULL == findclientbypid( OurSiteID, req->clm_authpid ) )
    {
	struct reclock *a;

	DBPRINTF((DBG_CONFIG, "bounce: drop 0x%x (%s) lockid=0x%x pid=%d",
		  msg, printbin(rl_entry.rl_name, rl_entry.rl_namelen),
		  req->clm_lockinfo.lockid, req->clm_pid));

	/*
	 * purge will not clean this lock if it's not on a resource.
	 * It will only not be on a resource if it's an inflighted
	 * LOCK request - don't blow away unlocks!!!
	 */

	if ( req->clm_type == DLM_LOCK )
	{
	    a = find_reclock( DLM_RES_VMS, req->clm_lockreq.lockinfo.lockid );
	    if ( (a != NULL) && mylock( a, req ) )
	    {
		bsdlog(LOG_DEBUG,"Freeing bounced inflighted reclock 0x%x "
		        "r=%s", req->clm_lockreq.lockinfo.lockid,
		        printbin(&req->clm_lockreq.nameinfo.name[0],
				 req->clm_lockreq.nameinfo.namelen));
		free_le(DLM_RES_VMS, a);
	    }
	}


	/* the bounce logic should take this off the inflight queue */
	clm_vers_reset(msg);

	CLM_ASSERT((msg->cm_private == NULL),
		   "bounce_request: cm_private not NULL");
	msg->cm_private = (void *)MSG_BNCE_DONE;

	return FALSE;
    }

    dir_node = clm_direct(rl_entry.rl_name, rl_entry.rl_namelen);
    msg->cm_node = dir_node;
    msg->cm_port = PORT_DIRECT;

    clm_vers_reset(msg);


    /*
     * Here if local directory bounced our request.  If we bounce this
     * one back, this thread will never release the lock necessary to let
     * a directory update message in to give us the new location of the
     * resource being searched for.  So we acknowledge the request and
     * allow it to complete.  When the completion callback runs, we'll put
     * the message on a queue that will be flushed when our directory
     * update comes in.  Setting cm_private indicates the need to resend.
     */

    CLM_ASSERT((msg->cm_private == NULL),
	       "dir_bounce_request: cm_private not NULL");
    msg->cm_private = (void *)MSG_BOUNCE;
    DBPRINTF((DBG_CONFIG, "bounce: queue 0x%x (%s) lockid=0x%x", msg,
	      printbin(rl_entry.rl_name, rl_entry.rl_namelen),
	      req->clm_lockinfo.lockid));

    release_lock( LK_ENTRY, kproc_pid );
    return 0;
}

/*
 * Name
 *	dir_proc_request
 * Description
 *	Use the RLDB interface to lookup the master node information for the
 *	resource specified in the transaction structure.  If the resource is
 *	not found, create new entry in the RLDB for the resource and return
 *	the site id of the requesting site.  Otherwise, return the site id of
 *	the master node for the resource.
 * Arguments
 *	req - lock request in search of master node information
 * Returns
 *	site id of the master node for the specified resource
 * Global Data Affected
 * 	RLDB
 */
static int
dir_proc_request(clm_msg_t	*msg)
{
	rl_info_t		rl_entry;	/* RLDB entry structure */
	clm_trans_t	*req = msg->cm_data;
	int			cached = FALSE;
	int			dir_node;
	int			ret;

    STAMP(req);
    DBPRINTF((DBG_DIRECT, "[%s:%s] %s 0x%lx hop=%d from %d\n",
	      current->comm, __FUNCTION__, "enter",
	      (unsigned long) msg, msg->cm_hops, msg->cm_node));

    if ( ! dlm_stuff_rl_info( req, & rl_entry ) )
    {
	bsdlog( LOG_CRIT, "[%s:%s] Invalid directory request type=%d",
	        current->comm, __FUNCTION__, req->clm_type );
	BUG();
    }

    DBPRINTF((DBG_DIRECT, "[%s:%s] dir - %u: \"%s\", type = %d\n",
	      current->comm, __FUNCTION__, req->clm_type,
	      printbin( rl_entry.rl_name, rl_entry.rl_namelen ),
	      rl_entry.rl_type) );
    

    if ( (msg->cm_hops > 80) && (req->clm_type == DLM_UNLOCK) )
    {
	bsdlog(LOG_ERR,	"Master unlock request for %d:%s exceeds 80 "
	       "hops.  Dropping request.",
	       req->clm_lockreq.nameinfo.namelen,
	       printbin(&req->clm_lockreq.nameinfo.name[0],
			req->clm_lockreq.nameinfo.namelen));
	goto missing_unlock;
    }

    /* If this message has been bouncing around a lot, then there is a
     * problem.  Messages that can safely be dropped should be recognized
     * before this point and dropped.
     */

    if (msg->cm_hops > 100) 
    {
	bsdlog(LOG_ERR, "dir_proc_request: hop count exceeds 100");
	abort("dir_proc_request 4");
    }

    /* check for mis-directed directory request */
    dir_node = clm_direct(rl_entry.rl_name, rl_entry.rl_namelen);
    if (OurSiteID != dir_node) 
    {
	DBPRINTF((DBG_DIRECT, "Not the directory! 0x%x (%s) -> %d\n",
		  msg, printbin(rl_entry.rl_name,
				rl_entry.rl_namelen), dir_node));
	msg->cm_node = dir_node;
	return dir_node;
    }

    /*
     * If we get a directory request while we are in the process of
     * reading in a direct dump (when we first come up), then someone has
     * somehow jumped the gun.  Bounce the request.  The sending node should
     * queue the request until reconfig is done or node up is over.
     */
    if ( ! clm_is_running() ) 
    {
	bsdlog(LOG_ERR, "[%s:%s] (%s) low pri hops=%d id=0x%x",
	       current->comm, __FUNCTION__,
	       printbin(rl_entry.rl_name, rl_entry.rl_namelen),
	       msg->cm_hops, req->clm_lockinfo.lockid);
	msg->cm_hops = 0;
	msg->cm_port = PORT_BOUNCE;
	return msg->cm_node;
    }


    cached = (DLM_NORMAL == rl_lookup( & rl_entry ));
    if ( cached 
	 && (rl_entry.rl_flags & RLINFO_DIRECTORY) 
	 && clm_node_active( rl_entry.rl_node ) )
    {
	/* If we think this request's resource is mastered here and it's got
	 * more than a few hops on it, then it's likely going to spin till the
	 * cows come home, because the resource is *probably* migrating to a
	 * different node.  So stash it until we are updated with the new
	 * location of the resource.
	 */

	if ((msg->cm_hops % 10) > 3) 
	{
	    msg->cm_port = PORT_BOUNCE;
	    DBPRINTF((DBG_CONFIG|DBG_DIRECT|DBG_MIGR,
		      "dir - bounce (%s) 0x%x back to %d from %d",
		      printbin(rl_entry.rl_name, rl_entry.rl_namelen),
		      req->clm_lockinfo.lockid, msg->cm_node,
		      rl_entry.rl_node));
	    return TRUE;
	}

	DBPRINTF((DBG_DIRECT, "dir - returning %d", rl_entry.rl_node));

	msg->cm_node = rl_entry.rl_node;
	msg->cm_port = PORT_MSTR;
	DBPRINTF((DBG_DIRECT,"Redirecting request to master on %d",
		  rl_entry.rl_node));

	STAMP(req);
	PRINTTIME(req);
	return rl_entry.rl_node;
    }

	/* here if no existing directory entry for the resource */
	DBPRINTF((DBG_DIRECT, "dir - not found"));
	switch (req->clm_type) {
	case DLM_LOCK:
		if (req->clm_locktype == DLM_RES_VMS &&
				req->clm_lockinfo.lk_flags & LKM_CONVERT) {
			bsdlog(LOG_ERR, MSGSTR(DROP_REQUEST_CONV,
				"Master convert request for %d:%s not\n found "
				"by directory server.  Dropping request."),
				rl_entry.rl_namelen, printbin(rl_entry.rl_name,
				rl_entry.rl_namelen));
			return 0;
		}
		rl_entry.rl_node = req->clm_lockinfo.site;
		break;
	case DLM_REGISTER:
		rl_entry.rl_node = req->clm_register.site;
		break;
	case DLM_UNLOCK:
		bsdlog(LOG_ERR, MSGSTR(DROP_REQUEST_UNLOCK,
			"Master unlock request for %d:%s not found by "
			"directory server.  Dropping request."),
			rl_entry.rl_namelen, printbin(rl_entry.rl_name,
			rl_entry.rl_namelen));
		goto missing_unlock;
	case DLM_PURGELOCKS:
		bsdlog(LOG_ERR, 
			"Master purge request shouldn't be in "
			"directory server.  Dropping request.");
		return 0;
	case DLM_RMIGR_OP:
	case DLM_STATS:
		bsdlog(LOG_ERR,
		       "Resource's Master Migrating");
		/*
		 * request is not going to be re-directed.  call cccp_msg_duplicate()
		 * to insure that we have a writeable copy of the message,
		 * reset the req pointer and enqueue the response back to
		 * the secondary.
		 */
		msg = cccp_msg_duplicate( msg );
		req = msg->cm_data;
		req->clm_direction = PTI_RESPONSE;
		clm_queue_respmsg(msg);
		req->clm_status = DLM_BADRESOURCE;

		return 0;

		break;
	default:
		/* XXX fix msgstr */
		bsdlog(LOG_CRIT,
			"Invalid trans type=%d for directory create",
			req->clm_type);
		abort("dir_proc_request 5");
	}
	rl_entry.rl_flags = RLINFO_DIRECTORY;

	/*
	 * cached should be set to true if we found a cache entry that was
	 * not a directory entry - not likely to happen, but if directory
	 * information has been shifting around because of reconfigs, it's
	 * worth checking for.  We can't create an existing entry, so modify it.
	 */
	if (cached) {
		if (DLM_NORMAL != (ret = rl_modify(&rl_entry))) {
			/* couldn't modify it either.  */
			/* XXX fix msgstr */
			bsdlog(LOG_CRIT,
				"dir_proc_request: rl_modify failed ret = %d",
				ret);
			abort("dir_proc_request 6");
		}
	} else {
		if (DLM_NORMAL != (ret = rl_create(&rl_entry))) {
			/* couldn't create it either.  damn.  */
			/* XXX fix msgstr */
			bsdlog(LOG_CRIT,
				"dir_proc_request: rl_create failed ret = %d",
				ret);
			abort("dir_proc_request 7");
		}
	}

	DBPRINTF((DBG_DIRECT, "dir - returning %d", rl_entry.rl_node));

	msg->cm_node = rl_entry.rl_node;
	msg->cm_port = PORT_MSTR_CR;
	DBPRINTF((DBG_DIRECT,"Redirecting request to master/create on %d",
		rl_entry.rl_node));

	STAMP(req);
	PRINTTIME(req);
	return rl_entry.rl_node;

missing_unlock:
	/*
	 * We've got an unlock request that we can't find any directory
	 * information for.  So it probably has been purged before we
	 * could unlock it.  The request will still be on the originator's
	 * inflight queue, so queue a harmless response to free inflight
	 * request.
	 */
	msg = cccp_msg_duplicate( msg );
	req = msg->cm_data;
	req->clm_direction = PTI_RESPONSE;
	clm_queue_respmsg(msg);
	req->clm_status = DLM_VOID;

	return 0;
}

/*
 * Name
 *	dir_read_request
 * Description
 * 	This function is called by the communications layer when a directory
 * 	request is read on a directory node.  It calls the dir_proc_request
 *	function with the transaction structure pointer that is contained
 *	in the message structure.  The function then reassigns the destination
 *	field of the message with the value returned from the dir_proc_request
 *	function, changes the port to PORT_MSTR and returns the message to
 *	the communications layer so it will be forwarded to the master.
 * Arguments
 *	msg - response message structure
 * Returns
 *	The function should always return a non-zero value so the request
 *	will be redirected to the master node for the resource.
 * Global Data Affected
 * 	None
 */
int
dir_read_request(clm_msg_t *msg)
{
#ifndef __linux__
	clm_trans_t	*req = (clm_trans_t *)msg->cm_data;
#endif /* !__linux__ */
	int			ret;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_DIR_READ,
                req->clm_pid, req->clm_sequence, msg, 0x00);

	/* Massage incoming message into readable form */
	clm_vers_rcv(msg);

	acquire_lock( LK_ENTRY, kproc_pid );
	ret = dir_proc_request(msg);
	release_lock( LK_ENTRY, kproc_pid );

	/* Reset incoming message to original form */
	clm_vers_reset(msg);

	TRCHKGT(CLLOCKD_HKID | T_EXIT_FUNC, T_DIR_READ,
                ret, req->clm_sequence, msg, 0x00);

	CLM_ASSERT((0 == ret || clm_node_active(msg->cm_node)),
			"redirect to invalid node");
	return  ret;
}


/*
 * Name
 *	pop_up
 * Description
 *	Pop a directory update off of the hold stack a return a pointer to it.
 * Arguments
 *	None
 * Returns
 *	Pointer to the top directory update if there is one.
 * Global Data Affected
 * 	None
 */
static rl_update_t *
pop_up(void)
{
    up_q_t      * curr;
    rl_update_t	* updt_req = NULL;


    curr = (up_q_t *) dlm_list_rem_head( & hold_up );
    if ( curr != NULL )
    {
	updt_req = curr->up_date;
	CLM_FREE( curr );
    }


    return updt_req;
}


/*
 * Name
 *	push_up
 * Description
 *	Take a directory update and put it on a hold queue until it is safe
 *	to send it.
 * Arguments
 *	updt_req - directory update request
 *		Node entry should be non-zero for an update of the entry.
 *		A node entry of 0 indicates that the entry should be removed.
 * Returns
 *	Nothing
 * Global Data Affected
 * 	RLDB
 */
static void
push_up( rl_update_t * updt_req )
{
    up_q_t * curr;

    curr = (up_q_t *) CLM_MALLOC( sizeof( *curr ) );
    if ( curr == NULL )
	abort("Malloc error in push_up");
	
    curr->up_date = updt_req;
    dlm_list_add_head( & hold_up, (dlm_listnode_t *) curr );
    return;
}

/*
 * Name
 *	hold_update
 * Description
 *	Take a directory update and put it on a hold queue until it is safe
 *	to send it.
 * Arguments
 *	updt_req - directory update request
 *		Node entry should be non-zero for an update of the entry.
 *		A node entry of 0 indicates that the entry should be removed.
 * Returns
 *	Nothing
 * Global Data Affected
 * 	RLDB
 */
static void
hold_update( const rl_update_t * updt_req )
{
    rl_update_t * save_up;

    save_up = CLM_MALLOC( sizeof( *save_up ) );
    if ( save_up == NULL )
    {
	abort("Malloc error in rl_update_t");
    }


    (void) memcpy( save_up, updt_req, sizeof( *save_up ) );
    push_up(save_up);
}


/*
 * Name
 *	dir_proc_update
 * Description
 *	Use the RLDB interface to update directory information for the
 *	resource specified in the directory update structure.
 * Arguments
 *	updt_req - directory update request
 *		Node entry should be non-zero for an update of the entry.
 *		A node entry of 0 indicates that the entry should be removed.
 * Returns
 *	0 for success
 * Global Data Affected
 * 	RLDB
 */
static int
dir_proc_update(rl_update_t *updt_req, int prio)
{
	rl_info_t	rl_entry;	/* RLDB entry structure */
	int		ret = 0;
	int		dir_node;
#ifdef DEBUG
	int dbg_code ;
	char *xstr ;
#endif

	rl_entry.rl_name = &updt_req->rl_name[0];
	rl_entry.rl_namelen = updt_req->rl_namelen;
	rl_entry.rl_type = updt_req->rl_type;

	/* check update request to be certain that it came to the right node */
	dir_node = clm_direct(rl_entry.rl_name, rl_entry.rl_namelen);
	if (dir_node != OurSiteID) {
		DBPRINTF((DBG_DIRECT, "%s: (%s)->%d Not the directory", __FUNCTION__,
				printbin(rl_entry.rl_name, rl_entry.rl_namelen),
				updt_req->rl_node));
		return dir_node;
	}

#if defined(DEBUG)
	if (prio == MIGAPRI || !updt_req->rl_node)
	{
	    dbg_code = DBG_MIGR2;
	    xstr = "MIGA | self" ;
	}
	else if ( ! clm_run_rcv())
	{
	    dbg_code = DBG_CONFIG ;
	    xstr = "!run_rcv" ;
	}
	else
	{
	    dbg_code = DBG_DIRECT ;
	    xstr = "dflt" ;
	}
	DBPRINTF((dbg_code, "[%s:%s] %s (%s) to %d mp=%d p=%d",
		  current->comm, __FUNCTION__, xstr,
		  printbin(rl_entry.rl_name, rl_entry.rl_namelen),
		  updt_req->rl_node, prio, 
		  dlm_recov_state_machine->curr->name));
#endif

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_PROC_UPDATE, rl_entry.rl_namelen, 
			((char *)(rl_entry.rl_name))[0],
			updt_req->rl_node, 0x00);

	if (0 != updt_req->rl_node) {
		rl_entry.rl_node = updt_req->rl_node;
		rl_entry.rl_flags = RLINFO_DIRECTORY;

		/*
		 * If resource is a local lock (in reconfig) set directory flag
		 */
		if (updt_req->rl_flags & RSRC_LOCAL)
		{
		        rl_entry.rl_flags |= RLINFO_LOCAL;
		}
	       
#if defined(DEBUG)
		if (prio == MIGAPRI) {
			CLM_ASSERT((DLM_NORMAL == rl_lookup(&rl_entry)),
					"No dir entry for migration");
			CLM_ASSERT((rl_entry.rl_flags & RLINFO_DIRECTORY),
					"Mig RLDB entry not directory");

			/* reset in case rl_lookup changed rl_entry */
			rl_entry.rl_name = &updt_req->rl_name[0];
			rl_entry.rl_namelen = updt_req->rl_namelen;
			rl_entry.rl_type = updt_req->rl_type;
			rl_entry.rl_node = updt_req->rl_node;
			rl_entry.rl_flags = RLINFO_DIRECTORY;
		} else if (prio == RESRPRI) {
			/* reset in case rl_lookup changed rl_entry */
			rl_entry.rl_name = &updt_req->rl_name[0];
			rl_entry.rl_namelen = updt_req->rl_namelen;
			rl_entry.rl_type = updt_req->rl_type;
			rl_entry.rl_node = updt_req->rl_node;
			rl_entry.rl_flags = RLINFO_DIRECTORY;
		}
#endif
		if (DLM_NORMAL != rl_modify(&rl_entry)) {
			/*
			 * here if no existing entry for the resource.  If
			 * the prio is non-zero, then we're rebuilding the
			 * directory, so create the entry
			 */
			if (prio <= MIGAPRI ||
				DLM_NORMAL != (ret = rl_create(&rl_entry))) {
				bsdlog(LOG_CRIT,
					"%s: not found r=%d p=%d", __FUNCTION__,
					ret, prio);
				abort("dir_proc_update: update error");
			}
		}
	} else {
#if defined(DEBUG)
		ret = rl_lookup(&rl_entry);
		CLM_ASSERT((ret == DLM_NORMAL), "delete not found");
		CLM_ASSERT((rl_entry.rl_flags & RLINFO_DIRECTORY),
				"Delete non-directory");
		CLM_ASSERT((rl_entry.rl_flags < 16),
				"dir_proc_update: bad flag value");
#endif
		if (DLM_NORMAL != (ret = rl_delete(&rl_entry))) {
			/* here if no existing entry for the resource */
			bsdlog(LOG_CRIT,
				"%s: (%s) not found ret = %d", __FUNCTION__,
				printbin(rl_entry.rl_name,
				rl_entry.rl_namelen), ret);
			abort("dir_proc_update: delete not found");
		}
	}

	return 0;
}

/*
 * Name
 *	dir_read_update
 * Description
 * 	This function is called by the communications layer when a directory
 * 	update request is read on a directory node.  It calls the
 *	dir_proc_update function with the update structure pointer that is
 *	contained in the message structure.
 *	This function is called while holding the serialization kernel lock
 *	LK_ENTRY.
 * Arguments
 *	msg - directory update message structure
 * Returns
 *	This function should always return zero because update requests
 *	should not be redirected.
 * Global Data Affected
 * 	None
 */
int
dir_read_update(clm_msg_t *msg)
{
	rl_update_t	*req = (rl_update_t *)msg->cm_data;
	int		ret;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_DIRUPDT_READ,
                msg, 0x00, 0x00, 0x00);

    /* 'ret' is used as the return value and as a flag.  If any of the
     * conditionals set 'ret' for a return value, then dir_proc_update will
     * NOT be called.  If 'ret' is left unmodified, then dir_proc_update
     * will be called, and its return value will be used for 'ret'.
     * 
     * The reason for this bit of strangeness is so that the routine can have
     * only a signle exit point.  Otherwise, the locking/unlocking would be
     * a bit more complex and quite a bit more error prone.
     */

    ret = -1;


    acquire_lock( LK_ENTRY, kproc_pid );

    /* If we get a directory update request while we are in the process of
     * reading in a direct dump (when we first come up), then someone has
     * somehow jumped the gun.  Bounce the request.  The comm layer should
     * queue the request until the priority level drops.
     */

    if (clm_in_reconfig())
    {
	/* drop any updates received in rebuild except dir deletes */
	if (msg->cm_prio == MIGAPRI)
	{
	    DBPRINTF((DBG_CONFIG, "Dropping update (%s)->%d mp=%d p=%d\n",
		      printbin(req->rl_name, req->rl_namelen),
		      req->rl_node, msg->cm_prio, MIGAPRI)) ;
	    ret = 0;
	}
	else if (req->rl_node == 0 && clmr_ok_to_del_dir())
	{
	    /* Handle situation where a dir delete is received
	     * before we've gotten all dir updates
	     */

	    DBPRINTF((DBG_CONFIG,
		      "premature dir delete (%s)->%d frm=%d mp=%d",
		      printbin(req->rl_name, req->rl_namelen),
		      req->rl_node, msg->cm_node, msg->cm_prio));
	    hold_update(req);
	    ret = 0;
	}
    }
    else if (clm_in_delivery() && msg->cm_node != OurSiteID)
    {
	/* Handle situation where a migration update is received
	 * while we're shuffling directory entries to a new node
	 */

	DBPRINTF((DBG_CONFIG, "update in integration (%s)->%d frm=%d mp=%d",
		  printbin(req->rl_name, req->rl_namelen),
		  req->rl_node, msg->cm_node, msg->cm_prio));
	msg->cm_hops = 0;
	ret = msg->cm_node;
    }

    if ( ret == -1 )
    {
	ret = dir_proc_update(req, msg->cm_prio);
	if ( ret != 0 )
	{
	    msg->cm_node = ret;
	    CLM_ASSERT( clm_node_active( msg->cm_node ),
		        "redirect to invalid node" );
	}
    }

    release_lock( LK_ENTRY, kproc_pid );

    TRCHKGT(CLLOCKD_HKID | T_EXIT_FUNC, T_DIRUPDT_READ,
	    msg, 0x00, 0x00, 0x00);

    return ret;
}


/*
 * Name
 *	resend_updates
 * Description
 * 	This function is called when a node in rebuild reaches the point
 *	where it can accept directory updates.  Any updates that have been
 *	placed on the hold queue will be popped up and handled as they would
 * 	have been originally.
 * Arguments
 *	None
 * Returns
 *	Nothing
 * Global Data Affected
 * 	RLDB
 */
void
resend_updates(void)
{
    clm_msg_t	*msg;
    int		dir_node;
    int		ret;
    rl_update_t	*updtp;


    while (NULL != (updtp = pop_up())) 
    {
	ret = dir_proc_update(updtp, RESRPRI);
	if (0 == ret) {
	    CLM_FREE(updtp);
	}
	else
	{
	    /* Figure out the node that contains the directory information
	     * for this resource name.
	     */

	    dir_node = clm_direct(updtp->rl_name, updtp->rl_namelen);
	    if (dir_node <= 0) {
		bsdlog(LOG_CRIT, MSGSTR(DIR_NODE_ERR,
					"Unable to determine directory node"));
		abort("resend_updates");
	    }

	    msg = cccp_msg_create( PORT_DIRUPDT, RESRPRI, dir_node, 0,
				  sizeof( rl_update_t ), dir_update_complete,
				  updtp, NULL );
	    if ( msg == NULL )
	    {
		bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC7,
					"Malloc error in resend_updates()"));
		abort("resend_updates");
	    }

	    updtp->rl_resptr = NULL;
	    ret = cccp_send(msg);
	    DBPRINTF((DBG_DIRECT,
		      "resend_updates: cccp_send to %d returned %d 0x%x 0x%x",
		      dir_node, ret, msg, msg->cm_data));
	}
    }

    return;
}


/*
 * Name
 *	update_directory
 * Description
 * 	This function is used by the migration module and by the master
 * 	module to either delete or change a directory entry.  This function
 *	should only be called on the node that is the master for the
 *	resource whose directory entry is to be changed.
 * Arguments
 *	rp - pointer to the resource structure
 *	node - node id of the resource master.  A node id of 0 indicates
 *		that the directory entry should be removed.
 * Returns
 *	Nothing
 * Global Data Affected
 * 	RLDB
 */
int
update_directory(struct resource *rp, int site, int prio, clm_msg_callback_t * cb)
{
        clm_msg_t	*msg;
	int		dir_node;
	int		ret;
	rl_update_t	*updtp;

	DBPRINTF((DBG_DIRECT,
			"Dir update: (%s) to site=%d", 
			printbin(rp->name, rp->namelen), site));
#if defined(DEBUG)
	if ( ! clm_run_rcv())
	{
		DBPRINTF((DBG_RECOV,
			"Dir update: (%s:%x) to site=%d",
			printbin(rp->name, rp->namelen), (rp->rh.rh&0xff),
			site));
		CLM_ASSERT((site), "Deleting dir entry in rebuild?!?!?\n");
	}
#endif
	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_UPDATE_DIR, rp->namelen, 
			rp->name[0], site, rp->master);

	/*
	 * Local Lock: Temporary Fix, there is no directory to update
	 * for a resource created with LKM_LOCAL flag so return OK
	 */
	if ((rp->rsrc_flags & RSRC_LOCAL) && ( ! clm_in_reconfig()))
	{
	        DBPRINTF((DBG_LOCAL,
			   "update_directory: RSRC_LOCAL set so returning\n"));
		return (0);
	}
	
	/*
	 * Figure out the node that contains the directory information for
	 * this resource name.
	 */
	dir_node = clm_direct(rp->name, rp->namelen);
	if (dir_node <= 0) {
		bsdlog(LOG_CRIT, MSGSTR(DIR_NODE_ERR,
				"Unable to determine directory node"));
		abort("update_directory");
	}

	/* fill in the message structure to be sent to the directory node */
	updtp = (rl_update_t *)CLM_MALLOC(sizeof(rl_update_t));
	if (NULL == updtp) {
		bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC7,
				"Malloc error in update_directory()"));
		abort("update_directory");
	}
    (void) memset( updtp, 0, sizeof(*updtp) );

	resname_bcopy(rp->name, &updtp->rl_name[0], rp->namelen);
	updtp->rl_namelen = rp->namelen;
	updtp->rl_flags = rp->rsrc_flags;
	updtp->rl_type = rp->type;
	updtp->rl_node = site;
	updtp->rl_resptr = rp;

    msg = cccp_msg_create( PORT_DIRUPDT, prio, dir_node, 0, 
			   sizeof( rl_update_t ), 
			   (cb != NULL) ? cb : dir_update_complete,
			   updtp, NULL );
    if ( msg == NULL ) 
    {
	bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC7,
				"Malloc error in update_directory()"));
	abort("update_directory");
    }

	ret = cccp_send(msg);
	DBPRINTF((DBG_DIRECT,
		"update_dir: cccp_send to %d returned %d 0x%x 0x%x", dir_node,
		ret, msg, msg->cm_data));
#if defined(DEBUG)
	if ( ! clm_run_rcv())
	{
		DBPRINTF((DBG_RECOV,
			"update_dir: cccp_send to %d returned %d 0x%x 0x%x",
			dir_node, ret, msg, msg->cm_data));
	}
#endif
	return ret;
}
