/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_group.c,v 1.15 2001/11/07 23:56:19 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <asm/uaccess.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "cltrc.h"
#include "clm_data.h"
#include "dlm_kernel.h"
#include "ast.h"
#include "clm_locksw.h"
#include "dlm_version.h"

#include <iterator.h>

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/*
 * How much do we bump the group array each time we have to increase its
 * size?  
 */
#define GROUP_INCR	10

/* arbitrary limit */
#define MAX_GROUPS	2040

/* get index from group id */
#define	ID2NDX(id)	((id) & 0xfff)

/* get site from group id */
#define	ID2SITE(id)	((id) >> 16)

/* generate group id from site, gen and index */
#define	GROUPID(site, gen, ndx)	(((site) << 16) | ((gen) << 12) | (ndx))

/* number of gen values and mask */
#define	GENMAX		(1 << 4)
#define	GENMASK		(GENMAX - 1)

/*
 *	Group id is:
 *		bit 0-11	= index		0x00000fff
 *		bit 12-15	= gen value	0x0000f000
 *		bit 16-31	= site		0xffff0000
 */

static int      	maxgroups = 0;	/* number of groups allocated */
static int      	ngroups = 0;	/* number of groups in use */
static int      	grp_gen = 0;	/* unique seq # or'ed into group id */
static clm_group_t	*groups;	/* group array */

static void res_grpify( struct resource * rp, clm_grpify_t * grp_msg );

/*
 * Name
 * 	clm_group_init 
 * Description
 * 	Initializes group variables
 * Arguments
 *	None
 * Returns
 *	< 0 for resource failure
 * 	0 for success
 * Global Data Affected
 *	maxgroups
 *	ngroups
 *	groups
 */
void
clm_group_init(void)
{
	maxgroups = 0;
	ngroups = 0;
	groups = NULL;
	grp_gen = 0 ;
	/* gen value doesn't have to be inited */
	/* but it -everything- -should- be (gwes) */
}


/*
 * Name
 * 	extend_groups 
 * Description
 * 	Extends the group array
 * Arguments
 *	None
 * Returns
 *	< 0 for resource failure
 * 	0 for success
 * Global Data Affected
 *	maxgroups
 *	groups
 */
int
extend_groups(void)
{
	int	old_size = maxgroups * sizeof(clm_group_t);
	int	new_size;

	/* let's not get ridiculous */
	if (maxgroups > MAX_GROUPS)
	{
	        bsdlog(LOG_ERR, 
		       "extend_groups: exceeded maximum number of groups");
		return -1;
	}

	/* set new max and new size */
	maxgroups += GROUP_INCR;
	new_size = maxgroups * sizeof(clm_group_t);

	/* extend the memory */
	groups = (clm_group_t *)REALLOC(groups, old_size, new_size);
	if (NULL == groups)
		return -1;

	return 0;
}


/*
 * Name
 *	res_grpify
 * Description
 * 	This function is called on each resource in the resource table
 * 	to change ownership of locks from a specified client to a group.
 * Arguments
 *	rp - resource pointer
 *	grp_msg - groupify structure pointer
 * Returns
 *	The function should always return a zero or the apply function will
 *	stop its processing.
 * Global Data Affected
 * 	None
 */
static void
res_grpify(struct resource *rp, clm_grpify_t *grp_msg)
{
    struct reclock	*a;
    int		touched = FALSE;
    clm_client_id_t clid;
    int i ;

    /* non-master copies only need to be changed on local node */
    if (rp->master && grp_msg->gi_site != OurSiteID)
	return;
	
    /* Use groupid as client id */
    set_client_id(clid, grp_msg->gi_gid);
	
    /* change ownership of granted, converting, waiting locks */
    for (i = RLQ_GRANT ; i < RLQ_MAX ; i++)
    {
	for ( EACH_DLM_LISTNODE( & rp->lockqs[i], a ) )
	{
	    if ( (a->lck.site == grp_msg->gi_site)
		 && (a->lck.pid == grp_msg->gi_pid) )
	    {
		/* update group client for local locks */
		if (a->lck.site == OurSiteID) 
		{
		    /* Delete lock from client list */
		    delete_client_reclock(a);
		    add_client_reclock( grp_msg->gi_site, &clid,
				        CLM_CLI_GROUP, a );
		}

		/* Change ownership to group id */
		a->lck.pid = grp_msg->gi_gid;
		touched = TRUE;
	    }
	}
    }

    /* non-zero return would break resource_table_apply() */
    return;
}


/*
 * Name
 *	grp_proc_request
 * Description
 * 	This function is called by the communications layer when a groupify
 * 	request is read on node.  This function is responsible for finding
 *	all locks on the local node owned by the client specified in the
 *	groupify message and reassociating them with the group.
 * Arguments
 *	msg - groupify message structure
 * Returns
 *	The function should always return a zero value so the request
 *	will not be redirected.
 * Global Data Affected
 * 	None
 */
int
grp_proc_request(clm_msg_t *msg)
{
    struct resource * res;
    iterator_t * iter;
    clm_grpify_t * grp_msg = (clm_grpify_t *)msg->cm_data;

    acquire_lock( LK_ENTRY, kproc_pid );
    iter = resource_table_iterator_create();
    while ( ! iterator_next( iter, & res ) )
    {
	res_grpify( res, grp_msg );
    }
    
    iterator_destroy( iter );
    release_lock( LK_ENTRY, kproc_pid );

    return  0;
}


/*
 * Name
 *	clm_groupify_confirm
 * Description
 *	This is the callback that should be called when a groupify
 *	message is received on the destination node.
 * Arguments
 *	msg - pointer to the message that has been received
 *	stat - status of the message
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
clm_groupify_confirm(clm_msg_t *msg, int stat)
{
	DBPRINTF((DBG_CONFIG, "Groupify message to node %d confirmed\n",
			msg->cm_node));

	/* check status */
	if (stat) {
		bsdlog(LOG_ERR, "Bad status %d in clm_groupify_confirm",
				stat);
	}

	/* free memory */
	if (msg) {
		if (msg->cm_data)
			CLM_FREE(msg->cm_data);
		cccp_msg_destroy( msg );
	}
}


#if defined(GROUP_ATTACH_ACQUIRES_ALL_PREVIOUSLY_OWNED_LOCKS)
/*
 * Name: clm_groupify
 *
 * Description:
 *	Sends notification to all active nodes that the specified client
 *	has joined a group.  This causes all locks on all active nodes
 *	that are owned by this client to become associated with the group.
 * Parameters:
 *	site	Site of the process
 *	pid	Pid of the process
 *	gid	Groupid of the group that the process has joined
 *
 * Return value:
 *	0 if successful
 *	-1 if error
 *
 * Side effects:
 *	None
 */
static int
clm_groupify(int site, int pid, int gid)
{
    unsigned       id;
    clm_msg_t    * msg;
    clm_grpify_t * grp_msg;
    iterator_t   * iter;
    int            status;


    status = 0;
    iter = dlm_nodeid_iterator_create();
    if ( iter == NULL )
    {
	bsdlog( LOG_CRIT, "[%s:%s] Unable to create node ID iterator\n",
	        current->comm, __FUNCTION__ );
	status = -1;
    }

    while ( (status == 0) && ! iterator_next( iter, & id ) )
    {
	if ( clm_node_active(id) )
	{
	    /* allocate groupify structure */
	    grp_msg = CLM_MALLOC( sizeof( clm_grpify_t ) );
	    if ( grp_msg == NULL) 
	    {
		bsdlog(LOG_CRIT, "%s [%d]: malloc(%d) failure\n",
		       __FILE__, __LINE__, sizeof(*msg));
		status = -1;
	    }
	    else
	    {
		grp_msg->gi_gid = gid;
		grp_msg->gi_site = site;
		grp_msg->gi_pid = pid;

		/* allocate message header */
		msg = cccp_msg_create( PORT_GROUP, BASEPRIO, id, CLMG_GROUPIFY,
				      sizeof( *grp_msg ), clm_groupify_confirm,
				      grp_msg, NULL );
		if ( msg == NULL)
		{
		    CLM_FREE( grp_msg );
		    bsdlog(LOG_CRIT, "%s [%d]: malloc(%d) failure\n",
			   __FILE__, __LINE__, sizeof(*msg));
		    status = -1;
		}
		else
		{
		    if ( cccp_send(msg) != CPERR_OK ) 
		    {
			bsdlog( LOG_CRIT, "%s: cccp_send(%d)\n", __FUNCTION__,
			        id );
		    }

		    DBPRINTF((DBG_GROUP, "Sent groupify message to %d\n", 
			      id));
		}
	    }
	}
    }


    iterator_destroy( iter );
    return( status );
}
#endif /* GROUP_ATTACH_ACQUIRES_ALL_PREVIOUSLY_OWNED_LOCKS */

/*
 * Name: grp_lookup
 *
 * Description:
 *	Returns the group id for a given pid if the pid is attached to a
 *	group.
 * Parameters:
 *	pid	Pid of the process
 *
 * Return value:
 *	Group id of the attached process
 *	0 if the process is not attached to any group
 *
 * Side effects:
 *	None
 */
static int
grp_lookup(int pid)
{
    int		i;
    grp_mem_t	*mptr;

    /* search member list of each active group for matching pid */
    for (i = 0; i < ngroups; i++) 
    {
	/* valid groups have non-zero group ids */
	if ( groups[i].grp_id != 0 )
	{
	    mptr = groups[i].grp_mbrs;
	    while (mptr) 
	    {
		if (mptr->gm_pid == pid)
		    return groups[i].grp_id;
		mptr = mptr->gm_next;
	    }
	}
    }

    /* no match */
    return 0;
}


/*
 * Name: grp_rmpid
 *
 * Description:
 *	Removes the specified process from the specified group.
 *
 * Parameters:
 *	gid	Group id
 *	pid	Pid of process to be removed
 *
 * Return value:
 *	CLM error code
 *
 * Side effects:
 *	None
 */
dlm_stats_t
grp_rmpid(int gid, int pid)
{
	int		idx = ID2NDX(gid);	/* group index */
	dlm_stats_t	stat = DLM_NORMAL;	/* return status */
	grp_mem_t	*mptr, *prev = NULL;	/* ptr to list member */
	clm_client_t    *cl;
	clm_client_id_t clid;

	/* validate group id */
	if (OurSiteID != ID2SITE(gid) || idx >= ngroups || 
						groups[idx].grp_id != gid) {
		abort("grp_rmpid: Invalid groupid");
		return DLM_IVGROUPID;
	}

	/* walk list of group members */
	mptr = groups[idx].grp_mbrs;
	while (mptr) {
		/* if we find the pid, deactivate the member */
		if (mptr->gm_pid == pid) {
			if (prev) {
				prev->gm_next = mptr->gm_next;
			} else {
				groups[idx].grp_mbrs = mptr->gm_next;
			}
			FREE(mptr);
			goto done;
		}

		prev = mptr;
		mptr = mptr->gm_next;
	}

	DBPRINTF((DBG_GROUP, "%s: Could not find pid %d in group %d",
              __FUNCTION__, pid, gid));
	stat = DLM_DENIED;

done:
	/* reset group id to 0 in pid's client record */
	set_client_id(clid, pid);
	cl = findclient (OurSiteID, &clid);	
	CLM_ASSERT( cl != NULL, "NULL client pointer" );
	cl->cli_groupid = 0;

	DBPRINTF((DBG_GROUP, "%s: Detach pid: %d from group: %d", 
              __FUNCTION__, pid, gid));

	if (groups[idx].grp_mbrs) {
		return stat;
	}

	/* here if we deleted last member of the group */

	DBPRINTF((DBG_GROUP, "%s: Purge group owned locks", __FUNCTION__));

	/* send a purge request for the group */

	internal_purge_request(OurSiteID, groups[idx].grp_id, PURGE_DEAD) ;

	/* clear groupid and flags to free up slot */
	groups[idx].grp_id = 0;
	groups[idx].grp_flags = 0;

	return stat;
}


/*
 * Name: grp_addpid
 *
 * Description:
 *	Adds the process specified by the pid to the group specified by the
 *	group id.
 * Parameters:
 *	gid	Group to add the process to
 *	pid	Pid of the process to be added
 *
 * Return value:
 *	CLM error code
 *
 * Side effects:
 *	None
 */
static dlm_stats_t
grp_addpid(int gid, int pid)
{
	int		idx = ID2NDX(gid);
	grp_mem_t	*mptr;
	clm_client_t	*cl;

	/* validate group id */
	if (OurSiteID != ID2SITE(gid) || idx >= ngroups ||
						groups[idx].grp_id != gid) {
		return DLM_IVGROUPID;
	}

	/* allocate member structure */
	mptr = (grp_mem_t *)MALLOC(sizeof(*mptr));
	if (NULL == mptr)
		return DLM_SYSERR;

	/* lookup client record */
	cl = findclientbypid(OurSiteID, pid);
	if (!cl) {
		FREE(mptr);
		return DLM_SYSERR;
	}

	/* set group id in client record */
	cl->cli_groupid = gid;

	/* set fields in group member structure - link into group list */
	mptr->gm_pid = pid;
	mptr->gm_next = groups[idx].grp_mbrs;
	groups[idx].grp_mbrs = mptr;

	DBPRINTF((DBG_GROUP, "%s: Attach pid: %d to group: %d",
              __FUNCTION__, pid, gid));

#if defined(GROUP_ATTACH_ACQUIRES_ALL_PREVIOUSLY_OWNED_LOCKS)
	/* notify all active nodes that locks owned by pid now belong to gid */
	clm_groupify(OurSiteID, pid, gid);
#endif

	return DLM_NORMAL;
}


/*
 * Name: grp_new
 *
 * Description:
 *	Creates a new group on the local node and returns the group id.
 *
 * Parameters:
 *	None
 *
 * Return value:
 *	CLM error code
 *
 * Side effects:
 *	None
 */
static int
grp_new(int flags)
{
    int	idx;
    clm_client_id_t clid;


    /* check for any old slots that are available */
    for ( idx = 0 ; (idx < ngroups) && (groups[idx].grp_id == 0) ; idx++ )
    {
	/* empty */
    }


    /* If an empty slot count not be found, extend the group table.
     */

    if ( idx == ngroups )
    {
	ngroups++;

	if (ngroups >= maxgroups && extend_groups() != 0) 
	{
	    ngroups--;
	    return 0;
	}
    }


    grp_gen++;
    grp_gen &= GENMASK;
    groups[idx].grp_id = GROUPID(OurSiteID, grp_gen, idx);
    groups[idx].grp_flags = flags;
    groups[idx].grp_mbrs = NULL;

    /* Use groupid as client id */
    set_client_id(clid, groups[idx].grp_id);
	
    /*
     * Group client structure needs to be explicitly added.  If the call to
     * addclient is unsuccessful, it is entirely possible that we have a
     * duplicate group id.  That is, a group with the same id is currently
     * being purged, so we have to wait.  Return 0 so that we will enter
     * the grace period, and wait for the purge to complete.
     */
    if ( ! addclient( OurSiteID, &clid, CLM_CLI_GROUP, 0, NULL ) ) 
    {
	groups[idx].grp_id = 0;
    }

    return groups[idx].grp_id;
}


/*
 * Name: clm_grp_op
 *
 * Description:
 *	Group utility function.  Used to perform all client group operations.
 *	NOTE:  gid needs to be copied in/out.
 * Parameters:
 *	grp_op		Group operation to perform
 *	usr_gid		Pointer to group id storage
 *	flags		Option flags
 *  pid         pid of client, must be sent in because of the
 *              Linux thread model (i.e, different threads are
 *              different pids!)
 *
 * Return value:
 *	CLM error code
 *
 * Side effects:
 *	Takes/releases global entry-point lock
 */
static int
clm_grp_op(int grp_op, int *usr_gid, int flags, int clpid)
{
	int		gid;
	int		tmpgid;
	dlm_stats_t	stat = DLM_NORMAL;

	DBPRINTF((DBG_CTI,"Copyin from 0x%x to 0x%x\n",usr_gid,&gid));
	if (copy_from_user(&gid, usr_gid, sizeof(gid))) {
	    bsdlog(LOG_ERR, MSGSTR(COPY_FAILED, "Copyin failed"));
	    return(DLM_SYSERR);
	}

	switch (grp_op) {
	case DLM_GRP_CREATE:
		/* are we already attached to a group? */
		gid = grp_lookup(clpid);
		if (0 != gid) {
			stat = DLM_DENIED;
			break;
		}

		/* 
		 * Need to create new group.  If the return code is 0, it is
		 * possible that a new group is being created before a current
		 * group with the same id is finished purging.  Enter the grace
		 * period phase until the purge is complete.
		 */
		gid = grp_new(flags);
		if (0 == gid) {
		        stat = DLM_DENIED_GRACE_PERIOD;
			break;
		}

		/* attach process to new group */
		stat = grp_addpid(gid, clpid);
		break;
	case DLM_GRP_ATTACH:
		/* are we already attached to a group? */
		tmpgid = grp_lookup(clpid);
		if (0 != tmpgid) {
			DBPRINTF((DBG_GROUP, 
				  "%s: Groupid lookup failed for pid = %d "
				  "during attach", __FUNCTION__, clpid));
			stat = DLM_DENIED;
			break;
		}

		/* attach process to existing group */
		stat = grp_addpid(gid, clpid);
		break;
	case DLM_GRP_DETACH:
		/* are we attached to a group? */
		gid = grp_lookup(clpid);
		if (0 == gid) {
			DBPRINTF((DBG_GROUP, 
				  "%s: Groupid lookup failed for pid = %d "
				  "during detach", __FUNCTION__, clpid));
			stat = DLM_DENIED;
			break;
		}

		/* detach this process from the group */
		stat = grp_rmpid(gid, clpid);
		break;
	default:
		bsdlog(LOG_ERR, "%s: Invalid op type", __FUNCTION__);
		stat = DLM_SYSERR;
	}

	DBPRINTF((DBG_CTI,"Copyout from 0x%x to 0x%x\n",&gid,usr_gid));
	if (copy_to_user(usr_gid, &gid, sizeof(gid))) {
		bsdlog(LOG_ERR,MSGSTR(COPY_OUT_FAILED, "Copyout failed"));
		return(DLM_SYSERR);
	}

	return stat;

}


/*
 * Name: k_clm_grp_op
 *
 * Description:
 *	A wrapper for clm_grp_op, to guarantee
 * correct locking.
 *
 * Parameters:
 *	grp_op		Group operation to perform
 *	gid		Pointer to group id storage
 *	flags		Option flags
 *  pid         Client pid
 *
 * Return value:
 *	CLM error code
 *
 * Side effects:
 *	Takes/releases global entry-point lock
 */
int
k_clm_group_op(int grp_op, int *gid, int flags, int clpid)
{
  int status;
  extern int kproc_pid;

  acquire_lock(LK_ENTRY, clpid);
  att_seg();
  if (kproc_pid) {
    if (clm_active) {
      status = clm_grp_op(grp_op, gid, flags, clpid);
    } else {
      status = DLM_DENIED_GRACE_PERIOD;
    }
  } else {
    status = DLM_NOLOCKMGR;
  }
  det_seg();
  release_lock(LK_ENTRY, clpid);

  return(status);
}

/****************************************************************************/
/* dlm_group_print - Print group table to a buffer
 *
 * Walk through the group table and print the values of each entry to the
 * provided buffer.
 *
 * RETURNS
 *
 * Total number of characters copied to the buffer.
 */

unsigned
dlm_group_print( char * buf, unsigned limit )
{
    grp_mem_t * mptr;
    unsigned    bytes;
    unsigned    i;


    bytes = snprintf( buf, limit, "Group ID | Index | Member PIDs\n" );
    if ( groups != NULL )
    {
	for ( i = 0 ; (i < ngroups) && (bytes < limit) ; i++ )
	{
	    if ( groups[i].grp_id != 0 )
	    {
		bytes += snprintf( & buf[ bytes ], limit - bytes, 
				   "%08X | %5u |", groups[ i ].grp_id, i );

		for ( mptr = groups[i].grp_mbrs; mptr; mptr = mptr->gm_next )
		{
		    bytes += snprintf( & buf[ bytes ], limit - bytes,
				       " %u", mptr->gm_pid );
		}

		bytes += snprintf( & buf[ bytes ], limit - bytes, "\n" );
	    }
	}
    }

    return( bytes );
}

