/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_heap.c,v 1.1.1.1 2001/02/14 20:57:51 wombat Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: 
 *		clm_heap_init
 *		clm_delete_heap
 *		att_seg
 *		det_seg
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/types.h>
#include <linux/errno.h>

#include "dlm.h"
#include "ast.h"
#include "clmstructs.h"
#include "dlm_kernel.h"

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/*
 * Refer to dlm_kernel.c for additional comments about the
 * heap support.  It is similar to the cache/slab support in
 * Linux, but, different enough to not translate directly.
 * For now, we will essentially gut this file, but keep the
 * code just ifdef'ed out.  Eventually we will use it to help
 * implement private cache support under Linux, but initially
 * it is essentially empty.
 */
/* ============================================================ */

#define LOCK_SEG_ADDR 0x30000000

#ifndef __linux__
/* file globals */
static vmid_t		vmid;
static vmhandle_t	vmhandle;
static caddr_t 		eff_addr;
static caddr_t 		lock_seg_addr;

/* extension globals */
heapaddr_t	lock_heap;
extern void	*kernel_heap;
#endif /* !__linux__ */

/* function prototypes */
void	att_seg(void);
void	det_seg(void);

/*
 * Function:
 *      clm_heap_init
 *
 * Description:
 *      Create a new virtual memory segment and initialize the
 *	heap for use by lockmanager client procs
 *
 * Arguments:
 *	NONE
 *
 * Return value:
 *      NONE
 *
 */
void
clm_heap_init(void)
{
#ifdef __linux__
  bsdlog(LOG_DEBUG, "Empty clm_init_heap for now.\n");
#else
    int		exit_code;
    struct heap	*da_heap;

#if defined(USE_KSEG)
    /* DEBUG - use kernel segment instead of our own */
    lock_heap = kernel_heap;
    return;
#endif

    /* Create segment of size SEGSIZE and get handle to it */
    exit_code = vms_create(&vmid, V_WORKING, 0, SEGSIZE, SEGSIZE, 0);
    if (0 != exit_code) {
	bsdlog(LOG_ERR, "Segment creation failed with exit code %d. " 
		 "Switching to kernel heap\n", exit_code);
	lock_heap = kernel_heap;
	return;
    }

    vmhandle = vm_handle(vmid, 1);

    /* Make segment part of cllockd address space */ 
    eff_addr = vm_att(vmhandle, 0);

    /* Save segment address for client checking */
    lock_seg_addr = eff_addr;

    /* Init the heap */
    lock_heap = init_heap((caddr_t)lock_seg_addr, SEGSIZE, NULL);

    /* pin the heap structure */
    da_heap = (struct heap *)lock_heap;
    exit_code = pin(da_heap, sizeof(*da_heap));
    if (exit_code != 0){
	    bsdlog(LOG_ERR, "Pin of lock_heap struct falied with "
		    "exit_code = %d\n", exit_code);
    }
    bsdlog(LOG_DEBUG, "Lock manager heap inited successfully");
#endif /* __linux__ */
}


/*
 * Function:
 *      clm_delete_heap
 *
 * Description:
 *      Detach, unpin the lock heap structure and delete the virtual
 *	memory segment.
 *
 * Arguments:
 *	NONE
 *
 * Return value:
 *      NONE
 *
 */
void
clm_delete_heap(void)
{
#ifdef __linux__
  bsdlog(LOG_DEBUG, "Empty clm_delete_heap for now.\n");
#else
    struct heap	*da_heap;
    int exit_code;

    /* check to be certain that we are using our own heap */
    if (lock_heap == kernel_heap) {
	return;
    }

    /* unpin the heap structure */
    da_heap = (struct heap *)lock_heap;
    exit_code = pin(da_heap, sizeof(*da_heap));
    if (exit_code != 0)
	    bsdlog(LOG_ERR, "Unpin of lock_heap struct falied "
		    "with exit_code = %d\n",exit_code);

    /* detach the segment */
    det_seg();

    /* delete the segment */
    exit_code = vms_delete(vmid);
    CLM_ASSERT((exit_code == 0), "Could delete VM segment.");

    /* zero out lock_heap - something may try to ref it after kproc exit */
    lock_heap = NULL;

    bsdlog(LOG_DEBUG, "Lock manager memory segment removed successfully");
#endif /* __linux__ */
}

/*
 * Function:
 *      att_seg
 *
 * Description:
 *      Used by lockmanager client procs to attach the lock segment.
 *	Aborts if segment not placed correctly.
 *
 * Arguments:
 *	NONE
 *
 * Return value:
 *      NONE
 *
 * Sets static 'eff_addr' to address where segment is placed.
 */
void
att_seg(void)
{
#ifdef __linux__  
  /*
   * Although the code should always just point to the kernel_heap
   * on Linux (at least for now) we keep the code and ifdef around
   * to both preserve history and as a signal in case we get to do
   * something in the future on Linux closer to the AIX multi-heap
   * capability.
   */

  return;
#else
    if (lock_heap == kernel_heap){
	return;
    }

    eff_addr = vm_att(vmhandle, 0);

    if (eff_addr != (caddr_t)lock_seg_addr){
	    bsdlog(LOG_ERR, "Can't attach virtual memory segment.  "
	    "Effective address = %x\n", eff_addr);
	    abort("att_seg");
    }
#endif /* __linux__ */    
}

/*
 * Function:
 *      det_seg
 *
 * Description:
 *      Detach the lock segment after user by client proc
 *
 * Arguments:
 *	NONE
 *
 * Return value:
 *      NONE
 *
 */
void
det_seg(void)
{
#ifdef __linux__  
  /*
   * Although the code should always just point to the kernel_heap
   * on Linux (at least for now) we keep the code and ifdef around
   * to both preserve history and as a signal in case we get to do
   * something in the future on Linux closer to the AIX multi-heap
   * capability.
   */

  return;
#else
    /* don't try to detach the segment if we aren't attached to it!!! */
    if (lock_heap == kernel_heap){
	return;
    }

    vm_det(eff_addr);
#endif /* __linux__ */
}

