/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_info.c,v 1.13 2001/10/23 22:21:44 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		clm_node_active
 *		clm_change_topology
 *		handle_new_topology
 *              print_node_list
 *              fix_low_high_version
 *              clm_version
 *              clm_low_version
 *              get_node
 *              initialize_new_instance
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/net.h>
#include <linux/inet.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "dlm_kernel.h"
#include "pti.h"
#include "cltrc.h"
#include "dlm_version.h"
#include "dlm_devmsg.h"
#include "clm_data.h"
#include "clm_info.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "hsm.h"
#include "dlm_recover.h"
#include "dlm_clust.h"

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */

extern void update_idhash( int * idmap );
extern void update_syncd( int * idmap );

void clm_change_topology(void) ;
static void fix_low_high_version(struct node_list *nl) ;
#ifdef DBPRINT
static void print_node_list(struct node_list *nl, const char *) ;
#endif


#ifdef __linux__
extern void cccp_node_down( uint id, uint node_failed );
extern void cccp_node_up( uint id );
extern void ifsDown ( uint id, ulong ifs_addr);
extern void cccp_change_topology(void);
#endif /* __linux__ */


/*
 * Function:
 *	clm_low_version
 * Description:
 *	Return the oldest version value of all active lock managers in the
 *	cluster
 * Arguments:
 *	None
 * Return value:
 *	Version in a single int.  16bit major, 16bit minor
 *	-1 if error
 * Side effects:
 *	NONE
 */
int
clm_low_version(void)
{
    return node_list->low_version;
}



/****************************************************************************/

static struct node_item *
get_node( struct node_list * inst, unsigned id )
{
    unsigned   i;
    struct node_item * node;


    node = NULL;
    for ( i = 0 ; i < inst->node_count ; i++ )
    {
	if ( inst->node[ i ].nodeid == id )
	{
	    node = & inst->node[ i ];
	    break;
	}
    }


    return( node );
}




/****************************************************************************/

static void
initialize_new_instance(
			 struct node_list * next_inst,
			 const struct topology_address_message * topology
		       )
{
    unsigned   i;
    unsigned   version;
    const MQ_DLM_USEADDR_t * node_addrs;


    /* Insert each of the nodes in the topology message in the new-node
     * list.
     */

    node_addrs = topology->addrs;

    (void) memset( next_inst, 0, sizeof( struct node_list ) );
    next_inst->low_version = 0x7fffffff;

    for ( i = 0 ; i < topology->n_nodes ; i++, node_addrs++ )
    {
	DBPRINTF((DBG_CONFIG, "   %d   %x.%x  %s\n",
		  node_addrs->nodeid, node_addrs->dlm_major,
		  node_addrs->dlm_minor,
		  in_ntoa( node_addrs->useaddr.s_addr ))) ;

	version = (topology->this_nodeid == node_addrs->nodeid)
	    ? DLM_VERSION
	    : (unsigned) (node_addrs->dlm_major << 16) 
		| node_addrs->dlm_minor;

	next_inst->node[ i ].nodeid = node_addrs->nodeid;
	next_inst->node[ i ].saddr = node_addrs->useaddr;
	next_inst->node[ i ].version = version;

	if ( version != 0 )
	{
	    if ( version < next_inst->low_version )
	    {
		next_inst->low_version = version;
	    }
	    
	    if ( version > next_inst->high_version )
	    {
		next_inst->high_version = version;
	    }
	}
    }

    next_inst->node_count = topology->n_nodes;
}




/*
 * Name
 *	handle_new_topology
 * Description
 *	on first call:
 *   initialize an array of nodeids to 0
 *   add all of the nodeids in the input array
 *   activate new configuration
 *   return
 *  on subsequent calls:
 *   initialize an array of nodeids to 0
 *   add all of the nodeids in the input array
 *   compare old and new topology call clms_node_del for entries in
 *   old table not present in new table, call clms_node_add for entries
 *   in new table not in old table.
 *
 *   compare old & new IP addresses
 *	== OK
 *	!= and both nonzero, OK - PROPAGATE IMMEDIATELY (address change)
 *      != and old zero, node is coming up, but do nothing (node selfstarts)
 *      != and new zero, node has gone down, SET BOTH to 0 to prevent
 *		further traffic
 *
 * Arguments
 *		pointer to an array of node information
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */

void
handle_new_topology( const struct topology_address_message * topology )
{
  int i,j;			/* loop control */
  ulong olda, newa ;	/* addresses for up/down check */
  int new_nodeid ;
  int old_nodeid ;
    struct node_item * node;

  DBPRINTF((DBG_CONFIG, "[%s:%s] node count %d  this nodeid %d  event "
	    "0x%x  msg ver %x\n",
	    current->comm, __FUNCTION__,
	    topology->n_nodes, topology->this_nodeid, topology->event,
	    topology->msg_version) );


    initialize_new_instance( new_node_list, topology );
#ifdef DBPRINT
    print_node_list(new_node_list, "new") ;
#endif


    /* The first time that the topology is changed there are some special
     * things that need to be done.  We need to set OutSiteID, and we need
     * to establish the initial topology setting (by calling
     * clm_change_topology).  We also need to send a CLM_MSG_NODE_INIT message
     * down into the recovery state-machine so that the first recovery
     * can begin.
     */

    if ( first_top_call )
    {
	struct rc_msg * m;

	first_top_call = FALSE;
	OurSiteID = topology->this_nodeid ;	/* also used as 'have top' flag */
	clm_change_topology() ;		/* make rec top active */

	m = rc_msg_alloc( CLM_MSG_NODE_INIT, 0, NULL );
	hsm_process_event( dlm_recov_state_machine, (hsm_msg_t *) m );

	return ;
    }


    /*
     * first, lets look for nodeids that have been dropped
     */
    /* check the whole table */
    for (j = 0 ; j < MAXNODES ; j++) 
    {
	old_nodeid = node_list->node[j].nodeid ;
	if ( ! old_nodeid)		/* if no one home */
	    continue;			/* don't process */

	node = get_node( new_node_list, old_nodeid );
	if ( node != NULL )
	{
	    /* Determine if a node went up or down by asking if it is
	     * accessable now and was it accessable before
	     */

	    newa = node->saddr.s_addr ;
	    olda = node_list->node[j].saddr.s_addr ;
	    if (newa != olda) 
	    {
		if ( olda == 0 )
		{
		    bsdlog( LOG_INFO, "[%s:%s] attempting to bring up node "
			    "%u, DLM is%s running\n",
			    current->comm, __FUNCTION__,
			    node->nodeid, clm_is_running() ? "" : " not" );

		    if (clm_is_running())
		    {
			/* We only want to update the topology if we are
			 * NOT in any kind of reconfig.  Adding a new node
			 * mid-rebuild will hose the rebuild process.
			 */

			CLM_ASSERT((old_nodeid != OurSiteID),
			       "attempt to reconnect to local node") ;

			/* IMMEDIATELY propagate access */
			node_list->node[j] = *node;
			fix_low_high_version(node_list) ;

#ifdef CONFIG_PROC_FS
			if ( node_list->node[j].proc_entry == NULL )
			{
			    node_list->node[j].proc_entry =
				dlm_proc_create_node_entry( node_list->node[ j ].nodeid );
			}
#endif /* CONFIG_PROC_FS */

			/* update CCCP connections */
			cccp_node_up( old_nodeid );
			/* node up handled by messages from that node */
		    }
		} 
		else if ( newa == 0 )
		{
		    /* version must also be 0 before node is down */
		    if ( (node->version == 0) ||
			old_nodeid != OurSiteID) 
		    {
			node_list->node[j].saddr.s_addr = 0 ;	/* KILL OLD */
			node_list->node[j].version = 0 ;
			fix_low_high_version(node_list) ;

			cccp_node_down( old_nodeid, TRUE );	/* TELL CCCP */
			clms_node_down(node_list->node[j].nodeid) ;
		    }
		}
		else
		{
		    /* make it visible immediately */
		    /* else address has changed */
		    /* first flush old conn */
		    ifsDown(old_nodeid, olda);

		    node_list->node[j].saddr.s_addr = newa ;
#ifndef __linux__
		    if (old_nodeid == OurSiteID)
			break ;		/* our node is always up */
		    openConnections() ;	/* then make new one */
#endif
		}
	    }
	}
	else
	{
	    clms_node_delete(old_nodeid) ;	/* tell the scheduler */
	}
    }
    

    /* second, lets look for new nodeids
     */

    for (i = 0; i<topology->n_nodes ; i++) 
    {
	new_nodeid = new_node_list->node[i].nodeid ;
	node = get_node( node_list, new_nodeid );
	if ( node == NULL )
	{
	    clms_node_add(new_nodeid) ;	/* tell the scheduler */
	}
    }

#ifdef DBPRINT
    print_node_list(node_list, "current") ;
#endif
}

/*
 * print_node_list
 *
 * debug print of struct node_list with a header string
 * conditional on compilation DBPRINT and dlm_debug DBG_CONFIG
 */
static void
print_node_list(struct node_list *nl, const char * attr)
{
#ifdef DBPRINT
    int i ;

    DBPRINTF((DBG_CONFIG, "  %s node list: nnodes %d  low_ver %x  high_ver %x\n",
		attr,
		nl->node_count,
		nl->low_version,
		nl->high_version)) ;
		
    for (i = 0 ; i < new_node_list->node_count ; i++)
	DBPRINTF((DBG_CONFIG, "   %u  %x  %s\n",
		nl->node[i].nodeid,
		nl->node[i].version,
		in_ntoa( nl->node[i].saddr.s_addr ))) ;
#endif
}

/*
 * fix_low_high_version
 *
 * scan node entries in struct node_list
 * determine highest and lowest clm versions present from valid nodes
 * update low_version and high_version structure header entries
 */
static void
fix_low_high_version(struct node_list *nl)
{
    unsigned   low_version;
    unsigned   high_version;
    unsigned   version;
    unsigned   i;


    high_version = 0;
    low_version = 0x7fffffff;
    
    for (i = 0 ; i < MAXNODES ; i++)
    {
	version = nl->node[ i ].version;

	if ( version != 0 )
	{
	    if ( version < low_version )
		low_version = version;

	    if ( version > high_version )
		high_version = version;
	}
    }

    nl->low_version = low_version;
    nl->high_version = high_version;
}

/*
 * name            clm_change_topology
 * purpose         change from prior topology to new topology
 * return values   none
 * side effects    none
 */

void
clm_change_topology(void)	/* make node_list point at new topology */
{
    struct node_list *temp;
    int i, j ;
    int idmap[MAXNODES];

    /* return if there is not topology update to propogate */
    if (0 == new_node_list->node_count)
	return;

    /* init map */
    for ( i = 0 ; i < MAXNODES ; i++ )
    {
	idmap[ i ] = -1;
    }


    /* Create a mapping of old node INDEX values to new node INDEX values
     * based on the node ID values.
     */

    for (i = 0 ; i < node_list->node_count ; i++) 
    {
	for (j = 0 ; j < new_node_list->node_count ; j++)
	{
	    if ( node_list->node[i].nodeid == new_node_list->node[j].nodeid )
	    {
		if ( new_node_list->node[j].version != 0 )
		{
		    idmap[j] = i;
		}

#ifdef CONFIG_PROC_FS
		new_node_list->node[j].proc_entry = node_list->node[i].proc_entry;
#endif /* CONFIG_PROC_FS */
	    }
	}
    }


#ifdef CONFIG_PROC_FS
    /* Make sure that all the node-level /proc entries exist.  At least on
     * 2.4.5 Linux kernels and before, the proc stuff is not very smart.  If
     * you try to create a proc entry that already exists, you'll end up with
     * TWO entries with the same name.  So, we try to be careful to make sure
     * that doesn't happen.
     */

    for ( j = 0 ; j < new_node_list->node_count ; j++ )
    {
	/* If the node is UP and there is not already a /proc entry for
	 * the node, create one.
	 */

	if ( (new_node_list->node[j].version != 0)
	    && (new_node_list->node[j].proc_entry == NULL) )
	{
	    new_node_list->node[ j ].proc_entry = 
		dlm_proc_create_node_entry( new_node_list->node[ j ].nodeid );
	}
    }
#endif /* CONFIG_PROC_FS */


    update_idhash(idmap);	/* correct for shifts in idhash table */
    update_syncd(idmap);	/* correct for shifts in syncd arrays */

    /*
     * swap node lists
     */
    temp = node_list;
    node_list = new_node_list;
    new_node_list = temp;

    (void) memset( temp, 0, sizeof( *temp ) );

    cccp_change_topology() ;	/* call comm layer to do the same again */
}


/*
 * Function:
 *	clm_version
 * Description:
 *	Given a node id, return the version of the lock manager running
 *	on that node.  If provided node id is 0, return local version.
 * Arguments:
 *	nodeid - Node id of the node whose lock manager version to return.
 * Return value:
 *	Version in a single int.  16bit major, 16bit minor
 *	-1 if error
 * Side effects:
 *	NONE
 */
int
clm_version(int nodeid)
{
    int ndx ;

    if ( ! clm_config_inited())
      return -1 ;

    ndx = NODEID_TO_INDEX(nodeid ? nodeid : OurSiteID) ;
    if (ndx < 0)
      return -1 ;
    
    return node_list->node[ndx].version ;
}

/*
 * Function:
 *	clm_node_active
 * Description:
 *	Given a node id, return TRUE if the lock manager on that node is
 *	active.  Otherwise, return FALSE.
 * Arguments:
 *	nodeid - Node id of the cluster to check for an active lock manager.
 * Return value:
 *	TRUE if the node is active, FALSE otherwise.
 * Side effects:
 *	NONE
 */
int
clm_node_active(int nodeid)
{
    int ndx ;

    ndx = NODEID_TO_INDEX(nodeid) ;
    if (ndx < 0)
      return 0 ;

#if defined(WICKED_VERBOSE_DEBUG)
    bsdlog(LOG_DEBUG,"clm_node_active: state is %d\n",
		DLM_STATE(&cdNode->cd_md[ndx]));
#endif

    return node_list->node[ndx].version != 0 ;
}




/****************************************************************************/
/* dlm_nodeid_next - Determine the next node ID in the cluster
 *
 * Searches for the active node in the cluste with the next numerically 
 * higher node ID.  This is performed using modulo arithmetic.  This is,
 * if nodes 2, 4, and 5 are the only nodes in the cluster, calling this
 * function with 5 as input will case 2 to be returned.
 *
 * RETURNS
 *
 * The node ID of the next active node in the cluster.  If the speicifed node
 * is the only active node in the cluster, its own node ID will be returned.
 * So, if node 6 is the only active node in the cluster and this function is
 * called with 6 as input, 6 will be returned.
 *
 * WARNING
 *
 * This function assumes that the input node ID is actually an active member
 * of the cluster.
 */

unsigned
dlm_nodeid_next( int me )
{
    unsigned   node_count;
    int        i;
    unsigned   node_id;
#ifdef DEBUG
    unsigned   count;
#endif

    node_count = clm_node_count();
#ifdef DEBUG
    count = node_count;
#endif
    CLM_ASSERT( node_count != 0, "DLM not yet initialized!" );
    CLM_ASSERT( me > 0, "Invalid initial node ID" );

    i = NODEID_TO_INDEX( me );
    CLM_ASSERT( i >= 0, "Invalid initial node ID" );

    do
    {
	i = (i + 1) % node_count;
	node_id = INDEX_TO_NODEID( i );

#ifdef DEBUG
	CLM_ASSERT( count != 0, "Too many loop iterations!" );
	count--;
#endif
    } while ( (node_id == 0) || ! clm_node_active( node_id ) );


    return( node_id );
}




/****************************************************************************/
/* dlm_node_get_proc_entry - Get the /proc entry for a node
 *
 * Gets the proc_dir_entry for the specified node ID.
 *
 * RETURNS
 *
 * The proc_dir_entry for the node or NULL if the node is unknown.
 */

#ifdef CONFIG_PROC_FS
struct proc_dir_entry *
dlm_node_get_proc_entry(
			 unsigned   node_id
		       )
{
    struct node_item * item;


    item = get_node( node_list, node_id );
    return( (item == NULL) ? NULL : item->proc_entry );
}
#endif /* CONFIG_PROC_FS */
