/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_libr.c,v 1.8 2001/10/10 23:19:14 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS:	MSGSTR
 *		clm_direct
 *		clm_direct_dirsend
 *		comm_retrans
 *		dump_msg
 *		downRetry
 *		pti_call
 *		pti_call_complete
 *		pti_call_reg
 *		pti_call_unix
 */

#ifndef lint
#endif

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/signal.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/errno.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "cltrc.h"
#include "clm_data.h"			/* external data definitions */
#include "dlm_kernel.h"
#include "dlm_version.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
extern clm_msg_t *inflight;
extern clm_msg_t *inflight_last;

/* clm_libr.c
 * consists of routines used for initialization, dumping, PTI
 */

/****************************************************************************/
/* dlm_pti_send - Send a pti request to a node
 * 
 */

int
dlm_pti_send( clm_trans_t * req, unsigned node, unsigned port )
{
    clm_msg_t  * msg;
    clm_msg_t ** save_next;
    clm_msg_t  * save_inf_lst;
    int          ret;


    CLM_ASSERT( (port == 0) || (port == PORT_MSTR) || (port == PORT_MSTR_CR),
	        "Invalid port for PTI request" );


    /* If the port is not set, then the caller had no idea which node in the
     * cluster is the master for the resource.  Calculate the directory node
     * and forward the request to the directory port of that node.
     */

    if ( port == 0 )
    {
	node = clm_direct( & req->clm_lockreq.nameinfo.name[0],
			   req->clm_lockreq.nameinfo.namelen );
	if ( node == 0 )
	{
	    bsdlog(LOG_CRIT,MSGSTR(DLM_DIRECT_NULL,
				   "clm_direct() returned NULL"));
	    return -1;
	}

	DBPRINTF((DBG_DIRECT, "[%s:%s] send dir request to %d", 
		  current->comm, __FUNCTION__, node));
	port = PORT_DIRECT;
    }


    ret = -1;
    msg = cccp_msg_create( port, 1, node, req->clm_type, sizeof( *req ),
			   pti_call_complete, req, NULL );
    if ( msg != NULL )
    {
	req->clm_direction = PTI_REQUEST;
	req->clm_sender = (void *) req;
	req->clm_vers = DLM_VERSION;


	/* Put the about-to-be-transmitted transaction to the end of inflight
	 * queue because if all of this happens locally, the response callback
	 * might try to match it before we even return from this cccp_send().
	 */

	req->clm_next = NULL;

	save_inf_lst = inflight_last;
	save_next = (inflight_last != NULL)
	    ? (clm_msg_t **) & ((clm_trans_t *) (inflight_last->cm_data))->clm_next
	    : (clm_msg_t **) & inflight;

	*save_next = msg;
	inflight_last = msg;

	ninflight++;
	DBPRINTF((DBG_PTI,"[%s:%s] Enqueued 0x%x on inflight queue", 
		  current->comm, __FUNCTION__, req));

	/* make any adjusments for version skew between lock managers */
	ret = clm_vers_snd(msg, clm_low_version());
	if (0 == ret) 
	{
	    ret = cccp_send(msg);
	    DBPRINTF((DBG_COMM, "[%s:%s] cccp_send returned %d (port/node -> "
		      "%u/%u)\n", current->comm, __FUNCTION__, ret,
		      msg->cm_port, msg->cm_node));
	}


	/* if the send failed, then we need to restore the inflight queue */
	if ( ret != CPERR_OK ) 
	{
	    *save_next = NULL;
	    inflight_last = save_inf_lst;
	    ninflight--;

	    cccp_msg_destroy( msg );
	    ret = -1;
	}
    }


    return( ret );
}


/****************************************************************************/
/* clm_direct - Determine which node is the directory master
 *
 * Determine which node in the cluster is the responsible for the directory
 * service of the resource.
 *
 * RETURNS
 *
 * The site ID of the directory master or zero on failure.  The only way that
 * this function can fail is if there are no nodes active in the cluster.
 */

unsigned
clm_direct(
	    /* Name of the resource.
	     */

	    const char * name,

	    /* Length of the resource name.
	     */

	    size_t namelen
	  )
{
    /* Fake it!  Calling clm_direct_dirsend with an invalid 'nodeup' value
     * will cause it to only consider active nodes, which is what we want
     * in this case.
     */

    return( clm_direct_dirsend( name, namelen, 0xffffffff ) );
}




/****************************************************************************/
/* clm_direct_dirsend - Determine which node is the directory master
 *
 * Determine which node in the cluster is the responsible for the directory
 * service of the resource.  This may include the node specified by 'nodeup'
 * even if that node is not an active DLM node.
 *
 * RETURNS
 *
 * The site ID of the directory master or zero on failure.  The only way that
 * this function can fail is if there are no nodes active in the cluster.
 */

unsigned
clm_direct_dirsend(
		    /* Name of the resource.
		     */

		    const char * name,

		    /* Length of the resource name.
		     */

		    size_t namelen,

		    /* Node ID of a new node to be considered up.
		     */

		    unsigned nodeup 
		  )
{
    unsigned   node;
    unsigned   maxnode;
    unsigned   i;
    unsigned   start;


    /* Determine the directory node by calling the hash function on the
     * resource name with the max node index as the modulus.
     */

    maxnode = clm_node_count();
    start = hashfunc( name, namelen, maxnode );
    for ( i = 0 ; i < maxnode ; i++ )
    {
	node = INDEX_TO_NODEID( (start + i) % maxnode );
	if ( clm_node_active( node ) 
	     || ((node == nodeup) && clm_node_reachable( node )) )
	{
	    return node;
	}
    }


    return( 0 );
}


/*
 * Name
 * 	pti_call_complete
 * Description
 *	This function is call by the communications layer when it has
 *  	received an acknowledgement for the outbound request.
 *	NOTE:  We do *NOT* want to free the transaction structure pointed
 *	to by the message structure here because it is enqueued on the
 *	inflight list and will be used and then freed after the response
 *	comes back.
 * Arguments
 *	msg - pointer to the message structure that was acknowledged
 *	stat - status returned
 * Returns
 *	Nothing
 * Global Data Affected
 *	NONE
 */
void
pti_call_complete(clm_msg_t	*msg, int	stat)
{
	clm_trans_t *req = (clm_trans_t *)msg->cm_data;

	CLM_ASSERT((msg->cm_node != 0), "Zero node in pti_call_complete");
	CLM_ASSERT(((ulong)msg->cm_private < 16),
			"pti_call_complete: msg on queue");

	/*
	 * set confirmed bit so further refs to this msg know that its send
	 * callback has been run.
	 */
	msg->cm_private = (void *)((ulong)msg->cm_private | MSG_CONFIRMED);

	/*
	 * This function should get called if we have an unacknowledged
	 * message that was sent to a node that has failed.  Check the status
	 * here to determine if the message needs to be resent.  HOWEVER,
	 * sometimes, we will get a node-failed acknowledgement for a message
	 * whose response has already been processed.  If this is the case,
	 * then the cm_private field of the message will be non-zero and we
	 * should *ignore* the failure status and free the message.
	 */
	if (CPERR_NODEDOWN == stat &&
				(~((ulong)msg->cm_private) & MSG_MATCHED)) {
		clm_trans_t	*req = msg->cm_data;

		DBPRINTF((DBG_CONFIG,"%s: %d message lost to failed node",
                  __FUNCTION__, msg->cm_type));

		switch (msg->cm_type) {
		case DLM_LOCK:
		case DLM_UNLOCK:
		case DLM_REGISTER:
		case DLM_SCN_OP:
		case DLM_RMIGR_OP:
		case DLM_STATS:
			break;
		case DLM_PURGELOCKS:
			/* best we can do is punt */
			bsdlog(LOG_CRIT,
				"%s: Node failed.  Drop purge request", __FUNCTION__);
			cccp_msg_destroy( msg );
			clm_free_trans(req);
			return;
		default:
			/* best we can do is punt */
			bsdlog(LOG_CRIT,
				"%s: Node failed.  Can't redirect type %ld",
				__FUNCTION__, msg->cm_type);
			cccp_msg_destroy( msg );
			clm_free_trans(req);
			return;
		}

		/*
		 * We are setting the destination node to 0 in case the node
		 * that failed happens (and believe me, it *has* happened)
		 * to come up before we finishing recovering from its failure.
		 * If we don't set this to 0, then when resend_inflight looks
		 * at this message and has to decide whether or not to resend
		 * it, it'll say "Oh, that node's up.  I don't have to resend
		 * this message" when, in fact, it does.  So there.
		 */
		msg->cm_node = 0;
		msg->cm_port = PORT_DIRECT;
		msg->cm_private = NULL;

		clm_queue_reqmsg(msg);
		return;

	} else if (stat && (~((ulong)msg->cm_private) & MSG_MATCHED)) {
		bsdlog(LOG_ERR, MSGSTR(PTI_CALL_ERR,
				"Error detected in pti_call_complete\n"));
		return;
	}

	/* we are purposely not freeing the req ptr - see func desc above */

	/*
	 * This is heinous, but it had to be done.  If private is set to
	 * MSG_BOUNCE, then this request is being held due to ongoing
	 * migration.  Don't free anything because it's going to be resent.
	 */
	if ((ulong)msg->cm_private & MSG_BOUNCE) {
		bounce_msg(msg);
		return;
	}

	/*
	 * More heinousness.  Check if bounce logic decided that the client
	 * died.  If so, dequeue its copy from the inflight queue and drop
	 * it on the floor.
	 */
	if ((ulong)msg->cm_private & MSG_BNCE_DONE) {
		DBPRINTF((DBG_CONFIG, "Request dropped request id=0x%x pid=%d",
			req->clm_lockinfo.lockid, req->clm_pid));
		req = match_request(req);
		clm_free_trans(req);
		return;
	}

	/*
	 * free the message structure - we check cm_private because here we
	 * use it like a flag.  The message structure either has to be freed
	 * by the match_request() or by this function, but we never know which
	 * function is going to be called first.  Whichever one *is* called
	 * first, sets the cm_private field of the message to be non-zero.
	 * So, if cm_private is not zero, then the other function has finished
	 * with the message and it can be safely freed.  Got it?
	 */
	if ((ulong)msg->cm_private & MSG_MATCHED) {
		cccp_msg_destroy( msg );
	}

	return;

}

/*
 * Name
 * 	downRetry
 * Description
 *	This function should be called with a pointer to a message that
 *  	failed in cccp_send because the destination node is down.  It
 *	tries to determine the directory node for the message and calls
 *	cccp_send with the new node and directory port.
 * Arguments
 *	msg - pointer to the message
 * Returns
 *	return code from cccp_send
 * Global Data Affected
 *	NONE
 */
int
downRetry( clm_msg_t * msg )
{
    rl_info_t     rl_entry;	/* RLDB entry structure */
    clm_trans_t	* req;
    int           dir_node;
    int           ret;


    DBPRINTF((DBG_CONFIG, "entering downRetry 0x%x hop=%d",
	      msg, msg->cm_hops));


    req = (clm_trans_t *) msg->cm_data;
    if ( ! dlm_stuff_rl_info( req, & rl_entry ) )
    {
	if ( req->clm_type != DLM_PURGELOCKS )
	{
	    bsdlog( LOG_CRIT, "[%s:%s] Invalid directory request type=%d\n",
		   current->comm, __FUNCTION__, req->clm_type );
	    BUG();
	}

	/* We were purging and the destination node left.  Since there's
	 * nothing tha can be done at this point (the node that left doesn't
	 * care about the purge anymore), just drop the message and move on.
	 */

	bsdlog( LOG_WARNING, "[%s:%s] Node failed.  Drop purge request",
	        current->comm,__FUNCTION__ );
	cccp_msg_destroy( msg );
	clm_free_trans(req);
	ret = CPERR_OK;
    }
    else
    {
	dir_node = clm_direct(rl_entry.rl_name, rl_entry.rl_namelen);
	msg->cm_node = dir_node;
	msg->cm_hops = 0;
	msg->cm_port = PORT_DIRECT;

	ret = cccp_send( msg );
    }


    return( ret );
}

/*
 * Name
 * 	pti_call
 * Description
 *	Given a pointer to a transaction structure and an optional value
 *  	for the site id, determine where a request needs to be sent and
 *	call cccp_send to send it.  The request may have to be sent to a
 *	directory on the PORT_DIRECT port or it may go straight to a master
 *	handler on the PORT_MSTR port.
 * Arguments
 *	req - the transaction to send
 *	master - master site id if it is known
 * Returns
 *	0 for sucess
 *	< 0 for failure
 * Global Data Affected
 *	NONE
 */
int
pti_call( clm_trans_t * req, unsigned master )
{
    unsigned    port;
    unsigned    node;


    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_PTI_CALL,
	    req->clm_pid, req->clm_sequence, 0x00, 0x00);

    DBPRINTF((DBG_FUNC, "pti_call: req = 0x%x\n", req));


    node = 0;
    port = 0;
    if ( (master != 0) && clm_node_active( master ) )
    {
	/* here if we know which node is the master */
	DBPRINTF((DBG_DIRECT, "pti_call: master known, send to %d", master));
	node = master;
	port = PORT_MSTR;
    }
    else 
    {
	rl_info_t	rl_entry;

	/* here to look in the local RLDB for the master */

	/* Fill in the name and type information in the rl_info_t.  If the
	 * type is not one that is valid, punt.
	 */

	if ( ! dlm_stuff_rl_info( req, & rl_entry )
	     || (req->clm_type == DLM_REGISTER) )
	{
	    bsdlog( LOG_CRIT, "[%s:%s] Invalid directory request type=%d\n",
		    current->comm, __FUNCTION__, req->clm_type );
	    BUG();
	}

	if ( (DLM_NORMAL == rl_lookup( & rl_entry ))
	     && clm_node_active( rl_entry.rl_node ) )
	{
	    node = rl_entry.rl_node;
	    port = PORT_MSTR;
	}
    }


    return( dlm_pti_send( req, node, port ) );
}


/*
 * Name
 * 	pti_call_reg
 * Description
 *	Given a pointer to a DLM_REGISTER transaction structure,
 *  	determine where the request needs to be sent and
 *	call cccp_send to send it.  The request may have to be sent to a
 *	directory on the PORT_DIRECT port or it may go straight to a master
 *	handler on the PORT_MSTR port.  A separate function is used for UNIX
 *	locks because the resource name information is stored differently.
 * Arguments
 *	req - the transaction to send
 * Returns
 *	0 for sucess
 *	< 0 for failure
 * Global Data Affected
 *	NONE
 */
int
pti_call_reg(clm_trans_t *req)
{
    rl_info_t	rl_entry;
    unsigned    port;
    unsigned    node;


    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_PTI_CALL,
	    req->clm_pid, req->clm_sequence, 0x00, 0x00);

    DBPRINTF((DBG_FUNC, "pti_call_reg: req = 0x%x\n", req));


    /* here to look in the local RLDB for the master */
    rl_entry.rl_name = &req->clm_register.unix_name[0];
    rl_entry.rl_namelen = req->clm_register.namelen;
    rl_entry.rl_type = DLM_RES_UNIX;

    node = 0;
    port = 0;
    if ( (DLM_NORMAL == rl_lookup( & rl_entry ))
	 && clm_node_active( rl_entry.rl_node ) ) 
    {
	node = rl_entry.rl_node;
	port = PORT_MSTR;
    }


    return( dlm_pti_send( req, node, port ) );
}

/*
 * Name
 * 	pti_call_unix
 * Description
 *	Given a pointer to a DLM_LOCK transaction structure for a UNIX lock
 *  	request, determine where the request needs to be sent and
 *	call cccp_send to send it.  The request may have to be sent to a
 *	directory on the PORT_DIRECT port or it may go straight to a master
 *	handler on the PORT_MSTR port.
 * Arguments
 *	req - the transaction to send
 * Returns
 *	0 for sucess
 *	< 0 for failure
 * Global Data Affected
 *	NONE
 */
int
pti_call_unix(clm_trans_t *req, struct resource *rp)
{
    rl_info_t	rl_entry;
    unsigned    port;
    unsigned    node;


    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_PTI_CALL,
	    req->clm_pid, req->clm_sequence, 0x00, 0x00);

    DBPRINTF((DBG_FUNC, "pti_call_unix: req = 0x%x\n", req));


    req->clm_lockreq.nameinfo.namelen = rp->namelen;
    resname_bcopy(rp->name, &req->clm_lockreq.nameinfo.name[0], rp->namelen);

    /* If we got this far, then we've already found the local copy of
     * the resource, so we should at least have an idea of who the master
     * is, but only send there if the node is active.
     */

    node = 0;
    port = 0;
    if (!rp->master || clm_node_active(rp->master)) 
    {
	DBPRINTF((DBG_DIRECT, "pti_call: master known, send to %d",
		  rp->master ? rp->master : OurSiteID));
	node = rp->master ? rp->master : OurSiteID;
	port = PORT_MSTR;
    }
    else
    {
	/* here to look in the local RLDB for the master */
	rl_entry.rl_name = &req->clm_register.unix_name[0];
	rl_entry.rl_namelen = req->clm_register.namelen;
	rl_entry.rl_type = DLM_RES_UNIX;

	if ( (DLM_NORMAL == rl_lookup( & rl_entry ))
	    && clm_node_active( rl_entry.rl_node ) )
	{
	    node = rl_entry.rl_node;
	    port = PORT_MSTR;
	}
    }


    return( dlm_pti_send( req, node, port ) );
}
