/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_locallock.c,v 1.8 2001/11/28 00:31:16 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: clmll_init
 *              clmll_create
 *              clmll_broadcast
 *              clmll_llb_call_complete
 *              clmll_read_nack_response
 *              clmll_nack_complete
 *              clmll_check_llb
 *              clmll_send_nack_response
 *              clmll_proc_response
 *              clmll_print_llb
 *              clmll_print_status
 */

/* ============================================================ */
/*
 *      clm_locallock.c
 *
 *        This module implements the local lock feature of the clmlock API
 *        call to the Cluster Lock Manager.  These functions are called
 *        from functions in clm_clmlock.c or from within.
 */

/*
 *        REQUEST
 *              LKM_LOCAL
 *                      master =   OurSiteID
 *                      port =     PORT_MSTR_CR      (pti_read_request_cr)
 *                      callback = pti_call_complete
 *              LKM_FINDLOCAL
 *                      master =   Every Node in Cluster
 *                      port =     PORT_MSTR         (pti_read_request)
 *                      callback = clmll_llb_call_complete
 *
 *        RESPONSE      clm_lock
 *              LKM_FINDLOCAL
 *                      master =
 *                      port =     PORT_NACK         (clmll_read_nack_response)
 *                      callback = clmll_nack_complete
 *
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/time.h>

#include "dlm.h"
#include "dlm_version.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "clm_data.h"
#include "dlm_kernel.h"

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/*
 * Linux note:  I do not know right yet what these comment
 * blocks mean.  THEY WERE NOT ADDED AS PART OF THE LINUX
 * PORT!  They were already here.  There is nobody around
 * that would know what these mean, i.e., is there, or is
 * there not, work to do here...  I will leave them here
 * to remind us that we need to decipher the meaning, and see
 * if there's anything here to care about!
 */

/*  TO DO LIST
 *  FAIL OVER
 *   - clm_reconfig.c, make sure directories get created OK for local
 *     resource and at end of reconfig blow away resource directories
 *  MISC
 *   - could speed up broadcast, by sending to local first.
 *   - clm_group.c see if effected by changes
 *
 */

/*
 *  OTHER FILES EDITED
 *  1. Changed clm_sender to void *, edited clm_batch.c clm_libr.c
 *  2. For Sync Locks: check in clm_lock_regresp for LKM_BLOCK & LKM_SYNCSTS
 *  3. clm_main.c kern_main() added clmll_init () and PORT_NACK declare
 *  4. kernel.c exit() clear PORT_NACK is this necessary?
 *  5. clm_cti.c clm_reply () do not free request transaction if LKM_FINDLOCAL
 *  6. clm_main.c pti_proc_response () call clmll_proc_response  also don't
 *     call clm_reply if multiple masters were found.
 *  7. clm_clmlock.c begin_lock_clm () clear LKM_LOCAL & LKM_FINDLOCAL if
 *     convert
 *  8. Added resource flag, RSRC_LOCAL, in clmstructs.h
 *  9. clm_migrate.c clm_cache () if RSRC_LOCAL don't migrate
 */
/* ============================================================ */

/* 
 *  Status (per node) of a Lock Request made from Local Lock Broadcast.
 */
enum llb_status {   /* Local Lock Broadcast Message             */
  LLB_NOT_SENT = 0, /* Has not been Sent                        */
  LLB_SENT,         /* Sent but not Received                    */
  LLB_VERSINCOMPAT, /* Not Sent Vers. Incompatible              */
  LLB_SENDERROR,    /* Error Sending Message                    */
  LLB_NODEFAILED,   /* Sent but Node Faied                      */
  LLB_RECEIVED,     /* Sent & Received but not Found            */
  LLB_FOUND,        /* Sent & Received & Found                  */
  LLB_NOT_MEMBER,   /* Node is not a member of the cluster      */
  LLB_MAGIC_COOKIE  /* Insures clm_sender returns valid llb_ptr */
};

/*
 *  Local Lock Broadcast structure used to keep track of the progress
 *  of locating a resource that was created using local lock feature.
 *  Maintains every node's lock request status  (see enum llb_status)
 */
struct locallock_broadcast {
    dlm_listnode_t    link;
    enum llb_status status[DLM_MAXNODES+1];/* Node's Lock Request Status */
    clm_trans_t     *trans_ptr;           /* Request Transaction ptr    */
};

/* GLOBAL VARIABLES */
static DECLARE_DLM_LIST( dlm_bcast_queue );

/*
 * Name
 *      clmll_init
 * Description
 *      This function is called from kern_main() in clm_main.c.
 *
 *      Initializes the LLB_HEAD, this is required for kload.
 *      Walks the LLB list and frees LLB structures & their transaction
 *      request until there are no more.
 *
 * Arguments
 *      None.
 * Returns
 *      Void
 * Global Data Affected
 *      None.
 */
void
clmll_init (void)
{
    dlm_listnode_t * curr;

    DBPRINTF((DBG_FUNC, "enter clmll_init\n"));
    
    while ( (curr = dlm_list_rem_head( & dlm_bcast_queue )) != NULL )
    {
	CLM_FREE( curr );
    }

    DBPRINTF((DBG_FUNC, "leave clmll_init\n"));

    return;
}


/*
 * Name
 *      clmll_create
 * Description
 *      This function gets called from begin_lock_clm() when LKM_LOCAL set.
 *
 *      Creates a local lock. It does this by not creating (or updating) 
 *      directory information, instead the resource is mastered on the local
 *      node.  Similar to pti_call.  
 *      It is understood that the resource does not already exist, however
 *      just for safety we check if it existes locally and return
 *      DLM_BADRESOURCE.
 *
 * Arguments
 *      req - the lock transaction request
 * Returns
 *      0 for success
 *     -1 for failure
 * Global Data Affected
 *      None.
 */

int
clmll_create( clm_trans_t * req )
{
    int                 ret;
    struct clm_regreq   regreq; /* see if resource already exists locally */
    struct reclock    * rl_ptr; /* reclock ptr  */ 

    
    DBPRINTF((DBG_FUNC, "enter clmll_create\n"));

    DBPRINTF((DBG_LOCAL, "clmll_create: create resource %s on node %d\n",
	       printbin(req->clm_lockreq.nameinfo.name,
			req->clm_lockreq.nameinfo.namelen),
	       OurSiteID));

    /* If resource already existes locally then return DLM_BADRESOURCE */
    (void) memset( & regreq, 0, sizeof( struct clm_regreq ) );
    regreq.name = req->clm_lockreq.nameinfo.name;
    regreq.namelen = req->clm_lockreq.nameinfo.namelen;
    regreq.type = req->clm_locktype; 

    if ( NULL != lookup_resource( & regreq ) )
    {
	/* Resource already exists locally so reject lock */
	DBPRINTF((DBG_LOCAL,
		  "clmll_create: resource %s already exists on node %d\n",
		  printbin(req->clm_lockreq.nameinfo.name,
			   req->clm_lockreq.nameinfo.namelen), OurSiteID));
	
	req->clm_status = DLM_BADRESOURCE; /* LKSB status */
	
	/* Need to copy the AST Route Info (& various other bits) from the
	 * request into the reclock so we can send AST back. 
	 */

	rl_ptr = find_reclock (DLM_RES_VMS, req->clm_lockinfo.lockid);
	(void) memcpy( & rl_ptr->route, & req->clm_route, 
		       sizeof( struct routeinfo ) );
	rl_ptr->lck = req->clm_lockinfo;
	rl_ptr->lck.site = OurSiteID;
	rl_ptr->route.asth.asth_pid = req->clm_pid;

	ret = -1;
    }
    else
    {
	ret = dlm_pti_send( req, OurSiteID, PORT_MSTR_CR );
    }
    
    
    DBPRINTF((DBG_FUNC, "leave clmll_create\n"));
    return( ret );
}


/*
 * Name
 *      clmll_broadcast
 * Description
 *      This function gets called from begin_lock_clm when LKM_FINDLOCAL set.
 *
 * ******* HACK *******
 * used to assume nodeids were compact starting at 1.
 * now use index + 1 !!!!!!!! which isn't the same thing AT ALL
 * since nodeids can be 1 .. MAXSHORT
 *
 *      Allocates a LLB structure on LLB list.
 *
 *      For every node in cluster
 *       - allocate & set up message
 *       - ask for lock on resource using cccp_send
 *       - if send fails, mark message status in LLB, and free message
 *         (otherwise the callback will free the message)
 *
 *      The clm_lock call will send back a NACK over PORT_NACK if resource
 *      doesn't exist.
 *
 *      If the resource does exist then master_lock_clm will handle normally &
 *      clmll_proc_response () will update the node's message status in LLB.
 *
 *      The clmll_llb_call_complete callback will free the message structure
 *      and if appropriate (message failed) will update the LLB structure.
 *
 *  COULD INCREASE PERFORMANCE BY CHECKING LOCAL NODE FIRST
 *
 * Arguments
 *      req - the lock transaction request
 * Returns
 *      0 for success
 *     -1 for failure
 * Global Data Affected
 *      None.
 */
int
clmll_broadcast (clm_trans_t *req)
{
    int node;                       /* Node to send request           */
    int ndx ;				/* index of node */
    int ret;                        /* Return Value                   */
    int maxnode = clm_node_count(); /* Number of Nodes in the Cluster */
    locallock_broadcast_t *llb_ptr; /* Ptr to LLB Structure           */
    clm_msg_t *msg;                 /* Message Structure              */
    
    DBPRINTF((DBG_FUNC, "enter clmll_broadcast\n"));

    /* Allocate LLB structure */
    llb_ptr = CLM_MALLOC( sizeof( locallock_broadcast_t ) );
    if (NULL == llb_ptr) /* Malloc failed */
    {
	abort ("clmll_broadcast: malloc failed");
    }
    else 
    {
	(void) memset( llb_ptr, 0, sizeof( locallock_broadcast_t ) );
    }

    dlm_list_add_head( & dlm_bcast_queue, (dlm_listnode_t *) llb_ptr );

    /* Initialize new LLB Structure */
    llb_ptr->trans_ptr = req;

    /*
     * We set the otherwise unused status[0] = LLB_MAGIC_COOKIE so functions
     * clmll_read_nack_response & clmll_proc_response can validate llb_ptr.
     * We also set the status of nodes not in the cluster to be LLB_NOT_MEMBER.
     */
    llb_ptr->status[0] = LLB_MAGIC_COOKIE;
    for (ndx = (maxnode + 1); ndx <= DLM_MAXNODES ; ndx++)
    {
	llb_ptr->status[ndx] = LLB_NOT_MEMBER;
    }
    
    /* For Each Node request lock on resource */
    for (ndx = 0 ; ndx < maxnode ; ndx++)
    {
	/* Allocate & Set up  message structure */
	node = INDEX_TO_NODEID( ndx );
	msg = cccp_msg_create( PORT_MSTR, BASEPRIO, node, req->clm_type,
			       sizeof( *req ), clmll_llb_call_complete,
			       req, NULL );
	if (NULL == msg) /* Malloc Failed */
	{
	    abort("clmll_broadcast: malloc failed");
	}
	
	req->clm_direction = PTI_REQUEST;
	req->clm_sender = (locallock_broadcast_t *)
	    dlm_list_get_head( & dlm_bcast_queue ); /* used by NACK */
	req->clm_next = NULL;
	
	DBPRINTF((DBG_LOCAL,
		"clmll_broadcast: send %d lock request on %s to node %d\n",
		req->clm_lockreq.lockinfo.mode,
		req->clm_lockreq.nameinfo.name, node));
	
	/* make any adjusments for version skew between lock managers */
	ret = clm_vers_snd(msg, clm_low_version());
	if (ret != DLM_NORMAL)
	{
	    DBPRINTF((DBG_LOCAL,
		       "clmll_broadcast: CLM vers incompat w/ nodes %d & %d\n",
		       node, OurSiteID));

	    /* free message */
	    cccp_msg_destroy( msg );

	    /* mark LLB vers incpt */
	    llb_ptr->status[ndx + 1] = LLB_VERSINCOMPAT ;
	}
	else
	{	/* mark LLB that message sent */
	    llb_ptr->status[ndx + 1] = LLB_SENT ;
	    ret = cccp_send(msg);
	    
	    DBPRINTF((DBG_LOCAL, "clmll_broadcast: cccp_send returned %d\n",
		       ret));
	    
	    if (ret != CPERR_OK)
	    {
		if (ret == CPERR_NODEDOWN)
		{
		    /* Mark Node's Status = Sent message to failed node */
		    llb_ptr->status[ndx + 1] = LLB_NODEFAILED;
		}
		else
		{
		    /* Mark Node's Status = Error sending message */
		    llb_ptr->status[ndx + 1] = LLB_SENDERROR;
		}
		
		/* Check LLB Structure */
		if (clmll_check_llb (llb_ptr)) /* call failed */
		{
		    DBPRINTF((DBG_LOCAL,
			       "clmll_broadcast: clmll_check_llb failed\n"));
		}
		if (msg) /* free message */
		{
		    cccp_msg_destroy( msg );
		}
	    }
	}
    }
    
    DBPRINTF((DBG_FUNC, "leave clmll_broadcast\n"));
    return (0);
}


/*
 * Name
 * 	clm_llb_call_complete
 * Description
 *      This is the callback function for the lock requests made from the 
 *      clmll_broadcast function (which gets called when LKM_FINDLOCAL set).
 *
 *      Return Value         Action
 *      ------------         ------
 *      CPERR_NODEFAILED     Mark node's status in LLB as LLB_NODEFAILED
 *      CPERR_OK             Message sent OK, continue.
 *      !CPERR_OK            Mark node's status in LLB as LLB_SENDERROR
 *      cm_private set       Mark node's status in LLB as LLB_SENDERROR
 *
 *      The LLB structure is checked everytime a node's status is updated.
 *      The message structure is freed at this point.
 *
 *      The request structure is used by every message sent in the 
 *      broadcast, it will only get freed when all responses are returned.
 *      The clmll_check_llb function handles this.
 *
 * Arguments
 *	msg - pointer to the message structure that was acknowledged
 *	stat - status returned
 * Returns
 *	void
 * Global Data Affected
 *	NONE
 */
void
clmll_llb_call_complete(clm_msg_t *msg, int stat)
{
    locallock_broadcast_t *llb_ptr; /* Ptr to LLB Structure */
    clm_trans_t           *req;     /* Ptr to Request       */
    int ndx ;

    llb_ptr = NULL;

    DBPRINTF((DBG_FUNC, "enter clmll_llb_call_complete\n"));

    if (NULL == msg)
    {
	DBPRINTF((DBG_LOCAL, "clmll_llb_call_complete: msg == NULL\n"));
	DBPRINTF((DBG_FUNC, "leave clmll_llb_call_complete\n"));
	return;
    }
    
    req = (clm_trans_t *)msg->cm_data;
    ndx = NODEID_TO_INDEX(msg->cm_node) ;
    
    /* If !CPERR_OK then mark LLB appropriately and check LLB */
    if (CPERR_OK != stat) /* error occurred sending message */
    {
      llb_ptr = (locallock_broadcast_t *)req->clm_sender;
      if (CPERR_NODEDOWN == stat) /* message sent to failed node */
        {
          DBPRINTF((DBG_LOCAL,
                    "clmll_llb_call_complete: Msg sent to Failed Node %d\n",
                    msg->cm_node));
          /* Mark Node's status in LLB as node failed */
          llb_ptr->status[ndx + 1] = LLB_NODEFAILED;
        }
      else
        {
          DBPRINTF((DBG_LOCAL,
		      "clmll_llb_call_complete: Msg return error %d\n", stat));
          /* Mark Node's message status in LLB as error sending message */
          llb_ptr->status[ndx + 1] = LLB_SENDERROR;
        }   
      /* Check LLB Structure */
      if (clmll_check_llb (llb_ptr)) /* call failed */
        {
          DBPRINTF((DBG_LOCAL, 
                    "clmll_llb_call_complete: clmll_check_llb failed\n"));
        }
    }
    
    /* we are purposely not freeing the req ptr - see func desc above */
    
    /*
     * Local Locks Do Not migrate, if BOUNCE_xxx flags set, then something
     * is wrong.  Mark send error LLB.
     */
    if ((NULL != msg->cm_private) &&
        ((ulong)msg->cm_private & MSG_BOUNCE ||
         (ulong)msg->cm_private & MSG_BNCE_DONE)) {
      DBPRINTF((DBG_LOCAL, "clmll_llb_call_complete: msg->cm_private = %d\n",
                msg->cm_private));
      /* Mark Node's Message Status in LLB as error sending message */
      llb_ptr->status[ndx + 1] = LLB_SENDERROR;
	
      /* Check LLB Structure */
      if (clmll_check_llb (llb_ptr)) /* call failed */
        { 
          DBPRINTF((DBG_LOCAL,
                    "clmll_llb_call_complete: clmll_check_llb failed\n"));
        }
	
    }
    
    if (msg) /* free message structure */
    {
	cccp_msg_destroy( msg );
    } 
    
    DBPRINTF((DBG_FUNC, "leave clmll_llb_call_complete\n"));
    return; 
}


/*
 * Name
 *	clmll_read_nack_response
 * Description
 * 	This function is called by the communications layer when a NACK
 *      response to a request is sent from a master node.  Update the LLB.
 * Arguments
 *	msg - response message structure
 * Returns 
 *	0 for success
 *     -1 for network error
 * Global Data Affected
 *	None
 */
int
clmll_read_nack_response(clm_msg_t *msg)
{
    locallock_broadcast_t *llb_ptr;  /* Ptr to LLB structure */
    int ndx ;
    
    DBPRINTF((DBG_FUNC, "enter clmll_read_nack_response\n"));

    if (NULL == msg)
    {
	DBPRINTF((DBG_LOCAL, "clmll_read_nack_response: msg == NULL\n"));
	DBPRINTF((DBG_FUNC, "leave clmll_read_nack_response\n"));
	return (-1);
    }
    
    acquire_lock( LK_ENTRY, kproc_pid );

    /* Might want to add version compatiblity here */
    
    /* The Message contains a pointer to the LLB Structure */
    llb_ptr = *((locallock_broadcast_t **)(msg->cm_data));
    ndx = NODEID_TO_INDEX(msg->cm_node) ;

    /* check to see if llb_ptr is valid */
    if (llb_ptr->status[0] != LLB_MAGIC_COOKIE)
    {
	/* llb_ptr doesn't point to a valid llb structure */
	DBPRINTF((DBG_LOCAL, "clmll_read_nack_response: invalid cookie\n"));
	return (-1);
    }

    llb_ptr->status[ndx + 1] = LLB_RECEIVED;
    
    if (clmll_check_llb (llb_ptr)) /* call failed */
    {
	DBPRINTF((DBG_LOCAL,
		   "clmll_read_nack_response: clmll_check_llb failed\n"));
    }
    release_lock( LK_ENTRY, kproc_pid );
    
    /* Message gets freed by communication layer */

    DBPRINTF((DBG_FUNC, "leave clmll_read_nack_response\n"));
    return (0);
}


/*
 * Name
 * 	clmll_nack_complete
 * Description
 *	This function is called by the communications layer when it has
 *      received an acknowledgement for an outbound NACK being sent to 
 *      a node that requested a lock with the LKM_FINDLOCAL flag set.
 *
 *      The message structure is freed aswell as its cm_data field.
 * Arguments
 *	msg - pointer to the message structure that was acknowledged
 *	stat - status returned
 * Returns
 *	Nothing
 * Global Data Affected
 *	NONE
 */
void
clmll_nack_complete(clm_msg_t *msg, int stat)
{
    DBPRINTF((DBG_FUNC, "enter clmll_nack_complete\n"));
    
    if ((CPERR_OK != stat) || (NULL == msg)) /* NACK message send failed */
    {
	DBPRINTF((DBG_LOCAL,
		   "clmll_nack_complete: NACK msg to %d returned error (%d)\n",
		   msg->cm_node, stat));
    }
    
    if (NULL != msg->cm_data) /* free cm_data ptr */
    {
	CLM_FREE (msg->cm_data);
    }
    if (NULL != msg) /* free message structure */
    {
	cccp_msg_destroy( msg );
    }
    
    DBPRINTF((DBG_FUNC, "leave clmll_nack_complete\n"));
    return;
}


/*
 * Name
 * 	clmll_check_llb
 * Description
 *	This function is called when the llb structure has been updated.
 *      Once all the nodes have been heard from (node's status is greater
 *      then sent) then the request transaction and the LLB structure can be
 *      freed.  The LLB list also needs to be updated.
 *
 *      If the resource was never found then a local lock should be created.
 *      If more than one node found then log error.
 * Arguments
 *     *llb_ptr - pointer to the Local Lock Broadcast Structure
 * Returns
 *	0 - for success
 *     -1 - for failure
 * Global Data Affected
 *	NONE
 */
int
clmll_check_llb (locallock_broadcast_t *llb_ptr)
{
    int ndx ;		/* node index */
    int maxnode = clm_node_count(); /* Number of Nodes in the Cluster       */
    int found_node = 0; /* 0 Not found; >1 Node where found; -1 Multi finds */
    clm_trans_t *req;   /* ptr to transaction request */
    
    DBPRINTF((DBG_FUNC, "enter clmll_check_llb\n"));

    clmll_print_llb (llb_ptr);

    for (ndx = 0 ; ndx < maxnode ; ndx++)
    {
	if (llb_ptr->status[ndx + 1] == LLB_NOT_SENT)
	{
	    /* This should not happen, could send message */
	    DBPRINTF((DBG_LOCAL, "clmll_check_llb: msg status == unsent\n"));
	    DBPRINTF((DBG_FUNC, "leave clmll_check_llb\n"));
	    return (0);
	}
	else if (llb_ptr->status[ndx + 1] == LLB_SENT)
	{
	    DBPRINTF((DBG_FUNC, "leave clmll_check_llb\n"));
	    return (0);
	}
	else if (llb_ptr->status[ndx + 1] == LLB_FOUND)
	{
	    if (0 == found_node)
	    {	
		found_node = INDEX_TO_NODEID(ndx) ; 
	    }
	    else 
	    {	
		/* Resource was found on multiple nodes */
		found_node = -1;
	    } 
	}
    }
    
    /* At this point all nodes have been heard from */
    
    if (-1 == found_node) 
    {
	/*
	 * The first copy of the resource will return an AST the others
	 * won't pti_read_response handles this, so we just print error
	 */
	DBPRINTF((DBG_LOCAL, "clmll_check_llb: res found on multi nodes\n"));
    }
    else if (0 == found_node)
    {
	/* resource wasn't found, create local lock with LKM_LOCAL flag */
	req = llb_ptr->trans_ptr;
	req->clm_lockinfo.lk_flags &= ~LKM_FINDLOCAL;
	req->clm_lockinfo.lk_flags |= LKM_LOCAL;
	
	if (0 != clmll_create (req)) 
	{
	    struct reclock *a = find_reclock(DLM_RES_VMS,
				    req->clm_lockinfo.lockid);

	    /* local lock request failed */
	    DBPRINTF((DBG_LOCAL, "clmll_check_llb: Local lock request on "
			"Resource %s failed\n",
			printbin(req->clm_lockreq.nameinfo.name,
			req->clm_lockreq.nameinfo.namelen)));
	    
	    /* clear UDP flag and send AST */
	    req->clm_respflags &= ~RESP_UDP;
	    req->clm_respflags |= RESP_AST;	
	    clm_reply(req, 0);

	    /* free the reclock */
	    if (a)
	    {
		free_le(DLM_RES_VMS, a);
	    }
	}
    }
    
    dlm_list_remove( (dlm_listnode_t *) llb_ptr );

    /* Free the LLB structure */
    CLM_FREE (llb_ptr);

    DBPRINTF((DBG_FUNC, "leave clmll_check_llb\n"));
    return (0);
}



/*
 * Name
 * 	clmll_send_nack_response
 * Description
 *	This function is called from clm_lock.  At that point the lock
 *      request has reached the master, the resource isn't found, and
 *      the LKM_FINDLOCAL flag is set.  A NACK is sent to the node
 *      requesting the lock.
 * Arguments
 *      msg   - the lock request message
 * Returns
 *	0 - for success
 *     -1 - for failure
 * Global Data Affected
 *	NONE
 */
int
clmll_send_nack_response (clm_msg_t *msg)
{
    clm_trans_t	  *req = msg->cm_data;
    struct lockinfo *lip = &req->clm_lockreq.lockinfo;
    
    clm_msg_t  *nack_msg;  /* NACK message structure */
    void      **llb_ptr;   /* used to pass LLB ptr   */
    int         ret_stat;  /* return status          */
    
    DBPRINTF((DBG_FUNC, "enter clmll_send_nack_response\n"));

    
    /* Allocate ptr to store LLB structure address being passed */
    llb_ptr = (void *)CLM_MALLOC(sizeof(void *));
    if (NULL == llb_ptr) /* malloc failed */
    {
	abort ("clm_lock: CLM_MALLOC failed");
    } 
    
    /* 
     * clm_sender contains the address of the LLB structure that will get
     * updated when NACK gets handled on requesting site. cm_data field 
     * points to llb_ptr allocated to contain address in clm_sender.
     */
    *llb_ptr = req->clm_sender;
    
    /* Allocate NACK message */
    nack_msg = cccp_msg_create( PORT_NACK, BASEPRIO, lip->site, NACK_SEND,
			        sizeof( void * ), clmll_nack_complete,
			        llb_ptr, NULL );
    if (NULL == nack_msg) /* malloc failed */
    {
	CLM_FREE( llb_ptr );
	abort ("clm_lock: CLM_MALLOC failed");
    }
    
    ret_stat = cccp_send (nack_msg);
    if (ret_stat != CPERR_OK)
    {
	DBPRINTF((DBG_LOCAL, "clmll_send_nack_response: nack to %d failed\n",
		  lip->site));
	CLM_FREE( msg->cm_data );
	cccp_msg_destroy( msg );
	DBPRINTF((DBG_FUNC, "leave clmll_send_nack_response\n"));
	return (-1);
    }

    DBPRINTF((DBG_FUNC, "leave clmll_send_nack_response\n"));
    return (0);
}


/*
 * Name
 * 	clmll_proc_response
 * Description
 *	This function is called from pti_proc_response. 
 *
 *      We have just received a (resource found) response to a lock request we
 *      made from clmll_broadcast (LKM_FINDLOCAL set).  We update the LLB
 *      structure, the clm_sender field points to it.  We want to match sure 
 *      this was the first response for the request and then do a clm_reply().
 * Arguments
 *      *msg  - ptr to message structure
 *      *resp - ptr to response transaction structure
 * Returns
 *	clm_trans_t* - ptr to request transaction (NULL if error)     
 * Global Data Affected
 *	NONE
 */
clm_trans_t*
clmll_proc_response( unsigned node_id, const clm_trans_t * resp, 
		     int *multiple_masters )
{
    int node ;	/* Node id */
    int ndx ;	/* node index */
    int maxnode = clm_node_count(); /* Number of Nodes in the Cluster */
    locallock_broadcast_t *llb_ptr;

    DBPRINTF((DBG_FUNC, "enter clmll_proc_response\n"));

    *multiple_masters = 0;

    if (NULL == resp->clm_sender) 
    {
	DBPRINTF((DBG_LOCAL, "clmll_proc_response: clm_sender == NULL\n"));
	DBPRINTF((DBG_FUNC, "leave clmll_proc_response\n"));
	return (NULL);
    }
    
    /* Get ptr to LLB structure form clm_sender field */
    llb_ptr = (locallock_broadcast_t *)resp->clm_sender;
  
    /* check to see if llb_ptr is valid */
    if (llb_ptr->status[0] != LLB_MAGIC_COOKIE)
    {
	/* llb_ptr doesn't point to a valid llb structure */
	DBPRINTF((DBG_LOCAL, "clmll_proc_response: invalid cookie\n"));
	DBPRINTF((DBG_FUNC, "leave clmll_proc_response\n"));
	return (NULL);	
    }
    
    ndx = NODEID_TO_INDEX( node_id );
    if (ndx < 0)
    {
	DBPRINTF((DBG_LOCAL, "clmll_proc_response: invalid source node\n")) ;
    }
    llb_ptr->status[ndx + 1] = LLB_FOUND; 

    /* Check LLB to see if resource ready, if so then AST already sent */
    for (ndx = 0 ; ndx < maxnode ; ndx++)
    {
	node = INDEX_TO_NODEID(ndx) ;
	if ((node != node_id) && (llb_ptr->status[ndx + 1] == LLB_FOUND))
	{
	    DBPRINTF((DBG_LOCAL,
		       "clmll_proc_request: resource has %d masters\n",
		      *multiple_masters));
	    /*
	     * increment multiple_masters so pti_proc_response() won't call
	     * clm_reply () preventing another AST from being sent.
	     */
	    (*multiple_masters)++;
	}
    }

    DBPRINTF((DBG_FUNC, "leave clmll_proc_response\n"));
    return (llb_ptr->trans_ptr); /* Returning ptr to request transaction */ 
}


/*
 * Name
 * 	clmll_proc_response
 * Description
 *      Writes to syslog the local lock broadcast structure. 
 * Arguments
 *      *llb_ptr - pointer to local lock broadcast structure
 * Returns
 *      void
 * Global Data Affected
 *	NONE
 */
void
clmll_print_llb (locallock_broadcast_t *llb_ptr)
{
    int node;
   
    DBPRINTF((DBG_FUNC, "enter clmll_print_llb\n"));
    DBPRINTF((DBG_LOCAL, "Local Lock Broadcast Structure\n"));
    DBPRINTF((DBG_LOCAL, "------------------------------\n"));

    for (node = 0; node <= DLM_MAXNODES; node++)
    {
	DBPRINTF((DBG_LOCAL, "  status[%d]=%s\n",
		   node, clmll_print_status(llb_ptr->status[node])));
    }
    
    if (NULL == llb_ptr->trans_ptr) {
	DBPRINTF((DBG_LOCAL, "  *trans_ptr -> NULL\n"));
    } else {
	DBPRINTF((DBG_LOCAL, "  *trans_ptr -> 0x%x\n", llb_ptr->trans_ptr));
    }
    
    DBPRINTF((DBG_LOCAL, "------------------------------\n"));
    
    DBPRINTF((DBG_FUNC, "leave clmll_print_llb\n"));
    return;
}

/*
 * Name
 * 	clmll_print_status
 * Description
 *      converts llb_status number into meaningful string
 * Arguments
 *      status - llb_status
 * Returns
 *	char* - llb status message string
 * Global Data Affected
 *	NONE
 */
char*
clmll_print_status (int status)
{
    switch (status)
    {
    case LLB_NOT_SENT:
	return ("Message Not Sent");
	break;
    case  LLB_SENT:
	return ("Message Sent");
	break;
    case LLB_VERSINCOMPAT:
	return ("CLM Versions Incompatible");
	break;
    case LLB_SENDERROR:
	return ("Message Send Error");
	break;
    case LLB_NODEFAILED:
	return ("Node Failed");
	break;
    case LLB_RECEIVED:
	return ("Resource Not Found");
	break;
    case LLB_FOUND:
	return ("Resource Found");
	break;
    case LLB_NOT_MEMBER:
	return ("Node Not in Cluster");
	break;
    case LLB_MAGIC_COOKIE:
	return ("Magic Cookie, Yummie!");
	break;
    default:
	return ("Unknown");
    }
    
    return ("Switch Failed");
}

