/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_lock.c,v 1.6 2001/09/27 22:28:59 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		add_orphan_ref
 *		add_queue
 *		begin_purge
 *		check_queue
 *		lck_set_seq
 *		master_purge
 *		pti_purge
 *		pti_purge_complete
 *		purgelist
 *		purgelocks
 *		remove_queue
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/types.h>
#include <linux/signal.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "cltrc.h"
#include "clm_data.h"			/* external data definitions */
#include "dlm_kernel.h"
#include "clm_locksw.h"
#include "ast.h"
#include <iterator.h>

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/* recover rlq struct at ofs in reclock x */
#define OFSP(x, ofs) ((struct rlq *) (((char *) (x)) + (ofs)))

static int purgelist( struct resource *rp, dlm_list_t * queue,
		      int site, int pid, int flags, int which_queue );

void
add_orphan_ref(register struct resource *rp,
               register int site,
               register int pid);

/*
 * Name
 *	mylock
 * Description
 *	Give a pointer to a reclock and a pointer to a transaction structure,
 *	return TRUE if the lock is owned (grouply or singly) by the client
 *	indicated by the transaction.  Otherwise return FALSE.
 * Arguments
 *	a   - pointer to a valid reclock structure
 *	req - pointer to a valid transaction structure
 * Returns
 *	See description.
 * Global Data Affected
 *	NONE
 */
bool_t
mylock( const struct reclock * a, const clm_trans_t * req )
{
    return( (a->lck.pid == req->clm_authpid)
	    || (a->lck.pid == req->clm_pid) );
}

/*
 * check_queued_lock
 *
 * DEBUG
 * check one lock for correctness for check_queues()
 * abort()s if problems found
 * conditional compilation with SOMETHING_IS_TERRIBLY_WRONG
 */

#ifdef SOMETHING_IS_TERRIBLY_WRONG

void
check_queued_lock(
    char *s,			/* queue name string */
    struct resource *rp,	/* resource on which lock is queued */
    struct reclock *a,		/* lock being examined */
    int bits)			/* queue-indicating bits for check *unused* */
{
    /* verify correct site value */
    if (rp->type == DLM_RES_VMS)
    {
	if (a->lck.remlockid)
	{
	    if ((a->lck.remlockid >> 28) != (a->lck.site & 0xF) &&
			    ~rp->rsrc_flags & RSRC_FRZEN)
	    {
		abort("check_handles:  Invalid site value");
	    }
	}
	else
	{
	    if ((a->lck.lockid >> 28) != (a->lck.site & 0xF) &&
			    ~rp->rsrc_flags & RSRC_FRZEN)
	    {
		abort("check_handles:  Invalid site value");
	    }
	}
    }

    /* verify lock not remote on non-master copy */
    if (a->lck.site != OurSiteID && rp->master &&
		    clm_is_running() &&
		    ~rp->rsrc_flags & RSRC_FRZEN)
    {
	    abort("check_handles: remote convert on non-master");
    }

    /* verify resource handle match */
    if (a->lck.rh.rh != rp->rh.rh)
	    abort("check_handles: convert");

	/* this check assumes a four node cluster */
    if (NODEID_TO_INDEX(a->lck.site) < 0)
	abort("check_handles: convert site out of range");
}
#endif

/*
 * Name
 *	check_handles
 * Description
 *	Walk all queues for a resource and ensure that all of the resource
 *	handles on all of the locks match the resource's resource handle.
 *	Abort if one is found that doesn't match.
 * Arguments
 *	rp - pointer to resource referenced by the lock
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
check_handles(struct resource *rp, int flag)
{
#ifdef SOMETHING_IS_TERRIBLY_WRONG
	struct reclock	*a;
	int	site[] = {0,0,0,0,0,0,0,0,0};
	int	count = 0;

	/* master should never be set to the local nodeid */
	if (rp->master == OurSiteID) {
		abort("master set local");
	}

	for ( EACH_DLM_LISTNODE( & rp->lockqs[RLQ_GRANT], a ) )
	{
		count++;
		check_queued_lock("grant", rp, a, LOCK_GRANTED) ;

		/* verify converting locks have convert bit set in state */
		if (a->lck.state & LOCK_CONVERTING)
			abort("check_handles: conv bit set on lock queue");

		/* verify converting locks have convert bit set in state */
		if (a->lck.state & LOCK_WAITING)
			abort("check_handles: wait bit set on lock queue");
	}

	for ( EACH_DLM_LISTNODE( & rp->lockqs[RLQ_CONV], a ) )
	{
		count++;
		check_queued_lock("convert", rp, a, LOCK_CONVERTING) ;
	}

	for ( EACH_DLM_LISTNODE( & rp->lockqs[RLQ_WAIT], a ) )
	{
		count++;
		check_queued_lock("wait", rp, a, LOCK_WAITING) ;
	}

	if ( ! (flag & CH_NOREF) && rp->refcount != count)
	{
		abort("check_handles: bad refcount");
	}
#endif /* SOMETHING_IS_TERRIBLY_WRONG */
}

/*
 * Name
 *	lck_set_seq
 * Description
 *	Determine if the resource specified is the master copy.  If so,
 *	scan the lock queues and assign the sequence number of the specified
 *	lock to be one more than the highest existing value.  If the value
 *	we assign wraps back to 0, then make a note in the log, but it's
 *	*extremely* unlikely that it will cause a problem.
 * Arguments
 *	rp - pointer to resource referenced by the lock
 *	a - the lock that may need its sequence number set
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
lck_set_seq(struct resource *rp, struct reclock *a)
{
    unsigned         highest = 0;
    unsigned         i;
    struct reclock * tail;


    /* If we're not the master copy of the resource, then the sequence
     * number of the reclock should've already been set and we shouldn't
     * touch it.
     */

    if ( rp->master == 0 )
    {
	/* We rely on the fact that the tail of the queues should have the
	 * highest sequence values.
	 */

	for (i = 0 ; i < RLQ_MAX ; i++)
	{
	    tail = (struct reclock *) dlm_list_get_tail( & rp->lockqs[ i ] );
	    if ( (tail != NULL) && (tail->lck.seqnum > highest) )
	    {
		highest = tail->lck.seqnum;
	    }
	}


	a->lck.seqnum = highest + 1;
	if (a->lck.seqnum < highest)
	{
	    bsdlog(LOG_NOTICE, "lck_set_seq: sequence wrap on %x\n",
		   a->lck.lockid);
	}
    }
}

/*
 * Name
 *	add_queue
 * Description
 *	add a reclock to the tail of the specified doubly linked queue
 * Arguments
 *	rlqp - pointer to queue head/tail struct
 *	a - the lock to be enqueued
 *	ofs - offset to the double-link structure in the object linked
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
add_queue( dlm_list_t * queue, struct reclock *a, size_t ofs)
{
    dlm_listnode_t * node;


    DBPRINTF((DBG_LOCK, "add_queue: adding\n"));
#ifdef DBPRINT
    clm_printlock(a);
#endif
#if defined(DEBUG)
    /* lock should not already be on queue */
    if ( check_queue( queue, a, ofs ) )
    {
	DBPRINTF((DBG_BASIC|DBG_RECOV,
		  "add_queue: lock 0x%x w/ lockid %x already on queue\n",
		  a, a->lck.lockid));
	abort("dup lock in add_queue");
    }
#endif

    node = (dlm_listnode_t *) OFSP( a, ofs );
    dlm_list_add_tail( queue, node );
}


/*
 * Name
 *	remove_queue
 * Description
 *	remove a reclock from the specified doubly linked queue
 * Arguments
 *	rlqp - pointer to queue head/tail struct
 *	a - the lock to be dequeued
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
remove_queue( dlm_list_t * queue, struct reclock *a, size_t ofs)
{
    dlm_listnode_t * node;


    DBPRINTF((DBG_LOCK, "remove_queue: dequeueing\n"));
#ifdef DBPRINT
    clm_printlock(a);
#endif

#if defined(DEBUG)
    /* If lock not found, we have a problem */
    if (! check_queue( queue, a, ofs ) )
    {
	DBPRINTF((DBG_BASIC|DBG_RECOV,
		  "remove_queue: lock 0x%x w/ lockid %x not on queue\n",
		  a, a->lck.lockid));
	abort("lock not found in remove_queue");
    }
#endif

    node = (dlm_listnode_t *) OFSP( a, ofs );
    dlm_list_remove( node );
}

/*
 * Name
 *	purgelocks
 * Description
 *	Purge all locks associated with site and pid
 * 	Pid 0 matches any pid. Call purgelist() for
 *	each queue on each resource. Remove the client registration.
 *	Cancel any pending ASTs
 * Arguments
 *	site - the site associated with the client to be purged
 *	pid - the pid of the client to be purged - 0 matches any pid
 *	flags - purge flags
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
purgelocks(int site,
	int pid,
	int flags)
{
    iterator_t * iter;
    struct resource * rp;
    unsigned   cnt;

    /* Loop counter.  Because of the way that it is used, this MUST BE
     * a signed integer.
     */

    int    k;


    cnt = 0;

    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_PURGELOCKS,
	    site, pid, flags, 0x00);

    DBPRINTF((DBG_LOCK|DBG_CONFIG|DBG_PURGE,
	      "Enter %s for site %d pid %d flags 0x%x\n",
	      __FUNCTION__, site, pid, flags));

    iter = resource_table_iterator_create();
    CLM_ASSERT( iter != NULL, "Unable to get restab iterator" );

    while ( ! iterator_next( iter, & rp ) )
    {
	if ( rp->refcount > 0 )
	{
	    unsigned   sub;


	    /* Remove from wait & convert first because purgelist has a side
	     * effect of moving OTHER locks around (mainly to grant q).
	     */

	    sub = 0;
	    for (k = RLQ_MAX - 1 ; k >= 0 ; k--) 
	    {
		sub += purgelist( rp, &rp->lockqs[k],
				  site, pid, flags, k );
	    }

	    cnt += sub;
	    /* FIXME - this is a bug waiting to happen! */
	    for (k = 0 ; k < sub ; k++)
		decr_delete_resource(rp, CH_NOREF) ;
	}
    }
    
    iterator_destroy( iter );


    if (site == OurSiteID)
    {
	int nast;

	nast = ASTpurge(pid, 0);
	DBPRINTF((DBG_LOCK, "%s: removed %d ASTs\n", __FUNCTION__, nast));

	if (flags & PURGE_DEAD) 
	{
	    /* if this client has req's on the bounce queue - kill em */
	    unbounce(pid);
	    /* ditto for the resend queue */
	    unresend(pid);
	}
    }

    DBPRINTF((DBG_LOCK, "%s: %d locks purged for site %d pid %d\n", 
	      __FUNCTION__, cnt, site, pid));
}

/*
 * Name
 *	purgelist
 * Description
 * 	Common list purging logic used by all three lock queues
 *	Calls lock model dependent wakeup if a lock is not being orphaned
 *	Does not delete UNIX NL locks unless DEAD because they are registrations
 * Arguments
 *	rp - the resource off which the speicified queue is linked
 *	rlqp - pointer to queue head/tail struct
 *		for reclock chain to be walked
 *	site - site number to match when looking for reclocks to purge
 *	pid - pid to match when looking for reclocks to purge - 0 matches all
 *	flags - purge flags
 *		PURGE_DEAD?? means purge all locks including orphans
 * Returns
 * 	number of locks purged
 * Global Data Affected
 *	NONE
 */
static int 
purgelist(struct resource *rp, dlm_list_t * queue,
		int site, int pid, int flags, int which_queue )
{
    unsigned         cnt = 0;
    struct reclock * curr;
    struct reclock * next;


    for ( curr = (struct reclock *) dlm_list_get_head( queue )
	  ; curr != NULL 
	  ; curr = next )
    {
	next = (struct reclock *) dlm_list_succ( (dlm_listnode_t *) curr );

	if ( (curr->lck.site == site)
	     && ((pid == 0) || (curr->lck.pid == pid)) )
	{
	    /* If the lock is orphanable and this call came from the purge
	     * daemon and the lock isn't on the wait queue, then orphan it.
	     */

	    if ( (curr->lck.lk_flags & LKM_ORPHAN)
		 && (flags & PURGE_DEAD)
		 && (which_queue != RLQ_WAIT) )
	    {
		DBPRINTF((DBG_LOCK, "%s: orphaning lock 0x%x\n",
			  __FUNCTION__, curr->lck.lockid));

		curr->lck.state |= LOCK_ORPHANED;

		/* If it's on the convert queue, put it back on the grant
		 * queue at its old mode
		 */

		if ( which_queue == RLQ_CONV )
		{
		    curr->lck.state &= ~LOCK_CONVERTING;
		    remove_queue( queue, curr, RL_RQ_OFS );
		    lck_set_seq(rp, curr);
		    delete_timeout_reclock( curr );
		    add_queue(&rp->lockqs[RLQ_GRANT], curr, RL_RQ_OFS);
		    curr->lck.state |= LOCK_GRANTED;
		    check_handles(rp, 0) ;

		    /* unblock any freed requests */
		    convert_locks(rp, curr);
		}

		/* The lock's purge function should only be called if the call
		 * came from purged or if it was made from a client and the
		 * lock is already an orphan OR if the client is purging
		 * itself.
		 *
		 * DO NOT purge NL locks on UNIX resources for clients because
		 * these represent registrations
		 */
	    }
	    else if ( (flags & PURGE_DEAD) 
		      || (curr->lck.state & LOCK_ORPHANED) 
		      || ((flags & PURGE_ME) 
			  && ((rp->type != DLM_RES_UNIX) 
			      || (curr->lck.mode != LKM_NLMODE))) )
	    {
		DBPRINTF((DBG_LOCK, "%s:\n", __FUNCTION__));
		clm_printlock( curr );
		cnt++ ;

		(lockswitch[rp->type].purge)(rp, queue, curr);
		/* don't decr refcount here - do at top level */
	    }
	}
    }

    return cnt;
}


/*
 * Name
 *	check_queue
 * Description
 *	Check a queue for consistency
 * 	Is the queue NULL terminated in both directions?
 *	Is the lock ptr to be enqueued or dequeued non-NULL?
 *      Check if lock is on queue.  If so, set found=TRUE.
 * Arguments
 *      rlqp - pointer to queue head/tail struct
 *      a - the lock to be enqueued or dequeued
 *      pfound - pointer to lock found on queue flag 
 *	ofs - where queue links are in struct reclock
 * Returns
 * 	Aborts if queue not OK otherwise,
 *      NOTHING
 * Global Data Affected
 *      NONE
 */

bool_t
check_queue( dlm_list_t * queue, struct reclock *a, size_t ofs )
{
    dlm_listnode_t * curr;
    struct reclock * plock;


    /* loop through list */
    for ( curr = dlm_list_get_head( queue )
	  ; curr != NULL
	  ; curr = dlm_list_succ( curr ) )
    {
	plock = (struct reclock *) (((void *) curr) - ofs);
	if ( plock == a )
	{
	    /* found lock on list already */
	    return( TRUE );
	}

	CLM_ASSERT((plock->lck.lockid != a->lck.lockid),
		   "check_queue: found other lock with same lockid");
    }


    return( FALSE );
}
