/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_lockid.c,v 1.6 2001/11/02 22:15:03 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: HASHFUNC
 *		LOCKIDTOINDEX
 *		LOCKIDTOSEG
 *		MSGSTR
 *		addmap
 *		free_hashent
 *		freemap
 *		get_hashent
 *		lockidtoindex
 *		mapid
 *		update_idhash
 */

#ifndef lint
#endif

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/types.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "clm_data.h"			/* external data definitions */
#include "dlm_kernel.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */

#define LOCKIDTOINDEX(id)	((id)&0xffff)
#define LOCKIDTOSEG(id)		(((id)>>16)&0xff)
#define LOCKIDTOSITE(id)	((id)>>28)

#ifndef __linux__
/* 
 * an entry in the hash table
 */
struct hashent {
	struct hashent *next;
	int remoteid;
	int localid;
};
#endif /* !__linux__ */

void free_hashlist(struct hashent *hep) ;
struct hashent *idhash[MAXNODES] = { NULL } ;	/* the lockid hash table */
struct hashent *hashent_free = NULL;  /* hash entry freelist */

int idhashsize;

#define HASHFUNC(remoteid)	((remoteid) % (idhashsize))



/*
 * Name
 *	clm_lockid_init
 * Description
 *	Initialize globals in clm_lockid.c
 * Arguments
 *	None
 * Returns 
 *	nothing
 * Global Data Affected
 *	hashent_free
 */
void
clm_lockid_init(void)
{
    (void) memset( idhash, 0, sizeof idhash );
    hashent_free = NULL;
    idhashsize = 0;
}

/*
 * clm_lockid_node_down
 *
 * deletes & frees any hash entries for a node which is down
 */

void
clm_lockid_node_down(int nodeid)
{
    int idx ;
    struct hashent *htp ;
    int i ;

    idx = NODEID_TO_INDEX(nodeid) ;
    if (idx == -1)
	return ;

    if (! idhashsize)
	return ;

    if ( ! (htp = idhash[idx]))
	return ;
    
    for (i = 0 ; i < idhashsize ; i++)
	if (htp[i].next)
	    free_hashlist(htp[i].next) ;

    (void) memset( idhash[idx], 0, idhashsize * sizeof *idhash[0] );
}

/*
 * Name
 *	update_idhash
 * Description
 *	Update idhash table because of topology changes
 * Arguments
 *	Pointer to map array
 * Returns 
 *	nothing
 * Global Data Affected
 *	idhash
 */
void
update_idhash(int	*idmap)
{
    int	i;
    struct hashent *new_idhash[MAXNODES];	/* new hash table */
						/* size MUST = idhash */
    /* init new table */
    (void) memset( new_idhash, 0, sizeof new_idhash );

    /* copy old pointers into new table */
    for (i = 0; i < MAXNODES; i++) {
	if (idmap[i] >= 0) {
	    new_idhash[i] = idhash[idmap[i]];
	    idhash[idmap[i]] = NULL;
	}
    }

    /* free up any leftover hash tables from removed nodes */
    for (i = 0; i < MAXNODES; i++) {
	if (idhash[i]) {
	    FREE(idhash[i]);
	}
    }

    /* copy new table over on top of old one */
    (void) memcpy( idhash, new_idhash, sizeof( idhash ) );


    return;
}


/*
 * Name
 *	free_hashent
 * Description
 *	free hash table entry
 * Arguments
 *	hash entry pointer
 * Returns 
 *	nothing
 * Global Data Affected
 *	hashent_free
 */
void
free_hashent(struct hashent	*hep)
{
	hep->next = hashent_free;
	hashent_free = hep;
}

/*
 * Name
 *	free_hashlist
 * Description
 *	free a list of hash table entries
 * Arguments
 *	hash entry pointer
 * Returns 
 *	nothing
 * Global Data Affected
 *	hashent_free
 */
void
free_hashlist(struct hashent *hep)
{
    struct hashent *next ;
	/* search for last entry in list */
    for (next = hep ; next->next ; next = next->next)
	;
    next->next = hashent_free ;
    hashent_free = hep ;
}

/*
 * Name
 *	get_hashent
 * Description
 *	allocate another hash table entry
 * Arguments
 *	none
 * Returns 
 *	pointer to the table entry
 *	0 for error
 * Global Data Affected
 *	hashent_free
 */
struct hashent *
get_hashent(void)
{
    struct hashent	*hep;

	if (hashent_free) {
		/* here if allocating from the freelist */
		hep = hashent_free;
		hashent_free = hep->next;
		hep->next = NULL;
		return hep;
	} else {
		/*
		 * here if freelist is empty - link in new block of
		 * hash entry structures and hand back first
		 */
		register int i;
		struct hashent *hashblock = (struct hashent *)MALLOC(
				1024 * sizeof(struct hashent));

#if defined(MEMCHK)
		bsdlog(LOG_DEBUG, "Allocating new block of map hash entries\n");
#endif
		if (!hashblock)
			return NULL;
		for (i=1, hashent_free=hep=&hashblock[1]; i<1024; hep++, i++) {
			hep->next = hep + 1;
		}
		hashblock->next = (--hep)->next = NULL;
		return hashblock;
	}
}

/*
 * Name
 *	freemap
 * Description
 *	remove a lockid from the map table - table entry must match both
 *	lock ids.
 * Arguments
 *	remoteid - the remote lockid that is to be unmapped
 *	localid - the local lockid that is to be unmapped
 * Returns 
 *	0 for success
 *	-1 for error (no map allocated or map entry not found)
 * Global Data Affected
 *	idhash
 */
int
freemap(uint remoteid, uint localid, int remotenode)
{
	struct hashent *hep;
	struct hashent *prev;
	struct hashent *table ;
	int idx ;

	if ((idx = NODEID_TO_INDEX(remotenode)) == -1)
	{
	    bsdlog(LOG_CRIT, "freemap: node %d not active\n", remotenode) ;
	    return -1 ;
	}

	if ((table = idhash[idx]) == NULL)
	{
		bsdlog(LOG_ERR, MSGSTR(FREEMAP_ERR, "freemap: not allocated"));
		abort("freemap: hash table not allocated");
	}

	prev = &table[HASHFUNC(remoteid)];

	DBPRINTF((DBG_BASIC, "freemap: freeing remoteid 0x%x slot 0x%x\n", 
		remoteid, prev - table));
		
	hep = prev->next;
	while (hep != NULL) {
		if (hep->remoteid == remoteid && hep->localid == localid)
			break;
		prev = hep;
		hep = hep->next;
	}
	if (hep == NULL) {
		bsdlog(LOG_CRIT, MSGSTR(REMOTE_FREEMAP, 
			"freemap: can't find map for remoteid id 0x%x"),
			remoteid);
		CLM_ASSERT((0), "freemap");
		return -1;
	}
	prev->next = hep->next;
	free_hashent(hep);
	return 0;
}

/*
 * Name
 *	addmap
 * Description
 * 	Add a lockid to the map table. It is a fatal error to attempt to
 * 	map the same remote lockid more than once
 *	Well, it used to be fatal.  We now allow remote ids to be mapped
 *	more than once because in situations where we have migrated a
 *	resource to a different node and are waiting for the ACK.  In the
 *	meantime, a lock local to the node where we migrated to (that we
 *	have a mapping for) is unlocked and then reused on a resource that
 *	is mastered here.
 * Arguments
 * 	remoteid - the remote lockid that is to be mapped
 * 	localid - the id to which the remote should be mapped
 * Returns
 *	0 for success
 *	-1 for error (malloc failure)
 * Global Data Affected
 *	idhash
 */
int
addmap(uint remoteid, uint localid, int remotenode)
{
	struct hashent *hep;
	struct hashent *slotp, *prev;
	struct hashent *table ;
	struct hashent *slot ;
	int idx ;
	int slotidx ;

	if ((idx = NODEID_TO_INDEX(remotenode)) == -1)
	{
	    bsdlog(LOG_CRIT, "addmap: node %d not active\n", remotenode) ;
	    return -1 ;
	}

	/* allocate hash table first time in */
	if ( ! (table = idhash[idx]))
	{
		idhashsize = lock_seg_size * 8;
  		table = (struct hashent *)
                                 MALLOC(idhashsize * sizeof(struct hashent));
		if ( ! table)
		{
			bsdlog(LOG_ERR, MSGSTR(HASHTABLE_ERR,
					"addmap: can't malloc hash table"));
			return -1;
		}
		(void) memset( table, 0, idhashsize * sizeof(struct hashent) );
		idhash[idx] = table ;
	}

	/*
	 * We used to only execute this block for debugging, but have since
	 * seen situations caused by migrating resource where a new lock
	 * request will come in with a remote lock id that we already (or
	 * still) have because it's queued to a resource that has migrated
	 * to another node and is being reused and sent to us before the
	 * migration acknowledgement has reached us.  Blech.
	 */
	slotidx = HASHFUNC(remoteid) ;
	slot = table + slotidx ;	/* for debug keep record */
	for (prev = slot, slotp = prev->next ; slotp ;
				prev = slotp, slotp = slotp->next)
	{
		/*
		 * if there is a duplicate map entry, then we can assume
		 * that the existing entry is only there until a migration
		 * acknowledgement is received.  Therefore, we want to add the
		 * new entry onto the hash chain *before* the existing one so
		 * that any lookups will find the new entry.
		 */
		if (slotp->remoteid == remoteid) {
			bsdlog(LOG_CRIT, MSGSTR(HASHED_REMOTE_ID,
					"ERROR: remoteid 0x%x already hashed"), 
					remoteid);
			/* this will leave prev pointing to entry point */
			break;
		}
	}

#if defined(DEBUG)
    if ( (remotenode == 0) || !clm_node_active( remotenode ) )
    {
	abort("Invalid remote id");
    }
#endif

	if ((hep = (struct hashent *)get_hashent()) == NULL) {
		bsdlog(LOG_ERR, "addmap: can't malloc hash entry");
		return -1;
	}

	hep->next = prev->next;
	prev->next = hep;
	hep->remoteid = remoteid;
	hep->localid = localid;
	DBPRINTF((DBG_BASIC,
		"addmap: added site %d remote 0x%x local 0x%x slot 0x%x\n", 
		remotenode, remoteid, localid, slotidx));
	
	return 0;
}

/*
 * Name
 *	mapid
 * Description
 * 	Map a secondary lockid to a local lockid. It is assumed that
 *	this routine will only run on the primary
 * Arguments
 *	remoteid - the remote lockid to be looked up
 * Returns
 * 	the local lockid 
 *	0 for error
 * Global Data Affected
 *	NONE
 */
uint
mapid(uint remoteid, int remotenode)
{
	struct hashent *hep, *ret_ptr = NULL;
	struct hashent *table ;
	int idx ;

	if ((idx = NODEID_TO_INDEX(remotenode)) == -1)
	{
	    bsdlog(LOG_CRIT, "mapid: node %d not active\n", remotenode) ;
	    return -1 ;
	}

	if ( ! (table = idhash[idx]))
	{
	    bsdlog(LOG_CRIT, MSGSTR(MAPID_ERR, "mapid: not allocated"));
	    return 0;
	}
		
	hep = &table[HASHFUNC(remoteid)];
	DBPRINTF((DBG_BASIC, "mapid: slot 0x%x\n", hep - table));
	for (hep = hep->next; hep != NULL; hep = hep->next) {
		if (hep->remoteid == remoteid) {
			DBPRINTF((1, "mapping remote 0x%x to local 0x%x\n",
				hep->remoteid, hep->localid));
			if (!ret_ptr) {
				ret_ptr = hep;
			} else {
				bsdlog(LOG_DEBUG, "mapid: found another map for "
					"0x%x -> 0x%x\n", remoteid,
					hep->localid);
			}
		}
	}

	if (ret_ptr) {
		return ret_ptr->localid;
	}

	DBPRINTF((DBG_LOCK, "mapid: can't find map for remoteid 0x%x",
			remoteid));
	return 0;
}


/*
 * Name
 *	lockidtoindex
 * Description
 * 	given a lockid, return an index into the lock table
 *	associated with the lockid. It is up to the caller to
 *	to determine which lock table is relevant. The segment
 *	argument will be set to the correct segment.  No error
 *	checking is performed.
 * Arguments
 * 	id - lockid
 *	seg - returns segment value
 * Returns
 *	an index into the lock table
 * Global data Affected
 *	NONE
 */
int
lockidtoindex(int id, unsigned int *seg)
{
	*seg = LOCKIDTOSEG(id);

	return LOCKIDTOINDEX(id);
}




/****************************************************************************/
/* dlm_map_table_print - Print the lock ID map table
 *
 * Walk through the lock ID map table and print the values of each entry to
 * the provided buffer.
 *
 * RETURNS
 *
 * Total number of characters copied to the buffer.
 */

unsigned
dlm_map_table_print( char * buf, unsigned limit )
{
    struct hashent * slotp;
    struct hashent * table;
    unsigned         index;
    unsigned         i;
    unsigned         bytes;
    unsigned         node;


    bytes = snprintf( buf, limit, 
		      "Node ID | Index | Remote ID | Local ID\n" );

    for (i = 0 ; (i < MAXNODES) && (bytes < limit) ; i++)
    {
	if ((table = idhash[i]))
	{
	    node = INDEX_TO_NODEID( i );

	    for ( index = 0 
		  ; (index < idhashsize)  && (bytes < limit)
		  ; index++ )
	    {
		for (slotp = table[index].next ; slotp ;
		     slotp = slotp->next)
		{
		    bytes += snprintf( & buf[ bytes ], limit - bytes,
				       "%7u | %5u |  %08X | %08X\n",
				       node, index,
				       slotp->remoteid, slotp->localid );
		}
	    }
	}
    }


    return( bytes );
}
