/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_main.c,v 1.29 2001/11/28 00:33:20 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		clm_sleep
 *		get_lockd_pid
 *		kern_main
 *		sig_handle
 *		start_lockd
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/types.h>
#include <linux/param.h>
#include <linux/ipc.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/ioctl.h>
#include <linux/time.h>
#include <linux/timer.h>
#include <linux/resource.h>
#include <linux/socket.h>
#include <linux/poll.h>
#include <linux/in.h>
#include <linux/tcp.h>
#include <asm/uaccess.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,18)
# define DECLARE_WAIT_QUEUE_HEAD(x) struct wait_queue *x = NULL
# define init_waitqueue_head(x)
#endif

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "cltrc.h"
#include "clm_data.h"
#undef LOCK_FREE
#include "dlm_kernel.h"
#include "clm_locksw.h"
#include "clm_reconfig.h"
#include "cccp.h"
#include "dlm_version.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_workqueue.h"
#include "dlm_recover.h"
#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
#define DEAD_COUNT		2   /* main loop cycles per ddlk check */

#ifdef DEBUG_ALL_PROTO /* hide from prototype file but not debugger */
#define STATIC
#else
#define STATIC static
#endif

static int kern_main( void * not_used ) __attribute__ ((noreturn));

static int clm_batchsleep = 250000;	/* 1/4 sec in us */

struct callbacklist
{
    int cbl_port ;		/* cccp 'port' number to listen on */
    cccp_callback_t * cbl_routine;	/* message handler routine */
    int cbl_nlscode ;		/* NLS lookup code for error message */
    char *cbl_errmessage ;	/* error message if listen fails */
} ;

/*
 * nls-style messages
 */
#define MSGTAB(code, str) code, "cllockd: unable to init " str " port"
static const struct callbacklist cb_tab[] =
{
	{ PORT_MSTR, pti_read_request,
			MSGTAB(CANT_INIT_PTI_PORT, "PTI request") },
	{ PORT_DIRECT, dir_read_request,
			MSGTAB(CANT_INIT_DIR_PORT, "directory") },
	{ PORT_RESPONSE, pti_read_response,
			MSGTAB(CANT_INIT_RES_PORT, "PTI response") },
	{ PORT_DIRUPDT, dir_read_update,
			MSGTAB(CANT_INIT_RES_PORT, "dir update") },
	{ PORT_MSTR_CR, pti_read_request,
			MSGTAB(CANT_INIT_RES_PORT, "master create") },
	{ PORT_BOUNCE, dir_bounce_request,
			MSGTAB(CANT_INIT_RES_PORT, "bounce") },
	{ PORT_GROUP, grp_proc_request,
			MSGTAB(CANT_INIT_RES_PORT, "group") },
	{ PORT_NACK, clmll_read_nack_response,
			MSGTAB(CANT_INIT_RES_PORT, "nack") },
	{ PORT_SYNC, clm_inband_sync,
			MSGTAB(CANT_INIT_RES_PORT, "sync") },
	{ PORT_SYNCRET, clm_inband_sync_resp,
			MSGTAB(CANT_INIT_RES_PORT, "syncret") },
	{ -1, 0, 0 }
} ;

arg_struct_t *args;

STATIC void clm_master_loop(void) ;

static dlm_workqueue_t   dlm_private_work_queue;
dlm_workqueue_t * dlm_master_work_queue = & dlm_private_work_queue;
bool_t dlm_is_going_down = FALSE;

static DECLARE_MUTEX_LOCKED( dlmdk_is_inited );

/* ============================================================ */

static int the_u_error;
int setuerror(int error)
{
  return(the_u_error = error);
}
int getuerror(void)
{
  return(the_u_error);
}

/* ============================================================ */
/*
 * Function:
 *	kern_main
 *
 * Description:
 *	Does all the stuff that the old main() used to do after argument
 * processing, which wasn't very well-documented before either.  Basically
 * it's just a bunch of initialization and information-gathering stuff.
 *
 * clm_active ("we are alive" flag) set when DIRECTORY has been updated,
 *    NOT earlier
 *
 * 	return is not used - exit is in dlm_exit() in dlm_kernel.c
 *
 * Arguments:
 *	argc and argv, just as though it's a user program
 *
 * Return value:
 *	NONE
 *
 * Side effects:
 *	Oh, not much.  Just the minor effect that while we're in this routine
 * the lock manager is functional and when we exit it's not.
 */

char main_compiled[MAXLINELEN];

extern int cccp_init( void );
extern unsigned long sys_signal(int sig, __sighandler_t handler);

static int
kern_main(void *not_used)
{
  static int	been_there = FALSE;

  snprintf(main_compiled, sizeof( main_compiled ),
	   "DLMK Major/Minor(%d/%d/%p) Time(%s) Date(%s)\n",
	   DLM_MAJOR_VERSION,
	   DLM_MINOR_VERSION,
	   (void *)DLM_VERSION,
	   __TIME__,
	   __DATE__);


  dlm_set_thread_name( current, args->dlmprocname );

  bsdlog( LOG_INFO, main_compiled );

  dlm_daemonize();


  TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_KERN_MAIN,
          0x00, 0x00, 0x00, 0x00);

  /* init lock manager memory segment before anything else */
  clm_heap_init();

  clm_data_init();	/* clear EVERYTHING global in sight */
  locks_init();
  ast_init();
  clm_mem_init();
  clm_alloc_init();
  clm_batch_init();
  clm_client_init();
  clm_direct_init();
  clm_group_init();
  clm_lockid_init();
  clm_migrate_init();
  clm_resource_init( args->ressegsize, args->resprealloc, args->hashsize );
  clm_rldb_init();
  clm_timeout_init();
  sched_queue_init() ;
	
  clmll_init(); /* Local Locks: Initialize LLB List */


  /* check to see if we have been used */
  if (been_there) {
    DBPRINTF((DBG_CONFIG|DBG_BASIC,
              "Discarded kernel extension.  Attempt to recycle."));
  } else {
    been_there = TRUE;	/* done that */
  }

  /* what else did our predecessor leave lying around? */
  if (kproc_pid) {
    bsdlog(LOG_DEBUG, "kern_main: kproc_pid = %d\n", kproc_pid);
  }

  dlm_debug	= args->debug;
  local_debug	= args->local_debug;
  clm_batchsleep	= args->batch_sleep;
  lock_seg_size	= args->locksegsize;
  kmem_cutoff	= args->kmem_cutoff;

  DBPRINTF((DBG_BASIC, "Resource segment size set to %d resources "
	    "with %d segments preallocated\n", 
	    args->ressegsize, args->resprealloc));

  DBPRINTF((DBG_BASIC, "Lock table segment size set to %d locks\n",
			lock_seg_size));

  acquire_lock(LK_ENTRY, kproc_pid);

  /*
   * At this point, what we need to set up is a "queue" where
   * our user-side admin partner can dump in updates to us.
   * We mostly just cancel stuff out for now to get through
   * a very basic mainline.
   */

  dlm_workqueue_init( dlm_master_work_queue );


  setuerror(0);

  /* Install the listening callback functions.  */
  if ( CPERR_OK != dlm_cccp_connection( TRUE ) ) {
      abort("cllockd: cccp_listen");
  }


  /* initialize the RLDB hash table */
  if (DLM_NORMAL != rl_init(100000)) {
    bsdlog(LOG_CRIT, MSGSTR(CANT_INIT_RLDB,
                            "cllockd: unable to init RLDB"));
    abort("cllockd: rl_init");
  }

  /* initialize the migration module */
  if (CPERR_OK != clmm_init(args->recalc_time, args->decay_rate)) {
    bsdlog(LOG_CRIT, MSGSTR(CANT_INIT_CLMM,
                            "cllockd: unable to init CLMM"));
    abort("cllockd: clmm_init");
  }

  /* initialize the deadlock detection module */
  if (CPERR_OK != clmdd_init()) {
    bsdlog(LOG_CRIT, MSGSTR(CANT_INIT_DEADLOCK_DET,
                            "cllockd: unable to init deadlock detection"));
    abort("cllockd: clmdd_init");
  }

  /* initialize the lock manager recovery module */
    if ( dlm_recov_init() != 0 )
    {
	bsdlog(LOG_CRIT,"cllockd: error integrating new lock manager");
	abort("cllockd: dlm_recov_init");
    }


  /* Let the parent process (i.e., the user-mode daemon that started us)
   * know that it can continue on.
   */

  release_lock( LK_ENTRY, kproc_pid );
  up( & dlmdk_is_inited );


  clm_master_loop() ;		/* main code */
  dlm_exit();
  /*NOTREACHED*/
}


void
clm_alarm_handler( unsigned timer )
{
    struct timer_list * t;
    
    
    t = (struct timer_list *) timer;
    t->data = 0UL;

    up( & dlm_master_work_queue->sema );
}

void
clm_set_alarm( void )
{
    static struct timer_list  t = { data: 0UL };
    
    if ( t.data == 0UL )
    {
	init_timer( & t );
	t.expires = jiffies + (3 * HZ);
	t.function = (void *) clm_alarm_handler;
	t.data = (unsigned long) & t;

	add_timer( & t );
    }
}




/*
 * The following loop is the heart of control in the
 * lock manager. Execution blocks in the fp_poll system call
 * waiting for I/O from another lock manager, the cluster manager,
 * or the deadman.
 * If no I/O is pending, the fp_poll call will return after timing out.
 * If for any of the following reasons we need to we check:
 *	The transaction batching timer has fired and it is time
 *		to transmit to the primary
 *	A grace period is in effect and the grace timer needs to be
 *		counted down
 *	It is time to perform global deadlock detection
 *
 * The only return is via longjmpx from an error condition or
 * a shutdown message.
 */

/*
 * these are blobs of related functionality extracted from the loop
 */
STATIC void do_timer_stuff( void );

STATIC void
clm_master_loop(void)
{
    int   t;
    int   ret;
    dlm_workunit_t * wu;

    bsdlog(LOG_INFO,
           "in %s/%d\n", __FUNCTION__, __LINE__);
    
    acquire_lock( LK_ENTRY, kproc_pid );
    clm_set_alarm();

    while ( ! dlm_is_going_down ) 
    {
	/* note: we assume it is always okay to write */

	pti_flush_responses();
	release_lock( LK_ENTRY, kproc_pid );

	wu = dlm_workqueue_get_work( dlm_master_work_queue, 0 );
	acquire_lock( LK_ENTRY, kproc_pid );

	if ( wu == NULL )
	{
	    /* BUGBUG In the AIX code, this case would also call 
	     * BUGBUG clm_handle_signals.  Something needs to be done with
	     * BUGBUG that here, too.
	     */

	    do_timer_stuff();
	    clm_set_alarm();

	    if ( STATE_CURR( dlm_recov_state_machine ) == CLM_ST_RUN )
	    {
		purge_list();
	    }
	}
	else
	{
	    dlm_process_cm_message( wu );
	    dlm_workunit_dispose( wu );
	    wu = NULL;
	}


	ret = new_change_request(&t);
	if ( (0 != ret) 
	    && ( (t == CLM_MSG_NODE_DOWN)
		|| (STATE_CURR( dlm_recov_state_machine ) == CLM_ST_RUN) ) )
	{
	    rc_msg_t * dat;


	    seen_chg(ret, t) ;
	    dat = rc_msg_alloc( t, ret, NULL );
	    hsm_process_event( dlm_recov_state_machine, (hsm_msg_t *) dat );
	}
    }
}




/****************************************************************************/

int
dlm_cccp_connection( const bool_t connect_or_disconnect )
{
    unsigned   i;
    int        status;
    cccp_callback_t * cb;


    status = CPERR_OK;
    for ( i = 0 ; (cb_tab[i].cbl_port >= 0) && (status == CPERR_OK) ; i++)
    {
	cb = (connect_or_disconnect) ? cb_tab[i].cbl_routine : NULL;
	status = cccp_listen( cb_tab[i].cbl_port, cb );
    }


    if ( status != CPERR_OK )
    {
	bsdlog(LOG_CRIT, MSGSTR(cb_tab[i].cbl_nlscode,
				cb_tab[i].cbl_errmessage));
    }

    return( status );
}

static int inited = 0;

/*
 * Function:
 *	start_lockd
 *
 * Description:
 *	This is the kernel routine called from user-land to get the lock
 * manager ball rolling.  Its only real reponsibility is to start the kernel
 * process.
 *
 * Arguments:
 *	arg_struct *argptr;		-- the processed results of whatever
 *					-- the user put on the command line,
 *					-- along with the port numbers for
 *					-- various services, since those are
 *					-- easier to get from user-land
 *	int argsz;			-- size of *argptr
 *
 * Return value:
 *	pid of kproc, or -1 on error
 *
 * Side effects:
 *	NONE
 */

int
start_lockd (arg_struct_t *argptr, int argsz)
{
  int pid;

  if (argsz != sizeof(arg_struct_t))
	return(ERROR);

  if (NULL == (args = kmalloc(sizeof(arg_struct_t), GFP_KERNEL))) {
    bsdlog(LOG_CRIT,
           "Cannot allocate args struct.\n");
    return(ERROR);
  }

  if (copy_from_user(args, argptr, argsz) == ERROR)
	return(ERROR);

  /* Running without reinitializing can be disastrous.  -jjd- */
  if (clm_active) {
	bsdlog(LOG_CRIT, "Lock manager already running!!!");
	/* if we call our "exit", we may stop something we don't want to */
	_exit(1);
  } else if (inited) {
	bsdlog(LOG_CRIT, "Lock daemon is being restarted");
  }
  inited = 1;

#ifdef __linux__
  /*
   * uproc_pid is set in dlm_kerndd.c
   *
   * Something interesting here, as the kernel thread inherits the
   * open file descriptors from the user process (the dlmdu user
   * side admin client!)  This causes a problem if dlmdu exits,
   * since that should cause the /dev/haDLM/admin device to be
   * closed and tell us to shut down.  But, since when that happens
   * the kernel thread is still around, we don't see the device
   * get closed.
   *
   * Sigh.  Down in kern_main() need to close the descriptors.
   */
  pid = kernel_thread(kern_main, NULL, 0);
  bsdlog(LOG_DEBUG, "We are pid %d created kernel thread %d\n",
         uproc_pid,
         pid);

  /* Wait here until the daemon is fully initialized.
   */

  down( & dlmdk_is_inited );
  bsdlog( LOG_INFO, "[%s:%s] kernel thread initialization complete\n",
	  current->comm, __FUNCTION__ );

#else
  pid = creatp();
  if (pid == ERROR)
	return(ERROR);

  if (initp(pid,kern_main,NULL,0,"k_cllockd") == ERROR)
	return(ERROR);

  /*
   * Pid of our user-mode admin partner.
   */
  uproc_pid = current->pid;
#endif /* __linux__ */
  return(pid);
}


/*
 * Function:
 *	clm_sleep
 *
 * Description:
 *	Sleep for the specified number of seconds, but check the 
 *	nsleep return for signal interruptions and ignore all but KILL
 *
 * Arguments:
 *	number of seconds to sleep
 *
 * Return value:
 *	Not currently used.
 *
 * Side effects:
 *	NONE
 *
 */
void
clm_sleep(int	secs)
{
    long   wakeup;
    DECLARE_WAIT_QUEUE_HEAD( wait );


    /* The sleep call returns the number of jiffies left to sleep from
     * the original timeout.
     */
    wakeup = secs * HZ;

    init_waitqueue_head( & wait );
    do
    {
      /* BUGBUG We should just hit the signal handler if any hit, unless
       * BUGBUG they're blocked...  If they are blocked, do we even care
       * BUGBUG to get them?
       */
#if 0
      if (signal_pending()) {
        sig_handle(current->sigpending);
      }
#endif
	wakeup = interruptible_sleep_on_timeout( & wait, wakeup );
    } while( wakeup > 0 );
}

/*
 * Name
 *	clm_data_init
 * Description
 * 	Init all static structures in clm_data.c
 * Arguments
 *	None
 * Returns
 *	Nothing
 * Global Data Affected
 *	freelocks
 */
void
clm_data_init(void)
{
	OurSiteID = -1;			/* lock manager site id */
	nclients = 0;			/* number of registered applications */

	nbatched = 0;     /* number of entries waiting to be xmitted */
	ninflight = 0;    /* number of entries xmitted but not returned */
	batchsize = 1;    /* send after this many pending */

	bounce_list = NULL;

	dlm_debug = 0 ;		/* enables debug stream if > 0 */
	local_debug = 0 ;	/* pseudo host id */
	grace_period = 0 ;

	lock_seg_size = 0 ;	/* lock table segment size */
	clm_lock_warned = 0 ;
	clm_res_warned = 0 ;    /* # times warned for no more space for res */
	clm_items_max = 0;

#if defined(TIME_STATS)
	(void) memset( global_stats, 0, sizeof global_stats );
#endif
	clm_active = FALSE ;	/* TRUE is lock manager is running */
}

/*
 * external interface routines
 * provide status to control other module behaviors
 * entry points to signal topology change
 */

/*
 * clm_in_reconfig
 *
 * TRUE if master state is in reconfig (but not in drainoff)
 * FALSE otherwise
 */
int
clm_in_reconfig(void)
{
    return dlm_recov_in_reconfig;
}

/*
 * clm_is_running
 *
 * TRUE if master state is running (but not in drainoff)
 * FALSE otherwise
 */

int
clm_is_running(void)
{
    return( STATE_CURR( dlm_recov_state_machine ) == CLM_ST_RUN );
}

/*
 * clm_in_delivery
 *
 * TRUE if master state is delivering directory data
 * FALSE otherwise
 */

int
clm_in_delivery(void)
{
    return( STATE_CURR( dlm_recov_state_machine ) == RC_DIR_INIT ); 
}

/*
 * clm_run_rcv
 *
 * TRUE if in a state where it is legal to receive transactions
 * or to continue them
 */
int
clm_run_rcv(void)
{
    return dlm_recov_can_transact;
}


/*
 * do_timer_stuff
 *
 * all time or iteration based stuff here:
 *    request timeout
 *    warnings for late msgs from other lock managers
 *    deadlock detection
 *    purge of anything held by exited processes
 *
 * takes encoded fp_poll() results as input
 */

STATIC void
do_timer_stuff( void )
{
  static int dead_count = 0;

  /*
   *  process deadlock list once every DEAD_COUNT iterations
   *  or if timer expires.
   */
    if ( STATE_CURR( dlm_recov_state_machine ) == CLM_ST_RUN )
    {
      /*
       * we have to check the timeout queue to process any
       * timeouts that may have occured since the last time we
       * were here.
       */
      timer_tick();

      /* check bounced requests to cover timing window */
      resend_held(NULL);

      dead_count++;
      if ( dead_count >= DEAD_COUNT )
        {
          dead_count = 0;
          clmdd_task();
        }
    }
}
