/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_migrate.c,v 1.15 2001/10/29 21:49:46 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: clmm_bundle
 *		clmm_bundle_compat
 *		clmm_cache
 *		clmm_callback
 *		clmm_chain
 *		clmm_chain_compat
 *		clmm_copy
 *		clmm_ctrl
 *		clmm_ftoi
 *		clmm_incoming
 *		clmm_init
 *		clmm_lookup
 *		clmm_master2slave
 *		clmm_params
 *		clmm_resource
 *		clmm_slave2master
 *		clmm_strip
 *		clmm_trigger
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/types.h>
#include <linux/time.h>
#include <linux/socket.h>
#include <linux/in.h>

#include "dlm_version.h"
#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#undef LOCK_FREE
#include "dlm_kernel.h"
#include "clm_data.h"

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
#ifdef UNIT_TEST
/* MIGRATION/PURGE UNIT TEST */
/* used to test correct overlap between migration and purge */
int unit_test_hold_migrations = 0 ;	/* 1 to delay any migration */
int unit_test_force_migrations = 0 ;	/* 1 to force migration when allowed */
#endif

/* File scope variables */
static struct clmm_cb Clmm_cb;			/* Module control block */

/* Function Prototypes */
int clmm_incoming(clm_msg_t *msg_in);

static void clmm_replace_leftovers( dlm_list_t * queue, struct resource *rp );

#define KFLOAT_1	0x3f800000	/* floating 1.0 as binary */

/*
 | ~~~~~~~~~~~~~~~~~~~~~~
 | RMM External Functions
 | ~~~~~~~~~~~~~~~~~~~~~~
 */

/*
 | Name
 |	clm_migrate_init
 | Description
 |	This function inits static variables in clm_migrate.c
 | Arguments
 |	None
 | Returns
 |	Nothing
 | Global Data Affected
 |      None
 */
void
clm_migrate_init(void)
{
}

/*
 | Name
 |	clmm_trigger
 | Description
 |	This function provides the only upper level entry point for
 |	triggering resource migration. This routine increments the
 | 	statistics required for the resource, and triggers a migration
 |	if necessary. 
 | 	Note that node ids are remapped to their cell values!
 | 	[Node ID 1 is stored in array cell 1, node ID 0 is bogus]
 *
 * Use 'kfl_' fake floating point in kernel - real float ops
 * are nowhere allowed.
 *
 | Arguments
 |	None.
 | Returns
 |	0 if no resource migration is necessary.
 |      -1 in case of error
 |      Value returned from clmm_resource() if migration was triggered.
 | Global Data Affected
 |      Clmm_cb
 */
int
clmm_trigger(void)
{
  struct resource *rptr;
  clmm_info_t *iptr;			/* migration info ptr	*/
  int node;				/* nodeid to be used	*/
  int i;				/* ACME ints & indices	*/
  int ndx ;				/* node index */
  int best_ndx = 0 ;			/* index of node to migr to */
  int our_ndx = NODEID_TO_INDEX(OurSiteID) ;
  int maxnodes = clm_node_count() ;	/* number of nodes in cluster */
  int ret ;


    bsdlog( LOG_DEBUG, "[%s:%s] DLM is%s running, our index = %d, "
	    "maxnodes = %d\n", current->comm, __FUNCTION__,
	    (clm_is_running()) ? "" : " not", our_ndx, maxnodes );

    rptr = Clmm_cb.mcb_rcache;
    node = Clmm_cb.mcb_ncache;

    if ( (rptr == NULL) || (node == 0) || ! clm_is_running()
#ifdef UNIT_TEST
	 || (unit_test_hold_migrations && node != OurSiteID)
#endif
	) 
    {
	return -1 ;
    }

    /* reset cache */
    Clmm_cb.mcb_rcache = NULL;
    Clmm_cb.mcb_ncache = 0;
    
    if ( clm_in_reconfig() )
    {
	return 0 ;
    }


    /* check if active */
    if ( ~(Clmm_cb.mcb_state) & MCB_SEND )
    {
	DBPRINTF((DBG_MIGR, "[%s:%s] module disabled <0x%x>\n",
		  current->comm, __FUNCTION__, Clmm_cb.mcb_state));
	return -1;
    }

    ndx = NODEID_TO_INDEX( node );
    if ( ndx < 0 )
    {
	DBPRINTF((DBG_MIGR, "[%s:%s] node id [%d] out of range\n",
		  current->comm, __FUNCTION__, node));
	return -1;
    }

  check_handles(rptr, 0) ;
  /*
   * check for stuck or updating flag in case previous migration incomplete
   * and make sure that we *are* the master copy!!
   */
  if (rptr->master || (rptr->rsrc_flags & (RSRC_STUCK | RSRC_UPDATING))) {
	return -1;
  }

  /* first update stats */
  iptr = &rptr->clminfo;
  iptr->mi_counter[ndx]++;
  iptr->mi_event_ttl++;
    
  /* check for recalc/migration */
  if (iptr->mi_event_ttl >= Clmm_cb.mcb_event_run
#ifdef UNIT_TEST
      || unit_test_force_migrations
#endif
      )
    {
      struct timeval t1, tdelta;
      uint64_t best_aps, total_aps;
      int best_site = OurSiteID;
      uint64_t decay_factor, aps_decay, calc_aps ;

      iptr->mi_event_ttl = 0;		/* reset for next run */
      best_aps = total_aps = 0;

      /* calculate time delta */
      if (iptr->mi_timestamp.tv_sec == 0) 
	    gettimeofday(&iptr->mi_timestamp, NULL);
      gettimeofday(&t1, NULL);
      timeval_sub(&t1, &iptr->mi_timestamp, &tdelta);
      if (tdelta.tv_sec == 0)
	    tdelta.tv_sec++;

      DBPRINTF((DBG_MIGR,
                "c %d, tsp.sec = %d, td.sec = %d\n",
                t1.tv_sec,
                iptr->mi_timestamp.tv_sec, tdelta.tv_sec));

      /* calculate the decay factor once */
      aps_decay = Clmm_cb.mcb_aps_decay ;
      for (decay_factor = KFLOAT_1, i = 0; i < tdelta.tv_sec; i++)
        {
          decay_factor = kfl_mul(decay_factor, aps_decay) ;
        }

      /* update aps for all nodes and track highest */
      for (i = 0 ; i < maxnodes ; i++)
        {
          iptr->mi_aps[i] = kfl_mul(decay_factor, iptr->mi_aps[i]) ;
          iptr->mi_aps[i] = kfl_add(iptr->mi_aps[i],
                                    (kfl_idivi(iptr->mi_counter[i], tdelta.tv_sec))) ;

          DBPRINTF((DBG_MIGR, "[%s:%s] i %d, ctr %d, td %d, aps %s\n",
		    current->comm, __FUNCTION__, i, iptr->mi_counter[i],
                    tdelta.tv_sec, kfl_ftos(iptr->mi_aps[i])));

          iptr->mi_counter[i] = 0;	    	/* clear old counters	*/
          if (iptr->mi_aps[i] > best_aps) {	/* check if best	*/
            best_aps = iptr->mi_aps[i];
            best_ndx = i ;
          }
          total_aps = kfl_add(total_aps, iptr->mi_aps[i]) ;
        }
    
      DBPRINTF((DBG_MIGR, "[%s:%s] target %d, ctr %d, aps %s\n",
		current->comm, __FUNCTION__, best_site,
		iptr->mi_counter[best_site],
		kfl_ftos(iptr->mi_aps[best_site]))) ;

      /* prime new clock value */
      gettimeofday(&iptr->mi_timestamp, NULL);

      /*
       * Resource Statistics:
       *        cs_total_aps  - resource's total aps
       *        cs_aps[]      - aps per node
       */
      rptr->stats.cs_total_aps = total_aps ;
      /* could cause a problem because array sizes different */
      for (i = 0 ; i < maxnodes && i < DLM_MAXNODES ; i++)
        {
          rptr->stats.cs_aps[i] = iptr->mi_aps[i] ;
        }

      best_site = INDEX_TO_NODEID(best_ndx) ;
      /*
       * Use 'kfl' fake floating point in kernel - real float ops
       * are nowhere allowed.
       */
      calc_aps =
		kfl_add(
                iptr->mi_aps[our_ndx],
                kfl_mul(
                        kfl_idivi(iptr->mi_stickiness, MAX_STICK),
                        kfl_sub(total_aps, iptr->mi_aps[our_ndx]))) ;
#ifdef UNIT_TEST
      /*
       * Code to force migration every time a remote node requests a lock
       */
      if (unit_test_force_migrations && node != OurSiteID)
        {
          best_aps = calc_aps + 1 ;
          best_site = node ;
        }
#endif

      /* send resource on its way */
      if (best_site != OurSiteID
          && best_aps > calc_aps
          && Clmm_cb.mcb_state & MCB_SEND
          && rptr->refcount <= MAX_LOCKS)
        {
          /* increment client migr counts to block purges until done */
          if (res_check_client_purge(rptr))
            {
              DBPRINTF((DBG_MIGR|DBG_PURGE,
                        "clmm_trigger: xfer of (%s) blocked by purge 0x%x\n",
                        printbin(rptr->name, rptr->namelen), rptr)) ;
            }
          else
            {
              DBPRINTF((DBG_MIGR,
                        "clmm_trigger: transferring 0x%x to node %d\n",
                        rptr, best_site));
              /*
               *  Resource Statistics:
               *         cs_migrations - resource migrated
               */
              rptr->stats.cs_migrations++;

              ret = clmm_resource(rptr, best_site) ;
              if (ret)
                res_allow_client_purge(rptr) ;
#ifdef UNIT_TEST
              unit_test_send_saved_purge() ;
#endif
            }
        }
    }

  DBPRINTF((DBG_MIGR,
            "clmm_trigger: %d events [%d threshold]\n",
            iptr->mi_event_ttl, Clmm_cb.mcb_event_run));
  return 0 ;
}

/*
 | Name
 |	clmm_init
 | Description
 |	This function initializes the migration module.
 | Arguments
 |	None.
 | Returns
 |	CPERR_OK if no errors
 |      CCCP error code otherwise
 | Global Data Affected
 |      Clmm_cb
 */
int
clmm_init(unsigned int recalc_time,
#ifdef CPROTO
          /*
           * Although uint64_t happily exists in linux/types.h,
           * cproto barfs on it!  Since I want to use cproto to
           * generate prototypes, it skips this function, if I
           * don't use this construction.  Stupid tool.
           */
          unsigned long long int decay_rate
#else
          uint64_t decay_rate
#endif /* CPROTO */
)
{
    (void) memset( & Clmm_cb, 0, sizeof( Clmm_cb ) );
    Clmm_cb.mcb_state = MCB_MASK ;
    Clmm_cb.mcb_event_run = recalc_time;
    Clmm_cb.mcb_aps_decay = decay_rate;

    DBPRINTF((DBG_MIGR, "clmm_init: Clmm_cb s 0x%x er %d aps %s\n",
	      Clmm_cb.mcb_state, Clmm_cb.mcb_event_run,
	      kfl_ftos(Clmm_cb.mcb_aps_decay)));

    return( cccp_listen( PORT_MIGR, clmm_incoming ) );
}

/*
 | Name
 |      clmm_params
 | Description
 |	This function is included for backwards compatibility to the spec.
 |      From a design standpoint, it is better to have a single ioctl()
 |      style function to set/get any values.
 | Arguments
 |      rptr - pointer to resource structure being accessed.
 |	data - pointer to the data structure to be used.
 |	op - type of operation to be performed (GET/SET)
 | Returns
 |	Value returned from clmm_ctrl()
 |      or -1 in case of an unsupported operation.
 | Global Data Affected
 |      None
 */
int
clmm_params(struct resource *rptr, void  *data, clmm_ops_t op)
{
    switch(op) {
    case GETS:
	return(clmm_ctrl(data, CLMM_GETSTK));
    case SETS:
	return(clmm_ctrl(data, CLMM_SETSTK));
    default:
	return(-1);
    }
}

/*
 | Name
 |      clmm_cache
 | Description
 |	This function receives a pointer to a resource from the master
 |      routine.  This pointer is cached until after the master routine
 |      has completed in order to avoid having the resource yanked out
 |      from under it.
 * It's a HACK, Jim...
 | Arguments
 |	rp - pointer to resource structure being cached
 |	node - pointer to the node to which the structure is being moved.
 | Returns
 | 	0 if successful
 |	-1 if the cache is previously occupied.
 | Global Data Affected
 |      Clmm_cb
 */
int
clmm_cache(struct resource *rp, int node)
{
  /* don't migrate if resource created with LKM_LOCAL */
  if ((!rp) || (!rp->refcount) || (rp->rsrc_flags & RSRC_LOCAL))
	return -1;

  if (Clmm_cb.mcb_rcache != NULL || Clmm_cb.mcb_ncache != 0) {
      DBPRINTF((DBG_MIGR,
		"clmm_cache: occupied cache [0x%x, %d]\n",
		Clmm_cb.mcb_rcache,
		Clmm_cb.mcb_ncache));
    return(-1);
  } else if (rp == NULL || node == 0) {
	bsdlog( LOG_ERR, "clmm_cache: rp = 0x%08lx node = %d invalid", 
	        (unsigned long) rp, node );
	return -1;
  } else {
    Clmm_cb.mcb_rcache = rp;
    Clmm_cb.mcb_ncache = node;
    return(0);
  }
}

/*
 | Name
 |      clmm_ctrl
 | Description
 |	This function provides the only method whereby external
 |	modules of the cluster lock manager may access the control
 |	block. 
 | Arguments
 |	addr - a pointer to an operation dependent structure.
 |	op - the type of clmm_ops_t operation to be performed.
 | Returns
 | 	0 if successful
 |	-1 in case of error.
 | Global Data Affected
 |      Clmm_cb
 */
int
clmm_ctrl(void *addr, clmm_ops_t op)
{
    clmm_cb_t *cb = (clmm_cb_t *) addr;

    if (addr == NULL) {
	DBPRINTF((DBG_MIGR, "clmm_ctrl: NULL addr\n"));
	return(-1);
    }

    switch(op) {

    case CLMM_SETGCB:
	if (cb->mcb_state & ~MCB_MASK ||
	    cb->mcb_event_run == 0 ||
	    cb->mcb_aps_decay < KFL_DECAY_MIN ||
	    cb->mcb_aps_decay > KFL_DECAY_MAX) {
	    DBPRINTF((DBG_MIGR,
		      "clmm_ctrl: bad addr [0x%x] %d %s\n",
		      cb->mcb_state, cb->mcb_event_run,
		      kfl_ftos(cb->mcb_aps_decay))) ;
	    return(-1);
	}
	(void) memcpy( &Clmm_cb, cb, sizeof(clmm_cb_t) );
	break;

    case CLMM_GETGCB:
	(void) memcpy( cb, &Clmm_cb, sizeof(clmm_cb_t) );
	break;

    case CLMM_UPDGCB:
	if (cb->mcb_event_run == 0 ||
	    cb->mcb_aps_decay < KFL_DECAY_MIN ||
	    cb->mcb_aps_decay > KFL_DECAY_MAX) {
	    DBPRINTF((DBG_MIGR,
		      "clmm_ctrl: bad addr [0x%x] %d %s\n",
		      cb->mcb_state, cb->mcb_event_run,
		      kfl_ftos(cb->mcb_aps_decay))) ;
	    return(-1);
	}
	
	Clmm_cb.mcb_event_run = cb->mcb_event_run;
	Clmm_cb.mcb_aps_decay = cb->mcb_aps_decay;
	break;

    default:
	DBPRINTF((DBG_MIGR,
		  "clmm_ctrl: bad op [0x%x]\n",
		  op));
	return -1 ;
    }

    return 0 ;
}

/*
 | ~~~~~~~~~~~~~~~~~~~~~~
 | RMM Internal Functions
 | ~~~~~~~~~~~~~~~~~~~~~~
 */

/*
 | Name
 |      clmm_resource
 | Description
 |	This function is accessed by the only upper level entry point
 |	for resource migration.  This routine is multi-threaded.
 | 	In order to provide for multithreading the callback table for the
 | 	migrations in progress is used.
 | Arguments
 |	rptr - a pointer to the resource being migrated.
 |	node - nodeid of the target node.
 | Returns
 | 	0 if successful
 |	-1 in case of error.
 | Global Data Affected
 |	None
 |
 */
int
clmm_resource(struct resource *rptr, int node)
{
    clm_msg_t *msg_xmit = NULL;			/* outgoing msg		*/
    int	ret;

    DBPRINTF((DBG_MIGR,
	      "clmm_resource(rptr 0x%x, node %d)\n",
	      rptr, node));


    check_handles(rptr, 0) ;
    CLM_ASSERT((rptr->master == 0), "Migrating non-master");
    /* freeze resource & change master now */
    rptr->rsrc_flags |= RSRC_FRZEN;
    rptr->master = node;

    /* bundle the resource and callback data */
    msg_xmit = clmm_bundle( rptr, node );
    if ( msg_xmit == NULL ) {
	DBPRINTF((DBG_MIGR,
		  "clmm_resource: clmm_bundle(%s)\n",
		  printbin(rptr->name, rptr->namelen)));
	goto cleanup;
    }

#ifdef VERBOSE_MIGR
    clm_dumpres(rptr);
#endif
    /* send it off */
    if ((ret = cccp_send(msg_xmit)) != CPERR_OK) {
	DBPRINTF((DBG_MIGR,
		  "clmm_resource: cccp_send (%s) returned %d\n",
		  printbin(rptr->name, rptr->namelen), ret));
	goto cleanup;
    }
    DBPRINTF((DBG_MIGR|DBG_MIGR2,
	      "SEND MASTER (%s:%x) to %d complete\n",
	      printbin(rptr->name, rptr->namelen), (rptr->rh.rh&0xff), node));
    return 0 ;

cleanup:
    /* maybe we should free the message pointer, hmmm? */
    if (msg_xmit)
	cccp_msg_destroy( msg_xmit );

    /* if aborting, we attempt to restore the unfrozen state */
    CLM_ASSERT((rptr->master), "Invalid reset");
    rptr->rsrc_flags &= ~RSRC_FRZEN ;
    rptr->master = 0 ;
    return -1 ;
}

/*
 * clmm_send_fake_sync
 *
 * TRAPDOOR
 * this code coerces a pre 4.2 node to send an inband response to
 * an inband message by requesting a migration of a nonexistent
 * resource
 */

int
clmm_send_fake_sync(
    int node,		/* node causing reconfiguration */
    int sync_point_val)	/* stage code to send */
{
    struct clm_msg *msg ;
    clmm_data_t *mig_data;		/* migration data struct	*/
    struct reclock *fake_lock ;
    int ret ;

    DBPRINTF((DBG_SYNC, "%s: send to %d\n", __FUNCTION__, node));

    /* allocate and clear the memory for fake msg */

    if ( ! (mig_data = (struct clmm_data *)
			CLM_MALLOC(sizeof *mig_data + sizeof *fake_lock)))
    {
	DBPRINTF((DBG_MIGR, "clmm_send_fake_sync: malloc) failed\n"));
	return 1;
    }

    (void) memset( mig_data, 0, sizeof *mig_data + sizeof *fake_lock );

    /*
     * Variable buffer starts at data pointer in migration data structure.
     */
    fake_lock = (struct reclock *) &mig_data->cmig_data ;
    fake_lock->lck.type = DLM_RES_NTYPES+5 ; /* invalid type is better */
    fake_lock->lck.remlockid = -1 ;	/* UNSPEAKABLE lockid */

    /* set up header */
    mig_data->cmig_version = MPROTO_VERSION;
    mig_data->cmig_csize = 1 ;
    mig_data->cmig_rptr = (struct resource *) -4 ;	/* special FLAG VALUE */

    /* allocate message header */

    msg = cccp_msg_create( PORT_MIGR, RESRPRI, node, MIGR,
			   sizeof( *mig_data ) + sizeof( *fake_lock ),
			   NULL, mig_data, NULL );
    if ( msg == NULL )
    {
	DBPRINTF((DBG_MIGR,
		  "%s: malloc(%d) failed\n", __FUNCTION__, sizeof(clm_msg_t)));
	CLM_FREE( mig_data );

        return 1 ;
    }

    /* send it off */
    if ((ret = cccp_send(msg)) != CPERR_OK)
    {
	DBPRINTF((DBG_MIGR, "%s: cccp_send returned %d\n", __FUNCTION__, ret));
	CLM_FREE(msg->cm_data) ;
	cccp_msg_destroy( msg );
	return 1 ;
    }
    return 0 ;
}

/*
 | Name
 |      clmm_ack_confirm
 | Description
 |	This function is used to ensure that a resource doesn't migrate
 |	*again* before all ACKs are received.
 | Arguments
 |	reply - a pointer to a clm_msg_t.
 |	status - the status of the message sent.
 | Returns
 | 	None
 | Global Data Affected
 |	None
 |
 */
void
clmm_ack_confirm(clm_msg_t *reply, int status)
{
    struct resource *rptr;
    clmm_ack_t *ack;

    DBPRINTF((DBG_MIGR, "clmm_ack_confirm(reply 0x%x, status %d)\n",
	      reply, status));

    if (status == 0) {

	/* get ack pointer */
	ack = (clmm_ack_t *)reply->cm_data;
	if (ack->cma_status != MIG_ACK) {
		/* here if resource didn't migrate */
		goto done;
	}

        /* recover resource */
        rptr = (struct resource *) reply->cm_private;

	CLM_ASSERT((rptr), "No resource to unstick");
	check_handles(rptr, 0) ;

	/* rebuild might have blown resource migrating state */
	if (~rptr->rsrc_flags & RSRC_ACTIVE) {
		goto done;
	}

	CLM_ASSERT((rptr->master == 0 || ~rptr->rsrc_flags & RSRC_STUCK),
			"Unstick non-master copy");

	/* un-stick the resource */
	rptr->rsrc_flags &= ~RSRC_STUCK;

	/*
	 * resource may have been purged during migration attempt but left
	 * around in order for migration to clean up.
	 */
	if (rptr->refcount == 0) {
		deallocate_resource(rptr);
	}
    }

done:
    /* free the message and data pointers here 'cause the comms layer won't */
    if (reply->cm_length && reply->cm_data)
        CLM_FREE(reply->cm_data);

    cccp_msg_destroy( reply );
}


/*
 | Name
 |      clmm_callback
 | Description
 |	This function is used to ensure handling consistency in the
 |	face of a failure to deliver a migration request.
 | Arguments
 |	reply - a pointer to a clm_msg_t.
 |	status - the status of the message sent.
 | Returns
 | 	None
 | Global Data Affected
 |	None
 |
 */
void
clmm_callback(clm_msg_t *reply, int status)
{
    struct resource *rptr;

    DBPRINTF((DBG_MIGR, "clmm_callback(reply 0x%x, status %d)\n",
	      reply, status));
    if (status == 0) {
	goto done;
    }

    /* verify message type */
    if (reply->cm_type != MIGR) {
	DBPRINTF((DBG_MIGR, "clmm_callback: msg type 0x%x\n", reply->cm_type));
	goto done;
    }

    /* recover resource */
    rptr = (struct resource *) reply->cm_private;

    CLM_ASSERT((rptr && rptr->master && rptr->rsrc_flags & RSRC_ACTIVE),
		"bad resource pointer in clmm_callback");

    check_handles(rptr, 0) ;

    /* thaw out resource */
    res_allow_client_purge(rptr);
    rptr->rsrc_flags &= ~RSRC_FRZEN;
    rptr->master = 0;

done:
    /* free the message and data pointers here 'cause to comms layer won't */
    if (reply->cm_length && reply->cm_data)
        CLM_FREE(reply->cm_data);

    cccp_msg_destroy( reply );
}

/*
 | Name
 |      clmm_bundle
 | Description
 |	This function is used to package migration data and send
 |	it off to a target node.
 *	It assumes that the recipient is version compatible to this node.
 *	If not, a higher level routine is responsible for converting the msg.
 | Arguments
 |	resptr - a pointer to the resource structure being migrated.
 |	node - the target node.
 |	msg_out - a pointer to the message template to be filled.
 | Returns
 | 	The number of bytes sent
 |	-1 in case of error.
 | Global Data Affected
 |	None
 |
 */
clm_msg_t *
clmm_bundle(struct resource *resptr, int node )
{
    clm_msg_t * msg_out;
    clmm_data_t *mig_data;		/* migration data struct	*/
    struct reclock *bptr, *top ;	/* infamous buffer ptrs	*/
    struct reclock *hptr;		/* walk the reclock chains	*/
    int datasz,				/* data size to be sent		*/
	nlocks,				/* refcount */
	oops = TRUE,			/* new master has dropped all locks */
	i ;				/* universal counter - mostly RLQ */

    DBPRINTF((DBG_MIGR,
	      "clmm_bundle(rptr 0x%x, node %d)\n",
	      resptr, node));

    check_handles(resptr, 0) ;


    /* allocate and clear the memory for it all */
    datasz = sizeof(clmm_data_t) + resptr->refcount * sizeof(struct reclock) ;

    if ( ! (mig_data = (struct clmm_data *)CLM_MALLOC(datasz)))
    {
	DBPRINTF((DBG_MIGR, "clmm_bundle: malloc(%d) failed\n", datasz));
	return( NULL );
    }

    (void) memset( mig_data, 0, datasz );
    
    /* set up header */
    mig_data->cmig_version = MPROTO_VERSION;
    mig_data->cmig_csize = clm_node_count()+1;
    mig_data->cmig_rptr = resptr;
    mig_data->cmig_flags = resptr->rsrc_flags;
    mig_data->cmig_flags &= ~RSRC_FRZEN;	/* new copy is thawed	*/
    (void) memcpy( &mig_data->cmig_value, &resptr->value, 
		   sizeof mig_data->cmig_value );
    (void) memcpy( &mig_data->cmig_info, &resptr->clminfo, 
		   sizeof mig_data->cmig_info );

    /*
     * fill in resource stats fields
     * back conversion will delete if necessary
     */
    mig_data->cmig_lastnode = resptr->lastnode ;
    (void) memcpy( &mig_data->cmig_stats, &resptr->stats, 
		   sizeof(mig_data->cmig_stats) );

    /*
     * Variable buffer starts at data pointer in migration data structure.
     */
    bptr = (struct reclock *) &mig_data->cmig_data ;

    /*
     * save ptr to top of data section for absurd method of passing
     * resource id via lock offset
     */
    top = bptr;

    /* fill in the data (grant, convert, wait) */
    for (i = 0, nlocks = 0 ; i < RLQ_MAX ; i++)
    {
	for ( EACH_DLM_LISTNODE( & resptr->lockqs[i], hptr ) )
	{
	    CLM_ASSERT((~hptr->lck.state & LOCK_STALE),
			    "Stale lock in clmm_bundle");
	    /* don't include LOCK_LOCAL_ONLY lock in migration packet */
	    if ( (i != RLQ_GRANT) || !(hptr->lck.state & LOCK_LOCAL_ONLY) )
	    {
		if (mig_data->cmig_cache == 0 && hptr->lck.site == node)
		{
		    mig_data->cmig_cache = ((uint)bptr - (uint)top) ;
		    oops = FALSE ;
		}

		(void) memcpy( bptr, hptr, sizeof( *bptr ) );
		bptr++;

		mig_data->cmig_q_num[i]++ ;
	    }
	}
	nlocks += mig_data->cmig_q_num[i] ;
    }
    
    /*
     * maybe we shouldn't migrate this if the destination node no longer has
     * a copy of its resource
     */
    if (oops) {
	DBPRINTF((DBG_MIGR,
		"clmm_bundle: Abort migration of (%s) to %d",
		printbin(resptr->name, resptr->namelen), node));
	CLM_FREE((char *) mig_data) ;
	return( NULL );
    }
    DBPRINTF((DBG_MIGR,
	      "clmm_bundle: backpatching header ref %d l %d c %d w %d\n",
	      resptr->refcount,
	      mig_data->cmig_q_num[RLQ_GRANT], mig_data->cmig_q_num[RLQ_CONV],
	      mig_data->cmig_q_num[RLQ_CONV]));

    /* make sure counts don't exceed refcount */
    CLM_ASSERT((nlocks <= resptr->refcount), "Bad refcount");

    /* set up the actual message */
    msg_out = cccp_msg_create( PORT_MIGR, RESRPRI, node, MIGR, datasz,
			       clmm_callback, mig_data, resptr );

    if ( msg_out == NULL )
    {
	CLM_FREE( mig_data );
	DBPRINTF((DBG_MIGR, "[%s:%s] malloc(%d) failed\n",
		  current->comm, __FUNCTION__,
		  sizeof(clm_msg_t)));
    }

    DBPRINTF((DBG_MIGR,
	      "clmm_bundle: complete [%d]\n",
	      datasz));
    return( msg_out );
}

/*
 | Name
 |      clmm_lookup
 | Description
 |	This function maps resource information embedded in a message,
 |	to a pointer to that resource.
 | Arguments
 |	resptr - a pointer to the resource structure being migrated.
 |	node - the target node.
 |	msg_out - a pointer to the message template to be filled.
 | Returns
 | 	A pointer to the resource.
 |	NULL in case of error.
 | Global Data Affected
 |	None
 |
 */
static struct resource *
clmm_lookup(clm_msg_t *msg)
{
    struct resource *tptr = NULL;
#if defined(DEBUG)
    static int	badid = 0;
#endif

    DBPRINTF((DBG_MIGR,
	      "clmm_lookup(msg 0x%x)\n",
	      msg));

    switch(msg->cm_type)
    {

    case MIGA:			/* retrieve from cache		*/
	{
	    clmm_ack_t *aptr = (clmm_ack_t *)(msg->cm_data);

	    tptr = aptr->cma_resptr;
	}
	break;

    case MIGR:
	{		/* track it down via offset	*/
	    struct reclock *rptr;
	    struct reclock *a;
	    clmm_data_t *cmdp;

	    cmdp = (clmm_data_t *) msg->cm_data;
	    rptr = (struct reclock *)((uint)&cmdp->cmig_data +
				    (uint)cmdp->cmig_cache);

	    DBPRINTF((DBG_MIGR,
		    "clmm_lookup: MIGR mig_data=0x%x bufptr=0x%x reclock=0x%x\n",
		    cmdp, &cmdp->cmig_data, rptr));
#ifdef DBPRINT
	    if (rptr)
		clm_printlock(rptr);
#endif

	    a = find_reclock(rptr->lck.type, rptr->lck.remlockid);
	/*
	 * We need to examine the lock information passed in to ensure
	 * we pick up the local copy of the correct resource.
	 */
	    if (!a || !a->lck.rh.rh || a->lck.state & LOCK_GONE)
#ifdef NOTDEF
		|| clmm_purge_chk(rptr->lck.site, rptr->lck.pid)
#endif
	    {
		DBPRINTF((DBG_UNIXLOCK, "clmm_lookup: 0x%x bad lock from %d",
			rptr->lck.remlockid, msg->cm_node));
#if defined(DEBUG)
		if (!a && rptr->lck.remlockid == badid)
		{
		    bsdlog(LOG_ERR, "Repeat bad lockid");
		}
		else
		{
		    badid = rptr->lck.remlockid;
		}
#endif
	    }
	    else if (a->lck.pid != rptr->lck.pid)
	    {
#if defined(DEBUG)
		bsdlog(LOG_ERR, "clmm_lookup: 0x%x pid mismatch from %d",
			rptr->lck.remlockid, msg->cm_node);
		if (rptr->lck.remlockid == badid)
		{
		    bsdlog(LOG_ERR, "Repeat bad lockid");
		}
		else
		{
		    badid = rptr->lck.remlockid;
		}
#endif
	    }
	    else if (a->lck.site != rptr->lck.site)
	    {
#if defined(DEBUG)
		bsdlog(LOG_ERR, "clmm_lookup: 0x%x site mismatch from %d",
			rptr->lck.remlockid, msg->cm_node);
		if (rptr->lck.remlockid == badid)
		{
		    bsdlog(LOG_ERR, "Repeat bad lockid");
		}
		else
		{
		    badid = rptr->lck.remlockid;
		}
#endif
	    }
	    else
	    {
		tptr = resptr(a->lck.rh);
	    }
	}
	break ;

    default:
	tptr = NULL;
	break ;
    }

    if (tptr == NULL)
	DBPRINTF((DBG_MIGR,
		  "clmm_lookup:  failure for type 0x%x\n",
		  msg->cm_type));
    return(tptr);
}

/*
 | Name
 |      clmm_strip
 | Description
 |	This function unlinks and frees all non-local reclocks from a
 |	reclock chain while updating the reference count and list of the
 |	resource.
 | Arguments
 |	rp - pointer to the resource.
 |	qp - pointer to the queue structure
 | 	mflag - TRUE if this is a migration operation
 | Returns
 | 	The number of remaining reclocks.
 | Global Data Affected
 |	None
 |
 */
static int
clmm_strip(struct resource *rp, dlm_list_t * queue, int mflag)
{
    struct reclock * curr;
    struct reclock * next;
    int i = 0;

    /* clear out from the start of the chain */
    for ( curr = (struct reclock *) dlm_list_get_head( queue )
	  ; curr != NULL
	  ; curr = next )
    {
	next = (struct reclock *) dlm_list_succ( (dlm_listnode_t *) curr );

	if (mflag)	/* only adjust migrcount for migration operation */
	    lock_allow_client_purge( curr );	/* decr client migr count */
	if ( curr->lck.site != OurSiteID )
	{
	    remove_queue( queue, curr, RL_RQ_OFS) ;
	    rp->refcount-- ;	/* MUST NOT delete resource here */
	    free_le( curr->lck.type, curr );
	}
	else
	    i++;
    }

    return i ;
}


/*
 | Name
 |      clmm_copy
 | Description
 |	This function copies reclock data from an incoming reclock to
 |	another that is freshly allocated.
 | Arguments
 |	recp - a pointer to the incoming reclock.
 |	tmpp - a pointer to the newly allocated reclock.
 | Returns
 |	None
 | Global Data Affected
 |	None
 |
 */
void
clmm_copy( const struct reclock * recp, struct reclock * tmpp )
{
    (void) memcpy( &tmpp->route, &recp->route, sizeof( recp->route ) );

    tmpp->lck.mode = recp->lck.mode;
    tmpp->lck.reqmode = recp->lck.reqmode;
    tmpp->lck.site = recp->lck.site;
    tmpp->lck.lk_flags = recp->lck.lk_flags;
    tmpp->lck.pid = recp->lck.pid;
    tmpp->lck.state = recp->lck.state;
    tmpp->lck.lb = recp->lck.lb;
    tmpp->lck.ub = recp->lck.ub;
    tmpp->lck.seqnum = recp->lck.seqnum;
    tmpp->lck.xid = recp->lck.xid;
}

/*
 | Name
 |      clmm_chain
 | Description
 |	This function is used to take a block of reclock structures
 |	from a clmm_data_t and restore them into a new master.  Note
 |	that local reclocks override migrated reclocks.
 | Arguments
 |	rp - a pointer to the local resource structure
 |	msg - a pointer to the incoming clmm_data_t.
 | Returns
 |	0 on success
 |	-1 otherwise.
 | Global Data Affected
 |	None
 |
 */

/*
 * structure defining table of bits which must be set or clear
 * in a lock when it is on a particular queue of a resource
 */
struct lkq_bits
{
    int set ;		/* bits which must be set */
    int clear ;		/* bits which must be clear */
} ;

static const struct lkq_bits lkq_mask[RLQ_MAX] =
{
   { LOCK_GRANTED, LOCK_WAITING | LOCK_CONVERTING },
   { LOCK_CONVERTING, LOCK_GRANTED | LOCK_WAITING },
   { LOCK_WAITING, LOCK_GRANTED | LOCK_CONVERTING }
} ;

char *rlq_names[] = { "GRANT", "CONVERT", "WAIT" } ;

int
clmm_chain(struct resource *prp, clm_msg_t *msg)
{
    clmm_data_t *data = (clmm_data_t *)msg->cm_data;
    struct reclock *rp, *tmpp ;
    int i, j ;
    dlm_list_t   old_queue;
    dlm_list_t   queue;

    DBPRINTF((DBG_MIGR,
	      "clmm_chain(prp 0x%x, msg 0x%x  ref %d  g %d  c %d  w %d)\n",
	      prp, msg, 0,
	      data->cmig_q_num[RLQ_GRANT],
	      data->cmig_q_num[RLQ_CONV],
	      data->cmig_q_num[RLQ_WAIT]
	      )) ;

    check_handles(prp, CH_NOREF) ;	/* ref count is 0 now! */

    /*
     * take all locks on the queues for this resource and store them on a
     * temporary queue.  This is done to track any locks that we may have
     * locally that are not in the incoming migration packet.
     */
    
    dlm_list_init( & old_queue );
    for (i = 0 ; i < RLQ_MAX ; i++)
    {
	dlm_list_append( & old_queue, &prp->lockqs[i] );
    }

    rp = (struct reclock *)&data->cmig_data; /* start position */

    for (i = 0; i < RLQ_MAX ; i++)
    {
	dlm_list_init( & queue );
	tmpp = NULL;	/* clear tmp pointers	*/
	for (j = 0; j < data->cmig_q_num[i] ; j++, rp++) 
	{
	    if (rp->lck.site == OurSiteID) {	/* exists locally	*/
		tmpp = find_reclock(prp->type, rp->lck.remlockid);
		if (NULL == tmpp) {
		    DBPRINTF((DBG_MIGR2, "HEY!  Where's my reclock?!?!?"));
		    abort("clmm_chain: granted lock not found");
		    continue;
		}

		/*
		 * It's possible for a master copy of a resource to arrive here
		 * before the response that would put this lock on a queue.  If
		 * it's not queued, then we can't remove it.  We need to queue
		 * it on the local copy of the resource in case it tries to
		 * migrate away before the response does get here.
		 */
		/*
		 * Can't remove from old_queue unless it WAS queued before
		 * incoming migration.  Lock is always on a queue unless
		 * INFLIGHT_WAITING is set.
		 */
		if (~tmpp->lck.state & INFLIGHT_WAITING)
		{
			remove_queue( & old_queue, tmpp, RL_RQ_OFS) ;
#ifdef SPECIALCASE
		} else {
			tmpp->lck.rh.rh = rp->rh.rh;
#endif
		}
		/*
		 * local copy of the lock could've been on the convert queue,
		 * so clear the LOCK_* bits so a future response handler
		 * doesn't try to remove it from the wrong queue.
		 */
		tmpp->lck.state &= ~ lkq_mask[i].clear ;
		tmpp->lck.state |= lkq_mask[i].set ;
	    }
	    else
	    {				/* allocate and set	*/
		tmpp = get_le(prp->type);
		if (NULL == tmpp)
		    abort("can't get reclock in clmm_chain");
		if (rp->lck.site == msg->cm_node)
		    tmpp->lck.remlockid = rp->lck.lockid;
		else
		    tmpp->lck.remlockid = rp->lck.remlockid;
		(void) memcpy( &tmpp->lck.rh, &prp->rh, sizeof(tmpp->lck.rh) );
		clmm_copy(rp, tmpp);
		addmap(tmpp->lck.remlockid, tmpp->lck.lockid, tmpp->lck.site) ;
		tmpp->lck.state &= ~LOCK_GONE;
	    }
	    add_queue( & queue, tmpp, RL_RQ_OFS) ;
	    prp->refcount++ ;
	}
	
	dlm_list_move( & prp->lockqs[i], & queue );
	check_handles(prp, 0) ;
    }

    /* walk the queue of leftovers and requeue locks not in migr packet */
    clmm_replace_leftovers( & old_queue, prp) ;
    check_handles(prp, 0) ;
    return 0 ;
}

/*
 * clmm_replace_leftovers
 *
 * take a queue of mixed locks for one resource
 * replace them on the correct queues on that resource
 * destroys source queue
 */

static void
clmm_replace_leftovers(
		        /* Queue containing new locks.
			 */

		        dlm_list_t * queue,
		       
		        /* Resource on which to queue the locks.
			 */

		        struct resource *rp
		      )
{
    unsigned         i;
    struct reclock * lock;

    while ( ! dlm_list_is_empty( queue ) )
    {
	lock = (struct reclock *) dlm_list_rem_head( queue );
	
	for (i = 0 ; i < RLQ_MAX ; i++)
	{
	    if (lock->lck.state & lkq_mask[i].set)	/* which queue? */
	    {
		add_queue(&rp->lockqs[i], lock, RL_RQ_OFS) ;
		rp->refcount++ ;
		DBPRINTF((DBG_MIGR2,
				"Requeue 0x%x st=0x%x fl=0x%x on (%s) %s\n",
				lock->lck.lockid, lock->lck.state,
				lock->lck.lk_flags, printbin(rp->name,
				rp->namelen), rlq_names[i])) ;
		break ;
	    }
	}
#ifdef SPECIALCASE
/*
 * ??????? Why is this necessary?  Shouldn't LOCK_LOCAL_ONLY already be set?
 */
	if (i == RLQ_WAIT && lock->lck.state & LOCK_GONE) {
		/*
		 * Lock didn't come in with migration packet, so it
		 * is not on the master copy.  Set LOCK_LOCAL_ONLY in
		 * state so that if we try to migrate this resource,
		 * we know not to include this lock.
		 */
		oldh->lck.state |= LOCK_LOCAL_ONLY ;
	}
#endif /* SPECIALCASE */
	if (i == RLQ_MAX)
	    abort("lock w/no queue marker in clmm_chain") ;
    }
}


/*
 | Name
 |      clmm_master2slave
 | Description
 |	This function is to downgrade the local copy
 | 	of the resource into a slave resource.  First,
 | 	all nonlocal locks are removed, and the lock count is
 |	decremented accordingly.  Then the master field is updated.
 | Arguments
 |	rptr - a pointer to the local resource structure
 |	new_master - the nodeid of the new master
 |	mflag - TRUE if this is a migration operation.
 | Returns
 |	None
 | Global Data Affected
 |	None
 |
 */
void
clmm_master2slave(struct resource *rptr, ushort new_master, int mflag)
{
    rl_info_t	rl_entry;
    int i ;

    CLM_ASSERT((new_master), "new master not 0");

    check_handles(rptr, 0) ;

    DBPRINTF((DBG_MIGR,
	      "clmm_master2slave(rptr 0x%x, newnode %d)\n",
	      rptr, new_master));

    for (i = 0 ; i < RLQ_MAX ; i++)
	clmm_strip(rptr, &rptr->lockqs[i], mflag) ;

    /* update the master field - if there still *is* a resource */
    if (rptr->refcount > 0) {
	rptr->master = new_master;
    }
    /* update local RLDB cache */
    rl_entry.rl_name = rptr->name;
    rl_entry.rl_namelen = rptr->namelen;
    rl_entry.rl_node = rptr->master;
    rl_entry.rl_type = rptr->type;
    rl_update(&rl_entry);

#ifdef VERBOSE_MIGR
    clm_dumpres(rptr);
#endif
    check_handles(rptr, 0) ;
}

/*
 | Name
 |      clmm_slave2master
 | Description
 |	This function is to upgrade the local copy
 | 	of the resource into a master resource.  First, the print
 | 	entry is updated, then all nonlocal locks are removed,
 | 	and the lock count is decremented accordingly.
 | Arguments
 |	rptr - a pointer to the local resource structure
 |	msg - a pointer to the incoming clmm_data structure
 | Returns
 |	0 on success
 |	-1 on failure
 | Global Data Affected
 |	None
 |
 */
static int
clmm_slave2master(struct resource *rptr, clm_msg_t *msg)
{
    clmm_data_t *data = msg->cm_data;	/* types punned here for version */
    int i;
    rl_info_t	rl_entry;

    DBPRINTF((DBG_MIGR,
	      "clmm_slave2master(rptr 0x%x, msg 0x%x)\n",
	      rptr, msg));

    /* integrity checks */
    if (data->cmig_version > MPROTO_VERSION) {
	DBPRINTF((DBG_MIGR,
		  "clmm_slave2master: version skew %d [local %d]\n",
		  data->cmig_version, MPROTO_VERSION));
    }

    if (data->cmig_csize > clm_node_count()+1) {
	DBPRINTF((DBG_MIGR,
		  "clmm_slave2master: size skew %d [local %d]\n",
		  data->cmig_csize, clm_node_count()+1));
	goto cleanup;
    }

    check_handles(rptr, 0) ;

#ifdef VERBOSE_MIGR
    clm_dumpres(rptr);
#endif
    rptr->refcount = 0;

    /* copy in new resource struct data */
    rptr->rsrc_flags = data->cmig_flags | RSRC_ACTIVE;

    (void) memcpy( &rptr->value, &data->cmig_value, sizeof(rptr->value) );
    (void) memcpy( &rptr->clminfo, &data->cmig_info, sizeof(clmm_info_t) );

				/* set up lock chains */
				/* transfer locks from msg to res */
    i = clmm_chain(rptr, msg) ;
    if (i < 0) {
	DBPRINTF((DBG_MIGR, "clmm_slave2master: clmm_chain failed\n"));
	goto cleanup;
    }

    /* Passing Resource Statistics */
    (void) memcpy( &rptr->stats, &data->cmig_stats, sizeof(dlm_statistics_t) );
    rptr->lastnode = data->cmig_lastnode;

    gettimeofday(&rptr->clminfo.mi_timestamp, NULL);

#ifdef VERBOSE_MIGR
    clm_dumpres(rptr);
#endif

    /* this may be necessary if the migration cleaned up stale locks */
    convert_locks(rptr, NULL);

    /* stick this resource until the ACK is confirmed */
    rptr->rsrc_flags |= (RSRC_STUCK | RSRC_UPDATING);

    /* update local RLDB cache */
    rl_entry.rl_name = rptr->name;
    rl_entry.rl_namelen = rptr->namelen;
    rl_entry.rl_node = OurSiteID;
    rl_entry.rl_type = rptr->type;
    rl_update(&rl_entry);

    check_handles(rptr, 0) ;

    return 0 ;

cleanup:
    return -1 ;
}

/*
 | Name
 |      clmm_incoming
 | Description
 |	This function handles all incoming messages for the migration
 |	module.  Currently there may be two types of incoming messages:
 | 	1. Messages ack/nack-ing a previous request.
 | 	2. Messages requesting node to accept migration.
 *	3. TRAPDOOR messages from pre 4.2 nodes containing inband sync msgs
 *		for the reconfig functionality
 | Arguments
 |	msg_in - a pointer to the incoming message structure
 | Returns
 |	0 on success
 |	-1 on failure
 | Global Data Affected
 |	None
 |
 */
int
clmm_incoming(clm_msg_t *msg_in)
{
    struct resource *resptr;
    clmm_ack_t *ack;
    clm_msg_t *reply;
    int ret;
    int maxcells = clm_node_count()+1;	/* number of nodes in cluster */

    DBPRINTF((DBG_MIGR, "clmm_incoming(msg 0x%x) [type 0x%x]\n",
	      msg_in, msg_in->cm_type));

    /* lookup resource */
    resptr = clmm_lookup(msg_in);

    /* verify type and initialize */
    switch (msg_in->cm_type) {
	
    case MIGA: {
	ack = (clmm_ack_t *) msg_in->cm_data;

/******************
 *
 * TRAPDOOR to handle incoming fake sync replies from old nodes!
 *
 ******************/
	if (resptr == (struct resource *) -4)	/* FLAG VALUE */
	{
	    clm_fakesync_recv(msg_in->cm_node) ;
	    return 0 ;
	}

/* *** END TRAPDOOR *** */

	CLM_ASSERT((resptr), "No resource pointer for migration ACK");

	/* Do not accept incoming migration messages if in rebuild */
	if (ack->cma_status & MIG_ACK && clm_in_reconfig())
	{
		bsdlog(LOG_ERR, "Drop migr packet - in rebuild phase\n");
		return 0;
	}
    
	/* verify that we expect this ACK and drop it if we don't */
	if ((~resptr->rsrc_flags & RSRC_FRZEN) ||
			(~resptr->rsrc_flags & RSRC_ACTIVE)) {
		DBPRINTF((DBG_MIGR2,
			"ACK recvd for bad resource (%s) f=%d m=%d frm=%d\n",
				printbin(resptr->name, resptr->namelen),
				resptr->rsrc_flags, resptr->master,
				msg_in->cm_node));
		break;
	}

	if (ack->cma_status & MIG_ACK) {
	    clmm_master2slave(resptr, msg_in->cm_node, TRUE);
	} else {
	    /* why did request fail? */
	    if (ack->cma_csize != maxcells)
		DBPRINTF((DBG_MIGR,
			  "clmm_incoming: NACK size %d [%d]\n",
			  ack->cma_csize,
			  maxcells));
	    /* reset master designation in resource structure */
	    CLM_ASSERT((resptr->master), "Invalid reset");
	    resptr->master = 0;

	    /* didn't migrate, so reset migration/purge counters */
	    res_allow_client_purge(resptr) ;
	}
	if (ack->cma_version != MPROTO_VERSION) {
	    DBPRINTF((DBG_MIGR,
		      "clmm_incoming: version skew %d [%d]\n",
		      ack->cma_version,
		      MPROTO_VERSION));
	}

	/* thaw out resource */
	DBPRINTF((DBG_MIGR,
		  "clmm_incoming: thawing resource (%s)\n",
		  printbin(resptr->name, resptr->namelen)));
	resptr->rsrc_flags &= ~RSRC_FRZEN;

	if (ack->cma_status & MIG_ACK) {
	    DBPRINTF((DBG_MIGR|DBG_MIGR2,
		  "clmm_incoming: COMPLETED MASTERY of (%s:%x) -> %d\n",
		  printbin(resptr->name, resptr->namelen), (resptr->rh.rh&0xff),
		  msg_in->cm_node));
	} else {
	    DBPRINTF((DBG_MIGR|DBG_MIGR2,
		  "clmm_incoming: MIGRATION REFUSED (%s:%x) <- %d\n",
		  printbin(resptr->name, resptr->namelen), (resptr->rh.rh&0xff),
		  msg_in->cm_node));
	}

	/*
	 * resource may have been purged during migration attempt but left
	 * around in order for migration to clean up.
	 */
	if (resptr->refcount == 0) {
		deallocate_resource(resptr);
	}

	break;
    }

    case MIGR: {
	clmm_data_t *data = (clmm_data_t *) msg_in->cm_data;
#if 0
	struct reclock	*a = (struct reclock *)&data->cmig_data;
#endif

	CLM_ASSERT((!resptr || resptr->master), "I already *am* the master");

	/* setup reply */
	if ((ack = (clmm_ack_t *)CLM_MALLOC(sizeof(*ack))) == NULL) {
	    DBPRINTF((DBG_MIGR,
		      "clmm_incoming: malloc(%d) failure\n",
		      sizeof(*ack)));
	    return 0 ;
	}
	ack->cma_version = MPROTO_VERSION;
	ack->cma_resptr = data->cmig_rptr;
	ack->cma_csize = maxcells;
	
	/* allocate message header */
	reply = cccp_msg_create( PORT_MIGR, RESRPRI, msg_in->cm_node, MIGA,
				 sizeof( clmm_ack_t ), clmm_ack_confirm, ack,
				 resptr );
	if ( reply == NULL ) {
	    DBPRINTF((DBG_MIGR,
		      "clmm_incoming: malloc(%d) failed\n",
		      sizeof(clm_msg_t)));
	    CLM_FREE( ack );
	    return 0 ;
	}

	/*
	 * We accept an incoming migration if the recv bit is set in the
	 * state, and we have a resource pointer and the master value of the
	 * resource pointer is the node that the migration is coming from.
	 * There's no way we could be receiving a resource from one node when
	 * we think it's mastered on a another.
	 *
	 * If we're not in normal run state, fail the request
	 */
	if (clm_low_version() >= DLM_VERSION
	    && clm_is_running()
	    && (Clmm_cb.mcb_state & MCB_RCVE)
	    && resptr
	    && resptr->master == msg_in->cm_node
	    && clmm_slave2master(resptr, msg_in) >= 0)
	{
	    ack->cma_status = MIG_ACK;
	    /* save ack ptr, send when dir update completes */
	    resptr->migr_ack = reply;

	    DBPRINTF((DBG_MIGR|DBG_MIGR2,
		      "clmm_incoming: BEGINNING MASTERY of (%s:%x) <- %d\n",
		      printbin(resptr->name, resptr->namelen),
		      (resptr->rh.rh&0xff),
		      msg_in->cm_node));

	    /* update directory */
	    update_directory(resptr, OurSiteID, MIGAPRI, NULL );
	}
	else
	{
	    ack->cma_status = MIG_NACK;
	    /* if NACK, send immediately */
	    if ((ret = cccp_send(reply)) != CPERR_OK)
	    {
	        DBPRINTF((DBG_MIGR,
		      "clmm_incoming: cccp_send (%s) returned %d\n",
		      printbin(resptr->name, resptr->namelen), ret));
	    }

	    DBPRINTF((DBG_MIGR|DBG_MIGR2,
		      "clmm_incoming: DECLINED MASTERY of (%s)\n",
		      printbin(resptr->name,
		      resptr->namelen)));
	}
    }
    break ;

    default:
	DBPRINTF((DBG_MIGR,
		  "clmm_incoming: unknown type 0x%x\n",
		  msg_in->cm_type));
    }

    return 0 ;
}
