/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_msg.c,v 1.21 2001/11/09 20:04:03 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS:
 *		clm_getmsg
 *		clm_sync
 *		msg_initialize
 *		msg_read_config
 *		perror
 *		reset_syncd
 *		syncd_up
 */
#ifndef lint
#endif

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/types.h>
#include <linux/signal.h>
#include <linux/msg.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "cltrc.h"
#include "clm_info.h"
#include "clm_data.h"
#include "dlm_version.h"
#include "dlm_kernel.h"
#include "dlm_devmsg.h"

#include <iterator.h>

#include "dlm_workqueue.h"
#include "dlm_recover.h"
#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */

/* how long we wait for the cluster manager to feed us config info */
#define	INIT_TIMEOUT_SEC	30

/* macros for setting and reading sync points */
#define SYNC_NODE(s)	((s) >> 16)
#define SYNC_ID(s)	((s) & 0xff)
#define SYNC_POINT(n,p)	(((p) & 0xff) | ((n) << 16))

static void clm_newsync_callback( clm_msg_t * msg, int stat );
static void handle_sync( int sync_id, int from_node, int cluster_event,
			 int type );

/*
 * arrays to store sync status and lookahead sync statuses
 */
#define SYNC_FROM_CLIENT 0
#define SYNC_FROM_SERVER 1
#define NSYNCTYPES 2

static int syncd[MAXNODES][NSYNCTYPES] ;
static int la_syncd[MAXNODES][NSYNCTYPES] ;
static int sync_point_val ;	/* current point being serviced */
static int sync_point_event ;	/* current point being serviced */

static int current_event;	/* cluster mgr event # being serviced */




/****************************************************************************/
/* dlm_process_cm_message - Process a message from the cluster manager
 * 
 */

int
dlm_process_cm_message(
		        /* Message from the cluster manager.
			 */

		        struct dlm_workunit * msg
		      )
{
    const char * type_str;
    int          ret_value;
    extern const char * dlm_msgtab[];
    extern const int    dlm_nmsgs;
    
    
    type_str = ( msg->type < dlm_nmsgs ) 
	? dlm_msgtab[ msg->type ]
	: "unknown message type";

    bsdlog( LOG_DEBUG, "Read %s(%d) from CM\n", type_str, msg->type );

    ret_value = 0;
    switch( msg->type )
    {
	case MQ_EVENT_INFO_MSG:
	{
	    current_event = *((int *) msg->data);
	    break ;
	}


	case MQ_DLM_TOP_INFO_MSG:
	{
	    handle_new_topology( (MQ_DLM_TOP_INFO_t *) msg->data );
	    break;
	}


	/* No processing is actually done here for the client-exit message.
	 * This is because this routine is only called from clm_master_loop.
	 * After returning from this call, at least one loop through the
	 * state machine processing loop will be executed.  If the state is
	 * CLM_ST_RUN, purge_list will be called by the state machine
	 * processing.  If the state is not CLM_ST_RUN, then we can't do the
	 * purge (yet) anyway.
	 */

	case MQ_CLIENT_IS_DEAD_MSG:

	default:
	{
	    break;
	}
    }

    return ret_value;
}




/****************************************************************************/
/* dlm_sync_notify_recovery - Notify recovery that a barrier has been reached
 *
 * Sends notfication to the DLM recovery state-machine that a barrier (sync
 * point) has been reached.
 */

void
dlm_sync_notify_recovery( void )
{
    rc_msg_t * dat;


    /* set up for next sync & capture readahead 
     */

    reset_syncd(0);


    /* Send the got-barrier message down to the state-machine.
     */

    dat = rc_msg_alloc( RC_MSG_GOT_BARRIER, 0, NULL );
    hsm_process_event( dlm_recov_state_machine, (hsm_msg_t *) dat );
}




/*
 * Inter-lock-manager SYNC functionality
 *
 * Previous versions (< 4.2) used an out-of-band sync via messages
 * via the cluster manager.  This had loopholes allowing units to get
 * somewhat out of phase.
 *
 * Current version:
 *	send an inband sync message to each peer
 *	wait for -both-
 *	    a response to this message
 *		(showing that all messages sent have been received)
 *	and for
 *	    an inbound message from peer
 *		(showing that all peer's messages have been received)
 *
 *	For backwards compatibility, send out of band sync to old nodes
 *	after all expected -responses- to our inband sync messages.
 *	Ignore any out-of-band messages from new nodes.
 *
 *	allow for
 *	    reception of next stage inbound sync message
 *	    reception of stage 1 of new sync caused by additional node failure
 *
 *	A 'sync_id' is an int containing the tuple (failed_node, state_number).
 *	It doesn't fully disambiguate a node down-up-down.
 */

/*
 * Name
 *	print_sync
 * Description
 *	Prints out the sync array.
 * Arguments
 *		None
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */

unsigned
print_sync( char * buf, unsigned limit )
{
    int        ndx;
    int        maxnode = clm_node_count();
    unsigned   node_id;
    unsigned   bytes;


    bytes = 0;
    for ( ndx = 0 ; (ndx < maxnode) && (bytes < limit) ; ndx++ )
    {
	node_id = INDEX_TO_NODEID( ndx );
	if (clm_node_reachable( node_id ))
	{
	    bytes += snprintf( & buf[ bytes ], limit - bytes,
			       "n%d=[%d:%d %d:%d] ", 
			       node_id,
			       SYNC_NODE(syncd[ndx][0]),
			       SYNC_ID(syncd[ndx][0]),
			       SYNC_NODE(syncd[ndx][1]),
			       SYNC_ID(syncd[ndx][1]));
	}
	else 
	{
	    bytes += snprintf( & buf[ bytes ], limit - bytes, "n%d=[-] ",
			       node_id );
	}
    }

    return( bytes );
}

/*
 * Name
 *	syncd_up
 * Description
 *	Examine the syncd array to determine if all lock managers have
 *	reached the sync point.  This function assumes that the syncd
 *	array has been allocated.
 * Arguments
 *		None
 * Returns
 *	FALSE if any element corresponding to an active node in the syncd
 *	array is 0 indicating that the awaited sync message from that node
 *	has not been received.
 *	Otherwise, we are syncd with other nodes, so return TRUE.
 * Global Data Affected
 *	NONE
 */
int
syncd_up(void)
{
    int		ndx, maxnode = clm_node_count();
    int		ret, sync_val = 0;
    int i ;


    /* Walk through the syncd array.  We return TRUE if all array
     * entries for active nodes are equal.
     */

    ret = 0 ;
    for (ndx = 0; ndx < maxnode; ndx++) 
    {
	if ( clm_node_reachable( INDEX_TO_NODEID( ndx ) ) )
	{
	    for (i = 0 ; i < 2 ; i++)
	    {
		if (0 == syncd[ndx][i])
		{
		    return FALSE;
		}
		if (sync_val && syncd[ndx][i] != sync_val) {
		    bsdlog(LOG_CRIT, "%s: somehow got %d:%d expected %d:%d",
			   __FUNCTION__, SYNC_NODE(sync_val), SYNC_ID(sync_val),
			   SYNC_NODE(syncd[ndx][i]),
			   SYNC_ID(syncd[ndx][i]));

		    return FALSE;
		}
		sync_val = syncd[ndx][i] ;
	    }
	}
    }

    /* if we made it this far, then sync point has been reached */
    return TRUE ;
}


/*
 * Name
 *	update_syncd
 * Description
 *	Update syncd table because of topology changes
 * Arguments
 *	Pointer to map array
 * Returns 
 *	nothing
 * Global Data Affected
 *	syncd, la_syncd
 */
void
update_syncd(int	*idmap)
{
    int	i, j;
    int new_syncd[MAXNODES][NSYNCTYPES] ;
    int new_la_syncd[MAXNODES][NSYNCTYPES] ;

    /* init new table */
    (void) memset( new_syncd,    0, sizeof( new_syncd ) );
    (void) memset( new_la_syncd, 0, sizeof( new_la_syncd ) );

    /* copy old sync values into new table */
    for (i = 0; i < MAXNODES; i++) 
    {
	if (idmap[i] >= 0)
	{
	    for (j = 0; j < NSYNCTYPES; j++) 
	    {
		new_syncd[i][j] = syncd[idmap[i]][j];
		new_la_syncd[i][j] = la_syncd[idmap[i]][j];
	    }
	}
    }

    /* copy new tables over on top of old ones */
    (void) memcpy( syncd,    new_syncd,    sizeof( syncd ) );
    (void) memcpy( la_syncd, new_la_syncd, sizeof( la_syncd ) );

    return;
}


/*
 * Name
 *	reset_sync
 * Description
 *	Reset the syncd array.
 * Arguments
 *		None
 * Returns
 *	NOTHING
 * Global Data Affected
 *	syncd
 */
void
reset_syncd(int node)
{
	int		i, maxnode = clm_node_count();
	int j ;

	/* Do not clear syncs that have already arrived for this failure */
	for (i = 0; i < maxnode; i++)
	{
	    for (j = 0 ; j < NSYNCTYPES ; j++)
	    {
		if ( node != 0 )
		{
		    if (SYNC_NODE(syncd[i][j]) != node)
			    syncd[i][j] = 0;
		}
		else
		{
		    syncd[i][j] = la_syncd[i][j] ;
		}
	    }
	}

	(void) memset( la_syncd, 0, sizeof( la_syncd ) );

    if ( node == 0 )
    {
	/* don't reset these if resetting for a new failure */
	sync_point_val = 0 ;	/* no current point */
	sync_point_event = 0 ;	/* and no current event */
    }
}


/* clm_inband_sync - Handle a SYNC message
 *
 * Handles a SYNC message when received from CCCP.  The SYNC message is
 * processed and a response is sent back to the original sender's SYNCRET
 * port.
 *
 * RETURNS
 *
 * Zero is always returned to prevent CCCP from sending a redirect message.
 *
 * NOTES
 *
 * This function is registers as a callback for the PORT_SYNC message port
 * in CCCP.
 */

int
clm_inband_sync(
		 /* Message received from CCCP on the PORT_SYNC port.
		  */

		 clm_msg_t * msg
	       )
{
    struct clm_syncmsg * sdata;
    struct clm_msg     * nmsg;


    sdata = msg->cm_data ;
    if ( sdata == NULL )
    {
	bsdlog(LOG_ERR, "bad sync msg from node %d\n", msg->cm_node) ;
	return 0 ;
    }

    acquire_lock( LK_ENTRY, kproc_pid );
    handle_sync( sdata->sync_id, msg->cm_node, sdata->event, 
		 SYNC_FROM_CLIENT );
    release_lock( LK_ENTRY, kproc_pid );

    nmsg = cccp_msg_duplicate( msg );
    if ( nmsg != NULL )
    {
	sdata = (struct clm_syncmsg *) nmsg->cm_data ;
	sdata->sender = OurSiteID ;
	nmsg->cm_port = PORT_SYNCRET ;
	nmsg->cm_callback = clm_newsync_callback ;
	DBPRINTF((DBG_CONFIG, "returning server sync %x to %d\n",
		  sdata->sync_id, msg->cm_node)) ;
	cccp_send( nmsg );
    }

    return 0 ;	/* must return 0 to prevent comm redirect */
}




/* handle_sync - Handle a SYNC or SYNCRET message
 *
 * Handles a SYNC or SYNCRET message.
 * handle sync messages
 * update sync[] with this msg's info
 * handles 'next' sync point messages in la_sync
 *
 * WARNINGS
 *
 * This function assumes that the caller holds the LK_ENTRY mutex.
 */

static void
handle_sync(int sync_id, int from_nodeid, int cluster_event, int type)
{
    int fail_node = 0;
    int maxnode = clm_node_count();
    int ndx;
    int from_ndx ;
    int stored_sync_id ;
    int node ;

#ifdef DEBUG
    bsdlog(LOG_DEBUG, "Read %s sync %d:%d from %d",
	   type ? "server" : "client",
	   SYNC_NODE(sync_id), SYNC_ID(sync_id), from_nodeid);
#endif

    /* assert that sync message came from living lock manager */
    if (!clm_node_reachable(from_nodeid))
    {
	goto done;
    }

    /* validate nodeid */
    from_ndx = NODEID_TO_INDEX(from_nodeid) ;
    if (from_ndx < 0)
    {
	bsdlog(LOG_CRIT, "%s: nodeid %d out of range", __FUNCTION__, from_nodeid);
	goto done;
    }

    CLM_ASSERT((type >= 0 && type < 7), "sync type bad") ;
    fail_node = SYNC_NODE(sync_id);

    /* don't overwrite existing sync slot unless it's for a new failure */
    if (syncd[from_ndx][type] != 0)
    {
	stored_sync_id = SYNC_ID(syncd[from_ndx][type]) ;

	if (stored_sync_id != SYNC_ID(sync_id) && SYNC_ID(sync_id) == 1)
	{
	    bsdlog(LOG_CRIT, "%s: restart reconfig. Now rebuild %d",
		   __FUNCTION__, fail_node);
	    reset_syncd(fail_node);
	    syncd[from_ndx][type] = sync_id;
	    bsdlog(LOG_CRIT, "calling update_clm from handle_sync");
	    goto done;
	}
	else if (stored_sync_id == (SYNC_ID(sync_id) - 1))
	{
	    bsdlog(LOG_CRIT, "%s: Read ahead %s %d:%d from %d", __FUNCTION__,
		   type ? "client" : "server",
		   SYNC_NODE(sync_id), SYNC_ID(sync_id), from_nodeid);

	    /* update lookahead array and return */
	    la_syncd[from_ndx][type] = sync_id;
	    goto done;
	} else {
	    bsdlog(LOG_CRIT, "%s: got bad %d:%d from %d", __FUNCTION__, 
		   SYNC_NODE(sync_id), SYNC_ID(sync_id), from_nodeid) ;
	}
    }

    /* check sync array consistency */
    for (ndx = 0; ndx < maxnode; ndx++)
    {
	if (syncd[ndx][type] && (syncd[ndx][type] != sync_id))
	{
	    /* here if sync msg doesn't match an array entry */
	    if (SYNC_ID(sync_id) == 1)
	    {
		/* someone sent us a 1 so we restart */
		bsdlog(LOG_CRIT, "%s: restart rebuild", __FUNCTION__);
		node = SYNC_NODE(sync_id);
		reset_syncd(node);
	    } else {

		/* otherwise, what's going on? */
		bsdlog(LOG_CRIT, "%s: array huh? got %d:%d expected %d:%d",
		       __FUNCTION__, SYNC_NODE(sync_id), SYNC_ID(sync_id),
		       SYNC_NODE(syncd[ndx][type]),
		       SYNC_ID(syncd[ndx][type]));
		goto done;
	    }
	}
	else if (from_ndx == ndx && syncd[ndx][type])
	{
	    bsdlog(LOG_ERR, "%s: redundant sync message from %d",
		   __FUNCTION__, from_nodeid);
	    goto done;
	}
    }

    /* update array with received value and return */
    syncd[from_ndx][type] = sync_id;


    /* If all of the nodes have reached the barrier, send a message into the
     * recovery state-machine.
     */

    if ( syncd_up() )
    {
	dlm_sync_notify_recovery();
    }

done:
}

/* clm_sync - 
 *
 * Sends a (PORT_SYNC, MSYNC, RESRPRI) message with data (stage, cluster
 * event #) to all new nodes.  The receiver sends a (PORT_SYNCRET, MSYNC,
 * RESRPRI) in response.
 *
 * The calling routine must arrange to wait for all incoming syncs and decide
 * whether to change state.
 *
 * The sync array is 2D to hold info on -both- our server and client halves.
 * -Both- must be in sync before proceeding (major change).
 *
 * Arguments
 *	sync_id - sync number
 * Returns
 *	-1 for an error
 *	0 if the sync was started succcessfully
 *
 *  Global Data Affected
 *	NONE
 */
int
clm_sync(int node, int sync_id)
{
    clm_msg_t          * syncmsg;	/* inband version */
    int                  nodeid;
    iterator_t         * iter;
    int                  fail_event = current_event;
    struct clm_syncmsg * sdata ;
    unsigned   count;

#ifdef DEBUG
    DBPRINTF((DBG_SYNC, "%s:  node %d  id %d\n",
              __FUNCTION__, node, sync_id)) ;
#endif
	/* validate sync id */
	if (sync_id <= 0) {
		bsdlog(LOG_CRIT, "%s: called with sync_id = %d", __FUNCTION__, sync_id);
		return -1;
	}

	/*
	 * set module state so interrupt (message) routines know what
	 * point is currently being serviced
	 */
	if (SYNC_POINT(node, sync_id) == sync_point_val)
	    abort("duplicate sync attempted") ;
	sync_point_val = SYNC_POINT(node, sync_id) ;
	sync_point_event = fail_event ;


    /* Mark that we have the sync message and the sync-response message
     * from ourself.
     */

    handle_sync( sync_point_val, OurSiteID, fail_event, SYNC_FROM_CLIENT );
    handle_sync( sync_point_val, OurSiteID, fail_event, SYNC_FROM_SERVER );

    count = 0;
    iter = dlm_nodeid_iterator_create();
    while ( ! iterator_next( iter, & nodeid ) )
    {
	if ( clm_node_reachable( nodeid ) && (nodeid != OurSiteID) )
	{
	    DBPRINTF((DBG_CONFIG, "send inband client sync %x to %d\n",
		      sync_point_val, nodeid)) ;
	    sdata = CLM_MALLOC(sizeof (struct clm_syncmsg)) ;
	    if ( sdata == NULL )
	    {
		bsdlog(LOG_CRIT, "%s sync msg alloc 2 failed", __FUNCTION__) ;
		abort("sync msg alloc 2 failed\n") ;
	    }

	    syncmsg = cccp_msg_create( PORT_SYNC, RESRPRI, nodeid, MSYNC,
				      sizeof( *sdata ), 
				      clm_newsync_callback, sdata, 0 );
	    if ( syncmsg == NULL )
	    {
		CLM_FREE( sdata );
		bsdlog(LOG_CRIT, "%s sync msg alloc failed", __FUNCTION__);
		abort("sync msg alloc failed\n");
	    }

	    sdata->sender = OurSiteID ;
	    sdata->event = fail_event ;
	    sdata->sync_id = sync_point_val ;

	    cccp_send(syncmsg) ;
	    count++;
	}
    }

    iterator_destroy( iter );
    
    
    if ( count != 0 )
    {
	dlm_sync_notify_recovery();
    }

    bsdlog( LOG_INFO, "[%s:%s] exiting, count = %u\n", current->comm,
	    __FUNCTION__, count );

    return 0 ;
}

/*
 * clm_newsync_callback
 *
 * handle 'message sent' for inband sync messages
 * handle 'message could not be sent'
 */
static void
clm_newsync_callback(clm_msg_t *msg, int stat)
{
    DBPRINTF((DBG_CONFIG, "%s: inband sync message to node %d confirmed\n",
		  __FUNCTION__, msg->cm_node));

    /* check status */
    if (stat)
	bsdlog(LOG_ERR, "Bad msg status %d in %s", stat, __FUNCTION__);

    /* free memory */
    if (msg)
    {
	if ( msg->cm_data != NULL )
	{
	    CLM_FREE( msg->cm_data );
	}

	cccp_msg_destroy( msg );
    }
}

/*
 * clm_fakesync_recv
 *
 * called from trapdoor in migration module to handle expected
 * NACK of our MIGR of resource -1.
 * INBAND response to INBAND sync to old node.
 * call handle_sync()
 */
void
clm_fakesync_recv(int nodeid)
{
    handle_sync( sync_point_val, nodeid, 0, SYNC_FROM_SERVER );
}


/* clm_inband_sync_resp - Handle a SYNCRET message
 *
 * Handles a SYNCRET message when received from CCCP.  The SYNCRET message
 * is processed by handle_sync.
 *
 * RETURNS
 *
 * Zero is always returned to prevent CCCP from sending a redirect message.
 *
 * NOTES
 *
 * This function is registers as a callback for the PORT_SYNCRET message port
 * in CCCP.
 */

int
clm_inband_sync_resp(
		      /* Message received from CCCP on the PORT_SYNCRET port.
		       */

		      clm_msg_t * msg
		    )
{
    struct clm_syncmsg * sdata;


    sdata = msg->cm_data;
    if ( sdata != NULL )
    {
	acquire_lock( LK_ENTRY, kproc_pid );
	handle_sync( sdata->sync_id, msg->cm_node, 0, SYNC_FROM_SERVER );
	release_lock( LK_ENTRY, kproc_pid );
    }

    return 0 ;
}
