/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_pti.c,v 1.7 2001/11/28 00:32:57 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		pti_proc_request
 *		pti_proc_response
 *		pti_read_request
 *		pti_read_request_cr
 *		pti_read_response
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/errno.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "cltrc.h"
#include "clm_data.h"
#undef LOCK_FREE
#include "dlm_kernel.h"
#include "clm_locksw.h"
#include "clm_reconfig.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/****************************************************************************/
/* pti_proc_request - Handle messages on the master port
 *
 * This function is responsible for handling requests that come in from the
 * communications layer on the master port.  It is responsible for determining
 * the master function to handle the request and then handing the transaction
 * structure to the function.
 *
 * If the is message was received on the PORT_MSTR_CR port, then the
 * processing is done with the DLM_RES_CREATE flag set.  Messages to this port
 * are authorized to create a master copy of the resource on the local node.
 *
 * RETURNS
 *
 * Zero if the local node is the master.  Otherwise, the node index of the
 * master is returned on success, or less-than zero on failure.
 */

int 
pti_read_request( clm_msg_t * msg )
{
    clm_trans_t * req = msg->cm_data;
    dlm_ops_t     type = msg->cm_type;
    int	          ret;
    int           flags;
    
    
    flags = (msg->cm_port == PORT_MSTR_CR) ? DLM_RES_CREATE : 0;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_PTI_PROG,
                req->clm_pid, req->clm_sequence, msg, flags);

	DBPRINTF((DBG_BASIC, "[%s:%s] BEGIN REMOTE TRANSACTION %s\n",
		  current->comm, __FUNCTION__,
		  transaction_type_string( req->clm_type )));

	/* Massage incoming message into readable form */
	clm_vers_rcv(msg);

	acquire_lock( LK_ENTRY, kproc_pid );

	switch (type) {

	case DLM_REGISTER:
		ret = master_register(msg, flags);
		break;

	case DLM_LOCK:
		ret = (*lockswitch[req->clm_locktype].master_lock)(msg, flags);
		break;

	case DLM_UNLOCK:
		ret = (*lockswitch[req->clm_locktype].master_unlock)(msg);
		break;

	case DLM_SCN_OP:
		ret = master_scn_op(msg);
		break;

	case DLM_RMIGR_OP:
		ret = master_tune_op(msg);
		break;

	case DLM_GMIGR_OP:
		ret = remote_glob_tune_op(msg);
		break;
		
	case DLM_STATS:
		ret = master_stats_op(msg);
		break;

	case DLM_PURGELOCKS:
		ret = master_purge(msg);
		break;

	default:
		bsdlog(LOG_ERR, MSGSTR(ILLEGAL_REQ,
				"pti_prog_p: illegal request type %d seq 0x%x"),
				type, req->clm_sequence);
		abort("pti_prog_p");
	}

	/* Reset incoming message to original form */
	clm_vers_reset(msg);

	DBPRINTF((DBG_BASIC, "[%s:%s] END REMOTE TRANSACTION %s 0x%x "
		  "ret %d\n", current->comm, __FUNCTION__,
		  transaction_type_string( req->clm_type ),
		  req->clm_sequence, ret));

	/*
	 * A return of zero indicates that we were the master for this
	 * request and may have responses to send
	 */
	if (0 == ret) {
		pti_flush_responses();
	}

	release_lock( LK_ENTRY, kproc_pid );


    /* If the master handling function returned a positive integer, then the
     * request should be redirected to another site.  Change the destination
     * and port values in the message and return a non-zero value to inform
     * the comm layer that this message is supposed to go somewhere else.
     */

    if ( (0 < ret) && (msg->cm_port != PORT_BOUNCE) )
    {
	msg->cm_node = ret;
	msg->cm_port = PORT_DIRECT;
    }
    else if (0 > ret) 
    {
	bsdlog(LOG_CRIT, MSGSTR(PTI_REQ_ERR,
				"Error returned from pti_proc_request"));
	ret = 0;
    }


    TRCHKGT(CLLOCKD_HKID | T_EXIT_FUNC, T_PTI_PROG,
	    ret, req->clm_sequence, msg, 0x00);

    return ret;
}




/****************************************************************************/
/* pti_proc_response -
 *
 * This function is called by the pti_read_response function after it has
 * read a response from the communications layer.  It is reponsible for
 * processing the response transaction that has come from another node.
 */

static void
pti_proc_response( unsigned node, clm_trans_t * resp )
{
    int multiple_masters = 0;       /* used for local locks*/
    clm_trans_t *req = NULL;	/* the associated request */



    DBPRINTF((DBG_BASIC|DBG_PTI, "pti_proc_response from %d: %s\n",
	      node, transaction_type_string( resp->clm_type )));

    if ( (resp->clm_direction != PTI_RESPONSE)
	 && (resp->clm_direction != PTI_ASYNC) )
    {
	bsdlog( LOG_ERR,"bad direction (%d) in PTI response\n",
	        resp->clm_direction );
	resp->clm_status = DLM_BADARGS;
	return;
    }

    resp->clm_rt_master = node;


    /* If the incoming transaction is a response to a request of ours, then we
     * need to match them so the saved request can be retrieved.
     * match_request() also verifies the order of the resopnses.
     */

    if (resp->clm_direction == PTI_RESPONSE) 
    {
	if ( (resp->clm_type == DLM_LOCK)
	     && (resp->clm_lockinfo.lk_flags & LKM_FINDLOCAL) )
	{
	    /* Local Lock: Update LLB structure */
	    req = clmll_proc_response( node, resp, & multiple_masters );
	}
	else 
	{
	    req = match_request(resp);
	}
		
	if (req == NULL) 
	{
	    bsdlog(LOG_ERR, MSGSTR(STRAY_RESPONCE, "Got stray response!"));
	    DBPRINTF(( DBG_FORCE, "dir = %d\n", resp->clm_direction ));
	    return;
	}
	else if (req == (clm_trans_t *)(-1)) 
	{
	    bsdlog(LOG_ERR, MSGSTR(MATCH_REQ_INTERR,
				   "match_request interrupted"));
	    return;
	}
    }


    /* If we are in grace, then match_request requeued the current response
     * and we shouldn't try to process it any further
     */

    if (resp->clm_status != DLM_DENIED_GRACE_PERIOD)
    {
	/* There won't be a callback if this is a blocking AST
	 */

	if ( resp->clm_callback == NULL )
	{
	    clm_reply(req, resp);
	}
	else
	{
	    /* finish processing locally */
	    DBPRINTF((DBG_BASIC, "\nCONTINUE TRANSACTION %d\n", 
		      (req != NULL) ? req->clm_sequence : 0 ));
	    if ( multiple_masters == 0 )
	    {
		(*resp->clm_callback)(req,resp);
	    }
	}

	if ( (resp->clm_direction == PTI_RESPONSE)
	     && (resp->clm_type == DLM_LOCK)
	     && (resp->clm_lockinfo.lk_flags & LKM_FINDLOCAL) )
	{
	    /* response may be last for llb */
	    clmll_check_llb((locallock_broadcast_t *)resp->clm_sender);
	}
    }


    DBPRINTF((DBG_FUNC, "leave pti_proc_response\n"));
    return;
}




/****************************************************************************/
/* pti_read_response - Process response to a lock request
 *
 * This function is called by the communications layer when a response to a
 * request is sent from a master node.  It calls pti_proc_response
 * with the transaction structure pointer that is contained in the message.
 *
 * RETURNS
 * 
 * Zero is always returned to prevent message re-direct.
 */

int
pti_read_response(
		   /* Response message.
		    */

		   clm_msg_t * msg
		 )
{
    clm_trans_t	*resp;

    CLM_ASSERT( msg != NULL, "NULL message" );
    acquire_lock( LK_ENTRY, kproc_pid );

    /* Massage incoming message into readable form */
    clm_vers_rcv(msg);

    if ( msg->cm_data == NULL )
    {
	bsdlog( LOG_CRIT, "NULL data in message (0x%08lx/%lu)\n",
	        (unsigned long) msg, msg->cm_type );
	abort( "NULL data in message" );
    }
    else
    {
	resp = (clm_trans_t *)msg->cm_data;
	CLM_ASSERT( resp != NULL, "NULL response" );

	STAMP(resp);
	pti_proc_response( msg->cm_node, resp );

	/* Reset incoming message to original form */
	clm_vers_reset(msg);
    }

    release_lock( LK_ENTRY, kproc_pid );
    return 0;
}
