/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_purge.c,v 1.19 2001/10/29 21:49:46 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		free_dead
 *		get_dead
 *		proc_notify
 *		purge_list
 *		sig_handle
 */

/* ============================================================ */
/*
 * Linux note:  this file under AIX depends upon the prochadd()
 * kernel service, which allows us to register handlers to be
 * called when a process, in this case a lock client, dies.
 * Since Linux lacks this nice feature, we have to slightly
 * twist ourselves.
 *
 * Since our locking uses /dev entries, all clients that use
 * our API open a /dev (hidden in the library...) and if they
 * should exit, then we will get a 'close' on that device, and
 * this has basically the same effect as does prochadd(), but
 * we see it in dlm_kerndd.  So there, we will queue up the
 * calls, just as would prochadd(), and deal with them in about
 * the same fashion.
 */
/* ============================================================ */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/errno.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <asm/signal.h>

#include "dlm.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "cltrc.h"
#include "clm_data.h"
#undef LOCK_FREE
#include "dlm_kernel.h"
#include "dlm_devmsg.h"
#include "clm_locksw.h"
#include "clm_reconfig.h"
#include "dlm_version.h"
#include "dlm_workqueue.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) \
	catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), \
						CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */

#define DEADBUFSZ	512 /* number of zombie bufs for proc_notify */
#define DEAD_COUNT	2   /* main loop cycles per ddlk check */

static int purge_all(clm_client_t *cp,
                     unsigned int flags,
                     int notme);
static int account_for_reply(clm_client_t *cp,
                             int sender,		
                             unsigned int msgflags);


static int remove_client_inflight(clm_client_t *cp) ;

static struct dead_pid_desc {
    struct dead_pid_desc *dpd_next;
    int dpd_pid;
} *pid_list = NULL;
static struct dead_pid_desc *dead_list = NULL;
static struct dead_pid_desc *dead_block = NULL;

#ifndef __linux__
static struct proch proch;
#endif /* !__linux__ */

#ifdef UNIT_TEST
int unit_test_delay_purge = 0 ;		/* 1 to hold off purge until next migr */
struct transaction *unit_test_saved_purge = NULL ;	/* hold purge req here */
extern int unit_test_hold_migrations ;	/* 1 to hold off migr until next purge */
#endif

/*
 * Name
 *	purge_init
 * Description
 *	Clear out local vars at init time
 *	Link ourselves onto the process notification list
 */
void
purge_init(void)
{
    pid_list = NULL;
    dead_list = NULL;
    dead_block = NULL;
#ifndef __linux__
    proch.handler = proc_notify;
#endif /* !__linux__ */
}

void
purge_active(void)
{
#ifndef __linux__
    prochadd(&proch) ;
#endif /* !__linux__ */
}

/*
 * Name
 * 	get_dead
 * Description
 * 	This function was added after a system crash was traced to the point
 * 	where the proc_notify function was dereferencing memory allocated
 *	from xmalloc for the purpose of cleaning up after dead lock clients.
 *	We could not figure out why this was happening, so we stopped using
 *	xmalloc.  
 *	DESPITE THE ABOVE, this uses xmalloc() from the pinned heap
 * Arguments
 *	None
 * Returns
 *	A pointer to a struct dead_pid_desc or NULL of there's an error.
 * Global Data Affected
 *	dead_list - the head of the dead pool list
 */
static struct dead_pid_desc *
get_dead(void)
{
  struct dead_pid_desc *ret;

  ret = (struct dead_pid_desc *)kmalloc(sizeof(struct dead_pid_desc),
                                        GFP_KERNEL);
  if (NULL != ret)
    ret->dpd_next = NULL;

  return ret;
}

/*
 * Name
 * 	free_dead
 * Description
 * 	This function returns a dead pool structure to the pool.
 * Arguments
 *	dptr - pointer to the structure to be freed
 * Returns
 *	Nothing
 * Global Data Affected
 *	dead_list - the head of the dead pool list
 */
static void
free_dead(struct dead_pid_desc *dptr)
{
  kfree(dptr);

  return;
}

/*
 * Name
 * 	purge_check
 * Description
 * 	This function was written for debug purposes.
 * Arguments
 *	pid - pid of client to check for purge
 * Returns
 *	TRUE if the specified pid is not on the pid_list
 * Global Data Affected
 *	None
 */
int
purge_check(int pid)
{
    struct dead_pid_desc *zombie;

    zombie = pid_list;
    while (zombie) {
	if (zombie->dpd_pid == pid)
	    return 0;
	zombie = zombie->dpd_next ;
    }
    return 1;
}

/*
 * Function:
 *	purge_list
 * Description:
 *	This function is called to clear the list of dead processes that has
 *	been filled by the proc_notify function.
 *
 * Arguments:
 *	NONE
 *
 * Return value:
 *	NONE
 *
 * Side effects:
 *	NONE
 *
 */
void
purge_list (void)
{
    struct dead_pid_desc *zombie;
    clm_client_t	*cp;


    bsdlog(LOG_DEBUG, "[%s:%s] purge_list called, running [%d] \n",
	   current->comm, __FUNCTION__, clm_is_running());

    if (! clm_is_running())
    {
	DBPRINTF((DBG_RECOV, "[%s:%s] Called during rebuild\n",
		  current->comm, __FUNCTION__));
	return;
    }

    while (pid_list) 
    {
	zombie = pid_list;	/* intlk w/LK_ENTRY in proc_notify */
	pid_list = zombie->dpd_next ;

	/*
	 * See if dead process was a client of ours.  Be sure that it is an
	 * actual process.  The pid could also be a group or transaction id.
	 * We want to avoid the case where a real process on the system might
	 * have a pid that just happens to match a group or transaction id.  We
	 * don't want to purge locks that belong to a group by mistake here.
	 */
	if ((cp = findclientbypid(OurSiteID,zombie->dpd_pid))
	    && (CLM_CLI_PROCESS == cp->cli_type)) 
	{
	    bsdlog(LOG_INFO, "[%s:%s] purging pid %d\n",
		   current->comm, __FUNCTION__, zombie->dpd_pid);

	    /* check if client was a member of a group */
	    if (cp->cli_groupid) 
	    {
		/* remove this client from its group */
		grp_rmpid(cp->cli_groupid, zombie->dpd_pid);

		/* group function will purge locks if client was last member */
	    }

	    internal_purge_request(OurSiteID, zombie->dpd_pid, PURGE_DEAD) ;
	}

	free_dead(zombie);
    }
}

/*
 * internal_purge_request
 *
 * generate a purge transaction for an internally generated cause
 * (blame it on the kproc)
 *
 * used by purge_list and anyplace in the migration/purge/reconfig logic
 * which is allowing purges to continue.
 *
 * return 0 for success, 1 for failure
 */
int
internal_purge_request(int site, /* nodeid where process/group lives */
                       int pid,	 /* pid/group id being purged */
                       unsigned int flags)	/* flags determining semantics of message */
{
  clm_trans_t *req;

  /* set up a purge request for client owned locks */
  req = clm_alloc_trans();
  if (!req) {
	bsdlog(LOG_ERR, MSGSTR(FAIL_TO_ALLOCATE,
                           "Allocation failed in internal_purge_request"));
	return 1 ;
  } else {
	set_purge_req(req, site, pid, flags) ;
#ifdef UNIT_TEST
	if (unit_test_delay_purge) {
      unit_test_delay_purge = 0 ;
      unit_test_saved_purge = req ;
      DBPRINTF((DBG_PURGE, "purge %d/%d delayed for unit test\n",
                site, pid)) ;
      return 0 ;
    }
#endif
	DBPRINTF((DBG_BASIC,"Sending purge 0x%x\n",&req));
	clm_request(req);
  }
  return 0 ;
}

#ifdef UNIT_TEST
/*
 * unit_test_send_saved_purge
 *
 * Sends purge request saved by unit test code in internal_purge_request.
 * Resets saved purge flag
 */
void
unit_test_send_saved_purge(void)
{
     if (unit_test_saved_purge)
     {
	 DBPRINTF((DBG_BASIC,"Sending saved purge 0x%x\n",
					unit_test_saved_purge)) ;
	 clm_request(unit_test_saved_purge) ;
	 unit_test_saved_purge = 0 ;
     }
}
#endif

/*
 * Function:
 *	proc_notify
 *
 * Description:
 *	Add a process to the list of those which have died recently,
 * and prod the kproc so that it will clean up appropriately.
 * Notice that this routine is called in the context of the
 * process which is exiting, not in the kproc's.
 *
 * Under AIX this routine was called from the kernel's prochadd
 * handler.  Under Linux, we get called from the close handler
 * for the DLM device, which indicates a client closed its
 * connection.  Otherwise, the handling once we've been notified
 * pretty well works the same way in both.
 *
 * One note is handling death of the kernel thread.  AIX would
 * actually call this routine just as for a user process, but
 * not in Linux.  If our DLM kernel thread dies, our signal
 * handler will most likely get involved, unless we take down
 * the system (very likely.)
 *
 * Arguments:
 *	pid_t pid;			-- pid of affected process or thread
 *					   id of the affected thread
 *
 * Return value:
 *	NONE
 *
 * Side effects:
 *	Sends a SIGUSR1 to the kproc
 */

void
proc_notify (pid_t pid)
{
  struct dead_pid_desc *newpid;
  extern bool_t   dlm_is_going_down;

  bsdlog(LOG_INFO,
         "[%s:%s] dead pid [%d] being reported (kp/up[%d/%d].\n",
         current->comm, __FUNCTION__,
         pid, kproc_pid, uproc_pid);

    if (pid == kproc_pid) {
	kproc_pid = 0;
	while (pid_list) {
	    newpid = pid_list;
	    pid_list = newpid->dpd_next;
	    free_dead(newpid);
	}

	free_all();
    } else if (pid == uproc_pid) {
	/* Our step-parent died.  Commit suicide. */
	uproc_pid = 0;
	(void) dlm_kill_user_proc( kproc_pid, SIGKILL );
    } else {
	/* If the DLM is on its way down, then we do not (and should not)
	 * need to do any purges.  Starting purges at this point can only
	 * cause delays in the shutdown process.  In any case, the DLM is
	 * going down, so the other nodes will automatically purge our
	 * clients anyway!
	 */

	if ( ! dlm_is_going_down ) 
	{
	    newpid = get_dead();
	    if (newpid) {
		dlm_workunit_t * wu;
		extern dlm_workqueue_t * dlm_master_work_queue;


		newpid->dpd_pid = pid;
		acquire_lock(LK_ENTRY, current->pid) ;	/* use lock to be MP-safe */
		newpid->dpd_next = pid_list ;	/* add this desc block to */
		pid_list = newpid ;		/* head of pid_list */
		release_lock(LK_ENTRY, current->pid) ;


		/* Allocate and queue a work unit.  The worst that will happen
		 * if memory can't be allocated is purging of the client's
		 * locks will be delayed.  If we're out of memory, we'll
		 * probably crash soon, so all of our locks will be purged
		 * for us.  How nice.
		 */

		wu = CLM_MALLOC( sizeof( *wu ) );
		if ( wu != NULL )
		{
		    wu->type = MQ_CLIENT_IS_DEAD_MSG;
		    wu->data = NULL;
		    wu->free_data = NULL;

		    dlm_workqueue_put_work( dlm_master_work_queue, wu );
		}
	    }
	}
    }
}

/*
 * Name
 *	begin_purge
 * Description
 * 	local interface to purge logic, calls purgelocks
 *  NEW
 *	if all nodes 4.2b,
 *	     send 'want to purge' to all nodes (incl. self)
 *	this starts sequence of states which culminate in all nodes purged
 *	since this is not unitary, node up/down must be observed & handled
 * Arguments
 *	req - the transaction containing a purge request 
 * Returns
 * 	DLM_DENIED_GRACE_PERIOD if lock stall is in effect and
 *	client to be purged is not the staller
 *	DLM_NORMAL in the transaction
 * Global Data Affected
 * 	NONE	
 */
int
begin_purge(clm_trans_t *req)
{
  struct purgereq *prg = &req->clm_purgelocks ;
  clm_client_t *cp ;
  clm_client_t dummy ;
  int site = prg->purgesite ;
  int pid = prg->purgepid ;
  int flags = prg->flags ;
  int status ;

  TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_LOCAL_PURGE,
          site, pid, flags, 0x00);

  DBPRINTF((DBG_PURGE, "enter %s for site=%d pid=%d  flags %x\n",
            __FUNCTION__,
            site, pid, flags));

  flags &= ~PURGE_REMOTE;

  /* Add these to remove need for clinfo in API */
  req->clm_purgelocks.site = OurSiteID;
  if (site && clm_node_active(site) == FALSE) {
    req->clm_status = DLM_BADARGS;
    goto badreq;
  }

  /*
   * if request is not from purged AND it is a request
   * to purge a specific client, reject it if the client is still alive
   * For backwards compatibility, allow any purge to any other site
   */
  cp = findclientbypid(site, pid) ;
  if (((flags & PURGE_DEAD) == 0) &&
      site &&
      (site == prg->site) &&
      pid &&
      (pid != prg->pid) &&
      cp) {
    req->clm_status = DLM_DENIED;
    DBPRINTF((DBG_FORCE,
              "Purge request rejected: client not dead"));
    goto badreq ;
  }

    if ( cp == NULL )
    {
	/* fake client struct for purge everyone */
	if ( pid == 0 )
	{
	    (void) memset( & dummy, 0, sizeof dummy );
	    dummy.cli_site = site ;
	    set_client_id(dummy.cli_id, pid) ;
	    cp = &dummy ;
	}
	else
	{
	    if ( ! addclientbypid( site, pid, CLM_CLI_PURGE, & cp ) )
	    {
		DBPRINTF((DBG_PURGE, "can't add client for purge\n")) ;
		abort("begin_purge 2") ;
		/* NOTREACHED */
		req->clm_status = DLM_DENIED ;
		goto badreq ;
	    }

	    CLM_ASSERT( cp != NULL, "client added, but pointer is NULL!" );
	}
    }

    /* if client is purging itself, the set the PURGE_ME bit */
    if (site == prg->site && pid == prg->pid)
    {
	flags |= PURGE_ME;
    }

    flags |= PREQ_START ;


    /* always send request to master functionality */
    pti_purge(cp, flags | PURGE_REMOTE, site);

    req->clm_status = DLM_NORMAL;

badreq:
    status = req->clm_status ;

    /* If purges were to be made synchronous, this would have to be put
     * somewhere else
     */

    clm_reply(req, 0) ;

    return status ;
}


/*
 * Name
 * 	pti_purge_complete
 * Description
 *	This function is called by the communications layer when it has
 *  	received an acknowledgement for the outbound request.
 *	NOTE:  We do want to free the transaction structure pointed to by the
 *	message structure here because purge requests do not get queued
 *	on the inflight list so they won't be freed after a response
 *	that won't come back.
 *
 * NEW  if the purge is stateful, mark client structure & progress
 *	state machine.
 *
 * Arguments
 *	msg - pointer to the message structure that was acknowledged
 *	stat - status returned
 * Returns
 *	Nothing
 * Global Data Affected
 *	NONE
 */
void
pti_purge_complete(clm_msg_t *msg, int stat)
{
  clm_trans_t *req = (clm_trans_t *)msg->cm_data;
  struct purgereq *prg = &req->clm_purgelocks ;
  clm_client_t *cp ;

  switch (stat)
	{
	case 0:
      break ;

	case CPERR_NODEDOWN:
      if ( ! (prg->flags & PREQ_MASK)	/* ignore if old (stateless) vers */
           || (prg->flags & PREQ_MASK) == PREQ_DONE)	/* ign last msg */
	    {
          break ;
	    }
      cp = findclientbypid(prg->purgesite, prg->purgepid) ;
      if ( ! cp)
	    {
          bsdlog(LOG_ERR, "%s: error on nonexistent client req",
                 __FUNCTION__);
          break ;
	    }
      if ( ! (prg->flags & PRSP_BIT))	/* is this a request? */
	    {
          handle_purge_reply(cp, msg->cm_node, prg->flags) ;
	    }
      else
	    {
          /* master died - release everything */
          /* I don't think anything needs to happen here - cleanup in reconfig */
	    }

	default:
      bsdlog(LOG_ERR, MSGSTR(PTI_CALL_ERR,
                             "Error detected in pti_purge_complete\n"));
      break ;
	}

  clm_free_trans(req);

  /*
   * we do apparently after careful consideration have to free the
   * message pointer
   */
  cccp_msg_destroy( msg );
}

/*
 * Name
 * 	pti_purge
 * Description
 *	Given the site and pid of the client being purged,
 *	forward a purge request to the destination specified.
 *	That site (may be us) will do the next thing in the purge sequence.
 *	Message is QUEUED to be sent when current processing done.
 *	This prevents recursive calls, which cause synchronization problems.
 * Arguments
 *	cp - client struct for the client being purged
 *	flags - purge control flags
 *	dest - destination for this message
 * Returns
 *	0 for sucess
 *	< 0 for failure
 * Global Data Affected
 *	NONE
 */
int
pti_purge(clm_client_t *cp,
          unsigned int flags,
          int dest)
{
    int ret = 0 ;
    clm_msg_t	*msg;
    clm_trans_t	*req;
    int site, pid;

    DBPRINTF((DBG_FUNC, "Enter %s\n", __FUNCTION__));

    site = cp->cli_site ;
    pid = CLID2PID(cp->cli_id) ;

    DBPRINTF((DBG_PURGE, "%s: Send purge pid=%d site=%d flags=%x to %d",
	      __FUNCTION__, pid, site, flags, dest));

    req = clm_alloc_trans();
    if ( req == NULL )
    {
	bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC10,
				"malloc failed in pti_purge()\n"));
	abort("pti_purge");
    }
    
    msg = cccp_msg_create( PORT_MSTR, RESRPRI, dest, DLM_PURGELOCKS,
			   sizeof( *req ), pti_purge_complete, req, NULL );
    if ( msg == NULL )
    {
	bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC10,
				"malloc failed in pti_purge()\n"));
	abort("pti_purge");
    }
    

    set_purge_req(req, site, pid, flags) ;

    ret = clm_queue_msg(msg) ;
    DBPRINTF((DBG_COMM, "%s: cccp_send returned %d\n", __FUNCTION__, ret));
    return ret ;
}

/*
 * set_purge_req
 *
 * standard setup for struct transaction for a purge
 * called by purge_list() and pti_purge()
 */
void
set_purge_req(
	struct transaction *req,	/* request to fill in */
	int site,			/* site of process to be purged */
	int pid,			/* pid to purge */
	int flags)			/* request flags */
{
    req->clm_prog = DLM_PROG ;
    req->clm_vers = clm_low_version() ;
    req->clm_status = DLM_NORMAL ;
    req->clm_type = DLM_PURGELOCKS ;
    req->clm_direction = PTI_REQUEST ;
    /* clm_locktype? */
    req->clm_sequence = 7642 ;
    req->clm_route.respflags = RESP_UDP | RESP_NULL ;
    req->clm_pid = kproc_pid ;
    req->clm_purgelocks.purgesite = site ;
    req->clm_purgelocks.purgepid = pid ;
    req->clm_purgelocks.pid = kproc_pid ;
    /* .site?? */
    req->clm_purgelocks.flags = flags ;
}

/*
 * Name
 *	master_purge
 * Description
 *	Handle purge request sent from secondary.
 *
 * NEW  Handles clm_client_t state machine to drive purges
 *      Does both client and server operations,
 *	so things are a bit intertwingled.
 *
 *	For mixed clusters, multicast the "do purge" message to all nodes.
 *
 *	For new clusters:
 *		advance client state for responses & multicast next reqest
 *		advance server state for requests & do purge when OK
 *
 * Arguments
 *	msg - the purge message
 * Returns
 *	0
 *	Due to architecture restrictions in CCCP, no errors may be reported
 * Global Data Affected
 *	NONE
 */
int
master_purge(clm_msg_t *msg)
{
  clm_trans_t	*req = msg->cm_data;
  struct purgereq *prg = &req->clm_purgelocks ;
  clm_client_t *cp = NULL ;
  clm_client_t dummy ;
  int ret = 0;
  int err,
    site,
    pid ;
  unsigned int
    flags,
    oldclistate ;

  site = prg->purgesite ;
  pid = prg->purgepid ;
  flags = prg->flags ;

  DBPRINTF((DBG_FUNC|DBG_PURGE,
            "enter master_purge for site=%d pid=%d  flags %x\n",
            site, pid, flags));

  /*
   * Request to purge all clients on a node are funny.
   *
   *
   * Find client record for this PID.
   * If not found
   *    first time through - create (abort if fail)
   *    subsequently - ick
   */
    if ( ! pid) 
    {
	cp = &dummy ;
	(void) memset(&dummy, 0, sizeof dummy) ;
	dummy.cli_site = site ;
	set_client_id(dummy.cli_id, 0) ;
    }
    else if ( ! (cp = findclientbypid(site, pid))) 
    {
	if (((flags & PREQ_MASK) != PREQ_WANT)	/* legal start states */
	    && ((flags & PREQ_MASK) != PREQ_START)) 
	{
	    DBPRINTF((DBG_PURGE,
		      "%s: bad start - no client for %d/%d flags 0x%x\n",
		      __FUNCTION__, site, pid, flags)) ;
	    return 0 ;
	}

	if ( ! addclientbypid( site, pid, CLM_CLI_PURGE, & cp ) ) 
	{
	    bsdlog(LOG_ERR, "%s: can't add client %d/%d\n",
		   __FUNCTION__, site, pid) ;
	    abort("master_purge: client not found & not created") ;
	    return 0 ;
	}

	CLM_ASSERT( cp != NULL, "client added, but pointer is NULL!" );
    }

  oldclistate = cp->cli_state ;
  switch (flags & PREQ_MASK) {
  case 0:
    /*
     * This must be happening in a mixed version cluster.
     * this call to purgelocks will remove all locks owned by the
     * specified client.  Any resources left without references will be
     * removed.  If PURGE_DEAD flag is set, the client record will also
     * be deleted.  We have to perform this operation whether this
     * resource is mastered here or not, because the sender will assume
     * that the purge has been carried out on this node and will not
     * reissue the request for any subsequent resources that it thinks
     * are mastered here.
     */
    purgelocks(site, pid, flags) ;
    if (flags & PURGE_REMOTE) {
      /* forget trying to be smart.  Just blast them all! */
      err = purge_all(cp, (flags & ~PURGE_REMOTE), 1) ;
    }
    if (flags & PURGE_DEAD || cp->cli_type == CLM_CLI_PURGE) {
      remove_client_inflight(cp) ;
      removeclientbypid(site, pid) ;
    }
    break ;
	
  case PREQ_START:
    /*
     * CLIENT (node where process being purged lives)
     * Entry into state machine from single message sent by
     * proc_entry or user mode purge request
     *
     * Always executed on node == site field of message
     * which is where the 'main' client record should live.
     * Broadcast a 'WANT' to purge to all nodes (incl self).
     */
    if (cp->cli_site != OurSiteID) {
      DBPRINTF((DBG_PURGE,
                "%s: PREQ_START %d/%d to wrong node (%d)\n",
                __FUNCTION__, cp->cli_site, CLID2PID(cp->cli_id), OurSiteID)) ;
      abort("purge start sent to wrong node") ;
      break ;		/* DISCARD */
    }

    /*
     * check for priority request (DEAD) coming in during purge
     * processing if we originated the incoming purge and it's a
     * request to purge a user process that exited, mark it for
     * reprocessing after this purge is done
     */
    if (cp->cli_state & CLI_CODE) {
      if (prg->flags & PURGE_DEAD) {
        cp->cli_state |= CLI_NEED_DEAD ;
        DBPRINTF((DBG_PURGE,
                  "purge dead queued - purge in progress %d/%d\n",
                  cp->cli_site, CLID2PID(cp->cli_id))) ;
        break ;	/* QUEUED */
      }
      /*
       * conflict resolution - discard START if purge in progress
       */
      DBPRINTF((DBG_PURGE,
                "purge rejected - purge in progress %d/%d\n",
                cp->cli_site, CLID2PID(cp->cli_id))) ;
      break ;		/* DISCARD */
    }
    handle_purge_reply(cp, msg->cm_node, prg->flags) ;
#ifdef UNIT_TEST
    if (unit_test_hold_migrations) {
      unit_test_hold_migrations = 0 ;
      DBPRINTF((DBG_PURGE, "purge %d/%d allowing migrations\n",
                cp->cli_site, CLID2PID(cp->cli_id))) ;
    }
#endif
    break ;		/* HANDLED */

  case PREQ_WANT:
    /*
     * SERVER
     * new style 'want to purge' request
     *
     * if request to purge everything at a node,
     *	go call reconfig logic to step through this, because
     *	it's a global operation
     *	reconfig code will call purgelocks at the right time
     */
    if ( ! pid) {
      clms_node_purge(site) ;
      break ;
    }
    /*
     * check to see if allowed at this time,
     *    if so, reply GO AHEAD,
     *    else mark as needing a 
     *
     * if request to purge comes in when migration in progress for
     *   some resource on which this client has a lock,
     *   or reconfig is in process,
     *   delay purge until migration/reconfig done
     *   & hold request flags & originator in clm_client_t
     */
    if (cp->cli_migrcount || ! clm_run_rcv()) {
      cp->cli_state |= CLI_NEED_PURGE | (prg->flags & PURGE_USRMASK) ;
      /* QUEUED */
      DBPRINTF((DBG_PURGE,
				"%s: purge %d - %d deferred", __FUNCTION__,
				site, pid));
    }
    else {
      pti_purge(cp, (flags & ~ PREQ_MASK) | PRSP_GA, site) ;
      /* HANDLED */
    }
    CLI_SV_SET(cp, prg->flags) ;
    break ;		/* */
	
	/*
	 * CLIENT
	 * when all servers have said Go Ahead, send DOIT
	 * if process died in the meantime, add PURGE_DEAD to flags
	 */
  case PRSP_GA:
    if (cp->cli_state & CLI_NEED_DEAD) {
      DBPRINTF((DBG_PURGE,
                "%s merging PURGE_DEAD for %d/%d\n",
                cp->cli_site, CLID2PID(cp->cli_id))) ;
      cp->cli_state |= PURGE_DEAD ;
      cp->cli_state &= ~CLI_NEED_DEAD ;
      prg->flags |= PURGE_DEAD ;
    }
    handle_purge_reply(cp, msg->cm_node, prg->flags) ;
    break ;		/* HANDLED */

	/*
	 * CLIENT
	 * last response for this causes transition to idle
	 * see 'PREQ_DONE' hack for restart of anything needed
	 */
  case PRSP_PRGD:	
    handle_purge_reply(cp, msg->cm_node, prg->flags) ;
    if (CLI_CL_GET(cp) == PREQ_DONE)
      CLI_CL_SET(cp, 0) ;
    break ;		/* HANDLED - CLIENT EXIT */
	
	/*
	 * SERVER
	 * do the purge now
	 * reply 'did purge'
	 */
  case PREQ_DOIT:
    if (CLI_SV_GET(cp) != PREQ_WANT)
      abort("bad state when purge requested") ;
    purgelocks(site, pid, flags) ;
    pti_purge(cp, (flags & ~ PREQ_MASK) | PRSP_PRGD, site) ;
    CLI_SV_SET(cp, prg->flags) ;
    break ;		/* HANDLED */
	
	/*
	 * SERVER
	 * done - allow migrations
	 * if client dead or client structure set up for this purge,
	 *     delete it
	 * set server state to 0 (idle)
	 * ***HACK*** since this is the time when both client and
	 * server state both are 0, restart anything queued
	 */
  case PREQ_DONE:
    if (CLI_SV_GET(cp) != PREQ_DOIT)
      abort("bad state when purge requested") ;
    CLI_SV_SET(cp, 0) ;		/* set to idle */
    if (flags & PURGE_DEAD || cp->cli_type == CLM_CLI_PURGE) {
      remove_client_inflight(cp) ;
      removeclientbypid(site, pid) ;
      cp = NULL ;
    }

    DBPRINTF((DBG_PURGE, "Purge done.  cp = 0x%x", cp));
    /*
     * continue purge if new purge needed
     * start new purge if purge dead needed
     */
    if (cp && cp->cli_site == OurSiteID)
      reconfig_allow_client_purge(cp, (void *) __FUNCTION__) ;
    break ;		/* HANDLED - SERVER EXIT */

  case PRSP_FAIL:
    /*
     * CLIENT
     * a server has detected a conflict with another purge request
     * abandon this request after last response
     */
    DBPRINTF((DBG_PURGE,
              "Purge failed.  cp = 0x%x", cp));
    handle_purge_reply(cp, msg->cm_node, prg->flags) ;
    break ;		/* HANDLED */
  }

  DBPRINTF((DBG_FUNC, "leave master_purge ret = %d\n", ret));
  return ret;
}

/*
 * purge_all
 *
 * QUEUE purge message(site, pid) to all nodes that are up
 * qualify the message with the flag bits
 * record node state & msg sent in client struct
 * for each message sent record PREQ flags (for error checking) and 1
 *   to mark message sent in nr_resp field of cli_resp[i]
 * if 'notme' is set, don't send to self to prevent loop
 *
 */
static int
purge_all(
          clm_client_t *cp,		/* client being purged */
          unsigned int flags,		/* flags for purge msg & set in state */
          int notme)			/* 1 to prevent send-to-self */
{
    int site ;
    int pid ;
    int i ;

    site = cp->cli_site ;
    pid = CLID2PID(cp->cli_id) ;

    if (CLI_CL_GET(cp) == PREQ_WANT) 
    {
	/*
	 * first time for this purge - set up node array.  Subsequent passes
	 * will only send to nodes that we init'd on the first pass.  This
	 * prevents us from including nodes who come up during a purge.
	 */

	for (i = 0 ; i < MAXNODES ; i++) {
	    int node ;

	    node = INDEX_TO_NODEID(i) ;
	    if ( ! (node == OurSiteID && notme) )
	    {
		if (node != -1 && clm_node_active(node)) {
		    cp->cli_resp[i].nr_node = index_to_nodeid(i) ;
		    cp->cli_resp[i].nr_resp = 1 | (flags & PREQ_MASK) ;
		    pti_purge(cp, flags, node) ;
		}
		else 
		{
		    cp->cli_resp[i].nr_node = 0 ;
		    cp->cli_resp[i].nr_resp = 0 ;
		}
	    }
	}
    }
    else
    {
	/*
	 * here, we're either sending out the DOIT or the DONE msg.  Only
	 * send these to nodes that were set up in the array on the first
	 * pass and are still considered to be active
	 */
	for (i = 0 ; i < MAXNODES ; i++) 
	{
	    int node ;

	    node = INDEX_TO_NODEID(i) ;
	    if (cp->cli_resp[i].nr_node) {
		if (clm_node_active(node)) {
		    cp->cli_resp[i].nr_resp = 1 | (flags & PREQ_MASK) ;
		    pti_purge(cp, flags, node) ;
		} else {
		    /*
		     * if a node goes dead mid-purge, clear out the slot in
		     * the array so we won't try to include it again if it
		     * becomes active again
		     */
		    cp->cli_resp[i].nr_node = 0;
		}
	    }
	    else 
	    {
		cp->cli_resp[i].nr_resp = 0 ;
	    }
	}
    }


    return 0 ;
}

/*
 * handle_purge_reply
 *
 * counts current reply against client struct
 * when last client is accounted for, sets next state in master machine
 * & sends next purge message to all servers
 */
void
handle_purge_reply(clm_client_t *cp,	/* master client struct */
                   int sender,		/* node ID sending resp */
                   unsigned int msgflags)	/* flags from incoming message */
{
  int i;
  unsigned int reqcode ;	/* request code derived from msg */

  reqcode = msgflags & PREQ_MASK & ~PRSP_BIT ;
  /*
   * check for valid node replying
   * change state for replying node
   * check for any node not yet replied
   * don't check if this is the startup message
   */
  if ((msgflags & PREQ_MASK) != PREQ_START)
    i = account_for_reply(cp, sender, msgflags) ;
  else
    i = 0 ;
  if (i < 0) {
    DBPRINTF((DBG_PURGE, "%s: node %d not found in client array\n",
              __FUNCTION__, sender)) ;
  }
  if (i)			/* have all nodes responded ? */
    return ;		/*   no - finished */
  switch (reqcode) {
  case PREQ_CANCEL:
	/*
	 * clear to client idle
	 * if we were doing delayed replies, we'd send one here
	 */
	CLI_CL_SET(cp, 0) ;
	break ;

  case 0: /* really PREQ_START: */
  case PREQ_WANT:
  case PREQ_DOIT:
	/*
	 * if clm_client_t purge client state not same as message, punt
	 */
	if (CLI_CL_GET(cp) != reqcode) {
      DBPRINTF((DBG_PURGE, "%s: %d/%d in wrong state %x for msg %x\n",
                __FUNCTION__, cp->cli_site, CLID2PID(cp->cli_id),
                cp->cli_state, msgflags)) ;
      break ;
	}
	reqcode = PREQ_NEXT_CODE(reqcode) ;	/* next state / msg code */
	CLI_CL_SET(cp, reqcode) ;		/* save in clm_client_t */
	purge_all(cp, (msgflags & ~PREQ_MASK) | reqcode, 0) ;
	break ;

  default:
	DBPRINTF((DBG_PURGE, "wrong state %x in purge reply\n", msgflags)) ;
	break ;
  }
}

/*
 * account_for_reply
 *
 * checks sender of incoming message for validity
 * checks incoming reply message against state array set by purge_all
 * updates msg state array for sender of this reply
 *
 * if any error
 *	return -1
 * if all possible senders have sent a reply,
 *    return 0
 * else return 1
 */
static int
account_for_reply(
                  clm_client_t *cp,	/* master client struct */
                  int sender,		/* node ID sending resp */
                  unsigned int msgflags)	/* flags from incoming message */
{
    int               i;
    bool_t            found;
    bool_t            any_out;
    struct noderesp * nrp;
    unsigned          reqcode;	/* request code derived from msg */


    reqcode = msgflags & PREQ_MASK & ~PRSP_BIT ;

    /* Check for valid node replying.  Change state for replying node.  Check
     * for any node not yet replied
     */

    for (i = 0, nrp = cp->cli_resp, any_out = FALSE, found = FALSE ;
	 i < MAXNODES ; i++, nrp++) 
    {
	if ( nrp->nr_node != 0 )
	{
	    /* found right slot in resp array? */
	    if (nrp->nr_node == sender) 
	    {
		found = TRUE;

		/* already seen a reply !?!? */
		if ( ! (nrp->nr_resp & 01)) 
		{
		    DBPRINTF((DBG_PURGE, "%s: %d/%d  node %d already replied?\n",
			      __FUNCTION__, cp->cli_site, CLID2PID(cp->cli_id), sender)) ;
		}
		else 
		{
		    /* State match message?  Or is this a conflict detected by a
		     * server?
		     */

		    if ( ((nrp->nr_resp & PREQ_MASK) != reqcode)
			&& ((nrp->nr_resp & PREQ_MASK) != PREQ_WANT
			    || reqcode != PREQ_CANCEL) )
		    {
			DBPRINTF((DBG_PURGE,
				  "%s: node %d state %x wrong msg flags %x?\n",
				  __FUNCTION__, cp->cli_site, CLID2PID(cp->cli_id), sender,
				  nrp->nr_resp, msgflags)) ;
		    }
		    else 
		    {
			nrp->nr_resp = 0;	/* say msg received */
		    }
		}
	    }
	
	    if (nrp->nr_resp & 01)
		any_out = TRUE ;

	} /* if ( nrp->nr_node != 0 ) ... */
    }


    return( (! found) ? -1 : any_out );
}

/*
 * res_check_client_purge
 *
 * interlock between purge and migration
 *
 * checks all client structures for all reclocks queued on this resource
 * if any are being purged, return 1
 * else increment migr count for each client for each lock to
 *   prevent purge from starting before migration
 *   return 0
 */

int
res_check_client_purge(struct resource *rp)
{
  clm_client_t *cp ;
  struct reclock *a ;
  int i, anypurge ;

    for (i = 0, anypurge = 0 ; i < RLQ_MAX ; i++) 
    {
	for ( EACH_DLM_LISTNODE( & rp->lockqs[i], a ) )
	{
	    if ( ! (cp = findclientbypid(a->lck.site, a->lck.pid))) 
	    {
		if ( ! addclientbypid( a->lck.site, a->lck.pid, CLM_CLI_PURGE,
				       & cp ) )
		{
		    DBPRINTF((DBG_PURGE, "can't add client for migration\n"));
		    abort("res_check_client_purge");
		    /* NOTREACHED */
		}

		CLM_ASSERT( cp != NULL, "client added, but pointer is NULL!" );
	    }

	    if (cp->cli_state & CLI_CODE)	/* purging? */
		anypurge = 1 ;
	}
    }

    if (anypurge)				/* fail if any purging */
	return 1 ;

    /* Re-scan lock list and increment migration count
     */

    for (i = 0, anypurge = 0 ; i < RLQ_MAX ; i++)
    {
	for ( EACH_DLM_LISTNODE( & rp->lockqs[i], a ) )
	{
	    lock_incr_client_migr( a );
	}
    }


    return 0 ;
}

/*
 * lock_incr_client_migr
 *
 * find the client which owns this reclock
 * increment the migration count so purge won't happen until migr done
 */

int
lock_incr_client_migr(struct reclock *a)
{
    clm_client_t *cp ;

    if ( ! (cp = findclientbypid(a->lck.site, a->lck.pid)))
      return 1 ;
    cp->cli_migrcount++ ;
    return 0 ;
}

/*
 * res_allow_client_purge
 *
 * interlock between migration and purge
 *
 * for every lock on this resource (migrated or not),
 *    decrement migration count in corresponding clm_client_t
 *    check if client->migrcount == 0 && ->state == need a purge
 *	if so, initiate a purge for that client
 *
 * returns 0 for success, nonzero for ???failures???
 */
 
int
res_allow_client_purge(struct resource *rp)
{
    struct reclock * a;
    unsigned         i;
    bool_t           ret;


    ret = FALSE;
    for ( i = 0 ; i < RLQ_MAX ; i++ )
    {
	for ( EACH_DLM_LISTNODE( & rp->lockqs[i], a ) )
	{
	    if (lock_allow_client_purge(a))
		ret = TRUE;
	}
    }

    return ret ;
}

/*
 * lock_allow_client_purge
 *
 * interlock between migration and purge
 *
 * called for each lock no longer involved in migration (either directly
 * or from res_allow_client_purge())
 *
 * decrements client migrcount
 * initiates/continue purges if migrcount == 0
 *
 * returns nonzero for failure
 */
int
lock_allow_client_purge(struct reclock *a)
{
  int ret = 0 ;
  clm_client_t *cp ;
  struct resource *rp ;

  if ( ! (cp = findclientbypid(a->lck.site, a->lck.pid))) {
	if (!(rp = resptr(a->lck.rh))) {
      bsdlog(LOG_ERR, MSGSTR(NO_HANDLE,
                             "lock_allow_client_purge: can't find handle for 0x%x"),
             a->lck.lockid);
      return 1 ;
	}

	DBPRINTF((DBG_PURGE,
              "finishing migr %d:%s - can't find client record for %d/%d\n",
              rp->namelen, printbin(rp->name, rp->namelen),
              a->lck.site, a->lck.pid)) ;
	return 1 ;
  }
  if ( --cp->cli_migrcount < 0) {
	if (!(rp = resptr(a->lck.rh))) {
      bsdlog(LOG_ERR, MSGSTR(NO_HANDLE,
                             "lock_allow_client_purge2: can't find handle for 0x%x"),
             a->lck.lockid);
      return 1 ;
	}

	DBPRINTF((DBG_PURGE,
              "finishing migr %d:%s - migrcount negative for %d/%d\n",
              rp->namelen, printbin(rp->name, rp->namelen),
              a->lck.site, a->lck.pid)) ;
	abort("lock_allow_client_purge") ;
  }
  if ( ! cp->cli_migrcount && ! clm_in_reconfig())
	ret = reconfig_allow_client_purge(cp, "migration") ;
  return ret ;
}

/*
 * reconfig_free_client_purge
 *
 * This function is called for each client at the beginning of reconfig to
 * reset the migrcount and restart any purges that may be blocked.  All
 * purges must complete before reconfig can proceed.  It should be OK to
 * clear migrcount during reconfig because migrations are reset anyways.
 *
 *    clear migrcount and call reconfig_allow_client_purge to perform the
 *	remainder of the operation
 *
 * returns value from reconfig_allow_client_purge
 */
int
reconfig_free_client_purge(clm_client_t *cp, /* supplied by client_apply() */
                           void *arg)
{
    cp->cli_migrcount = 0;

    return reconfig_allow_client_purge(cp, arg);
}

/*
 * reconfig_allow_client_purge
 *
 * release purge interlock
 *
 * called by client_apply() for every lock on this resource (migrated or not),
 *	by reconfig when done
 * called by migration & purge itself to continue/restart
 *
 *    check if client->migrcount == 0 && ->state == need a purge
 *	if so, continue a purge for that client
 *		if NEED_DEAD, force PURGE_DEAD into client state
 *    else if NEED_DEAD, start a purge cycle
 *
 * returns 0 for success, nonzero for ???failures???
 */

int
reconfig_allow_client_purge(clm_client_t *cp, /* supplied by client_apply() */
                            void *arg)
{
    if (cp->cli_migrcount)
	return 0 ;

    /* not purging now? */
    if ( ! (cp->cli_state & CLI_CODE)) 
    {
	/*
	 * start a new purge cycle for a process requesting it while
	 * reconfig was in progress
	 */

	if (cp->cli_state & CLI_NEED_PURGE) 
	{
	    DBPRINTF((DBG_PURGE, "%s starting purge for %d/%d\n",
		      arg, cp->cli_site, CLID2PID(cp->cli_id))) ;

	    cp->cli_state &= ~ CLI_NEED_PURGE ;
	    internal_purge_request(cp->cli_site, CLID2PID(cp->cli_id), 0) ;
	}
	else if (cp->cli_state & CLI_NEED_DEAD) 
	{
	    /*
	     * start a new purge cycle for a process that died while
	     * (a) reconfig was in progress
	     * or
	     * (b) purging was in progress
	     */

	    DBPRINTF((DBG_PURGE, "%s starting purge dead for %d/%d\n",
		      arg, cp->cli_site, CLID2PID(cp->cli_id))) ;

	    internal_purge_request(cp->cli_site, CLID2PID(cp->cli_id),
				   PURGE_DEAD) ;
	}
    }
    else 
    {
	/*
	 * continue a purge held off because a migration was in progress
	 */

	if (cp->cli_state & CLI_NEED_PURGE) 
	{
	    DBPRINTF((DBG_PURGE, "%s continuing purge for %d/%d\n",
		      arg, cp->cli_site, CLID2PID(cp->cli_id))) ;

	    cp->cli_state &= ~ CLI_NEED_PURGE ;
	    pti_purge(cp, PRSP_GA | (cp->cli_state & PURGE_USRMASK),
		      cp->cli_site) ;
	}
    }

    return 0 ;
}

/*
 * reconfig_check_client_purge
 *
 * interlock between reconfig and purge
 *
 * called by client_apply() for every lock on this resource (migrated or not),
 *    return nonzero if purge in process (for drainout code)
 *
 * returns 0 for no purge in progress
 */

int
reconfig_check_client_purge(clm_client_t *cp, /* supplied by client_apply() */
                            void *arg)
  /* ARGSUSED */
{

    /* check for purging state for clients on active nodes */
    if (clm_node_active(cp->cli_site) && cp->cli_state & CLI_CODE)
        return 1 ;
    return 0 ;
}

/*
 * remove_inflight_client
 *
 * Just before deleting the client, remove all lock records
 * marked 'in flight' from the client structure
 * The struct reclock will be reclaimed by cont_lock_clm when it
 * finds no client struct for it.
 *
 * There shouldn't be any other locks remaining.
 */
static int
remove_client_inflight(clm_client_t *cp)
{
    dlm_listnode_t * curr;
    dlm_listnode_t * next;
    struct reclock * a;


    for ( curr = dlm_list_get_head( & cp->cli_queue )
	  ; curr != NULL 
	  ; curr = next) 
    {
	next = dlm_list_succ( curr );
	a = CLIENT_LINK_TO_RECLOCK( curr );


	/* If this client has a lock queued to it with its INFLIGHT bit set
	 * (which is unlikely) or has its LOCK_ORPHANED bit set (much more
	 * likely), then remove it from the client queue because it wasn't
	 * removed when the lock was dequeued from a resource.
	 */

	if (a->lck.state & (INFLIGHT_WAITING | LOCK_ORPHANED)) 
	{
	    remove_queue(&cp->cli_queue, a, RL_CL_OFS) ;
	}
	else
	{
	    DBPRINTF((DBG_PURGE,
		      "%s: lock %x on client %d/%d has state %x - skipped\n",
		      __FUNCTION__,
		      cp->cli_site, CLID2PID(cp->cli_id), a->lck.state)) ;
	    abort("remove_client_inflight") ;
	}
    }


    return 0 ;
}
