/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_resource.c,v 1.15 2001/10/29 21:49:46 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		RESNAME
 *		add_freelist
 *		allocate_res_seg
 *		allocate_resource
 *		begin_register
 *		cont_register
 *		deallocate_resource
 *		dump_reshash
 *		find_reshash
 *		get_free_res
 *		hashfunc
 *		lookup_resource
 *		master_register
 *		nmcmp
 *		receive_lock
 *		receive_resource
 *		resource_table_apply
 *		send_resource
 *		unregister_resource
 */

#ifndef lint
#endif

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/errno.h>
#include <linux/types.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "cltrc.h"
#include "clm_data.h"			/* external data definitions */
#include "dlm_kernel.h"
#include <iterator.h>

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */

#ifdef DEBUG_ALL_PROTO /* hide from prototype file but not debugger */
#define STATIC
#else
#define STATIC	static
#endif

#define RESNAME(rp)		printbin(rp->name, rp->namelen)
#define	DMP_SZ	40

/****************************************************************************/
/* restab_iterator is a concrete implementation of the abstract iterator
 * type.
 */

struct restab_iterator
{
    iterator_t   base;
    unsigned   next_seg;
    unsigned   next_cookie;
};

static bool_t resource_table_iterator_next( iterator_t * iter, 
					    void ** resource );
static void resource_table_iterator_destroy( iterator_t * iter );

static const struct iterator_methods restab_iterator_methods =
{
    destroy: resource_table_iterator_destroy,
    next: resource_table_iterator_next
};




/****************************************************************************/

/* This structure defines a resource hash bucket link lists of these hang
 * off the resource hash table array.
 */

struct reshash 
{
    dlm_listnode_t    link;
    struct resource * hash_rp;
};

static struct resource *restab[MAX_RES_SEGS];/* resource table segment array */
static DECLARE_DLM_LIST( reshash_free );
static dlm_list_t * reskeytab;	/* the resource hash table */
static struct resource *res_free = NULL;
int hashtablesize;		/* size of resource hash table */

/*
 * resindex and res_seg are used to track the next unused resource
 * structure.  their values should only be changed on the primary.
 */
static int resindex = 0;
static int res_seg = 0;

static int cur_seg = -1;	/* index of last allocated resource segment */

static int add_null_lock(struct transaction *req, struct resource *rp,
							int pid, int site) ;
/* 
 * 3/30/93:	Some aspects of resource management have changed to
 * allow for dynamic expansion of the resource table.  Details and
 * differences are at the end of this comment.
 *
 * This file implements the resource manager module.
 * All lockable objects are represented as resources. All resources
 * are maintained in the "restab" data structure. On the primary, all
 * resources existing in the cluster are present. On a secondary, only
 * those resources being used by local clients. Multiple instances of
 * resources with the same name can coexist provided each is associated with a 
 * different lock family (locktype). Otherwise, all references to a resource
 * resolve to the same entry in the table. Resources are allocated on the
 * primary. 
 *
 * A "resource handle" is passed back to the secondary. This handle
 * contains the index into the resource table to locate the resource. The 
 * secondary allocates the corresponding slot. In this way, all resources
 * across the cluster interleave into the global resource table. In the
 * event of primary failure, the table is reconstructed from the elements
 * known by each secondary. The "resource handle" is kept with a lock object
 * (reclock) and used for fast access to the resource entry in "restab". It
 * is sometimes necessary to search the table (only on the primary) when 
 * creating a new resource. To this end, a second table is maintained, 
 * "reskeytab". This is a hash table of pointers into "restab". This provides
 * quick search capability used for ADD,FIND, and DELETE operations.
 *
 * For some lock families, it is necessary to explicitly create (register)
 * a resource. For others, creation is implicit on first reference to a lock.
 * In the case of the former, a mechanism must exist to determine when a 
 * resource might be purged.
 * OLD:
 * This function is provided by resource 
 * reference (resref). Each time a local client references a resource, an entry
 * is made in the resref chain associated with the resource. When the number
 * of references to a resource reaches zero, the resource entry is removed.
 * NEW:
 * a NULL lock is added for each registration
 *
 * Dynamic expansion explanation:
 *
 * The resource hash table and its implementation are unchanged.  The
 * resource table, however, is now broken into segments.  The resource
 * handle now contains a segment field that indicates which segment a
 * given resource is in.  The restab variable is now an array of resource
 * segment pointers.  The size of the field in the resource handle allows
 * for 256 segments.  Resources are still allocated solely on the primary
 * and duplicated on the secondary.  If the primary exhausts its supply
 * of resources, it allocates another segment and continues on.  When the
 * primary hands a new resource back to the secondary, if resource handle
 * refers to a segment that doesn't exist on the secondary, that segment
 * is created on the secondary so that resources on the different nodes
 * still map to the same location.
 */

/* actions for reshash_find */
	
STATIC struct reshash *allocate_reshash( void );
STATIC struct reshash * find_reshash( const char * name, unsigned name_length,
    dlm_restype_t type, unsigned * hashp );

STATIC struct resource * allocate_res_seg( int segnum );
#define RESOURCE_GET(s,c)  (& restab[ s ][ c ])

/*
 * Name
 *	clm_resource_init
 * Description
 *	Init static variables in clm_resource.c
 * Arguments
 * 	None
 * Returns
 *	Nothing
 */

void
clm_resource_init( size_t seg_size, size_t prealloc, size_t table_size )
{
    unsigned   i;


    res_seg_size = seg_size;
    dlm_list_init( & reshash_free );
    res_free = NULL;

    resindex = 0;
    res_seg = 0;

    cur_seg = -1;	/* index of last allocated resource segment */

    (void) memset( restab, 0, sizeof( restab ) );

    /* Preallocate any resource segments indicated at startup.
     */

    maxresources = 0;
    clm_rescount = 0;
    res_seg_count = 0;
    for ( i = 0; i < prealloc; i++ )
    {
	if ( allocate_res_seg( i ) != NULL )
	{
	    res_seg_count++;
	}
    }

    /* allocate and initialize hashtable */
    reskeytab = MALLOC( table_size * sizeof( *reskeytab ) );
    if ( reskeytab == NULL )
    {
	bsdlog(LOG_CRIT, MSGSTR(CANT_ALLOC_RES_TAB, 
				"Unable to allocate resource tables of size %d"),
	       table_size );
	abort( "Unable to create resource key table" );
    }

    for ( i = 0 ; i < table_size ; i++ )
    {
	dlm_list_init( & reskeytab[ i ] );
    }

    hashtablesize = table_size;
}


/****************************************************************************/
/* deallocate_reshash - Return a reshash object to the free pool
 * 
 */

STATIC __inline__ void deallocate_reshash(struct reshash *rhp)
{
    dlm_list_add_head( & reshash_free, (dlm_listnode_t *) rhp );
}




/****************************************************************************/
/* resource_get - Get a pointer to a resource
 * 
 * Gets a pointer to a resource based on its segment number and its cookie
 * value.
 */

struct resource *
resource_get( unsigned segment, unsigned cookie )
{
    struct resource * r;
    
    
    r = NULL;
    if ( restab[ segment ] != NULL )
    {
	r = RESOURCE_GET( segment, cookie );
    }
    
    return( r );
}

/*
 * Name
 *	my_res
 * Description
 *	Returns true if the site/pid pair has a reference to the resource.
 * Arguments
 * 	site	- site of the client
 *	pid	- pid of the client
 *	rp	- resource pointer
 * Returns
 *	True if the site/pid has a resref for the spedified resource
 */
int
my_res(int site, int pid, struct resource *rp)
{
    struct reclock * p;

    for ( EACH_DLM_LISTNODE( & rp->lockqs[RLQ_GRANT], p ) )
    {
	if (p->lck.pid == pid && p->lck.site == site)
	    return TRUE;
    }

    return FALSE;
}

/*
 * Name
 *	nmcmp
 * Description
 *	Comparison routine for two resource names. This is used when
 *	walking a hash chain searching for a resource
 * Arguments
 * 	name1 - first name
 *	name1len - byte count of name1
 * 	name2 - second name
 *	name1len - byte count of name2
 * Returns
 *	0 if names are identical
 *	> 0 if name1 if longer or alphabetically > than name2
 *	< 0 if name1 if short or alphabetically < than name2
 */
STATIC int
nmcmp( const char * name1, int name1len, const char * name2, int name2len )
{
	int t;

	if (name1len != name2len)
		t = name1len - name2len;
	else
		t = memcmp(name1, name2, name1len);
	return t;
}

/*
 * Name
 *	allocate_res_seg
 * Description
 *	Allocate the resource table segment specified in the argument list
 * Arguments
 * 	segnum - which segment to allocate
 * Returns
 *	pointer to the new segment if successful
 *	0 if not successful
 * Global Data Affected
 *	cur_seg
 *	restab
 */
STATIC struct resource *
allocate_res_seg(int	segnum)
{
    struct resource * rp = NULL;
    const char      * err_msg;
    unsigned          i;


    DBPRINTF( (DBG_FUNC, "Entering " __FUNCTION__ " to allocate seg %d\n",
	       segnum) );


    /* Only so many segments can be allocated.
     */

    err_msg = "resource table overflow";
    if ( segnum < MAX_RES_SEGS )
    {
	/* Make sure ptr to new seg is empty.
	 */

	err_msg = "resource %d already allocated";
	if ( restab[ segnum ] == NULL )
	{
	    err_msg = "unable to expand restab";

	    rp = (struct resource *) MALLOC( res_seg_size * sizeof( *rp ) );
	    restab[ segnum ] = rp;
	    if ( restab[ segnum ] != NULL ) 
	    {
		err_msg = NULL;
	    }
	}
    }

    if ( err_msg != NULL ) 
    {
	bsdlog( LOG_ERR, err_msg, segnum );
    }
    else 
    {
	maxresources += res_seg_size;
	resindex = 0;

	DBPRINTF( (DBG_FUNC,
		   __FUNCTION__ " successful. Alloc segnum/maxres[%d/%d]\n",
		   segnum, maxresources) );

	(void) memset(rp, 0, res_seg_size * sizeof(struct resource));

	for ( i = 0 ; i < res_seg_size ; i++ )
	{
	    unsigned   j;


	    for ( j = 0 ; j < RLQ_MAX ; j++ )
	    {
		dlm_list_init( & RESOURCE_GET( segnum, i )->lockqs[ j ] );
	    }
	}
    }


    return rp;
}

/*
 * Name
 *	hashfunc
 * Description
 * 	hashing function used by resource hashing mechanism implemented in
 * 	this file.
 * Arguments
 *	name - hash key, does not need to be null terminated
 *	namelen - length of key
 *	modulus - return must be int less than this
 * Returns
 *	hash value for provided key between 0 and size of modulus
 * Global Data Affected
 *	NONE
 */
unsigned int
hashfunc( const char *name, int namelen, int modulus )
{
	register unsigned int hash = 0;
	register int i;
	register const unsigned int *iptr;
	register const char * ptr = name;

	iptr = (unsigned *)&name[0];
	for (i = 0; i < (namelen - 3); i += 4, iptr++) {
		hash += (unsigned int)*iptr;
	}
	ptr = &name[i];
	while (i < namelen) {
		hash += (unsigned char)*ptr;
		ptr++;
		if (++i < namelen) {
			hash = (hash << 8 | hash >> 26);
		} else {
			break;
		}
	}	
	DBPRINTF((DBG_RESOURCE,
			"hashfunc: %s hash=0x%x ret=%d mod=%d\n",
			printbin(name, namelen), hash, hash % modulus,
			modulus));
	return hash % modulus;
}
	


/*
 * Name
 *	lookup_resource
 * Description
 * 	Search the resource table for an existing entry matching the
 * 	provided resource creation request.
 * Arguments
 *	req - resource creation request
 * Returns
 * 	pointer to existing matching resource entry
 * 	NULL if not found
 * Global Data Affected
 *	NONE
 */
struct resource *
lookup_resource( const struct clm_regreq * req )
{
    struct reshash * rhp;
    unsigned         hash;


    rhp = find_reshash( req->name, req->namelen, req->type, & hash );
    return( (rhp == NULL) ? NULL : rhp->hash_rp );
}


/****************************************************************************/

struct resource *
find_resource(
	       /* Name of the resource.
		*/

	       const char * name,
	     
	       /* Length of the resource name.
		*/

	       unsigned name_length,

	       /* Resource type (i.e., DLM_RES_VAX or DLM_RES_UNIX).
		*/

	       dlm_restype_t type
	     )
{
    struct reshash * rhp;
    unsigned         hash;


    rhp = find_reshash( name, name_length, type, & hash );
    return( (rhp == NULL) ? NULL : rhp->hash_rp );
}




/****************************************************************************/
/* find_reshash - Find a reshash entry in the resource hash table
 *
 * Finds a reshash entry in the hash table.  The hash value of the supplied
 * name is also calculated.
 * 
 * RETURNS
 * 
 * If a matching entry is found in the resource hash table, a pointer to the
 * corresponding reshash object is returned.  Otherwise, NULL is returned.  In
 * any case, *hashp is set.
 */

STATIC struct reshash *
find_reshash(
	      /* Name of the resource.
	       */

	      const char * name,
	     
	      /* Length of the resource name.
	       */

	      unsigned name_length,

	      /* Resource type (i.e., DLM_RES_VAX or DLM_RES_UNIX).
	       */

	      dlm_restype_t type,

	      /* Location to store the hash value of the resource name.
	       */

	      unsigned * hashp
	    )
{
    struct reshash * rhp;
    unsigned         hash;


    hash = hashfunc( name, name_length, hashtablesize );
    for ( EACH_DLM_LISTNODE( & reskeytab[ hash ], rhp ) )
    {
	/* search for matching name in existing hash chain
	 */

	if ( (nmcmp( name, name_length,
		     rhp->hash_rp->name, rhp->hash_rp->namelen) == 0)
	     && (rhp->hash_rp->type == type) ) 
	{
	    break;
	}
    }


    *hashp = hash;
    return( rhp );
}


/*
 * Name: decr_delete_resource
 *
 * Description: decrement resource refcount; if == 0, delete resource
 *
 * Arguments: ->resource
 *
 * Returns: nothing
 *
 * Side effects: trashes resource if deleting it
 */

void
decr_delete_resource(struct resource *rp, int flag)
{
    CLM_ASSERT((rp->refcount > 0), "decr_delete_resource: refcount not > 0") ;
    --rp->refcount;
    check_handles(rp, flag) ;
    if ( ! rp->refcount) {
	deallocate_resource(rp) ;
    }
}

/*
 * Name
 *	get_free_res
 * Description
 *	Remove a resource structure from the resource freelist.
 *	This should only be called on the primary.
 * Arguments
 *	None
 * Returns
 *	resource pointer, NULL if list is empty.
 */
STATIC struct resource *
get_free_res(void)
{
	struct resource *rp;

	if (NULL != (rp = res_free)) {
		res_free = rp->free_next;
		rp->free_next = NULL;
	}
	return rp;
}

/*
 * Name
 *	add_freelist
 * Description
 *	Add a resource structure to the resource freelist.
 *	This should only be called on the primary.
 * Arguments
 *	rp - resource to free
 * Returns
 *	NOTHING
 */
STATIC void
add_freelist(struct resource *rp)
{
	rp->free_next = res_free;
	res_free = rp;
}

/*
 * Name
 *	count_res_freelist
 * Description
 *	Returns a count of the number of resource entries on the freelist.
 * Arguments
 *	None
 * Returns
 *	Number of entries on freelist.
 */
int
count_res_freelist(void)
{
	int		count = 0;
	struct resource	*rp = res_free;

	while (rp) {
		count++;
		rp = rp->free_next;
	}

	return count;
}

/*
 * Name
 *	allocate_resource
 * Description
 *	possibly create a new resource
 *	if a resource handle is provided, we are running on a secondary and
 *	duplicating a resource already created on the primary. At any rate,
 *	adjust resref count
 *
 * Arguments
 *	req - resource creation request
 *
 * Returns
 *	pointer to requested resource entry
 *	NULL of failure
 */
struct resource *
allocate_resource(struct clm_regreq *req)
{
	register struct resource *rp;
	union dlm_rh temp_rh;

	if (req->namelen <= 0) {
		bsdlog(LOG_ERR, MSGSTR(BAD_LEN_RES,
				"allocate_resource: bad length (%d)"),
			req->namelen);
		return NULL;
	}

	/* check freelist first */
	if (NULL != (rp = get_free_res())) {
		/*
		 * seg and cookie info for resources on the freelist
		 * have to be left in the resource record instead of
		 * trying to derive them from the pointer
		 */
		temp_rh.handle.site = OurSiteID;
		temp_rh.handle.type = req->type;
		temp_rh.handle.seg = rp->rh.handle.seg;
		temp_rh.handle.cookie = rp->rh.handle.cookie;
		DBPRINTF((DBG_DYNRES, "Res freelist. s=%d c=%d "
			"rescount=%d max=%d\n", rp->rh.handle.seg,
			rp->rh.handle.cookie, clm_rescount,
			maxresources));
	} else {
		if (clm_rescount >= maxresources) {
			/*
			 * here if all existing segments are full.
			 * the next section of code advances
			 * cur_seg to the next empty segment
			 * slot in restab and will wrap around
			 * to 0 once.
			 */
			cur_seg++;
			while (cur_seg < MAX_RES_SEGS &&
					restab[cur_seg])
				cur_seg++;

			if (cur_seg >= MAX_RES_SEGS) {
				cur_seg = 0;
				while (cur_seg < MAX_RES_SEGS &&
						restab[cur_seg])
					cur_seg++;

				if (cur_seg >= MAX_RES_SEGS) {
					bsdlog(LOG_ERR, MSGSTR(RES_SEG_FULL, 
			"%s: all resource segments full"), __FUNCTION__);
					return(NULL);
				}
			}
			if (!allocate_res_seg(cur_seg)) {
				bsdlog(LOG_ERR,
                       MSGSTR(RES_SEG_ALLOCATE_FAIL, 
                              "%s: failed to allocate res segment"),
                       __FUNCTION__);
				return(NULL);
			} else {
				if (cur_seg == res_seg_count)
					res_seg_count++;

				rp = RESOURCE_GET( cur_seg, resindex );

				DBPRINTF((DBG_DYNRES,
					"New seg:  s=%d c=%d "
					"rescount=%d max=%d\n",
					cur_seg, resindex,
					clm_rescount, maxresources));
			}

		} else {
			
			/* here if empty freelist but res available */

			/* if we're here after becoming primary,
			 * cur_seg may not be on an allocated seg, but
			 * there MUST be one, so find it.
			 */
			while (cur_seg < 0 || !restab[cur_seg] ||
					resindex >= res_seg_size) {
				cur_seg++;
				resindex = 0;
			}

			rp = RESOURCE_GET( cur_seg, resindex );

			while (resindex < res_seg_size) {
				if (rp->rh.rh <= 0) {
					/* here if this slot's empty */
					break;
				}
				/* here if current slot is used */
				resindex++;
				rp++;

				/* we may have to advance to new seg */
				if (resindex >= res_seg_size &&
					++cur_seg < res_seg_count) {
					while (!restab[cur_seg])
						cur_seg++;
					if (cur_seg >= res_seg_count) {
						bsdlog(LOG_ERR,
                               MSGSTR(RES_SEG_FIND_ERR, 
                                      "%s: failed to find res segment"),
                               __FUNCTION__);
						return(NULL);
					}
					resindex = 0;
					rp = RESOURCE_GET( cur_seg, resindex );
				}
			}
			if (resindex >= res_seg_size) {
				/* shouldn't happen - clm_rescount < maxresources */
				bsdlog(LOG_ERR,
                       MSGSTR(NO_MT_RES_SLOT, 
                              "%s: couldn't find empty resource slot"),
                       __FUNCTION__);
				return(NULL);
			}
		}
		temp_rh.handle.site = OurSiteID;
		temp_rh.handle.type = req->type;
		temp_rh.handle.seg = cur_seg;
		temp_rh.handle.cookie = resindex;
		DBPRINTF((DBG_DYNRES,
			"Incr resindex: s=%d c=%d rescnt=%d "
			"max=%d\n", cur_seg, resindex, clm_rescount,
			maxresources));
		resindex++;
	}

	if (rp->refcount > 0) {
		/* here if we already know about this resource */
		rp->refcount += req->count;
	} else {
	    struct reshash * rhp;
	    unsigned         hash;


	    rhp = find_reshash( req->name, req->namelen, req->type, & hash );
	    if ( rhp != NULL )
	    {
		bsdlog( LOG_ERR, "%s: found existing hash for new resource",
		        current->comm );
	    }
	    else
	    {
		/* here if this is the first time this slot is being used */
		rp->refcount = req->count;
		rp->rh.rh = temp_rh.rh;
		rp->type = req->type;
		CLM_ASSERT((req->master != OurSiteID), "res: Set master local");
		rp->master = req->master;

		/* resources start off as valid and zero */
		rp->rsrc_flags |= (RSRC_VALID | RSRC_ACTIVE);
		(void) memset(rp->value, 0, sizeof(rp->value));

		rp->name = (char *)CLM_MALLOC(req->namelen);
		if (!rp->name) {
			bsdlog(LOG_ERR,
                   MSGSTR(ERR_MALLOC12, 
                          "%s: malloc error getting resource name"),
                   __FUNCTION__);
			return(NULL);
		} /* end if */
		resname_bcopy(&req->name[0], &rp->name[0], req->namelen);

		rp->namelen = req->namelen;
		rp->refs = NULL;
		dlm_list_init( & rp->lockqs[RLQ_GRANT] );

		/* add new resource to end of hash chain */
		rhp = allocate_reshash();
		if (rhp == NULL) 
		{
		    bsdlog( LOG_CRIT, "%s: can't allocate reshash\n",
			    current->comm );
		    abort("find_reshash 2");
		}
		else
		{
		    /* push new entry on head of chain */
		    dlm_list_add_head( & reskeytab[hash], 
				       (dlm_listnode_t *) rhp );

		    rhp->hash_rp = rp;
		    clm_rescount++;
		}
	    }
	}

	/* how sticky do we want to be? */
	rp->clminfo.mi_stickiness = DEF_STICK;

	/*
	 * Resource Statistics:
	 *        Zero out resource statistics structure.
	 */
	rp->lastnode = 0;
	(void) memset( &(rp->stats), 0, sizeof (dlm_statistics_t) );

	return(rp);
}

/*
 * Name
 *	deallocate_resource
 * Description
 *	free a resource entry
 *	remove entry for pointer table, zero entry in resource table
 *
 * Arguments
 *	rp - resource to free
 * Returns
 *	NOTHING
 *	Abort if resource still has associated locks
 */
void
deallocate_resource(struct resource *rp)
{
    struct reshash * rhp;
    unsigned         hash;


	CLM_ASSERT((rp->rsrc_flags & RSRC_ACTIVE), "Free inactive resource");

	/*
	 * If the lock manager is not in a run state, then dir entries and
	 * resources could be in a state of change.  It's a bad idea at this
	 * time to try to remove the master copy of a resource.  Leaving an
	 * empty one behind shouldn't hurt anything.
	 */
	if (!clm_is_running() && !rp->master)
	{
		DBPRINTF((DBG_RECOV, "deallocate_resource: (%s) m=%d f=0x%x",
			printbin(rp->name, rp->namelen), rp->master,
			rp->rsrc_flags));
		return;
	}

	/* check for non-empty lock queues */
	if ( ! dlm_list_is_empty( & rp->lockqs[RLQ_GRANT] ) )
	{
		bsdlog(LOG_CRIT, MSGSTR(NON_MT_LOCKLIST,
				"%s: non-empty lock list for %s\n"), __FUNCTION__,
			RESNAME(rp));

		abort("deallocate_resource");
	}

	/*
	 * if resource was migrating, then we don't want to free it until
	 * the ACK or NACK is returned
	 */
	if (rp->rsrc_flags & (RSRC_FRZEN | RSRC_STUCK | RSRC_UPDATING)) {
#if defined(DEBUG)
		if (clm_in_reconfig())
		{
			DBPRINTF((DBG_RECOV,
			"Attempt to free migrating res (%s:%x) m=%d f=0x%x",
			printbin(rp->name, rp->namelen), (rp->rh.rh&0xff),
			rp->master, rp->rsrc_flags));
		}
#endif
		DBPRINTF((DBG_MIGR2,
			"Attempt to free migrating res (%s:%x) m=%d f=0x%x",
			printbin(rp->name, rp->namelen), (rp->rh.rh&0xff),
			rp->master, rp->rsrc_flags));
		return;
	}

	/*
	 * if this node is the master for the resource about to be removed,
	 * then the directory information for the resource needs to be
	 * deleted also.
	 */
	if (0 == rp->master) {
		DBPRINTF((DBG_RESOURCE, "Send dir update for %s\n",
			RESNAME(rp)));
		update_directory(rp, DIR_DELETE, RESRPRI, NULL );
	}
#if defined(DEBUG)
	if (clm_in_reconfig())
	{
		DBPRINTF((DBG_RECOV, "removing (%s) m=%d f=0x%x",
			printbin(rp->name, rp->namelen), rp->master,
			rp->rsrc_flags));
	}
#endif

	/*
	 * Up until this point, we haven't touched the resource yet. 
	 * Actually, we might have - and that's what this check is for.  The
	 * call that we just made to update_directory might have freed up
	 * a request that was held on the bounce queue.  So if the resource
	 * that we are about to delete -NOW- has a positive refcount, then
	 * some other request triggerred by the update call queued a lock
	 * onto it and we cannot remove it.  You'll notice too that the
	 * grant queue is no longer empty and it WAS as the start of this
	 * function.  -bg-
	 */
	if (rp->refcount > 0) {
	    DBPRINTF((DBG_RESOURCE, "HEY!!!! - (%s) is not dead afterall!!",
			printbin(rp->name, rp->namelen)));
	    return;
	}

    rhp = find_reshash( rp->name, rp->namelen, rp->type, & hash );
    if ( rhp != NULL )
    {
	dlm_list_remove( (dlm_listnode_t *) rhp );
	deallocate_reshash( rhp );
	clm_rescount--;
    }
    else 
    {
	bsdlog(LOG_CRIT, MSGSTR(CANT_FIND_DELETE, 
				"find_reshash: can't find/delete %s"),
	       RESNAME(rp));
	abort("find_reshash 3");
    }

    CLM_FREE(rp->name);

	/* fields cleared here just for defensive coding */
	rp->name = NULL;
	rp->namelen = 0;
	rp->rh.handle.site = 0;
	rp->rh.handle.type = 0;
	rp->refcount = 0;
	rp->rsrc_flags = 0;
	add_freelist(rp);
}

/*
 * Name
 *	master_register
 * Description
 *	register a resource, this is called only on the master node
 *	to create a new resource entry. if the resource already exists, 
 *	reference the existing copy, otherwise create an entry. This entry 
 *	point is used by the UNIX lock interface
 *
 * Arguments
 *	req - a client registration request. Uses clm_register union member
 * Returns
 *	0 if successful
 *	Site id of directory node if resource mastered elsewhere.
 */
int
master_register(clm_msg_t *msg, int flags)
{
	clm_trans_t *req = msg->cm_data;
	struct resource *rp;
	int		site;
	int ret ;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_CLM_STATS_T,
		0x00, 0x00, 0x00, 0x00);

	if (req->clm_register.count != 1)
	    abort("resource registered with count != 1") ;

	req->clm_register.name = req->clm_register.unix_name;
	DBPRINTF((DBG_RESOURCE, "master_register: name %s type %ld\n",
			printbin(req->clm_register.name, 
				req->clm_register.namelen), 
				req->clm_register.type));

	if (NULL != (rp = lookup_resource(&req->clm_register))) {
		DBPRINTF((1, "master_register: found rh 0x%x refcount %d\n",
				rp->rh.rh, rp->refcount));

		if (rp->master != 0) {
			/*
			 * here if we are not master of the resource, so
			 * determine the site id of the directory node so we
			 * can send a redirect
			 */
			site = clm_direct(req->clm_register.name,
					req->clm_register.namelen);
			if (!site) {
				bsdlog(LOG_CRIT, MSGSTR(NULL_RETURNED, 
					"clm_lock:clm_direct() returned NULL"));
				return -1;
			}
			
			return site;
		}

		/*
		 * At this point, we know that we are going to send a
		 * response for this message and not redirect to the directory
		 * server, so copy the message and queue the response.
		 */
		msg = cccp_msg_duplicate( msg );
		req = msg->cm_data;
		msg->cm_node = req->clm_register.site;
		req->clm_direction = PTI_RESPONSE;
		clm_queue_respmsg(msg);

		/*
		 * At this point, we know that we are the resource master that
		 * is going to handle this request.  Set the master field in the
		 * route information of the request so when the secondary lock
		 * manager reads the response, it'll know where it was mastered.
		 */
		req->clm_rt_master = OurSiteID;

		rp->refcount += req->clm_register.count;
		ret = add_null_lock(req, rp,
					req->clm_register.pid,
					req->clm_register.site) ;
		req->clm_handle.rh = rp->rh.rh;
		req->clm_status = ret ? ret : DLM_NORMAL ;
		return 0 ;

	} else if (req->clm_register.site != OurSiteID) {
		/*
		 * Here if we couldn't find the resource and the request is
		 * from a different node.  A master copy of a resource should
		 * not be *created* on a node other than the one that made the
		 * request to cause the creation.  So if we're here, then
		 * either the mastery of the resource moved since the
		 * directory told us where it was, or the master copy of the
		 * resource went away since we got our directory information.
		 * So we need to check with the directory again.
		 */
		site = clm_direct(req->clm_register.name,
				req->clm_register.namelen);
		if (!site) {
			bsdlog(LOG_CRIT, MSGSTR(NULL_RETURNED, 
				"clm_lock:clm_direct() returned NULL"));
			return -1;
		}
		
		return site;
	}

	/*
	 * At this point, we know that we are going to send a
	 * response for this message and not redirect to the directory
	 * server, so copy the message and queue the response.
	 */
	msg = cccp_msg_duplicate( msg );
	req = msg->cm_data;
	msg->cm_node = req->clm_register.site;
	req->clm_direction = PTI_RESPONSE;
	clm_queue_respmsg(msg);

	/*
	 * At this point, we know that we are the resource master that
	 * is going to handle this request.  Set the master field in the
	 * route information of the request so when the secondary lock
	 * manager reads the response, it'll know where it was mastered.
	 */
	req->clm_rt_master = OurSiteID;
	req->clm_register.master = 0;

	if (NULL != (rp = allocate_resource(&req->clm_register))) {
	 	DBPRINTF((DBG_RESOURCE, "master_register: allocated rh 0x%x\n", 
			rp->rh));
		ret = add_null_lock(req, rp,
					req->clm_register.pid,
					req->clm_register.site) ;
	/* do not update refcount because allocate_resource did ?!?!? */
		req->clm_handle.rh = rp->rh.rh;
		req->clm_status = ret ? ret : DLM_NORMAL ;
		return 0;
	}

	req->clm_handle.rh = 0L;
	req->clm_status = DLM_DENIED;
	return 0;
}

/* 
 * Name
 *	begin_register
 * Description
 *	send a register request to the primary
 * Arguments
 *	req - a client registration request. Uses clm_register union member
 * Returns
 *	0 for success
 * 	-1 for error
 */
int
begin_register(clm_trans_t *req)
{
	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_REMOTE_REGISTER,
		0x00, 0x00, 0x00, 0x00);

	/* Add this to remove need for clinfo in API */
	req->clm_register.site = OurSiteID;
	if (req->clm_register.count != 1)
	    abort("resource registered with count != 1") ;

	if (pti_call_reg(req) != 0) {
		bsdlog(LOG_ERR, MSGSTR(PTI_CALL_FAILED,
				"begin_register: failed call to PTI"));
		return -1;
	}
	return 0;
}

/*
 * Name
 *	cont_register
 * Description
 *	continue a register request that was sent to the master
 *	put the name in the local table and send a reply to the client
 * Arguments
 *	req - original request sent to the primary 
 *	      uses the clm_register union member
 *	resp - response coming back from the primary
 *	      uses the clm_register union member
 * Returns
 * 	NOTHING
 */
void
cont_register(clm_trans_t *req, clm_trans_t *resp)
{
	struct resource *rp;
	int ret = 0 ;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_CONT_REMOTE_REGISTER,
		resp->clm_status, 0x00, 0x00, 0x00);

	if (resp->clm_handle.rh == 0L) {
		bsdlog(LOG_ERR, MSGSTR(TOO_MANY_RESOURCES,
				"cont_remote_register: too many resources"));
		resp->clm_status = DLM_DENIED;
		goto done;
	}
	req->clm_register.rh.rh = resp->clm_handle.rh;
	req->clm_register.name = req->clm_register.unix_name;
	if (NULL != (rp = lookup_resource(&req->clm_register))) {
		DBPRINTF((1, "cont_register: found rh 0x%x refcount %d\n",
				rp->rh.rh, rp->refcount));

		if (rp->master == 0) {
			/*
			 * here if we are the master of the resource, we've
			 * already updated the local resource, so
			 * I guess we're done.
			 */
			
			req->clm_status = DLM_NORMAL;
			goto done;
		}

		/* update resource's notion of who is master */
		rp->master = resp->clm_rt_master;

		rp->refcount += req->clm_register.count;
		ret = add_null_lock(req, rp, req->clm_register.pid,
						req->clm_register.site) ;
		rp->mst_rh.rh = resp->clm_handle.rh;
		resp->clm_handle.rh = rp->rh.rh;
		resp->clm_status = ret ? ret : DLM_NORMAL ;
		goto done;

	}

	if (NULL != (rp = allocate_resource(&req->clm_register))) {
	 	DBPRINTF((DBG_RESOURCE, "master_register: allocated rh 0x%x\n", 
			rp->rh));

		/* update resource's notion of who is master */
		rp->master = resp->clm_rt_master;

		ret = add_null_lock(req, rp, req->clm_register.pid,
						req->clm_register.site) ;
	/* do not update refcount because allocate_resource did?!?! */
		rp->mst_rh.rh = resp->clm_handle.rh;
		resp->clm_status = ret ? ret : DLM_NORMAL ;
		resp->clm_handle.rh = rp->rh.rh;
		goto done;
	}

	bsdlog(LOG_ERR, MSGSTR(RESOURCE_FAILURE,
			"cont_remote_register: resource failure"));
	req->clm_handle.rh = 0L;
	req->clm_status = DLM_DENIED;

done:
	clm_reply(req, resp);
	return;
}

/*
 * add_null_lock
 *
 * adds a placeholder lock with null mode to "register" a resource
 * DOES NOT increment reference count - must be done by caller
 * (or implicitly by allocate_resource)
 * adds to END of grant queue so bounds checking in clm_unixlock work
 *
 * takes transaction causing lock, resource being registered,
 *	pid to blame, site where pid lives
 *
 * returns 0 for OK, 1 for failure
 */
static int
add_null_lock(struct transaction *req, struct resource *rp,
							int pid, int site)
{
    struct reclock *a ;

    if ( ! (a = get_le(DLM_RES_UNIX)))
    {
	bsdlog(LOG_ERR, MSGSTR(CANT_ALLOCATE_LK_ENTRY, 
	    "add_null_lock: can't allocate lock entry"));
	return 1 ;
    }
    else
    {
	if (site && site != OurSiteID)
	    a->lck.remlockid = (site << 28) | pid ;	/* FAKE remlockid */
	a->route = req->clm_route ;
	a->lck.site = site ;
	a->lck.pid = pid ;
	a->lck.mode = LKM_NLMODE ;
	a->lck.state |= LOCK_GRANTED ;
	a->lck.rh.rh = rp->rh.rh ;

	lck_set_seq(rp, a) ;
	add_queue(&rp->lockqs[RLQ_GRANT], a, RL_RQ_OFS) ;
	check_handles(rp, 0)  ;
    }
    return 0 ;
}

/****************************************************************************/
/* resource_table_iterator_create - Create an iterator for the resource table
 * 
 * Creates an iterator for the global resource table.
 */

struct iterator *
resource_table_iterator_create( void )
{
    struct restab_iterator * rt_iter;


    rt_iter = CLM_MALLOC( sizeof( struct restab_iterator ) );
    if ( rt_iter != NULL )
    {
	rt_iter->base.methods = & restab_iterator_methods;
	rt_iter->next_seg = 0;
	rt_iter->next_cookie = 0;
    }

    return( (iterator_t *) rt_iter );
}


static bool_t
resource_table_iterator_next( iterator_t * iter, void ** resource )
{
    struct restab_iterator * rt_iter;


    *resource = NULL;

    rt_iter = (struct restab_iterator *) iter;
    if ( rt_iter->next_seg < res_seg_count )
    {
	CLM_ASSERT( rt_iter->next_cookie < res_seg_size,
		    "Invalid iterator cookie!" );

	*resource = resource_get( rt_iter->next_seg,
				  rt_iter->next_cookie );
	rt_iter->next_cookie++;
	if ( rt_iter->next_cookie >= res_seg_size )
	{
	    rt_iter->next_cookie = 0;
	    rt_iter->next_seg++;
	}
    }

    return( (*resource == NULL) );
}



static void
resource_table_iterator_destroy( iterator_t * iter )
{
    if ( iter != NULL )
    {
	CLM_FREE( iter );
    }
}





/*
 * Name
 *	allocate_reshash
 * Description
 *	Return a reshash structure
 * Arguments
 *	NONE
 * Returns
 *	A pointer to a reshash structure
 *	NULL if error
 * Global Data Affected
 *	NONE
 */
STATIC struct reshash *
allocate_reshash(void)
{
    struct reshash * hep;
    unsigned         i;


    hep = (struct reshash *) dlm_list_rem_head( & reshash_free );
    if ( hep == NULL )
    {
	/* Here if freelist is empty - link in new block of hash entry
	 * structures and hand back first
	 */

	hep = (struct reshash *) MALLOC( 1024 * sizeof( struct reshash ) );
	if ( hep != NULL )
	{
	    for ( i = 0 ; i < 1024 ; i++ )
	    {
		dlm_list_add_tail( & reshash_free, (dlm_listnode_t *) hep );
		hep++;
	    }

	    hep = (struct reshash *) dlm_list_rem_head( & reshash_free );
	}
    }

    return( hep );
}
