/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_rldb.c,v 1.8 2001/11/07 23:56:19 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		clm_dump_rldb
 *		rl_alloc
 *		rl_create
 *		rl_decrement_dir_count
 *		rl_delete
 *		rl_enum
 *		rl_expand_freelist
 *		rl_find
 *		rl_free
 *		rl_init
 *		rl_lookup
 *		rl_modify
 *		rl_reclaim_entry
 *		rl_update
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/signal.h>
#include <linux/types.h>
#include <linux/errno.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "cltrc.h"
#include "clm_data.h"			/* external data definitions */
#include "dlm_kernel.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */

#define	RLDB_ALLOCSIZE	1000

static kmem_cache_t * rl_cache = NULL;
static dlm_list_t   * rldbtab = NULL;	/* RLDB hash table pointer */
static unsigned		rldbtabsize = 0;	/* size of RLDB hash table */
static unsigned		rl_dir_count = 0;	/* # of directory entries */


/*
 * Name
 *	clm_rldb_init
 * Description
 *	Init static variables in RLDB
 * Arguments
 *	None
 * Returns
 *	Nothing
 * Global Data Affected
 * 	None
 */
void
clm_rldb_init(void)
{
    rl_cache = kmem_cache_create( "RL Info", sizeof( rl_info_t ),
				  0, 0, NULL, NULL );
    rldbtab = NULL;
    rldbtabsize = 0;
    rl_dir_count = 0;
}


/*
 * Name
 *	clm_rldb_cleanup
 * Description
 *	Init static variables in RLDB
 * Arguments
 *	None
 * Returns
 *	Nothing
 * Global Data Affected
 * 	None
 */
void
clm_rldb_cleanup(void)
{
    rldbtabsize = 0;
    rl_dir_count = 0;

    kmem_cache_destroy( rl_cache );
    rl_cache = NULL;

    if ( rldbtab != NULL )
	CLM_FREE( rldbtab );

    rldbtab = NULL;
}


/*
 * Name
 *	rl_reclaim_entry
 * Description
 *	Search the current hash table for an entry that isn't a directory
 *	entry and hasn't been referenced recently.  Remove it from the hash
 *	table and return it.
 * Arguments
 *	None
 * Returns
 *	Address of the new rl_info structure.
 *	NULL if there's an error.
 * Global Data Affected
 * 	rldbtab
 */
static rl_info_t *
rl_reclaim_entry(void)
{
    static int   hand_index = 0;
    rl_info_t  * rl_entry;


    DBPRINTF((DBG_DIRECT|DBG_FUNC, "Entering rl_reclaim_entry"));
    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_RLDB_RECLAIM, 0x00, 0x00,
	    0x00, 0x00);

    while (1) 
    {
	/* bump it one more so we start fresh */
	hand_index = (hand_index + 1) % rldbtabsize;

	/* increment clock hand to next non-NULL hash pointer */
	while ( dlm_list_is_empty( & rldbtab[hand_index] ) )
	{
	    hand_index = (hand_index + 1) % rldbtabsize;
	}

	/* walk list, look for clear bit, clear ones that we check */
	
	for ( EACH_DLM_LISTNODE( & rldbtab[ hand_index ], rl_entry ) )
	{
	    if ( ((rl_entry->rl_flags & RLINFO_TOUCHED) == 0)
		 && ((rl_entry->rl_flags & RLINFO_DIRECTORY) == 0) )
	    {
		break;
	    }
	    
	    rl_entry->rl_flags &= ~RLINFO_TOUCHED;
	}

	if ( rl_entry != NULL ) 
	{
	    dlm_list_remove( (dlm_listnode_t *) rl_entry );

	    DBPRINTF((DBG_DIRECT, "rc: Messing s=%d %s 0x%x\n",
		      hand_index, printbin(rl_entry->rl_name,
					   rl_entry->rl_namelen),
		      rl_entry ));

	    CLM_FREE(rl_entry->rl_name);
	    rl_entry->rl_name = NULL;
	    break;
	}
    }

    DBPRINTF((DBG_DIRECT|DBG_FUNC, "Exit from rl_reclaim_entry"));
    TRCHKGT(CLLOCKD_HKID | T_EXIT_FUNC, T_RLDB_RECLAIM, 0x00, 0x00,
	    0x00, 0x00);

    return( rl_entry );
}

/*
 * Name
 *	rl_alloc
 * Description
 *	Return a new rl_info structure from the free pool.  If the freelist
 *	is empty, allocate another block of structures and extend the pool.
 * Arguments
 *	None
 * Returns
 *	Address of the new rl_info structure.
 *	NULL if there's an allocation error.
 */
static rl_info_t *
rl_alloc( const void * name, size_t namelen, unsigned type, unsigned node,
	  unsigned flags )
{
    rl_info_t * rl_entry;
    
    
    rl_entry = kmem_cache_alloc( rl_cache, 0 );
    if ( rl_entry == NULL )
    {
	rl_entry = rl_reclaim_entry();
    }

    if ( rl_entry != NULL )
    {
	rl_entry->rl_type = type;
	rl_entry->rl_node = node;
	rl_entry->rl_flags = flags | RLINFO_TOUCHED;
	rl_entry->rl_namelen = namelen;
	
	rl_entry->rl_name = CLM_MALLOC( namelen );
	if ( rl_entry->rl_name != NULL )
	{
	    (void) memcpy( rl_entry->rl_name, name, namelen );
	}
	else
	{
	    kmem_cache_free( rl_cache, rl_entry );
	    rl_entry = NULL;
	}
    }

    return( rl_entry );
}

/*
 * Name
 *	rl_free
 * Description
 *	Put the specified rl_info structure back into the free pool.
 * Arguments
 *	rl_entry - Entry to be freed
 * Returns
 *	Nothing
 */
static void
rl_free(rl_info_t	*rl_entry)
{
    TRCHKGT(CLLOCKD_HKID, T_RLDB_FREE, rl_entry, 0x00, 0x00, 0x00);
#ifdef DEBUG
    (void) memset( rl_entry, 0, sizeof( rl_info_t ) );
#endif

    kmem_cache_free( rl_cache, rl_entry );
    return;
}

/*
 * Name
 *	rl_find
 * Description
 *	Given a name and length, return a pointer to a matching entry in the
 *	table if it exists.
 * Arguments
 *	name		Name to find
 *	namelen		Length of name to find
 *	type		Type of resource to find
 *	r_slot		Return slot index if needed
 * Returns
 *	Address of the table entry.
 *	NULL if a match is not found.
 * Global Data Affected
 * 	None
 */
rl_info_t *
rl_find( const void * name, int namelen, int type, unsigned *r_slot)
{
    unsigned	slot = hashfunc(name, namelen, rldbtabsize);
    rl_info_t	*rl_entry;

    if (r_slot)
	*r_slot = slot;

    DBPRINTF((DBG_DIRECT, "rl_find: %s type=%d slot=%d:0x%08lx\n",
	      printbin(name, namelen), type, slot, & rldbtab[ slot ] ));
    for ( EACH_DLM_LISTNODE( & rldbtab[ slot ], rl_entry ) )
    {
	if ( (rl_entry->rl_namelen == namelen)
	     && (rl_entry->rl_type == type)
	     && (0 == memcmp(rl_entry->rl_name, name, namelen)) )
	{
	    break;
	}
    }


    return rl_entry;
}

/*
 * Name
 *	rl_init
 * Description
 *	Allocate and initialize the memory for rldbtab
 * Arguments
 *	Size of rldbtab
 * Returns
 *	DLM_NORMAL	Entry created, successful return
 *	DLM_SYSERR	Malloc error
 * Global Data Affected
 * 	rldbtab
 */
dlm_stats_t
rl_init(unsigned	size)
{
    unsigned   i;


    rldbtabsize = size;

    rldbtab = (dlm_list_t *)CLM_MALLOC(rldbtabsize * sizeof( dlm_list_t ));
    if (!rldbtab) {
	bsdlog(LOG_CRIT, MSGSTR(NO_MEM_FOR_RLDB, 
				"rl_init: unable to allocate memory for RLDB"));
	return DLM_SYSERR;
    }

    for ( i = 0 ; i < size ; i++ )
    {
	dlm_list_init( & rldbtab[ i ] );
    }

    DBPRINTF((DBG_DIRECT|DBG_FUNC, "rl_init: init RLDB tabsize = %d",
	      rldbtabsize));
    TRCHKGT(CLLOCKD_HKID, T_RLDB_INIT, rldbtabsize, 0x00, 0x00, 0x00);

    return DLM_NORMAL;
}

/*
 * Name
 *	rl_create
 * Description
 *	Given a populated rl_info structure, create an entry in the RLDB
 *	with the same values.
 * Arguments
 *	rl_new - Entry to be created
 * Returns
 *	DLM_NORMAL		Entry created, successful return
 *	DLM_IVBUFLEN		Invalid name length (check not enabled)
 *	DLM_BADRESOURCE		Entry already exists
 * Global Data Affected
 * 	rldbtab
 */
dlm_stats_t
rl_create( const rl_info_t * rl_new )
{
    rl_info_t    * rl_entry;
    unsigned	   slot;
    dlm_stats_t   status;


    CLM_ASSERT((rl_new), "Null entry passed to rl_create");
    CLM_ASSERT((rl_new->rl_namelen > 0 && rl_new->rl_namelen <= 255),
	       "Adding rl_entry w/ invalid resource name");

    status = DLM_BADRESOURCE;
    rl_entry = rl_find( rl_new->rl_name, rl_new->rl_namelen, rl_new->rl_type,
		        & slot );
    if ( rl_entry == NULL ) 
    {
	status = DLM_SYSERR;
	/* allocate and fill in new rldb entry */
	rl_entry = rl_alloc( rl_new->rl_name, rl_new->rl_namelen,
			    rl_new->rl_type, rl_new->rl_node,
			    rl_new->rl_flags );
	if ( rl_entry != NULL )
	{
	    /* track number of directory entries in rldb */
	    if (rl_entry->rl_flags & RLINFO_DIRECTORY) 
	    {
		rl_dir_count++;
		DBPRINTF((DBG_DIRECT, "create 0x%x %s in slt=%d",
			  rl_entry, printbin(rl_entry->rl_name, 
					     rl_entry->rl_namelen),
			  slot));
	    }

	    /* add it to the beginning of the hash list */
	    dlm_list_add_head( & rldbtab[ slot ], 
			       (dlm_listnode_t *) rl_entry );

	    TRCHKGT(CLLOCKD_HKID, T_RLDB_CREATE, rl_entry, rl_entry->rl_flags,
		    rl_entry->rl_node, 0x00);
	    DBPRINTF((DBG_DIRECT,"rl_create: %d:%s flags=0x%x node=%d\n",
		      rl_entry->rl_namelen, printbin(rl_entry->rl_name,
						     rl_entry->rl_namelen), rl_entry->rl_flags,
		      rl_entry->rl_node));

	    status = DLM_NORMAL;
	}
    }

    return( status );
}

/*
 * Name
 *	rl_delete
 * Description
 *	Given an rl_info structure containing a name and length,
 *	remove the matching entry from the RLDB.
 *
 *	Find the matching entry in the hash table.  Remove it from the
 *	hash table and free the entry back into the free pool.
 * Arguments
 *	rl_old - Entry to be deleted
 * Returns
 *	DLM_NORMAL		Entry created, successful return
 *	DLM_IVBUFLEN		Invalid name length
 *	DLM_BADRESOURCE		Could not find resource entry
 * Global Data Affected
 * 	rldbtab
 */
dlm_stats_t
rl_delete(rl_info_t	*rl_old)
{
    rl_info_t	*rl_entry;
    unsigned	slot;

    CLM_ASSERT((rl_old), "Null entry passed to rl_delete");
    CLM_ASSERT((rl_old->rl_namelen > 0 && rl_old->rl_namelen <= 255),
	       "Adding rl_entry w/ invalid resource name");

    rl_entry = rl_find( rl_old->rl_name, rl_old->rl_namelen, rl_old->rl_type,
		        & slot );
    if (!rl_entry) {
	return DLM_BADRESOURCE;
    }
	
    dlm_list_remove( (dlm_listnode_t *) rl_entry );

    /* track number of directory entries in rldb */
    if (rl_entry->rl_flags & RLINFO_DIRECTORY) {
	rl_dir_count--;
	DBPRINTF((DBG_DIRECT,
		  "delete %s from slt=%d nde=%d 0x%x",
		  printbin(rl_entry->rl_name, rl_entry->rl_namelen),
		  slot, rl_entry->rl_node, rl_entry));
    }

    CLM_FREE(rl_entry->rl_name);
    rl_free(rl_entry);

    return DLM_NORMAL;
}

/*
 * Name
 *	rl_lookup
 * Description
 *	Given an rl_info structure containing a name and length,
 *	return a copy of the matching entry from the RLDB.
 * Arguments
 *	rl_search - Entry to be found
 * Returns
 *	DLM_NORMAL		Entry created, successful return
 *	DLM_IVBUFLEN		Invalid name length
 *	DLM_BADRESOURCE		Could not find resource entry
 * Global Data Affected
 * 	rldbtab
 */
dlm_stats_t
rl_lookup(rl_info_t	*rl_search)
{
	rl_info_t	*rl_entry;
	
	CLM_ASSERT((rl_search), "Null entry passed to rl_lookup");
	CLM_ASSERT((rl_search->rl_namelen > 0 && rl_search->rl_namelen <= 255),
			"Adding rl_entry w/ invalid resource name");

	rl_entry = rl_find(rl_search->rl_name, rl_search->rl_namelen,
		rl_search->rl_type, 0);
	if (!rl_entry) {
		return DLM_BADRESOURCE;
	}

#ifndef __linux__
found:
#endif /* !__linux__ */
	rl_entry->rl_flags |= RLINFO_TOUCHED;
	rl_search->rl_node = rl_entry->rl_node;
	rl_search->rl_flags = rl_entry->rl_flags;

	return DLM_NORMAL;
}

/*
 * Name
 *	rl_modify
 * Description
 *	Given an rl_info structure containing a name and length,
 *	find the entry in the RLDB with the same values and update it
 *	with the new information.
 * Arguments
 *	rl_mod - Entry to be modified
 * Returns
 *	DLM_NORMAL		Entry created, successful return
 *	DLM_IVBUFLEN		Invalid name length
 *	DLM_BADRESOURCE		Could not find resource entry
 * Global Data Affected
 * 	rldbtab
 */
dlm_stats_t
rl_modify(rl_info_t	*rl_mod)
{
	rl_info_t	*rl_entry;
	unsigned	slot;
	
	CLM_ASSERT((rl_mod), "Null entry passed to rl_modify");
	CLM_ASSERT((rl_mod->rl_namelen > 0 && rl_mod->rl_namelen <= 255),
			"Adding rl_entry w/ invalid resource name");

	rl_entry = rl_find(rl_mod->rl_name, rl_mod->rl_namelen,
		rl_mod->rl_type, &slot);
	if (!rl_entry) {
		return DLM_BADRESOURCE;
	}

	/* track number of directory entries in rldb */
	if ((rl_entry->rl_flags & RLINFO_DIRECTORY) &&
			(rl_mod->rl_flags & RLINFO_DIRECTORY) == 0) {
		rl_dir_count--;
	} else if ((rl_mod->rl_flags & RLINFO_DIRECTORY) &&
			(rl_entry->rl_flags & RLINFO_DIRECTORY) == 0) {
		rl_dir_count++;
	}

	CLM_ASSERT((rl_mod->rl_node), "Node = 0 in rl_modify");
	rl_entry->rl_node = rl_mod->rl_node;
	rl_entry->rl_flags = rl_mod->rl_flags;

	return DLM_NORMAL;
}

/*
 * Name
 *	rl_update
 * Description
 *	Given an rl_info structure containing a name and length,
 *	find the entry in the RLDB with the same values and update it
 *	with the new information.  Create an entry with the provided
 *	information if one doesn't exist.
 *	NOTE:  rl_update should not be used to modify DIRECTORY info.
 *	It should only be used to update local cache values.  If the
 *	entry is found to be a directory entry and the update value would
 *	change the node information, then the function will return an error.
 * Arguments
 *	rl_mod - Entry to be updated
 * Returns
 *	DLM_NORMAL		Entry created, successful return
 *	DLM_IVBUFLEN		Invalid name length
 *	DLM_BADRESOURCE		Could not find resource entry
 * Global Data Affected
 * 	rldbtab
 * 
 * NOTES
 * 
 * If an existing entry is updated, the rl_flags field is not modified.  
 * Except, of course, to set the RLINFO_TOUCHED flag.
 */
dlm_stats_t
rl_update( rl_info_t * rl_update )
{
    rl_info_t * rl_entry;
    unsigned    slot;


    CLM_ASSERT((rl_update), "Null entry passed to rl_update");
    CLM_ASSERT((rl_update->rl_namelen > 0 && rl_update->rl_namelen <= 255),
	       "Adding rl_entry w/ invalid resource name");

    rl_entry = rl_find(rl_update->rl_name, rl_update->rl_namelen,
		       rl_update->rl_type, &slot);

    if ( rl_entry == NULL )
    {
	CLM_ASSERT((clm_direct(rl_update->rl_name, rl_update->rl_namelen)
		    != OurSiteID),
		   "Creating cache entry on dir node");

	/* allocate and fill in new rldb entry */

	CLM_ASSERT((rl_update->rl_node), "Node = 0 in rl_update");

	rl_entry = rl_alloc( rl_update->rl_name, rl_update->rl_namelen,
			    rl_update->rl_type, rl_update->rl_node,
			    rl_update->rl_flags );
	if (!rl_entry)
	    return DLM_SYSERR;

	dlm_list_add_head( & rldbtab[ slot ], (dlm_listnode_t *) rl_entry );

	DBPRINTF((DBG_DIRECT,
		  "upd: Messing with slot=%d %s 0x%x\n", slot,
		  printbin(rl_entry->rl_name, rl_entry->rl_namelen),
		  rl_entry));

    }
    else
    {
	if ((rl_entry->rl_flags & RLINFO_DIRECTORY) &&
	    rl_entry->rl_node != rl_update->rl_node) {
	    return DLM_BADARGS;
	}

	CLM_ASSERT((rl_update->rl_node), "Node = 0 in rl_update");
	rl_entry->rl_node = rl_update->rl_node;

	/* rl_update should only update node information */
	rl_entry->rl_flags |= RLINFO_TOUCHED;
    }

    return DLM_NORMAL;
}




/****************************************************************************/
/* dlm_rldb_print - Print RLDB table to a buffer
 *
 * Walk through the RLDB hash table and print the values of each entry to the
 * provided buffer.
 *
 * RETURNS
 *
 * Total number of characters copied to the buffer.
 */

unsigned
dlm_rldb_print( char * buf, unsigned limit )
{
    rl_info_t * rl_entry;
    unsigned    slot;
    unsigned    bytes;


    bytes = snprintf( buf, limit,
		      "    Slot |             Name |    Type | Flags | Node\n" );

    for ( slot = 0 ; (slot < rldbtabsize) && (bytes < limit) ; slot++ )
    {
	for ( EACH_DLM_LISTNODE_COND( & rldbtab[ slot ], rl_entry,
				      bytes < limit ) )
	{
	    bytes += snprintf( & buf[ bytes ], limit - bytes,
			       "%8u | %16s | %7s |  %04X | %4u\n", slot,
			       printbin( rl_entry->rl_name,
					 rl_entry->rl_namelen ),
			       dlm_get_lock_type_string( rl_entry->rl_type ),
			       rl_entry->rl_flags, rl_entry->rl_node );
	}
    }

    return( bytes );
}

/*
 * rl_decrement_dir_count
 *
 * decrement global counter of # directories held
 * for reconfig
 */
void
rl_decrement_dir_count(int i)
{
	rl_dir_count -= i;
}

/*
 * rl_enum
 *
 * for a subset or all entries in the directory
 *	call func() with arg of entry found
 *
 * arg type:
 *	RLENUM_ALLNODES:   all entries
 *	RLENUM_SPECNODES:  only entries pointing to 'node'
 */
void
rl_enum(int (*rl_enum_func)(rl_info_t *), unsigned type, int node)
{
	int		slot;
	rl_info_t	*rl_entry;
	rl_info_t	*next;

    switch (type) 
    {
	case RLENUM_ALLNODES:
	{
	    for (slot = 0; slot < rldbtabsize; slot++) {
		for ( rl_entry = (rl_info_t *) dlm_list_get_head( & rldbtab[slot] )
		     ; rl_entry != NULL
		     ; rl_entry = next )
		{
		    next = (rl_info_t *) dlm_list_succ( (dlm_listnode_t *) rl_entry );
		    (*rl_enum_func)(rl_entry);
		}
	    }

	    break;
	}

	case RLENUM_SPECNODES:
	{
	    for (slot = 0; slot < rldbtabsize; slot++) 
	    {
		for ( rl_entry = (rl_info_t *) dlm_list_get_head( & rldbtab[slot] )
		     ; rl_entry != NULL
		     ; rl_entry = next )
		{
		    next = (rl_info_t *) dlm_list_succ( (dlm_listnode_t *) rl_entry );

		    if (rl_entry->rl_node == node)
			(*rl_enum_func)(rl_entry);
		}
	    }
	    break;
	}

	default:
	bsdlog(LOG_ERR, MSGSTR(BAD_ARGS, "Bad arg passed to rl_enum\n"));
    }
}

static void *save_func ;	/* paranoia */
static unsigned ctd_slot ;		/* local context to resume search */

void
rl_enum_init()
{
    ctd_slot = 0 ;
    save_func = NULL ;
}

/*
 * rl_enum_counted
 *
 * see rl_enum() for base functionality
 *
 * This routine will only process 'count' slots in the RLDB hash table
 *	before returning, saving its place in a static variable.
 * Calling this routine again will resume processing where it left off.
 * Call rl_enum_init() to restart scan from beginning of RLDB.
 * Stores via 'done': 0 if scan not complete, 1 after processing last entry.
 */
void
rl_enum_counted(int (*rl_enum_func)(rl_info_t *), unsigned type, int node,
							int count, int *done)
{
    unsigned    slot;
    unsigned    index;
    rl_info_t * rl_entry;
    rl_info_t * next;

    if ( save_func && save_func != (void *) rl_enum_func)
	abort("rl_enum_counted: mixed calls") ;

    switch (type)
    {
    case RLENUM_ALLNODES:
	node = -1 ;
	break ;

    case RLENUM_SPECNODES:
	break;

    default:
	bsdlog(LOG_ERR, MSGSTR(BAD_ARGS, "Bad arg passed to rl_enum\n")) ;
	return ;
    }


    bsdlog( LOG_DEBUG, "[%s:%s] ctd_slot = %d, rldbtabsize = %u\n",
	    current->comm, __FUNCTION__,
	    ctd_slot, rldbtabsize );

    for ( slot = ctd_slot, index = 0
	  ; (slot < rldbtabsize) && (index < count) 
	  ; slot++ )
    {
	for ( rl_entry = (rl_info_t *) dlm_list_get_head( & rldbtab[slot] )
	      ; rl_entry != NULL
	      ; rl_entry = next )
	{
	    next = (rl_info_t *) dlm_list_succ( (dlm_listnode_t *) rl_entry );

	    if (node < 0 || rl_entry->rl_node == node)
	    {
		(*rl_enum_func)(rl_entry);
		index++;
	    }
	}
    }
    ctd_slot = slot;
    *done = (slot >= rldbtabsize);	/* set done flag if no more */

    bsdlog( LOG_DEBUG, "[%s:%s] %sdone\n", current->comm, __FUNCTION__,
            (*done) ? "" : "not " );
}
