/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_scn.c,v 1.6 2001/10/17 18:44:30 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		begin_scn_op
 *		cont_scn_op
 *		master_scn_op
 *		scn_add
 *		scn_cmp
 *		scn_mask
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/types.h>
#include <linux/errno.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "clm_data.h"			/* external data definitions */
#include "dlm_kernel.h"
#include "clm_locksw.h"
#include "cltrc.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/*
 * Name
 *	scn_mask
 * Description
 *	Given an SCN ptr and a bit length, mask out unwanted bits
 * Arguments
 *	scnptr - scn value to be masked
 *	bit_len - bit length
 * Returns
 *	NOTHING
 * Global Data Affected
 * 	NONE
 */
static void
scn_mask(scn_t *scnptr, int bit_len)
{
	DBPRINTF((DBG_SCN, "Mask <%d><%d><%d><%d> to len = %d\n",
		scnptr->wrap3, scnptr->wrap2, scnptr->wrap1, scnptr->base,
		bit_len));
	if (bit_len < 96) {
		scnptr->wrap3 = 0;
		if (bit_len < 64) {
			scnptr->wrap2 = 0;
			if (bit_len < 32) {
				scnptr->wrap1 = 0;
				scnptr->base &=
					((unsigned)(-1) >> (32 - bit_len));
			} else {
				scnptr->wrap1 &=
					((unsigned)(-1) >> (64 - bit_len));
			}
		} else {
			scnptr->wrap2 &= ((unsigned)(-1) >> (96 - bit_len));
		}
	} else {
		scnptr->wrap3 &= ((unsigned)(-1) >> (128 - bit_len));
	}
	DBPRINTF((DBG_SCN, "Result = <%d><%d><%d><%d> \n",
		scnptr->wrap3, scnptr->wrap2, scnptr->wrap1, scnptr->base));
}

/*
 * Name
 *	scn_add
 * Description
 * 	Take two scn pointers.  Add the first to the second.
 * Arguments
 *	inc_scn - scn value to be added to destination scn
 *	dst_scn - destination scn
 *	len - bit length
 * Returns
 *	NOTHING
 * Global Data Affected
 * 	NONE
 */
void
scn_add(scn_t *inc_scn, scn_t *dst_scn, int len)
{
	unsigned int	*dst, *inc, result, carry = 0;
	int	i, bit;

	DBPRINTF((DBG_SCN, "Add <%d><%d><%d><%d> to <%d><%d><%d><%d>\n",
		inc_scn->wrap3, inc_scn->wrap2, inc_scn->wrap1, inc_scn->base,
		dst_scn->wrap3, dst_scn->wrap2, dst_scn->wrap1, dst_scn->base));
	for (i = 0, bit = 0, inc = (unsigned int *)inc_scn,
					dst = (unsigned int *)dst_scn;
					(i < 4) && (bit < len);
					i++, bit += 32, inc++, dst++) {
		result = *inc + *dst + carry;
		if ((*inc || carry) && result <= *dst)
			carry = 1;
		else
			carry = 0;
		*dst = result;
		DBPRINTF((DBG_SCN,"result is %d  carry is %d\n",
					result, carry));
	}
	DBPRINTF((DBG_SCN, "and you get <%d><%d><%d><%d> for len = %d\n",
		dst_scn->wrap3, dst_scn->wrap2, dst_scn->wrap1, dst_scn->base,
		len));
}

/*
 * Name
 *	scn_cmp
 * Description
 * 	Compare two SCNs
 * Arguments
 *	scn_one - first SCN
 *	scn_two - second SCN
 * Returns
 *	< 0 if scn_one is less than scn_two
 *	0 if they are equal
 *	> 0 if scn_one is greater than scn_two
 * Global Data Affected
 * 	NONE
 */
int
scn_cmp(scn_t *scn_one, scn_t *scn_two)
{
	unsigned int	*one = (unsigned int *)scn_one;
	unsigned int	*two = (unsigned int *)scn_two;
	int		i, result = 0;

	DBPRINTF((DBG_SCN, "Compare <%d><%d><%d><%d> to <%d><%d><%d><%d>\n",
		scn_one->wrap3, scn_one->wrap2, scn_one->wrap1, scn_one->base,
		scn_two->wrap3, scn_two->wrap2, scn_two->wrap1, scn_two->base));
	for (i = 0; i < 4; i++, one++, two++) {
		if (*one < *two) {
			result -= (1 << i);
		} else if (*one > *two) {
			result += (1 << i);
		}
	}

	DBPRINTF((DBG_SCN,"Return %d\n", result));
	return result;
}

/*
 * Name
 *	begin_scn_op
 * Description
 * 	Send an scn request to the primary
 * Arguments
 *	req - the scn request
 * Returns
 *	return status from pti_call 
 * Global Data Affected
 * 	NONE
 */
int
begin_scn_op(clm_trans_t *req)
{
	struct reclock *a;
	struct resource *rp;
	int	stat;

	DBPRINTF((DBG_SCN|DBG_FUNC, "enter begin_scn_op\n"));

	/* Add this to remove need for clinfo in API */
	req->clm_scninfo.site = OurSiteID;

	/* check for an invalid lock id */
	if (req->clm_scninfo.lockid == 0) {
		DBPRINTF((DBG_SCN, "begin_scn_op: got bad id 0x%x", 
			req->clm_scninfo.lockid));
		req->clm_status = DLM_IVLOCKID;
		goto done;
	} else if (!(a = find_reclock(DLM_RES_VMS, req->clm_scninfo.lockid)) ||
			!mylock(a, req)) {
		DBPRINTF((DBG_CLMLOCK,
			"begin_scn_op: can't find reclock for 0x%x", 
			req->clm_scninfo.lockid));
		req->clm_status = DLM_IVLOCKID;	
		goto done;
	}

	/* get the resource pointer for the lock being referenced */
	if (a->lck.rh.rh == 0) {
		bsdlog(LOG_CRIT, MSGSTR(NULL_RES_HANDLE, 
		    "begin_scn_op: NULL resource handle for 0x%x"),
			 req->clm_scninfo.lockid);
		req->clm_status = DLM_IVLOCKID;	
		goto done;
	}
	if (!(rp = resptr(a->lck.rh))) {
		bsdlog(LOG_ERR, MSGSTR(CANT_GET_HANDLE, 
			"begin_scn_op: can't find handle for 0x%x"),
			req->clm_scninfo.lockid);
		req->clm_status = DLM_IVLOCKID;	
		goto done;
	}
	req->clm_scninfo.nameinfo.namelen = rp->namelen;
	resname_bcopy(rp->name, &req->clm_scninfo.nameinfo.name[0], rp->namelen);

	stat = pti_call(req, rp->master ? rp->master : OurSiteID);

	DBPRINTF((DBG_SCN|DBG_FUNC, "leave begin_scn_op\n"));
	return stat;
done:

	sendudp(req);
	
	return 0;
}

/*
 * Name
 *      cont_scn_op
 * Description
 *      Continue an scn request coming back from the primary
 *      If the status was DLM_NORMAL, call local_scn_op
 * Arguments
 *      req - the request that was sent to the primary
 *      resp - the response from the primary
 * Returns
 *      NOTHING
 * Global Data Affected
 *      NONE
 */
void
cont_scn_op(clm_trans_t *req, clm_trans_t *resp)
{
	struct reclock *a;
	struct resource *rp;
	struct scninfo *scn = &resp->clm_scninfo;

	DBPRINTF((DBG_SCN|DBG_FUNC, "enter cont_scn_op\n"));

	if (resp->clm_status != DLM_NORMAL)
		goto done;
	
	if (!(a = find_reclock(DLM_RES_VMS, scn->lockid)) ||
			!mylock(a, resp)) {
		DBPRINTF((DBG_CLMLOCK,
			"local_scn_op: can't find reclock for 0x%x", 
			resp->clm_scninfo.lockid));
		resp->clm_status = DLM_IVLOCKID;	
		goto done;
	}
	else if (a->lck.rh.rh == 0) {
			bsdlog(LOG_CRIT, MSGSTR(NULL_RES_HANDLE2,
			    "local_scn_op: NULL resource handle for 0x%x"),
				 resp->clm_scninfo.lockid);
			resp->clm_status = DLM_IVRESHANDLE;	
			abort("cont_scn_op");
			goto done;
	}
	if (!(rp = resptr(a->lck.rh))) {
		bsdlog(LOG_ERR, MSGSTR(CANT_GET_HANDLE2,
				"local_scn_op: can't find handle for 0x%x"),
			resp->clm_scninfo.lockid);
		resp->clm_status = DLM_IVRESHANDLE;	
		goto done;
	}

	/* update local value with new copy if we're not the master */
    if (0 != rp->master)
	(void) memcpy( rp->value, scn->lkvalue, sizeof( rp->value ) );

done:

	clm_reply(req, resp);

	DBPRINTF((DBG_SCN|DBG_FUNC, "leave cont_scn_op\n"));
}

/*
 * Name
 *	master_scn_op
 * Description
 * 	perform an SCN operation on the LVB of the resource referred to
 *	by the provided lockid
 * Arguments
 *	req - the SCN request
 * Returns
 *	DLM_IVLOCKID - bad lock id specified
 *	DLM_IVRESHANDLE - bad resource handle in lock
 *	DLM_BLOCKED - tried to perform SCN on resource w/ non-NULL locks
 *	DLM_VALNOTVALID - okay, but value block is invalid
 *	DLM_NORMAL - okay
 * 		Output Structure: lockinfo
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
master_scn_op(clm_msg_t *msg)
{
	clm_trans_t *req = msg->cm_data;
	struct reclock *a;
	struct resource *rp;
	register uint id;
	struct scninfo *scn = &req->clm_scninfo;
	static scn_t inc_scn = { 1, 0, 0, 0 };
	int	site;

	DBPRINTF((DBG_SCN|DBG_FUNC, "enter local_scn_op\n"));
	/*
	 * if we are processing a request from another node, map lockid 
	 */
	if (req->clm_scninfo.site != OurSiteID)
		id = mapid(req->clm_scninfo.lockid, req->clm_scninfo.site) ;
	else
		id = req->clm_scninfo.lockid;

	if (id == 0) {
		/* resource must've migrated - redirect */
		site = clm_direct(req->clm_scninfo.nameinfo.name,
				req->clm_scninfo.nameinfo.namelen);

		if (site)
			return site;
		else
			return -1;
	}

	else if (!(a = find_reclock(DLM_RES_VMS, id)) ||
			!mylock(a, req)) {
		DBPRINTF((DBG_CLMLOCK,
			"local_scn_op: can't find reclock for 0x%x", 
			req->clm_scninfo.lockid));
		req->clm_status = DLM_IVLOCKID;	
		goto error;
	}
	else if (a->lck.rh.rh == 0) {
			bsdlog(LOG_CRIT, MSGSTR(NULL_RES_HANDLE2,
			    "local_scn_op: NULL resource handle for 0x%x"),
				 id);
			req->clm_status = DLM_IVRESHANDLE;	
			abort("local_scn_op");
			goto error;
	}
	if (!(rp = resptr(a->lck.rh))) {
		bsdlog(LOG_ERR, MSGSTR(CANT_GET_HANDLE2,
				"local_scn_op: can't find handle for 0x%x"),
			req->clm_scninfo.lockid);
		req->clm_status = DLM_IVRESHANDLE;	
		goto error;
	}

	/* check to make sure resource is mastered here, otherwise redirect */
	if (0 != rp->master) {
		site = clm_direct(rp->name, rp->namelen);

		if (site)
			return site;
		else
			return -1;
	}

	/*
	 * request is not going to be re-directed.  call cccp_msg_duplicate() to insure
	 * that we have a writeable copy of the message, reset the req pointer
	 * and enqueue the response back to the secondary.
	 */
	msg = cccp_msg_duplicate( msg );
	req = msg->cm_data;
	scn = &req->clm_scninfo;
	msg->cm_node = scn->site;
	req->clm_direction = PTI_RESPONSE;
	clm_queue_respmsg(msg);

	/* this operation must not continue if non-NULL lock are granted */
    for ( EACH_DLM_LISTNODE( & rp->lockqs[RLQ_GRANT], a ) )
    {
	if (a->lck.mode != LKM_NLMODE) {
	    req->clm_status = DLM_BLOCKED;
	    goto done;
	}
    }

	/* check for invalid value block */
	if (!(rp->rsrc_flags & RSRC_VALID) && (scn->scn_op != SCN_SET)) {
		req->clm_status = DLM_VALNOTVALID;
		goto done;
	}

	/* switch on the operation type */
	switch (req->clm_scninfo.scn_op) {

	case SCN_INC:
		scn_add(&inc_scn, (scn_t *)rp->value, scn->bit_len);
		break;
	case SCN_ADD:
		scn_add((scn_t *)&scn->lkvalue[0], (scn_t *)rp->value,
				scn->bit_len);
		break;
	case SCN_ADJ:
		scn_mask((scn_t *)scn->lkvalue, scn->bit_len);
		if (scn_cmp((scn_t *)scn->lkvalue, (scn_t *)rp->value) > 0) {
		    (void) memcpy( rp->value, scn->lkvalue, 
				   sizeof( rp->value ) );
		}
		break;
	case SCN_SET:
		(void) memcpy( rp->value, scn->lkvalue, sizeof( rp->value ) );
		rp->rsrc_flags |= RSRC_VALID;
		break;

	}

	/* trim stored lvb to length specified in call */
	scn_mask((scn_t *)rp->value, scn->bit_len);

	/* current value is almost always returned */
	(void) memcpy( scn->lkvalue, rp->value, sizeof( rp->value ) );

	DBPRINTF((DBG_SCN|DBG_FUNC, "leave local_scn_op\n"));
	req->clm_status = DLM_NORMAL;	

#ifdef DBPRINT
	clm_dumpres(rp);
#endif

done:
	return 0;

error:
	msg = cccp_msg_duplicate( msg );
	req = msg->cm_data;
	req->clm_direction = PTI_RESPONSE;
	clm_queue_respmsg(msg);

	return 0;

}

