/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_timeout.c,v 1.6 2001/09/20 23:06:32 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		clm_timeout_init
 *		send_timeout
 *		set_lock_timer
 *		timeout_dequeue
 *		timer_tick
 *		unset_lock_timer
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/signal.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <asm/param.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "clm_data.h"			/* external data definitions */
#include "dlm_kernel.h"
#include "clm_locksw.h"
#include "cltrc.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/*
 * queue node for lock timeout queue
 */
typedef struct tq_node {
    dlm_listnode_t   link;
    struct reclock * tq_rl;
    int   tq_time;
} tq_node_t;

/* ============================================================ */

/* Subtract timestructs */
#define	timersub(tvp, svp, rvp)					\
	{								\
		if((svp).tv_nsec > (tvp).tv_nsec)  {			\
			(rvp).tv_sec = ((tvp).tv_sec - 1) - (svp).tv_sec;\
			(rvp).tv_nsec = ((int)(((uint)((tvp).tv_nsec) + NS_PER_SEC) - (uint)((svp).tv_nsec)));\
		}							\
		else  {							\
			(rvp).tv_sec = (tvp).tv_sec - (svp).tv_sec;	\
			(rvp).tv_nsec = (tvp).tv_nsec - (svp).tv_nsec;	\
		}							\
									\
	}


/* These two values control the timeout queue.  If timeout_queue is NOT empty,
 * then ticks_on_top is the number of ticks remaining until the top element
 * should fire.  This will ALWAYS be larger than zero (unless a new tick is
 * being processed in timer_tick).  If timeout_queue is empty, then the value
 * of ticks_on_top is undefined.
 */

static int ticks_on_top;
static DECLARE_DLM_LIST( timeout_queue );

/* 1 "tick" is 1/100th of a sec */

static int time_set;		/* true when lasttime is valid */
#ifdef __linux__
static unsigned long lasttime; /* used to calc delta time */
#else
static struct timestruc_t lasttime; /* used to calc delta time */
#endif /* __linux__ */


static void timeout_dequeue( struct reclock * a );
static void send_timeout( struct reclock * treq );

/*
 * Name
 *	clm_timeout_init
 * Description
 *	This function inits all statics in this file.
 * Arguments
 *	None
 * Returns
 *	NOTHING
 * Global Data Affected
 *	
 */
void
clm_timeout_init(void)
{
	time_set = FALSE;

	/* init and pin this as its ref'd often */
	ticks_on_top = 0 ;
#ifndef __linux__
	pin(ticks_on_top, sizeof ticks_on_top) ;
#endif /* !__linux__ */
}

/*
 * Name
 *	timer_tick
 * Description
 *	This function should be called from the main loop 100 times per sec.
 *	Presently, it is only called at least once per second.  It calculates
 *	the amount of time that has passed since it was last called and
 *	subtracts that amount from ticks_on_top.  If ticks_on_top
 *	goes to 0, then the reclock at the head of the timeout queue (and all
 *	others behind it with delta time 0) has expired.
 * Arguments
 *	NONE
 * Returns
 *	NOTHING
 * Global Data Affected
 *	ticks_on_top, tq - timeout request queue.
 */
void
timer_tick(void)
{
  tq_node_t		* curr;	/* current queue entry */
  tq_node_t		* next;	/* next queue entry */
  struct reclock	*a;		/* current reclock */
#ifdef __linux__
  unsigned long scurtime;	/* working time struct */
  unsigned long	delta;		/* delta time struct */
#else
  struct timestruc_t	scurtime;	/* working time struct */
  struct timestruc_t	delta;		/* delta time struct */
#endif /* __linux__ */

    /* pardon me, but do you have the time? */
    scurtime = jiffies;

    /* init lasttime on 1st pass after start or resumption of timer */
    if (!time_set) {
	time_set = TRUE;
	goto done;
    }

    delta = scurtime - lasttime;

    /* Do not attempt to do anything if the delta < 0.1 secs.  Deltas that
     * are too small round to 0 and ticks_on_top might not be decremented.
     * This would starve the timer and nothing would time out.
     */

    if ((HZ / 10) > delta) 
    {
	return;	/* return without updating lasttime */
    }


#ifdef DBPRINT
    if ( ticks_on_top <= delta )
    {
	DBPRINTF((DBG_TIMEOUT, "Delta (jiffies/secs)%lu:%lu  Top = %d",
		  delta, delta / HZ, ticks_on_top));
    }
#endif

    /* If the timeout queue is empty or the next timeout has not yet expired,
     * then there is nothing to be done.
     */

    ticks_on_top -= delta;
    while ( ! dlm_list_is_empty( & timeout_queue ) && (ticks_on_top <= 0) )
    {
	/* Here if at least one timeout has expired - dequeue entries until we
	 * find the first one with time remaining
	 */

	for ( curr = (tq_node_t *) dlm_list_get_head( & timeout_queue )
	     ; (curr != NULL) && (curr->tq_time <= 0) 
	     ; curr = next )
	{
	    /* Dequeue from time running queue 
	     */

	    dlm_list_remove( (dlm_listnode_t *) curr );
	    next = (tq_node_t *) dlm_list_get_head( & timeout_queue );


	    a = curr->tq_rl;
	    CLM_ASSERT((a && (a->lck.state & (LOCK_WAITING | LOCK_CONVERTING))
			&& (a->lck.state < 0x800)),
		       "Bad lock timing out");

	    /* Becuse timeout is a cancel and cancels clear timeouts, clearing
	     * the tq_node field in the reclock prevents the timeout from
	     * trying to recurse back on itself.
	     */

	    a->tq_node = NULL;

	    /* time out the request */
	    send_timeout(a);

	    CLM_FREE( curr );
	}


	/* Now, either the queue is empty or the head hasn't expired yet.
	 * Update ticks_on_top to the time value on the head entry.
	 */

	if ( curr != NULL )
	{
	    ticks_on_top += curr->tq_time;	/* ticks_on_top might be < 0, so '+=' */
	    curr->tq_time = 0;		/* node on top has 0 time */
	}
    }


    CLM_ASSERT((dlm_list_is_empty( & timeout_queue ) || ticks_on_top > 0),
	       "Bad timeout queue values");

done:
    lasttime = scurtime;	/* save curtime in lasttime for next delta */
    return;
}

/*
 * Name
 *	unset_lock_timer
 * Description
 *	This function is called to cancel a timeout for a specified lock
 * Arguments
 *	a - pointer to the reclock structure
 * Returns
 *	NOTHING
 * Global Data Affected
 *	tq, ticks_on_top
 */
void
unset_lock_timer(struct reclock *a)
{
    DBPRINTF((DBG_FUNC|DBG_TIMEOUT, "Clearing timeout for 0x%x\n",
	      a->lck.lockid));

    timeout_dequeue(a);
}

/*
 * Name
 *	set_lock_timer
 * Description
 *	This function is called to set a timeout for a specified lock
 * Arguments
 *	a - pointer to the reclock structure
 * Returns
 *	NOTHING
 * Global Data Affected
 *	tq, ticks_on_top
 */
void
set_lock_timer(struct reclock *a)
{
    tq_node_t * tq_node;	/* new queue entry pointer */
    tq_node_t * curr;
    tq_node_t * pred;
    unsigned long this_to; /* Convert incoming seconds to jiffies */
#ifdef DBPRINT
    struct resource *rp = resptr(a->lck.rh);

    if (rp != NULL) {
	DBPRINTF((DBG_TIMEOUT, "Setting timeout for 0x%x (%s)to %d.\n",
		  a->lck.lockid, printbin(rp->name, rp->namelen),
		  a->lck.li_to));
    } else {
	DBPRINTF((DBG_TIMEOUT, "Bad resptr"));
    }
#endif

    /* How many jiffies before THIS request times out.  The incoming value is
     * number of seconds relative to right now.
     */

    this_to = a->lck.li_to * HZ;
  

    /* Allocate a timeout queue entry 
     */

    tq_node = (tq_node_t *) CLM_MALLOC( sizeof( tq_node_t ) );
    if (NULL == tq_node) 
    {
	abort("Unable to allocate timeout struct in set_lock_timer");
    }

    /* timeout queue node and reclock point at each other */
    tq_node->tq_rl = a;
    a->tq_node = tq_node;

    CLM_ASSERT((dlm_list_is_empty( & timeout_queue ) || ticks_on_top > 0),
	       "Bad timeout queue values");

    /* Does this timeout belong at the top of the list?
     */
    if ( dlm_list_is_empty( & timeout_queue )
	 || (this_to < ticks_on_top) )
    {
	tq_node->tq_time = 0;
	ticks_on_top = this_to;

	dlm_list_add_head( & timeout_queue, (dlm_listnode_t *) tq_node );
    }
    else 
    {
	/* here if we belong somewhere past the top of the list */

	/* it's not on the head, so sub 1st time val from the new one */
	tq_node->tq_time = this_to - ticks_on_top;

	/*
	 * Walk thru the list.  If next delta < ours, sub it from ours and
	 * advance.
	 */

	pred = NULL;
	for ( EACH_DLM_LISTNODE_COND( & timeout_queue,
				      curr,
				      (curr->tq_time <= tq_node->tq_time) ) )
	{
	    tq_node->tq_time -= curr->tq_time;
	    pred = curr;
	}


	/* The only way that pred could be NULL is if the new timeout element
	 * belongs at the head of the queue.  However, if the new element
	 * belongs at the head, then we should NOT be in this execution path
	 * at all!
	 */

	CLM_ASSERT( pred != NULL, "Timeout queue corrupt" );


	/* We are either at the end of the list, or the new delta falls
	 * between tq_prev and tq_cur.  Insert here.
	 */

	dlm_list_insert( & timeout_queue, (dlm_listnode_t *) tq_node,
			 (dlm_listnode_t *) pred );


	/* If not at the end of the queue, subtract the new delta from curr
	 * time.
	 */

	if ( curr != NULL ) 
	{
	    curr->tq_time -= tq_node->tq_time;
	}
    }
}

/*
 * Name
 *	timeout_dequeue
 * Description
 *	This function is called to check for and remove a reclock that
 *	might be on the timeout queue.  If removing from the middle of the
 *	queue, update the time value of the next timeout.  If removing the
 *	head entry, update ticks_on_top.  Free the timeout queue struct and
 *	zero the pointer to it in the reclock.
 * Arguments
 *	a - pointer to the reclock structure
 * Returns
 *	NOTHING
 * Global Data Affected
 *	tq, ticks_on_top
 */
static void
timeout_dequeue(struct reclock *a)
{
    tq_node_t * node;
    tq_node_t * next;


    node = a->tq_node;
    if ( node != NULL ) 
    {
	/* Is this the head of the timeout queue?
	 */

	next = (tq_node_t *) dlm_list_succ( (dlm_listnode_t *) node );

	if ( node == (tq_node_t *) dlm_list_get_head( & timeout_queue ) ) 
	{
	    dlm_list_remove( (dlm_listnode_t *) node );

	    if ( next != NULL )
	    {
		ticks_on_top += next->tq_time;
		next->tq_time = 0;
	    }
	    else 
	    {
		ticks_on_top = 0;
	    }
	}
	else
	{
	    dlm_list_remove( (dlm_listnode_t *) node );


	    /* If there's a timeout after us, add our delta to his.
	     */

	    if ( next != NULL )
	    {
		next->tq_time += node->tq_time;
	    }
	}


	CLM_FREE( node );
	a->tq_node = NULL;
    }


    return;
}

/*
 * Name
 *	send_timeout
 * Description
 *	This function takes a pointer a reclock that needs to be timed out.
 *	If this machine is the master, then the timeout notification is sent 
 *	to the owner of the timed-out lock.  Otherwise, a cancel/timeout
 *	unlock request is created and sent to the master.
 * Arguments
 *	treq - pointer to the reclock structure
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
static void
send_timeout( struct reclock * treq )
{
  clm_trans_t *req;
  struct resource *rp = NULL;

  rp = resptr(treq->lck.rh);
  if (rp == NULL) {
    bsdlog(LOG_ERR, MSGSTR(BAD_RESPTR2, 
                           "send_timeouts: bad resptr, rh=0x%lx\n"),
           treq->lck.rh.rh);
    goto done;
  }

  DBPRINTF((DBG_TIMEOUT, "Lockid 0x%x is timing out\n",
		    treq->lck.lockid));
  /* if we are primary, notification is sent from here */
  if (rp && rp->master == 0) {
    /*
     * if lock is converting, take it off the
     * convert queue and put it back on the grant
     * queue at its previously granted mode.
     */
    if (treq->lck.state & LOCK_CONVERTING) {
      remove_queue(&rp->lockqs[RLQ_CONV], treq, RL_RQ_OFS);
      lck_set_seq(rp, treq);
      treq->lck.state &= ~LOCK_CONVERTING;
      add_queue(&rp->lockqs[RLQ_GRANT], treq, RL_RQ_OFS);
      treq->lck.state |= LOCK_GRANTED;
	if (treq->lck.site == OurSiteID)
	{
	    (void) memset( & treq->route.cbasth, 0,
			   sizeof( treq->route.cbasth ) );
	}

      /*
       * blocked converts can block other requests.  Call
       * convert_locks() to check for any grantable converts
       */
      convert_locks(rp, NULL);
    } else if (treq->lck.state & LOCK_WAITING) {
      /* !!!!!! where does it go ? */
      remove_queue(&rp->lockqs[RLQ_WAIT], treq, RL_RQ_OFS);
      rp->refcount-- ;
    } else {
      DBPRINTF((DBG_CLMLOCK,
				"Tried to time out a granted request.\n"));
      goto done;
    }

    /*  If it's on the deadlock queue, remove it. */
    delete_timeout_reclock(treq);

    /* send timeout AST */
    notify(treq, RESP_AST, DLM_TIMEOUT);

    if (treq->lck.state & LOCK_WAITING) {
      /* deallocate reclock */
      free_le(DLM_RES_VMS, treq);
    }
    check_handles(rp, 0);
  } else {
    /*
     * we are not primary cancel/timeout unlock
     * request must be sent to primary to cause lock
     * timeout.
     */
    /* state may have changed since timeout was queued */
    if ((treq->lck.state & (LOCK_WAITING |
                            LOCK_CONVERTING)) == 0){
      DBPRINTF((DBG_TIMEOUT,
                "Tried to time out a granted"
                " request.\n"));
      goto done;
    }
    /* allocate transaction */
    req = clm_alloc_trans();
    if (!req) {
	bsdlog(LOG_ERR, 
	       MSGSTR(CANT_ALLOCATE_TRANS, 
		      "send_timeouts: could not allocate transaction") );
	goto done;
    }

    /*
     * build a transaction that looks like a
     * cancel request, but give it a LKM_TIMEOUT
     * flag so it will be seen as a timeout.
     */
    req->clm_locktype = DLM_RES_VMS;
    req->clm_respflags = 0;
    req->clm_lockinfo = treq->lck;
    req->clm_route = treq->route;
    req->clm_callback = cont_unlock_clm;
    req->clm_lockinfo.lk_flags = LKM_CANCEL | LKM_TIMEOUT;
    req->clm_type = DLM_UNLOCK;
    req->clm_pid = treq->lck.pid;
    req->clm_status = DLM_FORWARD;

    /*
     * if we don't have a local resource copy, then
     * we've got a problem.
     */
    if (NULL == rp) {
      abort("No resource in send_timeout()");
    } else {
	req->clm_lockreq.nameinfo.namelen = rp->namelen;
	(void) memcpy( & req->clm_lockreq.nameinfo.name[0], rp->name, 
		       rp->namelen );
    }
    DBPRINTF((DBG_CLMLOCK,
              "Sending timeout to primary for 0x%x\n",
              treq->lck.lockid));
    pti_call(req, rp ? rp->master : 0);
  }
 done:
  return;
}

