/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_tune.c,v 1.9 2001/10/29 21:53:09 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		begin_glob_tune_op
 *		begin_tune_op
 *		cont_tune_op
 *		master_tune_op
 *		remote_glob_tune_op
 *		send_glob_params
 *		send_glob_params_complete
 *              begin_stats_op
 *              cont_stats_op
 *              master_stats_op
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/types.h>
#include <linux/errno.h>
#include <asm/uaccess.h>

#include "dlm.h"
#include "ast.h"
#include "clm_tune.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "clm_data.h"			/* external data definitions */
#include "dlm_kernel.h"
#include "clm_locksw.h"
#include "dlm_version.h"
#include "cltrc.h"

#include <iterator.h>

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/*
 * Name
 *	begin_tune_op
 * Description
 * 	Determine the master node for the resource whose migration parameters
 *	are to be tuned and forward the request to that node.
 * Arguments
 *	req - the tune request
 * Returns
 *	return status from pti_call 
 * Global Data Affected
 * 	NONE
 */
int
begin_tune_op(clm_trans_t *req)
{
    struct resource *rp;
    int	stat;


    DBPRINTF((DBG_STATS|DBG_FUNC, "enter %s n=%s t=%d\n", __FUNCTION__,
	      printbin(req->clm_rmpinfo.rmp_resname,
		       req->clm_rmpinfo.rmp_namelen),
	      req->clm_rmpinfo.rmp_type));

    /* Try to find a pointer to the specified resource
     */

    rp = find_resource( req->clm_rmpinfo.rmp_resname,
		        req->clm_rmpinfo.rmp_namelen,
		        req->clm_rmpinfo.rmp_type );

    if ( NULL != rp ) 
    {
	DBPRINTF((DBG_MIGR, "begin_tune_op: found rh 0x%x\n", rp->rh.rh));

	req->clm_rmpinfo.rmp_rp = rp;
	stat = pti_call(req, rp->master ? rp->master : OurSiteID) ;
    }
    else 
    {
	req->clm_status = DLM_BADRESOURCE;
	sendudp(req);
		
	stat = 0;
    }


    DBPRINTF((DBG_MIGR|DBG_FUNC, "leave begin_tune_op\n"));
    return stat;
}


/*
 * Name
 *      cont_tune_op
 * Description
 *      Continue resource tune request that has come back from the master
 *      node.
 * Arguments
 *      req - the request that was sent to the master
 *      resp - the response from the master
 * Returns
 *      NOTHING
 * Global Data Affected
 *      NONE
 */
void
cont_tune_op(clm_trans_t *req, clm_trans_t *resp)
{
    struct resource *rp = req->clm_rmpinfo.rmp_rp;


    DBPRINTF((DBG_MIGR|DBG_FUNC, "enter cont_scn_op\n"));

    if (resp->clm_status == DLM_NORMAL)
    {
	/* update local value with new copy if we're not the master */
	if (0 != rp->master) 
	{
	    rp->clminfo.mi_stickiness =
		resp->clm_rmpinfo.rmp_params.cr_stickiness;
	}
    }

    clm_reply(req, resp);
    DBPRINTF((DBG_MIGR|DBG_FUNC, "leave cont_scn_op\n"));
}

/*
 * Name
 *	master_tune_op
 * Description
 * 	 Make sure master node of resource, other redirect.  If
 *       DLM_TUNE_SET is set then set stickness else get stickness.
 * Arguments
 *	msg -
 * Returns
 *	0    - OK
 *      site - redirect request to site
 *      -1   - error 
 * Global Data Affected
 *	NONE
 */
int
master_tune_op(clm_msg_t *msg)
{
    clm_trans_t *req = msg->cm_data;
    struct resource *rp;
    int	site;


    DBPRINTF((DBG_MIGR|DBG_FUNC, "enter master_tune_op\n"));

    if ( req->clm_rmpinfo.rmp_site != OurSiteID )
    {
	/* Try to find a pointer to the specified resource
	 */

	rp = find_resource( req->clm_rmpinfo.rmp_resname,
			    req->clm_rmpinfo.rmp_namelen,
			    req->clm_rmpinfo.rmp_type );
    }
    else 
    {
	rp = req->clm_rmpinfo.rmp_rp;
    }

    /* check to make sure resource is mastered here, otherwise redirect */
    if (NULL == rp || 0 != rp->master) 
    {
	site = clm_direct(&req->clm_rmpinfo.rmp_resname[0],
			  req->clm_rmpinfo.rmp_namelen);
	return( (site == 0) ? -1 : site );
    }

    DBPRINTF((DBG_MIGR, "master_tune_op: found rh 0x%x\n", rp->rh.rh));

    /* Request is not going to be re-directed.  call cccp_msg_duplicate() to
     * insure that we have a writeable copy of the message, reset the req
     * pointer and enqueue the response back to the secondary.
     */

    msg = cccp_msg_duplicate( msg );
    req = msg->cm_data;

    msg->cm_node = req->clm_rmpinfo.rmp_site;
    req->clm_direction = PTI_RESPONSE;
    clm_queue_respmsg(msg);

    if (req->clm_rmpinfo.rmp_flags & DLM_TUNE_SET) 
    {
	if (req->clm_rmpinfo.rmp_params.cr_valid & DLM_RES_STICKINESS) 
	{
	    rp->clminfo.mi_stickiness =
		req->clm_rmpinfo.rmp_params.cr_stickiness;
	    DBPRINTF((DBG_TUNE, "set stick to %d\n",
		      rp->clminfo.mi_stickiness));
	}
    }
    else 
    {
	/* return current values */
	DBPRINTF((DBG_TUNE, "return stick of %d\n", 
		  rp->clminfo.mi_stickiness));
	req->clm_rmpinfo.rmp_params.cr_stickiness = rp->clminfo.mi_stickiness;
    }

    DBPRINTF((DBG_MIGR|DBG_FUNC, "leave master_tune_op\n"));
    req->clm_status = DLM_NORMAL;	

#ifdef DBPRINT
    clm_dumpres(rp);
#endif
    return 0;
}


/*
 * Name
 *	send_glob_params_complete
 * Description
 * 	Callback function for global parameter transmit.
 * Arguments
 *	msg - message structure pointer
 *	stat - send status
 * Returns
 *	Nothing
 * Global Data Affected
 * 	NONE
 */
void
send_glob_params_complete(clm_msg_t *msg, int stat)
{
	if (stat)
		bsdlog(LOG_ERR, MSGSTR(ERROR_IN_PROC,
				"Error detected in send_glob_params_complete"));

	/* the data area was allocated with the message structure */
	if (msg) {
		if (msg->cm_data) {
			CLM_FREE(msg->cm_data);
		}

		cccp_msg_destroy( msg );
	}
	
	return;
}


/*
 * Name
 *	send_glob_params
 * Description
 * 	Handle a set/get global migration tuning parameters request.  If it's
 *	a set request, set the parameters locally and forward the request
 *	to all other active lock managers
 * Arguments
 *	nodeid - node to send global migration parameter update to
 *	params - pointer to dlm_globparams structure containing new values
 * Returns
 *	Nothing
 * Global Data Affected
 * 	NONE
 */
static void
send_glob_params(int nodeid, dlm_globparams_t *params)
{
    clm_msg_t		*msg;
    dlm_globparams_t	*data;
    int			stat;

    data = (dlm_globparams_t *)CLM_MALLOC(sizeof(dlm_globparams_t));
    if ( data == NULL )
    {
	bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC14,
				"%s: malloc failed"), __FUNCTION__);
	return;
    }

    msg = cccp_msg_create( PORT_MSTR, 1, nodeid, DLM_GMIGR_OP,
			   sizeof( dlm_globparams_t ),
			   send_glob_params_complete, data, NULL );
    if ( msg == NULL )
    {
	bsdlog(LOG_CRIT, MSGSTR(ERR_MALLOC14,
				"%s: malloc failed"), __FUNCTION__);
	CLM_FREE( data );
	return;
    }


    *data = *params;
    stat = cccp_send(msg);
    if (CPERR_OK != stat) {
	bsdlog(LOG_CRIT, MSGSTR(CCCP_SEND_RETURNED,
				"%s: cccp_send returned %d\n"), __FUNCTION__, stat);
    }

    return;
}


/*
 * Name
 *	begin_glob_tune_op
 * Description
 * 	Handle a set/get global migration tuning parameters request.  If it's
 *	a set request, set the parameters locally and forward the request
 *	to all other active lock managers
 * Arguments
 *	req - the tune request
 * Returns
 *	return status from pti_call 
 * Global Data Affected
 * 	NONE
 */
int
begin_glob_tune_op(clm_trans_t *req)
{
    dlm_globparams_t * parm_ptr = & req->clm_gmpinfo.gmp_params;

    DBPRINTF((DBG_MIGR|DBG_FUNC, "enter begin_glob_tune_op decay=%s "
	      "event=%d valid=0x%x flag=0x%x\n",
	      kfl_ftos(parm_ptr->cg_decay_rate), parm_ptr->cg_recalc_time,
	      parm_ptr->cg_valid, req->clm_gmpinfo.gmp_flags));

    if (req->clm_gmpinfo.gmp_flags & DLM_TUNE_SET) 
    {
#ifndef DEBUG
	if (parm_ptr->cg_recalc_time < 5)
	    parm_ptr->cg_recalc_time = 5;
#endif
		
	if (clm_low_version() >= DLM_VERSION) 
	{
	    iterator_t * iter;
	    unsigned     node;


	    /*
	     * only permit migration to be set if all nodes are at
	     * current version or later.
	     */
	    iter = dlm_nodeid_iterator_create();
	    while ( ! iterator_next( iter, & node ) )
	    {
		if ( clm_node_active( node ) )
		{
		    send_glob_params( node, parm_ptr );
		}
	    }

	    iterator_destroy( iter );
	}
    }
    else 
    {
	clmm_cb_t   mcb;


	/* Here if we only need to read the local params and return them.
	 */

	clmm_ctrl(&mcb, CLMM_GETGCB);

	parm_ptr->cg_decay_rate = mcb.mcb_aps_decay;
	parm_ptr->cg_recalc_time = mcb.mcb_event_run;
    }

    clm_reply(req, 0);
    return(0);
}


/*
 * Name
 *	remote_glob_tune_op
 * Description
 * 	set global migration parameters from message broadcast by a client
 * Arguments
 *	req - the set parameters request
 * Returns
 *	DLM_NORMAL - okay
 * Global Data Affected
 *	NONE
 */
int
remote_glob_tune_op(clm_msg_t *msg)
{
	clmm_cb_t		mcb;	/* local copy migration control block */
	dlm_globparams_t	*parm_ptr = msg->cm_data;
	int			ret = 0;

	DBPRINTF((DBG_MIGR|DBG_FUNC,
		"enter remote_glob_tune_op decay=%s, event=%d valid=%d\n",
		kfl_ftos(parm_ptr->cg_decay_rate),
		parm_ptr->cg_recalc_time, parm_ptr->cg_valid));

	/*
	 * here to set global migration parameters.
	 */
	clmm_ctrl(&mcb, CLMM_GETGCB);
	if (parm_ptr->cg_valid & CLMTUNE_GLOB_RECALC)
		mcb.mcb_event_run = parm_ptr->cg_recalc_time;
	if (parm_ptr->cg_valid & CLMTUNE_GLOB_DECAY)
		mcb.mcb_aps_decay = parm_ptr->cg_decay_rate;
	ret = clmm_ctrl(&mcb, CLMM_SETGCB);
	if (0 != ret)
		bsdlog(LOG_ERR, "Bad return from clmm_ctrl\n");

	return 0;
}



/*
 * Name
 *	begin_stats_op
 * Description
 * 	Determine the master node for the resource whose statistic
 *	information is requested and forward the request to that node.
 * Arguments
 *	req - the resource statistics request
 * Returns
 *	return status from pti_call
 * Global Data Affected
 * 	NONE
 */
int
begin_stats_op(clm_trans_t *req)
{
    struct resource *rp;
    int	stat;

	
    DBPRINTF((DBG_STATS|DBG_FUNC, "enter %s n=%s t=%d\n",  __FUNCTION__,
	      printbin(req->clm_statsinfo.resname,
		       req->clm_statsinfo.namelen),
	      req->clm_statsinfo.type));

    /* Try to find a pointer to the specified resource
     */

    rp = find_resource( req->clm_statsinfo.resname,
		        req->clm_statsinfo.namelen,
		        req->clm_statsinfo.type );

    if ( NULL != rp ) 
    {
	DBPRINTF((DBG_STATS, "begin_stats_op: found rh 0x%x\n", rp->rh.rh));

	stat = pti_call(req, rp->master ? rp->master : OurSiteID) ;
    }
    else 
    {
	req->clm_status = DLM_BADRESOURCE;
	sendudp(req);

	stat = 0;
    }


    DBPRINTF((DBG_STATS|DBG_FUNC, "leave begin_stats_op\n"));
    return( stat );
}

/*
 * Name
 *      cont_stats_op
 * Description
 *      Continue resource statistic information request that has come
 *      back from the master node
 * Arguments
 *      req - the request that was sent to the master
 *      resp - the response from the master
 * Returns
 *      NOTHING
 * Global Data Affected
 *      NONE
 */
void
cont_stats_op(clm_trans_t *req, clm_trans_t *resp)
{
    DBPRINTF((DBG_STATS|DBG_FUNC, "enter cont_stats_op\n"));

    if (resp->clm_status == DLM_NORMAL)
    {
	/* what should happen here */
	DBPRINTF((DBG_STATS, "cont_stats_op (master's response status = %d)\n",
		  resp->clm_status));
    }


    clm_reply(req, resp);
    DBPRINTF((DBG_STATS|DBG_FUNC, "leave cont_stats_op\n"));
}

/*
 * Name
 *	master_stats_op
 * Description
 *      return the statistic information on the resource
 * Arguments
 *	req - the statistics request
 * Returns
 *	DLM_NORMAL - okay
 * Global Data Affected
 *	NONE
 */
int
master_stats_op(clm_msg_t *msg)
{
    clm_trans_t *req = msg->cm_data;
    struct resource *rp;
    int	site;


    DBPRINTF((DBG_STATS|DBG_FUNC, "enter master_stats_op\n"));

    /* Try to find a pointer to the specified resource
     */

    rp = find_resource( req->clm_statsinfo.resname,
		        req->clm_statsinfo.namelen,
		        req->clm_statsinfo.type );

    /* check to make sure resource is mastered here, otherwise redirect */
    if (NULL == rp || 0 != rp->master) 
    {
	site = clm_direct( &req->clm_statsinfo.resname[0],
			   req->clm_statsinfo.namelen);
	return( (site == 0) ? -1 : site );
    }

    DBPRINTF((DBG_STATS, "master_stats_op: found rh 0x%x\n", rp->rh.rh));


    /* Request is not going to be re-directed.  call cccp_msg_duplicate() to
     * insure that we have a writeable copy of the message, reset the req
     * pointer and enqueue the response back to the secondary.
     */

    msg = cccp_msg_duplicate( msg );
    req = msg->cm_data;


    /* Copy resource statistics into repsonse
     */

    req->clm_statsinfo.statistics = rp->stats;
    req->clm_direction = PTI_RESPONSE;
    clm_queue_respmsg(msg);

    DBPRINTF((DBG_STATS|DBG_FUNC, "leave master_stats_op\n"));
    req->clm_status = DLM_NORMAL;

#ifdef DBPRINT
    clm_dumpres(rp);
#endif
    return 0;
}

/*
 * clm_global_stats (system call)
 *
 * Return global lock manager memory usage statistics.
 * If kernel==1, directly copy info, otherwise sending data to a
 * user program.
 */
int
clm_global_stats(struct dlm_stats_buf *buf,
                 int kernel)
{
  struct dlm_stats_buf a , *b;
#ifndef __linux__
    extern void *lock_heap;
#endif /* !__linux__ */
    int err;

    att_seg();

    if (kernel) {
      b = buf;
    } else {
      b = &a;
    }
  
    b->maxresources = maxresources ;
    b->reclocks_used = clm_reclocks_used ;
    b->reclocks_denied = clm_reclocks_denied ;
    b->reclocks_max = clm_reclocks_max ;
    b->items_max = clm_items_max ;
    b->lock_warned = clm_lock_warned ;
    b->res_warned = clm_res_warned ;
    b->res_used = clm_rescount;
#ifndef __linux__
    b->kvmem_used = ((struct heap *) lock_heap)->amount ;
    b->kvmem_max = ((struct heap *) lock_heap)->numpages << PGSHIFT ;
#endif /* !__linux__ */

    if (!kernel) {
      err = copy_to_user((char *) buf,  (char *) & a, sizeof *buf) ;
      if (err) {
        setuerror(err) ;
        det_seg();
        return -1 ;
      }
    }
    det_seg();
    return 0 ;
}
