/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: clm_unixlock.c,v 1.7 2001/09/12 20:29:19 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		add_reclock
 *		begin_lock_unix
 *		begin_unlock_unix
 *		blocked
 *		cancel
 *		cont_lock_unix
 *		cont_unlock_unix
 *		delete_reclock
 *		find_insert
 *		insert_le
 *		inside
 *		local_lock_unix
 *		local_unlock_unix
 *		master_lock_unix
 *		master_unlock_unix
 *		overlap
 *		purge_unix
 *		same_bound
 *		same_lock
 *		same_op
 *		search_block_lock
 *		search_lock
 *		wakeup_unix
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/signal.h>
#include <linux/types.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "dlm_kernel.h"
#include "clm_locksw.h"
#include "clm_data.h"			/* external data definitions */
#include "cltrc.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
#define FAKEREMID(a) ((a->lck.site << 28) | a->lck.pid)
static struct reclock *search_lock(struct lockinfo *a) ;
static int insert_le( dlm_list_t * queue, struct reclock * insrtp,
		      struct lockinfo *lockinfo );

/*
 * this file contains the high level actions of the lock manager:
 * LOCK,UNLOCK,CANCEL
 * Each action is implemented through three functions: begin_xxx(),
 * master_xxx(), and cont_xxx()
 *
 * The begin function is called at the start of the request where initial
 * state for the request is set before it is forwarded to the communications
 * layer to be sent to the master node for the resource.
 *
 * The master function is called on the node that should be the master for
 * the resource.  If the mastery has moved elsewhere, the master function
 * must be able to determine this and reject the request so that it will be
 * resent to the directory node.
 *
 * The continue request is called on the node from where the request originated
 * to duplicate state information passed back from the master node.  The
 * continue function must check to see if the resource is mastered locally
 * so it will not duplicate work or corrupt the local state that was already
 * set by the master function.  The continue function is also responsible
 * for sending the response to the client that will cause the clm_regionlock
 * function to return.
 *
 * Resource 'registration' is done by creation of a lock in NL mode.
 * These locks are always at the END of the GRANT queue, so they do
 * not participate in bounds checking.
 */

/*
 * Name 
 *	master_lock_unix
 * Description
 * 	add a lock locally
 * 	blocked() will assign insertp as a side effect
 * 	add lock at position indicated by insertp
 *	if upper bound is specified as 0 this means "infinity"
 *
 * Arguments
 *	req - transaction containing lock request
 * Returns in req->clm_status
 * 	DLM_DENIED_NOLOCKS for resource failures
 * 	DLM_WORKING for blocked lock request
 * 	DLM_DENIED for nonblocking requests that fail
 * 	DLM_NORMAL for successful lock request
 * Returns
 *	0 for successful completion
 *	site id of directory if we are not the master of the resource.
 * Global Data Affected
 *	NONE
 */
int
master_lock_unix(clm_msg_t *msg, int flags)
{
	clm_trans_t *req = msg->cm_data;
	struct resource *rp;
	struct reclock *insrtp;
	struct reclock *a;
	register struct lockinfo *lockinfo = &req->clm_lockinfo;


	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_LOCAL_LOCK_UNIX,
                req->clm_pid, req->clm_sequence, 0x00, 0x00);

	DBPRINTF((DBG_FUNC, "enter master_lock_unix\n"));

	if (!(rp = resptr(lockinfo->rh)) || rp->refcount <= 0 ||
			!my_res(lockinfo->site, req->clm_pid, rp) || 
			rp->rh.rh != lockinfo->rh.rh ||
			0 != rp->master) {
		int	site;

		DBPRINTF((DBG_UNIXLOCK,
				"master_lock_unix: bad resptr, rh=0x%x\n",
				lockinfo->rh));
		/*
		 * The local copy of this resource handle was already found
		 * by the begin routine or we wouldn't be here.  So the
		 * resource must exist.  Consult the directory.
		 */
		site = clm_direct(req->clm_lockreq.nameinfo.name,
				req->clm_lockreq.nameinfo.namelen);
		if (0 == site) {
			bsdlog(LOG_ERR, MSGSTR(CANT_DETERMINE_DIR_NODE,
					"Can't determine directory node"));
			return -1;
		}

		return site;
	}

	/*
	 * At this point, we know that we are going to send a
	 * response for this message and not redirect to the directory
	 * server, so copy the message and queue the response.
	 */
	msg = cccp_msg_duplicate( msg );
	req = msg->cm_data;
	lockinfo = &req->clm_lockinfo;
	msg->cm_node = req->clm_lockinfo.site;
	req->clm_direction = PTI_RESPONSE;
	clm_queue_respmsg(msg);

	DBPRINTF((DBG_UNIXLOCK, "master_lock_unix: lb %x, ub %x\n",
			lockinfo->lb, lockinfo->ub));
	
	
	/* let the secondary know who's boss */
	req->clm_rt_master = OurSiteID;

	if (lockinfo->ub == 0 || lockinfo->ub < lockinfo->lb)
		lockinfo->ub = UINT_MAX;
	

	if (blocked(rp, &insrtp, lockinfo))
	{					/*blocked*/
		if(lockinfo->lk_flags & LKM_NOQUEUE) {
			req->clm_status = DLM_DENIED;
		} else {
			if ((a = get_le(DLM_RES_UNIX)) == NULL) {
				bsdlog(LOG_ERR, MSGSTR(CANT_ALLOCATE_LK_ENTRY, 
				"master_lock_unix: can't allocate lock entry"));
				req->clm_status = DLM_DENIED_NOLOCKS;
			} else {
				uint	id;

				lockinfo->state |= LOCK_WAITING;
				id = a->lck.lockid;
				a->lck = *lockinfo;
				a->lck.lockid = id;
				a->route = req->clm_route;
				if (a->lck.site && a->lck.site != OurSiteID)
				    a->lck.remlockid = FAKEREMID(a) ;

				/* add to wait queue for this resource */
				lck_set_seq(rp, a);
				add_queue(&rp->lockqs[RLQ_WAIT], a, RL_RQ_OFS) ;
				rp->refcount++;
				check_handles(rp, 0) ;
				req->clm_status = DLM_WORKING;
			}
		}
	}
	else { /* not blocked*/
		if(add_reclock(rp, insrtp, lockinfo) == -1) 
			req->clm_status = DLM_DENIED_NOLOCKS;
		else {
			req->clm_status = DLM_NORMAL;
#ifdef nada
			clmm_cache(rp, msg->cm_node);
#endif
		}

	}

	DBPRINTF((DBG_FUNC, "leave master_lock_unix\n"));
	return 0;
}


/*
 * Name 
 *	local_lock_unix
 * Description
 * 	add a lock locally
 * 	blocked() will assign insertp as a side effect
 * 	add lock at position indicated by insertp
 *	if upper bound is specified as 0 this means "infinity"
 *
 * Arguments
 *	req - transaction containing lock request
 * Returns in req->clm_status
 * 	DLM_DENIED_NOLOCKS for resource failures
 * 	DLM_WORKING for blocked lock request
 * 	DLM_DENIED for nonblocking requests that fail
 * 	DLM_NORMAL for successful lock request
 * Returns
 *	0 for successful completion
 *	site id of directory if we are not the master of the resource.
 * Global Data Affected
 *	NONE
 */
int
local_lock_unix(clm_trans_t *req)
{
	struct resource *rp;
	struct reclock *insrtp;
	struct reclock *a;
	register struct lockinfo *lockinfo = &req->clm_lockinfo;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_LOCAL_LOCK_UNIX,
                req->clm_pid, req->clm_sequence, 0x00, 0x00);

	DBPRINTF((DBG_FUNC, "enter local_lock_unix\n"));

	if (req->clm_rt_master != OurSiteID) {
		lockinfo->rh = req->clm_route.lcl_rh;
	}

	if (!(rp = resptr(lockinfo->rh)) || rp->refcount <= 0 ||
			rp->rh.rh != lockinfo->rh.rh) {
		DBPRINTF((DBG_UNIXLOCK,
				"local_lock_unix: bad resptr, rh=0x%x\n",
				lockinfo->rh));
		/* this shouldn't happen */
		bsdlog(LOG_CRIT, MSGSTR(CANT_FIND_RESOURCE,
				"Can't find resource in local_lock_unix"));
		abort("local_lock_unix");
	}

	/*
	 * if the continue function is run on the node that is the master
	 * for the resource, then all local processing has already been
	 * completed
	 */
	DBPRINTF((DBG_UNIXLOCK,	"local_lock_unix: rp->master %x, lockinfo->ub %x \n", rp->master, lockinfo->ub));

	if (0 == rp->master)
		goto done;

	if (lockinfo->ub == 0)
		lockinfo->ub = MAXEND;

	if (blocked(rp, &insrtp, lockinfo) == 1) {  /*blocked*/
	    DBPRINTF((DBG_UNIXLOCK,	"local_lock_unix: blocked\n"));

	    
		if(lockinfo->lk_flags & LKM_NOQUEUE) {
	    DBPRINTF((DBG_UNIXLOCK,	"local_lock_unix: noqueue: denied\n"));
			req->clm_status = DLM_DENIED;
		}
		else {
			if ((a = get_le(DLM_RES_UNIX)) == NULL) {
				bsdlog(LOG_ERR, MSGSTR(CANT_ALLOCATE_LK_ENTRY,
				"local_lock_unix: can't allocate lock entry"));
				req->clm_status = DLM_DENIED_NOLOCKS;
			} else {
				int	id;

				lockinfo->state |= LOCK_WAITING;
				id = a->lck.lockid;
				a->lck = *lockinfo;
				a->lck.lockid= id;
				a->route = req->clm_route;
				if (a->lck.site && a->lck.site != OurSiteID)
				    a->lck.remlockid = FAKEREMID(a) ;

				/* add to wait queue for this resource */
				lck_set_seq(rp, a);
				add_queue(&rp->lockqs[RLQ_WAIT], a, RL_RQ_OFS) ;
				rp->refcount++;
				check_handles(rp, 0) ;
				req->clm_status = DLM_WORKING;
				DBPRINTF((DBG_UNIXLOCK,
						"local_lock_unix: working\n"));
			}
		}
	}
	else { /* not blocked*/
	    DBPRINTF((DBG_UNIXLOCK,	"local_lock_unix: non-blocked\n"));
		if(add_reclock(rp, insrtp, lockinfo) == -1) 
			req->clm_status = DLM_DENIED_NOLOCKS;
		else 
			req->clm_status = DLM_NORMAL;
	}

	DBPRINTF((DBG_FUNC, "exit local_lock_unix (%d)\n", req->clm_status));

done:
	return req->clm_status;
}


/*
 * Name
 *	begin_lock_unix
 * Description
 * 	send a remote lock request after some trivial screening on this end
 * Arguments
 *	req - the lock request
 * Returns
 *	0 for success - with the following status passed back in the request
 *		DLM_NORMAL - desired region is already locked
 *			or request was posted to primary
 *	-1 if request not sent to primary  - with the following status passed 
 *		back in the request:
 *		DLM_IVRESHANDLE - bad resource handle
 * 		DLM_DENIED - if blocked and LKM_NOQUEUE specified
 * Global Data Affected
 *	NONE
 */
int
begin_lock_unix(clm_trans_t *req)
{
	struct resource *rp;
	struct reclock *insrtp;
	register struct lockinfo *lockinfo = &req->clm_lockinfo;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_REMOTE_LOCK_UNIX,
                req->clm_pid, req->clm_sequence, 0x00, 0x00);

	DBPRINTF((DBG_FUNC, "enter begin_lock_unix\n"));

	/* Add this to remove need for clinfo in API */
	req->clm_lockinfo.site = OurSiteID;

	if (lockinfo->mode != LKM_EXMODE && lockinfo->mode != LKM_PRMODE)
	{
	    req->clm_status = DLM_BADARGS ;
	    clm_reply(0, req) ;
	    goto done ;
	}

	if (!(rp = resptr(lockinfo->rh)) || rp->refcount <= 0 ||
			!my_res(lockinfo->site, req->clm_pid, rp) || 
			rp->rh.rh != lockinfo->rh.rh) {
		DBPRINTF((DBG_UNIXLOCK,
				"begin_lock_unix: bad resptr, rh=0x%x\n",
				lockinfo->rh));
		req->clm_status = DLM_IVRESHANDLE;
		clm_reply(0,req);
		goto done;
	}

	if(blocked(rp, &insrtp, lockinfo)) 
	{ 
	    DBPRINTF((DBG_UNIXLOCK, "begin_lock_unix: blocked! Noqueue %d\n",
			(lockinfo->lk_flags & LKM_NOQUEUE)));
		/*blocked*/
		if(lockinfo->lk_flags & LKM_NOQUEUE) {
			req->clm_status = DLM_DENIED;
			clm_reply(0,req);
			goto done;
		}
	}

	/*
	 * Check to see if request is from a process that already has a lock
	 * that spans the new request, if it is then don't bother calling
	 * the master, simply grant the lock.
	 */

	if(insrtp != NULL && same_proc(&insrtp->lck, lockinfo) && 
		 same_op(&insrtp->lck, lockinfo) && 
			inside(lockinfo, &insrtp->lck)) 
	{
	    DBPRINTF((DBG_UNIXLOCK, "begin_lock_unix: lock exists\n"));
		/*lock exists*/
		req->clm_status = DLM_NORMAL;
		clm_reply(0,req);
		goto ok;
	}
	else 
	{
		if (0 != rp->master) {
			/* save local resource handle */
			req->clm_route.lcl_rh = lockinfo->rh;
			lockinfo->rh = rp->mst_rh;
		}
		/* 
		 * Consult with master
		 */ 
		DBPRINTF((DBG_UNIXLOCK,
				"begin_lock_unix: consulting master (%d)\n",
				rp->master));
		pti_call_unix(req, rp);
	}
ok:
	DBPRINTF((DBG_FUNC, "exit begin_lock_unix OK\n"));
	return 0;
done:
	DBPRINTF((DBG_FUNC, "exit begin_lock_unix NOT OK\n"));
	return -1;
}

/*
 * Name 
 *	cont_lock_unix
 * Description
 *	Continue a lock request coming back from the primary
 *	If the status was DLM_NORMAL, call local_lock_unix
 * Arguments
 *      req - the request that was sent to the primary
 *      resp - the response from the primary
 * Returns
 *      NOTHING
 * Global Data Affected
 *      NONE
 */
void
cont_lock_unix(clm_trans_t *req, clm_trans_t *resp)
{
	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_CONT_REMOTE_LOCK_UNIX,
                req->clm_pid, req->clm_sequence, resp->clm_status, 0x00);

	DBPRINTF((DBG_FUNC, "enter cont_lock_unix (status %d)\n", resp->clm_status));
	if (resp->clm_status == DLM_NORMAL)
		local_lock_unix(resp);
	clm_reply(req,resp);
	DBPRINTF((DBG_FUNC, "exit cont_lock_unix\n"));
}
	

/*
 * Name
 *	master_unlock_unix
 * Description
 *	Free a lock region on the local machine. Wake any process blocked 
 *	on this lock region
 * Arguments
 *	req - the unlock request
 * Returns
 *	DLM_IVRESHANDLE - bad resource handle
 *	DLM_DENIED - no locks matched region specified
 *	DLM_NORMAL - success
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
master_unlock_unix(clm_msg_t *msg)
{
	clm_trans_t *req = msg->cm_data;
	struct resource *rp;
	struct reclock *insrtp;
	register struct lockinfo *lockinfo = &req->clm_lockinfo;
	register int nfreed = 0;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_LOCAL_UNLOCK_UNIX,
                req->clm_pid, req->clm_sequence, 0x00, 0x00);

	DBPRINTF((DBG_FUNC, "enter master_unlock_unix\n"));

	if (!(rp = resptr(lockinfo->rh))  || rp->refcount <= 0 ||
			!my_res(lockinfo->site, req->clm_pid, rp) || 
			rp->rh.rh != lockinfo->rh.rh ||
					0 != rp->master) {
		int	site;

		DBPRINTF((DBG_UNIXLOCK,
				"master_unlock_unix: bad resptr, rh=0x%x\n",
				lockinfo->rh));
		/*
		 * The local copy of this resource handle was already found
		 * by the begin routine or we wouldn't be here.  So the
		 * resource must exist.  Consult the directory.
		 */
		site = clm_direct(req->clm_lockreq.nameinfo.name,
				req->clm_lockreq.nameinfo.namelen);
		if (0 == site) {
			bsdlog(LOG_ERR, MSGSTR(CANT_DETERMINE_DIR_NODE,
					"Can't determine directory node"));
			return -1;
		}

	DBPRINTF((DBG_UNIXLOCK, "master_unlock_unix: consulted dir site %d\n", site));
		return site;
	}

	/*
	 * At this point, we know that we are going to send a
	 * response for this message and not redirect to the directory
	 * server, so copy the message and queue the response.
	 */
	msg = cccp_msg_duplicate( msg );
	req = msg->cm_data;
	lockinfo = &req->clm_lockinfo;
	msg->cm_node = req->clm_lockinfo.site;
	req->clm_direction = PTI_RESPONSE;
	clm_queue_respmsg(msg);

	/* let the secondary know who's boss */
	req->clm_rt_master = OurSiteID;
	
	insrtp = NULL;
	nfreed = delete_reclock(rp, &insrtp, lockinfo);

	DBPRINTF((DBG_UNIXLOCK, "master_unlock_unix: nfreed (from delete_reclock) %d\n", nfreed));
	
	if (nfreed > 0)	{
		req->clm_status = DLM_NORMAL;
#ifdef nada
		clmm_cache(rp, msg->cm_node); 
#endif
		/* here if any locks were deleted */
		wakeup_unix(rp, lockinfo);
	} else
		req->clm_status = DLM_DENIED;

	DBPRINTF((DBG_FUNC, "exit master_unlock_unix\n"));
	
	return 0;
}


/*
 * Name
 *	local_unlock_unix
 * Description
 *	Free a lock region on the local machine. Wake any process blocked 
 *	on this lock region
 * Arguments
 *	req - the unlock request
 * Returns
 *	DLM_IVRESHANDLE - bad resource handle
 *	DLM_DENIED - no locks matched region specified
 *	DLM_NORMAL - success
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
local_unlock_unix(clm_trans_t *req)
{
	struct resource *rp;
	struct reclock *insrtp;
	register struct lockinfo *lockinfo = &req->clm_lockinfo;
	register int nfreed = 0;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_LOCAL_UNLOCK_UNIX,
                req->clm_pid, req->clm_sequence, 0x00, 0x00);

	DBPRINTF((DBG_FUNC, "enter local_unlock_unix\n"));

	if (req->clm_rt_master != OurSiteID) {
		lockinfo->rh = req->clm_route.lcl_rh;
	}

	if (!(rp = resptr(lockinfo->rh)) || rp->refcount <= 0 ||
			!my_res(lockinfo->site, req->clm_pid, rp) || 
			rp->rh.rh != lockinfo->rh.rh) {
		DBPRINTF((DBG_UNIXLOCK,
				"local_unlock_unix: bad resptr, rh=0x%x\n",
				lockinfo->rh));
		/* this shouldn't happen */
		bsdlog(LOG_CRIT, MSGSTR(CANT_FIND_RESOURCE,
				"Can't find resource in local_unlock_unix"));
		abort("local_unlock_unix");
	}

	/*
	 * if the continue function is run on the node that is the master
	 * for the resource, then all local processing has already been
	 * completed
	 */
	if (0 == rp->master)
		goto done;

	insrtp = NULL;
	nfreed = delete_reclock(rp, &insrtp, lockinfo);
	DBPRINTF((DBG_UNIXLOCK, "local_unlock_unix: nfreed (delete_reclock) %d\n", nfreed));
	
	if (nfreed > 0)	{
		req->clm_status = DLM_NORMAL;
		/* here if any locks were deleted */
		wakeup_unix(rp, lockinfo);
	} else
		req->clm_status = DLM_DENIED;

	DBPRINTF((DBG_FUNC, "exit local_unlock_unix (%d)\n", req->clm_status));
done:
	return req->clm_status;
}

/*
 * Name
 *	begin_unlock_unix
 * Description
 * 	Send an unlock request to the primary
 * Arguments
 *	req - the unlock request
 * Returns
 *	return status from pti_call 
 *	-1 and status in request set to DLM_IVRESHANDLE if supplied resource 
 *		handle is invalid
 * Global Data Affected
 * 	NONE
 */
int
begin_unlock_unix(clm_trans_t *req)
{
	register struct lockinfo *lockinfo = &req->clm_lockinfo;
	struct resource	*rp;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_REMOTE_UNLOCK_UNIX,
                req->clm_pid, req->clm_sequence, 0x00, 0x00);

	DBPRINTF((DBG_FUNC, "enter begin_unlock_unix\n"));

	if(NULL == (rp = resptr(lockinfo->rh)) || rp->refcount <= 0 ||
			!my_res(lockinfo->site, req->clm_pid, rp) || 
			rp->rh.rh != lockinfo->rh.rh) {
		DBPRINTF((DBG_UNIXLOCK, "begin_unlock: bad resptr, rh=0x%x\n",
					lockinfo->rh));
		req->clm_status = DLM_IVRESHANDLE;
		clm_reply(0,req);
		return -1;
	}

	if (lockinfo->ub == 0)
		lockinfo->ub = UINT_MAX;

	if (0 != rp->master) {
		/* save local resource handle */
		req->clm_route.lcl_rh = lockinfo->rh;
		lockinfo->rh = rp->mst_rh;
	}

	pti_call_unix(req, rp);

	DBPRINTF((DBG_FUNC, "exit begin_unlock_unix\n"));
	return 0;
}

/*
 * Name
 *      cont_unlock_unix
 * Description
 *      Continue an unlock request coming back from the primary
 *      If the status was DLM_NORMAL, call local_unlock_unix
 * Arguments
 *      req - the request that was sent to the primary
 *      resp - the response from the primary
 * Returns
 *      NOTHING
 * Global Data Affected
 *      NONE
 */
void
cont_unlock_unix(clm_trans_t *req, clm_trans_t *resp)
{
	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_CONT_REMOTE_UNLOCK_UNIX,
                req->clm_pid, req->clm_sequence, resp->clm_status, 0x00);

	DBPRINTF((DBG_FUNC, "exit cont_unlock_unix\n"));
	if (resp->clm_status == DLM_NORMAL)
		local_unlock_unix(resp);
	clm_reply(req,resp);
	DBPRINTF((DBG_FUNC, "exit cont_unlock_unix\n"));
}
	

/* Name
 *	blocked
 * Description
 *	Checks whether a new lock (a) will be blocked by 
 * 	any previously granted lock (owned by another process).
 * 
 * 	all struct reclocks are in [lb, ub) in increasing lb order.
 *
 * Arguments
 *	rp - the resource associated with the lock
 *	insrtp - address of insert pointer
 *	a - the lock to be inserted
 * Returns
 *	NULL if "a" is not blocked and sets the contents of insrtp pointing  
 * 		to the lock before which the  new lock should be added
 *	1  "a" is blocked and leaves insrtp pointing to the first
 * 		blocking lock
 * Global Data Affected
 * 	NONE
 */
int
blocked(struct resource *rp, struct reclock **insrtp, struct lockinfo *a)
{
	register struct reclock *nl;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_BLOCKED,
                a->pid, a->rh.rh, 0x00, 0x00);

	DBPRINTF((DBG_FUNC, "enter blocked\n"));

	/*set up initial insrtp value*/
	*insrtp = NULL;

    for ( EACH_DLM_LISTNODE( & rp->lockqs[RLQ_GRANT], nl ) )
    {
	if ( (nl->lck.mode == LKM_NLMODE) || (nl->lck.lb > a->lb) )
	{
	    break;
	}

	*insrtp = nl;

	/* identify an overlapped lock owned by same process */
	if (same_proc(&nl->lck, a) && a->lb <= nl->lck.ub)
	    break;

	/* identify and reject an incompatible overlap */
	if (!same_proc(&nl->lck,a) &&  a->lb < nl->lck.ub 
	    && (nl->lck.mode == LKM_EXMODE || a->mode == LKM_EXMODE))
	{
	    DBPRINTF((DBG_FUNC, "exit blocked YES!\n"));
	    return 1;
	}
    }


    for ( /* empty */
	  ; (nl != NULL) && (nl->lck.mode != LKM_NLMODE)
	     && (nl->lck.lb < a->ub)
	  ; nl = (struct reclock *) dlm_list_succ( (dlm_listnode_t *) nl ) )
    {
	/* identify and reject an incompatible overlap */
	if (!same_proc(&nl->lck,a) && a->lb < nl->lck.ub 
	    && (nl->lck.mode == LKM_EXMODE || a->mode == LKM_EXMODE))
	{
	    *insrtp = nl;
	    DBPRINTF((DBG_FUNC, "exit blocked YES\n"));
	    return 1;
	}
    }


    DBPRINTF((DBG_FUNC, "exit blocked NO\n"));
    return 0;
}

/* Name
 *	add_reclock
 * Description
 *	add_reclock modifies existing struct reclock list pointed to by 
 * 	rp->lockqs[RLQ_GRANT] and adds new lock request "a" 
 * 	starting from position pointed to by insrtp;
 *
 * 	first delete all locks in specified region that are owned by the
 * 	same process, then add new lock. This might have the side effect
 * 	off downgrading a region, i.e. releasing an exclusive lock.
 * Arguments
 *	rp - the resource associated with the lock
 *	insrtp - insertion point
 *	a - the lock 
 * Returns
 * 	0 for success
 * 	-1 if no more le entries are available or insert_le() fails
 * Global Data Affected
 *	NONE
 */
int
add_reclock(struct resource *rp, struct reclock *insrtp, struct lockinfo *a)
{
	struct reclock *nl;
	int ind;

	/* check whether an equivilant lock already exists for this client */
	if ( insrtp && same_proc(&insrtp->lck, a) && 
		 same_op(&insrtp->lck, a) && inside(a, &insrtp->lck)) 
	{ 
		/* lock exists */
		if (&insrtp->lck == a) { 
			/* extra protection */
			DBPRINTF((DBG_UNIXLOCK, __FUNCTION__ 
				  ": insrtp = a = %x error!!!\n", a));
			        
		}
		else 
		{
			/*a->rel = 1;*/
		}

		return(0);
	}

	/* delete all reclocks owned by the same process from *insrtp 
 	 * that overlap the desired range
	 */
	ind = delete_reclock(rp, &insrtp, a);

	/* check to see if a's lower bound is connected 
	 * with *insrtp's upper bound. If so, expand a's lower bound and 
	 * discard *insrtp.
	 */

    if (insrtp != NULL) 
    {
	nl = insrtp;
	if ( same_proc(&nl->lck, a) && same_op(&nl->lck, a) 
	    && nl->lck.ub == a->lb )
	{
	    a->lb = nl->lck.lb;

	    insrtp = (struct reclock *) dlm_list_succ( (dlm_listnode_t *)
						       insrtp );

	    remove_queue(&rp->lockqs[RLQ_GRANT], nl, RL_RQ_OFS) ;
	    rp->refcount-- ;
	    free_le(DLM_RES_UNIX, nl);
	}
    }

    /* check to see if a's upper bound is connected to another lock 
     * if so, expand a's upper bound and discard the other lock
     */

    for ( EACH_DLM_LISTNODE( & rp->lockqs[RLQ_GRANT], nl ) )
    {
	if ( (nl->lck.mode == LKM_NLMODE) || (nl->lck.lb > a->ub) )
	{
	    break;
	}

	if ( same_proc(&nl->lck, a) && same_op(&nl->lck, a) 
	    && nl->lck.lb == a->ub )
	{ /*extend*/
	    a->ub = nl->lck.ub;
	    remove_queue(&rp->lockqs[RLQ_GRANT], nl, RL_RQ_OFS) ;
	    rp->refcount-- ;
	    free_le(DLM_RES_UNIX, nl);
	    break;
	}
    }

    a->state &= ~LOCK_WAITING ;
    a->state |= LOCK_GRANTED ;
    if (insert_le(&rp->lockqs[RLQ_GRANT], insrtp, a) != 0)
	return -1;

    rp->refcount++;

    if (a->mode == LKM_PRMODE && ind == 2) /*lock_ex => lock_sh*/
	wakeup_unix(rp, a);

    return(0);
}

/* 
 * Name
 *	delete_reclock
 * Description
 *	delete locks in the struct reclock list ptr to by fp->rlckp
 * 	(starting from insrtp position)
 * 	that are owned by the same process as "a" and in the
 * 	ranges specified by "a".
 * 	if the lock pointed to  by *insrtp has been deleted, insrtp is modified
 * 	to point to the lock before; this parameter is used for add_reclock only
 * Arguments
 *	rp - the resource associated with the lock
 *	insrtp - starting point
 *	a - the lock 
 * Returns
 * 	-1 if deletion requires new reclock and no more available
 *	0 if no lock is deleted;
 * 	1 if only shared locks have been deleted;
 * 	2 if some exclusive locks have been deleted;
 *
 * 	this return value is used to determine if add_reclock caused a
 * 	a downgrade of an exclusive and should call wakeup() or if remote 
 * 	lock manager needs to be contacted for deletion. 
 * Global Data Affected
 *	NONE
 */
int
delete_reclock(struct resource *rp, struct reclock **insrtp, struct lockinfo *a)
{
	struct reclock *nl;
	struct reclock *next;
	int lock_ex, lock_sh;
	struct reclock *new_lock;

	lock_ex = 0;
	lock_sh = 0;

    nl = (*insrtp != NULL)
	? *insrtp
	: (struct reclock *) dlm_list_get_head( & rp->lockqs[RLQ_GRANT] );

    /* while a lock in the list overlaps with the victim...*/
    /* start at insert point if it exists, else head of grant q */

    for ( /* empty */
	  ; (nl != NULL)
	     && (nl->lck.mode != LKM_NLMODE)
	     && (nl->lck.lb < a->ub)
	  ; nl = next )
    {
	next = (struct reclock *) dlm_list_succ( (dlm_listnode_t *) nl );

	if (same_proc(&nl->lck, a) && nl->lck.ub > a->lb) 
	{
	    /*overlap*/
	    if (nl->lck.mode == LKM_EXMODE)
		lock_ex ++;
	    else
		lock_sh ++;

	    /* Existing lock inside request region delete it completely
	     */

	    if (inside(&nl->lck, a)) 
	    { /*delete complete*/
		if ( nl == *insrtp)
		    *insrtp = next;

		remove_queue(&rp->lockqs[RLQ_GRANT], nl, RL_RQ_OFS) ;
		rp->refcount-- ;
		free_le(DLM_RES_UNIX, nl);
	    }

	    /* Existing lock straddles request region so split into two
	     * discrete locks
	     */
	    else if ( nl->lck.ub > a->ub &&
		     nl->lck.lb < a->lb) { /*break into half*/
		uint	id;

		if ((new_lock = get_le(DLM_RES_UNIX)) == NULL) {
		    bsdlog( LOG_ERR,
			   MSGSTR(CANT_GET_LOCK_ENTRY, 
				  "%s: can't get lock entry"),
			   __FUNCTION__ );
		    return -1;
		}

		/* New lock gets top half.  First copy all misc data 
		 */

		id = new_lock->lck.lockid;
		*new_lock = *nl;
		new_lock->lck.lockid= id;

		/* the next line is implicit */
		/* new_lock->lck.ub = nl->lck.ub;*/

		new_lock->lck.lb = a->ub;
		new_lock->lck.remlockid = 0;

		/* nl gets the lower half */
		if (find_insert(rp, new_lock) != 0)
		    return -1;

		nl->lck.ub = a->lb;
	    }
	    else if (a->lb > nl->lck.lb)
	    { /* first half remains */
		nl->lck.ub = a->lb;
	    }
	    else if (a->ub < nl->lck.ub)
	    { /*second half remains */
		nl->lck.lb = a->ub;

		if (nl == *insrtp)
		{
		    *insrtp = next;
		}

		remove_queue(&rp->lockqs[RLQ_GRANT],
			     nl, RL_RQ_OFS) ;	/* unlink */

		rp->refcount-- ;

		if (find_insert(rp, nl) != 0)	/* relink */
		    return -1;
	    }
	    else 
	    {
		bsdlog( LOG_ERR, MSGSTR(IMPOSSIBLE, "%s: "
					"impossible!"), __FUNCTION__ );
	    }
	}
    }

    if (lock_ex > 0)	/* some exclusive lock has been deleted */
	return(2);
    if (lock_sh > 0)	/* some shared lock has been deleted */
	return(1);
    else			/* no deletion */
	return(0);
}

/*
 * Name
 *	cancel
 * Description
 * 	cancel a request that is waiting to be granted
 * Arguments
 *	rp - the resource associated with the request
 *	a - the request to be cancelled
 * Returns 	
 * 	0 for sucess (either not found or removed from wait queue)
 *	-1 if lock is already granted
 * Global Data Affected
 *	NONE
 */
int
cancel(struct resource *rp, struct lockinfo *a)
{
	struct reclock *nl;

	TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_CANCEL,
                a->pid, a->rh.rh, 0x00, 0x00);

	if ((nl = search_lock(a)) != NULL) {
		if ((nl->lck.state & LOCK_WAITING) == 0) {
			/* here if lock already granted */
			return -1 ;
		}
		else {
			a->state &= ~LOCK_WAITING;	
			/* unlink */
			remove_queue(&rp->lockqs[RLQ_WAIT], nl, RL_RQ_OFS) ;
			free_le(DLM_RES_UNIX, nl);
			decr_delete_resource(rp, 0) ;
		}
	}
	return 0;
}


/*
 * Name
 *	search_lock
 * Description
 *	search_lock locates an identical lock as a in either the grant queue
 *	or the wait queue of the resource associated with the specified lock
 * Arguments
 *	a - the lock to match
 * Returns
 *	NULL if no match found
 *	else returns matching lock
 * Global Data Affected
 *	NONE
 */
static struct reclock *
search_lock(struct lockinfo *a)
{
	struct resource *rp;
	struct reclock *nl;

	if (!(rp = resptr(a->rh)) || rp->refcount <= 0 ||
			!my_res(a->site, a->pid, rp) || 
			rp->rh.rh != a->rh.rh)
		bsdlog(LOG_ERR, MSGSTR(BAD_RESPTR_SRCH_LK,
				       "search_lock: bad resptr, rh=0x%lx\n"),
		       a->rh.rh);
		return(NULL);

	if (!blocked(rp, &nl, a)) {	/* not blocked */
		if (nl != NULL && same_proc(&nl->lck, a) &&
		 same_bound(&nl->lck, a) && same_op(&nl->lck, a))
			return(nl);		/* found in grant queue */
	}
	/* search in wait queue */
	return search_block_lock(rp, a);
}

/*
 * Name
 *	search_block_lock
 * Description
 * 	search wait queue of specified resource for a request that matches "a"
 * Arguments
 *	rp - the resource associated with the request
 *  	a - the lock request
 * Returns
 *	NULL if not found
 *	else returns the matching lock
 * Global Data Affected
 *	NONE
 */
struct reclock *
search_block_lock(struct resource *rp, struct lockinfo *a)
{
    struct reclock *nl;

    for ( EACH_DLM_LISTNODE( & rp->lockqs[RLQ_WAIT], nl ) )
    {
	if (same_lock(&nl->lck, a))
	    return nl ;
    }

    return NULL ;
}


/*
 * Name
 *	wakeup_unix
 * Description
 * 	Searches wait queue of specified resource for requests that are
 * 	waiting for area [a->lb, a->ub) and attempts to grant them.
 * 	This gets called when a delete lock operation is successful or when an
 * 	exclusive lock is downgraded to a shared lock.
 * Arguments
 *	rp - the associated resource
 *	a - a reclock containing the range that has been freed
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
wakeup_unix(struct resource *rp, struct lockinfo *a)
{
    dlm_listnode_t * curr;
    dlm_listnode_t * next;
    struct reclock * nl;
    struct reclock * insrtp;


    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_WAKEUP_UNIX,
	    0x00, 0x00, 0x00, 0x00);

    for ( curr = dlm_list_get_head( & rp->lockqs[RLQ_WAIT] )
	  ; curr != NULL
	  ; curr = next )
    {
	next = dlm_list_succ( curr );
	nl = (struct reclock *) curr;

	/* If "a"'s region overlaps this request and this request is not
	 * blocked
	 */

	if (overlap(&nl->lck, a) &&
	    blocked(rp, &insrtp, &nl->lck) == 0)
	{
#ifdef DBPRINT
	    DBPRINTF((DBG_UNIXLOCK,  "waking:" ));
	    clm_printlock(nl);
#endif
	    add_reclock(rp, insrtp, &nl->lck);
	    remove_queue(&rp->lockqs[RLQ_WAIT], nl, RL_RQ_OFS) ;
	    rp->refcount-- ;
	    notify(nl, RESP_UDP, DLM_NORMAL);
	    free_le(rp->type, nl);
	}
    }
}



/*
 * Name
 *	find_insert
 * Description
 *	find "a" in the proper order in the lock list and insert;
 * 	does not worry about overlap or merge with lock from same process;
 * 	it all should have been taken care of before find_insert is called;
 * Arguments
 *	rp - the resource associated with a
 *	a - the reclock describing the region to be inserted
 * Returns
 * 	0 for success
 * 	-1 for failure
 * Global Data Affected
 *	NONE
 */
int
find_insert(struct resource *rp, struct reclock *a)
{
    struct reclock *nl;
    struct reclock *insrtp;
    int	ret;


    insrtp = NULL;

    for ( EACH_DLM_LISTNODE( & rp->lockqs[RLQ_GRANT], nl ) )
    {
	if ( (nl->lck.mode == LKM_NLMODE) 
	     || (nl->lck.lb >= a->lck.ub) )
	{
	    break;
	}

	insrtp = nl;

	if (nl->lck.lb >= a->lck.lb)
	{
	    break;
	}
    }

    ret = insert_le( & rp->lockqs[RLQ_GRANT], insrtp, & a->lck );
    rp->refcount++;

    return( ret );
}




/*
 * Name 
 * 	insert_le
 * Description
 *	Allocates a reclock and adds it to the specified list 
 * 	after entry pointed to by insertp
 * 	if insrtp is NULL, put on front of list
 * Returns 
 *	0 for success
 * 	-1 for failure to allocate reclock
 */
static int
insert_le(
	   /* Queue in which the new reclock will be inserted.
	    */

	   dlm_list_t * queue,

	   /* reclock that comes just before the new reclock in the queue.
	    */

	   struct reclock * pre,

	   /* Lock info. used to initialize the reclock.
	    */

	   struct lockinfo * lockinfo
	 )
{
    struct reclock * a;
    unsigned         id;


    a = get_le( DLM_RES_UNIX );
    if ( a == NULL) 
    {
	bsdlog(LOG_ERR, MSGSTR(CANT_ALLOCATE_LK_ENTRY3,
			       "insert_le: can't allocate lock entry"));
	return -1;
    }


    id = a->lck.lockid;
    a->lck = *lockinfo;
    a->lck.lockid = id;
    a->lck.remlockid = 0;

    dlm_list_insert( queue, (dlm_listnode_t *) a, (dlm_listnode_t *) pre );
    return 0 ;
}

/*
 * Name
 *	inside
 * Description
 *	Determine if the range described by a lock is encompassed
 *	by that described another lock
 * Arguments
 *	a - the lock to check
 *	b - the lock which a must fall inside
 * Returns
 *	0 if a is not inside b
 *	non-zero otherwise
 * Global Data Affected
 *	NONE
 */
bool_t
inside(struct lockinfo *a, struct lockinfo *b)
{
	if (a == NULL || b == NULL)
		return(FALSE);
	else
		return(b->lb <= a->lb && b->ub >= a->ub);
}

/*
 * Name
 *	overlap
 * Description
 *	Determine if the range described by a lock is overlapped
 *	by that described another lock
 * Arguments
 *	a - the lock to check
 *	b - the lock which a must overlap
 * Returns
 *	0 if a doesn't overlap b
 *	non-zero otherwise
 * Global Data Affected
 *	NONE
 */
bool_t
overlap(struct lockinfo *a, struct lockinfo *b)
{
	if (a->ub <= b->lb || a->lb >= b->ub)
		return(FALSE);
	else
		return(TRUE);
}

/*
 * Name
 *	same_bound
 * Description
 *	Determine if the range described by a lock is identical
 *	to that described another lock
 * Arguments
 *	a - the lock to check
 *	b - the lock which must match
 * Returns
 *	0 if a doesn't have the same upper and lower bounds
 *	non-zero otherwise
 * Global Data Affected
 *	NONE
 */
bool_t
same_bound(struct lockinfo *a, struct lockinfo *b)
{
	return(a->lb == b->lb && a->ub == b->ub);
}

/*
 * Name
 *	same_op
 * Description
 *	Determine if the mode held by a lock is identical
 *	to that held another lock
 * Arguments
 *	a - the lock to check
 *	b - the lock whose mode must match
 * Returns
 *	0 if a doesn't have the same mode
 *	non-zero otherwise
 * Global Data Affected
 *	NONE
 */
bool_t
same_op(struct lockinfo *a, struct lockinfo *b)
{
	return (a->mode == b->mode);
}
/*
 * Name
 *	same_lock
 * Description
 *	Determine if two lockinfo structure describe the 
 *	same lock by comparing pid,op, and bounds
 * Arguments
 *	a,b - the two locks to compare
 * Returns
 *	1 if they are the same lock
 *	0 if no
 * Global Data Affected
 *	NONE
 */
bool_t
same_lock(struct lockinfo *a, struct lockinfo *b)
{
	return (same_proc(a, b) && same_op(a, b) && same_bound(a, b));
}


/*
 * Name
 *	unix_purge
 * Description
 *	Purge the specified lock
 * Arguments
 *	rp - resource on which lock resides
 * 	headpp - pointer to head of current queue for lock
 *	tailpp - pointer to tail ""
 * 	a - the lock itself
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
purge_unix(struct resource *rp, dlm_list_t * queue, struct reclock *a)
{
    remove_queue( queue, a, RL_RQ_OFS );
    (lockswitch[rp->type].wakeup)(rp, (struct reclock*)(&a->lck));
    free_le(rp->type, a);
}
