/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_base.c,v 1.7 2001/10/29 21:51:56 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		clm_addr_to_id
 *		clm_getlocalid
 *		clm_id_to_addr
 *		clm_node_active
 *		clm_node_count
 *		clm_which_net
 *		clm_change_topology
 *		nodeid_to_index
 *		index_to_nodeid
 *		handle_new_topology
 *		configInited
 */

/* ============================================================ */
#ifndef __KERNEL__
# define __KERNEL__
#endif
#ifndef MODULE
# define MODULE
#endif

#include <linux/kernel.h>
#include <linux/module.h>

#ifdef USING_FULL_VERSIONING
# if CONFIG_MODVERSIONS==1
#  define MODVERSIONS
#  include <linux/modversions.h>
# endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "dlm_kernel.h"
#include "pti.h"
#include "cltrc.h"
#include "dlm_version.h"
#include "dlm_devmsg.h"
#include "clm_data.h"
#include <iterator.h>

#include "dlm_clust.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */

int OurSiteID = -1;			/* lock manager site id */
int first_top_call;
static struct node_list s_node_list;
static struct node_list s_new_node_list;
struct node_list * node_list = &s_node_list;
struct node_list * new_node_list = &s_new_node_list;




/****************************************************************************/

struct dlm_nodeid_iterator
{
    iterator_t   base;
    unsigned     next_index;
};

static bool_t dlm_nodeid_iterator_next( iterator_t * iter, void ** data );
static void dlm_nodeid_iterator_destroy( iterator_t * iter );

static const struct iterator_methods   dlm_nodeid_iterator_methods =
{
    destroy: dlm_nodeid_iterator_destroy,
    next: dlm_nodeid_iterator_next
};




/*
 * Name
 *	clm_info_init
 * Description
 *	Init static in clm_info.c
 * Arguments
 * 	None
 * Returns
 *	Nothing
 * Global Data Affected
 *	syncd
 */
static int
clm_info_init(void)
{
    OurSiteID = -1 ;

    (void) memset( & s_node_list, 0, sizeof s_node_list );
    (void) memset( & s_new_node_list, 0, sizeof s_new_node_list );

    node_list->low_version = -1 ;
    node_list->high_version = -1 ;
    node_list->node_count = -1 ;
    new_node_list->low_version = -1 ;
    new_node_list->high_version = -1 ;
    new_node_list->node_count = -1 ;
    first_top_call = 1 ;

    printk( KERN_INFO __FUNCTION__ " done.\n" );
    return( 0 );
}

static void
clm_info_exit( void )
{
    printk( KERN_INFO __FUNCTION__ " done.\n" );
}





/*
 * Function:
 *	clm_node_count
 * Description:
 *	Return the number of configured nodes in the cluster
 * Arguments:
 *	None
 * Return value:
 *	The number of configured nodes in the cluster
 *	Returns 0 if cluster info not init'd yet.
 * Side effects:
 *	NONE
 */

unsigned
clm_node_count(void)
{
    if (OurSiteID != -1)
	return node_list->node_count ;
    else
	return 0 ;
}




/*
 * name            index_to_nodeid
 * purpose         translate an index to the corresponding nodeid
 * return values   -1 if index out of range or not in table, nodeid otherwise
 * side effects    none
 */

int
index_to_nodeid(int idx)
{
    int n ;

    if (idx < 0 || idx >= MAXNODES || ! (n = node_list->node[idx].nodeid)) {
      return -1 ;
    }
    return n ;
}

/*
 * name            nodeid_to_index
 * purpose         translate a nodeid to table index
 * return values   -1 if nodeid not in table, index of entry otherwise
 * side effects    none
 */

int
nodeid_to_index(int id)
{
    int i;

    for (i = 0 ; i < MAXNODES ; i++) {
      if (node_list->node[i].nodeid == id) {
	    return i ;
      }
    }
    return -1 ;
}




/*
 * Function:
 *	clm_node_reachable
 *
 * Description:
 *	Given a node ID, return TRUE iff CLLOCKD on that node is responding.
 *	This -does not- imply that it has finished all node start processing.
 */

bool_t
clm_node_reachable(int nodeid)
{
    int     ndx;


    ndx = NODEID_TO_INDEX(nodeid) ;
    return( (ndx >= 0) && (node_list->node[ ndx ].version != 0) );
}




/*
 * Function:
 *	clm_id_to_addr
 * Description:
 *	Given a node id, return the most suitable address to connect to it
 *	from the local node.
 * Arguments:
 *	Node id.  (not an index)
 * Return value:
 *	A ulong value that can be slapped into a struct sockaddr_in that
 *	corresponds to the best address through which to connect to the
 *	specified cluster node.
 *	Returns 0 if cluster info not init'd yet or if there is no way to
 *	connect to the specified cluster node.
 * Side effects:
 *	NONE
 */

ulong
clm_id_to_addr(int id)
{
    int ndx ;

    if ( ! clm_config_inited())
      return 0 ;
    ndx = NODEID_TO_INDEX(id) ;
    if (ndx < 0)
      return 0 ;

#ifdef __linux__
    return node_list->node[ndx].saddr.s_addr ;
#else
    return node_list->node[ndx].saddr ;
#endif /* __linux__ */
}

/* clm_getlocalid
 *
 * This function did not exist prior to the Linux port.  It was
 * listed in the externs, etc., but had no implementation.
 * 
 * Sets *id to to be the value of the local node ID.  If the DLM is not
 * yet initialized (see RETURNS, below), the value if *id is undefined.
 *
 * RETURNS
 * 
 * If the DLM is not initialized and the local node ID is not yet known,
 * then DLM_SYSERR is returned.  Otherwise, DLM_NORMAL is returned.
 */

int
clm_getlocalid( int * id )
{
    if ( ! clm_config_inited())
    {
	return( DLM_SYSERR );
    }
    else
    {
	*id = OurSiteID;
	return( DLM_NORMAL );
    }
}




/****************************************************************************/
/* dlm_nodeid_iterator_create - Create a new dlm_nodeid_iterator object
 *
 * Creates and initializes a new dlm_nodeid_iterator object.  This type of
 * iterator will iterate over the node IDs that are currently part of the
 * cluster.  That is, each time iterator_next is called, the data pointer
 * will be set to the next node ID in the cluster.
 *
 * Unlike with other iterators, the data value from iterator_next is an
 * integer and NOT A POINTER!
 *
 * RETURNS
 *
 * A pointer to a new iterator object on success, or NULL on failure.
 */

struct iterator *
dlm_nodeid_iterator_create( void )
{
    struct dlm_nodeid_iterator * iter;
    
    
    iter = CLM_MALLOC( sizeof( struct dlm_nodeid_iterator ) );
    if ( iter != NULL )
    {
	iter->base.methods = & dlm_nodeid_iterator_methods;
	iter->next_index = 0;
    }


    return( (iterator_t *) iter );
}




/****************************************************************************/
/* dlm_nodeid_iterator_next - Advance to the next element in a dlm_nodeid_iterator
 *
 * Advances to the next element in a dlm_nodeid_iterator.  This function must
 * NEVER be called directly.  It should only be called via the
 * 'methods->next' pointer in an iterator_t object (i.e., by calling
 * iterator_next).
 * 
 * RETURNS
 * 
 * If the iterator has been exhausted, TRUE is returned and *data is
 * undefined.  Otherwise, FALSE is returned and *data is set to the next
 * node ID.
 */

static bool_t
dlm_nodeid_iterator_next( iterator_t * iter, void ** data )
{
    bool_t   done;
    const struct node_item * node;
    struct dlm_nodeid_iterator * ni_iter;


    done = TRUE;
    ni_iter = (struct dlm_nodeid_iterator *) iter;
    
    if ( ni_iter->next_index < MAXNODES )
    {
	node = & node_list->node[ ni_iter->next_index ];
	if ( node->nodeid != 0 )
	{
	    *data = (void *) node->nodeid;
	    ni_iter->next_index++;
	    done = FALSE;
	}
    }


    return( done );
}




/****************************************************************************/
/* dlm_nodeid_iterator_destroy - Destroy a dlm_nodeid_iterator object
 *
 * Frees all of the resources associated with a dlm_nodeid_iterator object.
 * This function must NEVER be called directly.  It should only be called
 * via the 'methods->destroy' pointer in an iterator_t object (i.e., by
 * calling iterator_destroy).
 */

static void
dlm_nodeid_iterator_destroy( iterator_t * iter )
{
    if ( iter != NULL )
    {
	CLM_FREE( iter );
    }
}



/****************************************************************************/

#ifndef module_init
# define module_init( func ) int init_module( void ) { return( func () ); }
# define module_exit( func ) void cleanup_module( void ) { func (); }
#endif

module_init( clm_info_init )
module_exit( clm_info_exit )
