/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 * Contributions:
 *  >2.2 proc support: Alan Jones, alan.jones@sun.com
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_kerndd.c,v 1.17 2001/10/19 22:42:38 idr Exp $"

/*
 * Distributed Lock Manager for Linux - device driver module
 *
 * This is the base file that makes up the kernel portion, i.e.,
 * the 'device driver' for the DLM.  All of the "real" locking
 * mechanisms are contained in the overall kernel code, this
 * module acts as the "gateway" for data coming from and going
 * to user space and clients:
 * - dlmdu - the user-side administration representative for
 *  the DLM.  Handles configuration and startup of the DLM, and
 *  monitors the state of the cluster nodes (if we have a user-side
 *  cluster manager) and shuts down the DLM if told to.
 *  - data comes through write() calls on the device "admin"
 *      device.
 *  - we only allow one open() call against the "admin" device.
 *  - if the open admin() device is closed, we shut down all
 *      locking services.
 *
 * - libclm - the user-side locking API library.  Used by locking
 *  clients.  Parses API calls, packages the data and passes it
 *  through the through the "locks" device.  Also receives AST
 *  blocks from the kernel to report results back to clients.
 *  - data is passed to the kernel via write() calls on "locks".
 *  - data is passed to the clients via read() calls on "locks".
 *
 * This module is essentially just the "catcher" for all of the
 * traffic, and doesn't really do the core DLM work.  Take a look
 * at everything else for that!  Follow the calls made from this
 * module and you'll see the wonders unfold.
 */

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: haDLM_open
 *      haDLM_release
 *      haDLM_read
 *      haDLM_write
 *      haDLM_read_proc
 *      init_module
 *      cleanup_module
 *      proc_print_vdevice
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/fs.h>		/* File operations */
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/ioctl.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/timer.h>
#include <asm/uaccess.h>

#include "dlm.h"
#include "ast.h"
#include "ast_stub.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "pti.h"
#include "clm_data.h"			/* external data definitions */
#include "dlm_kernel.h"
#include "clm_tune.h"
#include "dlm_devmsg.h"
#include "dlm_workqueue.h"
#include "cltrc.h"

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/*
 * We expect to receive a number of pieces of information from
 * the user-side when this module is loaded:
 * - Our major number to use.  We expect that dlmdu will
 *  take care of /dev entries.
 * - Name of the DLM.  Use this in messages we print in the
 *  system log.
 * - Minor numbers for 'admin' and 'locks' devices.
 */

int     haDLM_major_number = 0;         /* Default major number */
char   *haDLM_name = NULL;              /* Name of our device */
int     haDLM_admin_minor = 0;          /* Minor number for "admin" device */
int     haDLM_locks_minor = 0;          /* Minor number for "locks" device */

MODULE_AUTHOR("Peter Wombat");
MODULE_DESCRIPTION("The ha Distributed Lock Manager device.");
MODULE_PARM(haDLM_major_number, "i");	/* an int */
MODULE_PARM_DESC(haDLM_major_number, "Incoming major number to use.");
MODULE_PARM(haDLM_name, "s");	/* a string */
MODULE_PARM_DESC(haDLM_name, "Incoming device name to use.");
MODULE_PARM(haDLM_admin_minor, "i");	/* an int */
MODULE_PARM_DESC(haDLM_admin_minor, "Device minor for admin device.");
MODULE_PARM(haDLM_locks_minor, "i");	/* an int */
MODULE_PARM_DESC(haDLM_locks_minor, "Device minor for locks device.");

/* ------------------------------------------------------------ */
/*
 * Method prototypes for the functions that interact with the
 * clients.
 */

static int haDLM_open(struct inode *inode,
                      struct file *file);
static int haDLM_release(struct inode *inode,
                         struct file *file);
static int haDLM_read(struct file *file,
                      char *buf,
                      size_t lbuf,
                      loff_t * ppos);
static int haDLM_write(struct file *file,
                       const char *buf,
                       size_t lbuf,
                       loff_t * ppos);
static int haDLM_ioctl(struct inode *inode,
                       struct file *file,
                       unsigned int cmd,
                       unsigned long arg);

/* ------------------------------------------------------------ */
/*
 * Write is supported on both 'admin' and 'locks'.  Read is
 * really just used on 'locks' (to get AST blocks.)  To "read"
 * admin and status data, use the /proc entry.
 */

struct file_operations haDLM_fops = {
  open:	haDLM_open,
  release: haDLM_release,
  read: haDLM_read,
  write: haDLM_write,
  ioctl: haDLM_ioctl
};


/* ------------------------------------------------------------ */
/*
 * Our virtual "locks" device definition, one per open call.
 *
 * Note on client pid management:
 * We use this device to manage client opens, before the client
 * becomes a full-fledged client.  In essence, this identifies
 * the thread that issues the open() against the DLM device.
 * If the admin minor, fine, that should be dlmdu.  If it is
 * the locks minor, it is a client.
 *
 * We record a pid here because we will need it.  On Linux,
 * in a multi-threaded client each thread has its own pid, so
 * if multiple threads make calls through the same device, if
 * we use current->pid to figure out the client, we won't get
 * the same value, and there is much code spread around that
 * uses this pid.  So we record it on the open, which creates
 * a locks device instance, and then for every call through
 * that device we pick up the saved pid, and ignore exactly
 * which thread is actually calling in, as there is nothing
 * in the interface which limits the client to using the same
 * thread every time.
 */

typedef struct our_vdevice {
  dlm_listnode_t   link;
  pid_t     device_open_pid;            /* For reference information */
  struct file  *device_filp;            /* Point back to file struct */
  /* locks, stats, etc. */
} our_vdevice_t;

static DECLARE_DLM_LIST( haDLM_devices_anchor );

/*
 * Work queue for delivering work to the kernel thread.
 */
extern dlm_workqueue_t * dlm_master_work_queue;

/* ------------------------------------------------------------ */
/* Counters */

static int haDLM_devices_count = 0;
static int haDLM_devices_cumulative_count = 0;

static int haDLM_open_counter = 0;
static int haDLM_admin_counter = 0;
static int haDLM_allow_locks = 0;

static unsigned long jiffies_per_second;

/* ------------------------------------------------------------ */
/*
 * Keep track of admin client.
 */
static our_vdevice_t *haDLM_admin_device = NULL;

/* ------------------------------------------------------------ */
/*
 * Routines to handle the "details" of the various ioctl
 * calls.
 */
static int handle_ioctl_sendtran(our_vdevice_t *_device,
                                 int cmd,
                                 unsigned int arg);
static int handle_ioctl_recvtran(our_vdevice_t *_device,
                                 int cmd,
                                 unsigned int arg);
static int handle_ioctl_groupop(our_vdevice_t *_device,
                                int cmd,
                                unsigned long arg);
static int handle_ioctl_setdebug(our_vdevice_t *_device,
                                 int cmd,
                                 unsigned long arg);
static int handle_ioctl_astget(our_vdevice_t *_device,
                               int cmd,
                               unsigned long arg);
static int handle_ioctl_globstat(our_vdevice_t *_device,
				 int cmd,
				 unsigned long arg);

/* ------------------------------------------------------------ */
/*
 * Single place to do allocations, allows us to keep error checks
 * and such in one place.  Return NULL if problems allocating.  We
 * use name to print an error message if we have any problems.
 *
 * For future, allows easy ability to do centralized control and
 * monitoring of memory management, valuable thing to have...
 */

static inline void *
kmalloc_something(unsigned int bytes,
                  char *name)
{
  void *_data;

  if (NULL ==
      (_data = kmalloc(bytes, GFP_KERNEL))) {
    printk("[%s] Error allocating [%s]\n",
           haDLM_name,
           name);
  } else {
    printk("[%s] Allocated [%d/0x%p/%s]\n",
           haDLM_name,
           bytes, _data, name);
  }
  return(_data);
}

static inline void
kfree_something(void *ptr,
                char *name)
{
  printk("[%s] Free [0x%p/%s]\n",
         haDLM_name,
         ptr, name);
  kfree(ptr);
}

static int
kfree_topo_block(void *ptr)
{
  bsdlog(LOG_INFO,
         "[%s] freeing topology block from workunit [%p]\n",
         haDLM_name,
         ptr);
  kfree(ptr);
  return(0);
}

/* ============================================================ */
/*
 * Our init routine.  Called when the module is loaded.
 *
 * Unless we have had a major number passed in on the module load,
 * which will be reflected in the haDLM_major_number field, then
 * dynamically get one, and hope the /dev/haDLM entry gets set up
 * properly.
 *
 * NOTE: undecide - do start_lockd() call here, or wait until we
 * receive the first message (in haDLM_write()) from dlmdu?
 */
int
init_module(void)
{
  int	_i;

  if ( haDLM_name == NULL ) {
    bsdlog( LOG_ERR, "haDLM_name not set by module loader.  Terminating.\n" );
    return( -1 );
  }
    
  printk("[%s] device [%i/%s], init_module loaded at 0x%p\n",
         haDLM_name,
         haDLM_major_number,
         haDLM_name,
         init_module);

  _i = register_chrdev(haDLM_major_number,
                       haDLM_name,
                       &haDLM_fops);
  if (-1 == _i) {
    printk("[%s] Error %i trying to register device [%d/%s]\n",
           haDLM_name,
           _i,
           haDLM_major_number,
           haDLM_name);
    return(_i);
  } else if (0 == haDLM_major_number) {
    haDLM_major_number = _i;
  }

  /*
   * Account for possible clock changes in jiffies maintenance.
   */

  jiffies_per_second = HZ;

#ifdef CONFIG_PROC_FS
  dlm_proc_start( haDLM_name );
#endif  /* CONFIG_PROC_FS */

  printk("[%s] Success, register device [%d/%s]\n",
         haDLM_name,
         haDLM_major_number,
         haDLM_name);

  return(0);
}

/* ------------------------------------------------------------ */
/*
 * Cleanup time!  Being unloaded, release memory etc.  However,
 * we shouldn't have any, as we shouldn't get unloaded if we
 * have active users, right?  They would all have to go away
 * before the unload, I believe!
 */

void
cleanup_module(void)
{
  int  _i;

  printk("[%s] unloading [%d/%s], cleanup_module loaded at 0x%p\n",
         haDLM_name,
         haDLM_major_number,
         haDLM_name,
         cleanup_module);

  if ( ! dlm_list_is_empty( & haDLM_devices_anchor ) ) {
    printk("[%s] unloading module, have %i devices\n",
           haDLM_name,
           haDLM_devices_count);
  }

  _i = unregister_chrdev(haDLM_major_number,
                         haDLM_name);

  if (-1 == _i) {
    printk("[%s] Error %i trying to unregister device [%d/%s]\n",
           haDLM_name,
           _i,
           haDLM_major_number,
           haDLM_name);
  } else {
    printk("[%s] Unregistered device [%i/%s]\n",
           haDLM_name,
           haDLM_major_number,
           haDLM_name);
  }

#ifdef CONFIG_PROC_FS
    dlm_proc_cleanup();
#endif
}

/* ======================================================= */
/*
 * A client is opening /dev/haDLM/admin or /dev/haDLM/locks
 * (note the actual names can be modified via config.)
 * Create a "unique" instance and store the information in
 * the struct file * file->private_data field, to
 * differentiate each open.
 *
 * We allow only a single open against the "admin" port, so,
 * if this is a subsequent attempt, return EAGAIN.
 *
 * If we haven't received the initialization info from the
 * dlmdu, then do not allow locks opens to succeed.  In
 * this case return EAGAIN and let them loop back later.
 *
 * Not much to do here regarding "DLM" function, as we do
 * not get any "real" data here.  That comes in via writes
 * for both admin and locks.
 */

static int
haDLM_open(struct inode *inode,
           struct file *filp)
{
  static int counter = 0;
  our_vdevice_t *_device;
  int major, minor;

  major = MAJOR(inode->i_rdev);
  minor = MINOR(inode->i_rdev);

  printk("[%s] OPEN, major/minor/inode/filp [%d/%d/0x%p/0x%p]\n",
         haDLM_name,
         major, minor,
         inode,
         filp);

  haDLM_open_counter++;
  counter++;

  printk("[%s] OPEN: opens: total/current[%d/%d]\n",
         haDLM_name,
         counter,
         haDLM_open_counter);

  if (haDLM_admin_minor == minor) {
    if (0 == haDLM_admin_counter) {
      if (NULL == 
          (_device = kmalloc_something(sizeof(our_vdevice_t),
                                       "our_vdevice_t"))) {
        return(-ENOMEM);
      }

      haDLM_admin_counter++;

      uproc_pid = current->pid;
      _device->device_open_pid = uproc_pid;
      _device->device_filp = filp;
      _device->link.succ = NULL;
      _device->link.pred = NULL;
      filp->private_data = _device;

      haDLM_admin_device = _device;

      printk("[%s] OPEN admin by pid[%d]: opens: total/current[%d/%d]\n",
             haDLM_name,
             haDLM_admin_device->device_open_pid,
             counter,
             haDLM_admin_counter);

      MOD_INC_USE_COUNT;
      return(0);
    } else {
      printk("[%s] OPEN admin: already open in filp[0x%p] by pid[%d]\n",
             haDLM_name,
             haDLM_admin_device,
             haDLM_admin_device->device_open_pid);
      return(-EAGAIN);
    }
  }

  /*
   * Sanity check, it wasn't admin, so, must be...
   */
  if (haDLM_locks_minor != minor) {
    bsdlog(LOG_NOTICE,
           "[%s] open request against unrecognised minor [%d]\n",
           haDLM_name,
           minor);
    return(-EINVAL);
  }

  /*
   * OK, it is a locks request.  Have we received the relevant
   * initialiation info?
   */
  if (!haDLM_allow_locks) {
    /* Nope */
    printk("[%s] OPEN locks: not yet initialized.\n",
           haDLM_name);
    return(-EAGAIN);
  }

  /*
   * Allocate a virtual device instance, and tie it to this file
   * struct.  Also, insert it at the head of the global chain of
   * virtual devices.
   */
  
  if (NULL == 
      (_device = kmalloc_something(sizeof(struct our_vdevice),
                                   "our_vdevice_t"))) {
    return(-ENOMEM);
  }

  _device->device_open_pid = current->pid;
  _device->device_filp = filp;
  filp->private_data = _device;

#ifdef DEBUG
  if ( dlm_list_is_empty( & haDLM_devices_anchor ) ) {
    if (0 != haDLM_devices_count) {
      printk("[%s] BAD: device list empty but count %d\n",
             haDLM_name,
             haDLM_devices_count);
    }
  }
#endif /* DEBUG */

  dlm_list_add_head( & haDLM_devices_anchor, (dlm_listnode_t *) _device );

  haDLM_devices_count++;
  haDLM_devices_cumulative_count++;

  printk("[%s] New vdevice new/count/cumulative[0x%p/%d/%d]\n",
         haDLM_name,
         _device,
         haDLM_devices_count,
         haDLM_devices_cumulative_count);

  MOD_INC_USE_COUNT;
  return(0);
}

/* ------------------------------------------------------------ */
/*
 * A close was issued against admin or locks.  If it is admin,
 * then we need to shut down all locking function and nuke our
 * clients.  If it is locks, we may need to do recovery if that
 * client had any active locks.
 *
 * Determine what was just closed, and interact with all of the
 * right DLM routines to do cleanup.
 *
 * NOTE: Our admin channel is important if we are getting cluster
 * manager (node up/down etc.) updates through it.  If we are not,
 * then, do we care?  Let it restart (hopefully) but in the
 * interim just keep going.  Note that if it received a signal
 * it could catch and exited for that reason, it would have
 * told us (via a write) before exiting/closing.  So, it either
 * got an uncatchable signal, or, it has a bug...  For now, we
 * only have user-space cluster managers, so, we will shut down.
 */

static int
haDLM_release(struct inode *inode,
              struct file *filp)
{
  our_vdevice_t *_device;  
  int major, minor;

  major = MAJOR(inode->i_rdev);
  minor = MINOR(inode->i_rdev);

  haDLM_open_counter--;

  _device = (struct our_vdevice *)filp->private_data;

  printk("[%s] CLOSE: inode/fileptr/vdev/[0x%p/0x%p/0x%p]\n",
         haDLM_name,
         inode,
         filp,
         _device);

  if (haDLM_admin_device == _device) {
    /*
     * Bad news, we will have to shut down the DLM, our
     * user-side client has gone away.
     */
    printk("[%s] Admin client pid[%d] closed device!  Shut down.\n",
           haDLM_name,
           haDLM_admin_device->device_open_pid);
    /*
     * shut down kernel thread
     * kill clients
     * release resources
     * kernel will unload us once we've cleaned up
     */
    haDLM_allow_locks = 0;

    proc_notify(uproc_pid);

    kfree_something(haDLM_admin_device, "our_vdevice_t");
    haDLM_admin_device = NULL;
    haDLM_admin_counter--;

    /* Notify the DLM kernel thread that it is time for it to shutdown.
     */

    dlm_start_shutdown();

  } else {
	
    /*
     * A client has exited or closed.  Since we don't export to
     * them the file descriptor (remains hidden in library) this
     * usually indicates the process died, however, it could also
     * be that they closed the wrong file descriptor.  No matter,
     * as all of their locks will now be destroyed.
     */

#ifdef DEBUG
    {
      dlm_listnode_t * curr;

      for ( EACH_DLM_LISTNODE( & haDLM_devices_anchor, curr ) ) {
	if ( curr == (dlm_listnode_t *) _device ) {
	  break;
	}
      }

      if ( curr != (dlm_listnode_t *) _device ) {
        bsdlog(LOG_CRIT,
	       "[%s] client died [%u], vdevice[%p] unknown!\n",
	       haDLM_name,
	       _device->device_open_pid,
	       _device);
	return(0);
      }
    }
#endif /* DEBUG */

    dlm_list_remove( (dlm_listnode_t *) _device );

    bsdlog(LOG_INFO,
	   "[%s] client died, vdevice[%p] pid[%d]\n",
	   haDLM_name,
	   _device,
	   _device->device_open_pid);

    /*
     * Tell the kernel thread that a client is dead, and needs
     * to be cleaned up.
     */
    proc_notify(_device->device_open_pid);

    kfree_something(_device, "our_vdevice_t");
    haDLM_devices_count--;

    printk("[%s] Now have %d virtual device%s\n",
           haDLM_name,
	   haDLM_devices_count,
	   ((1 == haDLM_devices_count) ? "," : "s,") );
  }


  MOD_DEC_USE_COUNT;

  return(0);
}

/* ------------------------------------------------------------ */
/*
 * So far I have this here, although I don't have much anything
 * to actually be read.  All API requests are handled through
 * ioctl calls (see haDLM_ioctl()), thus, any reads against the
 * non-admin device are summarily rejected.
 *
 * As to reading the admin, not yet sure what to send, if we
 * don't figure anything out, drop this routine.
 */

static int
haDLM_read(struct file *filp,
           char *buf,
           size_t lbuf,
           loff_t * ppos)
{
  int	_bytes;
  char *_data_buffer;
  struct our_vdevice *_device;  

  _device = (struct our_vdevice *)filp->private_data;

  bsdlog(LOG_INFO,
         "[%s] Entering READ, size/offset[%d/0x%p]\n",
         haDLM_name,
         lbuf,
         ppos);

  if (haDLM_admin_device != _device) {
    bsdlog(LOG_NOTICE,
           "[%s] Attempt to read locks device/owner[%p/%d] by pid[%d]\n",
           haDLM_name,
           _device,
           _device->device_open_pid,
           current->pid);
    return(-EACCES);
  }

  if (NULL ==
      (_data_buffer = kmalloc(lbuf, GFP_KERNEL))) {
    printk("[%s] READ, cannot allocate memory buffer!\n",
           haDLM_name);
    return(-ENOMEM);
  }
  memset(_data_buffer, '1', lbuf);      /* Random pattern for now. */
    
  if (0 != (_bytes = copy_to_user(buf, _data_buffer, lbuf))) {
    printk("[%s] Error on READ, bytes left %i\n",
           haDLM_name,
           _bytes);
    _bytes = -EFAULT;
  } else {
    printk("[%s] READ, sent out %i bytes\n",
           haDLM_name,
           lbuf);
  }
  kfree(_data_buffer);

  return _bytes;
}

/* ------------------------------------------------------------ */
/*
 * Ah, good stuff coming in!
 * - admin device:
 *  - initialization information, initial cluster node status,
 *      this allows us to open up locks and allow lock requests.
 *  - cluster node status updates - node up/down, IP address
 *      info.
 *  - shut down - signal caught, cluster mgr dead, etc.
 *
 * We only deal with writes on the admin device, never on any
 * of the locks devices, i.e., all API client requests are
 * done via ioctl calls (see haDLM_ioctl()).
 *
 * We essentially parse the input, then call the relevant DLM
 * routines and let them worry about it from there.
 *
 * There is no partial success here, if there is any problem with
 * the input buffer, we simply return an error and do nothing.
 *
 * Input: various command buffers
 *
 * Output: 0, if buffer contains valid command
 *         -EINVAL otherwise
 *
 * Incoming buffer always of form:
 *  int (4 bytes): command code
 *  rest of buffer: information, varies based on command code
 */

static int
haDLM_write(struct file *filp,
            const char *buf,
            size_t lbuf,
            loff_t * ppos)
{
  int	_bytes, _rc;
  char *_data_buffer;
  struct our_vdevice *_device;
  dlm_MSG_t *message;
  dlm_MSG_types_t command;
  MQ_DLM_TOP_INFO_t *topo;
  dlm_workunit_t *wu;

  _rc = lbuf;

  _device = (struct our_vdevice *)filp->private_data;

  printk("[%s] WRITE, size/buf/filp/vdev[%d/0x%p/0x%p/0x%p]\n",
         haDLM_name,
         lbuf,
         buf,
         filp,
         _device);

  /*
   * A locks device, never accept data written through here, data
   * must come in via ioctl calls.
   */
  if (haDLM_admin_device != _device) {
    bsdlog(LOG_NOTICE,
           "[%s] Attempt to write non-admin device/owner[%p/%d] by pid[%d]\n",
           haDLM_name,
           _device,
           _device->device_open_pid,
           current->pid);
    return(-EACCES);
  }

  if (NULL ==
      (_data_buffer = kmalloc_something(lbuf,
                                        "WRITE data buffer"))) {
    return(-ENOMEM);
  }

  if (0 != (_bytes = copy_from_user(_data_buffer, buf, lbuf))) {
    printk("[%s] Error on WRITE, bytes left %i\n",
           haDLM_name,
           _bytes);
    return(-EFAULT);
  } else {
    printk("[%s] WRITE, brought in %i bytes from user\n",
           haDLM_name,
           lbuf);
  }

  message = (dlm_MSG_t *)_data_buffer;
  command = message->msg_code;

  /*
   * We need to finalise the interfaces between this and the
   * global event queue, embodied in clm_main.c.  Then the
   * data that comes into here will be passed there and
   * queued for handling, e.g., node events will be handled
   * and cause all kinds of exciting things to happen.
   */
  switch(command) {
  case haDLM_MSG_Init:
    /*
     * Since start_lockd() wants the argument block, it works
     * well to do it here, where we can pass it in easily.
     */
    printk("[%s] INIT message received\n",
           haDLM_name);

    kproc_pid = start_lockd(message->arg_block,
                            message->msg_ptr_size);
    if (ERROR == kproc_pid) {
      bsdlog(LOG_INFO,
             "[%s] cannot start kernel thread.\n",
             haDLM_name);
      _rc = -ESRCH;
    } else {
      bsdlog(LOG_INFO,
             "[%s] dlmdk kthread started, pid [%d]\n",
             haDLM_name,
             kproc_pid);
    }
    haDLM_allow_locks = 1;
    break;
  case haDLM_MSG_Stop:
    printk("[%s] STOP message received\n",
           haDLM_name);
    break;
  case haDLM_MSG_CMGR:
    printk("[%s] CMGR message received\n",
           haDLM_name);
    if (NULL ==
        (topo = kmalloc_something(message->msg_ptr_size,
                                  "CMGR topology block"))) {
      return(-ENOMEM);
    }

    if (0 != (_bytes = copy_from_user(topo,
                                      message->topo_array,
                                      message->msg_ptr_size))) {
      bsdlog(LOG_NOTICE,
             "[%s] cannot copy in topo block, vdevice[%p] pid[%d]\n",
             haDLM_name,
             _device,
             current->pid);
      kfree_something(topo, "CMGR topology block");
      return(-EFAULT);
    } else {
      bsdlog(LOG_INFO,
             "[%s] copied in topo block, %i bytes from user\n",
             haDLM_name,
             lbuf);
    }
    if (NULL ==
        (wu = kmalloc_something(sizeof(dlm_workunit_t),
                                "WorkUnit for topology block"))) {
      return(-ENOMEM);
    }
    wu->data = topo;
    wu->free_data = kfree_topo_block;
    wu->type = MQ_DLM_TOP_INFO_MSG;
    dlm_workqueue_put_work(dlm_master_work_queue, wu);
    break;
  default:
    printk("[%s] unknown command [%d], ignoring it\n",
           haDLM_name,
           command);
    _rc = -EINVAL;
  }

  kfree_something(_data_buffer,
                 "WRITE data buffer");

  printk("[%s] WRITE exit code [%d]\n",
         haDLM_name,
         _rc);

  return(_rc);
}

/* ------------------------------------------------------------ */
/*
 * Support the kill monitoring interfaces for IOCTL calls.
 */

static int
haDLM_ioctl(struct inode *inode,
            struct file *filp,
            unsigned int cmd,
            unsigned long arg)
{
  int _rc;
  int major, minor;
  our_vdevice_t *_device;

  if (_IOC_TYPE(cmd) != DLM_IOCTL_BASE) {
    return(-EINVAL);
  }
  major = MAJOR(inode->i_rdev);
  minor = MINOR(inode->i_rdev);

  if (haDLM_admin_minor == minor) {
    bsdlog(LOG_NOTICE,
           "[%s] ioctl call (cmd %d) against DLM admin device, ignore\n",
           haDLM_name,
           cmd);
    return(-EINVAL);
  }

  _device = (struct our_vdevice *)filp->private_data;

  /*
   * These are ordered in what should be the approximate order
   * of most common to least common, at least, we believe at
   * this point.
   */
  switch (cmd) {
  case DLM_IOCTL_SENDTRAN:
    /*
     * Client is sending in a lock request.
     */
    _rc = handle_ioctl_sendtran(_device, cmd, arg);
    break;
  /* ------------------------------------------------------------ */
  case DLM_IOCTL_RECVTRAN:
    /*
     * Client wants to retrieve a response block for a lock request.
     */
    _rc = handle_ioctl_recvtran(_device, cmd, arg);
    break;
  /* ------------------------------------------------------------ */
  case DLM_IOCTL_ASTGET:
    /*
     * Retrieve the next AST block for the user.
     */
    _rc = handle_ioctl_astget(_device, cmd, arg);
    break;
  /* ------------------------------------------------------------ */
  case DLM_IOCTL_GROUPOP:
    /*
     * Issue a lock group operation.
     */
    _rc = handle_ioctl_groupop(_device, cmd, arg);
    break;
  /* ------------------------------------------------------------ */
  case DLM_IOCTL_SETDEBUG:
    /*
     * Set the lock manager's debug level.
     */
    _rc = handle_ioctl_setdebug(_device, cmd, arg);
    break;
  /* ------------------------------------------------------------ */
  case DLM_IOCTL_GLOBSTAT:
    /*
     * Return a "count" of resources and locks to client.
     */
    _rc = handle_ioctl_globstat(_device, cmd, arg);
    break;
  /* ------------------------------------------------------------ */
  default:
    bsdlog(LOG_INFO, "[%s] unrecognized ioctl command %d\n",
           haDLM_name,
           cmd);
    return(-EINVAL);
  }

  return(_rc);

}

/* ============================================================ */
/*
 * Deal with the different ioctl calls.  Break out to help
 * keep the code more readable.
 */

/*
 * Client is sending in a lock request.
 *
 * Input:
 *  device instance of client making request
 *  ioctl cmd
 *  ioctl argument length - argument is a transaction control block,
 *  which points to client's transaction and length.
 *
 * Output:
 *  return code from k_clm_send
 *
 * Side effect:
 *  Kernel thread will take the request and do something with it.
 *  Side effects could therefore be major, i.e., messages sent to
 *  nodes, locks established, all kinds of good stuff.
 */
static int
handle_ioctl_sendtran(our_vdevice_t *_device,
                      int cmd,
                      unsigned int arg)
{
  int _bytes;
  int _rc;
  sendtran_t _tran;

  if (sizeof(sendtran_t) != _IOC_SIZE(cmd)) {
    bsdlog(LOG_NOTICE,
           "[%s] DLM_IOCTL_SENDTRAN, input wrong size %d\n",
           haDLM_name,
           _IOC_SIZE(cmd));
    _rc = -EINVAL;
    return(_rc);
  }
  if (0 != (_bytes = copy_from_user(&_tran,
                                    (sendtran_t *)arg,
                                    _IOC_SIZE(cmd)))) {
    bsdlog(LOG_NOTICE, 
           "[%s] DLM_IOCTL_SENDTRAN, cannot get control block\n",
           haDLM_name);
    _rc = -EFAULT;
    return(_rc);
  }
  bsdlog(LOG_INFO,
         "[%s] DLM_IOCTL_SENDTRAN, get trans from [%p]\n",
         haDLM_name,
         _tran.buf);

  _rc = k_clm_send(_tran.buf,
                   _tran.len,
                   _device->device_open_pid);

  bsdlog(LOG_INFO,
         "[%s:%s] DLM_IOCTL_SENDTRAN finished, exit code [%d] pid [%d]\n",
         current->comm, __FUNCTION__,
         _rc, _device->device_open_pid);

  return(_rc);
}

/* ------------------------------------------------------------ */
/*
 * Client wants to retrieve a response block for a lock request.
 *
 * Input:
 *  device instance of client making request
 *  ioctl cmd
 *  ioctl argument length - argument points to a transaction control
 *  block, tells us a sequence number and where to put response.
 *
 * Output:
 *  return code from k_clm_recv
 *
 * Side effect:
 *  Response copied out to user address space.
 */
static int
handle_ioctl_recvtran(our_vdevice_t *_device,
                      int cmd,
                      unsigned int arg)
{
  int _bytes;
  int _rc;
  recvtran_t _tran;

  if (sizeof(recvtran_t) != _IOC_SIZE(cmd)) {
    bsdlog(LOG_NOTICE,
           "[%s] DLM_IOCTL_RECVTRAN, input wrong size %d\n",
           haDLM_name,
           _IOC_SIZE(cmd));
    _rc = -EINVAL;
    return(_rc);
  }
  if (0 != (_bytes = copy_from_user(&_tran,
                                    (recvtran_t *)arg,
                                    _IOC_SIZE(cmd)))) {
    bsdlog(LOG_NOTICE, 
           "[%s] DLM_IOCTL_RECVTRAN, cannot get control block\n",
           haDLM_name);
    _rc = -EFAULT;
    return(_rc);
  }
  bsdlog(LOG_INFO,
         "[%s] DLM_IOCTL_RECVTRAN, send trans to [%p], seq [%d]\n",
         haDLM_name,
         _tran.buf,
         _tran.seq);

  _rc = k_clm_recv(_tran.buf,
                   _tran.len,
                   _tran.seq,
                   _device->device_open_pid);

  bsdlog(LOG_INFO,
         "[%s:%s] DLM_IOCTL_RECVTRAN finished, exit code [%d] pid [%d]\n",
         current->comm, __FUNCTION__,
         _rc, _device->device_open_pid);

  return(_rc);
}

/* ------------------------------------------------------------ */
/*
 * Retrieve the next AST block for the user.
 *
 * Input:
 *  device instance of client making request
 *  ioctl cmd
 *  ioctl argument length - argument is user address for AST
 *
 * Output:
 *  return code from AST fetch routines
 *
 * Side effect:
 *  The client's next pending AST copied into its address space
 *  for execution.  If none, then nothing copied.
 */
static int 
handle_ioctl_astget(our_vdevice_t *_device,
                    int cmd,
                    unsigned long arg)
{
  int _rc;

  if (sizeof(struct poll_info) != _IOC_SIZE(cmd)) {
    bsdlog(LOG_NOTICE,
           "[%s] DLM_IOCTL_ASTGET, input wrong size %d\n",
           haDLM_name,
           _IOC_SIZE(cmd));
    _rc = -EINVAL;
    return(_rc);
  }
  bsdlog(LOG_INFO,
         "[%s] DLM_IOCTL_ASTGET, write AST to [%p]\n",
         current->comm,
         (struct poll_info *)arg);

  _rc = k_ASTget((struct poll_info *)arg,
                 _device->device_open_pid);

  bsdlog(LOG_INFO,
         "[%s:%s] DLM_IOCTL_ASTGET finished, exit code [%d] pid [%d]\n",
         current->comm, __FUNCTION__,
         _rc, _device->device_open_pid);

  return(_rc);
}

/* ------------------------------------------------------------ */
/*
 * Set the lock manager's debug level.
 *
 * Input:
 *  device instance of client making request
 *  ioctl cmd
 *  ioctl argument length -- argument is debug mask
 *
 * Output:
 *  return code from debug setting routines
 *
 * Side effect:
 *  Potentially lots and lots of debug entries
 */
static int 
handle_ioctl_setdebug(our_vdevice_t *_device,
                      int cmd,
                      unsigned long arg)
{
  int _bytes;
  int _rc;
  int debug_mask;

  if (sizeof(int) != _IOC_SIZE(cmd)) {
    bsdlog(LOG_NOTICE, "[%s] DLM_IOCTL_SETDEBUG, input wrong size %d\n",
           haDLM_name,
           _IOC_SIZE(cmd));
    _rc = -EINVAL;
    return(_rc);
  }
  if (0 != (_bytes = copy_from_user(&debug_mask,
                                    (int *)arg,
                                    _IOC_SIZE(cmd)))) {
    bsdlog(LOG_NOTICE, 
           "[%s] DLM_IOCTL_SETDEBUG, cannot get debug mask\n",
	   haDLM_name);
    _rc = -EFAULT;
    return(_rc);
  }

  _rc = k_set_debug(debug_mask);

  return(_rc);
}

/* ------------------------------------------------------------ */
/*
 * Issue a lock group operation.
 *
 * Input:
 *  device instance of client making request
 *  ioctl cmd
 *  ioctl argument length -- group operation definition block
 *
 * Output:
 *  return code from group operations
 *
 * Side effect:
 *  Depends on the specific command.  See clm_group.c.
 */
static int 
handle_ioctl_groupop(our_vdevice_t *_device,
                     int cmd,
                     unsigned long arg)
{
  int _bytes;
  int _rc;
  struct groupop *gop;

  if (sizeof(groupop_t) != _IOC_SIZE(cmd)) {
    bsdlog(LOG_NOTICE, "[%s] DLM_IOCTL_GROUPOP, input wrong size %d\n",
           haDLM_name,
           _IOC_SIZE(cmd));
    _rc = -EINVAL;
    return(_rc);
  }
  if (NULL == (gop = (kmalloc(sizeof(groupop_t), GFP_KERNEL)))) {
    bsdlog(LOG_WARNING,
           "[%s] DLM_IOCTL_GROUPOP, cannot allocate groupop_t\n",
	   haDLM_name);
    _rc = -ENOMEM;
    return(_rc);
  }
  if (0 != (_bytes = copy_from_user(gop,
                                    (groupop_t *)arg,
                                    _IOC_SIZE(cmd)))) {
    bsdlog(LOG_NOTICE, 
           "[%s] DLM_IOCTL_GROUPOP, cannot get groupop_t\n",
	   haDLM_name);
    _rc = -EFAULT;
    kfree(gop);
    return(_rc);
  }
  /*
   * Now have data, try to do operation.
   */
  _rc = k_clm_group_op(gop->op,
                       gop->gid,
                       gop->flags,
                       _device->device_open_pid);
  kfree(gop);

  return(_rc);
}

/* ------------------------------------------------------------ */
/*
 * Return a block of global statistics.
 *
 * Input:
 *  device instance of client making request
 *  ioctl cmd
 *  ioctl argument length - argument is user address for stats
 *  block
 *
 * Output:
 *  0 - stats copied
 *  -1 - error
 *
 * Side effect:
 *  None
 */
static int 
handle_ioctl_globstat(our_vdevice_t *_device,
                      int cmd,
                      unsigned long arg)
{
  int _rc;

  if (sizeof(struct dlm_stats_buf) != _IOC_SIZE(cmd)) {
    bsdlog(LOG_NOTICE,
           "[%s] DLM_IOCTL_GLOBSTAT, input wrong size %d\n",
           haDLM_name,
           _IOC_SIZE(cmd));
    _rc = -EINVAL;
    return(_rc);
  }
  bsdlog(LOG_INFO,
         "[%s] DLM_IOCTL_GLOBSTAT, write stats to [%p]\n",
	 haDLM_name,
         (struct dlm_stats_buf *)arg);

  _rc = clm_global_stats((struct dlm_stats_buf *)arg, 0);

  return(_rc);
}
