/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_kernel.c,v 1.15 2001/11/02 22:15:35 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: CHECK_RANGE
 *		MSGSTR
 *		abort
 *		acquire_lock
 *		dlm_exit
 *		dlm_start_shutdown
 *		gettimeofday
 *		init_stats
 *		locks_init
 *		read_stats
 *		release_lock
 *		resname_bcopy
 *		stat_add
 *		wait_for_it
 */

#ifndef lint
#endif

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */


#include <linux/types.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <asm/pgtable.h>
#include <asm/semaphore.h>

#include <stdarg.h>                     /* Varargs stuff */

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "cltrc.h"
#include "dlm_kernel.h"
#include "dlm_workqueue.h"
#include <clm_data.h>

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */

#ifndef __linux__
extern void *lock_heap;
#endif /* !__linux__ */

extern void cccp_stop ( void );

/* ============================================================ */


/*
 * Function:
 *	gettimeofday
 *
 * Description:
 *	Get the time since the last boot, down to the closest nanosecond.  This
 * is, of course, different from what the user-land version does, but that's
 * cool because we only use the result to calculate time *differences* anyway.
 *
 * Arguments:
 *	struct timeval *foo;	-- Place to put time
 *	void *bar;			-- Unused (included for compatibility)
 *
 * Return value:
 *	Number of seconds since boot, in case that granularity's good enough.
 *
 * Side effects:
 *	None.
 */

extern void do_gettimeofday(struct timeval *tv);

int
gettimeofday (struct timeval *foo, void *bar)
{
#ifdef __linux__
  do_gettimeofday(foo);
#else  
  curtime(foo);
#endif /* __linux__ */
  foo->tv_usec /= 1000; /* yes, this is actually nsec!!! */
  return(foo->tv_sec);
}

/* ============================================================ */
/*
 * Locking support.
 *
 * Linux Note:  Pids
 *
 * Under AIX, all of the calls here simply took the current
 * thread's pid, and went with it.  This was fine, because on
 * AIX every thread (assuming a multi-thread client) always
 * had the same pid.  But, Linux doesn't do this, so if we
 * try to depend on using current->pid, we could get different
 * values for different threads.
 *
 * This leads us to requiring that the thread to be used be
 * sent in, it is ultimately sent in from the client.  See
 * the discussion in dlm_kerndd.c for more on this.
 */

/* Simple locks array is initialized in locks_init() */
static jlock_t locks[3];

/*
 * locks_init
 *
 * initialize mutex(es) used to interlock access to shared memory data base
 */
void
locks_init(void)
{
  int i;
  int numlocks = sizeof(locks) / sizeof(locks[0]);
    
    for (i = 0 ; i < numlocks ; i++ )
    {
	DLM_MUTEX_INIT( & locks[ i ] );
    }
}

 
/*
 * Function:
 *	acquire_lock
 *
 * Description:
 * Attempt to provide some coverage to detect "nested" lock hits, where
 * a thread tries to grab a lock it owns, this is catastrophic.
 *
 * Arguments:
 *	int n;		-- index of lock to acquire (we only use a few)
 *
 * Return value:
 *	NONE
 *
 * Side effects:
 *	May call longjmp.
 */

void
acquire_lock (int n, int pid)
{
    DLM_MUTEX_LOCK( & locks[ n ] );
}

/*
 * Function:
 *	release_lock
 *
 * Description:
 *	Release a lock acquired with acquire_lock, above.
 *
 * Arguments:
 *	int n;		-- index of lock to release
 *
 * Return value:
 *	NONE
 *
 * Side effects:
 *	NONE
 */

void
release_lock (int n, int pid)
{
    DLM_MUTEX_UNLOCK( & locks[ n ] );
}


/* ============================================================ */
/* dlm_exit - Terminate the DLM kernel processes
 *
 * Terminates the DLM kernel process.  The clean-up function for each sub-part
 * of the DLM (lock entities, ASTs, clients, etc.) is called and all
 * connections to CCCP are severed.  The local site ID is reset to -1, and
 * all memory is released.  When this function completes, there is no DLM.
 *
 * WARNINGS
 *
 * On Linux, it is not enough to simply terminate the DLM!  The DLM kernel
 * modules must also be unloaded BEFORE attempting to re-start the DLM.
 *
 * NOTES
 *
 * This function probably does not belong in this file.  It would probably
 * be better as a static function in clm_main.c.
 */

void
dlm_exit( void )
{
    extern void clm_delete_heap(void);
    extern int OurSiteID;


    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_EXIT, 0x00, 0x00, 0x00, 0x00);
    
    clm_active = FALSE;

    /* Close all connections to CCCP.  Do this very early as this prevents
     * any extra messages from coming in that we don't want to handle.
     */

    dlm_cccp_connection( FALSE );
    dlm_recov_cleanup();

    /* Destroy all lock entities.
     */
	
    dlm_destroy_all_le();


    /* Dink any clients that may still be using the lock manager.
     */

    kill_clients();


    clm_rldb_cleanup();

    ast_cleanup();
    sched_queue_cleanup();
    clm_batch_cleanup();

    clm_sleep(1);
    OurSiteID = -1;
    free_all();
    clm_delete_heap();
    locks_init();

    cccp_stop();


    bsdlog( LOG_NOTICE, MSGSTR(EXIT_WITH_OP, "DLM Exiting\n") );
    _exit(0) ;				/* CROAK ! */
    /* NOTREACHED */
}




/****************************************************************************/

void
dlm_start_shutdown( void )
{
    extern bool_t dlm_is_going_down;
    extern dlm_workqueue_t * dlm_master_work_queue;


    /* We are not executing as the main DLM thread.  In this case we
     * need to tell the main thread to please stop running and shut
     * everything down.
     */

    dlm_is_going_down = TRUE;
    up( & dlm_master_work_queue->sema );
}




/****************************************************************************/
/* abort - Cause an immediate abort of the DLM
 *
 * This function is roughly anagous to the C library abort(3C) call.  On
 * Linux, the result is that a message will be printed with KERN_CRIT priority
 * and the current process will oops.
 */

void
abort( const char * s )
{
    unsigned long wakeup;
    DECLARE_WAIT_QUEUE_HEAD( wait );


    TRCHKGT(CLLOCKD_HKID | T_ENTR_FUNC, T_ABORT, 0x00, 0x00, 0x00, 0x00);

    printk(KERN_CRIT "Aborting (%s)",s);


    /* Sleep for some seconds to make sure that the logs have had a chance
     * to be flushed.
     */

    wakeup = (3 * HZ);
    interruptible_sleep_on_timeout(&wait, wakeup);

#ifdef __linux__
#if defined(BRK_ON_ABORT)
    brkpoint();
#endif
#endif /* !__linux__ */
    BUG();
    /*NOTREACHED*/
}


/*
 * Function:
 *	wait_for_it
 *
 * Description:
 *	Wait for a reply if one hasn't already been queued for this client.
 * This is one of two places (the other is sleep_on_sendq, below) where a
 * client can be interrupted by a signal.  Note that the global lock specified
 * by n is released and reacquired within this routine.  If this were not done,
 * the process that should wake us up will never be able to enter the critical
 * section and livelock would result.  Got it?  Good.
 *
 * Arguments:
 *	int n;			-- global lock to release/reacquire
 *	clm_client_t *cp;	-- client to wait for
 *  int clpid;      -- pid of client to wait for
 *
 * Return value:
 *	>0	if interrupted, caller should retry after handling signal
 *	 0	if successful
 *	<0	if an error occurs
 *
 * Side effects:
 *	May call abort().
 *
 */

int
wait_for_it (int n, clm_client_t *cp, int clpid)
{
  int status;


  /*
   * set timeout function to fire after 1 sec - this is to step around
   * a race condition where the wakeup is being issued before the sleep.
   *
   * Linux note (prb):
   * This is a "textbook" race, and I have found discussions to get
   * around this, with use of various sheduler routines.  But, I
   * have, for now, chosen to stick with their timeout scheme, in
   * case they miss the wakeup, since I'm lazy and pressed to try
   * and get *something* working.
   *
   * Another note.  The AIX routines take a lock, and atomically
   * drop/grab the lock for you, but, I am drawing a blank trying
   * to understand exactly which (if any) routines in Linux are
   * equivalents...
   *
   * Finally, I don't know where/how a timer gets set around here!
   */

  /*
   * Would like to drop/acquire this atomically as part of the
   * sleep...
   */

  release_lock(n, clpid);
  status = interruptible_sleep_on_timeout( & cp->cli_reply_event, HZ );
  acquire_lock(n, clpid);

  if (signal_pending(current)) {
#if defined(REALLY_VERBOSE)
    bsdlog(LOG_DEBUG, "sleep_on_sendq interrupted");
#endif
    return(1);
  } else {
    return(0);
  }
}


#if defined(KERNEL_STATS)
static struct sigma stats[NUMSTATS];
static struct timestruc_t start_time;

void init_stats (void)
{
    (void) memset( stats, 0, sizeof( struct sigma ) * NUMSTATS );

    curtime(&start_time);
    start_time.tv_nsec /= 1000;
}

void stat_add (int n)
{
    struct sigma *sp;
    struct timestruc_t ts;
    
    if (n >= NUMSTATS) {
	printf(MSGSTR(INVALID_STAT, "Invalid stat %d\n"),n);
	return;
    }

    sp = &stats[n];
    curtime(&ts);
    ts.tv_nsec /= 1000;
    while (ts.tv_nsec < start_time.tv_nsec) {
	ts.tv_nsec += 1000000;
	--ts.tv_sec;
    }
    ts.tv_sec -= start_time.tv_sec;
    ts.tv_nsec -= start_time.tv_nsec;

    sp->seconds += ts.tv_sec;
    sp->microns += ts.tv_nsec;
    while (sp->microns >= 1000000) {
	++sp->seconds;
	sp->microns -= 1000000;
    }
    ++sp->samples;
}

int read_stats (void *buf, int len)
{
    if (len != sizeof(stats))
	return(-1);

    copyout(stats,buf,len);
    return(0);
}
#endif /* KERNEL_STATS */


/*
 * resname_bcopy
 *
 * copy a resource name with defense against destination overrun
 * cause abort() if failure
 * args as bcopy
 */
void
resname_bcopy( const void * src, void * dest, size_t len)
{
    if (len < 1 || len > 255 )
	abort("resname_bcopy: length bad\n") ;
    
    (void) memcpy( dest, src, len );
    return;
}

/* ============================================================ */
/* Because NLS doesn't work under kernel, so we temporary fake out 
 * the functions associated with NLS.
 */

#ifdef HAVING_NLS_ON_LINUX
#define _H_MESG
#define _H_STDIO
#include <nl_types.h>

char *setlocale(int flag, char *type)
{
	return(NULL);
}

nl_catd catopen(char *file, int flag)
{
	return((nl_catd) -1);
}

char *catgets(nl_catd catd, int set_id, int msg_id, char *message)
{
	return(message);
}

int catclose(nl_catd catd)
{
	return(0);
}
#endif  /* HAVING_NLS_ON_LINUX */

