/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_memory.c,v 1.7 2001/10/19 22:42:20 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: me_dump
 *		bfree
 *		bmalloc
 *		my_free
 *		my_malloc
 *		my_realloc
 * 		clm_mem_init
 * 		free_all
 */

#ifndef lint
#endif

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <asm/pgtable.h>
#include <asm/semaphore.h>

#include <stdarg.h>                     /* Varargs stuff */

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#include "cltrc.h"
#include "dlm_kernel.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/*
 * This first comment was NOT added during the Linux port...
 * It was already here :-)
 *
 * Rather than try to fix all the memory-allocation dain bramage
 * throughout the lock daemon, we keep track of everything we
 * allocate so we can be sure to free it before we leave.  Big
 * chunks are tracked individually, while smaller stuff is
 * allocated in private heaps (which are themselves allocated on
 * demand from the generic kernel heap).  On exit, the large
 * chunks and the private heaps are all thrown away.
 */
/*
 * DLM memory management.
 *
 * Linux notes.  The structure here is an artifact of the
 * AIX kernel implementation.  There, the kernel supports
 * paging of kernel-resident memory (including parts of the
 * OS and its memory.)  It is also a strictly segment-based
 * model.  Thus, to help alleviate pressure on the base
 * kernel memory segment (a segment is a max. of 256 MB of
 * virtual memory) the DLM could allocate its own segments
 * for locks.
 *
 * Of course, this meant the locks were pageable, which
 * could result in performance issues, but it also meant we
 * could support large numbers of locks, since it was virtual,
 * and not real, memory.  This then is a potential issue for
 * the Linux port.  We will be grabbing memory out of the
 * kernel space, not a specific difficulty, but it means we
 * can drain the system of real memory if we aren't careful.
 * The number of locks we can support will also be dependent
 * upon the amount of real memory.
 *
 * The gain by being in the kernel is responsiveness and
 * reaction to events.
 *
 * Additionally, the use of "heaps" here may not be quite
 * intuitive.  The AIX interface 'xmalloc()' was designed to
 * allocate memory from either the kernel's default heap
 * (contained in the kernel's base memory segment) or from
 * a 'private' heap.  xmalloc() would then manage the private
 * heaps as memory pools.  A private heap could be less than a
 * segment in size, a full segment, or up to 2**32 bytes in
 * size (thus encompassing multiple segments.)
 *
 * Thus, under AIX, the DLM would create private segments,
 * and then set each up as a private heap, and allocate
 * there.  It also used some amount of memory in the base
 * kernel segment, to provide anchors to the other memory
 * regions.
 *
 * Under Linux, I'm trying to keep the 'externals' of this
 * structure unchanged, to isolate modifications to the
 * memory management here and in clm_heap.c.  I want the
 * bulk of the code to just work as it did before to save
 * time up front porting.
 *
 * If you replace the above discussion from AIX with the
 * terms "cache" and "slab" we have roughly similar support
 * in Linux.  However, (at least so far as I understand it)
 * it doesn't quite work the same way.  Linux allows us to
 * build a cache, which would be a set of fixed size blocks
 * that we could then allocate/free, and we could later just
 * free the whole cache.
 *
 * This is not too bad, as there are some basic data blocks
 * that we use, so with a cache for each size we would be
 * able to mostly replicate the AIX function, and it would
 * be overall faster.
 *
 * Bottom line.  For time purposes right now, I'm going to
 * just make these passthrough calls to kmalloc()/kfree()
 * and take the memory problems in stride.  ASAP I'll look
 * to adopt a cache model here.  To help with this I'm going
 * to keep the original code here, but simply use ifdef
 * calls to preprocess it out.
 */

#ifndef __linux__

#define		BIG_CHUNK	65536
#define		HEAP_SIZE	(BIG_CHUNK *16)
#define		MAXNBIG		512
#define		CHECK_RANGE(p,h,s) \
	(((int)(p) >= (int)(h)) && ((int)(p) < ((int)(h) + (int)(s))))

struct big_desc {
    void	*ptr;
    int		len;
};

struct heap_desc {
     struct heap_desc *	next;
     caddr_t		addr;
     heapaddr_t		id;
};

typedef struct malloc_entry {
	int			me_pad[4];
	char			me_file[32];
	short			me_line;
	ulong			me_size;
	ulong			me_cookie;
	struct malloc_entry	*me_next;
	struct malloc_entry	*me_prev;
} me_t;

#define COOKIE	0xc001bead

struct big_desc		big_list[MAXNBIG];
int			nbig = 0;
struct heap_desc *	heap_list = NULL;

static int old_percent = 0;

#ifdef MEMCHK
static me_t	*me_head = NULL;
static me_t	*me_tail = NULL;
static int	resident = 0;
static char	lastfunc[50];
static int	lastline, lasttype;
static int	me_count = 0;
static me_t	*lastptr;
#endif /* MEMCHK */

#endif /* !__linux__ */

/*
 * Function:
 *	mem_init
 *
 * Description:
 *	Initialize memory management variables
 *
 * Arguments:
 *	NONE
 *
 * Return value:
 *	Nothing
 *
 * Side effects:
 *	NONE
 *
 */
void
clm_mem_init(void)
{
#ifndef __linux__
	old_percent = 0;
	nbig = 0;
	heap_list = NULL;
#if defined(MEMCHK)
	me_head = NULL;
	me_tail = NULL;
	me_count = 0;
#endif
#endif /* !__linux__ */
}

#ifndef __linux__
/*
 * Function:
 *	get_new_heap
 *
 * Description:
 *	Create a new mini-heap to allocate small objects from.
 *
 * Arguments:
 *	NONE
 *
 * Return value:
 *	Descriptor for new mini-heap.
 *
 * Side effects:
 *	NONE
 *
 */

static struct heap_desc *
get_new_heap (void)
{
  struct heap_desc *fubar = NULL;

  int *test_ptr;

  fubar = malloc(sizeof(*fubar));
  if (!fubar) {
	printf(MSGSTR(HEAP_DESC_ALLOC_FAIL, "Heap_desc allocation failed\n"));
	return(NULL);
  }
  fubar->addr = xmalloc(HEAP_SIZE,PGSHIFT,lock_heap);
  if (!fubar->addr) {
	printf(MSGSTR(HEAP_SPEC_ALLOC_FAIL, "Heap space allocation failed\n"));
	free(fubar);
	return(NULL);
  }
  fubar->id = init_heap(fubar->addr,HEAP_SIZE,NULL);

  return(fubar);
}
#endif /* !__linux__ */


/*
 * Function:
 *	my_malloc
 *
 * Description:
 * Allocate a new chunk of memory, either in the system heap
 * if the chunk is large enough, or from one of (potentially)
 * several mini-heaps for smaller stuff.
 *
 * Arguments:
 *	int size;		-- size of chunk to allocate
 *
 * Return value:
 *	Pointer to new chunk of memory, NULL if allocation failed.
 *
 * Side effects:
 *	None.
 *
 */

/* Is there a system-provided define or variable for this? */
static const int max_kmalloc_size = 131072; 

void *
my_malloc (int size)
{
  void *fubar;

  /*
   * For now, simply grab memory from the kernel.  Worry about
   * more involved management using our own caches and slabs
   * later.
   */

  if (max_kmalloc_size < size) {
    fubar = vmalloc(size);
    if ((NULL != fubar) && ((unsigned long)fubar < VMALLOC_START)) {
      bsdlog(LOG_INFO,
             "Odd, high memory is %p and just allocated at %p\n",
             (void *) VMALLOC_START,
             fubar);
    }
  } else {
    fubar = kmalloc(size, GFP_KERNEL);
  }

  if (NULL == fubar) {
    bsdlog(LOG_INFO, "Heap allocation failed!\n");
  }

#ifndef __linux__
  struct heap_desc *hp;

  if (lock_heap == kernel_heap){

    if (size >= BIG_CHUNK) {
      fubar = malloc(size);
      if (fubar) {
        if (nbig < MAXNBIG) {
          big_list[nbig].ptr = fubar;
          big_list[nbig].len = size;
          ++nbig;
        } else
          printf(MSGSTR(TOO_MANY_BLOK_ALLOC,
                        "Too many big blocks allocated\n"));
      }
    } else {
      for (hp = heap_list; hp; hp = hp->next) {
        fubar = xmalloc(size,0,hp->id);
        if (fubar)
          return(fubar);
      }
      /* No space in any current heaps */
      hp = get_new_heap();
      if (!hp) {
        printf(MSGSTR(HEAP_ALLOC_FAIL, 
                      "Heap allocation failed\n"));
        return(NULL);
      }
      hp->next = heap_list;
      heap_list = hp;
      fubar = xmalloc(size,0,hp->id);
    }
  } else {

	int	mem_used, mem_max, mem_percent;

	mem_used = ((struct heap *) lock_heap)->amount;
	mem_max = (((struct heap *) lock_heap)->numpages << PGSHIFT);
	mem_percent = mem_used / (mem_max / 100);
	if ((mem_percent - 4) > old_percent) {
      old_percent = mem_percent;
      bsdlog(LOG_DEBUG, "%s:  Used memory %d/%d MB = %d/%d", __FUNCTION__,
             mem_used >> 20, mem_max >> 20, mem_percent,
             100);
	}

    fubar = xmalloc(size,0,lock_heap);

    if (fubar == NULL){
      printf(MSGSTR(HEAP_ALLOC_FAIL, "Heap allocation failed\n"));
      return(NULL);
    }
  }
#endif /* !__linux__ */

  return(fubar);
}

/*
 * Function:
 *	my_free
 *
 * Description:
 *	Free a chunk of memory.
 *
 * Arguments:
 *	void *addr;		-- memory to be freed
 *
 * Return value:
 *	NONE
 *
 * Side effects:
 *	None.
 *
 */

unsigned int
my_free (void *addr)
{

  if ((unsigned long)addr > VMALLOC_START) {
    vfree(addr);
  } else {
    kfree(addr);
  }

#ifndef __linux__
  int i, ret;
  struct heap_desc *hp;


  if (lock_heap == kernel_heap) {
    for (hp = heap_list; hp; hp = hp->next) {
      if CHECK_RANGE(addr,hp->addr,HEAP_SIZE) {
        ret = xmfree(addr,hp->id);
        if (ret) {
          printf("xmfree returned %d\n", ret);
          abort("bad return from xmfree");
        }
        return 0;
      }
    }

    for (i = 0; i < nbig; ++i) {
      if (big_list[i].ptr == addr) {
        bsdlog(LOG_DEBUG, MSGSTR(FREE_BYTES, 
                                 "Freeing %d bytes at 0x%x\n"),
               big_list[i].len, big_list[i].ptr);
        if (nbig-- > 1)
          big_list[i] = big_list[nbig];
        ret = free(addr);
        if (ret) {
          printf("free returned %d\n", ret);
          abort("bad return from free");
        }
        return 0;
      }
    }
  } else {
    ret = xmfree(addr, lock_heap);

    if (ret != 0){
      printf(MSGSTR(FREE_ALIEN_BLOK, 
                    "Bad free at 0x%x\n"),addr);
      return 0;
	}
  }
#endif /* !__linux__ */

  return 0;

}

/*
 * Function:
 *	my_realloc
 *
 * Description:
 *	Change the size of an already-allocated chunk of memory.
 * Similar in concept to the standard realloc() we all learned
 * about in college, in case it's not abundantly obvious already.
 * The only difference is that we have to have the caller give
 * us the block's old size since the AIX developers did not see
 * fit to provide a *real* realloc in the kernel that could figure
 * that out on its own.
 *
 * Arguments:
 *	void *ptr;		-- Memory block to change
 *	int old_size;		-- Old size of block
 *	int new_size;		-- New size of block
 *
 * Return value:
 *	New address of block in case we had to move it.
 *
 * Side effects:
 *  None.
 */

void *
my_realloc (void *ptr, int old_size, int new_size)
{
    void *fubar;

    /* Emulate the BSD behaviour of degenerating into malloc().  -jjd- */
    if (!ptr)
	return(my_malloc(new_size));

    fubar = my_malloc(new_size);
    if (fubar) {
	(void) memcpy( fubar, ptr, old_size );
        my_free(ptr);
    }
    return(fubar);
}


/*
 * Function:
 *	free_all
 *
 * Description:
 *	The raison d'etre of all this my_xxx() stuff.  This routine frees all
 * of the memory we've ever allocated using my_malloc and my_realloc so that
 * we can exit safe in the knowledge that we haven't sucked the system dry of
 * memory while we were here as the original code was heartily inclined to do.
 *
 * Arguments:
 *	NONE
 *
 * Return value:
 *	NONE
 *
 * Side effects:
 *	The system probably runs faster because we just freed up huge chunks
 * of kernel memory.
 */

void
free_all (void)
{
  bsdlog(LOG_DEBUG,
         "free_all no function for now.\n");
#ifndef __linux__
    int i;
    struct big_desc *bp = big_list;
    struct heap_desc *hp, *nhp;

    for (i = 0; i < nbig; ++i) {
	free(bp->ptr);
	++bp;
    }
    nbig = 0;

    for (hp = heap_list; hp; hp = nhp) {
	free(hp->addr);
	nhp = hp->next;
	free(hp);
    }
    heap_list = NULL;
#endif /* !__linux__ */
}




#ifdef MEMCHK
void
me_check(void)
{
	me_t	*me_ptr = me_head;
	int	count = 0;

#if !defined(SLOW_MEMCHK)
	return;
#endif
	while (count < me_count && me_ptr && me_ptr->me_next) {
		if (me_ptr->me_next->me_prev != me_ptr) {
			abort("mislinked list in me_check");
		}
		count++;
		me_ptr = me_ptr->me_next;
	}

	if (me_ptr != me_tail || me_ptr && count >= me_count) {
		abort("Invalid memcheck");
	}

	count = 0;

	while (count < me_count && me_ptr && me_ptr->me_prev) {
		if (me_ptr->me_prev->me_next != me_ptr) {
			abort("mislinked list in me_check");
		}
		count++;
		me_ptr = me_ptr->me_prev;
	}

	if (me_ptr != me_head || me_ptr && count >= me_count) {
		abort("Invalid memcheck");
	}
}

void *
bmalloc(size_t	size, char *file, int line)
{
	void	*ptr;
	me_t	*me_ptr;
	ulong	*cookie;
	char	*cptr;

	CLM_ASSERT((resident == 0), "Inconceivable!!");
	resident = 1;
	me_check();

	ptr = MALLOC(size + sizeof(me_t) + sizeof(int));
	if (NULL == ptr) {
		resident = 0;
		return ptr;
	}

	me_ptr = (me_t *)ptr;
	ptr = &me_ptr[1];
	cptr = (char *)ptr;
	strcpy(me_ptr->me_file, file);
	me_ptr->me_line = line;
	me_ptr->me_size = size;
	me_ptr->me_cookie = COOKIE;
	cookie = (ulong *)&cptr[size];
	*cookie = COOKIE;

	me_ptr->me_next = NULL;
	if (NULL == me_tail) {
		me_head = me_tail = me_ptr;
		me_ptr->me_prev = NULL;
	} else {
		me_tail->me_next = me_ptr;
		me_ptr->me_prev = me_tail;
		me_tail = me_ptr;
	}

	strcpy(lastfunc, file);
	lastline = line;
	lasttype = 1;
	lastptr = me_ptr;

	me_count++;

	me_check();

	resident = 0;
	return ptr;

}

uint
bfree(void *ptr, char *file, int line)
{
	me_t	*me_ptr;
	char	*cptr;
	ulong	*cookie;
	int	i;

	CLM_ASSERT((resident == 0), "Inconceivable!!");
	resident = 1;
	me_ptr = (me_t *)ptr;
	me_ptr--;

	me_check();

	CLM_ASSERT((me_ptr->me_cookie == COOKIE),
			"Bad cookie in mem header");
	cptr = (char *)ptr;
	cookie = (ulong *)&cptr[me_ptr->me_size];
	CLM_ASSERT((*cookie == COOKIE),
			"Bad cookie in mem boundary");

	if (me_ptr->me_line == 0xd000 && me_ptr->me_size == 0xd000) {
		abort("bfree: freeing same block twice");
	}

	if (NULL != me_ptr->me_prev) {
		me_ptr->me_prev->me_next = me_ptr->me_next;
	} else if (me_head == me_ptr) {
		me_head = me_ptr->me_next;
		if (me_head)
			me_head->me_prev = NULL;
	} else {
		printf("Blech, ugh, gasp, groan...\n");
		abort("bfree");
	}

	if (NULL != me_ptr->me_next) {
		me_ptr->me_next->me_prev = me_ptr->me_prev;
	} else if (me_tail == me_ptr) {
		me_tail = me_ptr->me_prev;
		if (me_tail)
			me_tail->me_next = NULL;
	} else {
		printf("Groan, gasp, ugh, blech...\n");
		abort("bfree");
	}

	for (i=0; i<me_ptr->me_size; i++)
		cptr[i] = -1;
	me_ptr->me_size = 0xd000;
	me_ptr->me_line = 0xd000;
	me_ptr->me_cookie = 0;
	strcpy(me_ptr->me_file, file);
	me_ptr->me_line = line;

	strcpy(lastfunc, file);
	lastline = line;
	lasttype = 2;
	lastptr = me_ptr;

	FREE(me_ptr);

	me_count--;

	me_check();

	resident = 0;
	return 0;
}
#endif /* MEMCHK */
