/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 * Contributions:
 *  >2.2 proc support: Alan Jones, alan.jones@sun.com
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_proc.c,v 1.3 2001/11/09 20:04:03 idr Exp $"

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/fs.h>		/* File operations */
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/ioctl.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/timer.h>
#include <asm/uaccess.h>

#include <dlm.h>
#include <clmstructs.h>
#include <clm_data.h>
#include <clm_tune.h>
#include <dlm_kernel.h>
#include <dlm_version.h>
#include <iterator.h>
#include "dlm_proto.h"


#if defined( CPROTO ) || !defined( DEBUG )
#define STATIC static
#else
#define STATIC
#endif

STATIC struct proc_dir_entry * proc_dlm;

STATIC int dlm_proc_version( char * page, char ** start, off_t off,
    int count, int * eof, void * data );

STATIC int dlm_proc_state( char * page, char ** start, off_t off,
    int count, int * eof, void * data );

STATIC int dlm_proc_stats( char * page, char ** start, off_t off,
    int count, int * eof, void * data );

STATIC int dlm_proc_rldb( char * page, char ** start, off_t off,
    int count, int * eof, void * data );

STATIC int dlm_proc_groups( char * page, char ** start, off_t off,
    int count, int * eof, void * data );

STATIC int dlm_proc_map( char * page, char ** start, off_t off,
    int count, int * eof, void * data );

STATIC int dlm_proc_client_grant( char * page, char ** start, off_t off,
    int count, int * eof, void * data );

STATIC int dlm_proc_global_locks( char * page, char ** start, off_t off,
    int count, int * eof, void * data );

STATIC int dlm_proc_global_timeout( char * page, char ** start, off_t off,
    int count, int * eof, void * data );

/****************************************************************************/

STATIC void
dlm_proc_entry_destroy( struct proc_dir_entry * entry )
{
    struct proc_dir_entry * next;
    struct proc_dir_entry * curr;


    for ( curr = entry->subdir ; curr != NULL ; curr = next )
    {
	next = curr->next;
	if ( curr->subdir != NULL )
	{
	    dlm_proc_entry_destroy( curr->subdir );
	}

	remove_proc_entry( curr->name, proc_dlm );
    }
}




/****************************************************************************/
/* dlm_proc_start - Create the top-level /proc entries for the DLM
 * 
 */

int
dlm_proc_start( 
	        /* Name of the top-level /proc entry.
		 */

	        const char * name 
	      )
{
    struct proc_dir_entry * sub;


    proc_dlm = proc_mkdir( name, NULL );


    /* Create the top-level "version" entry.  Notice that the data field is
     * an explicit cast of 0 to (void *) for a reason.  dlm_proc_version does
     * not use 'data' as a pointer.  It uses it as an integer, and a value of
     * 0 has special meaning.
     */

    sub = create_proc_read_entry( "version", 0, proc_dlm, dlm_proc_version,
				  (void *) 0 );
    sub->owner = THIS_MODULE;


    /* Create the top-level "state" entry.  This is how users (i.e., system
     * administrators) can determine the state of any in-progress DLM
     * recovery operations.
     */

    sub = create_proc_read_entry( "state", 0, proc_dlm, dlm_proc_state,
				  NULL );
    sub->owner = THIS_MODULE;

    sub = create_proc_read_entry( "stats", 0, proc_dlm, dlm_proc_stats,
				  NULL );
    sub->owner = THIS_MODULE;

    sub = create_proc_read_entry( "rldb", 0, proc_dlm, dlm_proc_rldb, 
				  NULL );
    sub->owner = THIS_MODULE;

    sub = create_proc_read_entry( "groups", 0, proc_dlm, dlm_proc_groups,
				  NULL );
    sub->owner = THIS_MODULE;

    sub = create_proc_read_entry( "map", 0, proc_dlm, dlm_proc_map,
				  NULL );
    sub->owner = THIS_MODULE;


    sub = create_proc_read_entry( "grant_queue", 0, proc_dlm,
				  dlm_proc_global_locks, (void *) RLQ_GRANT );
    sub->owner = THIS_MODULE;

    sub = create_proc_read_entry( "wait_queue", 0, proc_dlm,
				  dlm_proc_global_locks, (void *) RLQ_WAIT );
    sub->owner = THIS_MODULE;

    sub = create_proc_read_entry( "convert_queue", 0, proc_dlm,
				  dlm_proc_global_locks, (void *) RLQ_CONV );
    sub->owner = THIS_MODULE;

    return( 0 );
}




/****************************************************************************/

void
dlm_proc_cleanup( void )
{
    if ( proc_dlm != NULL )
    {
	/* Remove all the entries inside /proc/haDLM.
	 */

	dlm_proc_entry_destroy( proc_dlm );

	remove_proc_entry( proc_dlm->name, NULL );
	proc_dlm = NULL;
    }
}




/****************************************************************************/
/* dlm_proc_create_node_entry - Create a node-level /proc entry
 * 
 */

struct proc_dir_entry *
dlm_proc_create_node_entry(
			    unsigned site
			  )
{
    struct proc_dir_entry * dir;
    struct proc_dir_entry * sub;
    char   buffer[8];


    (void) snprintf( buffer, sizeof( buffer ), "%u", site );
    buffer[ sizeof( buffer ) - 1 ] = '\0';

    dir = proc_mkdir( buffer, proc_dlm );
    dir->owner = THIS_MODULE;


    /* Create the "version" entry.  dlm_proc_version does not use 'data' as a
     * pointer.  It uses it as an integer representing the node ID.
     */

    sub = create_proc_read_entry( "version", 0, dir, dlm_proc_version,
				  (void *) site );
    sub->owner = THIS_MODULE;
    
    
    /* If the node ID matches our local node ID, then set up a link to the
     * new directory from the 'local' directory.
     */

    if ( site == OurSiteID )
    {
	(void) remove_proc_entry( "local", proc_dlm );
	sub = proc_symlink( "local", proc_dlm, buffer );
	sub->owner = THIS_MODULE;
    }

    return( dir );
}




/****************************************************************************/
/* dlm_proc_destroy_node_entry - Destroy a node-level /proc entry
 * 
 */

void
dlm_proc_destroy_node_entry(
			     struct proc_dir_entry * base
			   )
{
    if ( base != NULL )
    {
	remove_proc_entry( "version", base );
	remove_proc_entry( base->name, base->parent );
    }
}




/****************************************************************************/
/* dlm_proc_version - Display the version of the specified node
 * 
 */

STATIC int
dlm_proc_version(
		  char * page,
		  char ** start __attribute__ ((__unused__)),
		  off_t off,
		  int count,
		  int * eof,
		  void * data
		)
{
    off_t      bytes;
    unsigned   major;
    unsigned   minor;


    /* If data is 0, then this is a short cut meaning the local node.
     * Otherwise, look up the version of the specified node.
     */
    
    if ( (unsigned) data != 0 )
    {
	int   version;


	acquire_lock( LK_ENTRY, kproc_pid );
	version = clm_version( (int) data );
	release_lock( LK_ENTRY, kproc_pid );
	
	if ( version == -1 )
	{
	    major = 0;
	    minor = 0;
	}
	else
	{
	    major = ((unsigned) version) >> 16;
	    minor = ((unsigned) version) & 0x0ffff;
	}
    }
    else
    {
	major = DLM_MAJOR_VERSION;
	minor = DLM_MINOR_VERSION;
    }
    

    bytes = snprintf( & page[ off ], count - off, "%u.%u\n", major, minor );
    *eof = 1;

    return( bytes );
}




/****************************************************************************/
/* dlm_proc_state - Display the current DLM recovery state
 * 
 */

STATIC int
dlm_proc_state(
	        char * page,
	        char ** start __attribute__ ((__unused__)),
		off_t off,
		int count,
		int * eof,
		void * data __attribute__ ((__unused__))
	      )
{
    *eof = 1;
    return( dlm_recov_print_state( & page[ off ], count - off ) );
}




/****************************************************************************/

STATIC int
dlm_proc_stats(
	        char * page,
	        char ** start __attribute__ ((__unused__)),
		off_t off,
		int count __attribute__ ((__unused__)),
		int * eof,
		void * data __attribute__ ((__unused__))
	      )
{
    struct dlm_stats_buf   stats;
    size_t                 bytes;

    
    clm_global_stats( & stats, 1 );

    count -= off;
    bytes = snprintf( & page[ off ], count,
		      "Usage information: counts: current/highest[%d]\n",
		      stats.maxresources);
    bytes += snprintf( & page[ off + bytes ], count - bytes,
		       " Locks: current count[%d] max allowed[%d] denied[%d] warns[%d]\n",
		       stats.reclocks_used, stats.reclocks_max,
		       stats.reclocks_denied, stats.lock_warned );

    bytes += snprintf( & page[ off + bytes ], count - bytes,
		       " Resources: In use[%d] max[%d] warns[%d]\n",
		       stats.res_used, stats.items_max, stats.res_warned );

    *eof = 1;
    return( bytes );
}




/****************************************************************************/

struct proc_dir_entry *
dlm_proc_create_client_entry( 
			      unsigned node_id,
			      clm_client_t * client
			    )
{
    struct proc_dir_entry * client_entry;
    struct proc_dir_entry * node_entry;
    struct proc_dir_entry * sub;
    char   buffer[17];


    (void) snprintf( buffer, sizeof( buffer ), "%08X%08X",
		    client->cli_id.clid_i[0], 
		    client->cli_id.clid_i[1] );
    buffer[16] = '\0';

    bsdlog( LOG_INFO, "[%s:%s] making PROC entry for client %u/%s "
	    "(%sactive)\n",
	    current->comm, __FUNCTION__,
	    node_id, buffer, clm_node_active( node_id ) ? "" : "not " );

    client_entry = NULL;
    node_entry = dlm_node_get_proc_entry( node_id );
    if ( node_entry != NULL )
    {
	client_entry = proc_mkdir( buffer, node_entry );
	client_entry->owner = THIS_MODULE;

	sub = create_proc_read_entry( "grant", 0, client_entry,
				      dlm_proc_client_grant, client );
	sub->owner = THIS_MODULE;
    }

    return( client_entry );
}




/****************************************************************************/

void
dlm_proc_destroy_client_entry( struct proc_dir_entry * entry )
{
    dlm_proc_entry_destroy( entry );
    remove_proc_entry( entry->name, entry->parent );
}




static const char * const locktypes[] = { "UNIX", "VMS", "INVALID" };

/* cproto 4.6 has problems with functions that return const pointers.  It
 * generates invalid prototypes.
 */

#ifndef CPROTO
const char *
dlm_get_lock_type_string( unsigned type )
{
    return( locktypes[ (type < DLM_RES_NTYPES) ? type : DLM_RES_NTYPES ] );
}
#endif /* CPROTO */

STATIC int 
dlm_proc_client_grant( char * page, char ** start, off_t off,
		       int count, int * eof, void * data )
{
    clm_client_t    * client;
    struct reclock  * lock;
    struct resource * res;
    size_t            bytes;
    const char      * type;


    client = data;

    count -= off;
    bytes = snprintf( & page[ off ], count,
		      "                 | Grant | Request | Bast | \n" );
    bytes += snprintf( & page[ off + bytes], count - bytes,
		      "            Name |  Mode |    Mode | Mode |    Type\n" );

    for ( EACH_DLM_LISTNODE( & client->cli_queue, lock ) )
    {
	res = resptr( lock->lck.rh );
	
	type = dlm_get_lock_type_string( lock->lck.type );

	bytes += snprintf( & page[ off + bytes ], count - bytes,
			  "%16s | %5s | %7s | %4s | %7s\n",
			  printbin( res->name, res->namelen ),
			  clm_printmode( lock->lck.mode ),
			  clm_printmode( lock->lck.reqmode ),
			  clm_printmode( lock->lck.bastmode ),
			  type
		        );
    }
    
    *eof = 1;
    return( bytes );
}




/****************************************************************************/

unsigned
dlm_lock_print( char * buf, unsigned limit, struct reclock * lock )
{
    unsigned   bytes;

    bytes = snprintf( buf, limit,
		      "%d: id=0x%07x rmid=0x%x pid=%d gt=%s rq=%s "
		      "flgs=0x%x nd=%d st=0x%x ",
		      lock->lck.seqnum, lock->lck.lockid,
		      lock->lck.remlockid, lock->lck.pid,
		      clm_printmode(lock->lck.mode),
		      clm_printmode(lock->lck.reqmode),
		      lock->lck.lk_flags, lock->lck.site, lock->lck.state );

    if ( lock->lck.lk_flags & LKM_XID)
    {
	bytes += snprintf( & buf[ bytes ], limit - bytes, "xid=0x%x%08x\n",
			   (uint)lock->lck.xid.xid_i[0], 
			   (uint)lock->lck.xid.xid_i[1] );
    }
    else
    {
	bytes += snprintf( & buf[ bytes ], limit - bytes, "req_pid=%d\n",
			   lock->route.asth.asth_pid );
    }

    return( bytes );
}




/****************************************************************************/
/* dlm_proc_client_wait - List locks the client is waiting to acquire
 * 
 */

STATIC int
dlm_proc_global_locks( char * page, char ** start, off_t off,
		       int count, int * eof, void * data )
{
    unsigned          offset;
    iterator_t      * iter;
    struct resource * res;
    dlm_list_t      * list;
    struct reclock  * lock;
    size_t            bytes;


    offset = (unsigned) data;
    count -= off;
    bytes = 0;

    iter = resource_table_iterator_create();
    while ( ! iterator_next( iter, & res ) && (bytes < count) )
    {
	if ( res->rsrc_flags & RSRC_ACTIVE )
	{
	    bytes += snprintf( & page[ off + bytes ], count - bytes,
			      "RESOURCE: %s (0x%08lx)\n", 
			      printbin( res->name, res->namelen ),
			      res->rh.rh );

	    if ( res->type == DLM_RES_UNIX) 
	    {
		bytes += snprintf( & page[ off + bytes ], count - bytes,
				  "MASTER HANDLE: 0x%08lx\n", res->mst_rh.rh );
	    }
	    else 
	    {
		bytes += snprintf( & page [ off + bytes ], count - bytes,
				  "VALUE BLOCK: %s\n",
				  printbin( res->value,	sizeof(res->value) ) );
	    }

	    bytes += snprintf( & page[ off + bytes ], count - bytes,
			     "INFO: refcount=%u flags=0x%08x master=%d "
			     "stick=%ld\n",
			     res->refcount, res->rsrc_flags, res->master,
			     res->clminfo.mi_stickiness );

	    list = & res->lockqs[ offset ];
	    for ( EACH_DLM_LISTNODE( list, lock ) )
	    {
		bytes += dlm_lock_print( & page[ off + bytes ], 
					 count - bytes, lock );
	    }
	}
    }

    iterator_destroy( iter );

    *eof = 1;
    return( bytes );
}




/****************************************************************************/

STATIC int
dlm_proc_global_timeout( char * page, char ** start, off_t off,
			 int count, int * eof, void * data )
{
    *eof = 1;
    return( dlm_print_timeout_queue( page + off, count ) );
}




/****************************************************************************/

STATIC int
dlm_proc_rldb( char * page, char ** start, off_t off,
	       int count, int * eof, void * data )
{
    *eof = 1;
    return( dlm_rldb_print( page + off, count ) );
}




/****************************************************************************/

STATIC int
dlm_proc_groups( char * page, char ** start, off_t off,
		 int count, int * eof, void * data )
{
    *eof = 1;
    return( dlm_group_print( page + off, count ) );
}




/****************************************************************************/

STATIC int
dlm_proc_map( char * page, char ** start, off_t off,
	      int count, int * eof, void * data )
{
    *eof = 1;
    return( dlm_map_table_print( page + off, count ) );
}
