/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, romanick@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_recover.c,v 1.15 2001/11/29 21:59:05 idr Exp $"

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/signal.h>
#include <linux/errno.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "clmstructs.h"
#undef LOCK_FREE
#include "dlm_kernel.h"
#include "clm_reconfig.h"
#include "clm_data.h"
#include "dlm_version.h"
#include <iterator.h>

#include "hsm.h"
#define IN_MACHINE
#include "dlm_recover.h"
#undef IN_MACHINE

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) \
	catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), \
		CLLOCKD, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "dlm_workqueue.h"
#include "dlm_proto.h"  /* Should be LAST included file. */

/* ============================================================ */

static hsm_t  private_dlm_recov_state_machine;
hsm_t * const dlm_recov_state_machine = & private_dlm_recov_state_machine;
static kmem_cache_t * dlm_recov_rc_msg_cache = NULL;
static atomic_t   dlm_recov_rc_msg_cache_count = ATOMIC_INIT( 0 );

static int dlm_recov_leaving_node;
static int dlm_recov_joining_node;
static int dlm_recov_source_node;
static bool_t dlm_recov_in_purge;
static bool_t dlm_recov_in_add_delete;
bool_t dlm_recov_in_reconfig = FALSE;
bool_t dlm_recov_can_transact = FALSE;

static unsigned dlm_buff_per_try = 32;
static unsigned dlm_buff_limit  = 500;
static atomic_t dlm_recov_updates_inflight;

/* TRUE if all directory info has been interchanges, so orphan directory
 * entries may be deleted.
 */

static bool_t dlm_recov_can_del_dir;


static void dlm_recov_dir_msg_retire( clm_msg_t * msg, int stat );
static void dlm_recov_restart_msg_retire( clm_msg_t * msg, int stat );
static void dlm_recov_set_retry_timer( void );

static void clmr_purge_res( struct resource * resptr );
static unsigned clmr_clean_queue( struct resource *rp, 
    dlm_list_t * plock_queue );
static void clmr_clean( struct resource * resptr );
static int clmr_rm_locallock_dirs( rl_info_t *rl_entry );
static int clmr_send_dir_entry( rl_info_t *rldbe );
static int clmr_dir_purge_chk( rl_info_t *rldbe );
static int clmr_dir_move_chk( rl_info_t *rldbe );
static int clmr_res( struct resource *resptr );

static int clmr_end_new_node( int node );
static int clmr_restore_2( void );
static int clmr_dir( struct resource * resptr );
static void clmr_mark_stale(struct resource *resptr, int site);
static void clmr_alloc_reclock( struct resource *rp, struct reclock *mrglk,
    dlm_list_t *plockq, struct reclock *curlk, int qtype );
static void clmr_chain_reclocks(struct resource *rp, struct reclock *mrglk,
    dlm_list_t * plock_queue, const clmr_data_t *pdata, int qtype );
static int clmr_chain( struct resource * rp, clmr_data_t * data );
static int clmr_rebuild_send(struct resource *resptr, int master);
static void clmr_bundle_reclocks(struct reclock **prptr, char **pbptr,
    int *preclock_ctr, int *pl_num, int *pq_num );
static void clmr_lkvlb( struct resource * resptr );
static void handle_resource_request( int type, clmr_data_t * data, unsigned node );
static void handle_dir_request( rc_msg_t * dat );
static void clmr_respond( struct resource *resptr, clmr_data_t * data,
    unsigned node );

static int clmr_message_handler( clm_msg_t * msg );

/****************************************************************************/

struct rc_msg *
rc_msg_alloc( int event, unsigned node, void * data )
{
    rc_msg_t * msg;


    msg = kmem_cache_alloc( dlm_recov_rc_msg_cache, 0 );
    if ( msg != NULL )
    {
	msg->base.evt = event;
	msg->node = node;
	msg->data = data;
	atomic_inc( & dlm_recov_rc_msg_cache_count );
    }

    return( msg );
}




void
rc_msg_free( const void * msg )
{
#ifdef DEBUG
    struct rc_msg * rc_msg;


    rc_msg = (struct rc_msg *) msg;
    if ( rc_msg->data != NULL )
    {
	bsdlog( LOG_ERR, "[%s:%s] State-machine message with non-NULL data "
	        "pointer being freed (0x%08lx / 0x%08lx / %u)\n",
	        current->comm, __FUNCTION__,
	        (unsigned long) rc_msg, (unsigned long) rc_msg->data,
	        rc_msg->base.evt );
    }
#endif /* DEBUG */

    CLM_ASSERT( atomic_read( & dlm_recov_rc_msg_cache_count ) != 0,
	        "Recovery message cache has no allocated messages!" );
    atomic_dec( & dlm_recov_rc_msg_cache_count );
    kmem_cache_free( dlm_recov_rc_msg_cache, (void *) msg );
}




/****************************************************************************/
/* dlm_recov_init - 
 *
 * Initialize the comm ports that will be used to rebuild lock manager
 * resource information in the event of a node failure.  It also performs
 * the necessary startup tasks of querying for and reading in directory
 * information that belongs on the local node but was stored somehwere else
 * before this node came up.
 *
 * This function also installs the callback function that will be called by
 * the clminfo layer whenever any cluster lock manager changes its state.
 */

int
dlm_recov_init( void )
{
    dlm_recov_rc_msg_cache = kmem_cache_create( "DLM recovery msg",
					        sizeof( rc_msg_t ), 0, 0,
					        NULL, NULL );

    cccp_listen( PORT_RECOVERY, clmr_message_handler );

    HSM_INIT( dlm_recov_state_machine, "DLM recovery state machine",
	     CLM_ST_TOP );
    hsm_start( dlm_recov_state_machine );

    return( 0 );
}



/****************************************************************************/
/* dlm_recov_cleanup - Clean up all the resources used by recovery
 * 
 * Cleans up all of the resources used by the DLM recovery state machine.
 */

void
dlm_recov_cleanup( void )
{
    cccp_listen( PORT_RECOVERY, NULL );
    
    if ( dlm_recov_rc_msg_cache != NULL )
    {
	bsdlog( LOG_INFO, "[%s:%s] destroying %s cache (%d)\n",
	        current->comm, __FUNCTION__, "DLM recovery msg",
	        atomic_read( & dlm_recov_rc_msg_cache_count ) ); 
	kmem_cache_destroy( dlm_recov_rc_msg_cache );
	dlm_recov_rc_msg_cache = NULL;
    }
}




/****************************************************************************/
/* clmr_ok_to_del_dir -
 * 
 * RETURNS
 * 
 * If all directory information has been interchanged, TRUE is returned.
 * Otherwise, FALSE is returned.
 */

bool_t
clmr_ok_to_del_dir( void )
{
    return( dlm_recov_can_del_dir );
}




/****************************************************************************/
/* dlm_recovery_send_node_up_nak - Send a NAK for a CLM_MSG_NODE_UP
 * 
 * Sends a NAK to a node that has send a CLM_MSG_NODE_UP message.  The NAK can
 * either be a RC_MSG_DIR_NAK or RC_MSG_DIR_NOT_UP message.
 */

static void
dlm_recovery_send_node_up_nak( unsigned node, unsigned nak_type )
{
    clm_msg_t * nak_msg;
    int         status;


    CLM_ASSERT( (nak_type == RC_MSG_DIR_NAK)
	        || (nak_type == RC_MSG_DIR_NOT_UP),
	        "Invalid NAK type for CLM_MSG_NODE_UP message." );

    /* BUGBUG What if nak_msg is NULL? */
    nak_msg = cccp_msg_create( PORT_RECOVERY, DIRRPRI, node, nak_type,
			       0, NULL, NULL, NULL );
    if ( nak_msg != NULL )
    {
	status = cccp_send( nak_msg );
	/* BUGBUG What if status is non-zero? */
    }
}




#ifndef CPROTO

/****************************************************************************/
/* clm_st_top - Handle the CLM_ST_TOP state
 *
 * The entire recovery state machine is implemented as a heirarchical state
 * machine.  That is, some states actually contain sub-states.  At the outer
 * most level there is a single state that contains the entire state machine.
 * This is CLM_ST_TOP.
 *
 * This is very convenient because it serves as a "catch all" for any
 * unhandled messages.  This is why the default case in the event swtich in
 * clm_st_top is to log a warning message.  This also means that the code to
 * handle CLM_MSG_NODE_DOWN events and CLM_MSG_NODE_UP events only has to be
 * in one place.  CLM_MSG_NODE_UP events are only handles in the CLM_ST_RUN
 * state, and all other states ignore that message and pass it all the way up
 * to the top.  The CLM_MSG_NODE_DOWN message causes the current recovery to
 * abort (from any state) and go to the CLM_ST_DRAIN state.
 */

hsm_msg_t const *
clm_st_top( hsm_t * me, hsm_msg_t const * msg )
{
    switch( msg->evt )
    {
        case START_EVT:
	{
	    STATE_START( me, CLM_ST_INIT );
	    break;
	}

        case ENTRY_EVT:
        case EXIT_EVT:
        {
            break;
        }

	case CLM_MSG_NODE_DOWN:
	{
	    dlm_recov_in_add_delete = FALSE;
	    dlm_recov_in_purge = FALSE;
	    dlm_recov_leaving_node = ((rc_msg_t *) msg)->node;
	    STATE_TRAN( me, CLM_ST_DRAIN );
	    rc_msg_free( msg );
	    break;
	}

	/* The DLM is not currently in a state where it can handle a directory
	 * request.  Send a RC_MSG_DIR_NAK message back to the sender.  This
	 * will tell them to wait a little bit before asking again.
	 */

	case CLM_MSG_NODE_UP:
	{
	    dlm_recovery_send_node_up_nak( ((rc_msg_t *) msg)->node,
					   RC_MSG_DIR_NAK );
	    rc_msg_free( msg );
	    break;
	}

	default:
	{
	    bsdlog( LOG_WARNING, "Unhandled message type %d in state %s\n",
		    msg->evt, STATE_CURR( me )->name );
	    rc_msg_free( msg );
	    break;
	}
    }

    return( NULL );
}




/****************************************************************************/
/* clm_st_reconfig - Handle the CLM_ST_RECONFIG state
 *
 */

hsm_msg_t const *
clm_st_reconfig( hsm_t * me, hsm_msg_t const * msg )
{
    rc_msg_t * dat;


    dat = (rc_msg_t *) msg;

    switch( msg->evt )
    {
        case START_EVT:
        {
	    dlm_recov_in_reconfig = TRUE;
	    STATE_START( me, RC_BARRIER2 );
            break;
        }

        case ENTRY_EVT:
	{
	    dlm_recov_in_reconfig = TRUE;
	    break;
	}

        case EXIT_EVT:
        {
	    dlm_recov_in_reconfig = FALSE;
            break;
        }

	case RC_MSG_DIR_DATA:
	{
	    struct resource * resptr;
	    rl_update_t * prlq;


	    prlq = (rl_update_t *) dat->data;

	    /* count response for master state check */
	    atomic_dec( & dlm_recov_updates_inflight );


	    DBPRINTF((DBG_RECOV, "[%s:%s] from %d (%s) m-> %d",
		      current->comm, __FUNCTION__,
		      dat->node, printbin( prlq->rl_name, prlq->rl_namelen ),
		      prlq->rl_node));


	    /* continue handling */
	    if ( prlq->rl_node != 0 )
	    {
		resptr = find_resource( prlq->rl_name, prlq->rl_namelen,
				        prlq->rl_type );
		if ( resptr != NULL )
		{
		    if ( prlq->rl_node == OurSiteID )
		    {
			DBPRINTF((DBG_RECOV, "[%s:%s] %d says %d died.  "
				  "Reassign mastery of (%s)",
				  current->comm, __FUNCTION__, dat->node,
				  resptr->master,
				  printbin(resptr->name, resptr->namelen)));

			if (resptr->master != 0)
			{
			    resptr->rsrc_flags |= RSRC_NEW_MASTER;
			}

			resptr->master = 0;
			resptr->rsrc_flags &= ~RSRC_FRZEN;

			/* Clean out the queues in the resource.
			 */

			clmr_clean( resptr );
		    }
		    else
		    {
			clmr_rebuild_send(resptr, prlq->rl_node);
		    }
		}
	    }

	    break;
	}

	case RC_MSG_DIR_REQUEST:
	{
	    handle_dir_request( dat );
	    break;
	}

	case RC_MSG_RES_DATA:
	{
	    clr_res_t * clres;
	    struct resource *rptr;


	    clres = dat->data;
	    rptr = clres->clr_resptr;

	    (void)memcpy( &rptr->mst_rh, &clres->clr_mst_rh, 
			  sizeof( union dlm_rh ) );
	    rptr->master = dat->node;
	    CLM_ASSERT((rptr->master), "Rebuild response w/ zero master");
	    break;
	}

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* clm_st_drain - Handle the CLM_ST_DRAIN state
 *
 */

hsm_msg_t const *
clm_st_drain( hsm_t * me, hsm_msg_t const * msg )
{
    switch( msg->evt )
    {
        case START_EVT:
        {
	    STATE_START( me, RC_BARRIER1 );
            break;
        }

        case ENTRY_EVT:
	{
	    break;
	}

        case EXIT_EVT:
        {
	    dlm_recov_can_transact = FALSE;
            break;
        }

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* clm_st_send_dir - Handle the CLM_ST_SEND_DIR state
 *
 */

hsm_msg_t const *
clm_st_send_dir( hsm_t * me, hsm_msg_t const * msg )
{
    switch( msg->evt )
    {
        case START_EVT:
        case EXIT_EVT:
        {
            break;
        }

        case ENTRY_EVT:
        {
	    rc_msg_t * send_more_msg;


	    rl_enum_init();

	    /* Send a RC_MSG_SEND_MORE message to the machine.  This is
	     * kind of a hack, but it works.
	     */

	    send_more_msg = rc_msg_alloc( RC_MSG_SEND_MORE, 0, 0 );
	    hsm_process_event( dlm_recov_state_machine, (hsm_msg_t *) send_more_msg );
            break;
        }

        case RC_MSG_SEND_MORE:
        {
	    int   done;


	    bsdlog( LOG_INFO, "[%s:%s] got RC_MSG_SEND_MORE, sending "
		    "dir info to %u",
		    current->comm, __FUNCTION__, dlm_recov_joining_node );

	    rl_enum_counted( clmr_send_dir_entry, RLENUM_ALLNODES, 0,
			     dlm_buff_per_try, & done );
	    if ( done )
	    {
		clmr_end_new_node( dlm_recov_joining_node );
		STATE_TRAN( me, CLM_ST_RUN );
	    }

	    rc_msg_free( msg );
            break;
        }

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* clm_st_run - Handle the CLM_ST_RUN state
 *
 */

hsm_msg_t const *
clm_st_run( hsm_t * me, hsm_msg_t const * msg )
{
    switch( msg->evt )
    {
        case START_EVT:
        case EXIT_EVT:
        {
            break;
        }

        case ENTRY_EVT:
	{
	    grace_period = 0;
	    clm_active = TRUE;
	    dlm_recov_can_transact = TRUE;
	    purge_active();
	    break;
	}

        case CLM_MSG_NODE_PURGE:
        {
	    dlm_recov_in_add_delete = FALSE;
	    dlm_recov_in_purge = TRUE;
	    STATE_TRAN( me, CLM_ST_DRAIN );
	    rc_msg_free( msg );
            break;
        }

        case CLM_MSG_NODE_ADD:
        case CLM_MSG_NODE_DELETE:
        {
	    dlm_recov_in_add_delete = TRUE;
	    dlm_recov_in_purge = FALSE;
	    STATE_TRAN( me, CLM_ST_DRAIN );
	    rc_msg_free( msg );
            break;
        }

        case CLM_MSG_NODE_UP:
        {
	    dlm_recov_joining_node = ((rc_msg_t *) msg)->node;
	    STATE_TRAN( me, CLM_ST_SEND_DIR );
	    rc_msg_free( msg );
            break;
        }

        case CLM_MSG_NODE_DOWN:
        {
	    dlm_recov_in_add_delete = FALSE;
	    dlm_recov_in_purge = FALSE;
	    dlm_recov_leaving_node = ((rc_msg_t *) msg)->node;
	    STATE_TRAN( me, CLM_ST_DRAIN );
	    rc_msg_free( msg );
            break;
        }

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* clm_st_init - Handle the CLM_ST_INIT state
 *
 */

hsm_msg_t const *
clm_st_init( hsm_t * me, hsm_msg_t const * msg )
{
    switch( msg->evt )
    {
        case START_EVT:
        {
	    STATE_START( me, RC_INIT_WAIT );
            break;
        }

        case ENTRY_EVT:
        case EXIT_EVT:
        {
            break;
        }

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* rc_barrier6 - Handle the RC_BARRIER6 state
 *
 */

hsm_msg_t const *
rc_barrier6( hsm_t * me, hsm_msg_t const * msg )
{
    switch( msg->evt )
    {
        case START_EVT:
        case EXIT_EVT:
        {
            break;
        }

        case ENTRY_EVT:
        {
	    clm_sync( dlm_recov_leaving_node, 6 );
            break;
        }

	/* Last part of DLM reconfiguration.  Called for transition for
	 * CLM_ST_RECONFIG to CLM_ST_RUN.  Resends messages delayed by
	 * reconfig or node failure, wakes all locks, starts delayed purges.
	 */

        case RC_MSG_GOT_BARRIER:
        {
	    /* Grant requests that were blocked by clients on dead node.
	     */

	    wake_all_locks();


	    /* Resend any messages whose callback indicated failure.
	     */

	    pti_flush_resend();


	    /* Resend any messages whose callback indicated success, but were
	     * sent to a failed node and have not received a response.
	     */

	    resend_inflight();


	    /* If any purges are necessary but were delayed, do them now.
	     */

	    client_apply( reconfig_allow_client_purge, (void *) "reconfig",
			  NULL );

	    grace_period = 0;
	    clm_change_topology();

	    STATE_TRAN( me, CLM_ST_RUN );
	    rc_msg_free( msg );
            break;
        }

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* rc_barrier5 - Handle the RC_BARRIER5 state
 *
 */

hsm_msg_t const *
rc_barrier5( hsm_t * me, hsm_msg_t const * msg )
{
    switch( msg->evt )
    {
        case START_EVT:
        case EXIT_EVT:
        {
            break;
        }

        case ENTRY_EVT:
        {
	    clm_sync( dlm_recov_leaving_node, 5 );
            break;
        }

	/* Remove directories created for local locks during reconfig see
	 * clmr_rm_locallock_dirs().
	 */

        case RC_MSG_GOT_BARRIER:
        {
	    rl_enum( clmr_rm_locallock_dirs, RLENUM_ALLNODES, 0 );
	    STATE_TRAN( me, RC_BARRIER6 );
	    rc_msg_free( msg );
            break;
        }

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* rc_barrier4 - Handle the RC_BARRIER4 state
 *
 */

hsm_msg_t const *
rc_barrier4( hsm_t * me, hsm_msg_t const * msg )
{
    switch( msg->evt )
    {
        case START_EVT:
        case EXIT_EVT:
        {
            break;
        }

        case ENTRY_EVT:
        {
	    clm_sync( dlm_recov_leaving_node, 4 );
            break;
        }

        /* Invalidate lock value blocks for any lock which was held EX or PW 
         * by nodes that have died. 
         */

        case RC_MSG_GOT_BARRIER:
        {
	    struct resource * res;
	    iterator_t * iter;


	    iter = resource_table_iterator_create();
	    while ( ! iterator_next( iter, & res ) )
	    {
		clmr_lkvlb( res );
	    }

	    iterator_destroy( iter );

	    STATE_TRAN( me, RC_BARRIER5 );
	    rc_msg_free( msg );
            break;
        }

	case RC_MSG_RES_REQ:
	case RC_MSG_RES_REQ_1ST:
	{
	    rc_msg_t * dat;


	    dat = (rc_msg_t *) msg;
	    handle_resource_request( msg->evt, dat->data, dat->node );
	    rc_msg_free( msg );
	    break;
	}

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* rc_resource_query - Handle the RC_RESOURCE_QUERY state
 *
 */

hsm_msg_t const *
rc_resource_query( hsm_t * me, hsm_msg_t const * msg )
{
    static iterator_t * iter = NULL;


    switch( msg->evt )
    {
        case START_EVT:
        {
            break;
        }

        case EXIT_EVT:
        {
	    if ( iter != NULL )
	    {
		iterator_destroy( iter );
		iter = NULL;
	    }

            break;
        }

        case ENTRY_EVT:
        {
	    rc_msg_t * send_more_msg;


	    dlm_recov_can_del_dir = TRUE;


	    /* Send a RC_MSG_SEND_MORE message to the machine.  This is
	     * kind of a hack, but it works.
	     */

	    iter = resource_table_iterator_create();
	    send_more_msg = rc_msg_alloc( RC_MSG_SEND_MORE, 0, 0 );
	    hsm_process_event( dlm_recov_state_machine, (hsm_msg_t *) send_more_msg );
            break;
        }

	/* This work used to be done in clmr_stage_3.
	 */

        case RC_MSG_SEND_MORE:
        {
	    struct resource * res;
	    unsigned   i;
	    bool_t   done;


	    /* Handle any directory updates read during early rebuild.
	     */

	    resend_updates();


	    /* Rebuild lost resources.
	     */

	    done = FALSE;
	    for ( i = 0 ; i < dlm_buff_per_try ; i++ )
	    {
		done = iterator_next( iter, & res );
		if ( done )
		{
		    break;
		}

		if ( (res->rsrc_flags & RSRC_ACTIVE) != 0 )
		{
		    clmr_res( res );
		}
	    }

	    if ( done && (atomic_read( & dlm_recov_updates_inflight ) == 0) )
	    {
		STATE_TRAN( me, RC_BARRIER4 );
	    }

	    rc_msg_free( msg );
            break;
        }

	case RC_MSG_RES_REQ:
	case RC_MSG_RES_REQ_1ST:
	{
	    rc_msg_t * dat;


	    dat = (rc_msg_t *) msg;
	    handle_resource_request( msg->evt, dat->data, dat->node );
	    rc_msg_free( msg );
	    break;
	}

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* rc_barrier3 - Handle the RC_BARRIER3 state
 *
 */

hsm_msg_t const *
rc_barrier3( hsm_t * me, hsm_msg_t const * msg )
{
    switch( msg->evt )
    {
        case START_EVT:
        case EXIT_EVT:
        {
            break;
        }

        case ENTRY_EVT:
        {
	    clm_sync( dlm_recov_leaving_node, 3 );
            break;
        }

        case RC_MSG_GOT_BARRIER:
        {
	    STATE_TRAN( me, RC_RESOURCE_QUERY );
	    rc_msg_free( msg );
            break;
        }

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* rc_rebuild_dir - Handle the RC_REBUILD_DIR state
 *
 */

hsm_msg_t const *
rc_rebuild_dir( hsm_t * me, hsm_msg_t const * msg )
{
    static iterator_t * iter = NULL;


    switch( msg->evt )
    {
        case START_EVT:
        {
            break;
        }

        case EXIT_EVT:
        {
	    if ( iter != NULL )
	    {
		iterator_destroy( iter );
		iter = NULL;
	    }

            break;
        }

        case ENTRY_EVT:
        {
	    rc_msg_t * send_more_msg;


	    reconfig_del_dead_clients();

	    if ( dlm_recov_in_add_delete )
	    {
		/* Add/delete node "happens" here.
		 */

		clm_change_topology();


		/* Remove dir entries which no longer hash to this node.
		 */

		rl_enum(clmr_dir_move_chk, RLENUM_ALLNODES, 0) ;
	    }


	    /* Send a RC_MSG_SEND_MORE message to the machine.  This is
	     * kind of a hack, but it works.
	     */

	    iter = resource_table_iterator_create();
	    send_more_msg = rc_msg_alloc( RC_MSG_SEND_MORE, 0, 0 );
	    hsm_process_event( dlm_recov_state_machine, 
			       (hsm_msg_t *) send_more_msg );
            break;
        }

        /* Scan resource table and send directory information for each
	 * resource to the appropriate node.
	 *
	 * In between each call, incoming messages update this node's
	 * directory. 
	 *
	 * These messages are generated by dlm_recov_dir_msg_retire when the
	 * number of in-flight messages drops below the magic threshold value.
         */

        case RC_MSG_SEND_MORE:
        {
	    struct resource * res;
	    unsigned   i;
	    bool_t   done;


	    done = FALSE;
	    for ( i = 0 ; i < dlm_buff_per_try ; i++ )
	    {
		done = iterator_next( iter, & res );
		if ( done )
		{
		    break;
		}

		if ( (res->rsrc_flags & RSRC_ACTIVE) != 0 )
		{
		    clmr_dir( res );
		}
	    }


	    /* If all of our directory updates have been send and retired,
	     * move on to the next state.
	     */

	    if ( done && (atomic_read( & dlm_recov_updates_inflight ) == 0) )
	    {
		STATE_TRAN( me, RC_BARRIER3 );
	    }

	    rc_msg_free( msg );
            break;
        }

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* rc_barrier2 - Handle the RC_BARRIER2 state
 *
 */

hsm_msg_t const *
rc_barrier2( hsm_t * me, hsm_msg_t const * msg )
{
    switch( msg->evt )
    {
        case START_EVT:
        case EXIT_EVT:
        {
            break;
        }

        case ENTRY_EVT:
        {
	    grace_period = 1;
	    clm_active = TRUE;
	    clm_sync( dlm_recov_leaving_node, 2 );
            break;
        }

        case RC_MSG_GOT_BARRIER:
        {
	    if ( dlm_recov_in_purge )
	    {
		STATE_TRAN( me, RC_BARRIER6 );
	    }
	    else
	    {
		STATE_TRAN( me, RC_REBUILD_DIR );
	    }

	    rc_msg_free( msg );
            break;
        }

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* rc_purge_wait - Handle the RC_PURGE_WAIT state
 *
 */

hsm_msg_t const *
rc_purge_wait( hsm_t * me, hsm_msg_t const * msg )
{
    switch( msg->evt )
    {
        case START_EVT:
        {
            break;
        }

        case ENTRY_EVT:
        {
	    rc_msg_t * send_more_msg;


	    /* This operation *must* complete on the first try.
	     */

	    client_apply( reconfig_free_client_purge, NULL, NULL );


	    /* Send a RC_MSG_PURGE_CHECK message to the machine.  This is
	     * kind of a hack, but it works.
	     */

	    send_more_msg = rc_msg_alloc( RC_MSG_PURGE_CHECK, 0, 0 );
	    hsm_process_event( dlm_recov_state_machine, (hsm_msg_t *) send_more_msg );
            break;
        }

        case EXIT_EVT:
        {
            break;
        }

	/* Check to see if there are any purge operations still in progress.
	 * If not, then move on to the next state.
	 */

        case RC_MSG_PURGE_CHECK:
        {
	    if ( ! client_apply(reconfig_check_client_purge, NULL, NULL) )
	    {
		STATE_TRAN( me, RC_BARRIER2 );
	    }

	    rc_msg_free( msg );
            break;
        }

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* rc_barrier1 - Handle the RC_BARRIER1 state
 *
 */

hsm_msg_t const *
rc_barrier1( hsm_t * me, hsm_msg_t const * msg )
{
    switch( msg->evt )
    {
        case START_EVT:
        case EXIT_EVT:
	{
	    break;
	}

	/* Clean out the in-flight queue, locks owned by dead node(s), and
	 * directory entries pointing to dead node(s).
	 */

        case ENTRY_EVT:
	{
	    struct resource * res;
	    iterator_t * iter;


	    dlm_recov_can_del_dir = FALSE;


	    /* Any requests that have node as a destination should be reset.
	     */

	    reset_inflight( dlm_recov_leaving_node );


	    /* Kill off any locks that belong to the dead node *BEFORE* it
	     * gets the chance to restart.  The "before" part is *VERY*
	     * important!!!!!
	     *
	     * this doesn't need to be counted because it's all internal
	     */

	    iter = resource_table_iterator_create();
	    while ( ! iterator_next( iter, & res ) )
	    {
		clmr_purge_res( res );
	    }

	    iterator_destroy( iter );


	    /* While we're cleaning up directories, why don't we go through
	     * our local RLDB and delete any directory entries that pointed to
	     * nodes that are no longer active.
	     */

	    rl_enum( clmr_dir_purge_chk, RLENUM_ALLNODES, 0 );

	    clm_sync( dlm_recov_leaving_node, 1 );
	    break;
	}

        case RC_MSG_GOT_BARRIER:
        {
	    STATE_TRAN( me, RC_PURGE_WAIT );
	    rc_msg_free( msg );
            break;
        }

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* rc_dir_init - Handle the RC_DIR_INIT state
 *
 */

hsm_msg_t const *
rc_dir_init( hsm_t * me, hsm_msg_t const * msg )
{
    rc_msg_t * rc_msg;


    rc_msg = (rc_msg_t *) msg;
    switch( msg->evt )
    {
        case START_EVT:
        case EXIT_EVT:
        {
            break;
        }

        /* Both states do the same thing.  RC_MSG_DIR_RETRY is an internally
	 * generated message. This is sent sometime after receiving a
	 * RC_MSG_DIR_NAK message from our chosen directory master. We need
	 * to as the master for directory data again. 
         */

        case ENTRY_EVT:
	{
	    purge_init();

	    dlm_recov_source_node = dlm_nodeid_next( OurSiteID );

	    /*FALLTHROUGH*/
	}

        case RC_MSG_DIR_RETRY:
        {
	    clm_msg_t * retry_msg;
	    int   status;


	    CLM_ASSERT( dlm_recov_source_node != OurSiteID,
		        "No directory recovery node available!" );

	    status = CPERR_NOMEM;
	    retry_msg = cccp_msg_create( PORT_RECOVERY, DIRRPRI,
					 dlm_recov_source_node,
					 CLM_MSG_NODE_UP, 0,
					 dlm_recov_restart_msg_retire, 
					 NULL, NULL );
	    if ( retry_msg != NULL )
	    {
		bsdlog( LOG_INFO, "[%s:%s] sending directory request to node "
		        "%u\n", current->comm, __FUNCTION__,
		        dlm_recov_source_node );

		status = cccp_send( retry_msg );
	    }


	    if ( status != CPERR_OK )
	    {
		bsdlog( LOG_CRIT, "%s: cccp_send(%d)\n", __FUNCTION__,
		       status );
		dlm_recov_set_retry_timer();
	    }


	    /* This code falls through from the ENTRY_EVT case.  We CANNOT
	     * free that type of message!
	     */

	    if ( msg->evt >= 0 )
	    {
		rc_msg_free( msg );
	    }

            break;
        }

        /* The directory master will send this type of message if it is not 
         * in a state where it can send us directory information. We'll wait 
         * for awhile and re-send our request. 
         */

        case RC_MSG_DIR_NAK:
        {
	    dlm_recov_set_retry_timer();
	    rc_msg_free( msg );
            break;
        }

        case RC_MSG_DIR_NOT_UP:
	{
	    unsigned   next;

	    next = dlm_nodeid_next( dlm_recov_source_node );
	    if ( next == OurSiteID )
	    {
		/* If dlm_recov_source_node is greater than our local node
		 * ID, then we must have the lowest active node ID in the
		 * cluster.  In that case, we can declare that nobody has
		 * any state.  In this case, we will act as if we were the
		 * only node in the cluster (see rc_init_wait).
		 */

		if (  dlm_recov_source_node > OurSiteID )
		{
		    dlm_recov_source_node = 0;
		    STATE_TRAN( me, CLM_ST_RUN );
		}
	    }
	    else
	    {
		dlm_recov_source_node = next;
	    }

	    if ( dlm_recov_source_node != 0 )
	    {
		dlm_recov_set_retry_timer();
	    }

	    rc_msg_free( msg );
            break;
	}

        /* The directory master sends this message after sending the last
         * message containing actual directory information. 
         */

        case RC_MSG_DIR_END:
        {
	    /* local request for migration control block */
	    clmm_cb_t      tmp_cb;

	    rl_summary_t * rls;


	    rls = (rl_summary_t *) rc_msg->data;
	    (void) memset( & tmp_cb, 0, sizeof( clmm_cb_t ) );
	    tmp_cb.mcb_event_run = rls->rls_event_run;
	    tmp_cb.mcb_aps_decay = rls->rls_aps_decay;
	    clmm_ctrl(&tmp_cb, CLMM_UPDGCB);

	    DBPRINTF((DBG_CONFIG,
		      "Read directory end msg from %d  run %d, aps %s\n",
		      rc_msg->node, tmp_cb.mcb_event_run,
		      kfl_ftos(tmp_cb.mcb_aps_decay)));

	    if ( rc_msg->node == dlm_recov_source_node )
	    {
		clmr_restore_2();
		STATE_TRAN( me, CLM_ST_RUN );
	    }

	    CLM_FREE( rc_msg->data );
	    rc_msg->data = NULL;
	    rc_msg_free( msg );
            break;
        }

        /* These are the messages that contain actual directory information. 
         * Use this data to build the local copy of the directory. 
         */

        case RC_MSG_DIR_RESTORE:
        {
	    rl_update_t * update;
	    rl_info_t     rldbe;
	    int           status;

	    
	    update = (rl_update_t *) rc_msg->data;

	    /* Insert the entry into local directory.
	     */

	    rldbe.rl_name = &update->rl_name[0];
	    rldbe.rl_namelen = update->rl_namelen;
	    rldbe.rl_node = update->rl_node;
	    rldbe.rl_type = update->rl_type;
	    rldbe.rl_flags = RLINFO_DIRECTORY | RLINFO_TOUCHED;

	    status = rl_create( & rldbe );
	    if ( DLM_NORMAL != status )
	    {
		/* If this fails here, it is likely because of a directory
		 * request that snuck in before we got our complete dump.
		 */

		bsdlog( LOG_ERR, "clmr_cntrl: rl_create error ret=%d", 
		        status );
		abort(__FUNCTION__);
	    }

	    DBPRINTF((DBG_DIRECT, "Got a directory entry from %d\n", 
		      rc_msg->node));

	    CLM_FREE( rc_msg->data );
	    rc_msg->data = NULL;
	    rc_msg_free( msg );
            break;
        }


	/* If the node is in RC_DIR_INIT then it has no information that it
	 * could possibly send to another new node.  Let the other node know
	 * that it needs to look elsewhere.
	 */

	case CLM_MSG_NODE_UP:
	{
	    dlm_recovery_send_node_up_nak( rc_msg->node,
					   RC_MSG_DIR_NOT_UP );
	    rc_msg_free( msg );
	    break;
	}

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/
/* rc_init_wait - Handle the RC_INIT_WAIT state
 *
 */

hsm_msg_t const *
rc_init_wait( hsm_t * me, hsm_msg_t const * msg )
{
    switch( msg->evt )
    {
        case START_EVT:
        case ENTRY_EVT:
        {
            break;
        }


	/* On the way out of this state we need to initialize CCCP.
	 */

        case EXIT_EVT:
        {
	    if ( clm_config_inited() )
	    {
		extern int cccp_init( void );


		(void) cccp_init();
	    }
	    else
	    {
		bsdlog( LOG_WARNING, "DLM configuration NOT initialized!\n" );
	    }

            break;
        }

        /* When we get the CLM_MSG_DO_NODE_INIT message, we have a decision 
         * to make. If we are the only node in the cluster, then we can 
         * simply enter the CLM_ST_RUN state. Otherwise, we have to ask some 
         * other node for directory information and entry the RC_DIR_INIT 
         * state.
         */

        case CLM_MSG_NODE_INIT:
        {
	    dlm_recov_source_node = dlm_nodeid_next( OurSiteID );
	    if ( dlm_recov_source_node == OurSiteID )
	    {
		/* We *must* be alone in the cluster.  In this case, we have
		 * all the directory state that is available, can should go
		 * directly to CLM_ST_RUN.
		 */

		dlm_recov_source_node = 0;
		STATE_TRAN( me, CLM_ST_RUN );
	    }
	    else
	    {
		/* We're not alone!  Move to RC_DIR_INIT!
		 */

		STATE_TRAN( me, RC_DIR_INIT );
	    }

	    rc_msg_free( msg );
            break;
        }

        default:
        {
            return( msg );
        }
    }

    return( NULL );
}




/****************************************************************************/

static void
dlm_recov_retry_timer( unsigned timer )
{
    struct timer_list * t;
    rc_msg_t * msg;


    msg = rc_msg_alloc( RC_MSG_DIR_RETRY, 0, NULL );

    t = (struct timer_list *) timer;
    t->data = 0UL;

    /* BUGBUG It may not be wise to do this in this context. */
    hsm_process_event( dlm_recov_state_machine, (hsm_msg_t *) msg );
}

void
dlm_recov_set_retry_timer( void )
{
    static struct timer_list  t = { data: 0UL };


    if ( t.data == 0UL )
    {
        init_timer( & t );
        t.function = (void *) dlm_recov_retry_timer;
        t.expires = jiffies + ( HZ * 3 );
        t.data = (ulong) & t;

        add_timer( & t );
    }
}

/****************************************************************************/
/* clmr_restore_2
 *
 * send RC_MSG_RES_RESTART (node join) message to all other lock managers
 * returns 0 for OK, 1 for error
 */

static int
clmr_restore_2(void)
{
    clm_msg_t  * msg;
    unsigned     node;
    iterator_t * iter;


    iter = dlm_nodeid_iterator_create();
    while ( ! iterator_next( iter, & node ) )
    {
	if ( (node != OurSiteID) && clm_node_active( node ) )
	{
	    /* allocate message header */
	    msg = cccp_msg_create( PORT_RECOVERY, DIRRPRI, node,
				   RC_MSG_RES_RESTART, 0,
				   dlm_recov_restart_msg_retire, NULL, NULL);
	    if ( msg == NULL) 
	    {
		bsdlog( LOG_CRIT, "[%s:%s]: Unable to create restart "
		        "message\n", current->comm, __FILE__ );
		iterator_destroy( iter );
		return 1 ;
	    } 

	    if ( cccp_send( msg ) != CPERR_OK )
	    {
		bsdlog(LOG_CRIT, "%s: cccp_send(%d)\n", __FUNCTION__, node);
	    }

	    DBPRINTF((DBG_CONFIG, "Sent restart message to %d\n", node));
	}
    }

    iterator_destroy( iter );

    /*
     * if any purges are necessary but were delayed, do them now
     */
    client_apply(reconfig_allow_client_purge, (void *) "dir read" , NULL) ;

    return 0 ;
}




/****************************************************************************/
/* clmr_dir_purge_chk - Delete the directory entry from dead node
 *
 * Deletes the directory entry if it points to a failed node.  This funcion
 * is intended to be called via the rl_enum function.
 *
 * RETURNS
 *
 * Zero is always returned.
 */

static int
clmr_dir_purge_chk(
		    /* Pointer to the RLDB entry to (possibly) be deleted.
		     */

		    rl_info_t * rldbe
		  )
{
    if ( ((rldbe->rl_flags & RLINFO_DIRECTORY) != 0) &&
	 !clm_node_active( rldbe->rl_node ) ) 
    {
	DBPRINTF((DBG_RECOV, "Purging dir entry (%s) -> %d\n",
		  printbin(rldbe->rl_name, rldbe->rl_namelen),
		  rldbe->rl_node));
	if (DLM_NORMAL != rl_delete(rldbe)) 
	{
	    bsdlog(LOG_CRIT, "Can't find stale dir entry in clmr_dir_purge");
	    abort("clmr_dir_purge_chk");
	}
    }

    return 0;
}




/****************************************************************************/
/* clmr_dir_move_chk - Delete migrated directory entries
 *
 * Deletes the directory entry if it has moved away from the local node.  This
 * funcion is intended to be called via the rl_enum function.
 *
 * RETURNS
 *
 * Zero is always returned.
 */

static int
clmr_dir_move_chk(
		   /* Pointer to the RLDB entry to (possibly) be deleted.
		    */

		   rl_info_t * rldbe
		 )
{
    int node ;

    if (rldbe->rl_flags & RLINFO_DIRECTORY)
    {
	node = clm_direct( rldbe->rl_name, rldbe->rl_namelen );
	if ( (node != 0) && (node != OurSiteID) )
	{
	    DBPRINTF((DBG_RECOV, "Purging moved dir entry (%s) -> %d\n",
		      printbin(rldbe->rl_name, rldbe->rl_namelen),
		      rldbe->rl_node)) ;
	    if (DLM_NORMAL != rl_delete(rldbe))
	    {
		bsdlog(LOG_CRIT, "Can't find stale dir entry in clmr_dir_move") ;
		abort("clmr_dir_move_chk") ;
	    }
	}
    }

    return 0 ;
}




/****************************************************************************/
/* clmr_dir_move_chk - Delete migrated directory entries
 *
 * Deletes the RLDB entry if it points to a failed node.  This funcion is intended
 * to be called via the rl_enum function.
 *
 * RETURNS
 *
 * Zero is always returned.
 */

static int
clmr_dir_purge(rl_info_t *rldbe)
{
    if (DLM_NORMAL != rl_delete(rldbe)) 
    {
	bsdlog( LOG_CRIT, "Can't find stale directory entry in clmr_dir_purge" );
	abort("clmr_dir_purge");
    }

    return 0;
}




/****************************************************************************/

static int
clmr_send_dir_entry(rl_info_t *rldbe)
{
    clm_msg_t *reply;
    rl_update_t *response;


    bsdlog( LOG_INFO, "[%s:%s] enter", current->comm, __FUNCTION__ );


    /* check if it should be shipped */
    if ( !(rldbe->rl_flags & RLINFO_DIRECTORY)
	 || clm_direct_dirsend(rldbe->rl_name, rldbe->rl_namelen, dlm_recov_joining_node )
	!= dlm_recov_joining_node )
    {
	bsdlog(LOG_INFO, "%s [%d]: not sending dir entry\n",
		  __FILE__, __LINE__ );
	goto done;
    }

    /* send it to the node */
    response = CLM_MALLOC( sizeof( *response ) );
    if ( response == NULL ) {
	bsdlog(LOG_CRIT, "%s [%d]: malloc(%d) failed\n",
		  __FILE__, __LINE__, sizeof( *response ));
	goto done;
    }


    /* Copy the correct information into the transfer structure.
     */

    resname_bcopy( rldbe->rl_name, response->rl_name, rldbe->rl_namelen );
    response->rl_namelen = rldbe->rl_namelen;
    response->rl_node = rldbe->rl_node;
    response->rl_type = rldbe->rl_type;


    /* Fill in the message header.
     */

    reply = cccp_msg_create( PORT_RECOVERY, DIRRPRI, dlm_recov_joining_node,
			     RC_MSG_DIR_RESTORE, sizeof( *response ),
			     dlm_recov_dir_msg_retire, response, NULL );
    if ( reply == NULL )
    {
	bsdlog(LOG_CRIT, "%s [%d]: malloc(%d) failed\n",
		  __FILE__, __LINE__, sizeof(clm_msg_t));
	CLM_FREE( response );
	goto done;
    }
    
    atomic_inc( & dlm_recov_updates_inflight );

    /* send it off */
    if ( cccp_send( reply ) != CPERR_OK )
    {
	bsdlog( LOG_CRIT, "%s: cccp_send(%s)\n",
	        __FUNCTION__, printbin(rldbe->rl_name, rldbe->rl_namelen));
    }

    DBPRINTF((DBG_DIRECT, "[%s:%s] %s dir entry sent to %d\n",
	      current->comm, __FUNCTION__,
	      printbin(rldbe->rl_name, rldbe->rl_namelen), 
	      dlm_recov_joining_node ));

    /* mark it as a cache entry */
    rldbe->rl_flags &= ~RLINFO_DIRECTORY;
    rl_decrement_dir_count(1);

done:
    return 0;
}




/****************************************************************************/
/* clmr_end_new_node
 *
 * finishes distribution of directory from this node to a node
 * which just started by sending a RC_MSG_DIR_END message to it
 * receiver (new node) will broadcast RC_MSG_RES_RESTART
 *
 * return 0 for OK, 1 for error
 */

static int
clmr_end_new_node( int node )
{
    rl_summary_t *rls;
    clm_msg_t *msg_xmit;
    clmm_cb_t tmp_cb;		/* local request for migration control block */
    int ret ;

    /* allocate rl_summary struct */
    if ((rls = CLM_MALLOC(sizeof(rl_summary_t))) == NULL)
    {
	bsdlog(LOG_CRIT, "%s [%d]: malloc(%d) failure\n",
		  __FILE__, __LINE__, sizeof(rl_summary_t));
	abort("clmr_end_new_node: malloc failed") ;
    }

    (void) memset( rls, 0, sizeof( rl_summary_t ) );

    /* prepare data */
    (void) memset( & tmp_cb, 0, sizeof( clmm_cb_t ) );
    clmm_ctrl(&tmp_cb, CLMM_GETGCB);
    rls->rls_event_run = tmp_cb.mcb_event_run;
    rls->rls_aps_decay = tmp_cb.mcb_aps_decay;

    /* prepare message */
    msg_xmit = cccp_msg_create( PORT_RECOVERY, DIRRPRI, node, RC_MSG_DIR_END,
			        sizeof(rl_summary_t), NULL, rls, NULL );
    if ( msg_xmit == NULL )
    {
	CLM_FREE( rls );
	bsdlog(LOG_CRIT, "%s [%d]: malloc(%d) failed\n",
		  __FILE__, __LINE__, sizeof(clm_msg_t));
	abort("clmr_end_new_node: malloc failed") ;
    }

    /* send it off */
    if ( (ret = cccp_send(msg_xmit)) != CPERR_OK )
    {
	bsdlog(LOG_CRIT, "%s: cccp_send to node %d got error %d\n",
		__FUNCTION__, node, ret) ;
	return 1 ;
    }

    /*
     * if any purges are necessary but were delayed, do them now
     */
    client_apply(reconfig_allow_client_purge, (void *) "dir dump" , NULL) ;

    return 0 ;
}




/****************************************************************************/
/* clmr_message_handler - Callback for PORT_RECOVERY
 *
 * All messages for the DLM recovery process are handled by this function.
 * For the most part, these messages are just passed into the DLM recovery
 * state-machine.  The sole exception is the RC_MSG_RES_RESTART message, which
 * is handled in-line.
 * 
 * RETURNS
 * 
 * Zero is always returned.
 */

static int
clmr_message_handler( clm_msg_t * msg )
{
    rc_msg_t * dat;
    void * data;


    DBPRINTF((DBG_FUNC, "Enter clmr_cntrl msg=0x%x type=%d", msg, msg->cm_type));

    switch ( msg->cm_type )
    {
	case CLM_MSG_NODE_UP:
	case RC_MSG_DIR_RESTORE:
	case RC_MSG_DIR_END:
	case RC_MSG_DIR_NAK:
	case RC_MSG_DIR_NOT_UP:
	case RC_MSG_DIR_DATA:      /* directory query response */
	case RC_MSG_DIR_REQUEST:   /* initial directory query */
	case RC_MSG_RES_REQ_1ST:   /* new data to merge */
	case RC_MSG_RES_REQ:       /* new data to merge */
	case RC_MSG_RES_DATA:      /* master handle response */
	{
	    /* Wrap the messages with a rc_msg_t and pass them into the state
	     * machine.
	     */

	    data = NULL;
	    if ( msg->cm_length != 0 )
	    {
		data = CLM_MALLOC( msg->cm_length );
		(void) memcpy( data, msg->cm_data, msg->cm_length );
	    }

	    dat = rc_msg_alloc( msg->cm_type, msg->cm_node, data );

	    acquire_lock( LK_ENTRY, kproc_pid );
	    hsm_process_event( dlm_recov_state_machine, (hsm_msg_t *) dat );
	    release_lock( LK_ENTRY, kproc_pid );
	    break;
	}

	/* new node starting	*/
	case RC_MSG_RES_RESTART:
	{
	    DBPRINTF((DBG_CONFIG, "Read directory restart msg from %d\n",
		      msg->cm_node));

	    /* Some node has just told us that it is open for business.  There's
	     * a remote possibility that we are *still* handling its node down
	     * event.  Just in case, wipe out any directory entries that point
	     * to it.  If it's just starting, then there's no way it should be
	     * mastering any resources yet.
	     */

	    if (msg->cm_node)
	    {
		acquire_lock( LK_ENTRY, kproc_pid );
		rl_enum(clmr_dir_purge, RLENUM_SPECNODES, msg->cm_node);
		release_lock( LK_ENTRY, kproc_pid );
	    }

	    break;
	}

	default:
	{
	    bsdlog(LOG_CRIT, "%s: unknown message type [0x%lx]\n",
		   __FUNCTION__, msg->cm_type);
	    break;
	}
    }


    return 0;
}




/****************************************************************************/
/* clmr_rm_locallock_dirs -
 *
 * This function removes directories temporarily created for local
 * locks during reconfig.
 */

static int
clmr_rm_locallock_dirs (rl_info_t *rl_entry)
{
    /* if directory for local lock delete */
    if (rl_entry->rl_flags & RLINFO_LOCAL) 
    {
	DBPRINTF((DBG_LOCAL, "Removing local dir entry (%s)->%d",
		  printbin(rl_entry->rl_name, rl_entry->rl_namelen),
		  rl_entry->rl_node));
	rl_delete (rl_entry);
    }

    return 0 ;
}

	    
	    
	    
/****************************************************************************/
/* clmr_check_lkvlb - Verify lock value block state for resource
 *
 * Verifies the state of the lock value block for the resource.  It first checks
 * for locks which constrain the lock value block's correctness on the grant and
 * convert queues.  The number of valid reclocks found is returned.
 *
 * RETURNS
 *
 * Number of valid reclocks found.
 */

static int
clmr_check_lkvlb(
		  /* Resource structure to check for rebuild.
		   */

		  struct resource * resptr
		)
{
    struct reclock *rp;
    int retval = 0;
    int qtype;          /* reclock queue type */

    for (qtype = 0; qtype <= RLQ_CONV; qtype++)
    {
	for ( EACH_DLM_LISTNODE( & resptr->lockqs[qtype], rp ) )
	{
	    if (rp->lck.mode >= LKM_CWMODE)
		retval++;
	}
    }

    return retval ;
}




/****************************************************************************/
/* clmr_lkvlb -
 *
 *	Called for each resource.  If the master of this resource has died
 *      and mastery of the resource has been reassigned to this node, check
 *      for correctness on the grant and convert queues.  If no lock w/ mode
 *      >= CW found, then invalidate lock value block, because former master
 *      could have held a PW or EX lock on the resource.
 */

void
clmr_lkvlb(
	    /* Pointer to a resource structure to check for lock value block
	     * correctness.
	     */

	    struct resource * resptr
	  )
{
    struct reclock * a;
    dlm_listnode_t * curr;
    dlm_listnode_t * next;
    unsigned         qtype;          /* reclock queue type */


    DBPRINTF((DBG_RECOV,
	     "%s: res=%s Checking lock value block status", __FUNCTION__,
	      printbin(resptr->name, resptr->namelen)));

    /* only need to operate on master copies */
    if (resptr->master == 0) {

	/* clear stale locks on each queue */
	for (qtype = 0; qtype < RLQ_MAX; qtype++)
	{
	    for ( curr = dlm_list_get_head( & resptr->lockqs[ qtype ] )
		  ; curr != NULL
		  ; curr = next )
	    {
		next = dlm_list_succ( curr );
		a = (struct reclock *) curr;

		if (a->lck.state & LOCK_STALE) {
			remove_queue(&(resptr->lockqs[qtype]), a, RL_RQ_OFS) ;
			resptr->refcount-- ;
			free_le(DLM_RES_VMS, a);
		} else {
			a->lck.state &= ~LOCK_REFRESHED;
		}
	    }
	}

	/* Only check lvb if we're the new master */
	if (resptr->rsrc_flags & RSRC_NEW_MASTER) {

		/* Any lock of mode >= CW exist? */
		if (clmr_check_lkvlb(resptr) == 0) {
		    DBPRINTF((DBG_RECOV,
			      "%s: res=%s Invalidating lock value block", __FUNCTION__,
			      printbin(resptr->name, resptr->namelen)));
		    clear_value_block(resptr);
		}

		resptr->rsrc_flags &= ~RSRC_NEW_MASTER;
	}
    } else {
	    /* bogus, I know, but you'll get over it */
	    clmm_master2slave(resptr, resptr->master, FALSE);
    }
}




/****************************************************************************/

int
clmr_dir(
	  struct resource * resptr
	)
{
    /* Clear possible leftover updating flag.
     */

    resptr->rsrc_flags &= ~RSRC_UPDATING;
    
    if ( resptr->master == 0 )
    {
	CLM_ASSERT( (resptr->rsrc_flags & RSRC_FRZEN) == 0,
		    "resource has RSRC_FRZEN flag in " __FUNCTION__ );
	
	/* Update the new directory.
	 */
	
	update_directory( resptr, OurSiteID, RESRPRI,
			  dlm_recov_dir_msg_retire );
	atomic_inc( & dlm_recov_updates_inflight );
    }
    else
    {
	clmm_master2slave( resptr, resptr->master, FALSE );
    }

    return( 0 );
}




void
dlm_recov_dir_msg_retire(
			  clm_msg_t * msg,
			  int stat
			)
{
    rc_msg_t * send_more_msg;


    if ( msg->cm_data )
    {
	CLM_FREE( msg->cm_data );
    }

    cccp_msg_destroy( msg );
    
    atomic_dec( & dlm_recov_updates_inflight );
    if ( atomic_read( & dlm_recov_updates_inflight ) < dlm_buff_limit )
    {
	/* There are fewer than the threshold number of messages still in
	 * flight, so more messages need to go out.  Queue a RC_MSG_SEND_MORE
	 * message in the state machine.
	 */

	send_more_msg = rc_msg_alloc( RC_MSG_SEND_MORE, 0, 0 );
	hsm_process_event( dlm_recov_state_machine, (hsm_msg_t *) send_more_msg );
    }
}




/*
 * Name
 * 	clmr_res
 * Description
 *	receives as input a list of nodes.
 *	It returns 0 upon success, and -1 in the event of failure.
 *
 *	For each resource in the resource table:
 *	We query the directory for that resource, asking
 *	which node is the master.
 *	The response will come in a callback that will handle
 *	the rest of the operation, but basically:
 *
 *	If the master of the resource is the local node, then
 *	if the resource is frozen we downgrade the mastery, and
 *	continue as if the resource was mastered by a remote node.
 *	Otherwise, we scan the the resource structure, removing all
 *	locks held by failed nodes.
 *
 *	If the directory appoints
 *	us as the master, then we proceed to act as such, otherwise
 *	we send all our info to the new master.
 *
 * Arguments
 *	resptr - pointer to a resource structure to check for rebuild
 * Returns
 *	Nothing
 * Global Data Affected
 *	NONE
 */
static int
clmr_res( struct resource * resptr )
{
    clm_msg_t *msg_xmit;
    int node = clm_direct(resptr->name, resptr->namelen);
    struct reclock	*a;
    bool_t	gone;
    int qtype;          /* reclock queue type */


    DBPRINTF((DBG_FUNC|DBG_RECOV,
		"%s: process (%s:%x) m=%d f=0x%x", __FUNCTION__,
		printbin(resptr->name, resptr->namelen), (resptr->rh.rh&0xff),
		resptr->master, resptr->rsrc_flags));

    /* if we're the master copy, then this isn't necessary */
    if (resptr->master == 0) 
    {
	/* Clean out the queues in the resource.
	 */

	clmr_clean( resptr );
	return 0;
    }

    /* check to see if all of the locks on this resource are LOCK_GONE */
    for ( gone = TRUE, qtype = 0 ; (qtype < RLQ_MAX) && gone ; qtype++)
    {
	for ( EACH_DLM_LISTNODE( & resptr->lockqs[qtype], a ) )
	{
	    if (~a->lck.state & LOCK_GONE)
	    {
		gone = FALSE;
		break;
	    }
	}
    }

    /*
     * if "gone" then we may have a local copy with no master - we can't
     * rebuild it and we shouldn't try.
     */
    if ( ! gone )
    {
	rl_update_t * update;


	/* prepare query */

	update = CLM_MALLOC( sizeof( *update ) );
	if ( update == NULL )
	{
	    bsdlog(LOG_CRIT, "%s [%d]: malloc(%d) failure\n",
		   __FILE__, __LINE__, sizeof( *update ));
	    return -1 ;
	}

	(void) memset( update, 0, sizeof( *update ) );
	resname_bcopy(resptr->name, update->rl_name, resptr->namelen);
	update->rl_namelen = resptr->namelen;
	update->rl_flags = resptr->rsrc_flags;
	update->rl_type = resptr->type;


	/* allocate msg struct */
	msg_xmit = cccp_msg_create( PORT_RECOVERY, RESRPRI, node,
				    RC_MSG_DIR_REQUEST, sizeof( *update ),
				    dlm_recov_dir_msg_retire,
				    update, NULL );
	if ( msg_xmit == NULL) {
	    bsdlog(LOG_CRIT, "%s [%d]: malloc(%d) failed\n",
		   __FILE__, __LINE__, sizeof(clm_msg_t));
	    return(-1);
	}

	DBPRINTF((DBG_RECOV, "clmr_res: send query for (%s:%x) to %d\n",
		  printbin(resptr->name, resptr->namelen), (resptr->rh.rh&0xff),
		  msg_xmit->cm_node));

	atomic_inc( & dlm_recov_updates_inflight );


	/* Send it off.
	 */

	if ( cccp_send(msg_xmit) != CPERR_OK )
	{
	    bsdlog(LOG_CRIT, "%s: cccp_send (%s)\n",
		   __FUNCTION__, printbin(resptr->name, resptr->namelen));

	    cccp_msg_destroy( msg_xmit );
	    CLM_FREE( update );
	    atomic_dec( & dlm_recov_updates_inflight );
	    return -1;
	}
    }

    return 0 ;
}




/****************************************************************************/
/* clmr_clean_queue - Clean one queue in a resource
 *
 * Removes all locks belonging to failed nodes from the specified queue.
 * All removed locks are properly freed.
 *
 * RETURNS
 *
 * The number of locks held on dead nodes that were in either the PW or EX
 * state.
 */

unsigned
clmr_clean_queue(
		  /* Resource to be cleaned.
		   */

		  struct resource * rp,


		  /* Queue in the resource to be cleaned.  Must be an
		   * element of the 'lockqs' array.
		   */

                  dlm_list_t * plock_queue 
		)
{
    struct reclock * lock;
    dlm_listnode_t * curr;
    dlm_listnode_t * next;
    unsigned         count;


    /* clear out from the start of the chain */
    
    count = 0;
    for ( curr = dlm_list_get_head( plock_queue )
	  ; curr != NULL
	  ; curr = next )
    {
	next = dlm_list_succ( curr );

	lock = (struct reclock *) curr;

	if ( ! clm_node_active( lock->lck.site ) )
	{
	    /* Check if failed node owned PW or EX lock.
	     */

	    if ( lock->lck.mode >= LKM_PWMODE )
	    {
		count++;
	    }

	    remove_queue(plock_queue, lock, RL_RQ_OFS) ;
	    rp->refcount-- ;
	    free_le( rp->type, lock );
	}
    }


    return( count );
}




/****************************************************************************/
/* clmr_clean - Clean dead nodes out of a resource
 *
 * This routine scans through the locks for the resource and frees any locks
 * held by failed nodes.  If any of the locks were held at either PW or EX,
 * the value block of the resource will be invalidated (cleared).
 */

void
clmr_clean(
	    /* Resource to be cleaned.
	     */

	    struct resource * resptr 
	  )
{
    unsigned   qtype;          /* reclock queue type */
    unsigned   count;


    count = 0;
    for ( qtype = 0 ; qtype < RLQ_MAX ; qtype++ )
    {
	count += clmr_clean_queue( resptr, &(resptr->lockqs[qtype]) );
    }


    /* If an EX or PW lock owned by a failed node is found, then invalidate
     * the value block.
     */

    if ( count != 0 )
    {
	clear_value_block( resptr );
    }
}




/****************************************************************************/
/* clmr_purge_res -
 *
 * Purge any locks on the provided resource that are owned by clients on
 * dead nodes.
 */

void
clmr_purge_res( struct resource * resptr )
{
    DBPRINTF((DBG_FUNC|DBG_RECOV, "%s: process (%s:%x) m=%d f=0x%x rc=%d",
	      __FUNCTION__, printbin(resptr->name, resptr->namelen),
	      (resptr->rh.rh&0xff), resptr->master, resptr->rsrc_flags,
	      resptr->refcount));


    /* Clean out the queues in the resource.
     */

    clmr_clean( resptr );

    resptr->rsrc_flags &= ~(RSRC_STUCK | RSRC_UPDATING | RSRC_FRZEN);
}




/* dlm_recov_restart_msg_retire
 *
 * This is the callback that should be called when a restart message is
 * received on the destination node.  We may need to track them to insure
 * that all other nodes come out of reconfiguration when we tell them to.
 */

void
dlm_recov_restart_msg_retire(
			      /* Message that has been received.
			       */

			      clm_msg_t *msg,


			      /* Status of the message.
			       */

			      int stat
			    )
{
    DBPRINTF((DBG_CONFIG, "%s: Restart message to node %d confirmed\n",
	      __FUNCTION__, msg->cm_node));

    /* check status */
    if (stat) 
    {
	bsdlog( LOG_ERR, "Bad status %d in %s", stat, __FUNCTION__ );
    }

    /* free memory */
    if ( msg != NULL )
    {
	if ( msg->cm_data != NULL )
	{
	    CLM_FREE( msg->cm_data );
	}

	cccp_msg_destroy( msg );
    }
}




/****************************************************************************/
/* clmr_mark_stale -
 *
 *	This function is called before resource rebuild information that
 *  	has been read from another node is incorporated into the local
 *	copy of the resource.  If checks to see if the locks owned by the
 *	specified site have already been marked and if not, marks them
 *	all as STALE.  This bit is cleared from each lock as it is updated
 *	from the rebuild packet(s).  Any STALE locks that are left after
 *	rebuild is complete will be purged.
 * Arguments
 *	resptr - pointer to the resource to check for dir update
 *	site - unmarked locks from this site should be marked STALE.
 * Returns
 *	Nothing
 * Global Data Affected
 *	NONE
 */

void
clmr_mark_stale(struct resource *resptr, int site)
{
    struct reclock	*a;
    int 		qtype;        /* reclock queue type */

    /* Check lock queues for stale locks */
    for (qtype = 0; qtype < RLQ_MAX; qtype++)
    {
	for ( EACH_DLM_LISTNODE( & resptr->lockqs[qtype], a ) )
	{
	    if (a->lck.site == site) {
		/* if any locks have been marked, no need to continue */
		if (a->lck.state & (LOCK_STALE | LOCK_REFRESHED)) {
		    return;
		} else {
		    a->lck.state |= LOCK_STALE;
		}
	    }
	}
    }
}




void 
handle_resource_request( int type, clmr_data_t * data, unsigned node )
{
    struct resource *rptr;


    /* retrieve the resource pointer */

    rptr = find_resource( data->creb_name, data->creb_namelen, 
			  data->creb_type );
    if ( rptr == NULL ) {
	bsdlog(LOG_CRIT, "%s: bad lookup name (%s)\n",
	       __FUNCTION__, printbin(data->creb_name,
				      data->creb_namelen));
	abort("clmm_merge: bad lookup");
	return;
    }


    /* if 1st message, then call clmr_mark_stale to handle stale locks
     */
    if ( type == RC_MSG_RES_REQ_1ST )
    {
	clmr_mark_stale(rptr, node);
    }


    DBPRINTF((DBG_RECOV, "Merge in remote copy of (%s:%x) from %d\n",
	      printbin(rptr->name, rptr->namelen), (rptr->rh.rh&0xff),
	      node));

    if (data->creb_csize > MAX_CELLS) {
	bsdlog(LOG_CRIT, "%s: size skew %d [local %d]\n",
	       __FUNCTION__, data->creb_csize, MAX_CELLS);
	return;
    }

    /* now merge the data */
    clmr_chain(rptr, data);

    /* return the handle */
    clmr_respond(rptr, data, node);
}





/* clmr_respond -
 *
 *	is used to send the master's resource handle back to the node
 *	that sent in relevant resource data.  This routine allocates
 *	a message and a response structure, proceeding to fill them both
 *	and transmit the message.
 * Arguments
 *	resptr - pointer to resource whose handle needs to be sent
 *	msg - pointer to the query message
 * Returns
 *	Nothing
 * Global Data Affected
 *	NONE
 */
void
clmr_respond(struct resource *resptr, clmr_data_t * data, unsigned node )
{
    clr_res_t *res;
    clm_msg_t *reply;


    /* setup response */
    if ((res = CLM_MALLOC(sizeof(*res))) == NULL) {
	bsdlog(LOG_CRIT, "%s [%d]: malloc(%d) failure\n",
		  __FILE__, __LINE__, sizeof(*res));
	return;
    }
    res->clr_resptr = data->creb_rptr;
    (void) memcpy( & res->clr_mst_rh, & resptr->rh, sizeof( union dlm_rh ) );

    /* allocate message header */
    reply = cccp_msg_create( PORT_RECOVERY, RESRPRI, node, RC_MSG_RES_DATA,
			     sizeof( clr_res_t ), NULL, res, NULL );
    if ( reply == NULL) {
	bsdlog( LOG_CRIT, "%s [%d]: malloc(%d) failure\n",
	        __FILE__, __LINE__, sizeof(*reply));
	return;
    }

    if ( cccp_send(reply) != CPERR_OK ) {
	bsdlog(LOG_CRIT, "%s: cccp_send(%s)\n",
		  __FUNCTION__, printbin(resptr->name, resptr->namelen));
    }
}




/* 	clmr_chain
 * Description
 *	This function is used to take a block of reclock structures from
 *	a clmr_data_t and restore them into a new master.  Note that local
 *	reclocks override migrated reclocks.
 * Arguments
 *	rp - pointer to the resource to merge the lock onto
 *	msg - message pointer containing reclocks to be merged
 * Returns
 *	0
 * Global Data Affected
 *	NONE
 */

int
clmr_chain(struct resource *rp, clmr_data_t * data )
{
    int qtype;          /* reclock queue type */
    struct reclock *reclock_ptr;

    reclock_ptr = (struct reclock *) &data->creb_data ;

    for (qtype = 0; qtype < RLQ_MAX; qtype++)
    {
	clmr_chain_reclocks(rp, reclock_ptr, &(rp->lockqs[qtype]), data, qtype);
	reclock_ptr += data->creb_lockq_len[qtype] ;
    }

    return 0 ;
}

/*
 * Name
 * 	clmr_chain_reclocks
 * Description
 * Arguments
 *	rp - pointer to the resource to merge the lock onto
 *      mrglk - reclock to be merged
 *      plock_queue - pointer to the reclock queue (GRANT, CONV or WAIT)
 *      pdata - message data
 *      qtype - reclock queue type
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
clmr_chain_reclocks(struct resource *rp, struct reclock *mrglk,
		    dlm_list_t * plock_queue, const clmr_data_t * pdata,
		    int qtype)
{
    struct reclock *curlk;		/* lk ptr to walk queues */
    struct reclock *addlk;		/* new local lock pointer */
    uint i, id;				/* index and tmp lockid */
    short found;			/* true if local copy already exists */

    curlk = (struct reclock *) dlm_list_get_head( plock_queue );

    for (i = 0; i < pdata->creb_lockq_len[qtype]; i++, mrglk++) {
	found = FALSE;

	/* if VMS and we can lookup this lock, then there's no need to dup */
	if (rp->type == DLM_RES_VMS)
	{
	    if ((id = mapid(mrglk->lck.lockid, mrglk->lck.site)) &&
			(addlk = find_reclock(DLM_RES_VMS, id))) {
		addlk->lck.state = (addlk->lck.state | LOCK_REFRESHED) &
					~LOCK_STALE;
		continue;
	    }

	    /* get right to the spot */
	    while ( curlk != NULL )
	    {
		/* BUGBUG Do I also need to do this for unix locks? */
		/* The unixlock code doesn't pay attention to the curlk here */
		if (curlk->lck.remlockid == mrglk->lck.lockid) {
		    found = TRUE;
		    break;
		}
		if (curlk->lck.seqnum > mrglk->lck.seqnum)
		    break;

		curlk = (struct reclock *) dlm_list_succ( (dlm_listnode_t *)
							  curlk );
	    }

	    /* if the reclock doesn't exist, allocate it */
	    if (!found)
	    {
		clmr_alloc_reclock(rp, mrglk, plock_queue, curlk, qtype);
	    }
	}
	else if (rp->type == DLM_RES_UNIX)
	{
	    if (qtype == RLQ_GRANT)
	    {
		struct reclock *insrtp;

		if (blocked(rp, &insrtp, &mrglk->lck)) {
		    bsdlog(LOG_CRIT,"%s: blocked UNIX lock in merge", __FUNCTION__);
		    abort(__FUNCTION__);
		}
		if (add_reclock(rp, insrtp, &mrglk->lck) == -1) {
		    bsdlog(LOG_CRIT,"%s: unable to add UNIX lock in merge", __FUNCTION__);
		    abort(__FUNCTION__);
		}
	    }
	    else if (qtype == RLQ_WAIT)
	    {
		found = FALSE;

		/* check to see if lock already exists */
		for ( EACH_DLM_LISTNODE( & rp->lockqs[RLQ_WAIT], curlk ) )
		{
		    /* identify an identical lock owned by same process */
		    if (same_proc(&curlk->lck, &mrglk->lck) &&
			    mrglk->lck.lb == curlk->lck.lb &&
			    mrglk->lck.ub == curlk->lck.ub)
		    {
			found = TRUE;
			break;
		    }
		}
		/* if the reclock doesn't exist, allocate it */
		if (!found) {
		    addlk = get_le(DLM_RES_UNIX);
		    if (!addlk) {
			abort("NOLOCKS in clmr_chain");
		    }
		    (void) memcpy( &addlk->lck.rh, &rp->rh,
				   sizeof(addlk->lck.rh) );
		    clmm_copy(mrglk, addlk);
		/* do not add to lockid map, but fix remlockid */
		    addlk->lck.remlockid = mrglk->lck.lockid ;
		    rp->refcount++;
		    add_queue(plock_queue, addlk, RL_RQ_OFS) ;
		}
	    }
	}
	/* ??? What about a default case here */
    }

    /* ??? */
    /* Unix lock shouldn't have lock value block */
    if (qtype == RLQ_GRANT || qtype == RLQ_CONV)
    {
	/* check the lock value block */
	if (!(rp->rsrc_flags & RSRC_VALID) && clmr_check_lkvlb(rp) > 0) {
	    (void) memcpy( rp->value, pdata->creb_value, sizeof(rp->value) );
	    rp->rsrc_flags |= RSRC_VALID;
	}
    }

    check_handles(rp, 0) ;
}

/*
 * Name
 * 	clmr_alloc_reclocks
 * Description
 * Arguments
 *	rp - pointer to the resource to merge the lock onto
 *      mrglk - reclock to be merged
 *      plockq - pointer to the reclock queue (GRANT, CONV or WAIT)
 *      curlk - lk ptr to walk queues
 *      qtype - reclock queue type
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void
clmr_alloc_reclock(struct resource *rp, struct reclock *mrglk,
		   dlm_list_t *plockq, struct reclock *curlk, int qtype)
{
    struct reclock *addlk;

    addlk = get_le(rp->type);
    if ( ! addlk)
	abort("get_le failed in clmr_alloc_reclocks") ;

    addlk->lck.remlockid = mrglk->lck.lockid;
    (void) memcpy( &addlk->lck.rh, &rp->rh, sizeof(addlk->lck.rh) );
    clmm_copy(mrglk, addlk);
    rp->refcount++;
    if (qtype != RLQ_GRANT)
    {
	add_timeout_reclock(addlk);
    }
    addmap(addlk->lck.remlockid, addlk->lck.lockid, mrglk->lck.site) ;
    dlm_list_insert( plockq, (dlm_listnode_t *) addlk, 
		     (dlm_listnode_t *) curlk );
}




/****************************************************************************/
/* clmr_rebuild_send
 *
 *	This routine sends a copy of the local resource data to the master,
 *	who accepts the data and integrates it with the picture he has.
 *	Note that the sequence numbers are used to preserve lock ordering.
 *	Also note that multiple messages are sent if more the MAX_LOCKS
 *	reclocks exist.
 * Arguments
 *	resptr - pointer to the resource that needs to be shipped out
 *	master - node id of lock manager that is the master for this resource
 * Returns
 *	Nothing
 * Global Data Affected
 *	NONE
 */

int
clmr_rebuild_send(struct resource *resptr, int master)
{
  clmr_data_t *reb_data;		/* rebuild data struct	*/
  int datasz;				/* data size to be sent		*/
  struct reclock *lptr[RLQ_MAX];	/* array of ptrs to lock queues	*/
  char *buffer, *bptr;		/* infamous buffer ptrs		*/
  int reclock_ctr = 0;		/* counter for reclock limit    */
  clm_msg_t *msg_xmit;		/* outgoing msg		*/
  int qtype;          /* reclock queue type */
  int	msg_type;			/* type of rebuild msg sent */

  DBPRINTF((DBG_RECOV|DBG_FUNC,
            "enter clmr_rebuild_send: (%s:%x) send to %d\n",
            printbin(resptr->name, resptr->namelen),
            (resptr->rh.rh&0xff), master));

  /*
   * if this function is being called on a local copy of a resource then
   * it's fairly safe to assume that we are not going to be the master copy.
   * So fry any locks that are not local.
   */
  clmm_master2slave(resptr, master, FALSE);

    /* assign working pointers */
    for (qtype = 0; qtype < RLQ_MAX; qtype++) 
    {
	lptr[qtype] = (struct reclock *) dlm_list_get_head( & resptr->lockqs[qtype] );
    }

    while (reclock_ctr < resptr->refcount) 
    {
	int l_num = 0;			/* local reclock ctr	*/

	/* allocate and clear the memory for all data */
	datasz = sizeof(clmr_data_t);
	datasz += (min((resptr->refcount - reclock_ctr), MAX_LOCKS)
		   * sizeof(struct reclock));

	if ((buffer = CLM_MALLOC(datasz)) == NULL) 
	{
	    bsdlog(LOG_CRIT, "%s [%d]: malloc(%d) failed\n",
		   __FILE__, __LINE__, datasz);
	    return(-1);
	}

	(void) memset( buffer, 0, datasz );
	reb_data = (clmr_data_t *)buffer;


	/* fill in the data (grant, convert, wait) */
	bptr = (char*)&(reb_data->creb_data);

	for (qtype = 0; qtype < RLQ_MAX; qtype++) 
	{
	    if (lptr[qtype]) 
	    {
		clmr_bundle_reclocks(&(lptr[qtype]), &bptr, &reclock_ctr,
				     &l_num, &reb_data->creb_lockq_len[qtype] );
	    }
	}

	DBPRINTF((DBG_MIGR, "%s: backpatching header ref %d l %d c %d w %d\n",
		  __FUNCTION__, resptr->refcount,
		  reb_data->creb_lockq_len[RLQ_GRANT],
		  reb_data->creb_lockq_len[RLQ_CONV],
		  reb_data->creb_lockq_len[RLQ_WAIT]));

	/* backpatch header */
	reb_data->creb_csize = MAX_CELLS;
	reb_data->creb_rptr = resptr;
	reb_data->creb_flags = resptr->rsrc_flags;
	reb_data->creb_flags &= ~RSRC_FRZEN;	/* new copy is thawed	*/
	(void) memcpy( &reb_data->creb_value[0], &resptr->value[0],
		       sizeof(reb_data->creb_value) );
	resname_bcopy(resptr->name, &reb_data->creb_name[0], resptr->namelen);
	reb_data->creb_namelen = resptr->namelen;
	reb_data->creb_type = resptr->type;

	/* update the global reclock counter */
	reclock_ctr += l_num;

	/* use type to indicate if this message complete resource rebuild */

	msg_type = (reclock_ctr <= MAX_LOCKS)
	    ? RC_MSG_RES_REQ_1ST	/* first message */
	    : RC_MSG_RES_REQ;	/* other message */

	/* set up the actual message */
	msg_xmit = cccp_msg_create( PORT_RECOVERY, RESRPRI, master, msg_type,
				   datasz, dlm_recov_restart_msg_retire,
				   buffer, NULL );
	if ( msg_xmit == NULL )
	{
	    CLM_FREE( buffer );
	    bsdlog(LOG_CRIT, "%s [%d]: malloc(%d) failed\n",
		   __FILE__, __LINE__, sizeof(clm_msg_t));
	    return(-1);
	}

	/* send it off */
	if ( cccp_send(msg_xmit) != CPERR_OK )
	{
	    bsdlog(LOG_CRIT, "%s: cccp_send(%s)\n", __FUNCTION__,
		   printbin(resptr->name, resptr->namelen));
	    return -1 ;
	}
    }

    DBPRINTF((DBG_RECOV, "clmr_rebuild_send: (%s:%x) send to %d\n",
	      printbin(resptr->name, resptr->namelen),
	      (resptr->rh.rh&0xff), master));

    return 0 ;
}




/* 	clmr_bundle_reclocks
 * Description
 * Arguments
 *	prptr - pointer to the place in the reclock queue
 *      pbptr - pointer to the buffer pointer
 *	preclock_ctr -
 *      pl_num -
 *      pq_num -
 *	master - node id of lock manager that is the master for this resource
 * Returns
 *	Nothing
 * Global Data Affected
 *	NONE
 */

void
clmr_bundle_reclocks(struct reclock **prptr, char **pbptr, int *preclock_ctr,
		int *pl_num, int *pq_num )
{
    struct reclock *rptr = *prptr;
    char *bptr = *pbptr;
    int l_num = *pl_num;
    int q_num = *pq_num;

    /* ??? */
    /* Yuck.  Will revisit this */
    while (rptr != NULL && l_num < MAX_LOCKS) 
    {
	if (rptr->lck.state & LOCK_GONE) 
	{
	    (*preclock_ctr)++;
	}
	else
	{
	    (void) memcpy( bptr, rptr, sizeof(struct reclock));
	    bptr += sizeof(struct reclock);
	    q_num++;
	    l_num++;
	}

	rptr = (struct reclock *) dlm_list_succ( (dlm_listnode_t *) rptr );
    }

    *pbptr = bptr;
    *prptr = rptr;
    *pl_num = l_num;
    *pq_num = q_num;
}




/****************************************************************************/

void
handle_dir_request( rc_msg_t * dat )
{
    rl_info_t     rldbe;
    rl_update_t * update;
    rl_update_t * response;
    clm_msg_t   * reply;
    int	          dir_node;


    update = (rl_update_t *) dat->data;

    /* lookup original resource */
    (void) memset( &rldbe, 0, sizeof( rl_info_t ) );
    rldbe.rl_name = update->rl_name;
    rldbe.rl_namelen = update->rl_namelen;
    rldbe.rl_type = update->rl_type;


    /* verify that this query was sent to the correct directory node */
    dir_node = clm_direct(rldbe.rl_name, rldbe.rl_namelen);
    if (dir_node != OurSiteID) {
	DBPRINTF((DBG_DIRECT, "%s: (%s) Not directory.  Direct to %d",
		  __FUNCTION__, printbin(rldbe.rl_name, rldbe.rl_namelen),
		  dir_node));

	/* BUGBUG This needs to be fixed! */
#if 0
	msg->cm_node = dir_node;
	return dir_node;
#endif
    }

    /* wayward query could be asking us about a resource that should
     * know about, but don't yet.  Send back a 0 so it will be ignored
     * and retried
     */
#if 0
    if (rebuild_phase < RC_RESOURCE_QUERY)
#else
    if ( (STATE_CURR( dlm_recov_state_machine ) != RC_RESOURCE_QUERY)
	 && (STATE_CURR( dlm_recov_state_machine ) != RC_BARRIER4)
	 && (STATE_CURR( dlm_recov_state_machine ) != RC_BARRIER5)
	 && (STATE_CURR( dlm_recov_state_machine ) != RC_BARRIER6) )
#endif
    {
	rldbe.rl_node = 0;
	bsdlog(LOG_ERR, "Query received from %d after restart\n",
	       dat->node);
    }
    else if (rl_lookup(&rldbe) != DLM_NORMAL) 
    {
	/* here if no entry exists, or the node has failed so set one */
	rldbe.rl_node = dat->node;
	rldbe.rl_flags = RLINFO_DIRECTORY;
	if ( update->rl_flags & RSRC_LOCAL )
	{
	    /* if resource local lock then set flag in directory */
	    rldbe.rl_flags |= RLINFO_LOCAL;
	}

	if (DLM_NORMAL != rl_create(&rldbe)) {
	    bsdlog(LOG_CRIT, "%s: rl_create error", __FUNCTION__);
	    abort(__FUNCTION__);
	}
    }
    else if ( ! clm_node_active( rldbe.rl_node )
	      || !(rldbe.rl_flags & RLINFO_DIRECTORY) )
    {
	rldbe.rl_node = dat->node;
	rldbe.rl_flags = RLINFO_DIRECTORY;
	if (DLM_NORMAL != rl_modify(&rldbe)) 
	{
	    bsdlog(LOG_CRIT, "%s: rl_modify error", __FUNCTION__);
	    abort(__FUNCTION__);
	}
    }

    update->rl_node = rldbe.rl_node;

    /* send back the result */

    response = CLM_MALLOC( sizeof( *response ) );
    if ( response == NULL )
    {
	bsdlog(LOG_CRIT, "%s [%d]: malloc(%d) failed\n",
	       __FILE__, __LINE__, sizeof( response ));
	abort(__FUNCTION__);
    }

    (void) memcpy( response, update, sizeof( *response ) );

    reply = cccp_msg_create( PORT_RECOVERY, RESRPRI, dat->node,
			     RC_MSG_DIR_DATA, sizeof( *response ), NULL,
			     response, NULL );
    if ( reply == NULL) 
    {
	bsdlog(LOG_CRIT, "%s [%d]: malloc(%d) failed\n",
	       __FILE__, __LINE__, sizeof(clm_msg_t));
	abort(__FUNCTION__);
    }

    /* send it off */
    if ( cccp_send(reply) != CPERR_OK )
    {
	bsdlog(LOG_CRIT, "%s: cccp_send(%s)\n",
	       __FUNCTION__, printbin(rldbe.rl_name, rldbe.rl_namelen));
    }

    DBPRINTF((DBG_RECOV, "clmr_query: told %d that (%s) is mastered on %d",
	      dat->node, printbin(rldbe.rl_name, rldbe.rl_namelen),
	      rldbe.rl_node));
}

#endif /* ! CPROTO */




/* 	dir_update_complete
 * Description
 *	This function is called by the communications layer when it has
 *  	received an acknowledgement for the outbound directory update.
 *	We do this so we can wait until all of our updates can complete
 *	before we go on with the reconfiguration.
 *	This function is located in this module because it needs to update
 *	static variables here.  If the message priority was not DIRRPRI,
 *	then it wasn't a rebuild dir update so don't adjust counters.
 * Arguments
 *	msg - pointer to the message structure that was acknowledged
 *	stat - status returned
 * Returns
 *	Nothing
 * Global Data Affected
 *	NONE
 */
void
dir_update_complete(clm_msg_t *msg, int stat)
{
	rl_update_t	*req = (rl_update_t *)msg->cm_data;

	/*
	 * NOTE: zero rl_node means dir wipe from dealloc_res, so we
	 * don't adjust dir_outs in this case.
	 */

	if (CPERR_NODEDOWN == stat && msg->cm_prio == RESRPRI && req->rl_node) {
		/* node down - we'll be restarting shortly */
		DBPRINTF((DBG_RECOV, "%s: Restarting", __FUNCTION__));
	} else if (stat) {
		/* what happened? */
		bsdlog(LOG_ERR, "Error %d detected in dir_update_complete\n",
				stat);
		return;
	} else if (msg->cm_prio == MIGAPRI) {
		struct resource	*rp = req->rl_resptr;

		/*
		 * Here if we need to clear the "directory updating" bit in
		 * the resource flags that gets set to prevent a resource
		 * from migrating again before a directory update from a
		 * previous migration is complete.
		 */
		if (rp) {
			/* send ACK for migr after dir update completes */
			if (rp->migr_ack) {
			    int	ret;
			    if ((ret = cccp_send(rp->migr_ack)) != CPERR_OK) {
				DBPRINTF((DBG_MIGR,
				      "Send ack: cccp_send (%s) returned %d\n",
				      printbin(rp->name, rp->namelen), ret));
			    }
			    rp->migr_ack = NULL ;
			    rp->master = 0;
			}
			CLM_ASSERT(((rp->master == 0 ||
					~rp->rsrc_flags & RSRC_STUCK) &&
					rp->rsrc_flags & RSRC_ACTIVE),
					"Unstick non-master copy");
			rp->rsrc_flags &= ~RSRC_UPDATING;

			/*
			 * resource may have been purged during migration
			 * attempt but left around in order for migration to
			 * clean up.
			 */
			if (rp->refcount == 0) {
				deallocate_resource(rp);
			}
		}
	}

	if (msg->cm_data)
		CLM_FREE(msg->cm_data);

	cccp_msg_destroy( msg );
}




/****************************************************************************/

unsigned
dlm_recov_print_state( char * buf, unsigned limit )
{
    return( snprintf( buf, limit, "DLM recovery state: %s\n",
		      dlm_recov_state_machine->curr->name ) );
}
