/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_sched.c,v 1.7 2001/10/16 18:55:24 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: 
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include "dlm_kernel.h"
#include "clmstructs.h"
#include "clm_reconfig.h"

#include "dlm_recover.h"

#include "dlm_proto.h"  /* Should be LAST included file. */
#include <dlm_list.h>
#include <dlm_workqueue.h>

#ifdef DEBUG_ALL_PROTO /* hide from prototype file but not debugger */
#define STATIC
#else
#define STATIC static
#endif

#define STRING_BASE (CLM_MSG_NODE_UP)
static const char * const action_names[] =
{
    "node up", "node down", "add node", "delete node", "purge node"
};

/*
 * this group of functions queues events from outside and
 * retrieves them in a 2-level order:
 * down is high priority -and- erases any other requests for that node
 *     ahead of it in the low priority queue
 * if -any- down messages are in the queue, they have priority
 * all others are low priority and are processed in order
 */

struct dlm_transition
{
    dlm_listnode_t   link;
    unsigned         nodeid;
    unsigned         code;
};

typedef struct dlm_transition dlm_transition_t;

/* BUGBUG This queue needs to be protected by a spin-lock.
 */

static dlm_list_t dlm_transition_queue;
static kmem_cache_t * dlm_transition_cache = NULL;

/*
 * sched_queue_init
 *
 * set up scheduler queue as empty
 */

void
sched_queue_init( void )
{
    dlm_transition_cache = kmem_cache_create( "DLM transition",
					      sizeof( dlm_transition_t ),
					      0, 0, NULL, NULL );
    dlm_list_init( & dlm_transition_queue );
}


/*
 * sget
 *
 * return a free scheduler queue entry structure
 * abort() if fail
 */
static __inline__ dlm_transition_t *
sget(void)
{
    dlm_transition_t * ret;

    ret = kmem_cache_alloc( dlm_transition_cache, 0 );
    CLM_ASSERT( ret != NULL, "Unable to allocate dlm_transition_t" );
    return( ret );
}

/*
 * sfree
 *
 * free a scheduler queue structure
 */

static __inline__ void
sfree(dlm_transition_t *p)
{
    kmem_cache_free( dlm_transition_cache, p );
}




/****************************************************************************/
/* sched_queue_cleanup - Clean up all scheduler queue resources
 *
 * Release all resources held in the node transition scheduler queue.
 */

void
sched_queue_cleanup( void )
{
    dlm_transition_t * t;


    if ( dlm_transition_cache != NULL )
    {
	while ( (t = (dlm_transition_t *) 
		 dlm_list_rem_head( & dlm_transition_queue )) != NULL )
	{
	    sfree( t );
	}

	bsdlog( LOG_INFO, "[%s:%s] destroying %s cache\n",
	        current->comm, __FUNCTION__, "DLM transition" );

	kmem_cache_destroy( dlm_transition_cache );
	dlm_transition_cache = NULL;
    }
}




/****************************************************************************/
/* sched_queue
 *
 * Enter a request on a nodeid into the event queue side effect if a node
 * down, and purge any other request ahead
 */

static void
sched_queue(
	     /* Node ID of node changing state.
	      */

	     int nodeid,


	     /* State change function code.
	      */

             int code
	   )
{
    dlm_transition_t * node;
    dlm_transition_t * curr;
    dlm_listnode_t   * next;
    extern dlm_workqueue_t * dlm_master_work_queue;


    bsdlog(LOG_DEBUG, "[%s:%s] clms start %s (code=%u,node=%u)\n",
	   current->comm, __FUNCTION__,
	   action_names[ code - STRING_BASE ], code, nodeid );

    /* sget aborts if it fails - no need to check "node" */
    node = sget();
    node->nodeid = nodeid ;
    node->code = code ;

    if ( code == CLM_MSG_NODE_DOWN )
    {
	for ( curr = (dlm_transition_t *) dlm_list_get_head( & dlm_transition_queue )
	     ; curr != NULL
	     ; curr = (dlm_transition_t *) next )
	{
	    next = dlm_list_succ( (dlm_listnode_t *) curr );
	    if ( curr->nodeid == nodeid )
	    {
		dlm_list_remove( (dlm_listnode_t *) curr );
		sfree( curr );
	    }
	}
    }


    dlm_list_add_tail( & dlm_transition_queue,
		       (dlm_listnode_t *) node );
    up( & dlm_master_work_queue->sema );
}




/*
 * seen_chg
 *
 * match a nodeid & state change code to scheduler queue
 * delete matching queue item as handled
 * abort() if no match
 */
void
seen_chg(int nodeid, int code)
{
    dlm_transition_t * curr;
    dlm_listnode_t   * next;


    for ( curr = (dlm_transition_t *) dlm_list_get_head( & dlm_transition_queue )
	 ; curr != NULL
	 ; curr = (dlm_transition_t *) next )
    {
	next = dlm_list_succ( (dlm_listnode_t *) curr );

	if ( (code == curr->code) && (curr->nodeid == nodeid) )
        {
	    dlm_list_remove( (dlm_listnode_t *) curr );
	    sfree( curr );
	    return ;
        }
    }

    abort("cllockd: seen_chg can't delete item") ;
}

/*
 * new_change_request
 *
 * retrieves highest priority event from scheduler queue
 * if no high priority event, return the first one
 *
 * returns request code via codep, nodeid as value
 *	0 for no events pending
 */
int
new_change_request(int *codep)
{
    dlm_transition_t * curr;
    dlm_listnode_t   * next;
    int                node_id;


    node_id = 0;
    
    curr = (dlm_transition_t *) dlm_list_get_head( & dlm_transition_queue );
    if ( curr != NULL )
    {
	/* Default to taking the head of the list.  If there is a
	 * CLM_MSG_NODE_DOWN element in the list, it will replace these
	 * values.
	 */

	*codep = curr->code ;
	node_id = curr->nodeid ;

	for ( /* empty */ ; curr != NULL ; curr = (dlm_transition_t *) next )
	{
	    next = dlm_list_succ( (dlm_listnode_t *) curr );

	    if ( curr->code == CLM_MSG_NODE_DOWN )
	    {
		*codep = curr->code;
		node_id = curr->nodeid;
		break;
	    }
	}
    }


    return node_id;
}




/*
 * clms_node_delete
 *
 * queues a 'node delete' event for later processing
 */

int
clms_node_delete(int node)
{
  sched_queue(node, CLM_MSG_NODE_DELETE) ;
  return 0 ;
}

/*
 * clms_node_add
 *
 * queues a 'node delete' event for later processing
 */

int
clms_node_add(int node)
{
  sched_queue(node, CLM_MSG_NODE_ADD) ;
  return 0 ;
}

/*
 * clms_node_down
 *
 * queues a 'node down' for later (high priority) processing
 */

int
clms_node_down(int node)
{
  sched_queue(node, CLM_MSG_NODE_DOWN) ;
  return 0 ;
}

/*
 * clms_node_purge
 *
 * queues a 'purge all locks from this node' for later processing
 */
int
clms_node_purge(int node)
{
  sched_queue(node, CLM_MSG_NODE_PURGE) ;
  return 0 ;
}

/*
 * clms_start_send_new_node
 *
 * NO node up routine, because there are messages for that which work
 * at low protocol layers
 *
 * This is stage 2 of node up (start sending directory to newly up node)
 */
int
clms_start_send_new_node(int node)
{
  sched_queue(node, CLM_MSG_NODE_UP) ;
  return 0 ;
}
