/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_utilities.c,v 1.18 2001/11/07 23:56:54 idr Exp $"

/*
 * Utilities that existed on AIX, or buried in distaff parts of
 * HACMP are included here.  These were written locally to fill
 * in the various blanks.
 */
/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: bsdlog
 *      no_trace
 *      printbin
 *      timeval_sub
 */

/* ============================================================ */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__                  /* Only dlm_kerndd.c sets version. */
#include <linux/module.h>
#include <linux/version.h>              /* Due to __NO_VERSION__ */
#ifdef USING_FULL_VERSIONING
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */
/* ============================================================ */

#include <linux/types.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/time.h>
#include <linux/ctype.h>
#include <linux/byteorder/generic.h>
#include <stdarg.h>

#include "dlm_kernel.h"
#include "dlm_cluster.h"
#include "clmstructs.h"

#define BUFFER_LEN    1024
static char fmt_buffer[ BUFFER_LEN ];
static char final_buffer[ BUFFER_LEN ];
static spinlock_t bsdlog_lock = SPIN_LOCK_UNLOCKED;

/* ------------------------------------------------------------ */
/*
 * Provide a bsdlog since there doesn't seem to be one.  Make
 * it a call through to printk, which is essentially what we
 * want.
 */

int
vbsdlog( int n, const char * fmt, va_list arglist )
{
    static const char * const messages[ LOG_DEBUG + 1 ] =
    {
	KERN_EMERG, KERN_ALERT, KERN_CRIT, KERN_ERR, KERN_WARNING,
	KERN_NOTICE, KERN_INFO, KERN_DEBUG
    };

    const char * pre;
    char *buffer;
    int len, ret;
    int fmt_len;


    spin_lock( & bsdlog_lock );

    /* Length of the format string, plus the length of a KERN_ message, plus
     * one for the NUL terminator, plus one for the '\n' that we might have
     * to add.
     */

    fmt_len = strlen( fmt );
    len = 3 + fmt_len + 1 + 1;
    buffer = fmt_buffer;

    pre = ( (n < LOG_EMERG) || (n > LOG_DEBUG) )
	? messages[ 0 ]
	: messages[ n ];

    (void) memcpy( buffer, pre, 3 );
    (void) memcpy( buffer + 3, fmt, fmt_len );
    
    buffer[ len - 2 ] = '\0';
    buffer[ len - 1 ] = '\0';

    if ( buffer[ len - 3 ] != '\n' )
    {
	buffer[ len - 2 ] = '\n';
    }

    vsnprintf( final_buffer, BUFFER_LEN, buffer, arglist );
    final_buffer[ BUFFER_LEN - 2 ] = '\n';
    final_buffer[ BUFFER_LEN - 1 ] = '\0';

    ret = printk( final_buffer );
    spin_unlock( & bsdlog_lock );
  
    return(ret);
}

int
bsdlog(int n, const char * fmt, ...)
{
    va_list   arglist;


    va_start( arglist, fmt );
    return( vbsdlog( n, fmt, arglist ) );
}


void no_trace(int hook)
{
#if 0 //def DEBUG
  bsdlog( LOG_INFO, "Trace call, hook id %d\n", hook );
#endif

  return;
}

/* ------------------------------------------------------------ */
/*
 * Return a pointer to the beginning of a printable version of
 * the given data pointer.
 */
char *
printbin( const unsigned char * data, unsigned len)
{
    /* The extra "slop" is because a non-printable character will use more
     * than a single byte in the buffer.
     */

    static char buf[MAXLINELEN + 8];
    unsigned   index;


    if ( len > (MAXLINELEN - 1) )
    {
	len = (MAXLINELEN - 1);
    }

    for ( index = 0 ; (len != 0) && (index < MAXLINELEN) ; len-- )
    {
        if (*data < ' ' || *data > '~') 
	{
            index += sprintf( & buf[ index ], "<%X>", *data);
        }
	else 
	{
	    buf[ index ] = *data;
            index++;
        }

        data++;
    }

    buf[ index ] = '\0';
    return buf;
}

/* ------------------------------------------------------------ */
/*
 * Return the current pid.  Do here to preserve some amount
 * of data hiding.
 */
int getpid(void)
{
  return(current->pid);
}

/* ------------------------------------------------------------ */
/*
 * Subtract the second time value from the first, and store
 * the result in the third.  If successful, return 0, if not,
 * return -1.
 */
int timeval_sub(struct timeval *t0,
                struct timeval *t1,
                struct timeval *t2)
{
    if (t0 == NULL ||
        t1 == NULL ||
        t2 == NULL)
        return(-1);
    t2->tv_usec = t0->tv_usec - t1->tv_usec;
    t2->tv_sec = t0->tv_sec - t1->tv_sec;
    if (t2->tv_usec < 0) {
      t2->tv_usec += 1000000L;
      t2->tv_sec--;
    }
    return(0);
}

/* ============================================================ */
/*
 * Although uint64_t happily exists in linux/types.h,
 * cproto barfs on it!  Since I want to use cproto to
 * generate prototypes, it skips these functions, if I
 * don't use unsigned long long int instead.  Stupid tool.
 * uint64_t only bothers cproto if it's in prototypes.
 */

/*
 * kfl_ftos
 *
 * convert a 'kfloat' to a string in a static buffer
 * return the buffer
 */

char *
kfl_ftos(/*uint64_t*/ unsigned long long int f)
{
#ifndef __linux__
    int tmp ;
    int i = 0, n ;
#endif /* !__linux__ */
    static char kfl_buf[100] ;

    (void) memset( kfl_buf, 0, 100 );
    (void) snprintf( kfl_buf, sizeof( kfl_buf ), "%llu", f );

#ifndef __linux__
    tmp = kfl_ftofi(f) ;
    if (tmp < 0)
    {
        kfl_buf[0] = '-' ;
        i++ ;
        tmp = -tmp ;
    }
    tmp += 1 ;          /* magic rounding factor */
    sprintf(kfl_buf + i, "%d.", tmp >> 16) ;
    i = strlen(kfl_buf) ;

    for (n = 0 ; n < 4 ; n++)
    {
        tmp &= 0xFFFF ;
        tmp *= 10 ;
        kfl_buf[i++] = '0' + (tmp >> 16) ;
    }
#endif /* !__linux__ */

    return kfl_buf ;
}
int kfl_ftofi(/*uint64_t*/ unsigned long long int f)
{
  return(f);
}
/*uint64_t*/
unsigned long long int kfl_fitof(int i)
{
  return(i);
}
/*uint64_t*/
unsigned long long int kfl_mul(/*uint64_t*/ unsigned long long int a,
                               /*uint64_t*/ unsigned long long int b)
{
  return(a*b);
}
/*uint64_t*/
unsigned long long int kfl_add(/*uint64_t*/ unsigned long long int a,
                               /*uint64_t*/ unsigned long long int b)
{
  return(a+b);
}
/*uint64_t*/
unsigned long long int kfl_sub(/*uint64_t*/ unsigned long long int a,
                               /*uint64_t*/ unsigned long long int b)
{
  return(a-b);
}
/*uint64_t*/
unsigned long long int kfl_idivi(int i, int j)
{
  return(i/j);
}




/****************************************************************************/
/* dlm_kill_user_proc - Signal a user-mode process
 *
 * Sends a signal to a user-mode process.  This is usually done for one of
 * two reasons.  The first reason is to kill a client process.  The other
 * reason is to notify a process of an AST.
 */

int
dlm_kill_user_proc( pid_t user_pid, int sig )
{
    if ( sig != 0 )
    {
#ifdef __linux__
	bsdlog( LOG_INFO, "[%s:%s] send signal [%d] to pid [%d] \n",
	        current->comm, __FUNCTION__, sig, user_pid );

	return( kill_proc( user_pid, sig, 1 ) );
#else
	pidsig( user_pid, sig );
#endif
    }

    return( 0 );
}




/****************************************************************************/

void
dlm_set_thread_name( struct task_struct * task, const char * name )
{
    (void) strncpy( task->comm, name, sizeof( task->comm ) );
    task->comm[ sizeof( task->comm ) - 1 ] = '\0';
}




/****************************************************************************/
/* dlm_daemonize - Disconnect from user-mode resources
 * 
 * This code is essentially copied from kernel/sched.c of the 2.4.x kernel.
 * The 'daemonize' function does exist in later 2.2.x kernels, but it does
 * not call exit_files, which is the most important part in the DLM's case.
 */

void
dlm_daemonize( void )
{
    struct fs_struct *fs;

    /*
     * If we were started as result of loading a module, close all of the
     * user space pages.  We don't need them, and if we didn't close them
     * they would be locked into memory.
     */
    exit_mm(current);

    current->session = 1;
    current->pgrp = 1;

    /* Become as one with the init task */

    exit_fs(current);       /* current->fs->count--; */
    fs = init_task.fs;
    current->fs = fs;
    atomic_inc(&fs->count);

    exit_files(current);
    current->files = init_task.files;
    atomic_inc(&current->files->count);
}




/****************************************************************************/

bool_t
dlm_stuff_rl_info( clm_trans_t * req, rl_info_t * rl_entry )
{
    bool_t  valid_type;


    valid_type = TRUE;
    switch ( req->clm_type )
    {
	case DLM_LOCK:
	case DLM_UNLOCK:
	{
	    rl_entry->rl_name = &req->clm_lockreq.nameinfo.name[0];
	    rl_entry->rl_namelen = req->clm_lockreq.nameinfo.namelen;
	    rl_entry->rl_type = req->clm_locktype;
	    break;
	}

	case DLM_REGISTER:
	{
	    rl_entry->rl_name = &req->clm_register.unix_name[0];
	    rl_entry->rl_namelen = req->clm_register.namelen;
	    rl_entry->rl_type = DLM_RES_UNIX;
	    break;
	}
	
	case DLM_SCN_OP:
	{
	    rl_entry->rl_name = &req->clm_scninfo.nameinfo.name[0];
	    rl_entry->rl_namelen = req->clm_scninfo.nameinfo.namelen;
	    rl_entry->rl_type = DLM_RES_VMS;
	    break;
	}

	case DLM_RMIGR_OP:
	{
	    rl_entry->rl_name = &req->clm_rmpinfo.rmp_resname[0];
	    rl_entry->rl_namelen = req->clm_rmpinfo.rmp_namelen;
	    rl_entry->rl_type = req->clm_rmpinfo.rmp_type;
	    break;
	}

	case DLM_STATS:
	{
	    rl_entry->rl_name = &req->clm_statsinfo.resname[0];
	    rl_entry->rl_namelen = req->clm_statsinfo.namelen;
	    rl_entry->rl_type = req->clm_statsinfo.type;
	    break;
	}

	case DLM_PURGELOCKS:
	default:
	{
	    valid_type = FALSE;
	    break;
	}
    }


    return( valid_type );
}
