/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */

#ifndef DLM_WORKQUEUE_H
#define DLM_WORKQUEUE_H

#ident "$Id: dlm_workqueue.h,v 1.3 2001/03/02 00:00:22 idr Exp $"

#include "dlm_list.h"

struct dlm_workqueue
{
    /* This is the queue that contains the work-units that need to
     * be processed.
     */

    dlm_list_t   work_queue;


    /* Mutex used to arbitrate access to work_queue.  Obviously, it is used
     * by both producers and consumers.
     */

    spinlock_t         mutex;


    /* Semaphore used by producers to signal consumers that a new element has
     * been added to the work_queue.
     */

    struct semaphore   sema;
};




/* Type of function used to free the data associated with a work-unit.  These
 * functions are used only by dlm_workunit_dispose (see below).  If 0 is
 * returned, dlm_workunit_dispose will free (using CLM_FREE) the
 * dlm_workunit_t that was associated with the user data.  
 *
 * If a positive number is returned, dlm_workunit_dispose will assume that
 * this function released the dlm_workunit_t.  The implementation works
 * this way to support user defined types such as:
 * 
 * struct my_workunit 
 * {
 *     dlm_workunit_t   unit;
 *     struct my_data   my_data;
 * };
 *
 * In this case, foo->unit->data would point to foo, and the free function
 * would return a number greater than zero.
 */

typedef int (dlm_workunit_free_t)( void * );




struct dlm_workunit
{
    /* Information to link the work-unit in the work-queue.
     */

    dlm_listnode_t   link;


    /* User-defined type for the work unit.  This is usually used so that
     * the workers will know what type of data '*data' references.
     */

    unsigned    type;


    /* User-defined data associated with the work unit.
     */

    void * data;
    
    
    /* Function used by dlm_workunit_dispose to release '*data'.
     */

    dlm_workunit_free_t * free_data;
};

typedef struct dlm_workqueue dlm_workqueue_t;
typedef struct dlm_workunit dlm_workunit_t;

static __inline__ void dlm_workqueue_init( dlm_workqueue_t * work )
{
    spin_lock_init( & work->mutex );
    sema_init( & work->sema, 0 );
    dlm_list_init( & work->work_queue );
}

extern int dlm_workqueue_put_work( dlm_workqueue_t * work,
    dlm_workunit_t * data );

extern dlm_workunit_t * dlm_workqueue_get_work( dlm_workqueue_t * work,
    bool_t  non_blocking );

extern int dlm_workunit_dispose( dlm_workunit_t * unit );

#endif /* DLM_WORKQUEUE_H */
