#include <linux/version.h>
#include <linux/kernel.h>
#include "hsm.h"

#ifndef NDEBUG
# define STATIC
#else
# define STATIC static
#endif

static hsm_msg_t const startMsg = { {NULL, NULL}, START_EVT};
static hsm_msg_t const entryMsg = { {NULL, NULL}, ENTRY_EVT};
static hsm_msg_t const exitMsg =  { {NULL, NULL}, EXIT_EVT};

STATIC unsigned hsm_find_lca( const hsm_state_t * curr, 
			      const hsm_state_t * next );
static void hsm_execute_transition( hsm_t * me );
static const hsm_state_t ** hsm_build_trace( const hsm_state_t * state,
    const hsm_state_t * top, const hsm_state_t ** trace );
static void hsm_process_event_private( hsm_t * me, hsm_msg_t const *msg );

#define send_event(me_, ctx_, msg_)\
    (*(me_)->hndlr)((ctx_), (msg_))




/****************************************************************************/

const hsm_state_t **
hsm_build_trace( const hsm_state_t * state, const hsm_state_t * top,
		 const hsm_state_t ** trace )
{
    while ( state != top )
    {
	HSM_ASSERT( state != NULL, "Bad super-state pointer" );

	*(trace++) = state;
	state = state->super;
    }

    return( trace );
}




/****************************************************************************/

void
hsm_execute_transition( hsm_t * me )
{
    const hsm_state_t  * entry_path[MAX_STATE_NESTING + 1];
    const hsm_state_t ** trace;


    HSM_ASSERT( me->next != NULL, "Next-state not set" );

    /* The new state might be a composite state.  That is, it might have some
     * sub states (i.e., there exists some state such that 
     * state->super == me->curr).  If this is the case, then a "default"
     * transition should happen when a transition happens to the (composite)
     * state.  We know that this has happened because the event handler will
     * have called STATE_START (which sets me->next) to invoke the default
     * transition.
     */

    while ( me->next != NULL )
    {
	/* me-curr can only be NULL here if this is the very first transition.
	 * That is, if we are here via hsm_start.
	 */

	printk( KERN_INFO "[HSM:%s] %s (%s -> %s)\n", __FUNCTION__,
	        me->name, 
	        (me->curr == NULL) ? "(empty)" : me->curr->name, 
	        me->next->name );

	/* Send the entry message to all of the states between the new state
	 * and the least-common-ancestor.
	 */

	entry_path[0] = NULL;
	trace = hsm_build_trace( me->next, me->curr, & entry_path[1] );

	while ( *(--trace) != NULL )
	{
	    (void) send_event( *trace, me, &entryMsg );
	}

	me->curr = me->next;
	me->next = NULL;


	/* Send the start message to the new state, and see if it sets a new
	 * next state.
	 */

	(void) send_event( me->curr, me, &startMsg );
    }
}




/****************************************************************************/
/* enter and start the top state */
void 
hsm_start(hsm_t * me)
{
    me->curr = NULL;
    me->next = NULL;

    STATE_START( me, me->top );
    hsm_execute_transition( me );
}




/****************************************************************************/
/* hsm_process_event_private - Internal routine used to process a message
 * 
 */

void 
hsm_process_event_private(hsm_t * me, hsm_msg_t const *msg)
{
    const hsm_state_t  * s;


    HSM_ASSERT( me != NULL, "NULL state-machine pointer" );
    HSM_ASSERT( me->curr != NULL, "NULL current state" );
    HSM_ASSERT( msg != NULL, "NULL event pointer" );

    printk( KERN_INFO "[HSM:%s] %s (%d -> %s)\n",
	    __FUNCTION__,
	    (char *) me->name, (int) msg->evt, (char *) me->curr->name );

    /* Work up the state heirarchy until either we run out of super states
     * or the message gets processed.
     */

    for ( s = me->curr ; (s != NULL) && (msg != NULL) ; s =  s->super ) 
    {
	msg = send_event( (hsm_state_t *) s, me, msg );
    }


    /* state transition taken? */
    if ( me->next != NULL )
    {
	HSM_ASSERT( msg == NULL, "Next-state set, but message not consumed" );
	hsm_execute_transition( me );
    }
}




/****************************************************************************/
/* hsm_process_event - Send an event into a state machine for processing
 * 
 */

void 
hsm_process_event(hsm_t * me, hsm_msg_t const *msg)
{
    bool_t      can_deliver;
    hsm_msg_t * event;


    printk( KERN_DEBUG "[HSM:%s] enter( me = 0x%08lx, msg = 0x%08lx )\n",
	    __FUNCTION__, (long) me, (long) msg );

    HSM_ASSERT( me->event_queue.tail == NULL, "Corrupted list header" );

    spin_lock( & me->event_mutex );
    dlm_list_add_tail( & me->event_queue, (dlm_listnode_t *) msg );
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    can_deliver = atomic_inc_and_test( & me->event_delivery_flag );
#else
    atomic_inc( & me->event_delivery_flag );
    can_deliver = (atomic_read( & me->event_delivery_flag ) == 0);
#endif
    spin_unlock( & me->event_mutex );


    if ( can_deliver )
    {
	while( TRUE )
	{
	    spin_lock( & me->event_mutex );

	    /* Get the next event from the event queue.  If the queue is
	     * empty, clear the in-delivery flag and return.
	     */

	    event = (hsm_msg_t *) dlm_list_rem_head( & me->event_queue );
	    if ( event == NULL )
	    {
		atomic_set( & me->event_delivery_flag, -1 );
		spin_unlock( & me->event_mutex );
		break;
	    }

	    spin_unlock( & me->event_mutex );


	    /* Send the event into the state machine.
	     */

	    hsm_process_event_private( me, event );
	}
    }
    
    printk( KERN_DEBUG "[HSM:%s] exit\n", __FUNCTION__ );
}




/****************************************************************************/
/* hsm_find_lca - Find number of levels to least-common-ancestor
 *
 * Finds the number of levels to the least-common-ancestor of curr and next.
 * 
 * RETURNS
 * 
 * The number of levels from curr to its least-common-ancestor with next.  If
 * the two states do not have a common ancestor, zero is returned.
 */

unsigned
hsm_find_lca(const hsm_state_t * curr, const hsm_state_t * next)
{
    const hsm_state_t  * src_path[ MAX_STATE_NESTING ];
    const hsm_state_t  * dst_path[ MAX_STATE_NESTING ];
    const hsm_state_t ** src_trace;
    const hsm_state_t ** dst_trace;
    unsigned       to_lca;


    /* By far the most common case is where both states have the same super-
     * state.  Let's optimize that case.
     */

    if ( curr->super == next->super )
    {
	to_lca = 1;
    }
    else
    {
	to_lca = 0;

	src_trace = hsm_build_trace( curr, NULL, src_path );
	dst_trace = hsm_build_trace( next, NULL, dst_path );
	

	/* The top level MUST match!  If it doesn't match, then the two
	 * state heirarchies are disjoint, which means that they are from
	 * two different state machines!  Bad news!
	 */

	HSM_ASSERT( *(src_trace - 1) == *(dst_trace - 1),
		    "Least-common-parent states do not match" );


	/* Iterate starting at the top of the heirarchy (which is the last
	 * entry in the 'trace' arrays).  At the start the super-state
	 * pointers should match.  When a non-match is found, it must be the
	 * least-common-ancestor.
	 */

	while( (src_trace != src_path) && (dst_trace != dst_path) )
	{
	    src_trace--;
	    dst_trace--;

	    if ( *src_trace != *dst_trace )
	    {
		to_lca = (src_trace - src_path) + 1;
		break;
	    }
	}
    }


    return( to_lca );
}




/****************************************************************************/
/* hsm_exit_private - Exit current (and all super) states
 *
 * Sends an exit message to the current state and all super states upto
 * (but not including) the least-common-ancestor of the current state and the
 * new state.
 * 
 * The 'curr' pointer of the HSM is set to point to least-common-ancestor
 * state.  This is done so that entry messages can be sent to the proper
 * super states.
 */

void 
hsm_exit_private( hsm_t * me, unsigned * toLca, const hsm_state_t * target )
{
    hsm_state_t * s;
    unsigned      count;
    

    if ( *toLca == 0 )
    {
	*toLca = hsm_find_lca( me->curr, target );
    }

    s = (hsm_state_t *) me->curr;
    for ( count = *toLca ; count > 0; --count )
    {
	(void) send_event(s, me, &exitMsg);
	s = (hsm_state_t *) s->super;
    }

    me->curr = s;
}
