/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/lesser.html
 */

#ifndef HSM_H
#define HSM_H

#include <linux/spinlock.h>
#include <asm/atomic.h>
#include <asm/page.h>

#include <dlm_list.h>

#ifndef FALSE
#define FALSE 0
#define TRUE (! FALSE)
#endif

#ifdef DEBUG

extern void abort( const char * );
#define HSM_ASSERT(cond,text) \
    do \
    { \
	if ( ! (cond) ) \
	{ \
	    printk( KERN_CRIT "ASSERTION FAILED: %s\n", (text) ); \
	    BUG(); \
	} \
    } while( 0 )
    
#else
#define HSM_ASSERT(cond,text)
#endif

typedef int hsm_event_t;

struct hsm_msg {
    dlm_listnode_t   link;
    hsm_event_t      evt;
};

typedef struct hsm_msg hsm_msg_t;

typedef struct hsm hsm_t;
typedef hsm_msg_t const *(hsm_handler_t)(hsm_t*, hsm_msg_t const*);

struct hsm_state;
typedef struct hsm_state hsm_state_t;

struct hsm_state 
{
    /* pointer to superstate */
    const hsm_state_t * super;

    /* state's handler function */
    hsm_handler_t * hndlr;

    char const *name;
};


/* Hierarchical State Machine base class */
struct hsm 
{
    /* Pointer to the static name for the state machine.
     */

    char const  * name;


    /* Pointer to the currently active state.
     */

    const hsm_state_t * curr;


    /* Pointer to the next state.  If this is non-NULL, then a transition
     * is to be taken.
     */

    const hsm_state_t * next;


    /* Top-most state object. 
     */

    const hsm_state_t * top;


    /* The mutex protects the queue and the flag.  Events are added to the
     * tail of the queue, and removed from the head.  If the flag is set,
     * then another process is already processing an event.
     */

    spinlock_t    event_mutex;
    dlm_list_t    event_queue;
    atomic_t      event_delivery_flag;
};

#define MAX_STATE_NESTING 8

/****************************************************************************/

#define STATE_INIT(s,n,sup,han) \
    do { \
	(s)->super = (sup); \
	(s)->hndlr = (han); \
	(s)->name = n; \
    } while( 0 )

#define STATE_ASSIGN(s,h,n) { (s), (h), (n) }

#define HSM_INIT(h,n,t) \
    do { \
	(h)->name = n; \
	spin_lock_init( & (h)->event_mutex ); \
	atomic_set( & (h)->event_delivery_flag, -1 ); \
	dlm_list_init( & (h)->event_queue ); \
	(h)->top = (t); \
	HSM_ASSERT( (t)->super == NULL, "top state has super-state!" ); \
    } while( 0 )




/****************************************************************************/

void hsm_start(hsm_t * me);  /* enter and start the top state */
void hsm_process_event(hsm_t * me, hsm_msg_t const * msg);       /* "HSM engine" */

/* protected: */
void hsm_exit_private( hsm_t * me, unsigned * toLca, const hsm_state_t * target );

#define STATE_CURR(me_) (((hsm_t *)me_)->curr)

#define STATE_START(me_, target_) \
    do { \
	HSM_ASSERT( ((hsm_t *)me_)->next == NULL, "Non-NULL next pointer!" ); \
	((hsm_t *)me_)->next = (target_); \
    } while( 0 )


#define STATE_TRAN(me_, target_) \
    do { \
	static unsigned toLca_ = 0; \
	HSM_ASSERT( ((hsm_t *)me_)->next == NULL, "Non-NULL next pointer!" ); \
	hsm_exit_private((hsm_t *)(me_), & toLca_, (target_) );\
	((hsm_t *)(me_))->next = (target_);\
    } while ( 0 )

	 
#define START_EVT ((hsm_event_t)(-1))
#define ENTRY_EVT ((hsm_event_t)(-2))
#define EXIT_EVT  ((hsm_event_t)(-3))

#endif /* HSM_H */

