/*
 * User-space monitor modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_cmdline.c,v 1.4 2001/07/31 23:49:06 idr Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: MSGSTR
 *		load_func
 *		parse_debug
 *		print_debug_options
 *		probe
 */

/* ============================================================ */
/*
 * Parse the command line and save options into appropriate
 * variables for later handling.  See the usage msg function
 * below for options and descriptive info.
 */
/* ============================================================ */

#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/socketvar.h>

#include "dlm.h"
#include "ast.h"
#include "dlm_cluster.h"
#include "dlm_kernel.h"
#include "dlm_devmsg.h"
#include "dlm_version.h"
#include "clmstructs.h"
#include "clm_migrate.h"
#include "pti.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
/* end of NLS */
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */

#undef LOCK_FREE

/* ============================================================ */

int parse_debug(char *flagstring);
void print_debug_options( void );
#if 0
static int probe (void);
#endif

extern char *main_compiled;

/* -------------------------------------------------- */

char *usgtab[] = {
  " -F[dumpfilename] -R[resource_seg_size] -P[resource_seg_preallocate]\n",
  " -L[lock_seg_size] -D[decay-rate] -r[recalculation-rate]\n",
  " -C[configuration file name]\n",
  " -C : absolute path name of configuration file to use, default /etc/haDLM\n",
  NULL
};

arg_struct_t args = {
  0,                                  /* debug		*/
  0,                                  /* local_debug		*/
  250000,                             /* clm_batchsleep	*/
  DEFLOCKSEGSIZE,                     /* def lock seg size	*/
  DEFRESSEGSIZE,                      /* def res seg size	*/
  0,                                  /* def res seg prealloc	*/
  DEFHASHSIZE,                        /* def hash table size	*/
  0,                                  /* message queue key	*/
  50,                                 /* kmem cutoff %	*/
  MCB_EVNT,                           /* migration recalc rate*/
  MCB_APSD,                           /* migration decay rate */
  DEFDLMMAJORNUMBER,                  /* DLM major number */
  DEFDLMADMINMINOR,                   /* DLM admin minor number */
  DEFDLMLOCKSMINOR,                   /* DLM locks minor number */
  "",                                 /* DLM config file name */
  "",                                 /* dump file name	*/
  "",                                 /* DLM device base name */
  "",                                 /* DLM device absolute name */
  "",                                 /* DLM base dirname in /proc */
  "",                                 /* DLM Cluster Manager load module */
  "",                                 /* DLM "admin" base name */
  "",                                 /* DLM "admin" absolute name */
  "",                                 /* DLM "locks" base name */
  ""                                  /* DLM "locks" absolute name */
};


/* ============================================================ */
/*
 * Print usage information.
 */
void usage(char *prog,
           char **instructions)
{
  int i;

  fprintf(stderr,
          "Usage: %s <optional parameters>",
          prog);
  for(i = 0; NULL != instructions[i]; i++) {
    fprintf(stderr,
            "%s",
            instructions[i]);
  }
  fprintf(stderr, "\n");
}

/*
 * Name
 * 	parse_options
 * Description
 * 	Process command line arguments.
 * Arguments
 *	argc - count of command line arguments in arg vector
 *	argv - the arg vector
 * Returns
 *	a non-zero status for error
 *	zero if all went well
 * Global Data Affected
 *	debug
 *	local_debug
 *	clm_batchsleep
 *	dbfp
 *	maxresources
 */

int parse_options(int argc, char *argv[])
{
	int c;
	extern int optind;
	extern char *optarg;
	int retval;

    retval = 0;

    main_compiled = malloc(MAXLINELEN);
    sprintf(main_compiled,
            "DLMK Major/Minor(%d/%d/%p) Time(%s) Date(%s)\n",
            DLM_MAJOR_VERSION,
            DLM_MINOR_VERSION,
            (void *)DLM_VERSION,
            __TIME__,
            __DATE__);
    printf(main_compiled);


    /* defaults */
    strcpy(args.dumpfilename, DEFDLMDUMPFILE);
    strcpy(args.dlmconfigfile, DEFDLMCONFIGFILE); /* DLM default config file */

	while ((c = getopt(argc,
                       argv,
                       "F:d:L:h:b:s:f:R:P:H:r:D:M:C:")) != EOF) {
		switch(c) {
		case 'r':
			sscanf(optarg, "%u", &args.recalc_time);
			if (args.recalc_time <= 0) {
				fprintf(stderr, 
					"-r must be a positive integer\n");
				args.recalc_time = MCB_EVNT;
				fprintf(stderr,
					"Using default recalc rate of %d\n",
					MCB_EVNT);
			}
			break;
		case 'D':
			sscanf(optarg, "%llu", &args.decay_rate);
			if (args.decay_rate > DECAY_MAX ||
					args.decay_rate < DECAY_MIN) {
				fprintf(stderr, MSGSTR(D_RANGE_ERROR,
				"-D can't exceed %5.3f or be less than %5.3f\n"),
					DECAY_MAX, DECAY_MIN);
				args.decay_rate = MCB_APSD;
				fprintf(stderr, MSGSTR(D_RANGE_DEF,
					"Using default rate of %5.3f\n"),
					MCB_APSD);
			}
			break;
		case 'd':
			args.debug = parse_debug(optarg);
			if (args.debug) {
				args.debug |= DBG_FORCE;
			} else {
				exit(2);
			}
			break;
		case 'l':
			sscanf(optarg, "%d", &args.local_debug);
			break;
		case 's':
			sscanf(optarg, "%d", &args.batch_sleep);
			break;
		case 'F':
			strcpy(args.dumpfilename, optarg);
			break;
#if 0
		case 'L':
			sscanf(optarg, "%d", &args.locksegsize);
			if (args.locksegsize > MAXLOCKSEGSIZE ||
					args.locksegsize <= 0) {
				fprintf(stderr, MSGSTR(L_CANT_EXCEED, 
				"-L can't exceed %d or be less than 1\n"), 
					MAXLOCKSEGSIZE);
				args.locksegsize = DEFLOCKSEGSIZE;
				fprintf(stderr,"Using default size of %d\n",
					DEFLOCKSEGSIZE);
			}
			break;
		case 'R':
			sscanf(optarg, "%d", &args.ressegsize);
			if (args.ressegsize > MAXRESSEGSIZE ||
					args.ressegsize <= 0) {
				fprintf(stderr, MSGSTR(R_CANT_EXCEED, 
				"-R can't exceed %d or be less than 1\n"), 
					MAXRESSEGSIZE);
				args.ressegsize = DEFRESSEGSIZE;
				fprintf(stderr,"Using default size of %d\n",
					DEFRESSEGSIZE);
			}
			break;
		case 'P':
			sscanf(optarg, "%d", &args.resprealloc);
			if (args.resprealloc > MAX_RES_SEGS ||
					args.resprealloc < 0) {
				fprintf(stderr, MSGSTR(P_CANT_EXCEED, 
				"-P can't exceed %d or be less than 0\n"), 
					MAX_RES_SEGS);
				args.resprealloc = 0;
				fprintf(stderr,
					"Not preallocating any segments\n");
			}
			break;
#endif
		case 'H':
			sscanf(optarg, "%d", &args.hashsize);
			if (args.hashsize > MAXHASHSIZE ||
					args.hashsize < 0) {
				fprintf(stderr, MSGSTR(H_CANT_EXCEED, 
				"-H can't exceed %d or be less than 1\n"), 
					MAXHASHSIZE);
				args.hashsize = DEFHASHSIZE;
				fprintf(stderr,"Using default size of %d\n",
					DEFHASHSIZE);
			}
			break;
		case 'M':
			sscanf(optarg, "%d", &args.kmem_cutoff);
			if (args.kmem_cutoff > 90 ||
					args.hashsize < 10) {
				fprintf(stderr, 
				    "-M can't exceed 90 or be less than 10\n");
				args.kmem_cutoff = 50;
				fprintf(stderr,"Using default size of 50\n");
			}
			break;
        case 'C':
          strcpy(args.dlmconfigfile, optarg);
          break;
		default:
			usage(argv[0], &usgtab[0]);
            retval = -1;
			break;
		}
	}

	/* Kernel processes are inherently detached.  -jjd- */
	if (!args.debug) {
#if 0   /* Nuke for Linux */
      detach();                       /* Something in Linux match this? */
#endif  /* Nuke for Linux */
		/* set nice value to minimum */
		if(setpriority(PRIO_PROCESS, 0, PRIO_MIN) == -1) {
			perror("setpriority");
		}
	}

    return(retval);
}

/*
 * Name
 *	parse_debug
 * Description
 * 	This function takes a string obtained from the command
 *	line, parses each debug option found and 
 *	returns a bit mask of all debugging options requested.
 *	Erroneous strings get reported on stderr.
 *	The option string is expected to be comma separated
 *	
 * Arguments
 *	flagstring - string from command line
 * Returns
 *	Mask of all selected debugging options
 * Global Data Affected
 *	NONE
 */
static struct options {
	char *opt_name;	
	int opt_mask;
} options[] = {
{ "ALL", 	~ (DBG_KDBX) },	/* these cause bpt */
{ "BASIC", 	DBG_BASIC },
{ "FUNC", 	DBG_FUNC },
{ "STATS", 	DBG_STATS },
{ "COMM", 	DBG_COMM },
{ "DEADLOCK", 	DBG_DEADLOCK },
{ "CTI",	DBG_CTI},
{ "CLMLOCK",	DBG_CLMLOCK},
{ "UNIXLOCK",	DBG_UNIXLOCK},
{ "LOCK",	DBG_LOCK},
{ "RESOURCE",	DBG_RESOURCE},
{ "DYNRES",	DBG_DYNRES},
{ "BATCH",	DBG_BATCH},
{ "TIMEOUT",	DBG_TIMEOUT},
{ "NETFAIL",	DBG_NETFAIL},
{ "CONFIG",	DBG_CONFIG},
{ "SCN",	DBG_SCN},
{ "DIRECT",	DBG_DIRECT},
{ "MIGR",	DBG_MIGR},
{ "MIGR2",	DBG_MIGR2},
{ "RECOV",	DBG_RECOV},
{ "TUNE",	DBG_TUNE},
{ "GROUP",	DBG_GROUP},
{ "AST",	DBG_AST},
{ "SYNC",       DBG_SYNC},
{ "LOCAL",      DBG_LOCAL},
{ "PURGE",	DBG_PURGE},
{ "KDBX",	DBG_KDBX},
{ "FAILSYNC",	DBG_FAILSYNC},
{ "CLIENT",	DBG_CLIENT}
};

int
parse_debug(char *flagstring)
{
    int i;
    int mask = 0;
    int notflag ;
    char *p = flagstring;
    char *comma;

    while (p && *p)
    {
	notflag = 0 ;
	comma = (char*)index(p, ',');
	if (comma != NULL)
	{
	    *comma = '\0';
	}
	if (*p == '-')
	{
	    notflag = 1 ;
	    p++ ;
	}
	for (i = 0; i < sizeof(options)/sizeof(options[0]); i++)
	{
	    if (strcmp(p, options[i].opt_name) == 0)
	    {
		if ( ! notflag)
		{
		    mask |= options[i].opt_mask;
		    fprintf(stderr, MSGSTR(DEBUG_ENABLE,
					    "%s debugging enabled\n"), 
					    options[i].opt_name);
		}
		else
		{
		    mask &= ~ options[i].opt_mask ;
		    fprintf(stderr, MSGSTR(DEBUG_ENABLE,
					    "%s debugging disabled\n"), 
					    options[i].opt_name);
		}
		break;
	    }
	}
	if (i >= sizeof(options)/sizeof(options[0]))
	{
	    fprintf(stderr, MSGSTR(BAD_OPTION, "bad option flag '%s'\n"), p);
	    print_debug_options();
	}
	if (comma)
	{
	    *comma = ',';
	    p = comma;
	    p++;	
	}
	else
	{
	    break;
	}
    }
    return mask;
}

/* 
 * Name
 *	print_debug_options
 * Description
 *	Print list of supported debug options
 * Arguments
 *	NONE
 * Returns
 *	NOTHING
 * Global Data Affected
 *	NONE
 */
void print_debug_options( void )
{
	int i;
	printf(MSGSTR(DEBUG_OPTIONS, "Debug Options are:\n"));
	 for (i = 0; i < sizeof(options)/sizeof(options[0]); i++) {
		printf("\t%s\n", options[i].opt_name);
	}
}


#if 0
/*
 * Function:
 *	probe
 *
 * Description:
 *	User-level interface to get_lockd_pid, defined in clm_main.c
 *  Linux: read this from /proc
 *
 * Arguments:
 *	NONE
 *
 * Return value:
 *	PID of lock manager kproc
 *
 * Side effects:
 *	NONE
 */

/* TEMPORARY until /proc */
pid_t get_lockd_pid(void)
{
  return(getpid());
}

static int
probe (void)
{
  /* Read from /proc! */
    return(get_lockd_pid());
}

/*
 * None of the following used on Linux port, delete eventually...
 */
typedef int (*funcp)();

funcp func_list[] = { start, probe };

/*
 * Function:
 *	load_func
 *
 * Description:
 *	This is the entry point defined for the clm_loader module.  See the
 * AIX manual for load() to see what the significance of that fact might be.
 * It returns a pointer to the array of function pointers used by the lock
 * manager glue code.
 *
 * Arguments:
 *	int n;		-- number of functions the glue thinks we provide
 *
 * Return value:
 *	Address of function table, or NULL on error
 *
 * Side effects:
 *	NONE
 */

funcp
load_func (int n)
{
    if (n < (sizeof(func_list)/sizeof(func_list[0])))
	return(func_list[n]);
    else
	return(NULL);
}
#endif /* 0 */
