/*
 * User-space monitor modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_config.c,v 1.6 2001/11/21 00:13:35 idr Exp $"

/* ============================================================ */
/*
 * Read the DLM configuration file.
 *
 * The configuration file contains the list of nodes that make
 * up the cluster of interest for this DLM.  The basic information
 * is:
 * - node number, an integral identifier in range 1-4095.
 * - node name (e.g., uname -n)
 * - node primary IP address, the IP address we should use for
 *  communicating with the node.
 *
 * We may also look into putting some of the current command-line
 * options into this file, but save that for later...
 *
 * Real code, download some open source .ini file parser.  For
 * now, hack up my db2 prototype code.
 *
 * Format:
 *  - NODECOUNT <number of nodes>
 *  - [<node number> <node name> <node IP>] one line for each node
 *      entries MUST be separated by ONLY a single space on each line...
 *  a "#" indicates a comment, blank lines acceptable
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "dlm_cluster.h"
#include "clmstructs.h"
#include "dlm_devmsg.h"
#include "dlm_kernel.h"

/* ============================================================ */

#define LINESIZE 256
#define KEY_NODECOUNT "NODECOUNT"
static int have_nodecount = 0;
#define KEY_DLMNAME "DLMNAME"
static int have_dlmname = 0;
#define KEY_DLMMAJOR "DLMMAJOR"
static int have_dlmmajor = 0;
#define KEY_DLMCMGR "DLMCMGR"
static int have_dlmcmgr = 0;
#define KEY_DLMADMIN "DLMADMIN"
static int have_dlmadmin = 0;
#define KEY_DLMLOCKS "DLMLOCKS"
static int have_dlmlocks = 0;

/* -------------------------------------------------- */
/*
 * Return 0 if first char is "#" or blank line.  If we find any
 * non-blank char (except "#") consider it a non-comment line.
 * If not a comment line, return index of first non-blank
 * character.
 */
int comment_line(char *line)
{
  int i;
  for (i = 0; i < LINESIZE; i++) {
    if ('\n' == line[i]) {              /* End of line... */
      break;
    }
    if ('#' == line[i]) {
      break;                            /* Comment! */
    }
    if (' ' != line[i])
      return(i);                        /* Something on line... */
  }
  return(-1);
}

/* -------------------------------------------------- */
/*
 * Combine strings "base" + "/" + "suffix" into buffer at targ.
 * Assume it fits...
 */
static void make_device_name(char *targ,
                             char *base,
                             char *suffix)
{
  strcpy(targ, base);
  strcat(targ, "/");
  strcat(targ, suffix);
}

/* -------------------------------------------------- */
int premature_end(void) 
{

#ifdef DO_WE_WANT_REQUIRE_THESE_ALL_IN_CONFIG
  if (have_dlmname &&
      have_nodecount &&
      have_dlmmajor &&
      have_dlmcmgr &&
      have_dlmadmin &&
      have_dlmlocks) {
    return(0);
  }

  fprintf(stderr, "End of configuration file.");
  if (!have_dlmname) {
    fprintf(stderr, "Missing DLMNAME name entry.");
  }
  if (!have_nodecount) {
    fprintf(stderr, "Missing NODECOUNT entry.");
  }
  if (!have_dlmmajor) {
    fprintf(stderr, "Missing DLMMAJOR entry.");
  }
  if (!have_dlmcmgr) {
    fprintf(stderr, "Missing DLMCMGR entry.");
  }
  if (!have_dlmadmin) {
    fprintf(stderr, "Missing DLMADMIN entry.");
  }
  if (!have_dlmlocks) {
    fprintf(stderr, "Missing DLMLOCKS entry.");
  }
#endif /* DO_WE_WANT_REQUIRE_THESE_ALL_IN_CONFIG */

  if (have_nodecount) {
    return(0);
  }

  fprintf(stderr, "End of configuration file.");
  if (!have_nodecount) {
    fprintf(stderr, "Missing NODECOUNT entry.");
  }

  return(1);
}

/* -------------------------------------------------- */
/*
 * Read the config file, and fill in the node configuration
 * array.  if successful, return a pointer to the array, if
 * not, return NULL.
 */
dlm_node_array_t *parse_config_file(arg_struct_t *args)
{
  char confline[LINESIZE];
  char name[LINESIZE];
  char value[LINESIZE/2];
  int value2, linenum, len;
  char ipvalue[LINESIZE];
  int first_char, parsing_nodes, num_nodes, nodeidx;
  dlm_node_t *node_entries = NULL;
  dlm_node_array_t *tnodes = NULL;
  struct in_addr inipadd;
  FILE *config_fp;

  fprintf(stderr,
          "Opening configuration file [%s]\n",
          args->dlmconfigfile);

  /*
   * Can we even read the file?  Try opening and see...
   */
  if (NULL == (config_fp = fopen(args->dlmconfigfile, "r"))) {
    fprintf(stderr, "Error [%d/%s] opening config file [%s]!\n",
            errno,
            strerror(errno),
            args->dlmconfigfile);
    return(NULL);
  }
  fprintf(stderr, "Opened configuration file.\n");

  /*
   * Walk to the end of file, reading each line.  Look for the
   * keywords on each line, make sure all is in right place and
   * right order.
   */
  linenum = 0;
  parsing_nodes = 0;
  nodeidx = 0;

  while ( NULL != fgets(confline, 256, config_fp) ) {
    linenum++;
#ifdef DEBUG
    fprintf(stderr,
            "Linenum: %d  Line: %s",
            linenum,
            confline);
#endif
    if (-1 == (first_char = comment_line(confline))) {
      continue;                         /* Comment line, ignore it. */
    }
    /*
     * Fun part.  If we saw NODECOUNT, then the next X non-comment
     * lines are nodes or partitons.  Otherwise, all lines are
     * KEYWORD value.
     */
    if (parsing_nodes) {
      sscanf(confline + first_char, "%d %s %s",
             &(node_entries->dlm_node_number),
             &name[0],
             &ipvalue[0]);
      if ((DLM_MINNODE_NUMBER > node_entries->dlm_node_number) ||
          (DLM_MAXNODE_NUMBER < node_entries->dlm_node_number)) {
        fprintf(stderr,
                "Node number [%d] not in range %d to %d, inclusive.\n",
                node_entries->dlm_node_number,
                DLM_MINNODE_NUMBER,
                DLM_MAXNODE_NUMBER);
        return(NULL);
      }

      len = strlen(name);
      if (MAX_NODENAME_LEN <= len) {
        fprintf(stderr,
                "Given nodename [%s] longer than limit of [%d]\n",
                name,
                MAX_NODENAME_LEN);
        return(NULL);
      }
      strcpy(node_entries->dlm_node_name, name);

      if (0 == inet_aton(ipvalue, &inipadd)) {
        fprintf(stderr, "Cannot convert IP address [%s]\n", ipvalue);
        return(NULL);
      }

      node_entries->dlm_node_address.s_addr = inipadd.s_addr;

      node_entries++;                   /* Next entry */

      if (++nodeidx == num_nodes) {
        parsing_nodes = 0;              /* Expect no more of these... */
      }
    
      continue;
    }

    /*
     * Whew.  Keyword value lines are all that are left.
     */
    sscanf(confline + first_char, "%s %s %d", name, value, &value2);
    if (0 == strcmp(KEY_NODECOUNT, name)) {
      have_nodecount = 1;
      parsing_nodes = 1;
      sscanf(value, "%d", &num_nodes);
      /*
       * One dlm_node_t comes with the array.
       */
      tnodes = (dlm_node_array_t *)malloc(sizeof(dlm_node_array_t) +
                                          ((num_nodes - 1) *
                                           sizeof(dlm_node_t)));
      tnodes->dlm_node_count = num_nodes;
      node_entries = (dlm_node_t *)&(tnodes->dlm_node_entries[0]);
    } else if (0 == strcmp(KEY_DLMNAME, name)) {
      have_dlmname = 1;
      strcpy(args->dlmdevicebase, value);
    } else if (0 == strcmp(KEY_DLMMAJOR, name)) {
      have_dlmmajor = 1;
      sscanf(value, "%d", &args->dlmdevicemajor);
    } else if (0 == strcmp(KEY_DLMCMGR, name)) {
      have_dlmcmgr = 1;
      strcpy(args->dlmclustermgr, value);
    } else if (0 == strcmp(KEY_DLMADMIN, name)) {
      have_dlmadmin = 1;
      strcpy(args->dlmadminbase, value);
      args->dlmadminminor = value2;
    } else if (0 == strcmp(KEY_DLMLOCKS, name)) {
      have_dlmlocks = 1;
      strcpy(args->dlmlocksbase, value);
      args->dlmlocksminor = value2;
    } else {
      fprintf(stderr,"Invalid configuration file entry!");
      fprintf(stderr,
              "Bad line: %s",
              confline);
      return(NULL);
    }
  } /* while ( NULL != fgets(confline, 256, config_fp) ) ... */

  if (premature_end()) {
    if ( NULL != tnodes ) {
      free( tnodes );
      tnodes = NULL;
    }
  } else {
    /*
     * Backfill with defaults if the config file didn't override
     * things.
     */
    if (!have_dlmname) {
      strcpy(args->dlmdevicebase, DEFDLMDEVICEBASE);
    }
    make_device_name(args->dlmdevicename,
		     "/dev",
		     args->dlmdevicebase);  /* DLM device name */
    make_device_name(args->dlmprocname,
		     "/proc",
		     args->dlmdevicebase);  /* DLM /proc top dirname */
    if (!have_dlmmajor) {
      args->dlmdevicemajor = DEFDLMMAJORNUMBER;
    }
    if (!have_dlmadmin) {
      strcpy(args->dlmadminbase, DEFDLMADMINBASE);
      args->dlmadminminor = DEFDLMADMINMINOR;
    }
    if (!have_dlmlocks) {
      strcpy(args->dlmlocksbase, DEFDLMLOCKSBASE);
      args->dlmlocksminor = DEFDLMLOCKSMINOR;
    }
    make_device_name(args->dlmadminname,
		     args->dlmdevicename,
		     args->dlmadminbase);  /* DLM admin name */
    make_device_name(args->dlmlocksname,
		     args->dlmdevicename,
		     args->dlmlocksbase);  /* DLM locks name */
  }

  if (0 != fclose(config_fp)) {
    fprintf(stderr, "Error [%d/%s] closing config file [%s]!\n",
            errno,
            strerror(errno),
            args->dlmconfigfile);
  }

  return(tnodes);
}
