/*
 * User-space monitor modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_daemon.c,v 1.13 2001/10/24 18:33:03 idr Exp $"

/*
 *   COMPONENT_NAME: DLM User side (dlmdu)
 *
 *   FUNCTIONS: MSGSTR
 *		catch_and_die
 *		main
 *
 */

/* ============================================================ */
/*
 * For the Linux version, this is the user-side "controller" for
 * the DLM.  The core of the actual lock management are contained
 * in the kernel code, with the device driver for the /dev/haDLM
 * entry.
 *
 * This controller module is not responsible for any actual DLM
 * function.  It does not mediate lock requests, it does not drive
 * protocols, or much anything else.  All of that is in the
 * kernel code.  What this module DOES do (and there you were,
 * wondering if I'd ever get to the point :-) is:
 * - read the configuration file to get the list of nodes.
 * - parse command line options that control settings.
 * - ensure that the device driver is loaded.
 * - pass in the relevant control settings to the kernel.
 * - connect to the specified cluster infrastructure.
 * - pass events from the cluster infrastructure to the DLM core.
 * - catch signals that may force us to stop the DLM function.
 *
 * Node Identification and Config File
 *
 * All of the code in the DLM expects nodes to be identified by
 * a single integral identifier, i.e., a node number.  The nodes
 * are tracked by these node numbers, and inter-node communication
 * is handled by using the node number to look up an IP address
 * for each node.
 *
 * The DLM expects a configuration file that conveys three pieces
 * of information for EACH node in the cluster:
 * - node name (e.g., uname -n value)
 * - node number
 * - node primary IP address
 * We then expect the cluster infrastructure to reflect node
 * status for this information, with an API that either uses the
 * node name or the node number.  If the cluster infrastructure
 * supports reporting of multiple IP address information and
 * changing these, we can accept those as well.
 *
 * Cluster Infrastructure
 *
 * The DLM code does not determine two rather key pieces of
 * information:
 * - which nodes are up and down in the cluster, and
 * - which IP address to use AT THIS TIME for each node.
 * We require some cluster infrastructure to dynamically
 * determine these for us.  Examples of existing cluster
 * infrastructures are Phoenix, Heartbeat and FailSafe.
 *
 * The up and down nodes are, of course, key in that we need to
 * know whether a node is allowed to participate or not.  When
 * a node comes up, we allow it to generate lock requests and
 * to master locks.  When it goes down, we need to recover its
 * locks.  The DLM core code is capable of switching to different
 * IP addresses for each node, thus, if for some reason a node's
 * "best" IP address changes, we need to be told.
 *
 * Note that the concept of node's changing IP addresses is
 * optional for the DLM, thus if the cluster infrastructure
 * does not support it, no worries, it simply means we're stuck
 * hoping that interface stays up, and tied to the physical node.
 * It is the node up/down status that we must have.
 *
 * Node Status
 *
 * We do not care what heartbeating methods are used, we do
 * expect that the cluster infrastructure maintains cluster-wide
 * coherency of node events, i.e., all nodes agree that a node
 * is down, and then agree that it is up.
 */
/* ============================================================ */

#ifndef lint
#endif

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <assert.h>

#ifndef DIRECT_CLUSTINT
#include <dlfcn.h>                      /* Dynamic load support. */
#endif /* !DIRECT_CLUSTINT */

#include "dlm.h"
#include "dlm_devmsg.h"
#include "dlm_kernel.h"
#include "dlm_version.h"
#include "cltrc.h"

#include "dlm_list.h"


#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "cllockd_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("cllockd.cat"), CLLOCKD, Num, Str)
/* end of NLS */
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */

/* ============================================================ */
/*
 * File descriptor for open file against the admin device port
 * to the DLM.
 */
int admin_fd;

static dlm_node_array_t *configured_nodes;

static dlm_node_array_t *copy_node_array(dlm_node_array_t *nodes);

int admin_calls;                   /* Number of blocks written to kernel. */

int this_node_number;              /* Local node number. */

/* ============================================================ */
/*
 * These are only ever to be called from this module, so, just
 * put the declarations here.
 */
int parse_options(int argc, char *argv[]);
dlm_node_array_t *parse_config_file(arg_struct_t *args);
int verify_dev_entries(arg_struct_t *args);
int load_dev_driver(arg_struct_t *args);

/* ============================================================ */
/*
 * Functions in this mainline manual that are used to interact
 * with the cluster interface code.
 */
int cluster_manager_setup(dlm_node_array_t *nodes);
#ifndef DIRECT_CLUSTINT
void *cluster_interface_call(void *arg);
#endif /* !DIRECT_CLUSTINT */
int drain_event_queue(MQ_DLM_TOP_INFO_t *node_events);
int find_configured_node(dlm_node_t *node);

/* ============================================================ */
/*
 * Thread items for dealing with the cluster manager module.
 *
 * The main thread waits on a semaphore, the second thread
 * deals with the cluster manager.  When it gets data from
 * the cluster code, it dumps it on a queue and wakes up the
 * main thread.
 */

pthread_t cmgr_thread_id = 0;
pthread_mutex_t notify_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t notify_cond = PTHREAD_COND_INITIALIZER;

typedef struct dlm_nodeq {
  dlm_listnode_t link;
  dlm_node_t *node;
} dlm_nodeq_t;

dlm_list_t  private_node_queue;
dlm_list_t * nodeq = & private_node_queue;
int  nodeq_count = 0;

#define EVENT_WAIT_TIME 300             /* How long to wait for events */

/* ============================================================ */
/*
 * Signal handling.
 */

#define ERROR (-1)

enum { EOK, ESYSCONFIG, ELOAD, EINTERNAL };

static int tCaughtSig;
static int tCaughtSigIO;
static int handleSigIO;

static int need_to_quit = 0;

/*
 * Not too much to do, basically we just close the
 * admin interface, and that tells the kernel side
 * to shut down all locking services.
 */
static void
cleanup( void )
{
  if (-1 == admin_fd) return;

  if (0 != close(admin_fd)) {
    fprintf(stderr,
            "Error closing admin file [%d], errno [%s]\n",
            admin_fd,
            strerror(errno));
  } else {
    fprintf(stderr,
            "Closed admin file [%d, %s]\n",
            admin_fd,
            args.dlmadminname);
    admin_fd = -1;
  }

  if (0 != cmgr_thread_id) {
#ifdef DIRECT_CLUSTINT
    cmgr_thread_id = 0;
    fprintf(stderr,
            "Shutting down cluster interface.\n");
    cluster_interface_shutdown();
#else /* DIRECT_CLUSTINT */
    (*cluster_interface_stop)();
#endif /* DIRECT_CLUSTINT */
  }

#ifdef _ON_AIX_DLM_
	mid_t kmid;
	int status;

	cltrc(L_USR, CLLOCKD_HKID, T_ENTR_FUNC, T_CLLOCKD_CATCHER, 0x00, 0x00,
		0x00, 0x00);

	/* kill the kernel process and wait for it to exit */
	while (kill(pid,0) != ERROR) {
		(void)kill(pid,SIGKILL);
		sleep(3);
		(void)waitpid(pid, &status, WNOHANG);
	}
#endif /* _ON_AIX_DLM_ */

}
void catch_and_die(int sig)
{
  ++tCaughtSig;

  fprintf(stderr,
          "Caught signal [%d] count [%d]\n",
          sig,
          tCaughtSig);

  if (1 < tCaughtSig) return;

  cleanup();

  need_to_quit = 1;

  /*exit(sig);*/
  return;
}
void handle_sigio(int sig)
{
  ++tCaughtSigIO;

  fprintf(stderr,
          "Caught SIGIO number %d\n",
          tCaughtSigIO);

  handleSigIO++;
}

void
initialize_node_events(MQ_DLM_TOP_INFO_t *node_events)
{
    dlm_node_t *entry;
    MQ_DLM_USEADDR_t *useaddr;

    useaddr = &node_events->addrs[0];

    for(entry = &configured_nodes->dlm_node_entries[0];
	entry < &configured_nodes->dlm_node_entries
	  [configured_nodes->dlm_node_count];
	entry++) {

        useaddr->nodeid = entry->dlm_node_number;
        useaddr->useaddr.s_addr = entry->dlm_node_address.s_addr;
	useaddr->dlm_major = 0;
	useaddr->dlm_minor = 0;

	useaddr++;
    }
}


/* ============================================================ */

char *main_compiled = NULL;

int main(int argc, char *argv[])
{
    int		rc, i, j, msize;
    struct stat statbuf;
    dlm_MSG_t message;
    MQ_DLM_TOP_INFO_t *node_events;
    struct timespec event_timer;
    int  time_since = 0;                /* Time between cluster events */
    pid_t   child_pid;

    admin_calls = 0;
    this_node_number = -1;

    dlm_list_init( nodeq );

#ifdef _ON_AIX_DLM_
    cltrc(L_USR, CLLOCKD_HKID, T_ENTR_FUNC, T_CLLOCKD_MAIN, 0x00,
		0x00, 0x00, 0x00);
#endif /* _ON_AIX_DLM_ */

    /* if we're not on a tty, then close all fd's */
    if (fstat(0, &statbuf) != 0 || !isatty(0)) {
      if (0 != (rc = close(0))) {
		fprintf(stderr, "Error %d closing 0\n", errno);
      }

      if (0 != (rc = close(1))) {
		fprintf(stderr, "Error %d closing 1\n", errno);
      }

      if (0 != (rc = close(2))) {
		fprintf(stderr, "Error %d closing 2\n", errno);
      }
    }

    /*
     * Play the field.  If someone sends a SIGKILL to the
     * dlmdu process, the kernel side will still shut down,
     * as our admin hookup will be closed when our process dies.
     * We catch these signals in the interest of completeness,
     * since the end effect will be the same.
     */
    signal(SIGHUP, catch_and_die);
    signal(SIGINT, catch_and_die);
    signal(SIGQUIT, catch_and_die);
    signal(SIGIOT, catch_and_die);
    signal(SIGTERM, catch_and_die);
    signal(SIGUSR1, catch_and_die);
    signal(SIGUSR2, catch_and_die);
    signal(SIGCHLD, SIG_IGN);
    signal(SIGIO, handle_sigio);

    /*
     * Simple, really.
     * - parse command line options.
     * - read and digest config file.
     * - verify (or create) /dev entries.
     * - load device driver (if necessary).
     * - contact cluster manager.
     * - initialize kernel side lock manager.
     * - sit here and wait for cluster manager messages.
     * Exit if any step returns an error.
     */

    if (0 != parse_options(argc, argv)) {
      fprintf(stderr,
              "Error parsing command line options.  Exiting.\n");
      exit(1);
    }

    if (NULL == (configured_nodes = parse_config_file(&args))) {
      fprintf(stderr,
              "Error reading configuration file.  Exiting.\n");
      exit(2);
    }

    if (0 != verify_dev_entries(&args)) {
      fprintf(stderr,
              "Error establishing /dev entries.  Exiting.\n");
      exit(3);
    }


    /* Create a child process so that the main process can return success
     * to the calling shell.  This is useful for init scripts.  The child
     * process will do the remainder of the initialization and run the
     * actual user-mode DLM duties.
     */

    child_pid = fork();
    if ( child_pid == -1 ) {
      fprintf(stderr,
	      "Unable to daemonize (fork failed).  Exiting.\n");
      exit(4);
    } else if ( child_pid != 0 ) {
      exit(0);
    }

    setsid();
    fclose( stdin );

    /*
     * Load device driver and open the "admin" device for
     * operations.
     */
    if (-1 == (admin_fd = load_dev_driver(&args))) {
      fprintf(stderr,
              "Error loading device driver.  Exiting.\n");
      exit(4);
    }

    printf("1st thread, pid [%d]\n",
           getpid());


    /* -------------------------------------------------- */
    /*
     * At this point, we have created the necessary /dev
     * structures, received our static config info, loaded
     * the device driver and successfully opened the
     * "admin" portal.
     *
     * Hoever, the kernel side will continue to refuse any
     * requests for locks until we write through "admin"
     * portal to provide the kernel side with the cluster
     * manager data:
     * - if user-side cluster manager, we connect and get
     *      cluster status, and pass it through and the
     *      kernel is ready to go.
     * - if kernel-side cluster manager, we pass that
     *      through, and basically our job is more or
     *      less done, we simply sit here and if we catch
     *      a signal, shut everything down.
     */
    /* -------------------------------------------------- */
    /*
     * Here is where we would dynamically load the appropriate
     * cluster manager hook-up and get connected (failing that,
     * we freeze, and no DLM function happens.)  In the hypothetical
     * case of a kernel-based cluster manager, that would bypass
     * this step, and go onto to load the kernel DLM and let IT
     * do a direct connect to the cluster manager.  In that case,
     * our purpose is basically null and void from here on out.
     */
    /*
     * We expect the dynamic module to provide an init() routine
     * that will do all work necessary for us to:
     * - connect to the cluster manager
     * - create a subsidiary thread to monitor the connection
     * - to issue any additional appropriate calls.
     *
     * The main thread will be here, and will use a pthread_mutex
     * to "monitor" input from the cluster manager thread.  It
     * shoud post any events and inform the main thread, which
     * will take the data and massage it for the kernel side.
     *OR
     * publish a routine, that the subsidiary thread calls and
     * it does the submission to the kernel.  The main thread
     * basically monitors for signals, and for anything "bad"
     * to happen.
     */
    /*
     * Now, for quickie testing, I'm just going to directly hook
     * to an rsct interface.
     */

    /*
     * Fire up the cluster manager interface module.  A
     * thread will be created and dedicated to that
     * work.  Once we are connected, start up the kernel
     * thread so that it is ready to work.
     */
    if (-1 == cluster_manager_setup(configured_nodes)) {
      fprintf(stderr,
              "Error loading cluster interface module.  Exiting.\n");
      cleanup();
      exit(5);
    }

    message.msg_code = haDLM_MSG_Init;
    message.msg_ptr_size = sizeof(args);
    message.arg_block = &args;
    i = write(admin_fd, &message, sizeof(message)); /* Tell kernel. */
    if (i != sizeof(message)) {
      fprintf(stderr,
              "Error on write, i/err[%d/%s], code/size[%d/%d]\n",
              i,
              strerror(errno),
              message.msg_code,
              sizeof(message));
      cleanup();
      assert(!"Error writing through DLM admin device");
    } else {
      admin_calls++;

      fprintf(stderr,
              "Wrote code/size[%d/%d], blocks so far [%d]\n",
              message.msg_code,
              sizeof(message),
              admin_calls);
    }
    node_events = NULL;

    /*
     * Nothing more to do here than to simply wait for the
     * cluster interface code to start sending us events.
     * Once we start getting those, we can then open up the
     * DLM to clients wanting to do work.
     */

    pthread_mutex_lock(&notify_mutex);


    /* initialize message block with all nodes down
     */
    msize = (configured_nodes->dlm_node_count *
	     sizeof(MQ_DLM_USEADDR_t)) + sizeof(MQ_DLM_TOP_INFO_t);
    node_events = (MQ_DLM_TOP_INFO_t *)malloc(msize);
    if (NULL == node_events) {
      fprintf(stderr,
	      "Cannot allocate node events block to drain node events!\n");
      cleanup();
      assert(!"Cannot allocate node events block to drain node events!");
    }

    /* initialize message buffer; reused for each updata
     */
    initialize_node_events(node_events);

    while(!need_to_quit) {

      if (0 < nodeq_count) {
        /*
         * Work to do!  Node block(s) waiting.
         */

        message.msg_code = haDLM_MSG_CMGR;
        message.msg_ptr_size = msize;
        message.topo_array = node_events;

        /*
         * This will return the actual number of nodes that are
         * in the block, as some events may have been dropped
         * for rollup or because they are for non-recognized
         * nodes.
         */
        j = drain_event_queue(node_events);

        if (0 < j) {

          rc = pthread_mutex_unlock(&notify_mutex);
          if (0 != rc) {
            fprintf(stderr,
                    "Cannot unlock notify_mutex in main loop, rc = [%d]\n",
                    rc);
            cleanup();
            exit(rc);
          }

          i = write(admin_fd, &message, sizeof(message)); /* Tell kernel. */

          if (i != sizeof(message)) {
            fprintf(stderr,
                    "Error on write, i/err[%d/%s], code/size[%d/%d]\n",
                    i,
                    strerror(errno),
                    message.msg_code,
                    sizeof(message));
            cleanup();
            assert(!"Error writing through DLM admin device");
          } else {
            admin_calls++;

            fprintf(stderr,
                    "Wrote code/size[%d/%d], blocks so far [%d]\n",
                    message.msg_code,
                    sizeof(message),
                    admin_calls);
          }

          rc = pthread_mutex_lock(&notify_mutex);
          if (0 != rc) {
            fprintf(stderr,
                    "Cannot lock notify_mutex in main loop, rc = [%d]\n",
                    rc);
            cleanup();
            exit(rc);
          }

          time_since = 0;                 /* Had events! */
        }

      } else {
        fprintf(stderr,
                "No work to do in handling event queue.\n");
      }

      /*
       * Mutex is held at this point, regardless of path above.
       */
      if (0 == nodeq_count) {
        /*
         * No blocks appeared while we were in the kernel, wait
         * on the condition.
         */
        event_timer.tv_sec = time(NULL) + EVENT_WAIT_TIME;
        event_timer.tv_nsec = 0;

        rc = pthread_cond_timedwait(&notify_cond,
                                    &notify_mutex,
                                    &event_timer);
        if (ETIMEDOUT == rc) {
          time_since += EVENT_WAIT_TIME;
          fprintf(stderr,
                  "Condition timeout, last cluster event [%d] seconds ago.\n",
                  time_since);
        } else if (0 != rc) {
          fprintf(stderr,
                  "Cannot wait on notify_cond main loop, rc = [%d]\n",
                  rc);
          cleanup();
          exit(rc);
        }
      }

      /*
       * Loop around anyway, and check to see if any events, even
       * if we timed out.  If there are none, we'll just get right
       * back to the conditon wait anyway.
       */

    } /* while(!need_to_quit) */

    /*
     * This is the message we may have, or may not have, sent
     * to the kernel.  Release it.
     */
    if (NULL != node_events) {
      free(node_events);
      node_events = NULL;
    }

#ifdef ALL_NIGHT_TEST 
    /* TEMP: run all night, see if anything weird happens. */
    j = haDLM_MSG_Init;
    while(1) {
      close(admin_fd);
      admin_fd = open(args.dlmadminname, O_RDWR);

      if (-1 == admin_fd) {
        fprintf(stderr,
                "Error opening [%s], errno [%s]\n",
                args.dlmadminname,
                strerror(errno));
        return(-1);
      } else {
        fprintf(stderr,
                "Opened [%s], filedes [%d]\n",
                args.dlmadminname,
                admin_fd);
      }

      switch(j) {
      case haDLM_MSG_Init:
        message.msg_code = j++;
        strcpy(message.msg_data, "haDLM_MSG_Init");
        break;
      case haDLM_MSG_Stop:
        message.msg_code = j++;
        strcpy(message.msg_data, "haDLM_MSG_Stop");
        break;
      case haDLM_MSG_API:
        message.msg_code = j++;
        strcpy(message.msg_data, "haDLM_MSG_API");
        break;
      case haDLM_MSG_CMGR:
        message.msg_code = j++;
        strcpy(message.msg_data, "haDLM_MSG_CMGR");
        break;
      default:
        j = haDLM_MSG_Init;
      }

      i = write(admin_fd, &message, sizeof(haDLM_MSG_t));
      if (i != sizeof(haDLM_MSG_t)) {
        fprintf(stderr,
                "Error on write, i=[%d], buffer[%d/%s]\n",
                i,
                message.msg_code,
                message.msg_data);
        exit(1);
      } else {
        fprintf(stderr,
                "Wrote buffer[%d/%s]\n",
                message.msg_code,
                message.msg_data);
      }

      fprintf(stderr, "Sleeping...");
      for (i = 0; i < 30; i++) {
        fprintf(stderr, ".");
        sleep(1);
      }
      fprintf(stderr, "\n");
    }
#endif /*  ALL_NIGHT_TEST */

    exit(EOK);
}

/* ============================================================ */
/*
 * drain_event_queue
 *
 * Drain the node/interface event queue, and roll-up events
 * where it makes sense:
 * - drop an interface down if a node down is also found in
 *  the queue.
 * 
 * This routine also needs to ensure each block represents a
 * node in our configured nodes table.  It is possible that
 * the cluster reports on more nodes than exist.  Although we
 * pass that table to the cluster interface code, we do not
 * completely trust (more kindly, we trust but verify) it to
 * filter non-configured nodes.  At worst, we duplicate the
 * work.  For now, we have so few nodes, it's not too big of
 * a problem, also, in a "decent" cluster we should not see
 * too many of these events anyway.
 *
 * The caller must hold the nodeq mutex upon entry, and it
 * will be held the entire time within this routine.  Upon
 * return from this routine, we will have constructed a
 * single data block containing all events to be passed into
 * the kernel.
 *
 * Collect events from the "end" of the queue (FIFO) and
 * order them in the data block we're given.
 *
 * NOTE: the code in the kernel (handle_new_topology() in
 * clm_info.c) that eventually processes this data seems to
 * expect the FULL topology each time...  Which is not what
 * I had coded in here.  For the immediate moment, running on
 * one node, not a big deal, but must fix this before going
 * live.
 */

int
drain_event_queue( MQ_DLM_TOP_INFO_t *node_events )
{
  static int need_local = 1;
  int count;
  dlm_node_t *qnode;
  dlm_nodeq_t *que;
  MQ_DLM_USEADDR_t *useaddr;

  count = 0;

  while ( (que = (dlm_nodeq_t *) dlm_list_rem_head( nodeq )) != NULL )
  {
    qnode = que->node;
    assert( qnode != NULL );

    /*
     * Rollup handling later...  For now just determine if
     * the node is in the config list.
     */
    if (find_configured_node(qnode)) {
      if (haDLM_state_local == qnode->dlm_node_state) {
        if (need_local) {
          this_node_number = qnode->dlm_node_number;
          need_local = 0;
        }
      } else {
	/* Find message slot and adjust liveness status (in major/minor and
	 * IP address).
	 */
	for (useaddr = &node_events->addrs[0];
	     useaddr < &node_events->addrs[configured_nodes->dlm_node_count];
	     useaddr++) {

	  if (useaddr->nodeid == qnode->dlm_node_number) {
	    if (qnode->dlm_node_state == haDLM_state_up) {
	      useaddr->dlm_major = DLM_MAJOR_VERSION;
	      useaddr->dlm_minor = DLM_MINOR_VERSION;
	      useaddr->useaddr.s_addr = qnode->dlm_node_address.s_addr;
	    } else {
	      useaddr->dlm_major = 0;
	      useaddr->dlm_minor = 0;
	      useaddr->useaddr.s_addr = 0;
	    }
	    break;
	  }
	}
        count++;
      }
    } else {
      /*
       * Reported node not found in config file, not a problem...
       * UNLESS, it is also reported as the local node!!!  This
       * tells us we are not in our own configuration, or, there
       * is a config file problem.  In any case, report it and
       * give up.
       */
      if ((!need_local) && (this_node_number == qnode->dlm_node_number)) {
        fprintf(stderr,
                "Local node is not in config file!\n");
        fprintf(stderr,
                "  number/ifaddr/name[%d/%x/%s]\n",
                qnode->dlm_node_number,
                qnode->dlm_node_address.s_addr,
                qnode->dlm_node_name);
        assert(!"Local node is not in config file!");
      }
    }

    nodeq_count--;
    free(qnode);
    free(que);
  }

  if ( 0 != nodeq_count )
  {
      fprintf( stderr,
	       "Done processing node queue, but count not zero (%d)\n",
	       nodeq_count );
      exit( 10 );
  }



  /*
   * Number of blocks being passed through, could have changed
   * if there were blocks we skipped due to them not being in
   * the configuration.
   */
  if (0 < count) {
    node_events->n_nodes = configured_nodes->dlm_node_count;
    node_events->msg_version = TAM_version;
    node_events->this_nodeid = this_node_number;
    node_events->event = admin_calls;
  }

  
  return(count ? configured_nodes->dlm_node_count : 0);
}

/* ------------------------------------------------------------ */
/*
 * Determine that the node given is in the configured node
 * table.  If it is, return 1, if not, return 0.
 *
 * If the incoming block has an empty name field, copy the
 * name from the configured node table into the block.
 * Likewise if it has no node number, copy the node number
 * from the configured node table.
 */
int
find_configured_node(dlm_node_t *node)
{
  int count, found, len, dot;
  char *dote, *dotn;
  dlm_node_t *entry;

  assert( node != NULL );

  count = configured_nodes->dlm_node_count;
  found = 0;

  for(entry = &configured_nodes->dlm_node_entries[0];
      ((!found) && (0 < count));
      entry++, count--) {
    if (-1 != node->dlm_node_number) {
      if (entry->dlm_node_number == node->dlm_node_number) {
        found = 1;
        if ((haDLM_state_local == node->dlm_node_state) &&
            (haDLM_state_local != entry->dlm_node_state)) {
          entry->dlm_node_state = haDLM_state_local;
        }
        break;
      }
    } else {
      len = strlen(node->dlm_node_name);
      if ((0 == len) ||
          (MAX_NODENAME_LEN <= len)) {
        continue;
      }
      fprintf(stderr,
              "compare [%s] and [%s]\n",
              entry->dlm_node_name,
              node->dlm_node_name);
      /*
       * If only one name has a dot in it, compare only up to
       * the len of the non-dot name.  If neither has a dot,
       * do full compare.  If both have a dot, do full compare.
       */
      dote = index(entry->dlm_node_name, '.');
      dotn = index(node->dlm_node_name, '.');
      if (((NULL != dote) && (NULL != dotn)) ||
          ((NULL == dote) && (NULL == dotn))) {
        if (0 == strcmp(entry->dlm_node_name,
                        node->dlm_node_name)) {
          found = 1;
          break;
        }
      } else {
        /* get len of string to dot */
        if (NULL == dote) {
          dot = dotn - node->dlm_node_name;
        } else {
          dot = dote - entry->dlm_node_name;
        }
        if (0 == strncmp(entry->dlm_node_name,
                         node->dlm_node_name,
                         dot)) {
          found = 1;
          break;
        }
      }
    }
  }

  if (found && ('\0' == node->dlm_node_name[0])) {
    strcpy(node->dlm_node_name, entry->dlm_node_name);
  }
  if (found && (-1 == node->dlm_node_number)) {
    node->dlm_node_number = entry->dlm_node_number;
  }
  if (found && (haDLM_state_null == node->dlm_comm_state)) {
    node->dlm_comm_state = haDLM_state_up;
    node->dlm_node_address = entry->dlm_node_address;
  }
  if (found) {
    fprintf(stderr,
            "found number/name[%d/%s/%s]\n",
            node->dlm_node_number,
            node->dlm_node_name,
            inet_ntoa(node->dlm_node_address));
  } else {
    fprintf(stderr,
            "not found\n");
  }

  return(found);
}

/* ============================================================ */
/*
 * Set up to call the cluster manager.
 *
 * NOTE: later need to fix for dynamic loading...
 */
int cluster_manager_setup(dlm_node_array_t *nodes)
{
  dlm_node_array_t *copy_nodes;
  int status = 0;
#ifndef DIRECT_CLUSTINT
  char libname[256];
  char *error;
#endif /* !DIRECT_CLUSTINT */

  copy_nodes = copy_node_array(nodes);  /* Send in copy. */

#ifndef DIRECT_CLUSTINT
  /*
   * Dynamically load the library, and resolve the addresses
   * we'll need.  Any problems, return with an error and
   * we'll give up.
   *
   * If we are not doing the dynamic load, then the reference
   * to the cluster interface call in the pthread_create will
   * resolve directly.
   */

  sprintf(libname, "dlm_%s.so", args.dlmclustermgr);
  fprintf(stderr,
          "Attempt to open cluster interface library [%s]\n",
          libname);
#if 0
  cluster_interface_handle = dlopen(libname,
                                    RTLD_NOW | RTLD_GLOBAL);
#endif
  cluster_interface_handle = dlopen(libname,
                                    RTLD_NOW);
  if (NULL == cluster_interface_handle) {
    fprintf(stderr,
            "Cannot load cluster interface lib [%s]\n",
            dlerror());
    status = -1;
  } else {
    fprintf(stderr,
            "Loaded cluster interface lib, handle [%p]\n",
            cluster_interface_handle);
    cluster_interface_start = dlsym(cluster_interface_handle,
                                    CLUSTER_INTERFACE_START);
    if ((error = dlerror()) != NULL)  {
      fprintf(stderr,
              "Cannot resolve [%s] in cluster lib, error [%s]\n",
              CLUSTER_INTERFACE_START,
              error);
      status = -1;
    } else {
      fprintf(stderr,
              "Resolved [%s] in cluster lib [%p]\n",
              CLUSTER_INTERFACE_START,
              cluster_interface_start);

      cluster_interface_stop = dlsym(cluster_interface_handle,
                                     CLUSTER_INTERFACE_STOP);
      if ((error = dlerror()) != NULL)  {
        fprintf(stderr,
                "Cannot resolve [%s] in cluster lib, error [%s]\n",
                CLUSTER_INTERFACE_STOP,
                error);
        status = -1;
      } else {
        fprintf(stderr,
                "Resolved [%s] in cluster lib [%p]\n",
                CLUSTER_INTERFACE_STOP,
                cluster_interface_stop);
      }
    }
  }

  fprintf(stderr, "Make sure linkages work...\n");
  (*cluster_interface_stop)();
#endif /* !DIRECT_CLUSTINT */

  if (0 == status) {
    /*
   * Create a second thread, and fire it off to run the
   * cluster manager interface work.  We are done, and
   * go back to await events before doing much more.
   */
    status = pthread_create(&cmgr_thread_id,
                            NULL,
#ifdef DIRECT_CLUSTINT
                            cluster_interface_handling,
#else /* DIRECT_CLUSTINT */
                            cluster_interface_call,
#endif /* DIRECT_CLUSTINT */
                            (void *)copy_nodes);

    if (0 != status) {
      fprintf(stderr,
              "Bad return creating cluster interface thread [%d]\n",
              status);
      status = -1;
    } else {
      fprintf(stderr,
              "Created thread, id [%ld]\n",
              cmgr_thread_id);
    }
  }

  return(status);
}

#ifndef DIRECT_CLUSTINT
/* ------------------------------------------------------------ */
/*
 * See if it works this way...
 */
void *cluster_interface_call(void *arg)
{
  void *ret;

  fprintf(stderr,
          "Second thread, calling interface support at [%p]...\n",
          cluster_interface_start);

  ret = (*cluster_interface_start)(arg);

  return(ret);
}
#endif /* !DIRECT_CLUSTINT */

/* ------------------------------------------------------------ */
/*
 * With a view to supporting dynamic reconfiguration of the
 * DLM's node list, make a local copy.  Later, figure out
 * how exactly to support the dynamic reconfig.
 */
dlm_node_array_t *copy_node_array(dlm_node_array_t *nodes)
{
  dlm_node_array_t *copy_nodes;
  int count = nodes->dlm_node_count;

  copy_nodes = (dlm_node_array_t *)malloc(sizeof(dlm_node_array_t) +
                                          ((count - 1) * sizeof(dlm_node_t)));
  if (NULL == copy_nodes) {
    fprintf(stderr,
            "Cannot allocate configured nodes array for RSCT module\n");
    exit(88);
  }

  copy_nodes->dlm_node_count = count;

  memcpy(&copy_nodes->dlm_node_entries[0],
         &nodes->dlm_node_entries[0],
         count * sizeof(dlm_node_t));

  return(copy_nodes);
}

/* ============================================================ */
/*
 * Provide handles for the dynamically-loaded cluster manager
 * interface routines to deliver data to us.
 */

/*
 * About to have a set of notifications delivered.  This does
 * nothing more than grab the mutex, however, due to the fact
 * that this is called across a dynamic module boundary, we
 * want to hide "details."
 */
void dlm_notify_begin(void)
{
  int rc;

  rc = pthread_mutex_lock(&notify_mutex);

  if (0 != rc) {
    fprintf(stderr,
            "Cannot lock notify_mutex, rc = [%d]\n",
            rc);
    exit(rc);
  }
}

/* ------------------------------------------------------------ */
/*
 * Deliver a single node/interface event block.  Caller must
 * hold the lock (called dlm_notify_begin()) and must not call
 * dlm_notify_end() until all blocks have been delivered.
 *
 * Maintain FIFO order of events as delivered, this routine
 * will always insert at front of list, and they will be
 * removed from the "end" of the list.
 */
void dlm_notify_deliver(dlm_node_t *info)
{
  dlm_nodeq_t *que;

  if (NULL == (que = malloc(sizeof(dlm_nodeq_t)))) {
    fprintf(stderr,
            "Cannot allocate dlm_nodeq_t element\n");
    exit(11);
  }

  if ( dlm_list_is_empty( nodeq ) && (0 != nodeq_count) )
  {
    fprintf(stderr,
	    "Bad nodeq mgmt, queue empty: count == [%d]\n",
	    nodeq_count );
    exit(10);
  }

  assert( info != NULL );
  que->node = info;
  dlm_list_add_tail( nodeq, & que->link );
  nodeq_count++;

  fprintf(stderr,
          "Node have [%d] nodeq elements queued\n",
          nodeq_count);
}

/* ------------------------------------------------------------ */
/*
 * All node/interface events have been queued, inform the
 * main thread of this.  This basically unlocks the mutex
 * and pops the condition variable, again, hide this across
 * module boundary.
 *
 * Assumes caller holds the lock (called dlm_notify_begin()),
 * lock will be dropped upon exit, so caller must not deliver
 * any more events without calling begin again.
 */
void dlm_notify_end(void)
{
  int rc;

  rc = pthread_cond_broadcast(&notify_cond);

  if (0 != rc) {
    fprintf(stderr,
            "Cannot broadcast on notify_cond, rc = [%d]\n",
            rc);
    exit(rc);
  }

  rc = pthread_mutex_unlock(&notify_mutex);

  if (0 != rc) {
    fprintf(stderr,
            "Cannot unlock notify_mutex, rc = [%d]\n",
            rc);
    exit(rc);
  }

#if 0
  /* Not in Linux?!? */
  pthread_yield();                      /* Give chance to other thread. */
#endif
}

/* ------------------------------------------------------------ */
/*
 * Dump a given string into log, used by the cluster library
 * code that we loaded, to free it from having to know where
 * to write data.
 */
int dlm_write_buffer(char *buffer)
{
  int ret;

  ret = fprintf(stderr, "%s", buffer);

  return(ret);
}
