/*
 * User-space monitor modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_device.c,v 1.1.1.1 2001/02/14 20:58:04 wombat Exp $"

/* ============================================================ */
/*
 * Verify - and create if necessary - the DLM device directory
 * entries, if that all goes ok then load the device driver,
 * unless it is already loaded.
 *
 * Configuration (for now, command line) options tell us the
 * device name and major number to use, we impose various
 * "subdirectories" (i.e., minor numbers) by fiat and set up
 * all of the relevant device entries.
 *
 * Using the default name (haDLM), we have:
 * /dev/haDLM - root dir, uses major number 250 by default
 * /dev/haDLM/admin - for admin contacts from cllockd.u,
 *  minor number 0
 * /dev/haDLM/locks - for lock usage (1 domain for now...)
 *  minor number 1
 *
 * FUTURE: multiple "locks" domains, each with a different
 *  minor (filename).
 *
 * MISSING: permission setting, right now, limit everything to
 *  root, which is ok for admin, not ok for locks.
 *
 * Consistency: note that the minor numbers - but not the major
 *  number - need to be consistent clusterwide, as do the names
 *  of these "subdirectories", to allow the DLM instances on each
 *  node to pass around lock data and keep it straight.
 */
/* ============================================================ */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/sysmacros.h>
#include <ctype.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "clmstructs.h"
#include "dlm_kernel.h"

/* ============================================================ */
/*
 * Ensure a device exists with the given name, major and minor
 * numbers.  If it doesn't, create it.  If it exists, but is
 * incorrect, delete and create it.  If cannot get it right,
 * return -1.
 *
 * We assume that all directories intermediate to this device
 * already exist, thus we neither check for, nor fix, cases of
 * missing intermediate directories.
 */

static
int verify_device(int major,
                  char *device,
                  int minor)
{
  int rc;
  struct stat statbuf;
  
  if (-1 == (rc = stat(device, &statbuf))) {
    /*
     * Maybe no problem, as it may not exist.
     */
    if (ENOENT == errno) {
      fprintf(stderr,
              "Device [%s] does not exist.  Create it.\n",
              device);
      if (-1 == (rc = mknod(device,
                            S_IFCHR|S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH,
                            makedev(major, minor)))) {
        fprintf(stderr,
                "Error [%s] creating device.\n",
                strerror(errno));
      }
    } else {
      fprintf(stderr, "Error on stat: [%s]\n", strerror(errno));
    }
  } else {
    /*
     * "Something" exists in the expected slot.  What is it?
     */
    if ((!S_ISCHR(statbuf.st_mode)) ||
        (minor != minor(statbuf.st_rdev)) ||
        (major != major(statbuf.st_rdev))) {
      fprintf(stderr,
              "Fixing device [%s], type/maj/min[0x%x/%d/%d] want maj/min[%d/%d]\n",
              device,
              statbuf.st_mode,
              major(statbuf.st_rdev),
              minor(statbuf.st_rdev),
              major,
              minor);
      if (-1 == (rc = unlink(device))) {
        fprintf(stderr,
                "Cannot unlink device [%s] error [%s].\n",
                device,
                strerror(errno));
      } else {
        if (-1 == (rc = mknod(device,
                              S_IFCHR|S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH,
                              makedev(major, minor)))) {
          fprintf(stderr,
                  "Error [%s] creating device.\n",
                  strerror(errno));
        }
      }
    }
  }

  return(rc);
}

/* ============================================================ */
/*
 * Verify that the dev entries exist and match the desired
 * configuration.  If they don't, then attempt to fix them.
 * We do not load the device driver yet, although we probably
 * could.
 *
 * What if, the device exists, but already has a different
 * major number?  Attempt to deconfigure everything, and to
 * redefine - which may involve trying to unload the device
 * driver.  It we fail, signal error and give up.
 *
 * Return zero if all is well, !0 if any problems.
 */

int verify_dev_entries(arg_struct_t *args)
{
  struct stat statbuf;

  /*
   * Stat the top level file and ensure it is a directory.
   * Then stat the device files in that and be sure they are
   * set with the appropriate major and minor numbers.
   */
  
  if (-1 == stat(args->dlmdevicename, &statbuf)) {
    /*
     * Maybe no problem, as it may not exist.
     */
    if (ENOENT == errno) {
      fprintf(stderr,
              "Lock directory does not exist [%s].  Create it.\n",
              args->dlmdevicename);
      if (-1 == mkdir(args->dlmdevicename,
                      S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH)) {
        fprintf(stderr,
                "Error [%s] creating lock directory.\n",
                strerror(errno));
        return(-1);
      }
    } else {
      fprintf(stderr, "Error on stat: %s\n", strerror(errno));
      return(-1);
    }
  } else {
    /*
     * "Something" exists in the expected slot.  What is it?
     */
    if (!S_ISDIR(statbuf.st_mode)) {
      fprintf(stderr,
              "Lock directory [%s] is not a directory, but [%x].\n",
              args->dlmdevicename,
              statbuf.st_mode);
      return(-1);
    }
    /*
     * MISSING: verify permissions, etc. yuck.
     */
  }

  /*
   * Top level either exists, or we created it.  Now need to verify
   * the actual admin and lock device entries.  Stat them, and then
   * verify or create as necessary.
   */
  if (-1 == verify_device(args->dlmdevicemajor,
                          args->dlmadminname,
                          args->dlmadminminor)) {
    return(-1);
  }
  if (-1 == verify_device(args->dlmdevicemajor,
                          args->dlmlocksname,
                          args->dlmlocksminor)) {
    return(-1);
  }
  
  return(0);
}

/* ============================================================ */
/*
 * Load the "device driver" for the haDLM device, if it isn't
 * already loaded.  This will not allow it to actually start
 * serving locks, as it will wait for the user side to
 * "initialize" it via serving cluster manager information.
 *
 * Once the device driver is loaded, open the "admin" port and
 * return the fd.  Return -1 if any problems.
 *
 * We basically let the module support handle this, then try
 * the open.
 *
 * The "name" of the module is the DLM device "base" name.
 */
int load_dev_driver(arg_struct_t *args)
{
  char command[256];
  char settings[128];
  int rc, filedes;

  /*
   * Try to open, driver may be there, OR, kernel may actually
   * load it for us on the open attempt.
   */
  filedes = open(args->dlmadminname, O_RDWR);

  if (-1 != filedes) {
    fprintf(stderr,
            "Opened [%s], filedes [%d]\n",
            args->dlmadminname,
            filedes);
    return(filedes);
  }

  strcpy(command, "modprobe -a -k ");
  strcat(command, args->dlmdevicebase);
  sprintf(settings,
          " haDLM_major_number=%d haDLM_name=%s haDLM_admin_minor=%d haDLM_locks_minor=%d",
          args->dlmdevicemajor,
          args->dlmdevicebase,
          args->dlmadminminor,
          args->dlmlocksminor);
  strcat(command, settings);
  fprintf(stderr,
          "Attempt to load DLM module, cmd [%s]\n",
          command);

  rc = system(command);

  if (0 != rc) {
    fprintf(stderr,
            "Error loading DLM module, rc [%d]\n",
            rc);
    return(-1);
  } else {
    fprintf(stderr,
            "Loaded DLM module!\n");
  }

  filedes = open(args->dlmadminname, O_RDWR);

  if (-1 == filedes) {
    fprintf(stderr,
            "Error opening [%s], errno [%s]\n",
            args->dlmadminname,
            strerror(errno));
    return(-1);
  } else {
    fprintf(stderr,
            "Opened [%s], filedes [%d]\n",
            args->dlmadminname,
            filedes);
  }

  return(filedes);
}

#ifdef FUNKY_TEST_TIME
/* ============================================================ */
/* Temp */
int main (int argc, char *argv[])
{
  verify_dev_entries(250,
                     "/dev/haDLM");
}
#endif /* FUNKY_TEST_TIME */
