/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: cccp.h,v 1.8 2001/11/02 22:16:32 idr Exp $"

/*
 *   COMPONENT_NAME: INCCLUSTER
 *
 *   FUNCTIONS: ASSERT
 *		acquire_lock
 *		cccp_listen
 */

#ifndef CCCP_H
#define CCCP_H

#include "dlm_list.h"

/* status codes*/
enum {
    CPERR_OK = 0,	/* no error*/
    CPERR_NOCCCP,       /* CCCP not yet initialized */
    CPERR_NODEDOWN,	/* destination went down without acknowledging*/
    CPERR_ILLEGAL,	/* illegal argument*/
    CPERR_NOSOCKET,	/* socket() failed*/
    CPERR_CNBIND,	/* bind() failed*/
    CPERR_NOMEM,	/* malloc() failed*/
    CPERR_QPKT,		/* could not queue packet for transmission*/
    CPERR_UNKNOWN	/* catchall for everything else (MUST BE LAST)*/
};

/* Valid port numbers are 1-255. */
#define CCCP_MAX_PORT    255

/* main structure for external interface*/

struct clm_msg;
typedef struct clm_msg clm_msg_t;

typedef int (cccp_callback_t)( clm_msg_t * );
typedef void (clm_msg_callback_t)( clm_msg_t *, int );
typedef void (cccp_forward_callback_t)( clm_msg_t *, unsigned, void * );

struct clm_msg {
    dlm_listnode_t  cm_link;
    unsigned	cm_port;
    unsigned	cm_prio;
    unsigned	cm_hops;
    unsigned	cm_node;

    ulong	cm_type;
    ulong	cm_length;
    clm_msg_callback_t * cm_callback;
    void	*cm_data;
    void	*cm_private;
    
    /* These fields are NOT to be modified or used in any way by clients of
     * CCCP!
     */
    
    unsigned     cm_sender;
    void       * cm_packet;
    void       * cm_node_ptr;
    cccp_forward_callback_t * cm_action[2];
};

/* external routines*/
extern int cccp_send( clm_msg_t * );
extern int cccp_listen( uint port, cccp_callback_t * cb );
extern clm_msg_t * cccp_msg_duplicate( const clm_msg_t * src );
extern clm_msg_t * cccp_msg_create( unsigned port, unsigned prio, unsigned node,
    unsigned type, unsigned length, clm_msg_callback_t * cb, void * data,
    void * private );
extern void cccp_msg_destroy( clm_msg_t * msg );

#if defined(DEBUG)

#ifdef __KERNEL__
#include "dlm_kernel.h"

#ifdef ASSERT

#define as_str(a) as_str_b(a)
#define as_str_b(a) #a
#define ASSERT(cond,msg)					\
    if (cond) {							\
	;							\
    } else {							\
	bsdlog(LOG_CRIT,"ASSERT FAILED: %s, file %s, line %d\n",	\
	   (msg), __FILE__, __LINE__);				\
	abort("ASSERT: " __FILE__ ", line " as_str(__LINE__));	\
    }

#endif

#else /* i.e. !defined(_KERNEL)*/

#ifdef ASSERT
#undef ASSERT
#endif
#define ASSERT(cond,msg)					\
    if (cond) {							\
	;							\
    } else {							\
	fflush(stdout);						\
	fprintf(stderr,"ASSERT FAILED: %s, file %s, line %d\n",	\
	   (msg), __FILE__, __LINE__);				\
	fflush(stderr);						\
	abort();						\
    }



#endif /* __KERNEL__ */

#else /* i.e. !defined(DEBUG)*/

#ifdef ASSERT
#undef ASSERT
#endif
#define ASSERT(cond,msg)

#endif


#ifdef __KERNEL__
#include <asm/semaphore.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#else
#define acquire_lock(a)	printf ("ACQUIRE_LOCK ( NOOP )\n");
#define release_lock(a)	printf ("RELEASE_LOCK ( NOOP )\n");
#endif

#define CPTRC_BASE		0x90000000

#define CPTRC_UINIT_ENTRY	(CPTRC_BASE + 0x00)
#define CPTRC_UINIT_EXIT	(CPTRC_BASE + 0x01)
#define CPTRC_USTOP_ENTRY	(CPTRC_BASE + 0x02)
#define CPTRC_USTOP_EXIT	(CPTRC_BASE + 0x03)
#define CPTRC_KINIT_ENTRY	(CPTRC_BASE + 0x04)
#define CPTRC_KINIT_EXIT	(CPTRC_BASE + 0x05)
#define CPTRC_KSTOP_ENTRY	(CPTRC_BASE + 0x06)
#define CPTRC_KSTOP_EXIT	(CPTRC_BASE + 0x07)

#define CPTRC_SEND_ENTRY	(CPTRC_BASE + 0x10)
#define CPTRC_SEND_EXIT		(CPTRC_BASE + 0x11)
#define CPTRC_LISTEN_ENTRY	(CPTRC_BASE + 0x12)
#define CPTRC_LISTEN_EXIT	(CPTRC_BASE + 0x13)
#define CPTRC_SEND_ENTRY_LK	(CPTRC_BASE + 0x14)

#define CPTRC_SNDCB_INVOKE	(CPTRC_BASE + 0x20)
#define CPTRC_SNDCB_RETURN	(CPTRC_BASE + 0x21)
#define CPTRC_RCVCB_INVOKE	(CPTRC_BASE + 0x22)
#define CPTRC_RCVCB_RETURN	(CPTRC_BASE + 0x23)
#define CPTRC_FORWARDED		(CPTRC_BASE + 0x24)

#define CPTRC_TRANSMIT		(CPTRC_BASE + 0x30)
#define CPTRC_GOTREDIR		(CPTRC_BASE + 0x31)
#define CPTRC_RETRY		(CPTRC_BASE + 0x32)
#define CPTRC_ENQUEUE		(CPTRC_BASE + 0x33)
#define CPTRC_DEQUEUE		(CPTRC_BASE + 0x34)
#define CPTRC_GOTACKS		(CPTRC_BASE + 0x35)

#define CPTRC_WRITE		(CPTRC_BASE + 0x40)
#define CPTRC_READ		(CPTRC_BASE + 0x41)
#define CPTRC_DUPLICATE		(CPTRC_BASE + 0x42)

#define CPTRC_POLL_ENTRY	(CPTRC_BASE + 0x50)
#define CPTRC_POLL_PROC		(CPTRC_BASE + 0x51)
#define CPTRC_POLL_ERROR	(CPTRC_BASE + 0x52)
#define CPTRC_POLL_DONE		(CPTRC_BASE + 0x53)

#define CPTRC_ACQUIRE_LOCK	(CPTRC_BASE + 0x60)

#endif /* defined CCCP_H */
