/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm.h,v 1.6 2001/08/31 23:11:35 wombat Exp $"

/*
 *   COMPONENT_NAME: INCCLUSTER
 *
 *   FUNCTIONS: Client API include file.
 */

#ifndef DLM_H
#define DLM_H

#include "dlm_cluster.h"

/*
 * Some handy global limits.
 */
#define MAXRESOURCELEN	255		/* max length of resource name */
#define DLM_MAXNODES    MAXNODES        /* max number nodes in cluster */

/*
 * status codes
 */
typedef enum dlm_stats {
    DLM_NORMAL,               /* request in progress */
    DLM_GRANTED,              /* request granted */
    DLM_DENIED,               /* request denied */
    DLM_DENIED_NOLOCKS,       /* request denied, out of system resources */
    DLM_WORKING,              /* async request in progress */
    DLM_BLOCKED,              /* lock request blocked */
    DLM_DENIED_GRACE_PERIOD,  /* topological change in progress */
    DLM_SYSERR,               /* system error */
    DLM_NOSUPPORT,            /* unsupported */
    DLM_CANCELGRANT,          /* can't cancel convert: already granted */
    DLM_IVLOCKID,             /* bad lockid */
    DLM_SYNC,                 /* synchronous request granted */
    DLM_BADTYPE,              /* bad resource type */
    DLM_BADRESOURCE,          /* bad resource handle */
    DLM_MAXHANDLES,           /* no more resource handles */
    DLM_NOCLINFO,             /* can't contact cluster manager */   
    DLM_NOLOCKMGR,            /* can't contact lock manager */
    DLM_NOPURGED,             /* can't contact purge daemon */
    DLM_BADARGS,              /* bad api args */
    DLM_VOID,                 /* no status */
    DLM_NOTQUEUED,            /* NOQUEUE was specified and request failed */
    DLM_IVBUFLEN,             /* invalid resource name length */
    DLM_CVTUNGRANT,           /* attempted to convert ungranted lock */
    DLM_BADPARAM,             /* invalid lock mode specified */
    DLM_VALNOTVALID,          /* value block has been invalidated */
    DLM_REJECTED,             /* request rejected, unrecognized client */
    DLM_ABORT,                /* blocked lock request cancelled */
    DLM_CANCEL,               /* conversion request cancelled */
    DLM_IVRESHANDLE,          /* invalid resource handle */
    DLM_DEADLOCK,             /* deadlock recovery refused this request */
    DLM_DENIED_NOASTS,        /* failed to allocate AST */
    DLM_FORWARD,              /* request must wait for primary's response */
    DLM_TIMEOUT,              /* timeout value for lock has expired */
    DLM_IVGROUPID,            /* invalid group specification */
    DLM_VERS_CONFLICT,        /* version conflicts prevent request handling */
    DLM_BAD_DEVICE_PATH,      /* Locks device does not exist or path wrong */
    DLM_NO_DEVICE_PERMISSION, /* Client has insufficient pers for device */
    DLM_NO_CONTROL_DEVICE,    /* Cannot set options on opened device */
    DLM_MAXSTATS              /* upper limit for return code validation */
} dlm_stats_t;

/*
 * resource types
 */
typedef enum dlm_restype {
	DLM_RES_UNIX,
	DLM_RES_VMS,
	DLM_RES_NTYPES
} dlm_restype_t;

/*
 * The following structure is used to get and set the global migration
 * parameters used to control resource migration.
 */
typedef struct dlm_globparams {
	unsigned	cg_valid;
	unsigned	cg_recalc_time;
	uint64_t	cg_decay_rate;
} dlm_globparams_t;

/*
 * The following structure is used to store resource statistics.
 */
typedef struct dlm_statistics {
	unsigned long   cs_requests;   /* number of lock requests        */
	unsigned long   cs_local;      /* number of local requests       */
	unsigned long   cs_remote;     /* number of remote requests      */
	unsigned long   cs_same;       /* number requests by same node   */
	unsigned long   cs_migrations; /* number of lock migrations      */
	unsigned long   cs_compat;     /* number compatible requests     */
	unsigned long   cs_incompat;   /* number incompatible requests   */
	unsigned long   cs_downgrade;  /* number down converts & unlocks */
	uint64_t        cs_total_aps;  /* total accesses-per-second      */
	uint64_t        cs_aps[DLM_MAXNODES]; /* accesses-per-second     */
} dlm_statistics_t;

/* The defines are the bit definitions for the cg_valid field */

#define CLMTUNE_GLOB_RECALC 0x01
#define CLMTUNE_GLOB_DECAY  0x02
#define	DECAY_MIN		0.0
#define	DECAY_MAX		1.0
#define KFL_DECAY_MIN	0
#define KFL_DECAY_MAX	0x3F800000

/*
 * The following structure is used to get and set the tuneable resource
 * parameters used to control resource migration.
 */
typedef struct {
	unsigned	cr_valid;
	unsigned	cr_stickiness;
} dlm_resparams_t;

/* The defines are the bit definitions for the cr_valid field */

#define	DLM_RES_STICKINESS	0x1
#define	MAX_STICKINESS		100


/*
 * The following section defines an internal representation for transaction
 * ids.  Essentially, it will always be passed and referenced as a pointer
 * to 8 bytes of memory.  How the client represents it does not matter.
 */
#define	XID_SIZE	8		/* 8 byte for XID (also deadlock id) */

typedef struct dlm_xid {
	union {
		char	xid_char[XID_SIZE];
		uint	xid_int[2];
	} xid_u;
} dlm_xid_t;

/*
 * We have an open requirement to support this as a 32 byte
 * field.  I believe that simply changing this define will
 * handle it (and a full rebuild of course, of the DLM and
 * of all clients.)  My thought though is to put an ifdef
 * here, and allow the installer to "decide".  A possible
 * issue with this would be mismatching DLM images on the
 * different nodes, which would be fatal and nasty to track
 * down.
 *
 * In any case, I'm going to leave it as is for now, and get
 * the bunch of other changes tested and checked in first.
 */
/* this must be 16 to be compatible with VMS */
#define MAXLOCKVAL	16		/* sizeof lock value block */

/*
 * lock status block returned by locking calls
 * VMS actually starts with
 *	unsigned short status;
 * 	unsigned short reserved;
 * we overlay this with a 32 bit status
 */
struct lockstatus {
	dlm_stats_t	status;	
	int 		lockid;
	char 		value[MAXLOCKVAL];	
	unsigned int	timeout;
};
	

/*
 * Resource Handles are 32 bit quantities
 *
 * this is used as a token in all lock transactions
 *
 * if any of these are modified from being unsigned to signed, you
 * will want to carefully check all uses to see if there are any
 * dependencies!
 */
union dlm_rh {
	unsigned long rh;
	struct {
		unsigned site:5;
		unsigned type:3;
		unsigned seg:8;
		unsigned cookie:16;
	} handle;
};

/*
 * flags and constants for locking requests
 */
/*
 * lock modes 
 */
#define LKM_IVMODE	(-1)		/* invalid mode */
#define LKM_NLMODE	0 		/* null lock */
#define	LKM_CRMODE	1 		/* concurrent read */
#define	LKM_CWMODE	2 		/* concurrent write */
#define	LKM_PRMODE	3 		/* protected read */
#define	LKM_PWMODE	4 		/* protected write */
#define	LKM_EXMODE	5		/* exclusive */
#define LKM_MAXMODE	5
#define LKM_MODEMASK	0xff

/*
 * lock flags 
 */
#define LKM_ORPHAN	0x10		/* this lock is orphanable */
#define LKM_PARENTABLE  0x20		/* this lock was orphaned */
#define LKM_BLOCK       0x40            /* blocking lock request */
#define LKM_LOCAL       0x80            /* local lock request */
#define	LKM_VALBLK	0x100		/* lock value block request */
#define LKM_NOQUEUE	0x200		/* non blocking request */
#define	LKM_CONVERT	0x400		/* conversion request */
#define	LKM_NODLCKWT	0x800		/* this lock wont deadlock */
#define	LKM_UNLOCK	0x1000		/* deallocate this lock */
#define LKM_CANCEL	0x2000		/* cancel conversion request */
#define LKM_DEQALL	0x4000		/* remove all locks held by proc */
#define LKM_INVVALBLK	0x8000		/* invalidate lock value block */
#define	LKM_SYNCSTS	0x10000		/* return synchronous status if poss */
#define	LKM_TIMEOUT	0x20000		/* lock request contains timeout */
#define LKM_SNGLDLCK	0x40000		/* request can self-deadlock */
#define LKM_FINDLOCAL   0x80000         /* find local lock request */
#define LKM_PROC_OWNED	0x100000       	/* owned by process, not group */
#define LKM_XID         0x200000        /* use transaction id for deadlock */
#define LKM_XID_CONFLICT 0x400000       /* do not allow lock inheritance */
#define LKM_FORCE	0x800000	/* force unlock flag */
#define LKM_FLAGMASK	0xffffff	/* all flags */
#define UNIX_FLAGMASK   ( LOCK_UN | LOCK_NB | LOCK_SH | LOCK_EX )
                                        /* all unix flags */

#define LKM_NOID		(-1)		/* nonexistent lock id */

/*
 * Specialized use of the lock-value block, a cluster-global
 * counter 'System Commit Number' (SCN).  Perform atomic
 * operations on an SCN.
 */
typedef enum scn_op {
	SCN_NULL=0,	/* Placeholder */
	SCN_CUR,	/* Return current value of SCN */
	SCN_INC,	/* Add 1 to and return new value for SCN */
	SCN_ADD,	/* Increment SCN by in_scnp */
	SCN_ADJ,	/* Adjust value to at least in_scnp */
	SCN_SET		/* Set the value of the SCN to in_scnp. */
} scn_op_t;

#ifdef __linux__
/*
 * Allow override of path to lock device using these environment
 * variables.  See api_glue.c.  If the installation uses defaults
 * for path names, then be sure to ignore these and do not use
 * these vars.
 */
#define OVERRIDE_DLM_NAME "OVERRIDE_DLM_NAME"
#define OVERRIDE_LOCKS_NAME "OVERRIDE_LOCKS_NAME"
#endif /* __linux__ */

/*
 * The actual usage of these, as described in the programmer's
 * guide, is reflected in the following.  Oddly enough, the
 * code in ASTpoll does not reflect this, as it always throws
 * all 3 arguments, which C fobs off by just dropping the
 * extraneous arguments at runtime for locking (B)ASTs.  Only
 * the unlock AST uses all three arguments.
 *
 * These are used in the function prototypes in the API.
 */
typedef void (dlm_astlockfunc_t)(void *);
typedef void (dlm_bastlockfunc_t)(void *, int);
typedef void (dlm_astunlockfunc_t)(void *, int, void *);

/*
 * Retrieve pending AST (normal or blocking) for lock requests.
 * Used only by user-space clients.
 */
int 
ASTpoll(int pid,
        int tid);

/*
 * Request a DLM lock asynchronously.
 * Used by both user-space and kernel-space clients.
 */
dlm_stats_t 
dlmlock(int mode,
        struct lockstatus *lksb,
        int flags,
        void *name,
        unsigned int namelen,
        dlm_astlockfunc_t * ast,
        void *astargs,
        dlm_bastlockfunc_t * bast );

/*
 * Request a DLM lock synchronously.
 * Used by only user-space clients.
 */
dlm_stats_t 
dlmlock_sync(int mode,
             struct lockstatus *lksb,
             int flags,
             void *name,
             unsigned int namelen,
             void *bastargs,
             dlm_bastlockfunc_t * bast);

/*
 * Request a DLM lock asynchronously and specify a transaction ID.
 * Used by both user-space and kernel-space clients.
 */
dlm_stats_t
dlmlockx(int mode,
         struct lockstatus *lksb,
         int flags,
         void *name,
         unsigned int namelen,
         dlm_astlockfunc_t * ast,
         void *astargs,
         dlm_bastlockfunc_t * bast,
         dlm_xid_t *xid);

/*
 * Request a DLM lock synchronously and specify a transaction ID.
 * Used by only user-space clients.
 */
dlm_stats_t
dlmlockx_sync(int mode,
              struct lockstatus *lksb,
              int flags,
              void *name,
              unsigned int namelen,
              void *bastargs,
              dlm_bastlockfunc_t * bast,
              dlm_xid_t *xid);

/*
 * Unlock a DLM lock asynchronously.
 * Used by both user-space and kernel-space clients.
 */
dlm_stats_t 
dlmunlock(int lockid,
          char *valueblock,
          int flags,
          dlm_astunlockfunc_t * unlockast,
          void *unlockastargs,
          void *extrap);

/*
 * Unlock a DLM lock synchronously.
 * Used by only user-space clients.
 */
dlm_stats_t 
dlmunlock_sync(int lockid,
               char *valueblock,
               int flags);

/*
 * Specify to the DLM which signal should be used to notify the
 * client when there are pending AST blocks.
 * Used by only user-space clients.
 */
dlm_stats_t 
dlm_setnotify(int newsig,
              int *oldsigp);

/*
 * Register a region for later UNIX lock usage via dlmregionlock.
 * Used by only user-space clients.
 */
union dlm_rh 
dlmregister(char *name);

/*
 * Request or release a UNIX lock on a region created with dlmregister.
 * Used by only user-space clients.
 */
dlm_stats_t 
dlmregionlock(union dlm_rh rh, 
              unsigned long offset,
              unsigned long len, 
              unsigned long flags);

/*
 * Return a null-terminated string describing the given DLM status code.
 * Used by both user-space and kernel-space clients.
 */
char *
dlm_errmsg( dlm_stats_t err );

/*
 * Print an error message appropriate for the current value in
 * dlm_errno and append the given null-terminated string to it.
 * Used by only user-space clients.
 */
void 
dlm_perror(const char *s);

/*
 * Manipulate a cluster-global counter (System Commit Number).
 * You can retrieve it, increment it, set it, or test and set it.
 * Used by both user-space and kernel-space clients.
 */
dlm_stats_t
dlm_scnop(int       lockid,             /* lockid to use */
          scn_op_t  op_type,            /* scn operation type */
          short     bit_len,            /* bit length of SCN */
          char     *in_lvb,             /* input LVB */
          char     *out_lvb);           /* output LVB */

/*
 * Attach the client process to an existing lock group.
 * Used by both user-space and kernel-space clients.
 */
dlm_stats_t
dlm_grp_attach(int gid,
               int flags);

/*
 * Create a lock group and attach the creating client to the group.
 * Used by both user-space and kernel-space clients.
 */
dlm_stats_t
dlm_grp_create(int *gid,
               int flags);

/*
 * Detach the client from a lock group.  The group will be destroyed
 * and locks belonging to it purged.
 * Used by both user-space and kernel-space clients.
 */
dlm_stats_t
dlm_grp_detach(int flags);

/*
 * Purge all locks for the calling client (pid != 0) or for the
 * specified node (pid == 0).
 * Used by both user-space and kernel-space clients.
 */
dlm_stats_t
dlm_purge(int site,
          int pid,
          int flags);

/*
 * Set the migration parameters for the specified resource,
 * i.e., prevent or allow the DLM to migrate the master location
 * for the resource ("lock stickiness").
 * Used by both user-space and kernel-space clients.
 */
dlm_stats_t
dlm_setresparams(char		*resname,	/* resource name */
                 short		namelen,	/* resource name length */
                 short		type,		/* resource type */
                 dlm_resparams_t	*params);	/* res migration parameters */

/*
 * Retrieve the stickiness setting for the specified lock.
 * Used by both user-space and kernel-space clients.
 */
dlm_stats_t
dlm_getresparams(char		*resname,	/* resource name */
                 short		namelen,	/* resource name length */
                 short		type,		/* resource type */
                 dlm_resparams_t	*params);	/* res migration parameters */

#ifdef _THREAD_SAFE
/*
 * Per thread dlm_errno is provided by the lock manager. Both the extern int
 * and the per thread value must be maintained.
 */
#define dlm_errno  (_Dlm_Errno())
#endif /* _THREAD_SAFE */

extern int dlm_errno;


#endif /* DLM_H */
