/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_cluster.h,v 1.1 2001/03/12 19:50:47 wombat Exp $"

/*
 *   COMPONENT_NAME: DLM
 *
 *   FUNCTIONS: INDEX_NOT_VALID
 *		INDEX_TO_NAME
 *		NAME_TO_INDEX
 *		defined
 */

#ifndef CLUSTER_H
#define CLUSTER_H

#ifndef __KERNEL__
#include <sys/types.h>
#include <sys/param.h>
#include <netinet/in.h>
#else
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/limits.h>
#endif /* __KERNEL__ */

#include <asm/param.h>

/*
 *   This is the include file for the Cluster wide commands. Utilities that
 *   need to access the cluster configuration and node configuration
 *   information should also include this file.
 *   All the data structures (and values) used to define nodes and interfaces
 *   are defined here.
 */

/*
 * The version of the cluster must be increased for every release that
 * changes the configuration or adds functionality that needs to be
 * suppressed until all the nodes in the cluster have been migrated
 * to the current version.
 *
 *     Version number       Release Name
 *            0.1            pre-release
 *
 */

#define CLUSTERVERSION 1

/* temporary definition */
#define MAXNODES	(8)
#define MAXCLSTRIFS (128)             /* Max. comm interfaces in cluster */
#define MAX_NODE_ID     (4095)       /* 2 bytes unsigned number */
#define MIN_NODE_ID     (1)
#define MAX_NETWORK_ID  (2147483647) /* 4 bytes signed number*/
#define MIN_NETWORK_ID  (0)

/*
 * Node number identifiers must be in this range, or the nodes
 * will be ignored.
 */
#define DLM_MINNODE_NUMBER 1
#define DLM_MAXNODE_NUMBER 4095

/*
 * This is the maximum length for all the names, like: adapter name,
 * network name, etc.
 */

#ifndef MAXDLMNAMELEN
#define MAXDLMNAMELEN 	(32)
#endif /* MAXDLMNAMELEN */

/*
 * The maximum length of a string of participating node names
 */
#ifndef MAX_PARTNODES_NAMELEN
#define MAX_PARTNODES_NAMELEN (MAXNODES*MAXDLMNAMELEN)
#endif /* MAX_PARTNODES_NAMELEN */

/*
 * MAXPATHLEN defines the longest permissable path length after expanding
 * symbolic links.
 */

#if !defined(MAXPATHLEN)
#define MAXPATHLEN      PATH_MAX
#endif

/*
 * This is the maximum length for a host name.
 */

#if !defined(MAXHOSTNAMELEN)
#define MAXHOSTNAMELEN  (32)
#endif 

/*
 * This is used in a number of different places as a generic buffer size.
 */

#if !defined(MAXLINELEN)
#define MAXLINELEN      (1024)
#endif

/*
 * This is the maximum length of the hardware address that can be
 * associated with an adapter.
 */

#ifndef MAXHARDADDRLEN
#define MAXHARDADDRLEN	(18)
#endif /* MAXHARDADDRLEN */

/*
 * This is the maximum length of the device file path that can be
 * associated with an adapter.
 */

#ifndef MAXDEVNAME
#define MAXDEVNAME  (32)
#endif /* MAXDEVNAME */

/*
 * Time string for dump resources command.
 */

#ifndef DUMPTIMELEN
#define DUMPTIMELEN 32
#endif /* DUMPTIMELEN */

/*
 *  Used by the dbprint module in libclstr.a.
 */

#define DBPRINT_SYSLOG        0x01
#define DBPRINT_SYSLOG_ALL    0x02
#define DBPRINT_FILE          0x04

/*
 * If an interface is PRIVATE, this means its a point to point connection
 * that can't be advertised to clients.  BOOT means it was listed as a
 * boot address in the config file.
 */

#define IFS_PRIVATE     0x01
#define IFS_PUBLIC      0x02
#define IFS_SERVICE     0x04
#define IFS_STANDBY     0x08
#define IFS_BOOT        0x10
#define IFS_SERIAL      0x20
#define IFS_SHARED      0x40

/*
 * This is the structure that describes a network interface.  These are
 * usually found in sets of doubly linked lists, each list pertaining
 * to a particular network on a machine.  There are three of these lists
 * associated with each network - one each for service, standby and boot
 * interfaces.  The head of the service interface list is the
 * service interface for that network.   When a service interface fails
 * the first alive standby interface is used to replace it. The standby
 * interface list is ordered so that dead standbys are always put at the end
 * of the list.
 */

struct IFSnode {

  /*
   * Pointer to the next IFSnode in the list.  If this is NULL,
   * then we are at the end of the list.
   */
    struct IFSnode *ifs_next;       /* ptr to next IFSnode, NULL if end */
    char ifs_name[MAXDLMNAMELEN];      /* name of interface */
    int ifs_addrtype;               /* union flag (0 for IP address) */
    char ifs_addr[MAXDEVNAME];      /* address to send or device file */
    char ifs_haddr[MAXHARDADDRLEN]; /* hardware address */
    char ifs_interfacename[MAXDLMNAMELEN]; /* hacmp/es use only */
    int ifs_network;                /* network */
    char ifs_netname[MAXDLMNAMELEN];   /* name of network */
    char ifs_type[MAXDLMNAMELEN];      /* physical type */
    char ifs_globalname[MAXDLMNAMELEN]; /* hacmp/es use only */
    unsigned char ifs_attr;          /* special flags */
    void *ifs_misc;                 /* user defined data */
  /*
   * Pointer to the MDnode this IFSnode structure is associated with.
   */

    struct MDnode *ifs_orig;

  /*  
   * The address to send to/from.  
   * ONLY HERE SO BUILDS WILL WORK
   */

    struct sockaddr_in ifs_saddr;

    /* more junk so we can build */
    struct IFSnode *ifs_prev;
    int ifs_index;
    int ifs_ecount;
    int ifs_network_id;
  
};


/*
 * This is the struct that defines a machine or CPU in the cluster.  In
 * particular it contains the node name and a list of service, standby and
 * boot interfaces.
 */
struct MDnode  {

    /*
     * Node name, unique within the cluster.
     */

    char md_name[MAXDLMNAMELEN];

    /*
     * How many networks there are. This also determines how many "cells"
     * we will allocate for md_svc_ifs, md_stby_ifs, and md_boot_ifs.
     */

    int md_nnets;

    /*
     * These are pointers to the (shared) service address(es).
     * When a node joins, this list is queried to make a new service interface.
     */

    struct IFSnode **md_svc_ifs;

    /*
     * Number of service address(es).
     */

    int md_nsvc;

    /*
     * These are pointers to the standby address(es).
     * On failure of a interface, if the head of this list contains a live
     * standby adapter it becomes the new service interface.
     */

    struct IFSnode **md_stby_ifs;

    /*  
     * Number of standby address(es).
     */

    int md_nstby;

    /*
     * For sites that use our IP takeover reintegration scheme, they will
     * need this.  When a node first comes up, its service adapters come up
     * on the boot addresses, since it is possible that their service
     * addresses may have been taken over by another system's adapter as
     * the result of a previous failover.  In the service address is not in
     * use, it then switches from the boot to the service address.
     */

    struct IFSnode **md_boot_ifs;

    /* !!! ONLY INCLUDED SO BUILDS WILL WORK !!! */
    int md_last_ifs[1];
    u_long md_last_ifs_addr[1];

    /*  
     * Number of boot address(es).
     */

    int md_nboot;

    /*
     * Pointer to user defined data, e.g. node state.
     */

    void *md_misc;

    /*
     * Node ID, unique within the cluster.
     */
    
    int md_node_id;
    

    /*
     * Major and Minor versions of cluster manager.
     */
    
    ushort md_majorversion;
    
    ushort md_minorversion;
    
    /*
     * Node handle, unique within the cluster.
     */

    int md_node_handle;

};

/*
 * This is the struct that defines a cluster information data on a node.
 */
struct CDnode  {
	
	/*
	 * The name of the cluster.
	 */

	char 	cd_name[MAXDLMNAMELEN];

	/*
	 * Cluster id.
	 */

	int	cd_id;

	/*
	 * The name of the local node.
  	 */

	char 	cd_nodename[MAXDLMNAMELEN];

        /*
         * The security level of the cluster.
         */

        char    cd_seclevel[MAXDLMNAMELEN];

	/*
	 * Number of configured nodes in this cluster.
	 */

	int 	cd_nnodes;

	/*
	 * Pointer to an array of configured nodes.
	 */

	struct 	MDnode *cd_md;

 	/*
	 * Number of total configured addresses in the entire cluster.
	 */

	int 	cd_nifs;

        /*
	 * Memory size this CDnode + MDnode + IFSnode takes.
	 */

	int	cd_size;

	/*
	 * Pointer to user defined data.
	 */

	void 	*cd_misc;

	/* 
	 * Node ID of the local node.
	 *
	 * This should be the last field in order to maintain compatibility
	 * with previous versions of the cluster manager.
	 */
	
	int cd_node_id;
        int cd_highest_node_id;
        int cd_highest_network_id; 
        int cd_handle;
        int cd_cluster_version;
};

 
/*
 * Node Name <--> index conversion macros.  
 */
#define INDEX_TO_NAME(cl_index, cl_cdnode)\
                     (index_to_name(cl_index, cl_cdnode))

#define NAME_TO_INDEX(cl_name, cl_cdnode, cl_i)\
                     (cl_i=name_to_index(cl_name, cl_cdnode))

/*
 * Node index not valid
 */
#define INDEX_NOT_VALID(cl_index, cl_max)\
	(((cl_index) < 0 || (cl_index) >= (cl_max)) ? 1:0)


/*
 * Cluster Event definition. This is referenced in clstrmgr, clsmuxpd, and
 * clverify utility.
 */

/*
 * The CM uses events to describe system changes (as well as a few CM
 * changes).  The events are called Topology Events (TE's).  Some of
 * these events are used internally to control the Configure mechanism.
 * All of them are used in communcating with RPC and AIX message queue
 * clients.  This enum contains all of the different type of events
 * that occur.
 *
 * IF YOU CHANGE ANYTHING IN THIS ENUM, YOU SHOULD MAKE THE
 * APPROPRIATE CHANGES TO CM_event_strings (in cldata.c).
 */

enum te_type
{
    /*
     * This is not an event.  It is used as an event 'wildcard' in calls
     * to 'fail_event_in_event_list ()'.
     */
    ANY_FAIL_EVENT = -1,

    /*
     * This is for the WAIT event; RPC generated.  Basically stalls out 
     * the cluster manager for a short time while other stuff happens
     * in the HACMP/ES world.
     */
    TE_WAIT = 0,

    /*
     * Interface events; a service interface fails and new interface
     * takes over.
     */
    TE_SWAP_ADAPTER,

    /*
     * A user triggered swap_adapter event for the PCI hot plug feature
     *
     */

    TE_SWAP_ADDRESS,


    /*
     * A new node comes online in the cluster.  This event
     * causes an associated action to be initiated (eg. the running of
     * a script by the event action dispatcher).
     */
    TE_JOIN_NODE,

    /*
     *
     * The network and/or interfaces are fixed after a failure has occurred.
     */
    TE_JOIN_NETWORK,

    /* This event is generated when a standby adapter comes back to
     * life after having died.  For example, if you disconnect a
     * standby from the network and then reconnect it, this event
     * will occur.
     */
    TE_JOIN_STANDBY,

    /*
     * All the networks in a single node fail, causing the node to 
     * be failed.  This is the way the Cluster Manager perceives a node
     * failure.
     */
    TE_FAIL_NODE,

    /*
     * Network Events;  All the interfaces for a machine (or all
     * machines) on a particular network fail or the network fails.
     */
    TE_FAIL_NETWORK,

    /*
     * This event is generated if an adapter failure occurs that does
     * not cause an adapter swap.  For example, if a standby adapter
     * fails while the primary adapter is still working, this
     * event will be generated.
     */
    TE_FAIL_STANDBY,

    /*
     * These are the 'complete' events which are run after the corresponding
     * event above finishes executing.
     */
    TE_JOIN_NETWORK_COMPLETE,
    TE_FAIL_NETWORK_COMPLETE,
    TE_WAIT_COMPLETE,
    TE_SWAP_ADAPTER_COMPLETE,
    TE_SWAP_ADDRESS_COMPLETE,
    TE_JOIN_NODE_COMPLETE,
    TE_FAIL_NODE_COMPLETE,

    /*
     * This event is generated if the cluster has been in
     * configuration for too long.
     */
    TE_CONFIG_TOO_LONG,

    /*
     * This event is generated if the cluster has been unstable for
     * too long.
     */
    TE_UNSTABLE_TOO_LONG,

    /*
     * This event is generated when an event-associated error occurs
     * (such as a script failure).
     */
    TE_EVENT_ERROR,

    /*
     * New primary was elected.  This happens if the old primary fails
     * and when the cluster first starts up.
     */
    TE_NEW_PRIMARY,

    /*
     * The sending clstrmgr has gone into a stable/unstable state.  Over
     * time these come in UNSTABLE->STABLE pairs.
     */
    TE_UNSTABLE,
    TE_STABLE,

    /*
     * These signify the beginning and end of the Configure Procedure.
     * Note that since its not for the individual configure methods, when
     * a CONFIG_COMPLETE is sent, no TE events exist on the clstrmgr's
     * internal queue; they have all been resolved by the methods.
     */
    TE_CONFIG_START,
    TE_CONFIG_COMPLETE,

    /*
     * Generated on address changes reported by jims.
     */
    TE_FAIL_ADAPTER,
    TE_JOIN_ADAPTER,

    /*
     * These are the dynamic reconfig topology events.
     */
    TE_DARE_TOPOLOGY,
    TE_DARE_TOPOLOGY_START,
    TE_DARE_TOPOLOGY_COMPLETE,
    TE_DARE_RESOURCE,
    TE_DARE_RESOURCE_RELEASE,
    TE_DARE_RESOURCE_ACQUIRE,
    TE_DARE_RESOURCE_COMPLETE,

    /*
     * Don't put anything after this entry.  It is just here so we have
     * a way of telling how many TE events have been defined.
     */
    TE_NUMBER_OF_EVENTS
};

#endif /* !defined(CLUSTER_H) */

