/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: dlm_devmsg.h,v 1.6 2001/10/19 22:43:38 idr Exp $"

/*
 * Distributed Lock Manager for Linux - messages
 *
 * These are the message definitions for messages that go
 * across the "device" interface, these are not the internode
 * messages that are passed among lock manager instances.
 *
 * These represent:
 * - cluster status updates, node up/down, IP address info.
 */

#ifndef DLM_DEVMSG_H
#define DLM_DEVMSG_H

#ifndef __KERNEL__
#include <pthread.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#else
#include <linux/in.h>
#include <linux/ioctl.h>
#endif /* __KERNEL__ */

#include "ast_stub.h"
#include "dlm_cluster.h"
#include "clm_tune.h"

/* ------------------------------------------------------------ */
/*
 * Used to hold the list of nodes as read from the configuration
 * file and also from the cluster manager notifications.
 *
 * If the cluster code does NOT report node number, then
 * the name field must be set to -1!  If it does not report
 * node name, then the name field must be set to null in the
 * first character.  If it does not report IP address, then
 * the comm_state field should be set to NULL and the address
 * field set to zero.
 */

#define MAX_NODENAME_LEN 32

typedef struct topology_address_element
{
    int nodeid;			/* For this Node ID	*/
    struct in_addr useaddr; /* Use this IP address	*/
    ushort dlm_major;		/* Major Version of the CLLOCKD	*/
    ushort dlm_minor;		/* Minor Version of the CLLOCKD	*/
} MQ_DLM_USEADDR_t;

#define TAM_version 1	/* Topology Address Message Version */

typedef struct topology_address_message
{
  int msg_version ;           /* message layout version */
  int n_nodes;                /* # nodes valid in addrs */
  int this_nodeid ;           /* cluster global id for THIS node */
  int event ;                 /* cluster manager global event # */
  MQ_DLM_USEADDR_t addrs[1];  /* dynamically allocate enough of these */
} MQ_DLM_TOP_INFO_t;

typedef enum {
  haDLM_state_null = 0,                 /* Field not used in block */
  haDLM_state_down = 1,
  haDLM_state_up = 2,
  haDLM_state_local = 4                 /* This is the local node */
} haDLM_state_t;

typedef struct dlm_node {
  haDLM_state_t dlm_node_state;         /* Node up, down or other */
  int   dlm_node_number;                /* Integral node identifier */
  char  dlm_node_name[MAX_NODENAME_LEN]; /* Name as given */
  haDLM_state_t dlm_comm_state;         /* null if not given, or up/down */
  struct in_addr dlm_node_address;      /* Primary IP address to use */
} dlm_node_t;

typedef struct dlm_node_array {
  int   dlm_node_count;                 /* How many nodes */
  dlm_node_t dlm_node_entries[1];       /* One per node */
} dlm_node_array_t;

/* ------------------------------------------------------------ */
/*
 * Command buffers, used to wrap the information passed to the
 * kernel side.
 *
 * NOTE: more of these to come, watch this space.
 */

typedef enum {
  haDLM_MSG_First=0,
  haDLM_MSG_Init=0,                     /* Initialize the kernel side */
  haDLM_MSG_Stop,                       /* Tell kernel side to stop */
  haDLM_MSG_CMGR,                       /* Cluster update (figure out later) */
  haDLM_MSG_Last                        /* Always last one */
} dlm_MSG_types_t;

typedef struct dlm_MSG {
  dlm_MSG_types_t msg_code;   /* What kind of info here? */
  int   msg_ptr_size;         /* Size of block below pointed to. */
  union {
    void               *msg_arg_block;
    dlm_node_array_t   *msg_node_array;
    MQ_DLM_TOP_INFO_t  *msg_topo_array;
  } _msg_data;
} dlm_MSG_t;

#define arg_block _msg_data.msg_arg_block
#define node_array _msg_data.msg_node_array
#define topo_array _msg_data.msg_topo_array

/* ============================================================ */
/*
 * IOCTL calls used to pass data into the kernel and retrieve
 * various data items.  These are used where we essentially
 * have a write/read relationship, i.e., user side (the API)
 * sends in address, kernel writes out a data block.  We do
 * not simply use a read on the device because read only
 * allows us to differentiate via size of block, and does not
 * tell us exactly what is being read.
 *
 * We could use a write here, and write in a block that tells
 * the kernel what is desired, and the kernel then puts that
 * data out to the user, but, this is not semantically clean,
 * as the user is "writing to read".
 *
 * This leads us to IOCTLs, since we can provide a specific
 * one for each needed data item, and these can imply that
 * we send data both directions.
 */
#define DLM_IOCTL_BASE 'd'

/*
 * Respectively, these do:
 * ASTGET: fetch and execute all pending ASTs for the client.
 * DUMPRES: dump resources to the DLM dump file.
 * SETDEBUG: set the DLM debug trace flag.
 * RESCOUNT: return number of resources pending.
 * GROUPOP: issue a lock group operation, return group id.
 * SENDTRAN: send a lock request to the DLM
 * RECVTRAN: collect a response to a lock request from the DLM
 * GLOBSTAT: collect a global stats summary from the DLM
 *
 * All return 0 for success, -1 for error (and set errno
 * in the error cases.)
 */
#define DLM_IOCTL_ASTGET _IOR(DLM_IOCTL_BASE, 1, struct poll_info)

#define DLM_IOCTL_SETDEBUG _IOW(DLM_IOCTL_BASE, 2, int)

typedef struct groupop {
  int op;
  int *gid;
  int pid;
  int flags;
} groupop_t;

#define DLM_IOCTL_GROUPOP _IOW(DLM_IOCTL_BASE, 3, struct groupop)

typedef struct sendtran {
  void *buf;
  int   len;
} sendtran_t;

#define DLM_IOCTL_SENDTRAN _IOW(DLM_IOCTL_BASE, 4, struct sendtran)

typedef struct recvtran {
  void *buf;
  int   len;
  int   seq;
} recvtran_t;

#define DLM_IOCTL_RECVTRAN _IOW(DLM_IOCTL_BASE, 5, struct recvtran)

#define DLM_IOCTL_GLOBSTAT _IOR(DLM_IOCTL_BASE, 6, struct dlm_stats_buf)

/* ============================================================ */
/*
 * This message queue "stuff" represents the full suite of
 * data we eventually need to pass through the 'admin'
 * interface.  We put it here to get rid of msgq.h, and to
 * start cleaning things up.
 */
    
/*
 * Constants for describing the various types of messages that may be passed
 * over the message queues.
 */

/* added one more msg to get TEnode info from cluster manager. 6/29/94    */
/* the order of the enumerations cannot change in order to preserve       */
/* version compatibility. Add new values to the end.                      */

enum mq_msg_types {
  MQ_FIRST_MSG = 0,           /* Read first message from queue */
  MQ_MDLIST_LEN_MSG,          /* length of next mdlist msg     */
  MQ_MDLIST_MSG,              /* packed cdnode/mdlist          */
  MQ_CD_INFO_MSG,             /* Cluster State change info     */
  MQ_NODE_INFO_MSG,           /* node state change info        */
  MQ_IFS_INFO_MSG,            /* interface state change info   */
  MQ_DLM_IN_MSG,              /* msg to local clm from clm     */
  MQ_DLM_OUT_MSG,             /* msg from local clm to clm     */
  MQ_DLM_INFO_MSG,            /* clm state change info         */
  MQ_TENODE_INFO_MSG,         /* node&ifs state change info    */
  MQ_CLNODE_INFO_MSG,         /* CLnode info                   */
  MQ_EVENT_INFO_MSG,          /* global event number           */
  MQ_DLM_TOP_INFO_MSG,        /* dlm topology info msg         */
  MQ_MDLIST_CHUNK_MSG,        /* MDlist msg fragment           */
  MQ_DLM_MUST_STOP_MSG,       /* Admin side says DLM must stop */
  MQ_CLIENT_IS_DEAD_MSG,      /* A client has died, handle it  */
  MQ_MSG_TYPE_MAX             /* highest msg type defined      */
};

/*
 * Structure used to describe a Lock Manager's current state.
 */
/*
 * States which the Lock Manager can assume.
 */
enum CLM_STATES
{
    CLM_DEAD = 0,
    CLM_ALIVE
} CLM_STATES_t;

typedef struct clminfo_struct
{
    int nodeid;
    int major;
    int minor;
    enum CLM_STATES clm_state;
} MQ_CLM_INFO;

/*
 * Structure used to describe a node's current state.
 */

typedef struct nodeinfo_struct
{
    int nodeid;
    ulong node_state;
} MQ_NODE_INFO;

/*
 * Structure used to describe an interface's current state.
 */

typedef struct ifsinfo_struct
{
    int	nodeid;
    int	network;
    char ifs_name[MAXDLMNAMELEN];
    ulong ifs_state;
    int ifs_owner;
} MQ_IFS_INFO;


typedef struct clm_msg_struct
{
    int nodeid;
    int sync_id;
} MQ_CLM_MSG;

#if 0   /* Nuke for Linux */
/*
 * This structure defines the contents of the messages which are sent over 
 * the message queues.  Since we have no way of knowing how many nodes will
 * be configured in the cluster, we declare 'mq_mdlist' as 1 for now - we 
 * will allocate the memory for it later.
 */

/* added TEnode struct into the struct to pass node & ifs info. 6/29/94 */

struct mq_msg_struct { 
  long mq_type;
  int mq_localNodeId;
  union {
	int mq_mdlist_len;
	char mq_mdlist[1];
	int mq_event_number;
#if 0   /* Nuke for Linux */
	enum CD_STTES mq_cd_state;
#endif  /* Nuke for Linux */
	int mq_cd_state;
	MQ_NODE_INFO mq_node_state;
	MQ_IFS_INFO  mq_ifs_state;
	MQ_CLM_INFO  mq_clm_state;
	MQ_CLM_MSG  mq_clm_msg;
	MQ_CLM_TOP_INFO mq_clm_top;
#if 0   /* Nuke for Linux */
	struct TEnode mq_tenode;
#endif  /* Nuke for Linux */ 
    dlm_node_t mq_tenode;
#if 0   /* Nuke for Linux */
	struct CLnode mq_clnode; 
#endif  /* Nuke for Linux */
  } mq_union;
};
#endif  /* Nuke for Linux */

#ifndef __KERNEL__

/* ============================================================ */
/*
 * Used in conjunction with the above to pass information
 * between the main core of dlmdu and the dynamiclly
 * loaded cluster manager interface code (from dlmdcl).
 *
 * The dynamically-loaded routine handling callbacks needs
 * to call these:
 * - lock the notfication queue (begin)
 * - deliver node/interface notification blocks
 * - unlock the queue, which will trigger the main thread (end)
 *
 * The goal is to "batch" as many notifications as possible,
 * since the DLM code is able to deal with multiple things
 * going on - i.e., if two nodes fail, deal with both of them
 * at once, rather than through two separate recovery operations.
 * This is all assuming that the cluster manager is able to
 * provide this level of information.
 *
 * Note that the FIRST notification to be delivered should tell
 * the DLM the identity of the local node, i.e., the node on
 * which this instance is running.
 */

extern void dlm_notify_begin(void);
extern void dlm_notify_deliver(dlm_node_t *info);
extern void dlm_notify_end(void);
extern int  dlm_write_buffer(char *buffer);

/* ============================================================ */
/*
 * Allow us to directly build in the cluster interface code,
 * for testing and other such reasons.  Production intent is
 * to have it dynamically loaded based on input from the
 * configuration file.
 */

#ifdef DIRECT_CLUSTINT
void *cluster_interface_handling(void *config_nodes);
void *cluster_interface_shutdown(void);
#else /* DIRECT_CLUSTINT */
void *cluster_interface_handle;
#define CLUSTER_INTERFACE_START "cluster_interface_handling"
#define CLUSTER_INTERFACE_STOP "cluster_interface_shutdown"
void *(*cluster_interface_start)(void *);
void *(*cluster_interface_stop)(void);
#endif /* DIRECT_CLUSTINT */

#endif /* !__KERNEL__*/

#endif /* !DLM_DEVMSG_H */
