/*
 * Core modules for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */

#ifndef DLM_LIST_H
#define DLM_LIST_H

#ident "$Id: dlm_list.h,v 1.7 2001/07/25 21:23:50 idr Exp $"

#if defined( __linux__ ) && defined( __KERNEL__ ) 
# include <linux/stddef.h>
#endif

#ifndef BOOL_T_DEFINED
#define BOOL_T_DEFINED
typedef int bool_t;
#endif /* BOOL_T_DEFINED */


struct dlm_list;
typedef struct dlm_list dlm_list_t;
struct dlm_listnode;
typedef struct dlm_listnode dlm_listnode_t;

struct dlm_list
{
    dlm_listnode_t * head;
    dlm_listnode_t * tail;
    dlm_listnode_t * tailpred;
};

struct dlm_listnode
{
    dlm_listnode_t * succ;
    dlm_listnode_t * pred;
};

static __inline__ void dlm_list_init( dlm_list_t * list )
{
    list->head = (dlm_listnode_t *) & list->tail;
    list->tail = NULL;
    list->tailpred = (dlm_listnode_t *) & list->head;
}

static __inline__ bool_t dlm_list_is_empty( const dlm_list_t * list )
{
    return( list->tailpred == (dlm_listnode_t *) list );
}

static __inline__ void dlm_list_add_tail( dlm_list_t * list,
					  dlm_listnode_t * node )
{
    node->succ = (dlm_listnode_t *) & (list->tail);
    node->pred = list->tailpred;
    
    node->pred->succ = node;
    node->succ->pred = node;
}


static __inline__ void dlm_list_add_head( dlm_list_t * list,
					  dlm_listnode_t * node )
{
    node->succ = list->head;
    node->pred = (dlm_listnode_t *) list;
    
    node->pred->succ = node;
    node->succ->pred = node;
}

static __inline__ void dlm_list_insert( dlm_list_t * list,
				        dlm_listnode_t * node,
				        dlm_listnode_t * pred )
{
    if ( pred == NULL )
    {
	pred = (dlm_listnode_t *) list;
    }

    dlm_list_add_head( (dlm_list_t *) pred, node );
}

static __inline__ void dlm_list_remove( dlm_listnode_t * node )
{
    node->succ->pred = node->pred;
    node->pred->succ = node->succ;
    node->succ = NULL;
    node->pred = NULL;
}


static __inline__ dlm_listnode_t *  dlm_list_rem_head( dlm_list_t * list )
{
    dlm_listnode_t * node;
    
    
    node = NULL;
    if ( ! dlm_list_is_empty( list ) )
    {
	node = list->head;
	dlm_list_remove( node );
    }

    return( node );
}


static __inline__ dlm_listnode_t * dlm_list_get_head( dlm_list_t * list )
{
    return( dlm_list_is_empty( list ) ? NULL : list->head );
}

static __inline__ dlm_listnode_t * dlm_list_get_tail( dlm_list_t * list )
{
    return( dlm_list_is_empty( list ) ? NULL : list->tailpred );
}

static __inline__ dlm_listnode_t * dlm_list_succ( dlm_listnode_t * node )
{
    return( (node->succ->succ == NULL) ? NULL : node->succ );
}

static __inline__ dlm_listnode_t * dlm_list_pred( dlm_listnode_t * node )
{
    return( (node->pred->pred == NULL) ? NULL : node->pred );
}

static __inline__ void dlm_list_append( dlm_list_t * dest, dlm_list_t * src )
{
    dlm_listnode_t * new_succ;
    dlm_listnode_t * new_tail;

    new_succ = src->head;
    if ( new_succ != (dlm_listnode_t *) src )
    {
	new_tail = src->tailpred;
	
	new_tail->succ = (dlm_listnode_t *) & dest->tail;
	new_succ->pred = dest->tailpred;

	dest->tailpred->succ = new_succ;
	dest->tailpred = new_tail;

	dlm_list_init( src );
    }
}


static __inline__ void dlm_list_move( dlm_list_t * dest, dlm_list_t * src )
{
    dlm_listnode_t * new_succ;
    dlm_listnode_t * new_tail;

    new_succ = src->head;
    if ( new_succ != (dlm_listnode_t *) src )
    {
	dest->head = new_succ;
	new_succ->pred = (dlm_listnode_t *) dest;
	
	new_tail = src->tailpred;
	dest->tailpred = new_tail;
	new_tail->pred = (dlm_listnode_t *) & dest->tail;
    }
}

static __inline__ unsigned dlm_list_count( dlm_list_t * list )
{
    dlm_listnode_t * node;
    unsigned         count;


    count = 0;
    for ( node = list->head ; node->succ != NULL ; node = node->succ )
    {
	count++;
    }

    return( count );
}

#define EACH_DLM_LISTNODE(l,c) \
	    c = (typeof(c)) dlm_list_get_head( (l) ) \
	     ; (c) != NULL \
	     ; (c) = (typeof(c)) dlm_list_succ( (dlm_listnode_t *) (c) )

#define EACH_DLM_LISTNODE_COND(l,c,cond) \
	    c = (typeof(c)) dlm_list_get_head( (l) ) \
	     ; (c) != NULL && (cond) \
	     ; (c) = (typeof(c)) dlm_list_succ( (dlm_listnode_t *) (c) )

#define DECLARE_DLM_LIST(l) \
	    dlm_list_t l = { (dlm_listnode_t *) & l . tail, \
		    NULL, (dlm_listnode_t *) & l . head }

#endif /* DLM_LIST_H */
