/*
 * Interface library for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * http://www.gnu.org/copyleft/lesser.html 
 */
#ident "$Id: api_errno.c,v 1.3 2001/11/06 13:32:46 bkon Exp $"

/*
 *   COMPONENT_NAME: DLM LIB
 *
 *   FUNCTIONS: _Dlm_Errno
 *              init_dlm_errno_hdl
 *		set_dlm_errno
 */

#ifdef __KERNEL__

# ifndef MODULE
#  define MODULE
# endif

# include <linux/kernel.h>
# define __NO_VERSION__
# include <linux/module.h>
# ifdef USING_FULL_VERSIONING
#  if CONFIG_MODVERSIONS==1
#   define MODVERSIONS
#   include <linux/modversions.h>
#  endif
# endif /* USING_FULL_VERSIONING */
#endif /* __KERNEL__ */




#ifdef	dlm_errno
#undef	dlm_errno
#endif	/* dlm_errno */

#ifndef __KERNEL__
/* get rid of a lot of stuff for kernel api build */

/*
 * Note: dlm_errno must be defined as int, not dlm_stats_t,
 * because dlm.h can't be included in this file.  Can't include
 * dlm.h, since if _THREAD_SAFE,
 *it has #define dlm_errno (*_Dlm_Errno())
 */
#ifdef _THREAD_SAFE
/* global dlm_errno value */
int    dlm_errno;
#else
extern int      dlm_errno;
#endif

extern int errno;

#include <errno.h>
#include "clmutex_r.h"

#ifdef  _THREAD_SAFE
static pthread_mutex_t dlmerrno_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t dlm_errno_hdl_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_key_t dlm_errno_hdl;
static pthread_key_t dlm_errno_hdl_inited = 0;

/*
 * Function:
 *	_Dlm_Errno
 *
 * Return value:
 *	value of the per-thread dlm_errno
 *
 * Description:
 *	Allow access to a per-thread clm_errno. 
 */
int
_Dlm_Errno()
{
    int err;

    /* Make sure per-thread data handle is initialized */
    if (!dlm_errno_hdl_inited)
	init_dlm_errno_hdl();
    
    err = (int)pthread_getspecific(dlm_errno_hdl);

    /* If no per-thread clm_errno value */
    if (errno)
        err = dlm_errno;

    return err;
}

/*
 * Function:
 *	init_dlm_errno_hdl
 *
 * Return value:
 *	return value from pthread_key_create
 *
 * Description:
 *	create dlm_errno per-thread data key: dlm_errno_hdl
 */
int init_dlm_errno_hdl ()
{
    int status;

    /* Check if per-thread data handle is already initialized */
    TS_LOCK(&dlm_errno_hdl_mutex);
    if (dlm_errno_hdl_inited)
    {
	TS_UNLOCK(&dlm_errno_hdl_mutex);
	return 0;
    }

    dlm_errno = 0;

    /* Get per-thread handle for clm_errno */
    if (pthread_key_create(&dlm_errno_hdl, NULL) == 0)
	dlm_errno_hdl_inited++;
    
    TS_UNLOCK(&dlm_errno_hdl_mutex);

    return 0;
}
#endif	/* _THREAD_SAFE */
#endif /* __KERNEL__ */

/*
 * Function:
 *	set_dlm_errno
 *
 * Parameters:
 *	error	- value to set dlm_errno to
 *
 * Return value:
 *	new value of dlm_errno
 *
 * Description:
 *	Set the per-thread dlm_errno.
 */


int
set_dlm_errno(int error)
{
#ifndef __KERNEL__
#ifdef	_THREAD_SAFE
    /* Check if per-thread data handle is already initialized */
    if (!dlm_errno_hdl_inited)
	init_dlm_errno_hdl();

    /* Set per-thread dlm_errno */
    /* Don't need to lock, since just altering per-thread value */
    pthread_setspecific(dlm_errno_hdl, (void *)error);
#endif	/* _THREAD_SAFE */

    TS_LOCK (&dlmerrno_mutex);
    /* set global value */
    dlm_errno = error;
    TS_UNLOCK (&dlmerrno_mutex);
#endif /*__KERNEL__*/
    return (error);
}
