/*
 * Interface library for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * http://www.gnu.org/copyleft/lesser.html 
 */
#ident "$Id: api_glue.c,v 1.8 2001/11/06 13:32:46 bkon Exp $"

/*
 *   COMPONENT_NAME: DLM LIB
 *
 *   FUNCTIONS: ASTget
 *		clm_recv
 *		clm_send
 *		init_glue
 *		set_debug
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "dlm.h"
#include "dlm_devmsg.h"
#include "ast.h"
#include "ast_stub.h"
#include "clmstructs.h"
#include "clmutex_r.h"
#include "api_data.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "general_msg.h"
#define MSGSTR(Num, Str) catgets(get_catd("general.cat"), GENERAL, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

/* ============================================================ */

#ifdef  _THREAD_SAFE
/*
 * initglue_mutex protects our device file descriptor.
 *
 * This is sufficient, since Clm_Init is called before any
 * other lock manager call
 */
static pthread_mutex_t initglue_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

/*
 * We use the file descriptor being set to determine if the
 * "glue" layer is ready.  The mutex above protects this, if
 * multi-threaded.
 */
static int locks_fd = -1;

/*
 * Now, an interesting conundrum.  We need to know the PATH of
 * the lock device, however, in the interest of being very
 * flexible (too flexible?) we allow the names to be customised
 * by an installation.  E.g., by default the path would be:
 *  /dev/haDLM/locks
 * but, both the dlm "name" (haDLM) and the "locks" name (locks)
 * can be optionally customized.
 *
 * If these are modified, then the specific names will be in
 * the DLM configuration file and we could simply look it up
 * in there.  By default, this is in /etc, but, its location
 * could have also been overridden when the DLM was initialized,
 * so we would need in that case the config file location.
 * Sigh.
 *
 * Eventually we may want to put something in the API to allow
 * the clients to specify these customisations.  For now, we
 * adopt the "cheap" way and use environment variables to
 * override the defaults.  If the vars are not set, then use
 * the defaults.
 */

static char device_name[MAXDLMNAMELEN];
static char locks_name[MAXDLMNAMELEN];
static char lock_device[MAXDLMNAMELEN];

/* ============================================================ */

/*
 * Function:
 *  initted
 *
 * Description:
 * Return 1 if client is initialised, and 0 otherwise.
 *
 * This is to "hide" the actual implementation of what it means
 * to be initialised from most routines.
 */
static int
initted(void)
{
  return(-1 != locks_fd);
}

/*
 * Function:
 *	init_glue
 *
 * Description:
 * Open a device for this client through the /dev/haDLM/locks
 * interface to hook us up to the lock manager.  This does not
 * actually issue any lock calls, but all data will flow through
 * that opening.
 *
 * Note that currently there is no "exposure" of the resultant
 * file descriptor to the client code, as that didn't exist in
 * the AIX client lib.  This would be useful to expose as it
 * would allow us to provide an "interrupt" through the device
 * to allow the client to use poll() or select() against it to
 * wait for responses.
 *
 * But, if we do this, the client must NOT do direct reads or
 * writes against the descriptor!  It must call the relevant
 * lock api routines.
 *
 * Arguments:
 *	NONE
 *
 * Return value:
 *	DLM_NORMAL - device opened
 *  DLM_NOLOCKMGR - 
 *
 * Side effects:
 *	
 */
dlm_stats_t
init_glue (void)
{
  int rc;
  dlm_stats_t retval = DLM_NORMAL;
  
  TS_LOCK(&initglue_mutex);
  /*
   * Check if initialized again since calling functions
   * don't lock 
   */
  if (initted()) {
	TS_UNLOCK(&initglue_mutex);
	return(0);
  }

  /*
   * See if we are overriding device path names.  See comment
   * earlier on this issue.
   */
  if (getenv(OVERRIDE_DLM_NAME)) {
    if (MAXNAMLEN <= strlen(OVERRIDE_DLM_NAME)) {
#ifdef NOT_DEFINED
      fprintf(stderr,
              "Overriding DLM name too long len/val[%d/%s]\n",
              strlen(OVERRIDE_DLM_NAME),
              OVERRIDE_DLM_NAME);
#endif
      return(DLM_BAD_DEVICE_PATH);
    } else {
      strcpy(device_name, OVERRIDE_DLM_NAME);
    }
  } else {
    strcpy(device_name, DEFDLMDEVICEBASE);
  }

  if (getenv(OVERRIDE_LOCKS_NAME)) {
    if (MAXNAMLEN <= strlen(OVERRIDE_LOCKS_NAME)) {
#ifdef NOT_DEFINED
      fprintf(stderr,
              "Overriding DLM/LOCKS name too long len/val[%d/%s]\n",
              strlen(OVERRIDE_LOCKS_NAME),
              OVERRIDE_LOCKS_NAME);
#endif
      return(DLM_BAD_DEVICE_PATH);
    } else {
      strcpy(locks_name, OVERRIDE_LOCKS_NAME);
    }
  } else {
    strcpy(locks_name, DEFDLMLOCKSBASE);
  }
  sprintf(lock_device, "/dev/%s/%s", device_name, locks_name);

  /*
   * That out of the way, try to open the device.  Various errors
   * are permission related, name mismatch, and the kernel code
   * not being loaded.  However, we don't analyse it much, mostly
   * we just pass the error back, if we get one, to the caller
   * and let them take care of it.
   */
#ifdef NOT_DEFINED
  fprintf(stderr,
          "Attempt to open lock device [%s]\n",
          lock_device);
#endif

  rc = open(lock_device,
            O_RDWR);
  if (-1 == rc) {
#ifdef NOT_DEFINED
    fprintf(stderr,
            "Error opening lock device, [%d/%s]\n",
            errno,
            strerror(errno));
#endif
    switch (errno) {
    case EAGAIN:
      retval = DLM_NOLOCKMGR;
      break;
    case ENXIO:
    case ENODEV:
    case ENOENT:
    case ENOTDIR:
      retval = DLM_BAD_DEVICE_PATH;
      break;
    case EACCES:
      retval = DLM_NO_DEVICE_PERMISSION;
      break;
    default:
      retval = DLM_SYSERR;
    }
  } else {
    /*
     * Device open, attempt to set some control flags to
     * more carefully control the device.  For now, the only
     * caution is to ensure the device is closed if the
     * client should exec.
     */
    int flags;
    locks_fd = rc;                      /* Save file desc. */

    flags = fcntl(locks_fd, F_GETFD, 0);
    if (flags >= 0) {     
        flags |= FD_CLOEXEC;
        rc = fcntl(locks_fd, F_SETFD, flags);
        if (-1 == rc) {
           fprintf(stderr,
                   "Cannot set close on exec, err [%d/%s]\n",
                   errno,
                   strerror(errno));
           close(locks_fd);                  /* No use to us! */
           retval = DLM_NO_CONTROL_DEVICE;
           locks_fd = -1;
         }
        
    }else{
      fprintf(stderr,
              "Cannot get flags to set close on exec, err [%d/%s]\n",
              errno,
              strerror(errno));
      close(locks_fd);                  /* No use to us! */
      retval = DLM_NO_CONTROL_DEVICE;
      locks_fd = -1;
    }

  }

  /*
   * Device is open, flags set, we are ready to go.  We have
   * opened a path to the DLM, but, we have not yet provided
   * it any "identification."  That is done by the higher
   * level code (see Dlm_Init() actually) as we just handle
   * the plumbing here.
   */

  TS_UNLOCK(&initglue_mutex);
  return(retval);
}


/*
 * These functions act as wrappers to set up for the various
 * ioctl calls to send/retrieve data related to lock requests,
 * AST blocks, group operations, and various "service" functions.
 */

/*
 * Function:
 *	clm_send
 * Description:
 *	Wrapper for clm_send routine in kernel.
 * Arguments:
 *	buf - transaction pointer
 *	len - byte length of transaction
 * Return value:
 *	Return code from kernel
 * Side effects:
 *	NONE
 */
int
clm_send (void *buf, int len)
{
  sendtran_t send;

  send.buf = buf;
  send.len = len;

  return(ioctl(locks_fd, DLM_IOCTL_SENDTRAN, &send));  
}

/*
 * Function:
 *	clm_recv
 * Description:
 *	Wrapper for clm_recv routine in kernel.
 * Arguments:
 *	buf - transaction pointer
 *	len - byte length of transaction
 *	seq - sequence number
 * Return value:
 *	Return code from kernel
 * Side effects:
 *	NONE
 */
int
clm_recv (void *buf, int len, int seq)
{
  recvtran_t recv;

  recv.buf = buf;
  recv.len = len;
  recv.seq = seq;

  return(ioctl(locks_fd, DLM_IOCTL_RECVTRAN, &recv));  
}

/*
 * Function:
 *	ASTget
 * Description:
 *	Allocate a request block, write through the locks interface and
 *  request that the kernel deliver an ast block to the address
 *  we specify.
 * Arguments:
 *	p - pointer to ASTpoll info
 * Return value:
 *	>0 - AST returned
 *   0 - no pending ASTs
 *  <0 - error
 * Side effects:
 *	NONE
 */
int
ASTget (void *p)
{
  if (!initted())
    init_glue();
#ifndef __linux__
  fp = (int (*)())func_tab[FP_ASTGET];
  return((*fp)(p));
#endif /* !__linux__ */

#ifdef NOT_DEFINED
  fprintf(stderr,
          "Retrieve AST to address [%p]\n", p);
#endif

  return(ioctl(locks_fd, DLM_IOCTL_ASTGET, p));
}

/*
 * Function:
 *	set_debug
 * Description:
 *	Wrapper function for kernel routine.
 * Arguments:
 *	debug_flags - debug mask to set
 * Return value:
 *	Value returned from ioctl.
 * Side effects:
 *	NONE
 */
int
set_debug (int debug_flags)
{
  if (!initted())
    init_glue();

  return(ioctl(locks_fd, DLM_IOCTL_SETDEBUG, debug_flags));
}
    
/*
 * Function:
 *	clm_grp_op
 * Description:
 *	Wrapper function for kernel routine.
 * Arguments:
 *	op - group operation to perform
 *	gid - pointer to group id
 * Return value:
 *	Value returned from ioctl.
 * Side effects:
 *	NONE
 */
int
clm_grp_op(int op, void *gid, int flags)
{
  groupop_t group_op;

  group_op.op = op;
  group_op.gid = gid;
  group_op.flags = flags;
  group_op.pid = OurPid;

  if (!initted())
    init_glue();

  return(ioctl(locks_fd, DLM_IOCTL_GROUPOP, &group_op));
}

