/*
 * Interface library for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * http://www.gnu.org/copyleft/lesser.html 
 */
#ident "$Id: api_group.c,v 1.3 2001/11/06 13:32:46 bkon Exp $"

/*
 *   COMPONENT_NAME: DLM LIB
 *
 *   FUNCTIONS: 
 */

#ifdef __KERNEL__

# ifndef MODULE
#  define MODULE
# endif

# include <linux/kernel.h>
# define __NO_VERSION__
# include <linux/module.h>
# ifdef USING_FULL_VERSIONING
#  if CONFIG_MODVERSIONS==1
#   define MODVERSIONS
#   include <linux/modversions.h>
#  endif
# endif /* USING_FULL_VERSIONING */


#else


#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>

#include "dlm.h"
#include "ast.h"
#include "clmstructs.h"
#include "pti.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "general_msg.h"
#define MSGSTR(Num, Str) catgets(get_catd("general.cat"), GENERAL, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */
#endif /* __KERNEL__ */

#include "api_data.h"
#include "api_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/*
 * Name
 *	dlm_grp_create
 * Description
 *	This function is used by lock clients to create a new lock group.
 *	A process is not allowed to create a group if it is currently
 *	attached to one.  The create function returns a group id value that
 *	other processes should use to attach to this lock group.  Upon
 *	successful completion, the process is attached to the created group.
 * Arguments
 *	gid - pointer to storage for the group id that is returned
 *	flags - option flags
 * Returns
 *	One of the following values in dlm_stats_t:
 *
 * 	DLM_NORMAL - Success.  Group created.
 *	DLM_DENIED - Process is already attached to another group.
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
dlm_grp_create(int *gid, int flags)
{
	int		tmp_gid;
	dlm_stats_t	ret;

	if (NULL == gid) {
		ret = DLM_BADARGS;
		goto done;
	}

	if ((ret = Dlm_Init()) != DLM_NORMAL) {
		goto done;
	}

again:
	ret = clm_grp_op(DLM_GRP_CREATE, &tmp_gid, flags);

	if (DLM_DENIED_GRACE_PERIOD == ret) {
	    api_sleep(GRACE_SLEEP);
	    goto again;
	}

	if (DLM_NORMAL == ret) {
		*gid = tmp_gid;
	}

done:
	return ret;
}

/*
 * Name
 *	dlm_grp_attach
 * Description
 *	This function is used by lock clients to attach to an existing
 *	lock group.  A process is not allowed to attach to a group if it is
 *	already attached to one.  The process must specify a group id that
 *	has been returned in a successful call to dlm_grp_create().
 * Arguments
 *	gid - group id of group to attach to
 *	flags - option flags
 * Returns
 *	One of the following values in dlm_stats_t:
 *
 * 	DLM_NORMAL - Success.  Attached to group.
 *	DLM_DENIED - Process is already attached to a group.
 *	DLM_IVGROUPID - Invalid group id.
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
dlm_grp_attach(int gid, int flags)
{
	int		tmp_gid = gid;
	dlm_stats_t	ret;

	if ((ret = Dlm_Init()) != DLM_NORMAL) {
		goto done;
	}

again:
	ret = clm_grp_op(DLM_GRP_ATTACH, &tmp_gid, flags);

	if (DLM_DENIED_GRACE_PERIOD == ret) {
	    api_sleep(GRACE_SLEEP);
	    goto again;
	}

done:
	return ret;
}


/*
 * Name
 *	dlm_grp_detach
 * Description
 *	This function is used by lock clients to detach from a group that
 *	it is currently associated with.
 * Arguments
 *	flags - option flags
 * Returns
 *	One of the following values in dlm_stats_t:
 *
 * 	DLM_NORMAL - Success.  Detached from group.
 *	DLM_DENIED - Process is not attached to a group.
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
dlm_grp_detach(int flags)
{
	int		tmp_gid;
	dlm_stats_t	ret;

	if ((ret = Dlm_Init()) != DLM_NORMAL) {
		goto done;
	}

again:
	ret = clm_grp_op(DLM_GRP_DETACH, &tmp_gid, flags);

	if (DLM_DENIED_GRACE_PERIOD == ret) {
	    api_sleep(GRACE_SLEEP);
	    goto again;
	}

done:
	return ret;
}


