/*
 * Interface library for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * http://www.gnu.org/copyleft/lesser.html 
 */
#ident "$Id: api_init.c,v 1.5 2001/11/29 13:20:22 bkon Exp $"

/*
 *   COMPONENT_NAME: DLM LIB
 *
 *   FUNCTIONS: init_module
 *              init_tq_struct
 *              cleanup_module
 *              Dlm_Init
 *		api_reqresp
 *		api_response
 *		appreg
 *		dlm_setdebug
 */

#ifdef __KERNEL__

#ifndef MODULE
# define MODULE
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include "clmstructs.h"
#ifdef USING_FULL_VERSIONING

#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif /* USING_FULL_VERSIONING */

extern int OurSiteID;
extern clm_client_t * findclientbypid(int site, int pid);

#else

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/stat.h>



#endif /* __KERNEL__ */

#include "dlm.h"
#include "ast.h"
#include "clmstructs.h"
#include "dlm_kernel.h"
#include "pti.h"
#include "clmutex_r.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "general_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("general.cat"), GENERAL, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "api_data.h"
#include "api_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
#ifdef __KERNEL__
    static spinlock_t clminit_mutex;
    static spinlock_t sequence_mutex;
    extern struct tq_struct dlm_keventd_task;
    static void init_tq_struct ( struct tq_struct *tqp);
#else /* __KERNEL__ */
    #ifdef  _THREAD_SAFE
        /* clminit_mutex protects OurPid, HostName and sequence */
        static pthread_mutex_t clminit_mutex = PTHREAD_MUTEX_INITIALIZER;
        static pthread_mutex_t sequence_mutex = PTHREAD_MUTEX_INITIALIZER;
    #endif
#endif /* __KERNEL__ */

static int sequence = 0;	/* UDP sequence number */
#ifndef __KERNEL__
static char HostName[MAXNAMLEN] = {0};
#endif /* __KERNEL__ */

static int appreg(void);

#ifdef __KERNEL__
/*
 * Name: init_tq_struct
 *
 * Description:
 *      Initializes the tq_struct passed to the keventd thread for kernel
 *      API handling of ASTs.
 *
 * Parameters:
 *      tqp             Ptr to the tq_struct
 *
 * Return value:
 *      void
 *
 * Side effects:
 *     
 */
static void
init_tq_struct ( struct tq_struct *tqp)
{
        DBPRINTF((DBG_FUNC, "enter init_tq_struct\n"));
        tqp->sync = 0;
        INIT_LIST_HEAD(&(tqp->list));
        tqp->routine =  kernapi_k_ASTget;
        tqp->data = NULL;
        DBPRINTF((DBG_FUNC, "exit init_tq_struct\n"));

        return;
}


/* ============================================================ */
/*
 * Our init routine.  Called when the module is loaded.
 *
 */
int
init_module(void)
{
    spin_lock_init (&clminit_mutex);
    spin_lock_init (&sequence_mutex);
    init_tq_struct(&dlm_keventd_task);
    return(0);
}

/* ============================================================ */
void
cleanup_module(void)
{
}
#endif /* __KERNEL__*/

/* ============================================================ */
/*
 * Name
 *	Dlm_Init
 * Description
 *	Initialize the API if necessary.
 * Original AIX implementation:
 * 	We use the OurPid variable to determine if this process
 * 	has initialized yet. This costs a getpid() call each time, but
 * 	it works across forks. Get cluster info, register with
 *	purged, register with lock manager. (Different threads within
 *  a process all return same pid for getpid() on AIX.)
 *
 * Linux note: On Linux, every thread has a different pid, which
 * means two different threads coming here would think that the
 * process had not yet initialized, resulting in bad things.  So
 * the OurPid scheme alone doesn't work. On the other hand, if we
 * use a simple variable, then we will believe a client across a
 * fork is still the initialized client.  Seeing no immediately
 * obvious solution here, we'll simply use a variable rather than
 * depend on getpid.  Which leads to...
 *
 * Limitation:  If a process initializes with DLM, and then it
 * forks, we will consider the child already initialized as part
 * of its parent.  Since the child inherits a copy of the open
 * file descriptors at fork time, either will be able to use the
 * DLM session.  In addition, since dlmdk depends upon seeing the
 * device session close, if the parent dies, but the child remains,
 * then dlmdk won't see the device be closed!  Note that this may
 * be, indeed, useful for the DLM clients, point is that we need
 * to be sure these semantics are documented.  Or, if we go with
 * the AIX semantics, that those are documented.
 *
 * Arguments
 *	NONE
 * Returns
 *	DLM_NORMAL for success or already initialized
 *	DLM_NOLOCKMGR
 *  DLM_BAD_DEVICE_PATH
 *  DLM_NO_DEVICE_PERMISSION
 *	any dlm_errno status set by GetClusterInfo
 * Global Data Affected
 *	OurPid
 *	sequence
 */
dlm_stats_t 
Dlm_Init(void)
{
#ifndef __KERNEL__
  register int pid;
  dlm_stats_t _init; 
#endif /* __KERNEL__ */

  TS_LOCK(&clminit_mutex);
#ifdef __KERNEL__
  lvb_old_behave = 0;
  /* Check if process  has already initialized */
  if ( NULL != findclientbypid(OurSiteID, OurPid)) {
      TS_UNLOCK(&clminit_mutex);
      return DLM_NORMAL;
  } 

#else
  /* Check if any thread has already initialized */
  if (0 != ClientInit) {
      TS_UNLOCK(&clminit_mutex);
      return set_dlm_errno(DLM_NORMAL);
  } 

  /*
   * First init call by any thread associated with this
   * address space (i.e., process).  Use whatever comes
   * back from getpid(), although it may not be the
   * pid associated with "thread 0" of the process.  It
   * doesn't matter, although the DLM (dlmdk) *DOES* use
   * this pid to identify a client, unlike on AIX it
   * *DOES NOT* monitor for client death using the pid.
   * Thus, so long as every request from this client has
   * the same id, it doesn't much matter what we use.
   */
  pid = getpid();

  /* sequence already protected by clminit_mutex */
  /* reset in case we got here via fork */
  sequence = 0;

  /*
   * No idea what precisely they meant here, but the
   * only place I can find any impact of this is in the
   * valueblock() routine in dlmdk/clm_clmlock.c, but it
   * does modify lock value block setting allowances.  My
   * assumption is that they had originally implemented a
   * semantic, then decided to 'fix' it later, but had to
   * maintain old behavior due to someone's dependence upon
   * the original implementation.  This would not be the
   * first such case in products I've dealt with!  For now,
   * tell the code to ignore it even if set, and thus we
   * always get "new" behavior.
   */
#ifdef __linux__
  lvb_old_behave = 0;
  if (getenv("CLM_LVB_VAR")) {
    fprintf(stderr,
            "Lock value behavior variable [%s] set, ignoring.\n",
            "CLM_LVB_VAR");
  }
#else
  /* check for LVB env variable - see CR960021 */
  if (getenv(CLM_LVB_VAR)) {
    /* if flag is set, set var to disable new behvaior */
    lvb_old_behave = LKM_OLD_LVB;
  } else {
    lvb_old_behave = 0;
  }
#endif /* __linux__ */

  gethostname(HostName,MAXNAMLEN);

#ifdef _THREAD_SAFE
  /* initialize dlm_errno per-thread data key */
  init_dlm_errno_hdl();
#endif	

  /*
   * The "glue" layer on AIX handled setup of the dynamic
   * system calls.  On Linux it handles opening up the
   * client lock device.  Any problems opening the device
   * are basically passed right back through to the client.
   */
  _init = init_glue();
  if (DLM_NORMAL != _init) {
    TS_UNLOCK(&clminit_mutex);
    return set_dlm_errno(_init);
  }      

  /* OurPid already protected by clminit_mutex */
  OurPid = pid;
#endif /* __KERNEL__ */
  if (appreg() != 0) {
    /* OurPid already protected by clminit_mutex */
#ifndef __KERNEL__
    OurPid = 0;
#endif /* __KERNEL__ */
    TS_UNLOCK(&clminit_mutex);
    return set_dlm_errno(DLM_NOLOCKMGR);
  }

#ifndef __KERNEL__
  ClientInit = 1;                       /* Client ready to go. */
#endif /* __KERNEL__ */

  TS_UNLOCK(&clminit_mutex);
  return set_dlm_errno(DLM_NORMAL);
}

/*
 * Name
 *	api_response
 * Description
 * 	wait for a response from the clm
 * Arguments
 *	response - transction buffer in which to receive response
 * Returns
 *	0 for success
 * 	-1 for failure 
 * Global Data Affected
 *	dlm_errno
 */
#ifndef __KERNEL__
int api_response(struct transaction *response)
{
  int seq;
    
	/* Note: This will have to be changed for unixlocks */
	/* Store sequence number in local variable, in case another 
	   thread increments sequence */
	TS_LOCK(&sequence_mutex);
	seq = ++sequence;
	TS_UNLOCK(&sequence_mutex);
	
	if (api_recvpkt(response, sizeof(* response), seq) != 0)  {
		set_dlm_errno(DLM_NOLOCKMGR);
		return -1;
	}
	if (response->clm_status == DLM_REJECTED) {
		set_dlm_errno(DLM_REJECTED);
		return -1;
	}
	if (response->clm_sequence != seq) {
		fprintf(stderr, MSGSTR(TRANS_MISMATCH,
			"Transaction mismatch detected in API\n"));
		return -1;
	}

	if (transdebug)
		printf("RECV TRANSACTION %d\n", response->clm_sequence);
	return 0;
}
#endif /*__KERNEL__*/

bool_t
api_reqresp(dlm_ops_t type, struct transaction *request,
	struct transaction *response)
{
	int seq;
	TS_LOCK(&sequence_mutex);
	seq = ++sequence;
	TS_UNLOCK(&sequence_mutex);

	request->clm_prog = DLM_PROG + local_debug;
	request->clm_vers = DLM_VERS;
	request->clm_type = type;
	request->clm_direction = DLM_REQUEST;
	request->clm_sequence = seq;
	request->clm_pid = OurPid;

	if (transdebug)
		print_debug("[%s] SEND TRANSACTION %d\n", __FUNCTION__, request->clm_sequence);

#ifdef __KERNEL__
        if (k_clm_send(request, sizeof(*request), OurPid)){
            return(-1);
        }
        if (k_clm_recv(response, sizeof(*response), seq, OurPid)){
            return(-1);
        }

#else /* __KERNEL__*/


	if (api_sendpkt(request, sizeof(* request)) != 0)  {
		set_dlm_errno(DLM_NOLOCKMGR);
		return -1;
	}

	if (api_recvpkt(response, sizeof(* response), seq) != 0)  {
		set_dlm_errno(DLM_NOLOCKMGR);
		return -1;
	}
#endif /* __KERNEL__ */

	if (response->clm_status == DLM_REJECTED) {
		set_dlm_errno(DLM_REJECTED);
		return -1;
	}
	if (response->clm_sequence != seq) {

		fprint_error( MSGSTR(TRANS_MISMATCH,
			"Transaction mismatch detected in API\n"));
		return -1;
	}

	if (transdebug)
		print_debug("[%s] RECV TRANSACTION %d\n", __FUNCTION__, response->clm_sequence);
	return 0;
}

/* 
 * Name
 *	appreg
 * Description
 *	register as a client with the lock manager
 * Arguments
 *	NONE
 * Returns
 *	0 for success
 *	-1 for error
 */
static int
appreg(void)
{
	struct transaction request;
	struct transaction response;
	request.clm_appreg.pid = OurPid;
	request.clm_respflags = RESP_UDP;

again:
	if (api_reqresp(DLM_APPREG,&request,&response) != 0)
		return -1;
	if (response.clm_status == DLM_DENIED_GRACE_PERIOD) {
		api_sleep(GRACE_SLEEP);
		goto again;
	}
	if (response.clm_status != DLM_NORMAL)
		return -1;
	return 0;
}

/*
 * Name
 *	dlm_setdebug
 * Description
 * 	This function takes a string as an argument,
 *	parses each debug option found and call the function
 *	to set the lock manager debug mask to the new debug value.
 *	Erroneous strings get reported on stderr.
 *	The option string is expected to be comma separated
 *	
 * Arguments
 *	flagstring - string from command line
 * Returns
 *	Mask of all selected debugging options
 * Global Data Affected
 *	NONE
 */
#ifndef __KERNEL__
static struct options {
	char *opt_name;	
	int opt_mask;
} options[] = {
{ "ALL", 	0xffffffff },
{ "BASIC", 	DBG_BASIC },
{ "FUNC", 	DBG_FUNC },
#if defined(TIME_STATS)
{ "STATS", 	DBG_STATS },
#endif
{ "COMM", 	DBG_COMM },
{ "DEADLOCK", 	DBG_DEADLOCK },
{ "CTI",	DBG_CTI},
{ "CLMLOCK",	DBG_CLMLOCK},
{ "UNIXLOCK",	DBG_UNIXLOCK},
{ "LOCK",	DBG_LOCK},
{ "RESOURCE",	DBG_RESOURCE},
{ "DYNRES",	DBG_DYNRES},
{ "BATCH",	DBG_BATCH},
{ "TIMEOUT",	DBG_TIMEOUT},
{ "NETFAIL",	DBG_NETFAIL},
{ "CONFIG",	DBG_CONFIG},
{ "SCN",	DBG_SCN},
{ "DIRECT",	DBG_DIRECT},
{ "MIGR",	DBG_MIGR},
{ "MIGR2",	DBG_MIGR2},
{ "RECOV",	DBG_RECOV},
{ "TUNE",	DBG_TUNE},
{ "SYNC",	DBG_SYNC},
{ "LOCAL",	DBG_LOCAL},
{ "KDBX",	DBG_KDBX},
{ "FAILSYNC",	DBG_FAILSYNC},
{ "AST",	DBG_AST}};

int
dlm_setdebug(char *flagstring)
{
	int i;
	int mask = 0;
	char *p = flagstring;
	char *comma;

	while (p && *p) {
		comma = (char *)index(p, ',');
		if (comma != NULL) {
			*comma = '\0';
		}
		for (i = 0; i < sizeof(options)/sizeof(options[0]); i++) {
			if (strcmp(p, options[i].opt_name) == 0) {
				mask |= options[i].opt_mask;
				fprintf(stderr, "%s debugging enabled\n", 
						options[i].opt_name);
				break;
			}
		}
		if (i >= sizeof(options)/sizeof(options[0])) {
		    fprintf(stderr,MSGSTR(BAD_OPTION,"bad option flag '%s'\n"), p);
		}
		if (comma) {
			*comma = ',';
			p = comma;
			p++;	
		} else {
			break;
		}
	}
	set_debug(mask);

    return(mask);
}
#endif /* __KERNEL__*/

