/*
 * Interface library for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * http://www.gnu.org/copyleft/lesser.html 
 */
#ident "$Id: api_tune.c,v 1.4 2001/11/29 13:20:22 bkon Exp $"

/*
 *   COMPONENT_NAME: DLM LIB
 *
 *   FUNCTIONS: MSGSTR
 *              dlm_getstats
 *		dlm_getglobparams
 *		dlm_getresparams
 *		dlm_setglobparams
 *		dlm_setresparams
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>

#include "dlm.h"
#include "ast.h"
#include "clmstructs.h"
#include "pti.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "general_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("general.cat"), GENERAL, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */

#include "api_data.h"
#include "api_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/*
 * Name
 *	dlm_getresparams
 * Description
 *	Top level of get migration parameters interface call
 * Arguments
 *	resname - resource name
 *	namelen - length of resource name
 *	type - type of resource
 *	params - pointer to resource migration parameters structure
 * Returns
 *	DLM_NORMAL for success
 *	DLM_BADRESOURCE - resource not found
 *	DLM_NOLOCKMGR
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
dlm_getresparams(char		*resname,	/* resource name */
                 short		namelen,	/* resource name length */
                 short		type,		/* resource type */
                 dlm_resparams_t	*params)	/* res migration parameters */
{
	struct transaction request;
	struct res_migr_params *vi = &request.clm_rmpinfo;
	struct transaction response;
	int ret_val;	/* value to be returned from function */
	int reconfig = 0;


	/* check clinfo initialization */
	/* dlm_errno set by Dlm_Init */
	if ((ret_val = Dlm_Init()) != DLM_NORMAL) {
		goto done;
	}

	if (namelen <= 0 || namelen > MAXNAMLEN) {
		ret_val = DLM_IVBUFLEN;
		goto done;
	}

	if (NULL == params) {
		ret_val = DLM_BADARGS;
		goto done;
	}

        memset(&request, 0, sizeof(request));

	/* set up request for lock manager */
	request.clm_locktype = type;
	request.clm_respflags = RESP_UDP;
	request.clm_pid = OurPid;
	memcpy(&vi->rmp_resname[0], resname, namelen);
	vi->rmp_namelen = namelen;
	vi->rmp_flags = 0;
	vi->rmp_type = type;
retry:
	if (api_reqresp(DLM_RMIGR_OP, &request, &response)) {
		ret_val = set_dlm_errno(DLM_NOLOCKMGR);
		goto done;
	}

	if (response.clm_status == DLM_WORKING) {
		printf(MSGSTR(DLM_WORKING, "DLM_WORKING_TER returned from API call!!\n"));
		if (api_response(&response) != 0) {
			ret_val = set_dlm_errno(DLM_NOLOCKMGR);
			goto done;
		}
	}

	if (dlm_errno == DLM_NORMAL && 
		response.clm_status == DLM_DENIED_GRACE_PERIOD) {
		reconfig = 1;
		sleep(GRACE_SLEEP);
		goto retry;
	}

	ret_val = set_dlm_errno(response.clm_status);

	if (ret_val == DLM_NORMAL)
		memcpy(params, &response.clm_rmpinfo.rmp_params, 
				sizeof(*params));

done:

	return ret_val;
}


/*
 * Name
 *	dlm_setresparams
 * Description
 *	Top level of set migration parameters interface call
 * Arguments
 *	resname - resource name
 *	namelen - length of resource name
 *	type - type of resource
 *	params - pointer to resource migration parameters structure
 * Returns
 *	DLM_NORMAL for success
 *	DLM_BADRESOURCE - resource not found
 *	DLM_NOLOCKMGR
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
dlm_setresparams(char		*resname,	/* resource name */
                 short		namelen,	/* resource name length */
                 short		type,		/* resource type */
                 dlm_resparams_t	*params)	/* res migration parameters */
{
	struct transaction request;
	struct res_migr_params *vi = &request.clm_rmpinfo;
	struct transaction response;
	int ret_val;	/* value to be returned from function */
	int reconfig = 0;


	/* check clinfo initialization */
	/* dlm_errno set by Dlm_Init */
	if ((ret_val = Dlm_Init()) != DLM_NORMAL) {
		goto done;
	}

	if (namelen <= 0 || namelen > MAXNAMLEN) {
		ret_val = DLM_IVBUFLEN;
		goto done;
	}

	if (NULL == params) {
		ret_val = DLM_BADARGS;
		goto done;
	}

	if ((params->cr_valid & DLM_RES_STICKINESS) &&
			params->cr_stickiness > MAX_STICKINESS) {
		ret_val = DLM_BADARGS;
		goto done;
	}

        memset(&request, 0, sizeof(request));

	/* set up request for lock manager */
	request.clm_locktype = type;
	request.clm_respflags = RESP_UDP;
	request.clm_pid = OurPid;
	memcpy(&vi->rmp_resname[0], resname, namelen);
	memcpy(&vi->rmp_params, params, sizeof(*params));
	vi->rmp_namelen = namelen;
	vi->rmp_flags = DLM_TUNE_SET;
	vi->rmp_type = type;
retry:
	if (api_reqresp(DLM_RMIGR_OP, &request, &response)) {
		ret_val = set_dlm_errno(DLM_NOLOCKMGR);
		goto done;
	}

	if (response.clm_status == DLM_WORKING) {
		printf(MSGSTR(DLM_WORKING, "DLM_WORKING_TER returned from API call!!\n"));
		if (api_response(&response) != 0) {
			ret_val = set_dlm_errno(DLM_NOLOCKMGR);
			goto done;
		}
	}

	if (dlm_errno == DLM_NORMAL && 
		response.clm_status == DLM_DENIED_GRACE_PERIOD) {
		reconfig = 1;
		sleep(GRACE_SLEEP);
		goto retry;
	}

	ret_val = set_dlm_errno(response.clm_status);

done:

	return ret_val;
}
