/*
 * Interface library for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * http://www.gnu.org/copyleft/lesser.html 
 */
#ident "$Id: api_unix.c,v 1.4 2001/11/29 13:20:22 bkon Exp $"

/*
 *   COMPONENT_NAME: DLM LIB
 *
 *   FUNCTIONS: dlmregionlock
 */
/*
 * Distributed Lock Manager Library 
 *
 * This file contains the functions that provide user access to the 
 * facilities of the distributed lock manager through a UNIX record
 * locking interface.
 *
 *
 */


#ifdef __KERNEL__

# ifndef MODULE
#  define MODULE
# endif
 
# include <linux/kernel.h>
# define __NO_VERSION__
# include <linux/module.h>
# ifdef USING_FULL_VERSIONING
#  if CONFIG_MODVERSIONS==1
#   define MODVERSIONS
#   include <linux/modversions.h>
#  endif
# endif /* USING_FULL_VERSIONING */
 
#else


#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>

#endif /*__KERNEL__*/

#include "dlm.h"
#include "ast.h"
#include "clmstructs.h"
#include "pti.h"

#include "api_data.h"
#include "api_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
/*
 * Name
 *	clmregionlock
 * Description
 *	UNIX lock/unlock call, validates arguments and sends
 *	request to cllockd, will reject locally for bad
 *	resource handle.
 * Arguments
 *	rh - resource handle as returned by Dlm_Register
 * 	offset - lower bound of lock region
 *	len - length of lock region
 *	flags - locking flags
 * Returns
 *	0 for success
 *	valid dlm_errno value
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
dlmregionlock(union dlm_rh rh,
              unsigned long offset,
              unsigned long len,
              unsigned long flags)

{
	struct transaction request;
	struct transaction response;
	struct lockinfo *lockinfo = &request.clm_lockinfo;
	dlm_ops_t cmd = (flags&LOCK_UN) ? DLM_UNLOCK : DLM_LOCK;

#if defined(DEBUG)
    if (debug) {
      fprintf(stderr,
              "dlmregionlock(0x%x,%u,%u,%u)\n",
              rh.rh, offset, len, flags);
    }
#endif

        memset(&request, 0, sizeof(request));
	
	/* Check to see if any invalid flags are set */
	if (flags & ~UNIX_FLAGMASK)
	        return (DLM_BADARGS);

	/* Don't allow offset = ULONG_MAX
	   This causes zero length lock */
	if (offset == ULONG_MAX)
	        return (DLM_BADARGS);

	if (Dlm_Init() != DLM_NORMAL)
		return DLM_DENIED;

	response.clm_status = DLM_DENIED;
	request.clm_locktype = DLM_RES_UNIX;
	request.clm_respflags = RESP_UDP;

	/* Don't allow zero length lock */
	if (len == 0)
		len = ULONG_MAX;
	lockinfo->lk_flags = 0;
	if (flags & LOCK_EX)
		lockinfo->mode = LKM_EXMODE;
	else if (flags & LOCK_SH)
		lockinfo->mode = LKM_PRMODE;
	if (flags & LOCK_NB)
		lockinfo->lk_flags |= LKM_NOQUEUE;


	lockinfo->rh = rh;
	lockinfo->pid = OurPid;
	lockinfo->lb = offset;
	lockinfo->ub = offset + len;

	/* check if upper bound is outside of integer range */
	/* if so set upper bound to max integer value */
	if (lockinfo->ub < lockinfo->lb)
	  lockinfo->ub = (unsigned int)(-1);

retry:
	if (api_reqresp(cmd, &request, &response))
		return dlm_errno;

	if (response.clm_status == DLM_WORKING)
		if (api_response(&response) != 0)
			return dlm_errno;

	if (dlm_errno == DLM_NORMAL && 
		response.clm_status == DLM_DENIED_GRACE_PERIOD) {
		api_sleep(GRACE_SLEEP);
		goto retry;
	}
	return response.clm_status;
}

/*
 * Name
 *	Dlm_Register
 * Description
 *	Request a resource handle for a UNIX lock
 * Arguments
 *	name - NULL terminated resource name
 * Returns
 *	valid resource handle for success
 *	NULL for failure
 * Global Data Affected
 *	NONE
 */
union dlm_rh 
clmregister(char *name)
{
	return register_resource(name, DLM_RES_UNIX);
}
