/*
 * Interface library for the IBM OSS Distributed Lock Manager
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/opensource/linux/projects/dlm
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * http://www.gnu.org/copyleft/lesser.html 
 */
#ident "$Id: api_vms.c,v 1.6 2001/11/29 13:20:22 bkon Exp $"

/*
 *   COMPONENT_NAME: DLM LIB
 *
 *   FUNCTIONS: MSGSTR
 *		clm_lock_reqresp
 *		dlmlock
 *      dlmlock_sync
 *		dlmunlock
 *		dlmunlock_sync
 *		fprintf
 *		get_avg
 *		set_handle
 */

#ifdef __KERNEL__

# ifndef MODULE
#  define MODULE
# endif

# include <linux/kernel.h>
# define __NO_VERSION__
# include <linux/module.h>
# ifdef USING_FULL_VERSIONING
#  if CONFIG_MODVERSIONS==1
#   define MODVERSIONS
#   include <linux/modversions.h>
#  endif
# endif /* USING_FULL_VERSIONING */

#else 

# ifdef _THREAD_SAFE
#  include <pthread.h>
# endif
# include <stdio.h>
# include <unistd.h>
# include <string.h>
# include <sys/types.h>
# include <sys/file.h>
# include <sys/stat.h>

#endif /* __KERNEL__ */

#include "dlm.h"
#include "ast.h"
#include "clmstructs.h"
#include "dlm_kernel.h"
#include "pti.h"

#ifdef HAVING_NLS_ON_LINUX
/* for NLS */
#include "general_msg.h"
#define MSGSTR(Num, Str) catgets((struct _catalog_descriptor*)get_catd("general.cat"), GENERAL, Num, Str)
#else   /* HAVING_NLS_ON_LINUX */
#define MSGSTR(Num, Str) Str
#endif  /* HAVING_NLS_ON_LINUX */
/* end of NLS */


#include "api_data.h"
#include "api_proto.h"  /* Should be LAST included file. */

/* ============================================================ */
#define TIMESTAMP

#if defined(TIMESTAMP)
#ifndef __KERNEL__
double	last_start_time;
double	total_time		= 0.0;
int	samples 		= 0;

extern double ticks(void);


double
get_avg (void)
{
    double avg;

    avg = total_time / (double)samples;
    total_time	= 0.0;
    samples	= 0;
    return(avg);
}
#endif /* __KERNEL__ */
#endif

/*
 * Function:
 *      set_handle
 * Description:
 *      Local function set_handle().
 * Arguments:
 *      user_h - pointer to user space AST handle
 *      funcp   - ast function pointer
 *      argp    - ast argument pointer
 *      lksbp   - pointer to lock status block
 *      pid     - pid of client
 *      tid     - thread id of client
 * Return value:
 *      NONE
 * Side effects:
 *      NONE
 */
void
set_handle(ast_handle_t *user_h,
           int pid,
           int tid,
           struct lockstatus *lksb,
           dlm_astfunc_t * funcp,
           void *argp)
{
    user_h->asth_pid = pid;
    user_h->asth_tid = tid;
    user_h->asth_dest = lksb;
    user_h->asth_funcp = funcp;
    user_h->asth_arg = argp;
}

/*
 * Name
 *	clm_lock_reqresp
 * Description
 *	Second layer of vms locking call
 * Arguments
 *	mode - lock mode
 *	lksb - pointer to lock status block
 *	flags - lock flags
 *	name - NON NULL terminated name
 *	namelen - bytelength of name
 *	asth - pointer to populated ast handle
 *	basth - pointer to populated bast handle
 * Returns
 *	DLM_NORMAL for success
 *	DLM_SYNC for synchronous completion
 *	valid dlm_errno status code otherwise
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
clm_lock_reqresp(int mode,
                 struct lockstatus *lksb,
                 int flags,
                 void *name,
                 unsigned int namelen,
                 ast_handle_t *asth,
                 ast_handle_t *basth,
                 dlm_xid_t *xid)
{
	struct transaction request;
	struct transaction response;
	struct lockreq *vi = &request.clm_lockreq;

        memset(&request, 0, sizeof(request));
	request.clm_locktype = request.clm_lockinfo.type = DLM_RES_VMS;
	request.clm_asth = *asth;

	/* If XID defined, set LKM_XID flag and 
	   make sure LK_XID_CONFLICT flag is set */
	if (xid)
	{
	    flags |= LKM_XID;
	    if (!(flags & LKM_XID_CONFLICT)) {
		return set_dlm_errno(DLM_BADARGS);
	    }
	}
	else
	{
	    /* If no XID, clear XID flag */
	    if (flags & LKM_XID)
	        flags &= ~LKM_XID;
	}

	if (flags & LKM_CONVERT) {
		request.clm_cbasth = *basth;
	} else {
		request.clm_basth = *basth;
		if (xid)
		    memcpy(&request.clm_lockinfo.xid, xid, sizeof(dlm_xid_t));
	}
	request.clm_signo = notify_signo;
	request.clm_lockinfo.bastmode = LKM_IVMODE;

	/* always send UDP response back - clear flag after it's sent */
	request.clm_respflags = RESP_UDP | RESP_AST;

	if (lksb == NULL)
		return set_dlm_errno(DLM_BADARGS);

	if (mode > LKM_MAXMODE || mode < 0) {
		return set_dlm_errno(DLM_BADPARAM);
	}

	if (flags & ~(LKM_FLAGMASK)) {
		return set_dlm_errno(DLM_BADARGS);
	}

	if ((flags & LKM_BLOCK) && (flags & LKM_SYNCSTS)) {
	        return set_dlm_errno(DLM_BADARGS);
	}

	if ((flags & LKM_CONVERT) == 0) {
		/* here for new lock */
		if (name == NULL) {
			return set_dlm_errno(DLM_BADARGS);
		}

		if (namelen == 0 || namelen > 31) {
			return set_dlm_errno(DLM_IVBUFLEN);
		}
			
		memcpy(vi->nameinfo.name, name, namelen);
		vi->nameinfo.namelen = namelen;
		vi->lockinfo.lockid = -1;
	} else {
		name = NULL;
		vi->nameinfo.namelen = 0;
		vi->lockinfo.lockid = lksb->lockid;
		if (flags & LKM_VALBLK)
			memcpy(vi->lockinfo.lkvalue, lksb->value, 
				sizeof(lksb->value));
	}
		
	vi->lockinfo.reqmode = mode;
	vi->lockinfo.mode = LKM_NLMODE;
	vi->lockinfo.lk_flags = flags | lvb_old_behave;
	vi->lockinfo.rh.rh = 0;
	vi->lockinfo.li_to = lksb->timeout;
	lksb->status = DLM_NORMAL;

	if (api_reqresp(DLM_LOCK, &request, &response))
		return dlm_errno;

	lksb->lockid = response.clm_lockinfo.lockid;

	/* if synchronous, lksb must be copied out of response */
	if (request.clm_lockinfo.lk_flags & LKM_BLOCK) {
		memcpy( &lksb->value[0],&response.clm_lockinfo.lkvalue[0],
				sizeof(lksb->value));
		lksb->status = response.clm_status;
	}

	set_dlm_errno(lksb->status);

	/*
	 * Neither DLM_NOTQUEUED nor
	 * DLM_VALNOTVALID should be returned to the client.  That
	 * information should be read out of the lksb.   *bg*
	 */
	if ((response.clm_status == DLM_VALNOTVALID) ||
			(response.clm_status == DLM_NOTQUEUED))
	   return DLM_NORMAL;
	else
	   return response.clm_status;
}

/*
 * Name
 *	dlmlockx
 * Description
 *	Top level of vms locking call that accepts transaction id's.  This
 *	call basically mimicks the dlmlock call, but takes the extra
 *	argument of a XID.
 *	NOTE:  Changes made to dlmlock will likely need to be applied here.
 * Arguments
 *	mode - lock mode
 *	lksb - pointer to lock status block
 *	flags - lock flags
 *	name - NON NULL terminated name
 *	namelen - bytelength of name
 *	ast - function pointer to AST routine
 *	astargs - argument to AST routine
 *	bast - function pointer to BAST routine
 * Returns
 *	DLM_NORMAL for success
 *	DLM_SYNC for synchronous completion
 *	valid dlm_errno status code otherwise
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
dlmlockx(int mode,
         struct lockstatus *lksb,
         int flags,
         void *name,
         unsigned int namelen,
         dlm_astlockfunc_t * ast,
         void *astargs,
         dlm_bastlockfunc_t * bast,
         dlm_xid_t *xid)
{
	dlm_stats_t status;
	ast_handle_t asth;
	ast_handle_t basth;
	int reconfig = 0;
	int ret_val;		/* value to be returned from function */

	set_dlm_errno(DLM_NORMAL);
	
	if (Dlm_Init() != DLM_NORMAL) {
		goto done;
	}

	/*
	 | set an AST handle through which the lock manager can later
	 | allocate an astinfo structure and fire an asynchronous callback.
	 */
#ifdef _THREAD_SAFE
	/* Currently, only pass thread id to lock manager for synchronous 
	   lock requests */
	if (flags & LKM_BLOCK)
	    set_handle(&asth, OurPid, (int)pthread_self(), lksb, ast, astargs);
	else
#endif
	    set_handle(&asth, OurPid, 0, lksb, (dlm_astfunc_t *)ast, astargs);
	set_handle(&basth, OurPid, 0, 0, (dlm_astfunc_t *)bast, astargs);

again:
	/* 
 	 *  We could have changed roles in between calls to
	 *  dlmlockx.  
 	 */

	status = clm_lock_reqresp(mode, lksb, flags, name,
	    namelen, &asth, &basth, xid);

	if (status == DLM_DENIED_GRACE_PERIOD) {
		reconfig = 1;
		api_sleep(GRACE_SLEEP);
		goto again;
	}

	/* don't let the user see a WORKING status */
	/* XXX should never actually get a WORKING status here */
	if (status == DLM_WORKING || status == DLM_DENIED ||
			status == DLM_CVTUNGRANT)
		status = DLM_NORMAL;

	set_dlm_errno(status);

done:
	ret_val = dlm_errno;

	return ret_val;
}


/*
 * Name
 *      dlmlockx_sync
 * Description
 *      Making a blocking lock request.  The LKM_BLOCK flag is
 *      set and the LKM_SYNCSTS flag is cleared.  Then dlmlockx
 *      is called with NULL as the astarg.  The status of 
 *      dlmlockx is returned from dlmlockx_sync.
 * Arguments
 *	mode     - lock mode
 *	lksb     - pointer to lock status block
 *	flags    - lock flags
 *	name     - NON NULL terminated name
 *	namelen  - bytelength of name
 *	bastargs - argument to BAST routine
 * Returns
 *	DLM_NORMAL for success
 *	valid dlm_errno status code otherwise
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
dlmlockx_sync(int mode,
              struct lockstatus *lksb,
              int flags,
              void *name,
              unsigned int namelen,
              void *bastargs,
              dlm_bastlockfunc_t * bast,
              dlm_xid_t *xid)
{
  dlm_stats_t status;

  if (flags & LKM_SYNCSTS) /* if SYNCSTS set clear */
    flags &= ~LKM_SYNCSTS;

  flags |= LKM_BLOCK;   /* set on */
 
  status = dlmlockx(mode, lksb, flags, name,
		   namelen, NULL, bastargs, bast, xid);

  flags &= ~LKM_BLOCK;  /* clear */
  
  return status;
}


/*
 * Name
 *	dlmlock
 * Description
 *	Top level of vms locking call
 * Arguments
 *	mode - lock mode
 *	lksb - pointer to lock status block
 *	flags - lock flags
 *	name - NON NULL terminated name
 *	namelen - bytelength of name
 *	ast - function pointer to AST routine
 *	astargs - argument to AST routine
 * Returns
 *	DLM_NORMAL for success
 *	DLM_SYNC for synchronous completion
 *	valid dlm_errno status code otherwise
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
dlmlock(int mode,
        struct lockstatus *lksb,
        int flags,
        void *name,
        unsigned int namelen,
        dlm_astlockfunc_t * ast,
        void *astargs,
        dlm_bastlockfunc_t * bast)
{
	dlm_stats_t status;
	ast_handle_t asth;
	ast_handle_t basth;
	int reconfig = 0;
	int ret_val;		/* value to be returned from function */

	set_dlm_errno(DLM_NORMAL);
	
	if ((ret_val = Dlm_Init()) != DLM_NORMAL) {
		goto done;
	}

	/*
	 | set an AST handle through which the lock manager can later
	 | allocate an astinfo structure and fire an asynchronous callback.
	 */

#ifdef _THREAD_SAFE
	/* Currently, only pass thread id to lock manager for synchronous 
	   lock requests */
	if (flags & LKM_BLOCK)
	    set_handle(&asth, OurPid, (int)pthread_self(),
                   lksb, (dlm_astfunc_t *)ast, astargs);
	else
#endif
	set_handle(&asth, OurPid, 0, lksb, (dlm_astfunc_t *)ast, astargs);
	set_handle(&basth, OurPid, 0, 0, (dlm_astfunc_t *)bast, astargs);

again:
	/* 
 	 *  We could have changed roles in between calls to
	 *  dlmlock.  
 	 */

	status = clm_lock_reqresp(mode, lksb, flags, name,
	    namelen, &asth, &basth, NULL);

	if (status == DLM_DENIED_GRACE_PERIOD) {
		reconfig = 1;
		api_sleep(GRACE_SLEEP);
		goto again;
	}

	/* don't let the user see a WORKING status */
	/* XXX should never actually get a WORKING status here */
	if (status == DLM_WORKING || status == DLM_DENIED ||
			status == DLM_CVTUNGRANT)
		status = DLM_NORMAL;

	set_dlm_errno(status);
	ret_val = status;

done:

	return ret_val;
}


/*
 * Name
 *      dlmlock_sync
 * Description
 *      Making a blocking lock request.  The LKM_BLOCK flag is
 *      set and the LKM_SYNCSTS flag is cleared.  Then dlmlock
 *      is called with NULL as the astarg.  The status of 
 *      dlmlock is returned from dlmlock_sync.
 * Arguments
 *	mode     - lock mode
 *	lksb     - pointer to lock status block
 *	flags    - lock flags
 *	name     - NON NULL terminated name
 *	namelen  - bytelength of name
 *	bastargs - argument to BAST routine
 * Returns
 *	DLM_NORMAL for success
 *	valid dlm_errno status code otherwise
 * Global Data Affected
 *	NONE
 */
dlm_stats_t
dlmlock_sync(int mode,
             struct lockstatus *lksb,
             int flags,
             void *name,
             unsigned int namelen,
             void *bastargs,
             dlm_bastlockfunc_t * bast )
{
  dlm_stats_t status;

  if (flags & LKM_SYNCSTS) /* if SYNCSTS set clear */
    flags &= ~LKM_SYNCSTS;

  flags |= LKM_BLOCK;   /* set on */
 
  status = dlmlock(mode, lksb, flags, name,
		   namelen, NULL, bastargs, bast);

  flags &= ~LKM_BLOCK;  /* clear */
  
  return status;
}


/*
 * remove a lock or cancel a blocked request
 *
 * Linux port, this has been renamed to dlmunlock_sync, and the
 * dlmunlock_async routine to simply dlmunlock.  This was to bring
 * them in line with the lock calls.
 */
#ifndef __KERNEL__
dlm_stats_t
dlmunlock_sync(int lockid,
               char *valueblock,
               int flags)
{
	struct transaction request;
	struct lockinfo *vi = &request.clm_lockinfo;
	struct transaction response;
	int ret_val;	/* value to be returned from function */
	int reconfig = 0;

	/* dlm_errno set by Dlm_Init */
	if ((ret_val = Dlm_Init()) != DLM_NORMAL) {
		goto done;
	}

        memset(&request, 0, sizeof(request));

	if (flags & LKM_VALBLK) {
		if (!valueblock) {
			ret_val = set_dlm_errno(DLM_BADARGS);
			goto done;
		} else
			memcpy(vi->lkvalue, valueblock, sizeof(vi->lkvalue)); 
	}
	/* lockid - acts as wildcard when LKM_DEQALL specified */
	if (lockid == 0 && !(flags & LKM_DEQALL)) {
		ret_val = set_dlm_errno(DLM_IVLOCKID);
		goto done;
	}

	request.clm_locktype = DLM_RES_VMS;
	request.clm_respflags = RESP_UDP;
	vi->lk_flags = (flags & ~LKM_TIMEOUT) | lvb_old_behave;
	vi->pid = OurPid;
	vi->lockid = lockid;
retry:
	if (api_reqresp(DLM_UNLOCK, &request, &response)) {
		ret_val = set_dlm_errno(DLM_NOLOCKMGR);
		goto done;
	}

	if (response.clm_status == DLM_WORKING) {
		printf(MSGSTR(DLM_WORKING, "DLM_WORKING_TER returned from API call!!\n"));
		if (api_response(&response) != 0) {
			ret_val = set_dlm_errno(DLM_NOLOCKMGR);
			goto done;
		}
	}

	if (dlm_errno == DLM_NORMAL && 
		response.clm_status == DLM_DENIED_GRACE_PERIOD) {
#if defined(DEBUG)
if (debug) {
	fprintf(stderr, "dlmunlock: grace sleep(%d)\n", GRACE_SLEEP);
}
#endif
		reconfig = 1;
		api_sleep(GRACE_SLEEP);
		goto retry;
	}

	ret_val = set_dlm_errno(response.clm_status);

done:

	return ret_val;
}
#endif /* __KERNEL__ */


/*
 * asynchronous call to remove a lock or cancel a blocked request
 *
 * Linux port, this has been renamed to dlmunlock, and the dlmunlock
 * routine to dlmunlock_sync.  This was to bring them in line with
 * the lock calls.
 */
dlm_stats_t
dlmunlock(int lockid,
          char *valueblock,
          int flags,
          dlm_astunlockfunc_t * unlockast,
          void *unlockastargs,
          void *extrap)
{
	struct transaction request;
	struct lockinfo *vi = &request.clm_lockinfo;
	struct transaction response;
	int ret_val;	/* value to be returned from function */
	int reconfig = 0;

	/* dlm_errno set by Dlm_Init */
	if ((ret_val = Dlm_Init()) != DLM_NORMAL) {
		goto done;
	}

        memset(&request, 0, sizeof(request));

	if (flags & LKM_VALBLK) {
		if (!valueblock) {
			ret_val = set_dlm_errno(DLM_BADARGS);
			goto done;
		} else
			memcpy(vi->lkvalue, valueblock, sizeof(vi->lkvalue)); 
	}
	/* lockid - acts as wildcard when LKM_DEQALL specified */
	if (lockid == 0 && !(flags & LKM_DEQALL)) {
		ret_val = set_dlm_errno(DLM_IVLOCKID);
		goto done;
	}

	request.clm_locktype = DLM_RES_VMS;
	request.clm_signo = notify_signo;
	request.clm_respflags = RESP_UDP | RESP_AST;
	request.clm_asth.asth_funcp = unlockast;
	request.clm_asth.asth_arg = unlockastargs;
	request.clm_basth.asth_arg = extrap;

	vi->lk_flags = flags & ~LKM_TIMEOUT;
	vi->pid = OurPid;
	vi->lockid = lockid;
retry:
	if (api_reqresp(DLM_UNLOCK, &request, &response)) {
		ret_val = set_dlm_errno(DLM_NOLOCKMGR);
		goto done;
	}

	if (dlm_errno == DLM_NORMAL && 
		response.clm_status == DLM_DENIED_GRACE_PERIOD) {
#if defined(DEBUG)
if (debug) {
	fprint_debug("dlmunlock: grace sleep(%d)\n", GRACE_SLEEP);
}
#endif
		reconfig = 1;
		api_sleep(GRACE_SLEEP);
		goto retry;
	}

	ret_val = set_dlm_errno(response.clm_status);

done:

	return ret_val;
}
