/*
 * Sample locking client
 *
 * Copyright (C) 2001, International Business Machines
 *
 * Authors: Peter R. Badovinatz, Ian D. Romanick
 * Contacts: wombat@us.ibm.com, idr@us.ibm.com
 * http://oss.software.ibm.com/developerworks/projects/dlm
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY;  without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program;  if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * http://www.gnu.org/copyleft/gpl.html
 */
#ident "$Id: client2.c,v 1.4 2001/08/31 22:55:34 wombat Exp $"

/* ============================================================ */
/*
 * This program is a very simple locking client.  Idea is to
 * basically just make a pass through a lock request and
 * handle an AST return.
 *
 * Non-interactive client:
 *  - command line input:
 *    - lock name (default RES-A): name of the lock.
 *    - lock mode (default CR): mode of the lock.
 *    - hold time (default 10 secs): after lock is granted, 
 *        how long to wait before unlocking it.
 */
/* ============================================================ */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>

#include <dlm.h>

/* ============================================================ */

struct lockstatus lksb[12];        /* array of lock status blocks */
int which_lock;                    /* index into array of lock status blocks */
dlm_astlockfunc_t ast_func;             /* AST routine */
dlm_bastlockfunc_t bast_func;           /* BAST routine */
void sig_func();                   /* signal handling routine */

dlm_stats_t status;
void * stat;
char * msg;                        /* for printable status code */
extern int transdebug;
char * progname;
int    ASTexecuted;

static int display_quiet = 0;

/* ------------------------------------------------------------ */

static char * lock_mode_tab[] =
{
  "NL", "CR", "CW", "PR", "PW", "EX"
} ;

int desired_lock_mode = LKM_CRMODE;
char desired_lock_name[MAXDLMNAMELEN] = "RES-A";
unsigned int desired_wait_time = 10;

/* ============================================================ */
/*
 * Print usage message.
 */
static void
print_usage(void)
{
  printf("DLM test program %s usage:\n",
         progname);
  printf("%s [-m lock_mode] [-n lock_name] [-w wait_seconds]\n",
         progname);
  printf(" lock_mode: one of NL, CR, CW, PR, PW, EX.\n");
  printf(" lock_name: maximum of 31 characters.\n");
  printf(" wait_seconds: maximum time to wait for AST to be returned,\n");
  printf("   0 (zero) means wait forever if necessary!\n");

  return;
}

/* ------------------------------------------------------------ */
/*
 * Parse command line and set the locking flags to be used.
 *
 * Return
 *  0 - no errors, global values saved as desired.
 *  !0 - errors, printed on stderr
 */
static int
parse_command_line(int argc, char *argv[])
{
  int ret = 0;
  int c;
  int i, good;

  while ((0 == ret) &&
         ((c = getopt(argc,
                      argv,
                      "m:n:w:q")) != EOF)) {
    switch(c) {
    case 'm':
      for (i = 0, good = 0; i <= LKM_MAXMODE; i++) {
        if (0 == strcmp(optarg, lock_mode_tab[i])) {
          desired_lock_mode = i;
          good = 1;
          break;
        }
      }
      if (!good) {
        fprintf( stderr, "[%s] Invalid lock mode given [%s]!\n",
		 progname, optarg );
        print_usage();
        ret = 1;
      }
      break;
    case 'n':
      if (MAXDLMNAMELEN <= strlen(optarg)) {
        fprintf( stderr, "[%s] Given lock name too long!\n", progname );
        print_usage();
        ret = 1;
      } else {
        strcpy(desired_lock_name, optarg);
        desired_lock_name[strlen(optarg)] = '\0';
      }
      break;
    case 'w':
      sscanf(optarg, "%u", &desired_wait_time);
      break;
    case 'q':
      display_quiet = 1;
      break;
    default:
      print_usage();
      ret = 1;
    }
  }

  if ( (0 == ret) && ! display_quiet ) {
    printf("[%s] Given options:\n",
           progname);
    printf("\tMode[%s] Name[%s] Waittime[%d].\n",
           lock_mode_tab[desired_lock_mode],
           desired_lock_name,
           desired_wait_time);
  }

  return(ret);
}

/* ============================================================ */

int
main(int argc, char *argv[])
{
  int astarg = 0;             /* astarg parameter */
  int ret;
  int waiting, ww;
  int unlock_flag;

  progname = argv[0];
  ASTexecuted = 0;

  if (0 != (ret = parse_command_line(argc, argv))) {
    exit(ret);
  }

  transdebug = !display_quiet;


  status = dlm_setnotify( SIGUSR1, NULL );
  if( status != DLM_NORMAL ) {
    dlm_perror("dlm_setnotify");
  }
  stat = signal( SIGUSR1, sig_func );
  if( SIG_ERR == stat ) {
    perror("signal");
  }

  which_lock = 0;
  astarg = which_lock;
  printf("[%s] Call dlmlock mode/name/astarg[%s/%s/%d]!\n",
         progname,
         lock_mode_tab[desired_lock_mode],
         desired_lock_name,
         astarg);
  status = dlmlock( desired_lock_mode,  /* mode */
                    &lksb[which_lock],  /* lock status block */
                    LKM_VALBLK,
                    desired_lock_name,  /* name */
                    (strlen(desired_lock_name)), /* namelen */
                    ast_func,           /* ast routine */
                    &astarg,            /* astargs */
                    bast_func);         /* bast */
  if ( status != DLM_NORMAL ) {
    dlm_perror( "dlmlock");
    exit(status);
  } else {
    printf("[%s] Return DLM_NORMAL from dlmlock, lksb.status = %s!\n",
           progname,
           dlm_errmsg(lksb[which_lock].status));
  }

  waiting = desired_wait_time;
  if ( ! display_quiet )
  {
    printf("[%s] Sleeping %d secs:", progname, waiting );
  }

  for(ww = 0; ww < waiting; ww++) {
    if ( ! display_quiet )
    {
      printf("%d.",ww);
      fflush(stdout);
    }

    sleep(1);
  }

 if ( ! display_quiet )
 {
    printf("\n");
 }

  if (ASTexecuted) {
    printf("[%s] AST block executed!\n",
           progname);
  } else {
    printf("[%s] AST block not executed after %d second%s\n",
           progname,
           desired_wait_time,
           ((1 == desired_wait_time) ? "!" : "s!"));
  }

  if (DLM_NORMAL == lksb[which_lock].status) {
    unlock_flag = LKM_FORCE|LKM_INVVALBLK;
  } else {
    unlock_flag = LKM_CANCEL|LKM_INVVALBLK;
  }
  printf("[%s] Call dlmunlock id/name/flag[%d(%x)/%s/%x]!\n",
         progname,
         lksb[which_lock].lockid,
         lksb[which_lock].lockid,
         desired_lock_name,
         unlock_flag);
  status = dlmunlock_sync( lksb[which_lock].lockid,  /* id of lock */
                           NULL,
                           unlock_flag); /* release the lock */
  if ( status != DLM_NORMAL ) {
    dlm_perror( "dlmunlock");
    exit(status);
  } else {
    printf("[%s] Return DLM_NORMAL from dlmunlock!\n",
           progname);
  }
  waiting = desired_wait_time/2;

  if ( ! display_quiet )
  {
    printf("[%s] Sleeping %d secs:",
         progname,
         waiting);
  }
    
  for(ww = 0; ww < waiting; ww++) {
    if ( ! display_quiet )
    {
      printf("%d.",ww);
      fflush(stdout);
    }

    sleep(1);
  }

  if ( ! display_quiet )
  {
    printf("\n");
  }

  exit(status);
}

/* ------------------------------------------------------------ */
/*
 * Signal handling routine; calls ASTpoll to trigger AST
 * routine.
 */
void sig_func()
{
  ASTpoll( getpid(), 0 );
}

/* ------------------------------------------------------------ */
/*
 * Normal AST routine that is triggered by ASTpoll.
 */
void ast_func(void *astarg)
{
  int *astval;
  astval = (int *)astarg;

  msg = dlm_errmsg( lksb[*astval].status );
  printf("[%s:%s] status/astarg/id[%s/%d/%d(%x)]\n",
         progname,
         __FUNCTION__,
         msg,
         *astval,
         lksb[*astval].lockid,
         lksb[*astval].lockid);
  ASTexecuted = 1;
}

/* ------------------------------------------------------------ */
/*
 * Blocking AST routine that is triggered by ASTpoll.
 */
void bast_func(void *astarg, int mode)
{
  int *astval;
  astval = (int *)astarg;

  msg = dlm_errmsg( lksb[*astval].status );
  printf("[%s:%s] status= %s; astval passed = %d; mode = %s\n",
         progname,
         __FUNCTION__,
         msg,
         *astval,
         lock_mode_tab[mode]);
  ASTexecuted = 1;
}
